/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file triple_cnf.cpp
 * Generates special CNF with 3n variables and all prime implicates having n
 * positive, n negative and n missing literals.
 */

#include <csignal>
#include <iostream>
#include <optional>

#include "assign.h"
#include "logstream.h"
#include "normalform.h"
#include "pccomp.h"
#include "prgutil.h"

const std::string SUBool::kPrgName = "balanced_cnf";
const std::string SUBool::kPrgDesc =
    "Generates special CNF with 3n variables and all prime implicates having n "
    "positive, n negative and n missing literals.";

struct config_t : public SUBool::PrgConfigBase
{
   std::string output_file{};
   unsigned n{2};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   SUBool::logs.DumpValue("\t", level, "n", n);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   SUBool::PrgOptions opts;
   opts.SinglePositional({&conf.output_file, "output"});
   po::options_description desc("Configuration options");
   desc.add_options()("block-size,n",
       po::value<unsigned>(&conf.n)->default_value(conf.n),
       "The number of variables is 3n");
   if (!opts.Parse(argc, argv, conf))
   {
      return {};
   }
   return conf;
}

void
term_func(int signal)
{
   SUBool::logs.TLog(0) << "Terminated by signal " << signal << std::endl;
   _Exit(signal);
}

SUBool::CNF
generate(unsigned n)
{
   SUBool::PartialAssignment alpha(3 * n);
   std::vector<SUBool::Clause> clauses;
   do
   {
      unsigned c_true = 0;
      unsigned c_false = 0;
      unsigned c_undef = 0;
      for (const auto &val : alpha)
      {
         switch (val)
         {
         case SUBool::PartialValue::UNDEF:
            ++c_undef;
            break;
         case SUBool::PartialValue::TRUE:
            ++c_true;
            break;
         case SUBool::PartialValue::FALSE:
            ++c_false;
            break;
         }
      }
      if (c_undef == c_true && c_undef == c_false)
      {
         clauses.emplace_back(alpha);
      }
   } while (alpha.Inc());
   return SUBool::CNF(std::move(clauses));
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   std::signal(SIGTERM, term_func);
   std::signal(SIGQUIT, term_func);
   std::signal(SIGINT, term_func);
   conf->Dump();
   auto cnf = generate(conf->n);
   try
   {
      SUBool::output(conf->output_file, cnf);
   }
   catch (SUBool::IOException &e)
   {
      SUBool::logs.TLog(0) << e.what() << std::endl;
      SUBool::logs.TLog(0) << "No output written" << std::endl;
      return 2;
   }
   return 0;
}
