/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfcanon.cpp
 * Program for constructing the canonical formula for a given CNF. It
 * can use SAT calls to find prime implicates.
 */

#include <fstream>
#include <iostream>
#include <optional>
#include <string>

#include "canon.h"
#include "cnfutil.h"
#include "cpuutil.h"
#include "dimacs.h"
#include "enum_opt.h"
#include "logstream.h"
#include "normalform.h"
#include "prgutil.h"
#include "sbexception.h"

const std::string SUBool::kPrgName = "cnfcanon";
const std::string SUBool::kPrgDesc =
    "constructs the canonical CNF for a given CNF";

enum class PROCESS
{
   NONE,
   PRIME,
   Last = PRIME
};

const SUBool::EnumAnnotation<PROCESS> kProcessAnnotation("PROCESS",
    {{PROCESS::NONE, "none", "no inprocessing."},
        {PROCESS::PRIME, "prime", "use SAT to find a prime subimplicate."}});

const char *kProcessOptDesc = "Clause inprocessing and preprocessing level.\n"
                              "none = No inprocessing,\n"
                              "prime = use SAT to find a prime subimplicate.";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   PROCESS process{PROCESS::PRIME};

   virtual void InnerDump(unsigned level) const override;
};

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   auto opt_process = SUBool::make_enum_option(kProcessAnnotation, "process,p",
       "Clause inprocessing and preprocessing level.", conf.process);
   std::string process;
   SUBool::PrgOptions opts;
   opts.InputOutput(&conf.input_file, &conf.output_file);
   po::options_description desc("Configuration options");
   desc.add(opt_process);
   opts.Add(desc);
   if (!opts.Parse(argc, argv, conf) || !opt_process.GetValue(conf.process))
   {
      return {};
   }
   return conf;
}

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   kProcessAnnotation.DumpValue("\t", level, "process", process);
}

std::unique_ptr<SUBool::AbsCanon>
canon_procedure(const config_t &conf, const SUBool::CNF &cnf)
{
   switch (conf.process)
   {
   case PROCESS::NONE:
      return std::unique_ptr<SUBool::AbsCanon>(new SUBool::CanonBase(cnf));
   case PROCESS::PRIME:
      return std::unique_ptr<SUBool::AbsCanon>(new SUBool::CanonSAT(cnf));
   }
   throw SUBool::BadParameterException(
       "canon_procedure", "Bad method specification");
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   conf->Dump();
   auto cnf = SUBool::input_or_die<SUBool::CNF>(
       conf->input_file, 0, 2, "Failed to parse the input file");
   auto canon = canon_procedure(*conf, cnf);
   SUBool::CPUTime cpu_time;
   canon->Compute();
   canon->PrintStatistics(2);
   SUBool::output(conf->output_file, canon->Cnf());
   return 0;
}
