/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfcomp.cpp
 * Program for comparing two functions represented by CNFs.
 */

#include "cnfcompare.h"
#include "cnfutil.h"
#include "cpuutil.h"
#include "dimacs.h"
#include "enum_opt.h"
#include "logstream.h"
#include "normalform.h"
#include "prgutil.h"

const std::string SUBool::kPrgName = "cnfcomp";
const std::string SUBool::kPrgDesc =
    "compare two functions represented by CNF formulas";

enum class RELATION
{
   EQUAL,
   LEQ,
   GEQ,
   Last = GEQ
};

const SUBool::EnumAnnotation<RELATION> kRelationAnnotation(
    "RELATION", {{RELATION::EQUAL, "eq", "equality"},
                    {RELATION::LEQ, "le", "less or equal"},
                    {RELATION::GEQ, "ge", "greater or equal"}});

enum class COMPARE_KIND
{
   IMPLICATE,
   ONE_PROVABLE,
   ABSORBED,
   Last = ABSORBED
};

const SUBool::EnumAnnotation<COMPARE_KIND> kCompareKindAnnotation(
    "COMPARE_KIND",
    {{COMPARE_KIND::IMPLICATE, "implicate",
         "check if the clauses of one CNF are implicates of the "
         "other (semantical comparison)"},
        {COMPARE_KIND::ONE_PROVABLE, "one_provable",
            "check if the clauses of one CNF are 1-provable in the other"},
        {COMPARE_KIND::ABSORBED, "absorbed",
            "check if the clauses in one CNF are absorbed by the other (not "
            "empowering)"}});

struct config_t : public SUBool::PrgConfigBase
{
   std::string file_left{};
   std::string file_right{};
   RELATION relation{RELATION::EQUAL};
   COMPARE_KIND compare_kind{COMPARE_KIND::IMPLICATE};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "file_left", file_left);
   SUBool::logs.DumpValue("\t", level, "file_right", file_right);
   kRelationAnnotation.DumpValue("\t", level, "relation", relation);
   kCompareKindAnnotation.DumpValue("\t", level, "compare_kind", compare_kind);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   auto opt_relation = SUBool::make_enum_option(
       kRelationAnnotation, "relation,r", "Relation to check.", conf.relation);
   auto opt_compare_kind = SUBool::make_enum_option(kCompareKindAnnotation,
       "compare-kind,k", "How the CNFs should be compared.", conf.compare_kind);

   // options.positional_help("[input-left [input-right]]");
   SUBool::PrgOptions opts;
   opts.InputPair(
       {&conf.file_left, "input-left"}, {&conf.file_right, "input-right"});
   po::options_description desc("Configuration options");
   desc.add(opt_relation);
   desc.add(opt_compare_kind);
   opts.Add(desc);
   if (!opts.Parse(argc, argv, conf) || !opt_relation.GetValue(conf.relation)
       || !opt_compare_kind.GetValue(conf.compare_kind))
   {
      return {};
   }
   return conf;
}

std::string
eq_to_string(config_t &conf)
{
   std::stringstream ss;
   switch (conf.compare_kind)
   {
   case COMPARE_KIND::IMPLICATE:
      ss << "The formulas are equal.";
      break;
   case COMPARE_KIND::ONE_PROVABLE:
      ss << "The formulas are equal with respect to 1-provability.";
      break;
   case COMPARE_KIND::ABSORBED:
      ss << "Each formula absorbs the other one.";
      break;
   default:
      throw SUBool::BadParameterException(
          "eq_to_string", "unknown compare kind");
   }
   return ss.str();
}

std::string
implies_to_string(
    config_t &conf, const std::string &l, const std::string &r, bool strict)
{
   std::stringstream ss;
   switch (conf.compare_kind)
   {
   case COMPARE_KIND::IMPLICATE:
      if (strict)
      {
         ss << "The formulas are not equal" << std::endl;
      }
      ss << "The " << l << " implies the " << r << ".";
      break;
   case COMPARE_KIND::ONE_PROVABLE:
      if (strict)
      {
         ss << "The formulas are not equal with respect to 1-provability"
            << std::endl;
      }
      ss << "Clauses of the " << r << " are 1-provable in the " << l << ".";
      break;
   case COMPARE_KIND::ABSORBED:
      if (strict)
      {
         ss << "The formulas are not equal with respect to empowerement"
            << std::endl;
      }
      ss << "Clauses of the " << r << " are absorbed (not empowering) by the "
         << l << ".";
      break;
   default:
      throw SUBool::BadParameterException(
          "eq_to_string", "unknown compare kind");
   }
   return ss.str();
}

std::string
not_implies_to_string(
    config_t &conf, const std::string &l, const std::string &r)
{
   std::stringstream ss;
   switch (conf.compare_kind)
   {
   case COMPARE_KIND::IMPLICATE:
      ss << "The " << l << " does not imply the " << r << ".";
      break;
   case COMPARE_KIND::ONE_PROVABLE:
      ss << "Clauses of the " << r << " are not all 1-provable in the " << l
         << ".";
      break;
   case COMPARE_KIND::ABSORBED:
      ss << "Clauses of the " << r
         << " are not all absorbed (some are empowering) by the " << l << ".";
      break;
   default:
      throw SUBool::BadParameterException(
          "eq_to_string", "unknown compare kind");
   }
   return ss.str();
}

bool
check_equality(
    config_t &conf, const SUBool::CNF &left_cnf, const SUBool::CNF &right_cnf)
{
   SUBool::CompareResult result = SUBool::CompareResult::INCOMPARABLE;
   switch (conf.compare_kind)
   {
   case COMPARE_KIND::IMPLICATE:
      result = SUBool::cnf_compare(left_cnf, right_cnf);
      break;
   case COMPARE_KIND::ONE_PROVABLE:
      result = SUBool::cnf_compare_up(left_cnf, right_cnf);
      break;
   case COMPARE_KIND::ABSORBED:
      result = SUBool::cnf_compare_absorb(left_cnf, right_cnf);
      break;
   default:
      throw SUBool::BadParameterException(
          "check_equality", "unknown compare kind");
   }
   switch (result)
   {
   case SUBool::CompareResult::INCOMPARABLE:
      SUBool::logs.TLog(1) << "(L<>R)" << std::endl;
      SUBool::logs.TLog(2) << "The formulas are incomparable" << std::endl;
      break;
   case SUBool::CompareResult::LESS:
      SUBool::logs.TLog(1) << "(L<R)" << std::endl;
      SUBool::logs.TLog(2) << implies_to_string(
          conf, "first formula (L)", "second one (R)", true)
                           << std::endl;
      break;
   case SUBool::CompareResult::GREATER:
      SUBool::logs.TLog(1) << "(L>R)" << std::endl;
      SUBool::logs.TLog(2) << implies_to_string(
          conf, "second formula (R)", "first one (L)", true)
                           << std::endl;
      break;
   case SUBool::CompareResult::EQUAL:
      SUBool::logs.TLog(1) << "(L=R)" << std::endl;
      SUBool::logs.TLog(2) << eq_to_string(conf) << std::endl;
   }
   return (result == SUBool::CompareResult::EQUAL);
}

bool
check_implication(config_t &conf, const SUBool::CNF &left_cnf,
    const SUBool::CNF &right_cnf, bool swapped)
{
   bool result = false;
   switch (conf.compare_kind)
   {
   case COMPARE_KIND::IMPLICATE:
      result = SUBool::cnf_implies(left_cnf, right_cnf);
      break;
   case COMPARE_KIND::ONE_PROVABLE:
      result = SUBool::cnf_implies_up(left_cnf, right_cnf);
      break;
   case COMPARE_KIND::ABSORBED:
      result = SUBool::cnf_absorbed(left_cnf, right_cnf);
      break;
   default:
      throw SUBool::BadParameterException(
          "check_equality", "unknown compare kind");
   }
   std::string first, second, l, r;
   if (swapped)
   {
      second = "first formula (L)";
      first = "second formula (R)";
      r = "L";
      l = "R";
   }
   else
   {
      first = "first formula (L)";
      second = "second formula (R)";
      l = "L";
      r = "R";
   }
   if (result)
   {
      SUBool::logs.TLog(1) << "(" << l << "<=" << r << ")" << std::endl;
      SUBool::logs.TLog(2) << implies_to_string(conf, first, second, false)
                           << std::endl;
   }
   else
   {
      SUBool::logs.TLog(1) << "NOT (" << l << "<=" << r << ")" << std::endl;
      SUBool::logs.TLog(2) << not_implies_to_string(conf, first, second)
                           << std::endl;
   }
   return result;
}

bool
check(config_t &conf, const SUBool::CNF &left_cnf, const SUBool::CNF &right_cnf)
{
   switch (conf.relation)
   {
   case RELATION::EQUAL:
      return check_equality(conf, left_cnf, right_cnf);
   case RELATION::LEQ:
      return check_implication(conf, left_cnf, right_cnf, false);
   case RELATION::GEQ:
      return check_implication(conf, right_cnf, left_cnf, true);
   default:
      throw SUBool::InvariantFailedException("compare", "Unknown check kind");
   }
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   conf->Dump();
   auto left_cnf = SUBool::input_or_die<SUBool::CNF>(conf->file_left);
   auto right_cnf = SUBool::input_or_die<SUBool::CNF>(conf->file_right);
   SUBool::CPUTime cpu_time;
   auto ret = !check(*conf, left_cnf, right_cnf);
   SUBool::logs.TLog(2) << "Time " << cpu_time << std::endl;
   return ret;
}
