/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/*@file cnfdiff.cpp
 * Program for comparing two CNFs as sets of clauses (not as representations of
 * functions).
 */
#include "dimacs.h"
#include "normalform.h"
#include "prgutil.h"

const std::string SUBool::kPrgName = "cnfdiff";
const std::string SUBool::kPrgDesc =
    "Program for comparing two CNFs as sets of "
    "clauses (not as representations of * functions).";

struct config_t : public SUBool::PrgConfigBase
{
   std::string file_left{};
   std::string file_right{};
   bool suppress_first{false};
   bool suppress_second{false};
   bool suppress_third{false};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "file_left", file_left);
   SUBool::logs.DumpValue("\t", level, "file_right", file_right);
   SUBool::logs.DumpValue("\t", level, "suppress_first", suppress_first);
   SUBool::logs.DumpValue("\t", level, "suppress_second", suppress_second);
   SUBool::logs.DumpValue("\t", level, "suppress_third", suppress_third);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   SUBool::PrgOptions opts;
   opts.InputPair(
       {&conf.file_left, "input-left"}, {&conf.file_right, "input-right"});
   po::options_description desc("Configuration options");
   desc.add_options()(",1", po::bool_switch(&conf.suppress_first),
       "Suppress printing the clauses unique to the first CNF.")(",2",
       po::bool_switch(&conf.suppress_second),
       "Suppress printing the clauses unique to the second CNF.")(",3",
       po::bool_switch(&conf.suppress_third),
       "Suppress printing the clauses unique to the third CNF.");
   opts.Add(desc);
   if (!opts.Parse(argc, argv, conf))
   {
      return {};
   }
   return conf;
}

void
compare(SUBool::CNF left, SUBool::CNF right, config_t &conf)
{
   left.Sort();
   right.Sort();
   auto i_left = left.cbegin();
   auto i_right = right.cbegin();
   SUBool::AnnotatedCNF only_first;
   SUBool::AnnotatedCNF only_second;
   SUBool::AnnotatedCNF both;
   while (i_left != left.cend() && i_right != right.cend())
   {
      if (*i_left < *i_right)
      {
         only_first.nf.PushBack(*i_left);
         ++i_left;
         continue;
      }
      if (*i_right < *i_left)
      {
         only_second.nf.PushBack(*i_right);
         ++i_right;
         continue;
      }
      both.nf.PushBack(*i_right);
      ++i_left;
      ++i_right;
   }
   while (i_left != left.end())
   {
      only_first.nf.PushBack(*i_left);
      ++i_left;
   }
   while (i_right != right.end())
   {
      only_second.nf.PushBack(*i_right);
      ++i_right;
   }
   if (!conf.suppress_first)
   {
      only_first.dimacs_annotation = "c Clauses unique to " + conf.file_left;
      std::cout << only_first << std::endl;
   }
   if (!conf.suppress_second)
   {
      only_second.dimacs_annotation = "c Clauses unique to " + conf.file_right;
      std::cout << only_second << std::endl;
   }
   if (!conf.suppress_third)
   {
      both.dimacs_annotation = "c Clauses in both files";
      std::cout << both << std::endl;
   }
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   conf->Dump();
   compare(SUBool::input_or_die<SUBool::CNF>(conf->file_left),
       SUBool::input_or_die<SUBool::CNF>(conf->file_right), *conf);
   return 0;
}
