/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfgraphx.cpp
 * Program for exporting a CNF as a graph.
 */

#include <iostream>

#include "cnfgraph.h"
#include "dimacs.h"
#include "enum_opt.h"
#include "logstream.h"
#include "prgutil.h"

enum class OUTPUT_TYPE
{
   GRAPHML,
   DOT,
   Last = DOT
};

const SUBool::EnumAnnotation<OUTPUT_TYPE> kOutputTypeAnnotation(
    "OUTPUT_TYPE", {{OUTPUT_TYPE::GRAPHML, "graphml", "GraphML"},
                       {OUTPUT_TYPE::DOT, "dot", "dot format"}});

const std::string SUBool::kPrgName = "cnfgraph";
const std::string SUBool::kPrgDesc = "export a CNF as a graph";
const std::string kDefaultIncidenceGraphVarColor = "red";
const std::string kDefaultIncidenceGraphClauseColor = "blue";

enum class GRAPH_KIND
{
   PRIMAL,
   DUAL,
   INCIDENCE,
   DINCIDENCE,
   Last = DINCIDENCE
};

const SUBool::EnumAnnotation<GRAPH_KIND> kGraphKindAnnotation("GRAPH_KIND",
    {{GRAPH_KIND::PRIMAL, "primal", "primal graph"},
        {GRAPH_KIND::DUAL, "dual", "dual graph"},
        {GRAPH_KIND::INCIDENCE, "incidence", "incidence graph"},
        {GRAPH_KIND::DINCIDENCE, "dincidence", "directed incidence graph"}});

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   GRAPH_KIND graph_kind{GRAPH_KIND::PRIMAL};
   OUTPUT_TYPE output_type{OUTPUT_TYPE::GRAPHML};
   std::string var_prefix{"x"};
   std::string var_color{};
   std::string clause_prefix{"c"};
   std::string clause_color{};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   kGraphKindAnnotation.DumpValue("\t", level, "graph_kind", graph_kind);
   kOutputTypeAnnotation.DumpValue("\t", level, "output_type", output_type);
   SUBool::logs.DumpValue("\t", level, "var_prefix", var_prefix);
   SUBool::logs.DumpValue("\t", level, "var_color", var_color);
   SUBool::logs.DumpValue("\t", level, "clause_prefix", clause_prefix);
   SUBool::logs.DumpValue("\t", level, "clause_color", clause_color);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   auto opt_output_type = SUBool::make_enum_option(kOutputTypeAnnotation,
       "output-type,t", "Type of the output.", conf.output_type);
   auto opt_graph_kind = SUBool::make_enum_option(kGraphKindAnnotation,
       "graph-kind,g", "Kind of the graph to be exported.", conf.graph_kind);
   boost::optional<std::string> var_color;
   boost::optional<std::string> clause_color;

   SUBool::PrgOptions opts(true);
   opts.InputOutput(&conf.input_file, &conf.output_file);
   po::options_description desc("Configuration options");
   desc.add(opt_output_type);
   desc.add(opt_graph_kind);
   desc.add_options()("var-prefix",
       po::value<std::string>(&conf.var_prefix)->default_value(conf.var_prefix),
       "Prefix of a name of a vertex corresponding to a variable.")(
       "clause-prefix",
       po::value<std::string>(&conf.clause_prefix)
           ->default_value(conf.clause_prefix),
       "Prefix of a name of a vertex corresponding to a clause.")("var-color",
       po::value(&var_color),
       ("Color of a vertex corresponding to a variable. "
        "The default depends on the graph kind. In case of primal and "
        "dual graph, it is no color, in case of an incidence graph, it is "
           + kDefaultIncidenceGraphVarColor + ". Empty string means no color.")
           .c_str())("clause-color", po::value(&clause_color),
       ("Color of a vertex corresponding to a clause. "
        "The default depends on the graph kind. In case of primal and "
        "dual graph, it is no color, in case of an incidence graph, it is "
           + kDefaultIncidenceGraphClauseColor
           + ". Empty string means no color.")
           .c_str());
   opts.Add(desc);
   if (!opts.Parse(argc, argv, conf)
       || !opt_output_type.GetValue(conf.output_type)
       || !opt_graph_kind.GetValue(conf.graph_kind))
   {
      return {};
   }
   if (var_color)
   {
      conf.var_color = *var_color;
   }
   if (clause_color)
   {
      conf.clause_color = *clause_color;
   }
   if (conf.graph_kind == GRAPH_KIND::INCIDENCE
       || conf.graph_kind == GRAPH_KIND::DINCIDENCE)
   {
      if (!var_color)
      {
         conf.var_color = kDefaultIncidenceGraphVarColor;
      }
      if (!clause_color)
      {
         conf.clause_color = kDefaultIncidenceGraphClauseColor;
      }
   }
   return conf;
}

template <typename T, typename... Args>
int
write_graph(
    std::ostream &os, OUTPUT_TYPE output_type, T &g, const Args &...args)
{
   switch (output_type)
   {
   case OUTPUT_TYPE::GRAPHML:
      g.WriteGraphML(os, args...);
      break;
   case OUTPUT_TYPE::DOT:
      g.WriteGraphViz(os, args...);
      break;
   default:
      throw SUBool::InvariantFailedException(
          "write_graph", "Unknown output type");
   }
   return 0;
}

int
do_graph(const config_t &conf, const SUBool::CNF &cnf)
{
   SUBool::OutputStream os(conf.output_file);
   if (!os)
   {
      SUBool::logs.TLog(0) << "Failed to open file " << conf.output_file
                           << " for writing." << std::endl;
      return 2;
   }
   switch (conf.graph_kind)
   {
   case GRAPH_KIND::PRIMAL:
   {
      SUBool::PrimalGraph g(cnf);
      return write_graph(
          os, conf.output_type, g, conf.var_prefix, conf.var_color);
   }
   case GRAPH_KIND::DUAL:
   {
      SUBool::DualGraph g(cnf);
      return write_graph(
          os, conf.output_type, g, conf.clause_prefix, conf.clause_color);
   }
   case GRAPH_KIND::INCIDENCE:
   {
      SUBool::IncidenceGraph g(cnf);
      return write_graph(os, conf.output_type, g, conf.var_prefix,
          conf.var_color, conf.clause_prefix, conf.clause_color);
   }
   case GRAPH_KIND::DINCIDENCE:
   {
      SUBool::DIncidenceGraph g(cnf);
      return write_graph(os, conf.output_type, g, conf.var_prefix,
          conf.var_color, conf.clause_prefix, conf.clause_color);
   }
   default:
      throw SUBool::InvariantFailedException("do_graph", "Unknown graph kind");
   }
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   conf->Dump();
   auto cnf = SUBool::input_or_die<SUBool::CNF>(conf->input_file);
   return do_graph(*conf, cnf);
}
