/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfprime.cpp
 * Program for constructing a prime CNF for a given CNF.
 */

#include <fstream>
#include <iostream>
#include <string>

#include <boost/optional.hpp>

#include "cnfutil.h"
#include "cpuutil.h"
#include "dimacs.h"
#include "empower.h"
#include "logstream.h"
#include "normalform.h"
#include "prgutil.h"
#include "sbexception.h"

const std::string SUBool::kPrgName = "cnfprime";
const std::string SUBool::kPrgDesc = "Making a CNF prime";
const std::string prg_help =
    "Makes a given CNF prime, possibly only with respect to 1-provability.";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   bool unitprop{false};

   virtual void InnerDump(unsigned level) const override;
};

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   SUBool::PrgOptions opts;
   opts.SetHelpDesc(prg_help);
   opts.InputOutput(&conf.input_file, &conf.output_file);
   po::options_description desc("Configuration options");
   desc.add_options()("unitprop,u", po::bool_switch(&conf.unitprop),
       "Use only unit propagation instead of SAT. It means "
       "that the result is not necessarily prime, but it is prime with "
       "respect to 1-provability (although not fully as only one round of "
       "prime subimplicate finding is performed");
   opts.Add(desc);
   if (!opts.Parse(argc, argv, conf))
   {
      return {};
   }
   return conf;
}

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   SUBool::logs.DumpValue("\t", level, "unitprop", unitprop);
}

SUBool::CNF
process(const config_t &conf, SUBool::CNF input)
{
   if (conf.unitprop)
   {
      return SUBool::up_prime_cnf(std::move(input));
   }
   else
   {
      return SUBool::prime_cnf(input);
   }
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   conf->Dump();
   SUBool::logs.TLog(2) << "Reading input" << std::endl;
   auto in_cnf = SUBool::input<SUBool::CNF>(conf->input_file);
   SUBool::logs.TLog(1) << "Starting computation" << std::endl;
   SUBool::CPUTime cpu_time;
   auto out_cnf = process(*conf, in_cnf);
   SUBool::logs.TLog(1) << "Computation finished, proc time: " << cpu_time
                        << std::endl;
   SUBool::output(conf->output_file, out_cnf);
   SUBool::logs.Log(2) << std::endl;
   SUBool::logs.TLog(2) << "Output written" << std::endl;
}
