/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfencode.cpp
 * Program for constructing URC and PC encodings of a given (smooth) DNNF.
 */

#include <csignal>
#include <iostream>

#include "dimacs.h"
#include "dpelim_opt.h"
#include "encoding.h"
#include "enum_opt.h"
#include "logstream.h"
#include "prgutil.h"

const std::string SUBool::kPrgName = "dpelim";
const std::string SUBool::kPrgDesc = "DP elimination of variables";
const std::string prg_help =
    "DP elimination of variables. By default it aims to eliminate all "
    "variables, use -e, -f, and -s options to limit the extent of elimination. "
    "You can also limit the maximum input variables (up to which no variable "
    "should be eliminated) with -x option. Note that the input is considered "
    "to be an encoding. Thus the variables which are allowed to be eliminated "
    "may have different indices in the output (they are considered auxiliary "
    "in that their names do not matter).";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   unsigned max_input_var{0};
   bool elim_minimize{false};
   bool final_minimize{false};
   SUBool::CNFEncoding::BLOCK_PRIME block_prime{
       SUBool::CNFEncoding::BLOCK_PRIME::NONE};
   SUBool::DPElimEliminationOptions::ELIMINATE_POLICY eliminate{
       SUBool::DPElimEliminationOptions::kDefaultEliminatePolicy};
   SUBool::DPEliminator::IGNORE_POLICY ignore_policy{
       SUBool::DPEliminator::IGNORE_POLICY::ABSORBED};
   SUBool::DPEliminator::INPROCESSING inprocessing{
       SUBool::DPEliminator::INPROCESSING::NONE};
   SUBool::DPEliminator::INCREASE_POLICY increase_policy{
       SUBool::DPEliminator::INCREASE_POLICY::AFTER_BOUND};
   unsigned minimize_size_step{5000};
   double minimize_size_factor{0.6};
   unsigned max_increase{SUBool::DPElimEliminationOptions::kDefaultMaxIncrease};
   long double factor{SUBool::DPElimEliminationOptions::kDefaultFactor};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   SUBool::logs.DumpValue("\t", level, "max_input_var", max_input_var);
   SUBool::logs.DumpValue("\t", level, "elim_minimize", elim_minimize);
   SUBool::CNFEncoding::kBlockPrimeAnnotation.DumpValue(
       "\t", level, "block_prime", block_prime);
   SUBool::DPElimEliminationOptions::kEliminatePolicyAnnotation.DumpValue(
       "\t", level, "eliminate", eliminate);
   SUBool::DPEliminator::kIgnorePolicyAnnotation.DumpValue(
       "\t", level, "ignore_policy", ignore_policy);
   SUBool::DPEliminator::kInprocessingAnnotation.DumpValue(
       "\t", level, "inprocessing", inprocessing);
   SUBool::DPEliminator::kIncreasePolicyAnnotation.DumpValue(
       "\t", level, "increase_policy", increase_policy);
   SUBool::logs.DumpValue(
       "\t", level, "minimize_size_step", minimize_size_step);
   SUBool::logs.DumpValue(
       "\t", level, "minimize_size_factor", minimize_size_factor);
   SUBool::logs.DumpValue("\t", level, "max_increase", max_increase);
   SUBool::logs.DumpValue("\t", level, "factor", factor);
   SUBool::logs.DumpValue("\t", level, "final_minimize", final_minimize);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;

   SUBool::PrgOptions opts(true);
   opts.InputOutput(&conf.input_file, &conf.output_file);
   opts.SetHelpDesc(prg_help);

   auto elim_opts = opts.MakeOptionsGroup<SUBool::DPElimEliminationOptions>();
   auto clause_manip_opts =
       opts.MakeOptionsGroup<SUBool::DPElimClauseManipOptions>();
   auto minimize_opts = opts.MakeOptionsGroup<SUBool::DPElimMinimizeOptions>();

   if (!opts.Parse(argc, argv, conf))
   {
      return {};
   }

   conf.max_input_var = elim_opts->MaxInputVar();
   conf.eliminate = elim_opts->EliminatePolicy();
   conf.increase_policy = elim_opts->IncreasePolicy();
   conf.max_increase = elim_opts->MaxIncrease();
   conf.factor = elim_opts->Factor();

   conf.block_prime = clause_manip_opts->BlockPrime();
   conf.inprocessing = clause_manip_opts->Inprocessing();
   conf.ignore_policy = clause_manip_opts->IgnorePolicy();

   conf.elim_minimize = minimize_opts->ElimMinimize();
   conf.final_minimize = minimize_opts->FinalMinimize();
   conf.minimize_size_step = minimize_opts->SizeStep();
   conf.minimize_size_factor = minimize_opts->SizeFactor();

   return conf;
}

void
eliminate(SUBool::CNFEncoding &enc, const config_t &conf)
{
   enc.SetEliminationPolicy(conf.ignore_policy, conf.inprocessing,
       conf.increase_policy, conf.block_prime);
   switch (conf.eliminate)
   {
   case SUBool::DPElimEliminationOptions::ELIMINATE_POLICY::ALL:
      if (conf.elim_minimize)
      {
         enc.EliminateAuxVariablesWithMaxIncMinimize(
             SUBool::CNFEncoding::kEliminateAll, conf.minimize_size_step,
             conf.minimize_size_factor);
      }
      else
      {
         enc.EliminateAuxVariablesWithMaxInc(
             SUBool::CNFEncoding::kEliminateAll);
      }
      break;
   case SUBool::DPElimEliminationOptions::ELIMINATE_POLICY::MAX_INCREASE:
      if (conf.elim_minimize)
      {
         enc.EliminateAuxVariablesWithMaxIncMinimize(conf.max_increase,
             conf.minimize_size_step, conf.minimize_size_factor);
      }
      else
      {
         enc.EliminateAuxVariablesWithMaxInc(conf.max_increase);
      }
      break;
   case SUBool::DPElimEliminationOptions::ELIMINATE_POLICY::FACTOR:
      if (conf.elim_minimize)
      {
         enc.EliminateAuxVariablesWithFactorMinimize(
             conf.factor, conf.minimize_size_step, conf.minimize_size_factor);
      }
      else
      {
         enc.EliminateAuxVariablesWithFactor(conf.factor);
      }
      break;
   }
}

void
term_func(int signal)
{
   SUBool::logs.TLog(0) << "Terminated by signal " << signal << std::endl;
   exit(signal);
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   std::signal(SIGTERM, term_func);
   std::signal(SIGQUIT, term_func);
   std::signal(SIGINT, term_func);
   conf->Dump();
   auto cnf = SUBool::input_or_die<SUBool::CNF>(conf->input_file);
   SUBool::PCEncoding enc(cnf, conf->max_input_var, false);
   eliminate(enc, *conf);
   if (conf->final_minimize && !conf->elim_minimize)
   {
      enc.Minimize();
   }
   SUBool::output(conf->output_file, enc);
   return 0;
}
