/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/*@file gdhorn.cpp
 * Program computing the GD basis of a given definite Horn CNF
 */

#include "cpuutil.h"
#include "dimacs.h"
#include "drenc.h"
#include "enum_opt.h"
#include "horn.h"
#include "prgutil.h"

const std::string SUBool::kPrgName = "gdbasis";
const std::string SUBool::kPrgDesc =
    "Program computing the GD basis of a given definite Horn CNF";

enum class IMPL_TYPE
{
   HORN,
   DUAL_RAIL,
   Last = DUAL_RAIL
};

const SUBool::EnumAnnotation<IMPL_TYPE> kImplTypeAnnotation("IMPL_TYPE",
    {//
        {IMPL_TYPE::HORN, "horn",
            "the input is a definite Horn CNF describing an implicational "
            "system whose basis should be computed"}, //
        {IMPL_TYPE::DUAL_RAIL, "dual_rail",
            "the input is a general CNF. The basis should be constructed for "
            "the dual rail encoding of the input CNF."}});
const IMPL_TYPE kDefaultImplType = IMPL_TYPE::DUAL_RAIL;

enum class OUTPUT_TYPE
{
   IMPLICATIONS,
   CNF,
   Last = CNF
};

const SUBool::EnumAnnotation<OUTPUT_TYPE> kOutputTypeAnnotation("OUTPUT_TYPE",
    {//
        {OUTPUT_TYPE::IMPLICATIONS, "implications",
            "output consists of a list of implications"},
        {OUTPUT_TYPE::CNF, "cnf",
            "output is a CNF after projecting the implications back to "
            "clauses"}});

const OUTPUT_TYPE kDefaultOutputType = OUTPUT_TYPE::IMPLICATIONS;

const std::string kDefaultSeparator = ":";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   IMPL_TYPE impl_type{kDefaultImplType};
   OUTPUT_TYPE output_type{kDefaultOutputType};
   std::string impl_sep{kDefaultSeparator};
   bool keep_body_in_closure{false};
   // Closure of the empty body should be part of
   // both the body and cons at the and
   bool omit_empty_closure{false};
   // Compute only body minimal representation
   bool body_minimal{false};
   // At the end, make it right irredundant
   bool right_irredundant{false};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   kImplTypeAnnotation.DumpValue("\t", level, "impl_type", impl_type);
   kOutputTypeAnnotation.DumpValue("\t", level, "output_type", output_type);
   SUBool::logs.DumpValue("\t", level, "impl_sep", impl_sep);
   SUBool::logs.DumpValue(
       "\t", level, "keep_body_in_closure", keep_body_in_closure);
   SUBool::logs.DumpValue(
       "\t", level, "omit_empty_closure", omit_empty_closure);
   SUBool::logs.DumpValue("\t", level, "body_minimal", body_minimal);
   SUBool::logs.DumpValue("\t", level, "right_irredundant", right_irredundant);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   auto opt_impl_type = SUBool::make_enum_option(kImplTypeAnnotation,
       "impl-type,I",
       "Specifies how the input implications are derived from the input CNF.",
       kDefaultImplType);
   auto opt_output_type = SUBool::make_enum_option(kOutputTypeAnnotation,
       "output-type,O", "Type of the output.", kDefaultOutputType);
   SUBool::PrgOptions opts(true);
   opts.InputOutput(&conf.input_file, &conf.output_file);
   po::options_description input_opts("Input system");
   input_opts.add(opt_impl_type);
   opts.Add(input_opts);
   po::options_description output_opts("Output system");
   output_opts.add(opt_output_type);
   output_opts.add_options() //
       ("impl-sep,s",
           po::value<std::string>(&conf.impl_sep)
               ->default_value(kDefaultSeparator),
           "String which separates the consequence part from the body in the "
           "implications output type") //
       ("keep-body-in-closure,k", po::bool_switch(&conf.keep_body_in_closure),
           "if present, the body will be part of the closure in the output "
           "basis, otherwise, body is subtracted from the closure") //
       ("omit-empty-closure,e", po::bool_switch(&conf.omit_empty_closure),
           "if present, the closure of the empty body will be omitted from the "
           "bodies and consequences in the implications. Otherwise, the "
           "closure of the empty body is part of every set of consequences and "
           "every body (in case of left saturation).") //
       ("body-minimal,m", po::bool_switch(&conf.body_minimal),
           "compute only body minimal representation, i.e. do not use "
           "left-saturation, options keep-body-in-closure and "
           "omit-empty-closure are ignored in this case (keep-body-in-closure "
           "is set, omit-empty-closure is unset)") //
       ("right-irredundant,r", po::bool_switch(&conf.right_irredundant),
           "make the implication system right irredundant at the end, options "
           "keep-body-in-closure and omit-empty-closure are ignored in this "
           "case (keep-body-in-closure is set, omit-empty-closure is unset), "
           "however left saturation is performed if not suppressed by option "
           "body-minimal.");
   opts.Add(output_opts);
   if (!opts.Parse(argc, argv, conf) || !opt_impl_type.GetValue(conf.impl_type)
       || !opt_output_type.GetValue(conf.output_type))
   {
      return {};
   }
   if (conf.right_irredundant || conf.body_minimal)
   {
      conf.omit_empty_closure = true;
      conf.keep_body_in_closure = false;
   }
   return conf;
}

template <class Var>
void
compute_body_minimal(SUBool::ImplSystem<Var> &is, const config_t &conf)
{
   is.MakeBodyMinimal();
   if (conf.right_irredundant)
   {
      is.MakeRightIrredundant();
      // TODO: somehow the empty bodies are missing
   }
}

template <class Var>
void
compute_gd(SUBool::ImplSystem<Var> &is, const config_t &conf)
{
   is.MakeGDBasis();
   if (conf.right_irredundant)
   {
      is.MakeRightIrredundant();
      return;
   }
   if (!conf.omit_empty_closure)
   {
      SUBool::logs.TLog(2) << "Adding back the closure of the empty body"
                           << std::endl;
      is.AddEmptyClosureToImpl(true, false);
   }
   if (conf.keep_body_in_closure
       && conf.output_type == OUTPUT_TYPE::IMPLICATIONS)
   {
      SUBool::logs.TLog(2) << "Adding bodies to the consequences" << std::endl;
      is.AddBodyToCons();
   }
}

template <class Var, class ToCNF>
int
print_is(const SUBool::ImplSystem<Var> &is, unsigned max_variable,
    const config_t &conf, ToCNF to_cnf)
{
   auto var_printer = [](const auto &v) { return v; };
   switch (conf.output_type)
   {
   case OUTPUT_TYPE::IMPLICATIONS:
      SUBool::output(conf.output_file,
          SUBool::make_annotated_is(
              // SUBool::AnnotatedImplSystem<Var, decltype(var_printer)>(
              is, var_printer, max_variable, "", ":"));
      break;
   case OUTPUT_TYPE::CNF:
      SUBool::output(conf.output_file, to_cnf(is));
      break;
   default:
      SUBool::logs.TLog(1) << "Unknown output type" << std::endl;
      return 1;
   }
   SUBool::logs.Log(2) << std::endl;
   SUBool::logs.TLog(2) << "Output written" << std::endl;
   return 0;
}

template <class Var, class OfCNF, class ToCNF>
int
process_is(SUBool::CNF cnf, const config_t &conf, OfCNF of_cnf, ToCNF to_cnf)
{
   SUBool::logs.TLog(1) << "Starting computation" << std::endl;
   SUBool::CPUTime cpu_time;
   auto is = of_cnf(cnf, false);
   if (conf.body_minimal)
   {
      compute_body_minimal(is, conf);
   }
   else
   {
      compute_gd(is, conf);
   }
   SUBool::logs.TLog(1) << "Computation finished, proc time: " << cpu_time
                        << std::endl;
   return print_is(is, cnf.MaxVariable(), conf, to_cnf);
}

int
process_horn(const SUBool::CNF &horn_cnf, const config_t &conf)
{
   if (!IsDefiniteHornCNF(horn_cnf))
   {
      SUBool::logs.TLog(2) << "The input is not a definite Horn CNF"
                           << std::endl;
      return 1;
   }
   return process_is<unsigned>(horn_cnf, conf, SUBool::horn_cnf_to_implsys,
       SUBool::implsys_to_horn_cnf);
}

int
process_dual_rail(const SUBool::CNF &cnf, const config_t &conf)
{
   return process_is<SUBool::Literal>(
       cnf, conf, SUBool::dual_rail, SUBool::dual_rail_to_cnf);
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   conf->Dump();
   SUBool::logs.TLog(2) << "Reading input" << std::endl;
   auto in_cnf = SUBool::input_or_die<SUBool::CNF>(
       conf->input_file, 0, 2, "Failed to parse the input file");
   switch (conf->impl_type)
   {
   case IMPL_TYPE::HORN:
      return process_horn(in_cnf, *conf);
   case IMPL_TYPE::DUAL_RAIL:
      return process_dual_rail(in_cnf, *conf);
   default:
      SUBool::logs.TLog(1) << "Unknown impl_type";
      return 1;
   }
   return 0;
}
