/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfencode.cpp
 * Program for constructing URC and PC encodings of a given (smooth) DNNF.
 */

#include <array>
#include <csignal>
#include <iostream>

#include "cardinal.h"
#include "cardinal_opt.h"
#include "dnnf.h"
#include "dnnf_enc.h"
#include "enum_opt.h"
#include "logstream.h"
#include "prgutil.h"

const std::string SUBool::kPrgName = "nnfencode";
const std::string SUBool::kPrgDesc =
    "Construct a CNF encoding of a given (smooth) DNNF.";
const std::string kPrgHelp =
    "Constructs a consistency checker (CC), domain consistent encoding (DC), "
    "unit refutation complete (URC), or propagation complete (PC) encoding";

enum class ENCODING_TYPE
{
   CC,
   DC,
   URC,
   PC,
   Last = PC
};

const SUBool::EnumAnnotation<ENCODING_TYPE> kEncodingTypeAnnotation(
    "ENCODING_TYPE",
    {//
        {ENCODING_TYPE::CC, "cc",
            "consistency checker (CC), requires a DNNF"}, //
        {ENCODING_TYPE::DC, "dc",
            "domain consistent (DC) encoding, requires a smooth DNNF"}, //
        {ENCODING_TYPE::URC, "urc",
            "unit refutation complete (URC) encoding, requires a smooth "
            "DNNF"}, //
        {ENCODING_TYPE::PC, "pc",
            "unit refutation complete (PC) encoding, requires a smooth "
            "DNNF"}});

const ENCODING_TYPE kDefaultEncodingType = ENCODING_TYPE::CC;
const SUBool::LEVEL_FUNC kDefaultLevelFunc = SUBool::LEVEL_FUNC::FROM_LEAVES;

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   ENCODING_TYPE encoding_type{kDefaultEncodingType};
   bool final_minimize{false};
   SUBool::LEVEL_FUNC level_func{kDefaultLevelFunc};
   bool check_smooth{false};
   SUBool::AmoEncoder::KIND amo_encoding{
       SUBool::CardinalOptionsGroup::kDefaultAmoEncoding};
   SUBool::ExOneEncoder::KIND exone_encoding{
       SUBool::CardinalOptionsGroup::kDefaultExOneEncoding};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   kEncodingTypeAnnotation.DumpValue(
       "\t", level, "encoding_type", encoding_type);
   SUBool::logs.DumpValue("\t", level, "final_minimize", final_minimize);
   SUBool::kLevelFuncAnnotation.DumpValue(
       "\t", level, "level_func", level_func);
   SUBool::logs.DumpValue("\t", level, "check_smooth", check_smooth);
   SUBool::AmoEncoder::kKindAnnotation.DumpValue(
       "\t", level, "amo_encoding", amo_encoding);
   SUBool::ExOneEncoder::kKindAnnotation.DumpValue(
       "\t", level, "exone_encoding", exone_encoding);
}

bool
needs_smooth(ENCODING_TYPE enc_type)
{
   return enc_type != ENCODING_TYPE::CC;
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;

   auto opt_encoding_type = SUBool::make_enum_option(kEncodingTypeAnnotation,
       "encoding,e", "Type of the encoding.", kDefaultEncodingType);
   auto opt_level_func = SUBool::make_enum_option(SUBool::kLevelFuncAnnotation,
       "level-func,L",
       "How to compute levels in case of construction of URC or PC encoding.",
       kDefaultLevelFunc);

   SUBool::PrgOptions opts(true);
   opts.SetHelpDesc(kPrgHelp);
   opts.InputOutput(&conf.input_file, &conf.output_file);
   po::options_description desc("Encoding construction");
   desc.add(opt_encoding_type);
   desc.add(opt_level_func);
   desc.add_options() //
       ("minimize,m", po::bool_switch(&conf.final_minimize),
           "Perform minimization of the constructed encoding.") //
       ("smooth,S", po::bool_switch(&conf.check_smooth),
           "Check for smoothness and make DNNF smooth before constructing the "
           "encoding if necessary");
   opts.Add(desc);
   auto cardinal_opt_group =
       opts.MakeOptionsGroup<SUBool::CardinalOptionsGroup>();
   if (!opts.Parse(argc, argv, conf)
       || !opt_encoding_type.GetValue(conf.encoding_type)
       || !opt_level_func.GetValue(conf.level_func))
   {
      return {};
   }
   conf.amo_encoding = cardinal_opt_group->AmoEncoding();
   conf.exone_encoding = cardinal_opt_group->ExOneEncoding();
   return conf;
}

std::unique_ptr<SUBool::CNFEncoding>
construct_encoding(const SUBool::Nnf &nnf, const config_t &conf)
{
   switch (conf.encoding_type)
   {
   case ENCODING_TYPE::CC:
   {
      return std::unique_ptr<SUBool::CNFEncoding>(
          new SUBool::URCEncoding(SUBool::DnnfInputUrcEncoder::URCEncode(nnf)));
   }
   break;
   case ENCODING_TYPE::DC:
   {
      return std::unique_ptr<SUBool::CNFEncoding>(new SUBool::PCEncoding(
          SUBool::SmoothDnnfInputPcEncoder::PCEncode(nnf)));
   }
   break;
   case ENCODING_TYPE::URC:
   {
      return std::unique_ptr<SUBool::CNFEncoding>(new SUBool::URCEncoding(
          encode_smooth_dnnf_urc(nnf, conf.level_func, 2)));
   }
   break;
   case ENCODING_TYPE::PC:
   {
      return std::unique_ptr<SUBool::CNFEncoding>(new SUBool::PCEncoding(
          encode_smooth_dnnf_pc(nnf, conf.level_func, 2)));
   }
   break;
   default:;
   }
   throw SUBool::InvariantFailedException(
       "construct_encoding", "Unknown encoding type");
}

void
term_func(int signal)
{
   SUBool::logs.TLog(0) << "Terminated by signal " << signal << std::endl;
   exit(signal);
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   std::signal(SIGTERM, term_func);
   std::signal(SIGQUIT, term_func);
   std::signal(SIGINT, term_func);
   conf->Dump();
   SUBool::AmoEncoder::sGlobalKind = conf->amo_encoding;
   SUBool::ExOneEncoder::sGlobalKind = conf->exone_encoding;
   auto nnf = SUBool::input_or_die<SUBool::Nnf>(conf->input_file);
   if (needs_smooth(conf->encoding_type) && conf->check_smooth)
   {
      SUBool::logs.TLog(3) << "Checking smoothness" << std::endl;
      nnf = SUBool::nnf_smooth(std::move(nnf));
   }
   auto enc = construct_encoding(nnf, *conf);
   if (conf->final_minimize)
   {
      enc->Minimize();
   }
   SUBool::output(conf->output_file, *enc);
   return 0;
}
