/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfsmooth.cpp
 * Program for making a given nnf smooth or checking whether it is smooth.
 */

#include <csignal>
#include <iostream>

#include "dnnf.h"
#include "logstream.h"
#include "nnf.h"
#include "prgutil.h"

const std::string SUBool::kPrgName = "nnfsmooth";
const std::string SUBool::kPrgDesc = "Make a given DNNF smooth.";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   bool check_only{false};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   SUBool::logs.DumpValue("\t", level, "check_only", check_only);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   SUBool::PrgOptions opts;
   opts.InputOutput(&conf.input_file, &conf.output_file);
   po::options_description desc("Configuration options");
   desc.add_options()("check-only,k", po::bool_switch(&conf.check_only),
       "Only check if the given NNF is smooth.");
   opts.Add(desc);
   if (!opts.Parse(argc, argv, conf))
   {
      return {};
   }
   return conf;
}

void
term_func(int signal)
{
   SUBool::logs.TLog(0) << "Terminated by signal " << signal << std::endl;
   exit(signal);
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   std::signal(SIGTERM, term_func);
   std::signal(SIGQUIT, term_func);
   std::signal(SIGINT, term_func);
   conf->Dump();
   auto nnf = SUBool::input_or_die<SUBool::Nnf>(conf->input_file);
   if (conf->check_only)
   {
      auto traits = SUBool::nnf_traits(nnf);
      if (traits.smooth)
      {
         SUBool::logs.TLog(0) << "NNF is smooth" << std::endl;
      }
      else
      {
         SUBool::logs.TLog(0) << "NNF is not smooth" << std::endl;
      }
      return 0;
   }
   auto smooth_nnf = SUBool::nnf_smooth(std::move(nnf));
   SUBool::logs.TLog(1) << "Finished" << std::endl;
   SUBool::output(conf->output_file, smooth_nnf);
   SUBool::logs.TLog(1) << "Output written" << std::endl;
   return 0;
}
