/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file pccheck.cpp
 * Program for checking if a formula is PC or URC. Can also be used to get the
 * CNF encoding PC/URC property.
 */

#include <array>
#include <chrono>
#include <csignal>
#include <ctime>
#include <fstream>
#include <future>
#include <iostream>

#include "cardinal.h"
#include "cardinal_opt.h"
#include "cnfcompress.h"
#include "dimacs.h"
#include "logstream.h"
#include "normalform.h"
#include "pc_opt.h"
#include "pccomp.h"
#include "pcenc.h"
#include "pclearncomp.h"
#include "prgutil.h"
#include "random_utils.h"
#include "resolve.h"
#include "uccomp.h"
#include "ucenc.h"
#include "unitprop.h"

const std::string SUBool::kPrgName = "pccheck";
const std::string SUBool::kPrgDesc = "Checking if a CNF is PC or URC";
const std::string prg_help =
    "Checks if a given CNF is propagation complete (PC) or unit refutation "
    "complete (URC). Can also output the SAT encoding of the required query.";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string encoding_file{};
   bool encode_only{false};
   bool no_up_backbones{false};
   unsigned initial_up_level{1};
   SUBool::AbsEncoder::AbsConfig enc_conf{};
   SUBool::AbsCheckerSAT::Timeouts timeouts{};
   SUBool::LIT_IMPL_SYSTEM impl_system{SUBool::LIT_IMPL_SYSTEM::Default};
   SUBool::PCCheckGeneralOptions::GOAL goal{
       SUBool::PCCheckGeneralOptions::kDefaultGoal};
   SUBool::AmoEncoder::KIND amo_encoding{SUBool::AmoEncoder::KIND::REPR};
   SUBool::ExOneEncoder::KIND exone_encoding{SUBool::ExOneEncoder::KIND::REPR};
   SUBool::SOLVER_TYPE solver_type{SUBool::SOLVER_TYPE::GLUCOSE};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::PCCheckGeneralOptions::kGoalAnnotation.DumpValue(
       "\t", level, "goal", goal);
   SUBool::logs.DumpValue("\t", level, "encode_only", encode_only);
   SUBool::logs.DumpValue("\t", level, "encoding_file", encoding_file);
   SUBool::logs.DumpValue("\t", level, "no_up_backbones", no_up_backbones);
   SUBool::logs.DumpValue("\t", level, "initial_up_level", initial_up_level);
   SUBool::kLitImplSystemAnnotation.DumpValue(
       "\t", level, "impl_system", impl_system);
   SUBool::logs.TLog(level) << "\tenc_conf:" << std::endl;
   enc_conf.Dump("\t\t", level);
   SUBool::logs.DumpValue("\t", level, "timeouts.timeout", timeouts.timeout);
   SUBool::logs.DumpValue(
       "\t", level, "timeouts.level_base", timeouts.level_base);
   SUBool::logs.DumpValue(
       "\t", level, "timeouts.level_factor", timeouts.level_factor);
   SUBool::kLitImplSystemAnnotation.DumpValue(
       "\t", level, "impl_system", impl_system);
   SUBool::AmoEncoder::kKindAnnotation.DumpValue(
       "\t", level, "amo_encoding", amo_encoding);
   SUBool::ExOneEncoder::kKindAnnotation.DumpValue(
       "\t", level, "exone_encoding", exone_encoding);
   SUBool::kSolverTypeAnnotation.DumpValue(
       "\t", level, "solver_type", solver_type);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   SUBool::PrgOptions opts(true);
   opts.SetHelpDesc(prg_help);
   opts.SinglePositional({&conf.input_file, "input"});
   auto general_opts = opts.MakeOptionsGroup<SUBool::PCCheckGeneralOptions>();
   auto solver_opts = opts.MakeOptionsGroup<SUBool::PCSolverOptions>();
   auto oneprove_opts = opts.MakeOptionsGroup<SUBool::PCOneProveOptions>(false);
   auto timeout_opts = opts.MakeOptionsGroup<SUBool::PCTimeoutsOptions>(false);
   auto cardinal_opt_group =
       opts.MakeOptionsGroup<SUBool::CardinalOptionsGroup>();
   if (!opts.Parse(argc, argv, conf))
   {
      return {};
   }
   conf.goal = general_opts->Goal();
   if (general_opts->EncodeFile())
   {
      conf.encode_only = true;
      conf.encoding_file = *(general_opts->EncodeFile());
   }
   conf.enc_conf.use_backmap = general_opts->Dump();
   conf.no_up_backbones = general_opts->NoUPBackbones();
   conf.initial_up_level = general_opts->InitialUPLevel();
   conf.solver_type = solver_opts->SolverType();
   conf.enc_conf.up_one_prove = oneprove_opts->OneProve();
   conf.enc_conf.up_level_bound = oneprove_opts->UPLevelBound();
   conf.enc_conf.log_conf = oneprove_opts->EncLogarithmicConf();
   conf.impl_system = oneprove_opts->ImplSystem();
   conf.timeouts = timeout_opts->EmpTimeouts();
   conf.amo_encoding = cardinal_opt_group->AmoEncoding();
   conf.exone_encoding = cardinal_opt_group->ExOneEncoding();
   return conf;
}

int
encode(const config_t &conf, const SUBool::CompHypothesis &hyp,
    SUBool::AbsCheckerSAT &checker)
{
   auto cnf = hyp.GetHypothesis(false);
   auto enc = checker.Encoder(conf.enc_conf.up_level_bound);
   SUBool::ClausesToLiterals cl2lit(cnf);

   std::shared_ptr<SUBool::EncISProvider> impl_system_provider;
   if (checker.RequiresImplSystem())
   {
      impl_system_provider = make_enc_is_provider(conf.impl_system);
      impl_system_provider->Cleanup(cnf);
      if (checker.RequiresSCCGraph())
      {
         impl_system_provider->RebuildSCCGraph();
      }
   }
   else
   {
      impl_system_provider = std::make_shared<SUBool::DummyISProvider>();
   }
   enc->BuildEncoding(cnf, cl2lit, impl_system_provider);
   SUBool::OutputStream os(conf.encoding_file);
   if (!os)
   {
      SUBool::logs.TLog(SUBool::LogStream::LOG_LEVEL::NONE)
          << "[encode] Failed to open file '" << conf.encoding_file
          << "' for writing" << std::endl;
      return 2;
   }
   enc->WriteEncoding(os);
   SUBool::logs.TLog(3) << "Encoding written" << std::endl;
   return 0;
}

int
empower(const config_t &conf, SUBool::CompHypothesis &hyp,
    SUBool::AbsCheckerSAT &checker)
{
   if (checker.RequiresImplSystem())
   {
      hyp.SetImplSystemProvider(make_enc_is_provider(conf.impl_system));
   }
   else
   {
      hyp.SetImplSystemProvider(std::make_shared<SUBool::DummyISProvider>());
   }
   auto emp_result = hyp.FindEmpowering(checker, false);
   if (emp_result)
   {
      auto emp_var = checker.LastWitnessVariable();
      // TODO: change the last parameter to true if consistency is checked
      // before calling empower
      auto clause = hyp.PrimeSubimplicate(*emp_result, emp_var, false);
      SUBool::logs.TLog(1) << "Found empowering implicate " << clause
                           << " with empowered variable " << emp_var
                           << std::endl;
      return 1;
   }
   SUBool::logs.TLog(1) << "No empowering implicate found, formula is "
                        << SUBool::PCCheckGeneralOptions::kGoalAnnotation.Name(
                               conf.goal)
                        << std::endl;
   return 0;
}

void
terminate(int signal)
{
   std::stringstream ss;
   ss << "terminated by signal " << signal;
   _Exit(2);
}

std::unique_ptr<SUBool::PCCheckerSAT>
pc_checker_sat(const config_t &conf)
{
   auto checker = std::make_unique<SUBool::PCCheckerSAT>();
   SUBool::PCEncoder::Config check_conf(conf.enc_conf);
   checker->Init(check_conf, false, conf.timeouts, conf.initial_up_level);
   checker->SetSolverType(conf.solver_type);
   return checker;
}

std::unique_ptr<SUBool::UCCheckerSAT>
urc_checker_sat(const config_t &conf)
{
   auto checker = std::make_unique<SUBool::UCCheckerSAT>();
   checker->Init(conf.enc_conf, false, conf.timeouts);
   checker->SetSolverType(conf.solver_type);
   return checker;
}

std::unique_ptr<SUBool::AbsCheckerSAT>
checker_sat(const config_t &conf)
{
   switch (conf.goal)
   {
   case SUBool::PCCheckGeneralOptions::GOAL::PC:
      return pc_checker_sat(conf);
   case SUBool::PCCheckGeneralOptions::GOAL::URC:
      return urc_checker_sat(conf);
   default:
      throw SUBool::BadParameterException("checker_sat", "Unknown goal");
   }
}

int
execute(const config_t &conf, SUBool::CNF cnf)
{
   SUBool::CPUTime cpu_time;
   auto repr = cnf;
   SUBool::CompHypothesis hyp(
       SUBool::MINIMIZE_METHOD::FULL, std::move(cnf), std::move(repr), true);
   if (!conf.no_up_backbones)
   {
      hyp.PropagateUPBackbones();
   }
   auto checker = checker_sat(conf);
   auto res = (conf.encode_only ? encode(conf, hyp, *checker)
                                : empower(conf, hyp, *checker));
   SUBool::logs.TLog(2) << "Total processor time: " << cpu_time << std::endl;
   return res;
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return 1;
   }
   auto cnf = SUBool::input_or_die<SUBool::CNF>(
       conf->input_file, 0, 2, "Failed to parse the input file");

   std::signal(SIGTERM, terminate);
   std::signal(SIGQUIT, terminate);
   std::signal(SIGINT, terminate);

   SUBool::logs.TLog(SUBool::LogStream::LOG_LEVEL::MINLINES)
       << "PCCheck" << SUBool::kVersionInfo << std::endl;

   conf->Dump();
   SUBool::AmoEncoder::sGlobalKind = conf->amo_encoding;
   SUBool::ExOneEncoder::sGlobalKind = conf->exone_encoding;
   auto ret = execute(*conf, std::move(cnf));
   return ret;
}
