/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file pccompile.cpp
 * Program for PC compilation via adding new empowering implicates.
 */

#include <array>
#include <chrono>
#include <csignal>
#include <ctime>
#include <fstream>
#include <future>
#include <iostream>

#include "cardinal.h"
#include "cardinal_opt.h"
#include "cnfcompress.h"
#include "dimacs.h"
#include "logstream.h"
#include "normalform.h"
#include "pc_opt.h"
#include "pccomp.h"
#include "pcenc.h"
#include "pclearncomp.h"
#include "prgutil.h"
#include "random_utils.h"
#include "resolve.h"
#include "uccomp.h"
#include "ucenc.h"
#include "unitprop.h"

const std::string SUBool::kPrgName = "pccompile";
const std::string SUBool::kPrgDesc =
    "compiling a CNF into a propagation complete or unit refutation complete "
    "CNF.";
const std::string prg_help =
    "Implements an incremental algorithm based on adding empowering "
    "implicates and a learning algorithm";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   bool minimize{false};
   SUBool::PCCompAlgorithmOptions::ALGORITHM algorithm{
       SUBool::PCCompAlgorithmOptions::kDefaultAlgorithm};
   bool level_minimize{true}; // minimization at level increase
   unsigned verbosity{3};     // 0 means no output
   unsigned pccheck_random_rounds{100};
   unsigned pccheck_random_budget{10};
   unsigned pccheck_random_budget_inc{1};
   SUBool::PCLearnComp::INITIAL_NEGATIVES initial_negatives{
       SUBool::PCLearnComp::INITIAL_NEGATIVES::ALL_EMPOWERING};
   SUBool::PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY
       candidates_from_negative_policy{
           SUBool::PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY::UNITPROP};
   bool use_cnf_hypothesis_for_empowerment{false};
   long double timeout{-1.0};
   SUBool::AbsComp::BaseConfig comp_base_config{};
   bool unfinished_write{true};
   SUBool::AmoEncoder::KIND amo_encoding{SUBool::AmoEncoder::KIND::REPR};
   SUBool::ExOneEncoder::KIND exone_encoding{SUBool::ExOneEncoder::KIND::REPR};

   virtual void InnerDump(unsigned level) const override;
};

std::shared_ptr<SUBool::AbsComp> the_compiler;
config_t the_conf;

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   SUBool::logs.DumpValue("\t", level, "minimize", minimize);
   SUBool::kMinimizeMethodAnnotation.DumpValue(
       "\t", level, "minimize_method", comp_base_config.minimize_method);
   SUBool::PCCompAlgorithmOptions::kAlgorithmAnnotation.DumpValue(
       "\t", level, "algorithm", algorithm);
   SUBool::kLitImplSystemAnnotation.DumpValue(
       "\t", level, "impl_system", comp_base_config.impl_system);
   SUBool::logs.TLog(level) << "\tcomp_base_config.minimize_steps: "
                            << comp_base_config.minimize_steps
                            << (comp_base_config.minimize_steps > 0
                                       ? ""
                                       : "(no minimization in intervals)")
                            << std::endl;
   SUBool::logs.DumpValue("\t", level, "level_minimize", level_minimize);
   SUBool::kLitImplSystemAnnotation.DumpValue("\t", level,
       "comp_base_config.impl_system", comp_base_config.impl_system);
   SUBool::kImplSystemRebuildAnnotation.DumpValue("\t", level,
       "comp_base_config.impl_system_rebuild",
       comp_base_config.impl_system_rebuild);
   SUBool::AbsComp::kPreprocessMethodAnnotation.DumpValue(
       "\t", level, "comp_base_config.preprocess", comp_base_config.preprocess);
   SUBool::logs.DumpValue("\t", level, "comp_base_config.propagate_backbones",
       comp_base_config.propagate_backbones);
   SUBool::logs.DumpValue(
       "\t", level, "pccheck_random_rounds", pccheck_random_rounds);
   SUBool::logs.DumpValue(
       "\t", level, "pccheck_random_budget", pccheck_random_budget);
   SUBool::logs.DumpValue(
       "\t", level, "pccheck_random_budget_inc", pccheck_random_budget_inc);
   SUBool::PCLearnComp::kInitialNegativesAnnotation.DumpValue(
       "\t", level, "initial_negatives", initial_negatives);
   SUBool::PCLearnComp::kCandidatesFromNegativePolicyAnnotation.DumpValue("\t",
       level, "candidates_from_negative", candidates_from_negative_policy);
   SUBool::logs.DumpValue("\t", level, "use_cnf_hypothesis_for_empowerment",
       use_cnf_hypothesis_for_empowerment);
   SUBool::logs.DumpValue("\t", level, "timeout", timeout);
   SUBool::logs.DumpValue("\t", level, "comp_base_config.timeouts.timeout",
       comp_base_config.timeouts.timeout);
   SUBool::logs.DumpValue("\t", level, "comp_base_config.timeouts.level_base",
       comp_base_config.timeouts.level_base);
   SUBool::logs.DumpValue("\t", level, "comp_base_config.timeouts.level_factor",
       comp_base_config.timeouts.level_factor);
   SUBool::kSolverTypeAnnotation.DumpValue("\t", level,
       "comp_base_config.solver_type", comp_base_config.solver_type);
   SUBool::logs.TLog(level) << "\tenc_conf:" << std::endl;
   comp_base_config.enc_conf.Dump("\t\t", level);
   SUBool::AmoEncoder::kKindAnnotation.DumpValue(
       "\t", level, "amo_encoding", amo_encoding);
   SUBool::ExOneEncoder::kKindAnnotation.DumpValue(
       "\t", level, "exone_encoding", exone_encoding);
}

int
parse_arguments(int argc, char *argv[], config_t &conf)
{
   SUBool::PrgOptions opts(true);
   opts.SetHelpDesc(prg_help);
   opts.InputOutput(&conf.input_file, &conf.output_file);
   auto algorithm_opts =
       opts.MakeOptionsGroup<SUBool::PCCompAlgorithmOptions>();
   auto pclearncomp_opts = opts.MakeOptionsGroup<SUBool::PCLearnCompOptions>();
   auto preprocess_opts =
       opts.MakeOptionsGroup<SUBool::PCCompPreprocessOptions>();
   auto solver_opts = opts.MakeOptionsGroup<SUBool::PCSolverOptions>();
   auto oneprove_opts = opts.MakeOptionsGroup<SUBool::PCOneProveOptions>(true);
   auto pccheck_random_opts =
       opts.MakeOptionsGroup<SUBool::PCRandomPCChecksOptions>();
   auto minimize_opts = opts.MakeOptionsGroup<SUBool::PCMinimizationOptions>();
   auto timeout_opts = opts.MakeOptionsGroup<SUBool::PCTimeoutsOptions>();
   auto cardinal_opt_group =
       opts.MakeOptionsGroup<SUBool::CardinalOptionsGroup>();
   if (!opts.Parse(argc, argv, conf))
   {
      return 1;
   }
   conf.algorithm = algorithm_opts->Algorithm();
   conf.initial_negatives = pclearncomp_opts->InitialNegatives();
   conf.candidates_from_negative_policy =
       pclearncomp_opts->CandidatesFromNegativePolicy();
   conf.use_cnf_hypothesis_for_empowerment =
       pclearncomp_opts->UseCNFHypothesisForEmpeworment();
   conf.comp_base_config.preprocess = preprocess_opts->PreprocessMethod();
   conf.comp_base_config.propagate_backbones =
       !preprocess_opts->NoBackbonePropagation();
   conf.comp_base_config.solver_type = solver_opts->SolverType();
   conf.comp_base_config.enc_conf.up_one_prove = oneprove_opts->OneProve();
   conf.comp_base_config.enc_conf.up_level_bound =
       oneprove_opts->UPLevelBound();
   conf.comp_base_config.enc_conf.log_conf =
       oneprove_opts->EncLogarithmicConf();
   conf.comp_base_config.impl_system = oneprove_opts->ImplSystem();
   conf.comp_base_config.impl_system_rebuild =
       oneprove_opts->ImplSystemRebuild();
   conf.pccheck_random_rounds = pccheck_random_opts->Rounds();
   conf.pccheck_random_budget = pccheck_random_opts->Budget();
   conf.pccheck_random_budget_inc = pccheck_random_opts->BudgetInc();
   conf.minimize = minimize_opts->FinalMinimize();
   conf.comp_base_config.minimize_method = minimize_opts->MinimizeMethod();
   conf.comp_base_config.minimize_steps = minimize_opts->MinimizeSteps();
   conf.level_minimize = minimize_opts->LevelMinimize();
   conf.timeout = timeout_opts->Timeout();
   conf.comp_base_config.timeouts = timeout_opts->EmpTimeouts();
   conf.amo_encoding = cardinal_opt_group->AmoEncoding();
   conf.exone_encoding = cardinal_opt_group->ExOneEncoding();
   return 0;
}

template <class T, class R>
R
wait_with_timeout(long double timeout, std::future<R> &fut, T &comp)
{
   if (timeout > 0.0)
   {
      if (fut.wait_for(std::chrono::duration<long double>(timeout))
          == std::future_status::timeout)
      {
         comp.Abort("timed out");
         comp.PrintStatistics(1);
         fut.wait();
         return SUBool::AbsComp::RESULT::TIMED_OUT; //_Exit(3);
      }
   }
   else
   {
      fut.wait();
   }
   return fut.get();
}

void
fill_base_conf(SUBool::AbsComp::BaseConfig &out_conf, const config_t &conf)
{
   out_conf = conf.comp_base_config;
   out_conf.final_minimize = conf.minimize;
   out_conf.level_minimize = conf.level_minimize;
}

// std::unique_ptr<SUBool::AbsIncrementalComp>
std::shared_ptr<SUBool::AbsIncrementalComp>
incremental_compiler(const config_t &conf, SUBool::CNF cnf)
{
   SUBool::AbsComp::BaseConfig a_conf;
   fill_base_conf(a_conf, conf);
   bool interleave_pc = false;

   switch (conf.algorithm)
   {
   case SUBool::PCCompAlgorithmOptions::ALGORITHM::INCREMENT_PC:
      return std::make_shared<SUBool::PCComp>(
          std::move(cnf), std::move(a_conf));
   case SUBool::PCCompAlgorithmOptions::ALGORITHM::INCREMENT_URC_PC:
      interleave_pc = true;
      [[fallthrough]];
   case SUBool::PCCompAlgorithmOptions::ALGORITHM::INCREMENT_URC:
   {
      SUBool::UCComp::Config uc_conf(a_conf);
      uc_conf.interleave_pc = interleave_pc;
      return std::make_shared<SUBool::UCComp>(
          std::move(cnf), std::move(uc_conf));
   }
   default:
      throw SUBool::BadParameterException(
          "compiler", "Bad algorithm specification");
   }
}

std::shared_ptr<SUBool::PCLearnComp>
learning_compiler(const config_t &conf, SUBool::CNF cnf)
{
   SUBool::PCLearnComp::Config l_conf;
   fill_base_conf(l_conf, conf);
   l_conf.pccheck_random_rounds = conf.pccheck_random_rounds;
   l_conf.pccheck_random_budget = conf.pccheck_random_budget;
   l_conf.pccheck_random_budget_inc = conf.pccheck_random_budget_inc;
   l_conf.initial_negatives = conf.initial_negatives;
   l_conf.candidates_from_negative_policy =
       conf.candidates_from_negative_policy;
   l_conf.use_cnf_hypothesis_for_empowerment =
       conf.use_cnf_hypothesis_for_empowerment;
   return std::make_shared<SUBool::PCLearnComp>(
       std::move(cnf), std::move(l_conf));
}

std::shared_ptr<SUBool::AbsComp>
compiler(const config_t &conf, SUBool::CNF cnf)
{
   if (conf.algorithm == SUBool::PCCompAlgorithmOptions::ALGORITHM::LEARNING)
   {
      return learning_compiler(conf, std::move(cnf));
   }
   return incremental_compiler(conf, std::move(cnf));
}

int
write_output()
{
   if (!the_compiler)
   {
      SUBool::logs.TLog(1)
          << "[write_output] the_compiler is nullptr, no output written"
          << std::endl;
   }
   try
   {
      std::stringstream ss;
      SUBool::print_version(ss, true);
      SUBool::output(
          the_conf.output_file, the_compiler->CurrentCNF(), ss.str());
   }
   catch (SUBool::IOException &e)
   {
      SUBool::logs.TLog(SUBool::LogStream::LOG_LEVEL::NONE)
          << "[write_output]" << e.what() << std::endl;
      return 2;
   }
   return 0;
}

int
compile(SUBool::CNF cnf)
{
   auto comp = compiler(the_conf, std::move(cnf));
   the_compiler = comp;
   auto fut = std::async(
       std::launch::async,
       [](std::shared_ptr<SUBool::AbsComp> comp) { return comp->Compile(); },
       comp);
   auto res = wait_with_timeout(the_conf.timeout, fut, *comp);
   if (res == SUBool::AbsComp::RESULT::COMPILED || the_conf.unfinished_write)
   {
      return write_output();
   }
   SUBool::logs.TLog(SUBool::LogStream::LOG_LEVEL::MINLINES)
       << "No output written" << std::endl;
   return 1;
}

void
terminate(int signal)
{
   std::stringstream ss;
   ss << "terminated by signal " << signal;
   if (the_compiler)
   {
      the_compiler->Abort(ss.str());
      the_compiler->PrintStatistics(1);
      if (the_conf.unfinished_write)
      {
         write_output();
      }
      else
      {
         SUBool::logs.TLog(3) << "c No output written" << std::endl;
      }
   }
   _Exit(2);
}

int
main(int argc, char *argv[])
{
   int r = parse_arguments(argc, argv, the_conf);
   if (r != 0)
   {
      return r;
   }
   auto cnf = SUBool::input_or_die<SUBool::CNF>(
       the_conf.input_file, 0, 2, "Failed to parse the input file");

   std::signal(SIGTERM, terminate);
   std::signal(SIGQUIT, terminate);
   std::signal(SIGINT, terminate);

   SUBool::logs.TLog(SUBool::LogStream::LOG_LEVEL::MINLINES)
       << "PCCompile " << SUBool::kVersionInfo << std::endl;

   the_conf.Dump();
   SUBool::AmoEncoder::sGlobalKind = the_conf.amo_encoding;
   SUBool::ExOneEncoder::sGlobalKind = the_conf.exone_encoding;
   auto ret = compile(std::move(cnf));
   the_compiler = nullptr;
   return ret;
}
