/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file pcminimize.cpp
 * Program for minimization with respect to propagation completeness. Iterative
 * removing absorbtions.
 */

#include <array>
#include <chrono>
#include <csignal>
#include <ctime>
#include <fstream>
#include <future>
#include <iostream>

#include "cardinal.h"
#include "cnfcompress.h"
#include "dimacs.h"
#include "logstream.h"
#include "normalform.h"
#include "pc_opt.h"
#include "pccomp.h"
#include "pcenc.h"
#include "pclearncomp.h"
#include "prgutil.h"
#include "random_utils.h"
#include "resolve.h"
#include "uccomp.h"
#include "ucenc.h"
#include "unitprop.h"

const std::string SUBool::kPrgName = "pcminimize";
const std::string SUBool::kPrgDesc =
    "minimizing a CNF with respect to propagation completeness by iterative "
    "removing of absorbed clauses.";

struct config_t : public SUBool::PrgConfigBase
{
   std::string input_file{};
   std::string output_file{};
   SUBool::MINIMIZE_METHOD minimize_method{SUBool::MINIMIZE_METHOD::FULL};
   bool backbones{false}; // Whether backbones should be propagated
   bool prime{false};     // make prime before minimization
   long double timeout{-1.0l};

   virtual void InnerDump(unsigned level) const override;
};

void
config_t::InnerDump(unsigned level) const
{
   SUBool::logs.DumpValue("\t", level, "input_file", input_file);
   SUBool::logs.DumpValue("\t", level, "output_file", output_file);
   SUBool::kMinimizeMethodAnnotation.DumpValue(
       "\t", level, "minimize_method", minimize_method);
   SUBool::logs.DumpValue("\t", level, "backbones", backbones);
   SUBool::logs.DumpValue("\t", level, "prime", prime);
   SUBool::logs.DumpValue("\t", level, "timeout", timeout);
}

std::optional<config_t>
parse_arguments(int argc, char *argv[])
{
   config_t conf;
   auto minimize_method = SUBool::minimize_method_opt();
   SUBool::PrgOptions opts(true);
   opts.InputOutput(&conf.input_file, &conf.output_file);
   po::options_description desc("Configuration options");
   desc.add(minimize_method);
   desc.add_options() //
       ("backbones,B", po::bool_switch(&conf.backbones),
           "Use backbone propagation before the minimization is executed") //
       ("prime,p", po::bool_switch(&conf.prime),
           "Make CNF prime before running the minimization (this implies using "
           "backbone propagation)") //
       ("timeout,t", po::value<long double>(&conf.timeout),
           "Timeout for compilation in seconds as a fractional number. After "
           "this amount of seconds, the process finishes. Value t<=0.0 means "
           "no timeout. The timer uses CLOCK_REALTIME.");
   opts.Add(desc);
   if (!opts.Parse(argc, argv, conf)
       || !minimize_method.GetValue(conf.minimize_method))
   {
      return {};
   }
   conf.backbones |= conf.prime;
   return conf;
}

template <class T, class R>
bool
wait_with_timeout(long double timeout, std::future<R> &fut, T &comp)
{
   if (timeout > 0.0)
   {
      if (fut.wait_for(std::chrono::duration<long double>(timeout))
          == std::future_status::timeout)
      {
         comp.Abort("timed out");
         comp.PrintStatistics(1);
         fut.wait();
         return false; //_Exit(3);
      }
   }
   else
   {
      fut.wait();
   }
   return fut.get();
}

int
write_output(const config_t &conf, const SUBool::CNF &cnf)
{
   try
   {
      std::stringstream ss;
      SUBool::print_version(ss, true);
      SUBool::output(conf.output_file, cnf, ss.str());
   }
   catch (SUBool::IOException &e)
   {
      SUBool::logs.TLog(SUBool::LogStream::LOG_LEVEL::NONE)
          << "[write_output]" << e.what() << std::endl;
      return 2;
   }
   return 0;
}

template <class F>
bool
run_with_timeout(
    bool no_timeout, long double timeout, F f, const std::string &desc)
{
   if (no_timeout)
   {
      f();
      return true;
   }
   SUBool::logs.TLog(2) << "timeut: " << timeout << std::endl;
   if (timeout > 0.0f)
   {
      auto fut = std::async(std::launch::async, f);
      if (fut.wait_for(std::chrono::duration<long double>(timeout))
          == std::future_status::timeout)
      {
         SUBool::logs.TLog(2) << "Timed out during " << desc << std::endl;
         SUBool::logs.mVerbLevel = 0;
         _Exit(2);
      }
      return true;
   }
   SUBool::logs.TLog(2) << "Negative timeout for " << desc << std::endl;
   SUBool::logs.mVerbLevel = 0;
   return false;
}

int
minimize(const config_t &conf, SUBool::CNF cnf)
{
   SUBool::CPUTime cpu_time;
   SUBool::CompHypothesis hyp(
       conf.minimize_method, SUBool::CNF(), std::move(cnf), false);
   if (conf.backbones
       && !run_with_timeout(
           conf.timeout < 0.0f, conf.timeout - cpu_time.TimeSpan(),
           [&hyp]() { hyp.PropagateBackbones(); }, "backbones propagation"))
   {
      return 2;
   }
   if (conf.prime
       && !run_with_timeout(
           conf.timeout < 0.0f, conf.timeout - cpu_time.TimeSpan(),
           [&hyp]() { hyp.MakeReprPrime(); }, "making CNF prime"))
   {
      return 2;
   }
   hyp.MinimizeRepresentation();
   if (!run_with_timeout(
           conf.timeout < 0.0f, conf.timeout - cpu_time.TimeSpan(),
           [&hyp]() { hyp.MinimizeRepresentation(); }, "minimization"))
   {
      return 2;
   }
   write_output(conf, hyp.GetRepresentation(true));
   SUBool::logs.TLog(2) << "Finished successfully" << std::endl;
   SUBool::logs.TLog(2) << "Total CPU time: " << cpu_time << std::endl;
   return 0;
}

void
terminate(int signal)
{
   std::stringstream ss;
   ss << "terminated by signal " << signal;
   _Exit(2);
}

int
main(int argc, char *argv[])
{
   auto conf = parse_arguments(argc, argv);
   if (!conf)
   {
      return {};
   }
   SUBool::logs.TLog(3) << "PCMinimize" << SUBool::kVersionInfo << std::endl;
   conf->Dump();
   auto cnf = SUBool::input_or_die<SUBool::CNF>(
       conf->input_file, 0, 2, "Failed to parse the input file");

   std::signal(SIGTERM, terminate);
   std::signal(SIGQUIT, terminate);
   std::signal(SIGINT, terminate);

   auto ret = minimize(*conf, std::move(cnf));
   return ret;
}
