/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file abscomp.cpp
 * Abstract compiler. Common properties of a compiler into
 * PC or UC formula.
 */

#include "abscomp.h"
#include "propagate.h"

const SUBool::AbsComp::StateValue SUBool::AbsComp::INITIALIZED{
    "INITIALIZED", SUBool::AbsComp::RESULT::NOT_RUN};
const SUBool::AbsComp::StateValue SUBool::AbsComp::PROPAGATE_BACKBONES{
    "PROPAGATE_BACKBONES", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::AbsComp::PREPROCESS{
    "PREPROCESS", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::AbsComp::FINISHED{
    "FINISHED", SUBool::AbsComp::RESULT::COMPILED};
const SUBool::AbsComp::StateValue SUBool::AbsComp::ABORTED{
    "ABORTED", SUBool::AbsComp::RESULT::TERMINATED};
const SUBool::AbsComp::StateValue SUBool::AbsComp::ERROR{
    "ERROR", SUBool::AbsComp::RESULT::ERROR};
const SUBool::AbsComp::StateValue SUBool::AbsComp::UNSAT{
    "UNSAT", SUBool::AbsComp::RESULT::COMPILED};
const SUBool::AbsComp::StateValue SUBool::AbsComp::MINIMIZE{
    "MINIMIZE", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::AbsComp::SOLVER_INTERRUPTED{
    "SOLVER_INTERRUPTED", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::AbsComp::TIMED_OUT{
    "TIMED_OUT", SUBool::AbsComp::RESULT::TIMED_OUT};

const SUBool::EnumAnnotation<SUBool::AbsComp::PREPROCESS_METHOD>
    SUBool::AbsComp::kPreprocessMethodAnnotation("PREPROCESS_METHOD",
        {                                                           //
            {PREPROCESS_METHOD::NONE, "none", "no preprocessing."}, //
            {PREPROCESS_METHOD::MINIMIZE, "minimize",
                "minimize the formula by removing the clauses absorbed by the "
                "others (not empowering with respect to propagation "
                "completeness)"}, //
            {PREPROCESS_METHOD::UP_PRIME_MINIMIZE, "up_prime_minimize",
                "make the CNF prime with respect to 1-provability, then "
                "minimize it."}, //
            {PREPROCESS_METHOD::PRIME_MINIMIZE, "prime_minimize",
                "make the CNF prime and then minimize it."}, //
            {PREPROCESS_METHOD::UP_PRIME, "up_prime",
                "make the CNF prime with respect to 1-provability with no "
                "minimization."}, //
            {PREPROCESS_METHOD::PRIME, "prime",
                "make the CNF prime with no minimization."}});

SUBool::AbsComp::AbsComp(
    CNF cnf, MINIMIZE_METHOD minimize_method, IMPL_SYSTEM_REBUILD is_rebuild)
    : mStateMutex(), mRTime(), mCPUTime(),
      mHypothesis(minimize_method, CNF(), std::move(cnf), false)
{
   mHypothesis.SetImplSystemRebuild(is_rebuild);
}

bool
SUBool::AbsComp::Over() const
{
   std::lock_guard<std::mutex> lock(mStateMutex);
   return LockedOver();
}

bool
SUBool::AbsComp::LockedOver() const
{
   return mState->result != RESULT::NOT_RUN
          && mState->result != RESULT::UNFINISHED;
}

bool
SUBool::AbsComp::CheckState(const StateValue &status) const
{
   std::lock_guard<std::mutex> lock(mStateMutex);
   return mState == &status;
}

bool
SUBool::AbsComp::State(const StateValue &status)
{
   std::lock_guard<std::mutex> lock(mStateMutex);
   if (LockedOver())
   {
      return false;
   }
   mState = &status;
   return true;
}

SUBool::AbsComp::RESULT
SUBool::AbsComp::StateResult() const
{
   std::lock_guard<std::mutex> lock(mStateMutex);
   return mState->result;
}

void
SUBool::AbsComp::PreCompile()
{
   mCPUTime.Reset();
   mRTime.Reset();
   if (GetBaseConfig().propagate_backbones)
   {
      PropagateBackbones();
      if (Over())
      {
         return;
      }
   }
   Preprocess();
}

SUBool::AbsComp::RESULT
SUBool::AbsComp::Compile()
{
   if (!CheckState(INITIALIZED))
   {
      logs.TLog(1) << "ERROR: Called AbsComp::Compile() when the status is not "
                      "INITIALIZED"
                   << std::endl;
      State(ERROR);
      return StateResult();
   }
   logs.TLog(2) << "Compilation starts" << std::endl;
   PreCompile();
   if (!Over())
   {
      CompilePreprocessed();
   }
   return StateResult();
}

void
SUBool::AbsComp::PropagateBackbones()
{
   if (!State(PROPAGATE_BACKBONES))
   {
      return;
   }
   if (!mHypothesis.PropagateBackbones())
   {
      State(UNSAT);
   }
   if (mHypothesis.Empty())
   {
      logs.TLog(1) << "The formula is empty after propagating backbones, the "
                      "compilation is finished"
                   << std::endl;
      State(FINISHED);
   }
}

void
SUBool::AbsComp::Preprocess()
{
   if (!State(PREPROCESS))
   {
      return;
   }
   bool minimize = false;
   switch (GetBaseConfig().preprocess)
   {
   case PREPROCESS_METHOD::NONE:
      break;
   case PREPROCESS_METHOD::MINIMIZE:
      minimize = true;
      break;
   case PREPROCESS_METHOD::UP_PRIME_MINIMIZE:
      minimize = true;
      [[fallthrough]];
   case PREPROCESS_METHOD::UP_PRIME:
      PreprocessUPPrime();
      break;
   case PREPROCESS_METHOD::PRIME_MINIMIZE:
      minimize = true;
      [[fallthrough]];
   case PREPROCESS_METHOD::PRIME:
      PreprocessPrime();
      break;
   default:
      throw BadParameterException(
          "AbsComp::Preprocess", "Unkown preprocess method");
   }
   if (minimize)
   {
      mHypothesis.MinimizeRepresentation();
   }
}

SUBool::AbsComp::PREPROCESS_METHOD
SUBool::AbsComp::PMRemoveMinimize(PREPROCESS_METHOD preprocess)
{
   switch (preprocess)
   {
   case PREPROCESS_METHOD::UP_PRIME_MINIMIZE:
      return PREPROCESS_METHOD::UP_PRIME;
   case PREPROCESS_METHOD::PRIME_MINIMIZE:
      return PREPROCESS_METHOD::PRIME;
   default:
      return preprocess;
   }
}

void
SUBool::AbsComp::PreprocessUPPrime()
{
   logs.TLog(2) << "Making CNF prime w.r.t. UP" << std::endl;
   CPUTime cpu_time;
   mHypothesis.MakeReprUPPrime();
   logs.TLog(2) << "Finished making CNF prime w.r.t. UP in time=" << cpu_time
                << std::endl;
}

void
SUBool::AbsComp::PreprocessPrime()
{
   logs.TLog(2) << "Making CNF prime" << std::endl;
   CPUTime cpu_time;
   mHypothesis.MakeReprPrime();
   // mHypothesis.ModifyRepresentation([](prime_cnf);
   logs.TLog(2) << "Finished making CNF prime in time=" << cpu_time
                << std::endl;
}

SUBool::AnnotatedCNF
SUBool::AbsComp::CurrentCNF() const
{
   if (CheckState(UNSAT))
   {
      return {CNF(Clause::EmptyClause()), Annotation()};
   }
   return {mHypothesis.GetHypothesis(true), Annotation()};
}

void
SUBool::AbsComp::Abort(const std::string &msg)
{
   std::lock_guard<std::mutex> lock(mStateMutex);
   if (LockedOver())
   {
      logs.TLog(2) << "Compilation is not running, cannot abort" << std::endl;
      return;
   }
   logs.TLog(1) << "Compilation " << msg << " during phase " << *mState
                << std::endl;
   mState = &ABORTED;
   CheckerSAT().Interrupt();
   SolverPool::DefaultPool().InterruptAll();
}

std::ostream &
SUBool::operator<<(std::ostream &os, const AbsComp::StateValue &sv)
{
   return os << sv.name;
}
