/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file abscomp.h
 * Abstract compiler. Common properties of a compiler into
 * PC or UC formula.
 */

#ifndef __ABSCOMP_H
#define __ABSCOMP_H

#include <mutex>

#include "cnfcompress.h"
#include "comp_hypothesis.h"
#include "cpuutil.h"
#include "drenc.h"
#include "enum.h"
#include "normalform.h"

namespace SUBool
{
   class Abortable
   {
    public:
      virtual ~Abortable() = default;
      virtual void Abort(const std::string &msg) = 0;
   };

   class AbsComp : public Abortable
   {
    public:
      enum class RESULT
      {
         NOT_RUN,
         COMPILED,
         UNFINISHED,
         EMP_FOUND,
         EMP_NOT_FOUND,
         ERROR,
         TIMED_OUT,
         TERMINATED,
         FAILED
      };

      enum class PREPROCESS_METHOD
      {
         NONE,
         MINIMIZE,
         UP_PRIME_MINIMIZE,
         PRIME_MINIMIZE,
         UP_PRIME,
         PRIME,
         Last = PRIME
      };

      static const EnumAnnotation<PREPROCESS_METHOD>
          kPreprocessMethodAnnotation;
      static PREPROCESS_METHOD PMRemoveMinimize(PREPROCESS_METHOD preprocess);

      struct StateValue
      {
         const char *name;
         RESULT result;
      };

      struct BaseConfig
      {
         unsigned minimize_steps{50}; // 0 means no minimization
         bool final_minimize{true};   // minimize at the end of compilation
         PREPROCESS_METHOD preprocess{
             PREPROCESS_METHOD::PRIME_MINIMIZE}; // Preprocessing method
         bool propagate_backbones{true};         // If the backbones should be
                                         // propagated before preprocessing
         MINIMIZE_METHOD minimize_method{MINIMIZE_METHOD::FULL};
         AbsEncoder::AbsConfig enc_conf{};
         AbsCheckerSAT::Timeouts timeouts{};
         bool level_minimize{false};
         LIT_IMPL_SYSTEM impl_system{LIT_IMPL_SYSTEM::Default};
         SOLVER_TYPE solver_type{SOLVER_TYPE::GLUCOSE};
         IMPL_SYSTEM_REBUILD impl_system_rebuild{
             IMPL_SYSTEM_REBUILD::CLAUSE_ADDED};
      };

    private:
      mutable std::mutex mStateMutex;
      const StateValue *mState{&INITIALIZED};

      void PropagateBackbones();
      void Preprocess();
      void PreprocessUPPrime();
      void PreprocessPrime();

      bool LockedOver() const;

    protected:
      static const StateValue INITIALIZED;
      static const StateValue PROPAGATE_BACKBONES;
      static const StateValue PREPROCESS;
      static const StateValue FINISHED;
      static const StateValue ABORTED;
      static const StateValue ERROR;
      static const StateValue UNSAT;
      static const StateValue MINIMIZE;
      static const StateValue SOLVER_INTERRUPTED;
      static const StateValue TIMED_OUT;

      bool Over() const;
      bool CheckState(const StateValue &status) const;
      bool State(const StateValue &status);
      void DumpState(std::ostream &os) const;
      RESULT StateResult() const;

      void PreCompile();
      virtual void CompilePreprocessed() = 0;
      virtual AbsCheckerSAT &CheckerSAT() = 0;
      virtual const AbsCheckerSAT &CheckerSAT() const = 0;

      RealTime mRTime;
      CPUTime mCPUTime;

      CompHypothesis mHypothesis; // Compressed clauses

      virtual std::string ResultDesc() const = 0;
      virtual std::string Annotation() const = 0;
      virtual const BaseConfig &GetBaseConfig() const = 0;

    public:
      AbsComp(CNF cnf, MINIMIZE_METHOD minimize_method,
          IMPL_SYSTEM_REBUILD is_rebuild);
      virtual ~AbsComp() = default;

      AbsComp(const AbsComp &other) = delete;
      AbsComp(AbsComp &&other) = delete;
      AbsComp &operator=(const AbsComp &other) = delete;
      AbsComp &operator=(AbsComp &&other) = delete;

      RESULT Compile();
      AnnotatedCNF CurrentCNF() const;

      virtual void Abort(const std::string &msg) override;
      virtual void PrintStatistics(unsigned level) const = 0;
   };

   std::ostream &operator<<(std::ostream &os, const AbsComp::StateValue &sv);
} // namespace SUBool

#endif
