/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file absenc.cpp
 * Abstract class describing interface of an encoding of checking if
 * the given CNF admits an empowering implicate with respect to
 * propagation completeness or unit refutation completeness.
 */

#include "absenc.h"
#include "drenc.h"
#include "logstream.h"
#include "prgutil.h"
#include "up_impl_level.h"
#include "up_log.h"
#include "up_max.h"
#include "up_min_lq.h"
#include "up_onelevel.h"
#include "up_quad.h"
#include "up_quad_impl.h"
#include "up_scc_bin_direct.h"
#include "up_scc_binary.h"
#include "up_scc_unary.h"

const SUBool::EnumAnnotation<SUBool::AbsEncoder::UP_ONE_PROVE>
    SUBool::AbsEncoder::kUPOneProveAnnotation("UP_ONE_PROVE",
        {//
            {UP_ONE_PROVE::QUADRATIC_EQUIV, "quadratic_equiv",
                "quadratic encoding with equivalences."}, //
            {UP_ONE_PROVE::QUADRATIC_ONE_SIDED, "quadratic_one_sided",
                "quadratic encoding with implications."}, //
            {UP_ONE_PROVE::LOGARITHMIC, "logarithmic",
                "based on encoding levels of unit propagations with binary "
                "encoding (logarithmic size)"}, //
            {UP_ONE_PROVE::MAX_LEVEL, "max_level",
                "encoding based on determining a maximum level of a literal in "
                "a clause"}, //
            {UP_ONE_PROVE::MIN_LQ, "min_lq",
                "each time, pick the smaller one from the quadratic encoding "
                "with implications (quadratic_one_sided) and the logarithmic "
                "encoding (logarithmic)"}, //
            {UP_ONE_PROVE::QUAD_IMPL_SYSTEM, "quad_impl_system",
                "quadratic encoding which uses an implication system to "
                "represent the dual rail encoding."}, //
            {UP_ONE_PROVE::IMPL_LEVEL, "impl_level",
                "encoding based on an implication system where levels are "
                "encoded in binary."}, //
            {UP_ONE_PROVE::SCC_UNARY, "scc_unary",
                "encoding based on the strongly connected components (SCC) of "
                "an implication system with unary encoding of levels."}, //
            {UP_ONE_PROVE::SCC_BINARY, "scc_binary",
                "encoding based on the strongly connected components (SCC) of "
                "an implication system with binary encoding of levels. Using "
                "level variables for each implication with a main SCC."}, //
            {UP_ONE_PROVE::SCC_BINARY_DIRECT, "scc_binary_direct",
                "encoding based on the strongly connected components (SCC) of "
                "an implication system with binary encoding of levels. Using "
                "direct inequalities between the levels in the body and tail "
                "of each implication in the main SCC."}} //
    );

SUBool::AbsEncoder::AbsEncoder(const AbsConfig &conf)
    : mVarStore(conf.use_backmap), mClauses(), mOneProveEncoder(),
      mCnfMaxVariable(0), mMaxUPLevelBound(0)
{
}

void
SUBool::AbsEncoder::BuildEncoding(const CNF &cnf,
    const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider> impl_system_provider)
{
   logs.TLog(4) << "Building encoding for (p cnf " << cnf.MaxVariable() << " "
                << cnf.Size() << ")" << std::endl;
   mCnfMaxVariable = cnf.MaxVariable();
   InitOneProveEncoder(cnf, cl2lit, impl_system_provider);
   AddClauses(cnf, cl2lit, impl_system_provider);
   mOneProveEncoder = nullptr;
}

unsigned
SUBool::AbsEncoder::MaxVariable() const
{
   return mVarStore.LastVariable();
}

unsigned
SUBool::AbsEncoder::Size() const
{
   return mClauses.size();
}

bool
SUBool::AbsEncoder::RequiresImplSystem() const
{
   return AbsConfig::OneProveRequiresImplSystem(GetConfig().up_one_prove);
}

bool
SUBool::AbsEncoder::RequiresSCCGraph() const
{
   return AbsConfig::OneProveRequiresSCCGraph(GetConfig().up_one_prove);
}

void
SUBool::AbsEncoder::InitOneProveEncoder(const CNF &cnf,
    const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider> impl_system_provider)
{
   auto &conf = GetConfig();
   switch (conf.up_one_prove)
   {
   case UP_ONE_PROVE::QUADRATIC_EQUIV:
      mOneProveEncoder = std::unique_ptr<UPEncQuad>(
          new UPEncQuad(UPEncQuad::BOTH, cnf, cl2lit, mVarStore));
      break;
   case UP_ONE_PROVE::QUADRATIC_ONE_SIDED:
      mOneProveEncoder = std::unique_ptr<UPEncQuad>(
          new UPEncQuad(UPEncQuad::IF, cnf, cl2lit, mVarStore));
      break;
   case UP_ONE_PROVE::LOGARITHMIC:
      mOneProveEncoder = std::unique_ptr<UPEncLogarithmic>(
          new UPEncLogarithmic(conf.log_conf, cnf, cl2lit, mVarStore));
      break;
   case UP_ONE_PROVE::MAX_LEVEL:
      mOneProveEncoder = std::unique_ptr<UPEncMaxLevel>(
          new UPEncMaxLevel(cnf, cl2lit, mVarStore));
      break;
   case UP_ONE_PROVE::MIN_LQ:
      mOneProveEncoder = std::unique_ptr<UPEncMinLQ>(
          new UPEncMinLQ(cnf, cl2lit, mVarStore, conf.log_conf,
              conf.min_lq_log_weights, conf.min_lq_quad_weights));
      break;
   case UP_ONE_PROVE::QUAD_IMPL_SYSTEM:
      mOneProveEncoder = std::make_unique<UPEncQuadImpl>(
          cnf, cl2lit, mVarStore, impl_system_provider->OneProveImplSystem());
      break;
   case UP_ONE_PROVE::IMPL_LEVEL:
      mOneProveEncoder = std::make_unique<UPEncImplLevel>(
          UPEncImplLevel::Config(conf.log_conf.use_bit_level), cnf, cl2lit,
          mVarStore, impl_system_provider->OneProveImplSystem());
      break;
   case UP_ONE_PROVE::SCC_UNARY:
      mOneProveEncoder = std::make_unique<UPSCCUnary>(cnf, cl2lit, mVarStore,
          impl_system_provider->OneProveImplSystem(),
          impl_system_provider->SCCGraph());
      break;
   case UP_ONE_PROVE::SCC_BINARY:
      mOneProveEncoder = std::make_unique<UPSCCBinary>(
          UPSCCBinary::Config(conf.log_conf.use_bit_level), cnf, cl2lit,
          mVarStore, impl_system_provider->OneProveImplSystem(),
          impl_system_provider->SCCGraph());
      break;
   case UP_ONE_PROVE::SCC_BINARY_DIRECT:
      mOneProveEncoder = std::make_unique<UPSCCBinaryDirect>(
          UPSCCBinaryDirect::Config(conf.log_conf.use_bit_level), cnf, cl2lit,
          mVarStore, impl_system_provider->OneProveImplSystem(),
          impl_system_provider->SCCGraph());
      break;
   default:
      throw BadParameterException("AbsEncoder::InitOneProveEncoder",
          "Unknown UP encoding type for 1-provability");
   }
   mMaxUPLevelBound = mOneProveEncoder->MaxLevelBound();
}

unsigned
SUBool::AbsEncoder::NextUPLevelBound(const CNF &cnf,
    const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider> impl_system_provider, unsigned level)
{
   InitOneProveEncoder(cnf, cl2lit, impl_system_provider);
   return mOneProveEncoder->NextLevelBound(level);
   mOneProveEncoder = nullptr;
}

void
SUBool::AbsEncoder::WriteEncoding(std::ostream &out) const
{
   out << "c " << Name() << std::endl;
   for (unsigned i = 1; i <= mCnfMaxVariable; ++i)
   {
      WriteMappingsOfVariable(out, i);
   }
   WriteHeader(out);
   for (const auto &clause : mClauses)
   {
      WriteClause(out, clause);
   }
}

void
SUBool::AbsEncoder::WriteHeader(std::ostream &out) const
{
   out << "p cnf " << mVarStore.LastVariable() << " " << mClauses.size()
       << std::endl;
}

void
SUBool::AbsEncoder::WriteMappingsOfVariable(std::ostream &out, unsigned i) const
{
   if (i == 0)
   {
      throw OutOfBoundsException(
          "AbsEncoder::WriteMappingsOfVariable", "Variable i is 0");
   }
   out << "c Mappings of variable " << i << std::endl;
}

void
SUBool::AbsEncoder::WriteClause(std::ostream &out, const Clause &clause) const
{
   mVarStore.WriteClauseComment(out, clause);
   out << clause << std::endl;
}

unsigned
SUBool::AbsEncoder::MaxUPLevelBound() const
{
   return mMaxUPLevelBound;
}

bool
SUBool::AbsEncoder::PassToSolver(SolverInterface &solver) const
{
   return solver.AddClauses(mClauses);
}

SUBool::CNF
SUBool::AbsEncoder::Encoding() const
{
   return mClauses;
}

std::ostream &
SUBool::operator<<(std::ostream &out, const AbsEncoder &enc)
{
   enc.WriteEncoding(out);
   return out;
}

void
SUBool::AbsEncoder::AbsConfig::Dump(
    const std::string &line_prefix, unsigned level) const
{
   kUPOneProveAnnotation.DumpValue(
       line_prefix, level, "up_one_prove", up_one_prove);
   logs.DumpValue(line_prefix, level, "optimize", optimize);
   logs.DumpValue(line_prefix, level, "up_level_bound", up_level_bound);
   logs.DumpValue(line_prefix, level, "use_backmap", use_backmap);
   logs.TLog(level) << line_prefix << "log_conf:" << std::endl;
   log_conf.Dump(line_prefix + "\t", level);
   logs.DumpValue(line_prefix, level, "min_lq_log_weights", min_lq_log_weights);
   logs.DumpValue(
       line_prefix, level, "min_lq_quad_weights", min_lq_quad_weights);
}

