/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file absenc.h
 * Abstract class describing interface of an encoding of checking if
 * the given CNF admits an empowering implicate with respect to
 * propagation completeness or unit refutation completeness.
 */

#ifndef __ABSENC_H
#define __ABSENC_H

#include <iostream>
#include <memory>

#include "cardinal.h"
#include "cl2lit.h"
#include "enc_is_provider.h"
#include "enum.h"
#include "glucose_bind.h"
#include "impl_system.h"
#include "impl_system_graph.h"
#include "normalform.h"
#include "up_min_lq.h"
#include "upenc.h"
#include "varstore.h"

namespace SUBool
{
   class AbsEncoder
   {
    public:
      enum class UP_ONE_PROVE
      {
         QUADRATIC_EQUIV,
         QUADRATIC_ONE_SIDED,
         LOGARITHMIC,
         MAX_LEVEL,
         MIN_LQ, // The smaller one from QUADRATIC_ONE_SIDED and
                 // LOGARITHMIC, levels have to be synchronized.
         QUAD_IMPL_SYSTEM,
         IMPL_LEVEL,
         SCC_UNARY,
         SCC_BINARY,
         SCC_BINARY_DIRECT
      };

      struct UPOneProveTraits
      {
         UP_ONE_PROVE one_prove;
         bool requires_impl_system;
         bool requires_scc_graph;
      };

      inline static const std::array kUPOneProveTraits{
          UPOneProveTraits{UP_ONE_PROVE::QUADRATIC_EQUIV, false, false},
          UPOneProveTraits{UP_ONE_PROVE::QUADRATIC_ONE_SIDED, false, false},
          UPOneProveTraits{UP_ONE_PROVE::LOGARITHMIC, false, false},
          UPOneProveTraits{UP_ONE_PROVE::MAX_LEVEL, false, false},
          UPOneProveTraits{UP_ONE_PROVE::MIN_LQ, false, false},
          UPOneProveTraits{UP_ONE_PROVE::QUAD_IMPL_SYSTEM, true, false},
          UPOneProveTraits{UP_ONE_PROVE::IMPL_LEVEL, true, false},
          UPOneProveTraits{UP_ONE_PROVE::SCC_UNARY, true, true},
          UPOneProveTraits{UP_ONE_PROVE::SCC_BINARY, true, true},
          UPOneProveTraits{UP_ONE_PROVE::SCC_BINARY_DIRECT, true, true}};

      static const EnumAnnotation<UP_ONE_PROVE> kUPOneProveAnnotation;

      static UP_ONE_PROVE ParseUPOneProve(const std::string &up_one_prove);

      struct AbsConfig
      {
         UP_ONE_PROVE up_one_prove{UP_ONE_PROVE::SCC_UNARY};
         bool optimize{false};

         // Maximum number of a level in an
         // encoding unit propagation for 1-provability.
         // If 0, then max level=#vars
         // In case of logarithmic encoding, this is the number of
         // bits of level variables.
         unsigned up_level_bound{0};
         bool use_backmap{false};
         UPEncLogarithmic::Config log_conf;
         UPEncMinLQ::SizeWeights min_lq_log_weights{2.0, 2.0, 0.0};
         UPEncMinLQ::SizeWeights min_lq_quad_weights{1.0, 1.0, 0.0};

         AbsConfig() noexcept : log_conf() {}
         virtual ~AbsConfig() = default;
         AbsConfig(const AbsConfig &conf) = default;
         AbsConfig(AbsConfig &&conf) = default;
         AbsConfig &operator=(const AbsConfig &conf) = default;
         AbsConfig &operator=(AbsConfig &&conf) = default;
         virtual void Dump(
             const std::string &line_prefix, unsigned level) const;
         static constexpr bool
         OneProveRequiresImplSystem(UP_ONE_PROVE op)
         {
            return kUPOneProveTraits.at(static_cast<unsigned>(op))
                .requires_impl_system;
         }
         static constexpr bool
         OneProveRequiresSCCGraph(UP_ONE_PROVE op)
         {
            return AbsEncoder::kUPOneProveTraits.at(static_cast<unsigned>(op))
                .requires_scc_graph;
         }
      };

    protected:
      VarStore mVarStore;
      std::vector<Clause> mClauses;
      std::unique_ptr<UPEncoder> mOneProveEncoder;
      unsigned mCnfMaxVariable;
      unsigned mMaxUPLevelBound;

      virtual void InitOneProveEncoder(const CNF &cnf,
          const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider);

      virtual void WriteHeader(std::ostream &out) const;
      virtual std::string Name() const = 0;
      virtual void WriteMappingsOfVariable(std::ostream &out, unsigned i) const;
      virtual void WriteClause(std::ostream &out, const Clause &clause) const;
      virtual void AddClauses(const CNF &cnf, const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider) = 0;
      virtual const AbsConfig &GetConfig() const = 0;

    public:
      AbsEncoder(const AbsConfig &conf = AbsConfig());
      virtual ~AbsEncoder() = default;
      virtual void BuildEncoding(const CNF &cnf,
          const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider);

      virtual unsigned MaxVariable() const;
      virtual unsigned Size() const;

      virtual bool RequiresImplSystem() const;
      virtual bool RequiresSCCGraph() const;
      virtual unsigned MaxUPLevelBound() const;
      virtual unsigned NextUPLevelBound(const CNF &cnf,
          const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider,
          unsigned level);
      // The meaning of the return value depends on descendant,
      // in any case it returns a witness variable (e.g. empowered)
      virtual unsigned ExtractClause(
          const SolverInterface &solver, Clause *clause) const = 0;

      virtual bool PassToSolver(SolverInterface &solver) const;
      virtual void WriteEncoding(std::ostream &out) const;
      virtual CNF Encoding() const;
   };

   std::ostream &operator<<(std::ostream &out, const AbsEncoder &enc);
} // namespace SUBool

#endif
