/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file absinccomp.cpp
 * Abstract iterative compiler. Common properties of an iterative compiler into
 * PC or UC formula.
 */
#include <chrono>
#include <ctime>

#include "absinccomp.h"
#include "cnfutil.h"
#include "empower.h"
#include "logstream.h"
#include "unitprop.h"

const SUBool::AbsComp::StateValue SUBool::AbsIncrementalComp::FIND_EMPOWERING{
    "FIND_EMPOWERING", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::AbsIncrementalComp::INPROCESSING{
    "INPROCESSING", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::AbsIncrementalComp::PROCESS_LEARNED{
    "PROCESS_LEARNED", SUBool::AbsComp::RESULT::UNFINISHED};

SUBool::AbsIncrementalComp::AbsIncrementalComp(
    CNF cnf, MINIMIZE_METHOD minimize_method, IMPL_SYSTEM_REBUILD is_rebuild)
    : AbsComp(cnf, minimize_method, is_rebuild), mCntEmpowering(0),
      mCntEmpoweringSAT(0), mCntLearned(0), mCntLearnedEmpowering(0),
      mCntAdded(0), mLastMinimizationCnt(0), mCPUTime(), mRTime()
{
}

void
SUBool::AbsIncrementalComp::CompilePreprocessed()
{
   logs.TLog(2) << "The main cycle of compilation starts" << std::endl;
   if (CheckerSAT().RequiresImplSystem())
   {
      mHypothesis.SetImplSystemProvider(
          make_enc_is_provider(GetBaseConfig().impl_system));
   }
   else
   {
      mHypothesis.SetImplSystemProvider(std::make_shared<DummyISProvider>());
   }
   mHypothesis.SetHypothesisToRepresentation();
   while (!Over())
   {
      ProcessLearned();
      // Use learned clauses from the calls to prime subimplicate
      auto emp = FindEmpowering();
      if (emp)
      {
         ++mCntEmpoweringSAT;
         auto emp_var = CheckerSAT().LastWitnessVariable();
         AddEmpowering(*emp, emp_var);
      }
      else if (CheckerSAT().TimedOut())
      {
         State(TIMED_OUT);
         break;
      }
      MinimizeHypothesis();
      if (!emp && !CheckerSAT().WillIncreaseLevel())
      {
         State(FINISHED);
      }
   }
   if (!CheckState(TIMED_OUT) && GetBaseConfig().final_minimize)
   {
      mHypothesis.MinimizeHypothesis();
   }
   PrintStatistics(1);
}

std::optional<SUBool::Clause>
SUBool::AbsIncrementalComp::FindEmpowering()
{
   CPUTime cpu_time;
   mTimeTillTheLastSATEqCheck = mCPUTime.TimeSpan();
   auto emp =
       mHypothesis.FindEmpowering(CheckerSAT(), GetBaseConfig().level_minimize);
   if (mHypothesis.HypMinimized())
   {
      mLastMinimizationCnt = mCntAdded;
   }
   mTimeOfLastSATEqCheck = cpu_time.TimeSpan();
   return emp;
}

std::string
SUBool::AbsIncrementalComp::ResultDesc() const
{
   switch (StateResult())
   {
   case RESULT::NOT_RUN:
   case RESULT::COMPILED:
      return "Compilation finished successfully, formula is "
             + PreambleResultFormulaKind();
      break;
   case RESULT::UNFINISHED:
      return "Compilation in progress";
      break;
   case RESULT::EMP_FOUND:
      return "An empowering implicate was found, formula is not "
             + PreambleResultFormulaKind();
      break;
   case RESULT::EMP_NOT_FOUND:
      return "An empowering implicate was not found, formula is "
             + PreambleResultFormulaKind();
      break;
   case RESULT::ERROR:
      return "Compilation failed due to an error";
      break;
   case RESULT::TIMED_OUT:
      return "Compilation timed out";
      break;
   case RESULT::TERMINATED:
      return "Compilation terminated";
      break;
   case RESULT::FAILED:
      return "Compilation failed";
      break;
   default:
      return "Unknown result";
   }
}

std::string
SUBool::AbsIncrementalComp::Annotation() const
{
   std::string s = "c Using the iterative algorithm with goal being "
                   + PreambleGoal() + "\n";
   if (CheckState(UNSAT))
   {
      s += "c The formula is UNSATISFIABLE";
      return s;
   }
   s += "c " + ResultDesc();
   return s;
}

SUBool::CNF
SUBool::AbsIncrementalComp::Inprocess(const Clause &emp, unsigned emp_var)
{
   logs.TLog(SUBool::LogStream::LOG_LEVEL::ALL)
       << "Inprocess " << emp << " (" << emp_var << ")" << std::endl;
   auto clause = mHypothesis.PrimeSubimplicate(emp, emp_var, true);
   return *clause;
}

void
SUBool::AbsIncrementalComp::AddEmpowering(const Clause &emp, unsigned emp_var)
{
   if (!State(INPROCESSING))
   {
      return;
   }
   auto emp_clauses = Inprocess(emp, emp_var);
   ++mCntEmpowering;
   logs.TLog(LogStream::LOG_LEVEL::ALL)
       << "Inprocessing finished, adding " << emp_clauses.Size() << " clauses"
       << std::endl;
   mCntAdded += emp_clauses.Size();
   for (const auto &clause : emp_clauses)
   {
      logs.Log(LogStream::LOG_LEVEL::EMPOWER)
          << mHypothesis.Compressor().UncompressClause(clause) << std::endl;
   }
   mHypothesis.Append(std::move(emp_clauses));
}

void
SUBool::AbsIncrementalComp::PrintStatistics(unsigned level) const
{
   logs.TLog(level) << ResultDesc() << std::endl;
   logs.TLog(level) << "Total time: " << mRTime << std::endl;
   logs.TLog(level) << "Total processor time: " << mCPUTime << std::endl;
   logs.TLog(level) << "Processor time until the last SAT based EQ check: "
                    << mTimeTillTheLastSATEqCheck << std::endl;
   logs.TLog(level) << "Processor time of the last SAT based EQ check: "
                    << mTimeOfLastSATEqCheck << std::endl;
   logs.TLog(level)
       << "Processor time of the last SAT solver call within an EQ check: "
       << CheckerSAT().LastSolveTime() << std::endl;
   logs.TLog(level) << "Total number of empowering clauses: " << mCntEmpowering
                    << std::endl;
   logs.TLog(level) << "Total number of added clauses: " << mCntAdded
                    << std::endl;
   logs.TLog(level) << "Total number of empowering clauses found by SAT: "
                    << mCntEmpoweringSAT << std::endl;
   logs.TLog(level) << "Total number of learned clauses used: " << mCntLearned
                    << std::endl;
   logs.TLog(level) << "Total number of learned clauses added as empowering: "
                    << mCntLearnedEmpowering << std::endl;
   logs.TLog(level) << "Total time of SAT based equivalence queries: "
                    << CheckerSAT().TotalTime() << "s" << std::endl;
   logs.TLog(level) << "SAT based equivalence with result SAT (time/count): "
                    << CheckerSAT().TotalTimeSAT() << "s / "
                    << CheckerSAT().SolveCountSAT() << std::endl;
   logs.TLog(level) << "SAT based equivalence with result UNSAT (time/count): "
                    << CheckerSAT().TotalTimeUNSAT() << "s / "
                    << CheckerSAT().SolveCountUNSAT() << std::endl;
   logs.TLog(level) << "Maximum UP level: "
                    << CheckerSAT().CurrentUPLevelBound() << std::endl;
}

void
SUBool::AbsIncrementalComp::MinimizeHypothesis()
{
   if (ShouldMinimize() && State(MINIMIZE))
   {
      mLastMinimizationCnt = mCntAdded;
      mHypothesis.MinimizeHypothesis();
   }
}

bool
SUBool::AbsIncrementalComp::ShouldMinimize()
{
   if (GetBaseConfig().level_minimize && CheckerSAT().WillIncreaseLevel())
   {
      return true;
   }
   unsigned steps = GetBaseConfig().minimize_steps;
   return (steps > 0 && (mCntAdded - mLastMinimizationCnt) >= steps);
}

void
SUBool::AbsIncrementalComp::ProcessLearned()
{
   if (!State(PROCESS_LEARNED))
   {
      return;
   }
   while (auto learned = mHypothesis.NextLearnedClause())
   {
      ++mCntLearned;
      logs.TLog(3) << "Checking learned clause (" << mHypothesis.LearnedSize()
                   << " remaining)" << std::endl;
      if (auto emp_lit = mHypothesis.EmpoweredLiteral(*learned))
      {
         ++mCntLearnedEmpowering;
         AddEmpowering(*learned, emp_lit->Variable());
      }
      else
      {
         logs.TLog(3) << "... not empowering any more" << std::endl;
      }
   }
   MinimizeHypothesis();
}

