/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file assignment.cpp
 * Assignment vector.
 */

#include <cstdlib>

#include "assign.h"
#include "random_utils.h"
#include "sbexception.h"

SUBool::Assignment::Assignment(size_t n_values, double probability_true)
    : BaseClass(n_values, false)
{
   std::discrete_distribution dd({1.0 - probability_true, probability_true});
   for (auto &&v : *this)
   {
      v = (dd(sub_rand) == 0);
   }
}

SUBool::PartialAssignment::PartialAssignment(
    size_t n_values, double probability_false, double probability_true)
    : BaseClass(n_values, PartialValue::FALSE)
{
   // double pf = probability_false + probability_true;
   std::discrete_distribution dd({1.0 - probability_false - probability_true,
       probability_false, probability_true});
   for (auto &&v : *this)
   {
      v = static_cast<PartialValue>(dd(sub_rand));
   }
}

SUBool::PartialAssignment::PartialAssignment(const std::string &s)
    : BaseClass(s.length(), PartialValue::UNDEF)
{
   for (unsigned i = 0; i < s.length(); ++i)
   {
      switch (s[i])
      {
      case '0':
         (*this)[i + 1] = PartialValue::FALSE;
         break;
      case '1':
         (*this)[i + 1] = PartialValue::TRUE;
         break;
      case '*':
         break;
      default:
         throw BadParameterException(
             "PartialAssignment::PartialAssignment(std::string)",
             "Unknown value");
      }
   }
}

SUBool::PartialAssignment::PartialAssignment(const Assignment &alpha)
    : BaseClass(alpha.Size(), PartialValue::UNDEF)
{
   for (unsigned i = 1; i <= alpha.Size(); ++i)
   {
      (*this)[i] = partial_value_of_bool(alpha[i]);
   }
}

SUBool::PartialAssignment::PartialAssignment(
    size_t n_values, std::vector<Literal> literals)
    : PartialAssignment(n_values, PartialValue::UNDEF)
{
   for (const auto &lit : literals)
   {
      SetLiteral(lit);
   }
}

void
SUBool::PartialAssignment::SetLiteral(const Literal &lit)
{
   (*this)[lit.Variable()] = partial_value_of_bool(lit.Positive());
}

void
SUBool::PartialAssignment::UnsetLiteral(const Literal &lit)
{
   (*this)[lit.Variable()] = PartialValue::UNDEF;
}

std::ostream &
SUBool::operator<<(std::ostream &out, const PartialValue &v)
{
   if (v == PartialValue::UNDEF)
   {
      out << "*";
   }
   else
   {
      out << bool_of_partial_value(v);
   }
   return out;
}

SUBool::PartialValue
SUBool::pv_negate(PartialValue pv)
{
   switch (pv)
   {
   case PartialValue::FALSE:
      return PartialValue::TRUE;
   case PartialValue::TRUE:
      return PartialValue::FALSE;
   case PartialValue::UNDEF:
      return PartialValue::UNDEF;
   }
   throw BadParameterException("operator!(PartialValue)", "Unknown value");
}

SUBool::PartialValue
SUBool::pv_and(PartialValue a, PartialValue b)
{
   switch (a)
   {
   case PartialValue::TRUE:
      return b;
   case PartialValue::FALSE:
      return PartialValue::FALSE;
   case PartialValue::UNDEF:
      return (
          b == PartialValue::FALSE ? PartialValue::FALSE : PartialValue::UNDEF);
   default:
      throw BadParameterException(
          "operator&(PartialValue, PartialValue)", "unknown value of a");
   }
}

SUBool::PartialValue
SUBool::pv_or(PartialValue a, PartialValue b)
{
   switch (a)
   {
   case PartialValue::TRUE:
      return PartialValue::TRUE;
   case PartialValue::FALSE:
      return b;
   case PartialValue::UNDEF:
      return (
          b == PartialValue::TRUE ? PartialValue::TRUE : PartialValue::UNDEF);
   default:
      throw BadParameterException(
          "operator|(PartialValue, PartialValue)", "unknown value of a");
   }
}

bool
SUBool::next_value(bool val)
{
   if (val)
   {
      throw OverflowException("next_value(bool)", "TRUE cannot be increased");
   }
   return true;
}

SUBool::PartialValue
SUBool::next_value(PartialValue val)
{
   switch (val)
   {
   case PartialValue::UNDEF:
      return PartialValue::FALSE;
   case PartialValue::FALSE:
      return PartialValue::TRUE;
   case PartialValue::TRUE:
      throw OverflowException(
          "next_value(PartialValue)", "TRUE cannot be increased");
   default:
      throw InvariantFailedException(
          "next_value(PartialValue)", "Invalid value");
   }
}

bool
SUBool::pa_equal(const PartialAssignment &alpha, const PartialAssignment &beta)
{
   size_t m = std::max(alpha.Size(), beta.Size());
   for (size_t i = 0; i < m; ++i)
   {
      if (alpha.XValue(i) != beta.XValue(i))
      {
         return false;
      }
   }
   return true;
}

