/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file assign.h
 * Assignment vector.
 */

#ifndef __ASSIGN_H
#define __ASSIGN_H

#include <iostream>
#include <vector>

#include "literal.h"

namespace SUBool
{
   /**@brief Assignment is basically a vector of values, but indexed
    * from 1. It is parameterized by the type of values it should
    * store.
    */
   template <class ValueType, ValueType MinValue, ValueType MaxValue,
       ValueType (*NextValue)(ValueType)>
   class AssignTemplate
   {
    public:
      using const_iterator = typename std::vector<ValueType>::const_iterator;
      using iterator = typename std::vector<ValueType>::iterator;
      using reference = typename std::vector<ValueType>::reference;
      using const_reference = typename std::vector<ValueType>::const_reference;

    private:
      std::vector<ValueType> mValues;

    public:
      AssignTemplate(size_t n_values, ValueType default_value)
          : mValues(n_values, default_value)
      {
      }
      AssignTemplate(std::vector<ValueType> values) : mValues(std::move(values))
      {
      }
      AssignTemplate(std::initializer_list<ValueType> values)
          : mValues(std::move(values))
      {
      }

      const_reference
      operator[](size_t index) const
      {
         return mValues[index - 1];
      }

      reference
      operator[](size_t index)
      {
         return mValues[index - 1];
      }

      size_t
      Size() const
      {
         return mValues.size();
      }

      void
      Resize(size_t n_values, ValueType default_value)
      {
         mValues.resize(n_values, default_value);
      }

      // Return true if inc successful, false if already max.
      bool Inc();
      void Reset(ValueType val = MinValue);

      iterator
      begin()
      {
         return mValues.begin();
      }
      iterator
      end()
      {
         return mValues.end();
      }
      const_iterator
      begin() const
      {
         return mValues.begin();
      }
      const_iterator
      end() const
      {
         return mValues.end();
      }

      friend std::ostream &
      operator<<(std::ostream &out,
          const AssignTemplate<ValueType, MinValue, MaxValue, NextValue> &alpha)
      {
         for (auto val : alpha.mValues)
         {
            out << val;
         }
         return out;
      }

      bool
      operator==(
          const AssignTemplate<ValueType, MinValue, MaxValue, NextValue> &other)
          const
      {
         return mValues == other.mValues;
      }
      bool
      operator!=(
          const AssignTemplate<ValueType, MinValue, MaxValue, NextValue> &other)
          const
      {
         return mValues != other.mValues;
      }
      bool
      operator<(
          const AssignTemplate<ValueType, MinValue, MaxValue, NextValue> &other)
          const
      {
         return mValues < other.mValues;
      }
      bool
      operator>(
          const AssignTemplate<ValueType, MinValue, MaxValue, NextValue> &other)
          const
      {
         return mValues > other.mValues;
      }
      bool
      operator>=(
          const AssignTemplate<ValueType, MinValue, MaxValue, NextValue> &other)
          const
      {
         return mValues >= other.mValues;
      }
      bool
      operator<=(
          const AssignTemplate<ValueType, MinValue, MaxValue, NextValue> &other)
          const
      {
         return mValues <= other.mValues;
      }
   };

   enum class PartialValue
   {
      UNDEF,
      FALSE,
      TRUE
   };

   bool next_value(bool val);

   class Assignment : public AssignTemplate<bool, false, true, next_value>
   {
      using BaseClass = AssignTemplate<bool, false, true, next_value>;

    public:
      Assignment(size_t n_values, bool default_value = false)
          : BaseClass(n_values, default_value)
      {
      }
      Assignment(std::vector<bool> values) : BaseClass(std::move(values)) {}
      Assignment(std::initializer_list<bool> values)
          : BaseClass(std::move(values))
      {
      }
      Assignment(size_t n_values, double probability_true);
   };

   PartialValue next_value(PartialValue val);

   class PartialAssignment
       : public AssignTemplate<PartialValue, PartialValue::UNDEF,
             PartialValue::TRUE, next_value>
   {
      using BaseClass = AssignTemplate<PartialValue, PartialValue::UNDEF,
          PartialValue::TRUE, next_value>;

    public:
      PartialAssignment(
          size_t n_values, PartialValue default_value = PartialValue::UNDEF)
          : BaseClass(n_values, default_value)
      {
      }
      PartialAssignment(std::vector<PartialValue> values)
          : BaseClass(std::move(values))
      {
      }
      PartialAssignment(std::initializer_list<PartialValue> values)
          : BaseClass(std::move(values))
      {
      }

      PartialAssignment(const Assignment &alpha);
      PartialAssignment(
          size_t n_values, double probability_false, double probability_true);
      explicit PartialAssignment(const std::string &s);
      PartialAssignment(size_t n_values, std::vector<Literal> literals);

      void SetLiteral(const Literal &lit);
      void UnsetLiteral(const Literal &lit);
      template <class InputIterator>
      void SetLiterals(InputIterator begin, InputIterator end);
      template <class InputIterator>
      void SetLiteralsNegated(InputIterator begin, InputIterator end);
      template <class InputIterator>
      void UnsetLiterals(InputIterator begin, InputIterator end);

      PartialValue
      XValue(size_t index) const
      {
         return index < Size() ? (*this)[index] : PartialValue::UNDEF;
      }
   };

   // Differs from the comparison operators in that if the lengths do not agree,
   // then the shorter assignment is considered to be padded with undefs.
   bool pa_equal(const PartialAssignment &alpha, const PartialAssignment &beta);

   std::ostream &operator<<(std::ostream &out, const PartialValue &v);

   PartialValue pv_negate(PartialValue pv);
   PartialValue pv_and(PartialValue a, PartialValue b);
   PartialValue pv_or(PartialValue a, PartialValue b);

   inline PartialValue
   operator!(PartialValue pv)
   {
      return pv_negate(pv);
   }
   inline PartialValue
   operator&(PartialValue a, PartialValue b)
   {
      return pv_and(a, b);
   }
   inline PartialValue
   operator|(PartialValue a, PartialValue b)
   {
      return pv_or(a, b);
   }

   constexpr PartialValue
   partial_value_of_bool(bool v)
   {
      return (v ? PartialValue::TRUE : PartialValue::FALSE);
   }

   constexpr bool
   bool_of_partial_value(PartialValue v, bool default_value = false)
   {
      return (v == PartialValue::TRUE    ? true
              : v == PartialValue::FALSE ? false
                                         : default_value);
   }
} // namespace SUBool

// Implementation
template <class ValueType, ValueType MinValue, ValueType MaxValue,
    ValueType (*NextValue)(ValueType)>
bool
SUBool::AssignTemplate<ValueType, MinValue, MaxValue, NextValue>::Inc()
{
   auto v = mValues.rbegin();
   while (v != mValues.rend() && *v == MaxValue)
   {
      *(v++) = MinValue;
   }
   if (v == mValues.rend())
   {
      return false;
   }
   *v = NextValue(*v);
   return true;
}

template <class ValueType, ValueType MinValue, ValueType MaxValue,
    ValueType (*NextValue)(ValueType)>
void
SUBool::AssignTemplate<ValueType, MinValue, MaxValue, NextValue>::Reset(
    ValueType val)
{
   mValues.assign(mValues.size(), val);
}

template <class InputIterator>
void
SUBool::PartialAssignment::SetLiterals(InputIterator begin, InputIterator end)
{
   for (auto it = begin; it != end; ++it)
   {
      SetLiteral(*it);
   }
}

template <class InputIterator>
void
SUBool::PartialAssignment::SetLiteralsNegated(
    InputIterator begin, InputIterator end)
{
   for (auto it = begin; it != end; ++it)
   {
      SetLiteral(it->Negation());
   }
}

template <class InputIterator>
void
SUBool::PartialAssignment::UnsetLiterals(InputIterator begin, InputIterator end)
{
   for (auto it = begin; it != end; ++it)
   {
      (*this)[it->Variable()] = PartialValue::UNDEF;
   }
}

#endif
