/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file canon.cpp
 * Constructing canonical CNF.
 */

#include "canon.h"
#include "cnfutil.h"
#include "empower.h"
#include "resolve.h"

SUBool::CanonBase::CanonBase(CNF cnf)
    : mOpen(), mClosed(), mInput(std::move(cnf))
{
   for (auto &clause : mInput)
   {
      mOpen.push(std::move(clause));
   }
}

void
SUBool::CanonBase::Compute()
{
   while (!mOpen.empty())
   {
      if (SubsumedByClosed(mOpen.front()))
      {
         mOpen.pop();
         continue;
      }
      for (const auto &closed : mClosed)
      {
         auto resolvent = resolve(mOpen.front(), closed);
         if (resolvent.second == 0)
         {
            continue;
         }
         ++mResolventCount;
         mOpen.push(std::move(resolvent.first));
      }
      mClosed.push_back(mOpen.front());
      mOpen.pop();
   }
}

bool
SUBool::CanonBase::SubsumedByClosed(const Clause &clause)
{
   auto i_closed = mClosed.begin();
   while (i_closed != mClosed.end())
   {
      auto res = i_closed->SubsetCompare(clause);
      switch (res)
      {
      case SetCompareResult::SUPERSET:
      {
         auto i_del = i_closed;
         ++i_closed;
         mClosed.erase(i_del);
         break;
      }
      case SetCompareResult::SUBSET:
      case SetCompareResult::EQUAL:
         return true;
      case SetCompareResult::INCOMPARABLE:
         ++i_closed;
         break;
      }
   }
   return false;
}

SUBool::CNF
SUBool::CanonBase::Cnf() const
{
   return CNF(mClosed.begin(), mClosed.end());
}

void
SUBool::AbsCanon::PrintStatistics(unsigned level) const
{
   logs.TLog(level) << "Total CPU time: " << mCPUTime << std::endl;
   logs.TLog(level) << "Number of resolvents: " << mResolventCount << std::endl;
}

SUBool::CanonSAT::CanonSAT(CNF cnf)
    : mOpen(), mClosed(), mInput(std::move(cnf)), mSolver(mInput)
{
}

void
SUBool::CanonSAT::Compute()
{
   CPUTime cpu_time;
   if (!mSolver.Solver().Solve())
   {
      logs.TLog(3) << "The formula is unsatisfiable" << std::endl;
      mClosed.push_back(Clause::EmptyClause());
      return;
   }
   for (auto &clause : prime_cnf(mSolver.Solver(), mInput))
   {
      mOpen.push(std::move(clause));
   }
   mPreprocTime = cpu_time.TimeSpan();
   while (!mOpen.empty())
   {
      if (std::find(mClosed.cbegin(), mClosed.cend(), mOpen.front())
          != mClosed.cend())
      {
         mOpen.pop();
         continue;
      }
      for (const auto &closed : mClosed)
      {
         auto resolvent = resolve(mOpen.front(), closed);
         if (resolvent.second == 0)
         {
            continue;
         }
         ++mResolventCount;
         cpu_time.Reset();
         mOpen.push(
             *prime_subimplicate(mSolver.Solver(), resolvent.first, 0, true));
         mInprocTime += cpu_time.TimeSpan();
      }
      mClosed.push_back(std::move(mOpen.front()));
      mOpen.pop();
      if (mClosed.size() % kClosedSizeLogStep == 0
          || (mResolventCount > 0
              && mResolventCount % kResolventCountLogStep == 0))
      {
         logs.TLog(4) << "closed/resolvents: " << mClosed.size() << "/"
                      << mResolventCount << std::endl;
      }
   }
}

SUBool::CNF
SUBool::CanonSAT::Cnf() const
{
   return CNF(mClosed.begin(), mClosed.end());
}

void
SUBool::CanonSAT::PrintStatistics(unsigned level) const
{
   AbsCanon::PrintStatistics(level);
   SUBool::logs.TLog(level)
       << "Preprocessing time: " << mPreprocTime << "s" << std::endl;
   SUBool::logs.TLog(level)
       << "Total inprocessing time: " << mInprocTime << "s" << std::endl;
}

