/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file cardinal.h
 * Encodings of cardinality constraints.
 */

#include "cardinal.h"

SUBool::AmoEncoder::KIND SUBool::AmoEncoder::sGlobalKind =
    SUBool::AmoEncoder::kDefaultGlobalKind;
SUBool::ExOneEncoder::KIND SUBool::ExOneEncoder::sGlobalKind =
    SUBool::ExOneEncoder::kDefaultGlobalKind;

const SUBool::EnumAnnotation<SUBool::AmoEncoder::KIND>
    SUBool::AmoEncoder::kKindAnnotation("AmoEncoder::KIND",
        {
            //
            {KIND::REPR, "repr", "the canonical representation"}, //
            {KIND::SEQ, "seq", "the sequential encoding"},
        });
const SUBool::EnumAnnotation<SUBool::ExOneEncoder::KIND>
    SUBool::ExOneEncoder::kKindAnnotation("ExOneEncoder::KIND",
        {
            //
            {KIND::REPR, "repr", "the canonical representation"}, //
            {KIND::SPLIT, "split", "the splitting encoding"}      //
        });

void
SUBool::amo_representation(
    const std::vector<Literal> &inputs, std::vector<Clause> &clauses)
{
   for (unsigned a = 0; a < inputs.size(); ++a)
   {
      Literal a_neg = inputs[a].Negation();
      for (unsigned b = a + 1; b < inputs.size(); ++b)
      {
         clauses.emplace_back(a_neg, inputs[b].Negation());
      }
   }
}

void
SUBool::exone_representation(
    const std::vector<Literal> &inputs, std::vector<Clause> &clauses)
{
   amo_representation(inputs, clauses);
   clauses.emplace_back(inputs);
}

std::vector<SUBool::Clause>
SUBool::ConstraintEncoder::Encode(
    const std::vector<Literal> &inputs, std::vector<Clause> clauses)
{
   EncodeAppend(inputs, clauses);
   return clauses;
}

void
SUBool::AmoRepresentation::EncodeAppend(
    const std::vector<Literal> &inputs, std::vector<Clause> &clauses)
{
   amo_representation(inputs, clauses);
}

void
SUBool::ExOneRepresentation::EncodeAppend(
    const std::vector<Literal> &inputs, std::vector<Clause> &clauses)
{
   exone_representation(inputs, clauses);
}

void
SUBool::AmoSequential::EncodeAppend(
    const std::vector<Literal> &inputs, std::vector<Clause> &clauses)
{
   if (inputs.size() <= mReprLimit)
   {
      amo_representation(inputs, clauses);
      return;
   }
   InnerEncode(inputs, clauses, 0);
}

void
SUBool::AmoSequential::InnerEncode(std::vector<Literal> inputs,
    std::vector<Clause> &clauses, unsigned next_aux_index)
{
   if (inputs.size() <= mBaseSize)
   {
      amo_representation(inputs, clauses);
      return;
   }
   mVarStore.NameStream() << "y_" << next_aux_index;
   unsigned y = mVarStore.NewVariable();
   auto l1 = inputs.back();
   inputs.pop_back();
   auto l2 = inputs.back();
   inputs.back() = Literal(y, true);
   amo_representation({l1, l2, Literal(y, false)}, clauses);
   InnerEncode(std::move(inputs), clauses, next_aux_index + 1);
}

void
SUBool::ExOneSplitting::EncodeAppend(
    const std::vector<Literal> &inputs, std::vector<Clause> &clauses)
{
   if (inputs.size() <= mReprLimit)
   {
      exone_representation(inputs, clauses);
      return;
   }
   unsigned next_aux_index = 0;
   InnerEncode(inputs, clauses, next_aux_index);
}

void
SUBool::ExOneSplitting::InnerEncode(std::vector<Literal> inputs,
    std::vector<Clause> &clauses, unsigned &next_aux_index)
{
   if (inputs.size() <= mBaseSize)
   {
      exone_representation(inputs, clauses);
      return;
   }
   mVarStore.NameStream() << "y_" << next_aux_index++ << std::endl;
   unsigned y = mVarStore.NewVariable();
   auto half = inputs.size() / 2;
   std::vector<Literal> right(inputs.begin() + half, inputs.end());
   inputs.resize(half);
   inputs.emplace_back(y, true);
   right.emplace_back(y, false);
   InnerEncode(inputs, clauses, next_aux_index);
   InnerEncode(right, clauses, next_aux_index);
}

std::string
SUBool::AmoRepresentation::Description() const
{
   return "AMO representation";
}

std::string
SUBool::AmoSequential::Description() const
{
   return "AMO sequential encoding";
}

std::string
SUBool::ExOneRepresentation::Description() const
{
   return "exactly-1 representation encoding";
}

std::string
SUBool::ExOneSplitting::Description() const
{
   return "exactly-1 splitting encoding";
}

std::unique_ptr<SUBool::AmoEncoder>
SUBool::AmoEncoder::NewEncoder(KIND kind, VarStore &var_store)
{
   switch (kind)
   {
   case KIND::REPR:
      return std::unique_ptr<AmoEncoder>(new AmoRepresentation());
   case KIND::SEQ:
      return std::unique_ptr<AmoEncoder>(new AmoSequential(var_store));
   default:
      throw BadParameterException(
          "AmoEncoder::NewEncoder", "Uknown encoding kind");
   };
}

std::unique_ptr<SUBool::ExOneEncoder>
SUBool::ExOneEncoder::NewEncoder(KIND kind, VarStore &var_store)
{
   switch (kind)
   {
   case KIND::REPR:
      return std::unique_ptr<ExOneEncoder>(new ExOneRepresentation());
   case KIND::SPLIT:
      return std::unique_ptr<ExOneEncoder>(new ExOneSplitting(var_store));
   default:
      throw BadParameterException(
          "ExOneEncoder::NewEncoder", "Uknown encoding kind");
   };
}

SUBool::NFSize
SUBool::AmoRepresentation::Size(const size_t input_size, unsigned max_input_var)
{
   auto c = (input_size * (input_size - 1)) / 2;
   return NFSize{max_input_var, c, 2 * c};
}

SUBool::NFSize
SUBool::AmoRepresentation::CountClauses(
    const size_t input_size, unsigned max_input_var) const
{
   return Size(input_size, max_input_var);
}

SUBool::NFSize
SUBool::ExOneRepresentation::Size(
    const size_t input_size, unsigned max_input_var)
{
   auto c = (input_size * (input_size - 1)) / 2;
   return NFSize{max_input_var, c + 1, 2 * c + input_size};
}

SUBool::NFSize
SUBool::ExOneRepresentation::CountClauses(
    const size_t input_size, unsigned max_input_var) const
{
   return Size(input_size, max_input_var);
}

SUBool::NFSize
SUBool::AmoSequential::CountClauses(
    const size_t input_size, unsigned max_input_var) const
{
   if (input_size <= mReprLimit || input_size <= mBaseSize)
   {
      return AmoRepresentation::Size(input_size, max_input_var);
   }
   auto d = input_size - mBaseSize;
   NFSize s{/* .max_variable = */ 1, /* .size = */ 3, /* .len = */ 6};
   return AmoRepresentation::Size(mBaseSize, max_input_var) + s * d;
}

SUBool::NFSize
SUBool::ExOneSplitting::CountClauses(
    const size_t input_size, unsigned max_input_var) const
{
   if (input_size <= mReprLimit)
   {
      return ExOneRepresentation::Size(input_size, max_input_var);
   }
   NFSize s{max_input_var, 0, 0};
   s += InnerCountClauses(input_size);
   return s;
}

SUBool::NFSize
SUBool::ExOneSplitting::InnerCountClauses(const size_t input_size) const
{
   if (input_size <= mBaseSize)
   {
      return ExOneRepresentation::Size(input_size, 0);
   }
   NFSize s{/* .max_variable = */ 1, /* .size = */ 0, /* .len = */ 0};
   auto half = input_size / 2;
   s += InnerCountClauses(half + 1);
   s += InnerCountClauses(input_size - half + 1);
   return s;
}
