/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file cardinal.h
 * Encodings of cardinality constraints.
 */

#ifndef __CARDINAL_H
#define __CARDINAL_H

#include <array>
#include <memory>
#include <vector>

#include "enum.h"
#include "nfliteralset.h"
#include "normalform.h"
#include "varstore.h"

namespace SUBool
{
   class ConstraintEncoder
   {
    public:
      virtual ~ConstraintEncoder() = default;

      virtual std::vector<Clause> Encode(const std::vector<Literal> &inputs,
          std::vector<Clause> clauses = std::vector<Clause>());
      virtual void EncodeAppend(
          const std::vector<Literal> &inputs, std::vector<Clause> &clauses) = 0;

      virtual std::string Description() const = 0;

      virtual NFSize CountClauses(
          const size_t input_size, unsigned max_input_var = 0) const = 0;

      std::vector<Clause>
      operator()(const std::vector<Literal> &inputs)
      {
         return Encode(inputs);
      }
   };

   class AmoEncoder : public ConstraintEncoder
   {
    public:
      enum class KIND
      {
         REPR,
         SEQ,
         Last = SEQ
      };

      static const EnumAnnotation<KIND> kKindAnnotation;

      inline static const KIND kDefaultGlobalKind{KIND::REPR};
      static KIND sGlobalKind;
      static std::unique_ptr<AmoEncoder> NewEncoder(
          KIND kind, VarStore &var_store);
      static std::unique_ptr<AmoEncoder>
      NewGlobalEncoder(VarStore &var_store)
      {
         return NewEncoder(sGlobalKind, var_store);
      }
   };

   class AmoRepresentation : public AmoEncoder
   {
    public:
      virtual std::string Description() const override;

      virtual void EncodeAppend(const std::vector<Literal> &inputs,
          std::vector<Clause> &clauses) override;
      virtual NFSize CountClauses(
          const size_t input_size, unsigned max_input_var = 0) const override;

      static NFSize Size(const size_t input_size, unsigned max_input_var);
   };

   class ExOneEncoder : public ConstraintEncoder
   {
    public:
      enum class KIND
      {
         REPR,
         SPLIT,
         Last = SPLIT
      };

      static const EnumAnnotation<KIND> kKindAnnotation;
      inline static const KIND kDefaultGlobalKind{KIND::REPR};
      static KIND sGlobalKind;
      static std::unique_ptr<ExOneEncoder> NewEncoder(
          KIND kind, VarStore &var_store);
      static std::unique_ptr<ExOneEncoder>
      NewGlobalEncoder(VarStore &var_store)
      {
         return NewEncoder(sGlobalKind, var_store);
      }
   };

   class ExOneRepresentation : public ExOneEncoder
   {
    public:
      virtual std::string Description() const override;
      virtual void EncodeAppend(const std::vector<Literal> &inputs,
          std::vector<Clause> &clauses) override;
      virtual NFSize CountClauses(
          const size_t input_size, unsigned max_input_var = 0) const override;
      static NFSize Size(const size_t input_size, unsigned max_input_var);
   };

   class AmoSequential : public AmoEncoder
   {
      VarStore &mVarStore;
      unsigned mBaseSize;
      unsigned mReprLimit;

      void InnerEncode(std::vector<Literal> inputs,
          std::vector<Clause> &clauses, unsigned next_aux_index);

    public:
      static const unsigned kDefaultBaseSize = 6;
      static const unsigned kDefaultReprLimit = 10;

      AmoSequential(VarStore &var_store, unsigned base_size = kDefaultBaseSize,
          unsigned repr_limit = kDefaultReprLimit)
          : mVarStore(var_store), mBaseSize(base_size), mReprLimit(repr_limit)
      {
      }
      AmoSequential(VarStore &&var_store, unsigned base_size,
          unsigned repr_limit) = delete;

      virtual std::string Description() const override;
      virtual void EncodeAppend(const std::vector<Literal> &inputs,
          std::vector<Clause> &clauses) override;
      virtual NFSize CountClauses(
          const size_t input_size, unsigned max_input_var = 0) const override;
   };

   class ExOneSplitting : public ExOneEncoder
   {
      VarStore &mVarStore;
      unsigned mBaseSize;
      unsigned mReprLimit;

      void InnerEncode(std::vector<Literal> inputs,
          std::vector<Clause> &clauses, unsigned &next_aux_index);
      NFSize InnerCountClauses(const size_t input_size) const;

    public:
      static const unsigned kDefaultBaseSize = 4;
      static const unsigned kDefaultReprLimit = 10;

      virtual std::string Description() const override;
      ExOneSplitting(VarStore &var_store, unsigned base_size = kDefaultBaseSize,
          unsigned repr_limit = kDefaultReprLimit)
          : mVarStore(var_store), mBaseSize(base_size), mReprLimit(repr_limit)
      {
      }
      ExOneSplitting(VarStore &&var_store, unsigned base_size,
          unsigned repr_limit) = delete;

      virtual void EncodeAppend(const std::vector<Literal> &inputs,
          std::vector<Clause> &clauses) override;
      virtual NFSize CountClauses(
          const size_t input_size, unsigned max_input_var = 0) const override;
   };

   void amo_representation(
       const std::vector<Literal> &inputs, std::vector<Clause> &clauses);

   void exone_representation(
       const std::vector<Literal> &inputs, std::vector<Clause> &clauses);

} // namespace SUBool

#endif
