/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfcompare.cpp
 * Various CNF comparison related utility functions.
 */

#include "cnfcompare.h"
#include "cnfutil.h"
#include "empower.h"
#include "util.h"

bool
SUBool::cnf_implies(const CNF &a, const CNF &b)
{
   IncSolverWrapper solver(a);
   size_t i = 0;
   bool log_progress = (b.Size() > LogStream::kMinLogCNFSize);
   for (const auto &clause : b)
   {
      if (log_progress)
      {
         ++i;
         if (i % LogStream::kLogStep == 0)
         {
            double p = (static_cast<double>(i) / static_cast<double>(b.Size()))
                       * 100.0;
            logs.TLog(4) << p << "%" << std::endl;
         }
      }
      if (!is_implicate(solver.Solver(), clause))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::cnf_implies_up(const CNF &a, const CNF &b)
{
   OneProve op(&a);
   size_t i = 0;
   bool log_progress = (b.Size() > LogStream::kMinLogCNFSize);
   for (const auto &clause : b)
   {
      if (log_progress)
      {
         ++i;
         if (i % LogStream::kLogStep == 0)
         {
            double p = (static_cast<double>(i) / static_cast<double>(b.Size()))
                       * 100.0;
            logs.TLog(4) << p << "%" << std::endl;
         }
      }
      if (!op.IsOneProvable(clause))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::cnf_absorbed(const CNF &a, const CNF &b)
{
   OneProve op(&a);
   size_t i = 0;
   bool log_progress = (b.Size() > LogStream::kMinLogCNFSize);
   for (const auto &clause : b)
   {
      if (log_progress)
      {
         ++i;
         if (i % LogStream::kLogStep == 0)
         {
            double p = (static_cast<double>(i) / static_cast<double>(b.Size()))
                       * 100.0;
            logs.TLog(4) << p << "%" << std::endl;
         }
      }
      if (!a.SubsumedByAny(clause) && op.IsEmpowering(clause))
      {
         return false;
      }
   }
   return true;
}

SUBool::CompareResult
SUBool::cmp_result_of_bool(bool leq, bool geq)
{
   return (leq ? (geq ? CompareResult::EQUAL : CompareResult::LESS)
               : (geq ? CompareResult::GREATER : CompareResult::INCOMPARABLE));
}

SUBool::CompareResult
SUBool::cnf_compare(const CNF &a, const CNF &b)
{
   return cmp_result_of_bool(cnf_implies(a, b), cnf_implies(b, a));
}

SUBool::CompareResult
SUBool::cnf_compare_up(const CNF &a, const CNF &b)
{
   return cmp_result_of_bool(cnf_implies_up(a, b), cnf_implies_up(b, a));
}

SUBool::CompareResult
SUBool::cnf_compare_absorb(const CNF &a, const CNF &b)
{
   return cmp_result_of_bool(cnf_absorbed(a, b), cnf_absorbed(b, a));
}

bool
SUBool::cnf_equal(const CNF &a, const CNF &b)
{
   return cnf_implies(a, b) && cnf_implies(b, a);
}

bool
SUBool::cnf_equal_up(const CNF &a, const CNF &b)
{
   return cnf_implies_up(a, b) && cnf_implies_up(b, a);
}

bool
SUBool::cnf_equal_absorb(const CNF &a, const CNF &b)
{
   return cnf_absorbed(a, b) && cnf_absorbed(b, a);
}
