/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfcompress.h
 * Compressing CNF. This means renaming variables so that each
 * variable appears in some clause.
 */

#include <vector>

#include "cnfcompress.h"

SUBool::CNFCompressor::CNFCompressor(const CNF &cnf,
                                     unsigned max_fixed_variable)
    : mMutex(), mVarMap(), mVarBackmap()
{
   InitVarMap(cnf, max_fixed_variable);
}

SUBool::CNFCompressor::CNFCompressor(const CNFCompressor &other)
    : mMutex(), mVarMap(), mVarBackmap()
{
   auto maps = other.Maps();
   mVarMap = std::move(maps.first);
   mVarBackmap = std::move(maps.second);
}

SUBool::CNFCompressor::CNFCompressor(CNFCompressor &&other)
    : mMutex(), mVarMap(), mVarBackmap()
{
   auto maps = std::move(other).Maps();
   mVarMap = std::move(maps.first);
   mVarBackmap = std::move(maps.second);
}

SUBool::CNF
SUBool::CNFCompressor::CompressCNF(CNF cnf) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   cnf.Transform(
       [this](const auto &clause) { return InnerCompressClause(clause); });
   return cnf;
}

SUBool::CNF
SUBool::CNFCompressor::UncompressCNF(CNF cnf) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   cnf.Transform(
       [this](const auto &clause) { return InnerUncompressClause(clause); });
   return cnf;
}

void
SUBool::CNFCompressor::InitVarMap(const CNF &cnf, unsigned max_fixed_variable)
{
   std::vector<bool> occur(cnf.MaxVariable() + 1, false);
   for (unsigned var = 1; var <= max_fixed_variable; ++var)
   {
      occur[var] = true;
   }
   for (const auto &clause : cnf)
   {
      for (const auto &lit : clause)
      {
         occur[lit.Variable()] = true;
      }
   }
   mVarMap.resize(cnf.MaxVariable() + 1);
   mVarBackmap.push_back(0);
   for (unsigned i = 1; i <= cnf.MaxVariable(); ++i)
   {
      if (occur[i])
      {
         mVarBackmap.push_back(i);
         mVarMap[i] = mVarBackmap.size() - 1;
      }
   }
}

unsigned
SUBool::CNFCompressor::InnerMaxVariable() const
{
   return mVarBackmap.size() - 1;
}

unsigned
SUBool::CNFCompressor::InnerUncompressVar(unsigned var) const
{
   return (var < mVarBackmap.size() ? mVarBackmap[var] : var);
}

SUBool::Literal
SUBool::CNFCompressor::InnerUncompressLiteral(const Literal &lit) const
{
   return Literal(InnerUncompressVar(lit.Variable()), lit.Positive());
}

SUBool::Literal
SUBool::CNFCompressor::InnerCompressLiteral(const Literal &lit) const
{
   return Literal(InnerCompressVar(lit.Variable()), lit.Positive());
}

unsigned
SUBool::CNFCompressor::InnerCompressVar(unsigned var) const
{
   return (var < mVarMap.size() ? mVarMap[var] : var);
}

SUBool::Clause
SUBool::CNFCompressor::InnerUncompressClause(const Clause &clause) const
{
   std::vector<Literal> uncomp_literals;
   for (const auto &lit : clause)
   {
      uncomp_literals.emplace_back(InnerUncompressVar(lit.Variable()),
                                   lit.Positive());
   }
   return Clause(uncomp_literals);
}

SUBool::Clause
SUBool::CNFCompressor::InnerCompressClause(const Clause &clause) const
{
   std::vector<Literal> comp_literals;
   for (const auto &lit : clause)
   {
      comp_literals.emplace_back(InnerCompressVar(lit.Variable()),
                                 lit.Positive());
   }
   return Clause(comp_literals);
}

unsigned
SUBool::CNFCompressor::MaxVariable() const
{
   std::lock_guard<std::mutex> lock(mMutex);
   return InnerMaxVariable();
}

unsigned
SUBool::CNFCompressor::UncompressVar(unsigned var) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   return InnerUncompressVar(var);
}

SUBool::Literal
SUBool::CNFCompressor::UncompressLiteral(const Literal &lit) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   return InnerUncompressLiteral(lit);
}

SUBool::Literal
SUBool::CNFCompressor::CompressLiteral(const Literal &lit) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   return InnerCompressLiteral(lit);
}

unsigned
SUBool::CNFCompressor::CompressVar(unsigned var) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   return InnerCompressVar(var);
}

SUBool::Clause
SUBool::CNFCompressor::UncompressClause(const Clause &clause) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   return InnerUncompressClause(clause);
}

SUBool::Clause
SUBool::CNFCompressor::CompressClause(const Clause &clause) const
{
   std::lock_guard<std::mutex> lock(mMutex);
   return InnerCompressClause(clause);
}

SUBool::CNFCompressor &
SUBool::CNFCompressor ::operator=(const CNFCompressor &other)
{
   std::lock_guard<std::mutex> lock(mMutex);
   auto maps = other.Maps();
   mVarMap = std::move(maps.first);
   mVarBackmap = std::move(maps.second);
   return *this;
}

SUBool::CNFCompressor &
SUBool::CNFCompressor ::operator=(CNFCompressor &&other)
{
   std::lock_guard<std::mutex> lock(mMutex);
   auto maps = std::move(other).Maps();
   mVarMap = std::move(maps.first);
   mVarBackmap = std::move(maps.second);
   return *this;
}

std::pair<std::vector<unsigned>, std::vector<unsigned>>
SUBool::CNFCompressor::Maps() const &
{
   std::lock_guard<std::mutex> lock(mMutex);
   return std::make_pair(mVarMap, mVarBackmap);
}

std::pair<std::vector<unsigned>, std::vector<unsigned>>
SUBool::CNFCompressor::Maps() &
{
   std::lock_guard<std::mutex> lock(mMutex);
   return std::make_pair(std::move(mVarMap), std::move(mVarBackmap));
}
