/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file graph.cpp
 * CNF graphs.
 */

#include <boost/graph/graphml.hpp>
#include <boost/graph/graphviz.hpp>

#include "cnfgraph.h"

const SUBool::PDCNFGraph::GraphType &
SUBool::PDCNFGraph::Graph() const
{
   return mGraph;
}

std::pair<boost::dynamic_properties, std::string>
SUBool::PDCNFGraph::DynamicProperties(
    const std::string &vertex_prefix, const std::string &vertex_color)
{
   boost::dynamic_properties dp;
   auto index = boost::get(boost::vertex_index, mGraph);
   auto index2name = [&](unsigned i) { return NameOfIndex(i, vertex_prefix); };
   dp.property(
       "name", boost::make_transform_value_property_map(index2name, index));
   if (!vertex_color.empty())
   {
      dp.property("color",
          boost::make_transform_value_property_map(
              [&vertex_color](unsigned) { return vertex_color; }, index));
   }
   return std::make_pair(std::move(dp), "name");
}

void
SUBool::PDCNFGraph::WriteGraphML(std::ostream &out,
    const std::string &vertex_prefix, const std::string &vertex_color)
{
   auto dp = DynamicProperties(vertex_prefix, vertex_color);
   boost::write_graphml(out, Graph(), dp.first, true);
}

void
SUBool::PDCNFGraph::WriteGraphViz(std::ostream &out,
    const std::string &vertex_prefix, const std::string &vertex_color)
{
   auto dp = DynamicProperties(vertex_prefix, vertex_color);
   boost::write_graphviz_dp(out, Graph(), dp.first, dp.second);
}

SUBool::PDCNFGraph::PDCNFGraph() : mGraph() {}

void
SUBool::PDCNFGraph::InitGraph(
    std::vector<std::pair<unsigned, unsigned>> edges, unsigned vertex_count)
{
   std::sort(edges.begin(), edges.end());
   auto it = std::unique(edges.begin(), edges.end());
   edges.resize(std::distance(edges.begin(), it));
   mGraph = GraphType(edges.begin(), edges.end(), vertex_count);
}

SUBool::PrimalGraph::PrimalGraph(const CNF &cnf)
{
   std::vector<std::pair<unsigned, unsigned>> edges;
   for (const auto &clause : cnf)
   {
      for (unsigned i = 0; i < clause.Size(); ++i)
      {
         for (unsigned j = i + 1; j < clause.Size(); ++j)
         {
            edges.emplace_back(
                clause[i].Variable() - 1, clause[j].Variable() - 1);
         }
      }
   }
   InitGraph(std::move(edges), cnf.MaxVariable());
}

std::string
SUBool::PrimalGraph::NameOfIndex(unsigned i, const std::string &prefix)
{
   return prefix + std::to_string(i + 1);
}

SUBool::DualGraph::DualGraph(const CNF &cnf)
{
   std::vector<std::vector<unsigned>> clauses2var(cnf.MaxVariable());
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      for (const auto &lit : cnf[i])
      {
         clauses2var[lit.Variable() - 1].push_back(i);
      }
   }
   std::vector<std::pair<unsigned, unsigned>> edges;
   for (const auto &clauses : clauses2var)
   {
      for (unsigned i = 0; i < clauses.size(); ++i)
      {
         for (unsigned j = i + 1; j < clauses.size(); ++j)
         {
            edges.emplace_back(clauses[i], clauses[j]);
         }
      }
   }
   InitGraph(std::move(edges), cnf.MaxVariable());
}

std::string
SUBool::DualGraph::NameOfIndex(unsigned i, const std::string &prefix)
{
   return prefix + std::to_string(i);
}
