/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfgraph.h
 * CNF graphs.
 */

#ifndef __CNFGRAPH_H
#define __CNFGRAPH_H

#include "normalform.h"

#include <boost/graph/adjacency_list.hpp>
#include <boost/property_map/dynamic_property_map.hpp>

namespace SUBool
{
   class PDCNFGraph
   {
    public:
      using GraphType =
          boost::adjacency_list<boost::vecS, boost::vecS, boost::undirectedS>;

    private:
      GraphType mGraph;

    protected:
      PDCNFGraph();
      virtual ~PDCNFGraph() = default;

      void InitGraph(std::vector<std::pair<unsigned, unsigned>> edges,
                     unsigned vertex_count);

      virtual std::string NameOfIndex(unsigned i,
                                      const std::string &prefix) = 0;

      virtual std::pair<boost::dynamic_properties, std::string>
      DynamicProperties(const std::string &vertex_prefix,
                        const std::string &vertex_color);

    public:
      const GraphType &Graph() const;
      virtual void WriteGraphML(std::ostream &out,
                                const std::string &vertex_prefix,
                                const std::string &vertex_color = "");
      virtual void WriteGraphViz(std::ostream &out,
                                 const std::string &vertex_prefix,
                                 const std::string &vertex_color = "");
   };

   class PrimalGraph : public PDCNFGraph
   {
    protected:
      virtual std::string NameOfIndex(unsigned i,
                                      const std::string &prefix) override;

    public:
      PrimalGraph(const CNF &cnf);
   };

   class DualGraph : public PDCNFGraph
   {
    protected:
      virtual std::string NameOfIndex(unsigned i,
                                      const std::string &prefix) override;

    public:
      DualGraph(const CNF &cnf);
   };

   template <class Directionality> class IncidenceGraphT
   {
    public:
      enum class VERTEX_TYPE
      {
         VARIABLE,
         CLAUSE
      };
      using VertexProperty =
          boost::property<boost::vertex_color_t, VERTEX_TYPE>;
      using GraphType = boost::adjacency_list<boost::vecS, boost::vecS,
                                              Directionality, VertexProperty>;

      struct Components
      {
         unsigned count = 0;
         // indexed from 0, var 1 is at index 0
         std::vector<unsigned> comp_variables{};
         std::vector<unsigned> comp_clauses{};
      };

    private:
      GraphType mGraph;
      unsigned mMaxVariable;

    protected:
      virtual std::pair<boost::dynamic_properties, std::string>
      DynamicProperties(const std::string &var_prefix,
                        const std::string &var_color,
                        const std::string &clause_prefix,
                        const std::string &clause_color);

    public:
      IncidenceGraphT(const CNF &cnf);
      virtual ~IncidenceGraphT() = default;

      const GraphType &Graph() const;
      void WriteGraphML(std::ostream &out, const std::string &var_prefix,
                        const std::string &var_color,
                        const std::string &clause_prefix,
                        const std::string &clause_color);
      void WriteGraphViz(std::ostream &out, const std::string &var_prefix,
                         const std::string &var_color,
                         const std::string &clause_prefix,
                         const std::string &clause_color);

      Components ConnectedComponents() const;
   };

   using IncidenceGraph = IncidenceGraphT<boost::undirectedS>;
   using DIncidenceGraph = IncidenceGraphT<boost::directedS>;
} // namespace SUBool

#include "cnfgraph_impl.h"

#endif
