/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfgraph_impl.h
 * CNF graphs.
 */

#include <boost/graph/connected_components.hpp>
#include <boost/graph/graphml.hpp>
#include <boost/graph/graphviz.hpp>

template <class Directionality>
const typename SUBool::IncidenceGraphT<Directionality>::GraphType &
SUBool::IncidenceGraphT<Directionality>::Graph() const
{
   return mGraph;
}

template <class Directionality>
void
SUBool::IncidenceGraphT<Directionality>::WriteGraphML(std::ostream &out,
    const std::string &var_prefix, const std::string &var_color,
    const std::string &clause_prefix, const std::string &clause_color)
{
   auto dp =
       DynamicProperties(var_prefix, var_color, clause_prefix, clause_color);
   boost::write_graphml(out, Graph(), dp.first, true);
}

template <class Directionality>
void
SUBool::IncidenceGraphT<Directionality>::WriteGraphViz(std::ostream &out,
    const std::string &var_prefix, const std::string &var_color,
    const std::string &clause_prefix, const std::string &clause_color)
{
   auto dp =
       DynamicProperties(var_prefix, var_color, clause_prefix, clause_color);
   boost::write_graphviz_dp(out, Graph(), dp.first, dp.second);
}

template <class Directionality>
SUBool::IncidenceGraphT<Directionality>::IncidenceGraphT(const CNF &cnf)
    : mGraph(), mMaxVariable(cnf.MaxVariable())
{
   std::vector<std::pair<unsigned, unsigned>> edges;
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      for (const auto &lit : cnf[i])
      {
         if (lit.Positive())
         {
            edges.emplace_back(cnf.MaxVariable() + i, lit.Variable() - 1);
         }
         else
         {
            edges.emplace_back(lit.Variable() - 1, cnf.MaxVariable() + i);
         }
      }
   }
   mGraph =
       GraphType(edges.begin(), edges.end(), cnf.MaxVariable() + cnf.Size());
   typename boost::graph_traits<GraphType>::vertex_iterator v, v_end;
   auto index = boost::get(boost::vertex_index, mGraph);
   for (boost::tie(v, v_end) = boost::vertices(mGraph); v != v_end; ++v)
   {
      std::stringstream ss;
      if (index[*v] < cnf.MaxVariable())
      {
         boost::put(boost::vertex_color_t(), mGraph, *v, VERTEX_TYPE::VARIABLE);
      }
      else
      {
         boost::put(boost::vertex_color_t(), mGraph, *v, VERTEX_TYPE::CLAUSE);
      }
   }
}

template <class Directionality>
std::pair<boost::dynamic_properties, std::string>
SUBool::IncidenceGraphT<Directionality>::DynamicProperties(
    const std::string &var_prefix, const std::string &var_color,
    const std::string &clause_prefix, const std::string &clause_color)
{
   boost::dynamic_properties dp;
   auto index = boost::get(boost::vertex_index, mGraph);
   auto index2name = [&](unsigned i)
   {
      return (i < mMaxVariable
                  ? var_prefix + std::to_string(i + 1)
                  : clause_prefix + std::to_string(i - mMaxVariable));
   };
   dp.property(
       "name", boost::make_transform_value_property_map(index2name, index));
   if (!var_color.empty() || !clause_color.empty())
   {
      auto vertex_type = boost::get(boost::vertex_color, mGraph);
      auto func = [&](const VERTEX_TYPE &vt) -> std::string
      { return (vt == VERTEX_TYPE::VARIABLE ? var_color : clause_color); };
      dp.property(
          "color", boost::make_transform_value_property_map(func, vertex_type));
   }
   return std::make_pair(dp, "name");
}

template <class Directionality>
typename SUBool::IncidenceGraphT<Directionality>::Components
SUBool::IncidenceGraphT<Directionality>::ConnectedComponents() const
{
   std::vector<unsigned> comp(boost::num_vertices(mGraph));
   Components components;
   components.count = boost::connected_components(mGraph, &comp[0]);
   components.comp_variables.assign(comp.begin(), comp.begin() + mMaxVariable);
   auto mv = mMaxVariable;
   std::transform(comp.begin() + mMaxVariable, comp.end(),
       std::back_inserter(components.comp_clauses),
       [&mv](unsigned cl) { return cl - mv; });
   return components;
}
