/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfutil.cpp
 * Various CNF related utility functions.
 * Making CNF prime. Either using SAT, or a weaker version just by
 * using unit propagation.
 */

#include <unordered_map>

#include "cnfutil.h"
#include "unitprop.h"

namespace SUBool
{
   // Assumes that the CNF is satisfiable.
   std::vector<Literal> masked_clause_assump(const Clause &clause,
                                             const std::vector<bool> &mask);
   Clause masked_clause(const Clause &clause, const std::vector<bool> &mask);
   std::optional<Clause> solve_with_learning(
       const CNF &prime, const CNF &cnf, unsigned cnf_start_index, CNF *learned,
       const Clause *clause, const std::vector<Literal> &assump);
   std::unordered_set<Literal>
   intersect_lits_with_assign(std::unordered_set<Literal> lits,
                              const PartialAssignment &alpha);

}; // namespace SUBool

SUBool::CNF
SUBool::masked_cnf(CNF cnf, const std::vector<bool> &mask)
{
   auto elements = std::move(cnf).Elements();
   unsigned j = 0;
   for (unsigned i = 0; i < elements.size(); ++i)
   {
      if (mask[i])
      {
         if (j < i)
         {
            elements[j] = std::move(elements[i]);
         }
         ++j;
      }
   }
   elements.resize(j);
   return elements;
}

std::vector<SUBool::Literal>
SUBool::masked_clause_assump(const Clause &clause,
                             const std::vector<bool> &mask)
{
   std::vector<Literal> assump;
   for (unsigned j = 0; j < clause.Size(); ++j)
   {
      if (mask[j])
      {
         assump.push_back(clause[j].Negation());
      }
   }
   return assump;
}

SUBool::Clause
SUBool::masked_clause(const Clause &clause, const std::vector<bool> &mask)
{
   std::vector<Literal> literals;
   for (unsigned i = 0; i < clause.Size(); ++i)
   {
      if (mask[i])
      {
         literals.push_back(clause[i]);
      }
   }
   return Clause(literals);
}

SUBool::PartialAssignment
SUBool::clause_falsifying(const Clause &clause, unsigned n)
{
   PartialAssignment alpha(n);
   for (const auto &lit : clause)
   {
      if (lit.Variable() > n)
      {
         break;
      }
      alpha[lit.Variable()] = lit.FalseValue();
   }
   return alpha;
}

bool
SUBool::is_implicate(const CNF &cnf, const Clause &clause)
{
   SimpSolverWrapper wrap;
   auto &g = wrap.Solver();
   g.AddCnf(cnf);
   return is_implicate(g, clause);
}

bool
SUBool::is_implicate(const SolverInterface &solver, const Clause &clause)
{
   std::vector<Literal> neg_literals;
   neg_literals.reserve(clause.Size());
   for (const auto &lit : clause)
   {
      neg_literals.push_back(lit.Negation());
   }
   SimpSolverWrapper wrap;
   return !solver.Solve(neg_literals);
}

std::optional<SUBool::Clause>
SUBool::prime_subimplicate(const CNF &cnf, const Clause &clause,
                           unsigned ignore_var, bool promised_sat)
{
   IncSolverWrapper solver(cnf);
   return prime_subimplicate(solver.Solver(), clause, ignore_var, promised_sat);
}

std::optional<SUBool::Clause>
SUBool::prime_subimplicate(const SolverInterface &solver, const Clause &clause,
                           unsigned ignore_var, bool promised_sat)
{
   std::vector<Literal> assump;
   std::transform(clause.begin(), clause.end(), std::back_inserter(assump),
                  literal_negate);
   auto r = solver.Solve(assump);
   if (r)
   {
      return {}; // not an implicate
   }
   auto subclause = solver.ConflictClause();
   std::vector<Literal> to_check(subclause.Literals());
   while (!to_check.empty() && (!promised_sat || subclause.Size() > 1))
   {
      auto l = to_check.back();
      to_check.pop_back();
      if (l.Variable() == ignore_var)
      {
         continue;
      }
      assump.clear();
      std::for_each(subclause.begin(), subclause.end(),
                    [&assump, &l](const auto &lit) {
                       if (lit != l)
                       {
                          assump.push_back(lit.Negation());
                       }
                    });
      std::shuffle(assump.begin(), assump.end(), sub_rand);
      if (!solver.Solve(assump))
      {
         auto old_size = subclause.Size();
         subclause = solver.ConflictClause();
         if (subclause.Size() + 1 < old_size)
         {
            std::vector<Literal> intersect;
            std::set_intersection(
                to_check.begin(), to_check.end(), subclause.Literals().begin(),
                subclause.Literals().end(), std::back_inserter(intersect));
            to_check = std::move(intersect);
         }
      }
   }
   return subclause;
}

std::optional<SUBool::Clause>
SUBool::solve_with_learning(const CNF &prime, const CNF &cnf,
                            unsigned cnf_start_index, CNF *learned,
                            const Clause *clause,
                            const std::vector<Literal> &assump)
{
   SimpSolverWrapper wrap;
   auto &g = wrap.Solver();
   g.AddCnf(prime);
   if (learned)
   {
      g.AddCnf(*learned);
   }
   if (clause)
   {
      g.AddClause(*clause);
   }
   if (cnf_start_index < cnf.Size())
   {
      for (auto i_clause = cnf.begin() + cnf_start_index; i_clause != cnf.end();
           ++i_clause)
      {
         g.AddClause(*i_clause);
      }
   }
   auto r = g.Solve(assump);
   if (learned)
   {
      learned->Append(g.Learnts());
   }
   if (!r)
   {
      return g.ConflictClause();
   }
   return {};
}

SUBool::CNF
SUBool::prime_cnf(const SolverInterface &solver, const CNF &cnf)
{
   CNF prime;
   bool promised_sat = false;
   for (const auto &clause : cnf)
   {
      if (prime.SubsumedByAny(clause))
      {
         continue;
      }
      auto p = prime_subimplicate(solver, clause, 0, promised_sat);
      if (!p)
      {
         return CNF(Clause::EmptyClause());
      }
      promised_sat = true;
      prime.PushBack(*p);
   }
   return prime;
}

bool
SUBool::is_satisfiable(const CNF &cnf)
{
   SimpSolverWrapper wrap;
   auto &g = wrap.Solver();
   g.AddCnf(cnf);
   return g.Solve();
}
