/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file cnfutil.h
 * Various CNF related utility functions.
 * Making CNF prime. Either using SAT, or a weaker version just by
 * using unit propagation.
 */

// TODO: irredundant cnf.

#ifndef __CNFUTIL_H
#define __CNFUTIL_H

#include <optional>
#include <queue>
#include <unordered_set>

#include "glucose_bind.h"
#include "logstream.h"
#include "normalform.h"
#include "random_utils.h"
#include "unitprop.h"

namespace SUBool
{
   bool is_satisfiable(const CNF &cnf);
   bool is_consistent(const CNF &cnf, const std::vector<Literal> &assump);
   bool is_consistent(const CNF &cnf, const PartialAssignment &assump);
   bool is_consistent(const CNF &cnf, const Assignment &assump);

   bool is_implicate(const CNF &cnf, const Clause &clause);
   // Solver must be initialized with a CNF already
   bool is_implicate(const SolverInterface &solver, const Clause &clause);
   // promised_sat = cnf is satisfiable.
   std::optional<Clause> prime_subimplicate(const CNF &cnf,
                                            const Clause &clause,
                                            unsigned ignore_var = 0,
                                            bool promised_sat = false);

   // Solver must be incremental and initialized with a CNF already
   std::optional<Clause> prime_subimplicate(const SolverInterface &solver,
                                            const Clause &clause,
                                            unsigned ignore_var = 0,
                                            bool promised_sat = false);

   // Clauses at index i for which pred(i) is satisfied are ignored
   template <class IgnorePred>
   CNF prime_cnf_with_ignore(const CNF &cnf, IgnorePred pred);

   // solver must be incremental and initialized with a CNF which implies @p cnf
   template <class IgnorePred>
   CNF prime_cnf_with_ignore(const SolverInterface &solver, const CNF &cnf,
                             IgnorePred pred);

   // solver must be incremental and initialized with a CNF which implies @p cnf
   CNF prime_cnf(const SolverInterface &solver, const CNF &cnf);

   inline CNF
   prime_cnf(const CNF &cnf)
   {
      IncSolverWrapper solver(cnf);
      return prime_cnf(solver.Solver(), cnf);
   }

   CNF masked_cnf(SUBool::CNF cnf, const std::vector<bool> &mask);
   PartialAssignment clause_falsifying(const Clause &, unsigned n);

} // namespace SUBool

// Template definition
template <class Pred>
SUBool::CNF
SUBool::prime_cnf_with_ignore(const CNF &cnf, Pred pred)
{
   IncSolverWrapper solver(cnf);
   return prime_cnf_with_ignore(solver.Solver(), cnf, pred);
}

template <class Pred>
SUBool::CNF
SUBool::prime_cnf_with_ignore(const SolverInterface &solver, const CNF &cnf,
                              Pred pred)
{
   if (!solver.Solve())
   {
      return CNF(Clause::EmptyClause());
   }
   CNF prime;
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      if (pred(i))
      {
         prime.PushBack(cnf[i]);
      }
   }
   // add nonignored clauses in a cycle, then pred(i) is checked before
   // subsumedbyany
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      if (pred(i) || prime.SubsumedByAny(cnf[i]))
      {
         continue;
      }
      auto p = prime_subimplicate(solver, cnf[i], 0, true);
      prime.PushBack(*std::move(p));
   }
   return prime;
}

inline bool
SUBool::is_consistent(const CNF &cnf, const std::vector<Literal> &assump)
{
   SimpSolverWrapper wrap(cnf);
   auto &g = wrap.Solver();
   return g.Solve(assump);
}

inline bool
SUBool::is_consistent(const CNF &cnf, const PartialAssignment &assump)
{
   SimpSolverWrapper wrap(cnf);
   auto &g = wrap.Solver();
   return g.Solve(assump);
}

inline bool
SUBool::is_consistent(const CNF &cnf, const Assignment &assump)
{
   SimpSolverWrapper wrap(cnf);
   auto &g = wrap.Solver();
   return g.Solve(assump);
}

#endif
