/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file comp_hypothesis.cpp
 * Class representing a hypothesis in the compilation algorithms (both iterative
 * and learning). The access is guarded by a mutex.
 */

#include "comp_hypothesis.h"
#include "cnfutil.h"
#include "cpuutil.h"
#include "propagate.h"

SUBool::CompHypothesis::CompHypothesis(MINIMIZE_METHOD minimize_method)
    : mRepresentation(), mHypothesis(), mSolver(), mSafe(false),
      mOneProveRepr(mRepresentation), mOneProveHyp(mHypothesis),
      mMinSubimplicant(&mRepresentation),
      mSemPropagator(
          mSolver.GetSolverPtr(), &(mOneProveRepr.UP()), &mMinSubimplicant),
      mMutex(), mMinimizeMethod(minimize_method)
{
}

SUBool::CompHypothesis::CompHypothesis(
    MINIMIZE_METHOD minimize_method, CNF cnf, CNF repr, bool safe)
    : mRepresentation(std::move(repr)), mHypothesis(std::move(cnf)),
      mSolver(safe ? mHypothesis : mRepresentation), mSafe(safe),
      mOneProveRepr(mRepresentation), mOneProveHyp(mHypothesis),
      mMinSubimplicant(safe ? &mHypothesis : &mRepresentation),
      mSemPropagator(mSolver.GetSolverPtr(),
          safe ? &(mOneProveHyp.UP()) : &(mOneProveRepr.UP()),
          &mMinSubimplicant),
      mMutex(), mMinimizeMethod(minimize_method)
{
}

size_t
SUBool::CompHypothesis::MinimizationLogProgress(size_t cnf_size)
{
   if (cnf_size < kMinimizationLogThreshold)
   {
      return 0;
   }
   return 100; // 100;
}

void
SUBool::CompHypothesis::ResetSemPropagator()
{
   if (mSafe)
   {
      mMinSubimplicant.Reset(&mHypothesis);
      mSemPropagator.Reset(
          mSolver.GetSolverPtr(), &(mOneProveHyp.UP()), &mMinSubimplicant);
      return;
   }
   mMinSubimplicant.Reset(&mRepresentation);
   mSemPropagator.Reset(
       mSolver.GetSolverPtr(), &(mOneProveRepr.UP()), &mMinSubimplicant);
}

void
SUBool::CompHypothesis::ResetSolver()
{
   ExtractLearnts();
   mSolver.Solver().Reset(mSafe ? mHypothesis : mRepresentation);
   ResetSemPropagator();
}

void
SUBool::CompHypothesis::SetImplSystemRebuild(
    IMPL_SYSTEM_REBUILD impl_system_rebuild)
{
   std::lock_guard<SharedMutex> lock(mMutex);
   mImplSystemRebuild = impl_system_rebuild;
}

void
SUBool::CompHypothesis::ClearHypothesis()
{
   std::lock_guard<SharedMutex> lock(mMutex);
   mHypothesis.Clear();
   mSafe = false;
   mHypMinimized = true;
   mOneProveHyp = OneProveM();
   mImplSystemInitialized = false;
   ResetSolver();
}

void
SUBool::CompHypothesis::SetHypothesis(CNF cnf, bool safe)
{
   std::lock_guard<SharedMutex> lock(mMutex);
   mHypothesis = std::move(cnf);
   mHypMinimized = false;
   mOneProveHyp = OneProveM(mHypothesis);
   mSafe = safe;
   mImplSystemInitialized = false;
   ResetSolver();
}

void
SUBool::CompHypothesis::SetHypothesisToRepresentation()
{
   SetHypothesis(GetRepresentation(false), true);
   std::lock_guard<SharedMutex> lock(mMutex);
   mHypMinimized = mReprMinimized;
}

SUBool::CNF
SUBool::CompHypothesis::AddBackbones(CNF cnf) const
{
   for (const auto &lit : mBackbone)
   {
      cnf.PushBack(Clause(lit));
   }
   return cnf;
}

SUBool::CNF
SUBool::CompHypothesis::GetHypothesis(bool uncompress) const &
{
   SharedLockGuard lock(mMutex);
   if (uncompress)
   {
      return AddBackbones(mCompressor.UncompressCNF(mHypothesis));
   }
   return mHypothesis;
}

SUBool::CNF
SUBool::CompHypothesis::GetHypothesis(bool uncompress) &&
{
   std::lock_guard<SharedMutex> lock(mMutex);
   if (uncompress)
   {
      return AddBackbones(mCompressor.UncompressCNF(std::move(mHypothesis)));
   }
   return std::move(mHypothesis);
}

void
SUBool::CompHypothesis::InnerSetRepresentation(CNF cnf)
{
   mRepresentation = std::move(cnf);
   mReprMinimized = false;
   mSafe = false;
   mOneProveRepr = OneProveM(mRepresentation);
   ResetSolver();
}

SUBool::CNF
SUBool::CompHypothesis::GetRepresentation(bool uncompress) const &
{
   SharedLockGuard lock(mMutex);
   if (uncompress)
   {
      return AddBackbones(mCompressor.UncompressCNF(mRepresentation));
   }
   return mRepresentation;
}

SUBool::CNF
SUBool::CompHypothesis::GetRepresentation(bool uncompress) &&
{
   std::lock_guard<SharedMutex> lock(mMutex);
   if (uncompress)
   {
      return AddBackbones(
          mCompressor.UncompressCNF(std::move(mRepresentation)));
   }
   return std::move(mRepresentation);
}

void
SUBool::CompHypothesis::MakeReprPrime()
{
   std::lock_guard<SharedMutex> lock(mMutex);
   auto prime = prime_cnf(mSolver.Solver(), mRepresentation);
   InnerSetRepresentation(std::move(prime));
}

void
SUBool::CompHypothesis::MakeReprUPPrime()
{
   std::lock_guard<SharedMutex> lock(mMutex);
   auto prime = up_prime_cnf(std::move(mRepresentation));
   InnerSetRepresentation(std::move(prime));
}

void
SUBool::CompHypothesis::SetSafe()
{
   std::lock_guard<SharedMutex> lock(mMutex);
   if (mSafe == false)
   {
      mSafe = true;
      ResetSolver();
   }
}

bool
SUBool::CompHypothesis::Safe() const
{
   SharedLockGuard lock(mMutex);
   return mSafe;
}

std::optional<SUBool::PartialAssignment>
SUBool::CompHypothesis::PropagateUP(const PartialAssignment &alpha) const
{
   std::lock_guard<SharedMutex> lock(mMutex);
   return mOneProveHyp.Propagate(alpha);
}

const SUBool::CNF &
SUBool::CompHypothesis::TheRightCNF() const
{
   return mSafe ? mHypothesis : mRepresentation;
}

std::optional<SUBool::PartialAssignment>
SUBool::CompHypothesis::PropagateAllSAT(const PartialAssignment &alpha) const
{
   std::lock_guard lock(mMutex);
   return mSemPropagator.Propagate(alpha);
}

bool
SUBool::CompHypothesis::UPClosed(const PartialAssignment &alpha) const
{
   std::lock_guard lock(mMutex);
   return mOneProveHyp.Closed(alpha);
}

std::optional<SUBool::PartialAssignment>
SUBool::CompHypothesis::Backbones()
{
   std::lock_guard lock(mMutex);
   return mSemPropagator.Backbones();
}

bool
SUBool::CompHypothesis::AddLearnt(Clause clause)
{
   auto insert_result = mLearnedCache.insert(clause);
   if (!insert_result.second)
   {
      return false;
   }
   mLearnedQueue.push(std::move(clause));
   return true;
}

bool
SUBool::CompHypothesis::HasLearnedClause() const
{
   return !mLearnedQueue.empty() || mSolver.Solver().LearntsSize() > 0;
}

size_t
SUBool::CompHypothesis::LearnedSize() const
{
   return mLearnedQueue.size() + mSolver.Solver().LearntsSize();
}

std::optional<SUBool::Clause>
SUBool::CompHypothesis::NextLearnedClause()
{
   if (mLearnedQueue.empty() && mSolver.Solver().LearntsSize() > 0)
   {
      ResetSolver();
   }
   if (mLearnedQueue.empty())
   {
      return {};
   }
   auto c = mLearnedQueue.front();
   mLearnedQueue.pop();
   return std::make_optional(std::move(c));
}

void
SUBool::CompHypothesis::ExtractLearnts()
{
   auto learnts = mSolver.Solver().Learnts();
   logs.TLog(3) << "Extracted " << learnts.size() << " learnts from the solver"
                << std::endl;
   size_t cnt = 0;
   for (auto &clause : learnts)
   {
      if (AddLearnt(std::move(clause)))
      {
         ++cnt;
      }
   }
   logs.TLog(3) << "Added " << cnt << " learnts" << std::endl;
}

void
SUBool::CompHypothesis::CompressLearnts(const PartialAssignment &alpha)
{
   auto learned_cache = std::move(mLearnedCache);
   size_t cache_cnt = 0;
   size_t cache_total = learned_cache.size();
   for (const auto &clause : learned_cache)
   {
      auto c = clause;
      switch (c.Apply(alpha))
      {
      case PartialValue::UNDEF:
         if (mLearnedCache.insert(mCompressor.CompressClause(std::move(c)))
                 .second)
         {
            ++cache_cnt;
         }
         break;
      case PartialValue::FALSE:
         throw InvariantFailedException(
             "CompHypothesis::CompressLearnts", "Formula is unsatisfiable");
      case PartialValue::TRUE:
      default:
         break;
      }
   }
   auto learned_queue = std::move(mLearnedQueue);
   mLearnedQueue = std::queue<Clause>();
   size_t queue_cnt = 0;
   size_t queue_total = learned_queue.size();
   while (!learned_queue.empty())
   {
      auto c = learned_queue.front();
      learned_queue.pop();
      switch (c.Apply(alpha))
      {
      case PartialValue::UNDEF:
         mLearnedQueue.push(mCompressor.CompressClause(std::move(c)));
         ++queue_cnt;
         break;
      case PartialValue::FALSE:
         throw InvariantFailedException(
             "CompHypothesis::CompressLearnts", "Formula is unsatisfiable");
      case PartialValue::TRUE:
      default:
         break;
      }
   }
   logs.TLog(3) << "Compressed learnts (cache: " << cache_cnt << "/"
                << cache_total << ", queue: " << queue_cnt << "/" << queue_total
                << ")" << std::endl;
}

void
SUBool::CompHypothesis::PropagatePartialAssignment(
    const PartialAssignment &alpha)
{
   CPUTime cpu_time;
   logs.TLog(2) << "Propagating backbones and compressing." << std::endl;
   mRepresentation.Apply(alpha);
   mHypothesis.Apply(alpha);
   mCompressor = CNFCompressor(mRepresentation);
   mRepresentation = mCompressor.CompressCNF(std::move(mRepresentation));
   mOneProveRepr = OneProveM(mRepresentation);
   mHypothesis = mCompressor.CompressCNF(std::move(mHypothesis));
   if (mRepresentation.Empty() && mHypothesis.Empty())
   {
      mSafe = true;
      mHypMinimized = true;
      mReprMinimized = true;
   }
   else
   {
      mReprMinimized = false;
      mHypMinimized = false;
   }
   mOneProveHyp = OneProveM(mHypothesis);
   mImplSystemInitialized = false;
   ResetSolver();
   CompressLearnts(alpha);
   logs.TLog(2) << "Backbones propagated, compressed cnf (p cnf "
                << mRepresentation.MaxVariable() << " "
                << mRepresentation.Size() << "), time=" << cpu_time
                << std::endl;
}

bool
SUBool::CompHypothesis::PropagateBackbones(bool use_sat)
{
   CPUTime cpu_time;
   logs.TLog(2) << "Finding " << (use_sat ? "" : "UP") << " backbones (p cnf "
                << mRepresentation.MaxVariable() << " "
                << mRepresentation.Size() << ")" << std::endl;
   auto alpha = (use_sat ? mSemPropagator.Backbones()
                         : mOneProveRepr.Propagate(PartialAssignment(0)));
   if (!alpha)
   {
      logs.TLog(2) << "No backbones found, the formula is unsatisfiable, time="
                   << cpu_time << std::endl;
      return false;
   }
   mBackbone = literals_of_assignment(*alpha);
   logs.TLog(2) << "Found " << mBackbone.size()
                << " backbones, time=" << cpu_time << std::endl;
   PropagatePartialAssignment(*alpha);
   return true;
}

bool
SUBool::CompHypothesis::PropagateBackbones()
{
   std::lock_guard lock(mMutex);
   return PropagateBackbones(true);
}

bool
SUBool::CompHypothesis::PropagateUPBackbones()
{
   std::lock_guard lock(mMutex);
   return PropagateBackbones(false);
}

std::optional<std::vector<SUBool::Literal>>
SUBool::CompHypothesis::AllEmpoweredLiterals(const Clause &clause) const
{
   std::lock_guard<SharedMutex> lock(mMutex);
   return mOneProveHyp.AllEmpoweredLiterals(clause);
}

std::optional<SUBool::Literal>
SUBool::CompHypothesis::EmpoweredLiteral(const Clause &clause) const
{
   std::lock_guard<SharedMutex> lock(mMutex);
   return mOneProveHyp.EmpoweredLiteral(clause);
}

bool
SUBool::CompHypothesis::InnerMinimizeHypothesis()
{
   if (mHypMinimized)
   {
      return false;
   }
   MinimizeCNF(true);
   mHypMinimized = true;
   if (mImplSystemRebuild == IMPL_SYSTEM_REBUILD::MINIMIZED
       || mImplSystemRebuild == IMPL_SYSTEM_REBUILD::ALWAYS)
   {
      mImplSystemInitialized = false;
   }
   return true;
}

bool
SUBool::CompHypothesis::MinimizeHypothesis()
{
   std::lock_guard<SharedMutex> lock(mMutex);
   return InnerMinimizeHypothesis();
}

bool
SUBool::CompHypothesis::MinimizeRepresentation()
{
   std::lock_guard<SharedMutex> lock(mMutex);
   if (mReprMinimized)
   {
      return false;
   }
   MinimizeCNF(false);
   mReprMinimized = true;
   return true;
}

std::vector<SUBool::ClauseWithEmpowered>
SUBool::CompHypothesis::MinimizeReprWithCollectingEmpowered(bool collect_all)
{
   std::lock_guard<SharedMutex> lock(mMutex);
   auto clauses_with_emp = mOneProveRepr.MinimizeWithCollectingEmpowered(
       collect_all, MinimizationLogProgress(mRepresentation.Size()));
   mRepresentation = mOneProveRepr.ExtractCNF();
   mReprMinimized = true;
   return clauses_with_emp;
}

void
SUBool::CompHypothesis::MinimizeCNF(bool use_hypothesis)
{
   CPUTime cpu_time;
   auto name = (use_hypothesis ? "hypothesis" : "representation");
   auto &cnf = (use_hypothesis ? mHypothesis : mRepresentation);
   auto &op = (use_hypothesis ? mOneProveHyp : mOneProveRepr);
   logs.TLog(2) << "Minimizing " << name << " (p cnf " << cnf.MaxVariable()
                << " " << cnf.Size() << ")" << std::endl;
   bool reprocess = true;
   switch (mMinimizeMethod)
   {
   case MINIMIZE_METHOD::FULL:
      op.MinimizeFull(MinimizationLogProgress(cnf.Size()));
      break;
   case MINIMIZE_METHOD::SIMPLE_INCREMENTAL:
      reprocess = false;
      [[fallthrough]];
   case MINIMIZE_METHOD::FULL_INCREMENTAL:
      op = OneProveM();
      op.AddWithIncrementalMinimization(
          cnf, reprocess, MinimizationLogProgress(cnf.Size()));
      break;
   default:
      throw InvariantFailedException(
          "CompHypothesis::MinimizeCNF", "unknown minimization method");
   }
   cnf = op.ExtractCNF();
   if (use_hypothesis == mSafe)
   {
      ResetSemPropagator();
   }
   logs.TLog(2) << "Finished minimization of " << name << " (p cnf "
                << cnf.MaxVariable() << " " << cnf.Size()
                << "), time=" << cpu_time << std::endl;
}

bool
SUBool::CompHypothesis::IsEmpowering(const Clause &c)
{
   std::lock_guard<SharedMutex> lock(mMutex);
   return mOneProveHyp.IsEmpowering(c);
}

bool
SUBool::CompHypothesis::IsImplicate(const Clause &clause) const
{
   SharedLockGuard lock(mMutex);
   return is_implicate(mSolver.Solver(), clause);
}

std::optional<SUBool::Clause>
SUBool::CompHypothesis::PrimeSubimplicate(
    const Clause &clause, unsigned ignore_var, bool promised_sat) const
{
   SharedLockGuard lock(mMutex);
   return prime_subimplicate(
       mSolver.Solver(), clause, ignore_var, promised_sat);
}

std::optional<SUBool::Clause>
SUBool::CompHypothesis::FindEmpoweringQuery(AbsCheckerSAT &checker)
{
   SharedLockGuard lock(mMutex);
   ClausesToLiterals cl2lit(mHypothesis);
   UpdateISProvider();
   if (checker.RequiresSCCGraph())
   {
      mImplSystemProvider->RebuildSCCGraph();
   }
   return checker.FindEmpowering(mHypothesis, cl2lit, mImplSystemProvider);
}

void
SUBool::CompHypothesis::UpdateISProvider()
{
   if (!mImplSystemInitialized)
   {
      logs.TLog(3) << "[CompHypothesis::FindEmpoweringQuery] Initializing "
                      "the implication system"
                   << std::endl;
      mImplSystemProvider->Cleanup(mHypothesis);
      mImplSystemInitialized = true;
   }
}

const SUBool::EncISProvider &
SUBool::CompHypothesis::ImplSystemProvider()
{
   SharedLockGuard lock(mMutex);
   UpdateISProvider();
   return *mImplSystemProvider;
}

std::optional<SUBool::Clause>
SUBool::CompHypothesis::FindEmpowering(
    AbsCheckerSAT &checker, bool level_minimize)
{
   std::optional<Clause> emp_result = FindEmpoweringQuery(checker);
   if (emp_result || !checker.WillIncreaseLevel())
   {
      return emp_result;
   }
   if (level_minimize)
   {
      MinimizeHypothesis();
   }
   return FindEmpoweringQuery(checker);
}

void
SUBool::CompHypothesis::PushBack(Clause c)
{
   std::lock_guard<SharedMutex> lock(mMutex);
   // c is always implicate of mRepresentation, so
   // adding it to the solver is always safe
   mSolver.Solver().AddClause(c);
   mHypothesis.PushBack(c);
   mHypMinimized = false;
   mImplSystemProvider->AddClause(c);
   if (mImplSystemRebuild == IMPL_SYSTEM_REBUILD::CLAUSE_ADDED
       || mImplSystemRebuild == IMPL_SYSTEM_REBUILD::ALWAYS)
   {
      mImplSystemInitialized = false;
   }
   mOneProveHyp.AddClause(std::move(c));
   if (mSafe)
   {
      ResetSemPropagator();
   }
}

bool
SUBool::CompHypothesis::PushIfEmpowering(
    const Clause &c, bool find_prime_subimplicate)
{
   auto l = EmpoweredLiteral(c);
   if (!l)
   {
      return false;
   }
   if (find_prime_subimplicate)
   {
      auto prime = PrimeSubimplicate(c, l->Variable(), true);
      if (!prime)
      {
         throw InvariantFailedException(
             "CompHypothesis::PushIfEmpowering", "Not an implicate");
      }
      PushBack(std::move(*prime));
   }
   else
   {
      PushBack(c);
   }
   return true;
}

unsigned
SUBool::CompHypothesis::MaxVariable() const
{
   SharedLockGuard lock(mMutex);
   return mHypothesis.MaxVariable();
}

size_t
SUBool::CompHypothesis::Size() const
{
   SharedLockGuard lock(mMutex);
   return mHypothesis.Size();
}

bool
SUBool::CompHypothesis::Empty() const
{
   SharedLockGuard lock(mMutex);
   return TheRightCNF().Empty();
}

unsigned
SUBool::CompHypothesis::ReprMaxVariable() const
{
   SharedLockGuard lock(mMutex);
   return mRepresentation.MaxVariable();
}

size_t
SUBool::CompHypothesis::ReprSize() const
{
   SharedLockGuard lock(mMutex);
   return mRepresentation.Size();
}

void
SUBool::CompHypothesis::Append(CNF cnf)
{
   for (auto &clause : cnf)
   {
      PushBack(std::move(clause));
   }
}
