/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file comp_hypothesis.h
 * Class representing a hypothesis in the compilation algorithms (both iterative
 * and learning). The access is guarded by a mutex.
 */

#ifndef __COMP_HYPOTHESIS_H
#define __COMP_HYPOTHESIS_H

#include "cnfcompress.h"
#include "drenc.h"
#include "empower2.h"
#include "enc_is_provider.h"
#include "pcchecker.h"
#include "propagate.h"
#include "smutex.h"

namespace SUBool
{

   enum class IMPL_SYSTEM_REBUILD
   {
      CLAUSE_ADDED,
      MINIMIZED,
      ALWAYS,
      Default = CLAUSE_ADDED
   };

   inline const EnumAnnotation<IMPL_SYSTEM_REBUILD>
       kImplSystemRebuildAnnotation("IMPL_SYSTEM_REBUILD", //
           {
               //
               {IMPL_SYSTEM_REBUILD::CLAUSE_ADDED, "clause_added",
                   "rebuild the implication system after a clause has been "
                   "added "
                   "to the hypothesis"}, //
               {IMPL_SYSTEM_REBUILD::MINIMIZED, "minimized",
                   "rebuild the implication system after the hypothesis has "
                   "been minimized"}, //
               {IMPL_SYSTEM_REBUILD::ALWAYS, "always",
                   "rebuild the implication system before each equivalence "
                   "test, but not after "} //
           });

   class CompHypothesis
   {
      CNF mRepresentation;
      CNF mHypothesis;
      IncSolverWrapper mSolver;
      bool mSafe;
      // should rename it and the propagator could be semantical?
      OneProveM mOneProveRepr;
      OneProveM mOneProveHyp;
      MinimalSubimplicant mMinSubimplicant;
      Propagator mSemPropagator;
      mutable SharedMutex mMutex;
      bool mHypMinimized{false};
      bool mReprMinimized{false};
      MINIMIZE_METHOD mMinimizeMethod{MINIMIZE_METHOD::FULL};
      std::vector<Literal> mBackbone{};
      CNFCompressor mCompressor{};
      std::shared_ptr<EncISProvider> mImplSystemProvider{};
      std::atomic<bool> mImplSystemInitialized{false};
      IMPL_SYSTEM_REBUILD mImplSystemRebuild{IMPL_SYSTEM_REBUILD::CLAUSE_ADDED};

      std::unordered_set<Clause> mLearnedCache{};
      std::queue<Clause> mLearnedQueue{};

      const CNF &TheRightCNF() const;

      static size_t MinimizationLogProgress(size_t cnf_size);
      void MinimizeCNF(bool use_hypothesis);
      void InnerSetRepresentation(CNF cnf);
      void ResetSolver();
      void ResetSemPropagator();
      CNF AddBackbones(CNF cnf) const;
      void ExtractLearnts();
      void CompressLearnts(const PartialAssignment &alpha);
      bool AddLearnt(Clause clause);
      bool InnerMinimizeHypothesis(); // true if minimization executed
      void UpdateISProvider();
      std::optional<Clause> FindEmpoweringQuery(AbsCheckerSAT &checker);
      bool PropagateBackbones(bool use_sat);
      void PropagatePartialAssignment(const PartialAssignment &alpha);

    public:
      static const unsigned kMinimizationLogThreshold = 3000;

      CompHypothesis(MINIMIZE_METHOD minimize_method);
      CompHypothesis(
          MINIMIZE_METHOD minimize_method, CNF cnf, CNF repr, bool safe);
      ~CompHypothesis() = default;

      void SetImplSystemRebuild(IMPL_SYSTEM_REBUILD impl_system_rebuild);

      void ClearHypothesis();
      void SetHypothesis(CNF cnf, bool safe);
      void SetHypothesisToRepresentation();
      CNF GetHypothesis(bool uncompress) const &;
      CNF GetHypothesis(bool uncompress) &&;
      CNF GetRepresentation(bool uncompress) const &;
      CNF GetRepresentation(bool uncompress) &&;
      bool
      HypMinimized() const
      {
         SharedLockGuard lock(mMutex);
         return mHypMinimized;
      }
      template <typename F>
      auto
      CallOnRepresention(F f) const -> decltype(f(mRepresentation))
      {
         SharedLockGuard lock(mMutex);
         return f(mRepresentation);
      }

      void MakeReprPrime();
      void MakeReprUPPrime();

      // Should do with a semantic propagator on the representation
      std::optional<PartialAssignment> Backbones();
      // Initializes mBacbone and mCompressor, return false if the formula is
      // unsatisfiable.
      bool PropagateBackbones();
      // Like PropagateBackbones, but only propagates literals which can be
      // derived by unit propagation
      bool PropagateUPBackbones();
      void SetSafe();
      bool Safe() const;

      std::optional<PartialAssignment> PropagateUP(
          const PartialAssignment &alpha) const;
      bool UPClosed(const PartialAssignment &alpha) const;
      std::optional<PartialAssignment> PropagateAllSAT(
          const PartialAssignment &alpha) const;
      std::optional<std::vector<Literal>> AllEmpoweredLiterals(
          const Clause &clause) const;
      std::optional<Literal> EmpoweredLiteral(const Clause &clause) const;
      bool MinimizeHypothesis();     // true if minimization executed
      bool MinimizeRepresentation(); // true if minimization executed
      std::vector<ClauseWithEmpowered> MinimizeReprWithCollectingEmpowered(
          bool collect_all);
      const CNFCompressor &
      Compressor() const
      {
         return mCompressor;
      }
      const EncISProvider &ImplSystemProvider();

      std::optional<Clause> NextLearnedClause();
      bool HasLearnedClause() const;
      size_t LearnedSize() const;

      bool IsEmpowering(const Clause &c);
      bool IsImplicate(const Clause &clause) const;
      std::optional<Clause> PrimeSubimplicate(
          const Clause &clause, unsigned ignore_var, bool promised_sat) const;
      std::optional<Clause> FindEmpowering(
          AbsCheckerSAT &checker, bool level_minimize);
      void PushBack(Clause c);
      // Returns true if pushed
      bool PushIfEmpowering(const Clause &c, bool find_prime_subimplicate);
      void Append(CNF cnf);
      template <typename InputIterator>
      void
      Append(InputIterator begin, InputIterator end)
      {
         std::lock_guard<SharedMutex> lock(mMutex);
         mHypothesis.Append(begin, end);
         mHypMinimized = false;
      }

      unsigned MaxVariable() const;
      size_t Size() const;
      bool Empty() const;
      unsigned ReprMaxVariable() const;
      size_t ReprSize() const;

      void SetImplSystemProvider(std::shared_ptr<EncISProvider> is_provider);
   };
} // namespace SUBool

inline void
SUBool::CompHypothesis::SetImplSystemProvider(
    std::shared_ptr<EncISProvider> is_provider)
{
   mImplSystemProvider = std::move(is_provider);
}

#endif /* end of include guard */
