/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file dimacs.h
 * DIMACS format helper functions.
 */

#ifndef __DIMACS_H
#define __DIMACS_H

#include <algorithm>

#include "literal.h"
#include "nfliteralset.h"
#include "normalform.h"
#include "sbexception.h"
#include "util.h"

namespace SUBool
{
   namespace Dimacs
   {
      // Skips comment and empty lines
      template <class InputIterator>
      InputIterator skip_comment_lines(InputIterator begin, InputIterator end);

      // blanks mean spaces and tab characters (isblank)
      template <class InputIterator>
      InputIterator skip_blanks(InputIterator begin, InputIterator end);

      // spaces include \r and \n
      template <class InputIterator>
      InputIterator skip_spaces(InputIterator begin, InputIterator end);

      template <class InputIterator>
      InputIterator skip_string(
          const std::string &s, InputIterator begin, InputIterator end);

      template <class InputIterator> struct ParsedUnsigned
      {
         unsigned value;
         InputIterator next_pos;
      };

      template <class InputIterator>
      ParsedUnsigned<InputIterator> parse_unsigned(
          InputIterator begin, InputIterator end);

      template <class InputIterator> struct ParsedLiteral
      {
         std::optional<Literal> lit;
         bool zero;
         InputIterator next_pos;
      };

      template <class InputIterator>
      ParsedLiteral<InputIterator> parse_literal(
          InputIterator begin, InputIterator end);

      template <class InputIterator> struct ParsedClause
      {
         std::optional<Clause> clause;
         bool zero;
         InputIterator next_pos;
      };

      template <class InputIterator>
      ParsedClause<InputIterator> parse_clause(
          InputIterator begin, InputIterator end);

      template <class InputIterator> struct ParsedCNFHeader
      {
         unsigned max_var;
         unsigned clause_size;
         InputIterator next_pos;
      };

      template <class InputIterator>
      ParsedCNFHeader<InputIterator> parse_cnf_header(
          InputIterator begin, InputIterator end);

      template <class InputIterator>
      CNF parse_cnf(InputIterator begin, InputIterator end);

      Literal parse_literal(std::istream &s);
      Clause parse_clause(std::istream &s);
      CNF parse_cnf(std::istream &s);

      Literal parse_literal(const std::string &s);
      Clause parse_clause(const std::string &s);
      CNF parse_cnf(const std::string &s);
   } // namespace Dimacs

   template <>
   inline CNF
   read_from_stream(std::istream &s)
   {
      return Dimacs::parse_cnf(s);
   }

} // namespace SUBool

// Template implementation
template <class InputIterator>
InputIterator
SUBool::Dimacs::skip_comment_lines(InputIterator begin, InputIterator end)
{
   begin = skip_spaces(begin, end);
   while (begin != end && *begin == 'c')
   {
      while (begin != end && *begin != '\n')
      {
         ++begin;
      }
      begin = skip_spaces(begin, end);
   }
   return begin;
}

template <class InputIterator>
InputIterator
SUBool::Dimacs::skip_blanks(InputIterator begin, InputIterator end)
{
   return std::find_if_not(
       begin, end, [](auto c) { return c != '\n' && std::isspace(c); });
}

template <class InputIterator>
InputIterator
SUBool::Dimacs::skip_spaces(InputIterator begin, InputIterator end)
{
   return std::find_if_not(begin, end, [](auto c) { return std::isspace(c); });
}

template <class InputIterator>
SUBool::Dimacs::ParsedLiteral<InputIterator>
SUBool::Dimacs::parse_literal(InputIterator begin, InputIterator end)
{
   begin = skip_spaces(begin, end);
   if (begin == end)
   {
      return {{}, false, begin};
   }
   bool positive = true;
   if (*begin == '-')
   {
      positive = false;
      ++begin;
      if (begin == end)
      {
         throw ParseException("Dimacs::parse_literal", "Nothing after a minus");
      }
   }
   unsigned var = 0;
   unsigned char_count = 0;
   while (begin != end && std::isdigit(*begin))
   {
      ++char_count;
      var = var * 10 + (*begin - '0');
      ++begin;
   }
   if (begin != end && !std::isspace(*begin))
   {
      throw ParseException("Dimacs::parse_literal",
          "Literal must be followed by a space, newline, or the end of file");
   }
   if (char_count == 0)
   {
      throw ParseException("Dimacs::parse_literal", "No digits in a literal");
   }
   if (char_count == 1 && var == 0)
   {
      if (!positive)
      {
         throw ParseException("Dimacs::parse_literal",
             "Zero variable is not allowed and the "
             "zero delimiter must not be negative");
      }
      return {{}, true, begin};
   }
   return {Literal(var, positive), false, begin};
}

template <class InputIterator>
SUBool::Dimacs::ParsedClause<InputIterator>
SUBool::Dimacs::parse_clause(InputIterator begin, InputIterator end)
{
   std::vector<Literal> literals;
   bool zero_reached = false;
   while (begin != end)
   {
      auto parsed_lit = parse_literal(begin, end);
      begin = parsed_lit.next_pos;
      zero_reached = parsed_lit.zero;
      if (zero_reached || !parsed_lit.lit)
      {
         break;
      }
      literals.push_back(*parsed_lit.lit);
   }
   if (!zero_reached && literals.empty())
   {
      return {{}, false, begin};
   }
   return {literals, zero_reached, begin};
}

template <class InputIterator>
SUBool::CNF
SUBool::Dimacs::parse_cnf(InputIterator begin, InputIterator end)
{
   begin = skip_comment_lines(begin, end);
   auto parsed_header = parse_cnf_header(begin, end);
   begin = parsed_header.next_pos;
   std::vector<Clause> clauses;
   clauses.reserve(parsed_header.clause_size);
   unsigned clause_count = 0;
   while (begin != end)
   {
      begin = skip_comment_lines(begin, end);
      auto parsed_clause = parse_clause(begin, end);
      begin = parsed_clause.next_pos;
      if (!parsed_clause.clause)
      {
         break;
      }
      if (!parsed_clause.clause->Trivial())
      {
         clauses.push_back(*parsed_clause.clause);
      }
      ++clause_count;
      if (!parsed_clause.zero)
      {
         break;
      }
   }
   CNF cnf(clauses);
   if (clause_count != parsed_header.clause_size)
   {
      logs.TLog(1) << "[Dimacs::parse_cnf] WARNING: Clause size does not match "
                      "the header value"
                   << std::endl;
   }
   if (cnf.MaxVariable() != parsed_header.max_var)
   {
      logs.TLog(1) << "[Dimacs::parse_cnf] WARNING: Number of variables does "
                      "not match the header value"
                   << std::endl;
   }
   return cnf;
}

template <class InputIterator>
SUBool::Dimacs::ParsedCNFHeader<InputIterator>
SUBool::Dimacs::parse_cnf_header(InputIterator begin, InputIterator end)
{
   begin = skip_string("p", begin, end);
   begin = skip_string("cnf", begin, end);
   auto parsed_max_var = parse_unsigned(begin, end);
   begin = parsed_max_var.next_pos;
   auto parsed_clause_size = parse_unsigned(begin, end);
   begin = parsed_clause_size.next_pos;
   begin = skip_blanks(begin, end);
   if (begin != end)
   {
      if (*begin != '\n')
      {
         throw ParseException(
             "Dimacs::parse_cnf_header", "Extra characters on the header line");
      }
      ++begin;
   }
   return {parsed_max_var.value, parsed_clause_size.value, begin};
}

template <class InputIterator>
InputIterator
SUBool::Dimacs::skip_string(
    const std::string &s, InputIterator begin, InputIterator end)
{
   begin = skip_blanks(begin, end);
   for (const auto &c : s)
   {
      if (begin == end)
      {
         throw ParseException(
             "Dimacs::skip_string", "Input finished prematurely");
      }
      if (*begin != c)
      {
         throw ParseException("Dimacs::skip_string", "Character mismatch");
      }
      ++begin;
   }
   return begin;
}

template <class InputIterator>
SUBool::Dimacs::ParsedUnsigned<InputIterator>
SUBool::Dimacs::parse_unsigned(InputIterator begin, InputIterator end)
{
   begin = skip_blanks(begin, end);
   unsigned val = 0;
   bool has_char = false;
   while (begin != end && std::isdigit(*begin))
   {
      val = val * 10 + (*begin - '0');
      ++begin;
      has_char = true;
   }
   if (!has_char)
   {
      throw ParseException("Dimacs::parse_unsigned", "No character");
   }
   return {val, begin};
}

inline SUBool::Literal
SUBool::Dimacs::parse_literal(const std::string &s)
{
   auto parsed_literal = parse_literal(s.begin(), s.end());
   if (!parsed_literal.lit)
   {
      throw ParseException("Dimacs::parse_literal", "Failed to parse literal");
   }
   return *parsed_literal.lit;
}

inline SUBool::Clause
SUBool::Dimacs::parse_clause(const std::string &s)
{
   auto parsed_clause = parse_clause(s.begin(), s.end());
   if (!parsed_clause.clause)
   {
      throw ParseException("Dimacs::parse_clause", "Failed to parse clause");
   }
   return *parsed_clause.clause;
}

inline SUBool::CNF
SUBool::Dimacs::parse_cnf(const std::string &s)
{
   return parse_cnf(s.begin(), s.end());
}

inline SUBool::Literal
SUBool::Dimacs::parse_literal(std::istream &s)
{
   std::istreambuf_iterator<char> begin(s), end;
   auto parsed_literal = parse_literal(begin, end);
   if (!parsed_literal.lit)
   {
      throw ParseException("Dimacs::parse_literal", "Failed to parse literal");
   }
   return *parsed_literal.lit;
}

inline SUBool::Clause
SUBool::Dimacs::parse_clause(std::istream &s)
{
   std::istreambuf_iterator<char> begin(s), end;
   auto parsed_clause = parse_clause(begin, end);
   if (!parsed_clause.clause)
   {
      throw ParseException("Dimacs::parse_clause", "Failed to parse clause");
   }
   return *parsed_clause.clause;
}

inline SUBool::CNF
SUBool::Dimacs::parse_cnf(std::istream &s)
{
   std::istreambuf_iterator<char> begin(s), end;
   return parse_cnf(begin, end);
}

#endif
