/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file dnnf.cpp
 * DNNF traits checking, manipulation, and propagation.
 */

#include <algorithm>
#include <memory>
#include <stack>

#include "dnnf.h"
#include "nnf.h"
#include "nnfvisit.h"
#include "sbexception.h"

namespace SUBool
{
   class DnnfSmooth : public NnfVisitorMut
   {
      NnfNode::tValueCache<std::vector<unsigned>> mNodeVariables;
      std::unordered_map<Literal, std::shared_ptr<NnfLiteral>> mLiteralNodes;
      Nnf mNnf;
      std::unordered_map<unsigned, NnfNode::tPointer> mDummyOrNodes;

      void MakeNodeSmooth(const NnfNode::tPointer &node, unsigned skip_var = 0);
      NnfNode::tPointer AddInputsToNode(const NnfNode::tPointer &node,
                                        const std::vector<unsigned> &inputs,
                                        unsigned skip_var = 0);
      NnfNode::tPointer DummyOr(unsigned var);
      NnfNode::tPointer LiteralNode(const Literal &lit);

    public:
      DnnfSmooth(Nnf nnf,
                 NnfNode::tValueCache<std::vector<unsigned>> node_variables,
                 std::unordered_map<Literal, std::shared_ptr<NnfLiteral>>
                     literal_nodes);
      virtual ~DnnfSmooth() = default;

      virtual void
      VisitLiteral(const std::shared_ptr<NnfLiteral> &node) override;
      virtual void VisitAnd(const std::shared_ptr<NnfGateAnd> &node) override;
      virtual void VisitOr(const std::shared_ptr<NnfGateOr> &node) override;
      virtual void
      VisitDecision(const std::shared_ptr<NnfGateDecision> &node) override;
      virtual void VisitTrue(const std::shared_ptr<NnfTrue> &node) override;
      virtual void VisitFalse(const std::shared_ptr<NnfFalse> &node) override;

      void MakeSmooth();

      const Nnf &
      GetNnf() const &
      {
         return mNnf;
      }

      Nnf &&
      GetNnf() &&
      {
         return std::move(mNnf);
      }
   };
} // namespace SUBool

SUBool::NnfTraits
SUBool::nnf_traits(const Nnf &nnf)
{
   NnfNodeInputsVisitor visitor;
   nnf.AcceptPostorder(visitor);
   return visitor.NodeTraits(nnf.Root());
}

SUBool::DnnfSmooth::DnnfSmooth(
    Nnf nnf, NnfNode::tValueCache<std::vector<unsigned>> node_variables,
    std::unordered_map<Literal, std::shared_ptr<NnfLiteral>> literal_nodes)
    : mNodeVariables(std::move(node_variables)),
      mLiteralNodes(std::move(literal_nodes)), mNnf(std::move(nnf)),
      mDummyOrNodes()
{
}

void
SUBool::DnnfSmooth::VisitLiteral(const std::shared_ptr<NnfLiteral> & /*node*/)
{
}

void
SUBool::DnnfSmooth::VisitAnd(const std::shared_ptr<NnfGateAnd> & /*node*/)
{
}

void
SUBool::DnnfSmooth::VisitOr(const std::shared_ptr<NnfGateOr> &node)
{
   MakeNodeSmooth(node);
}

void
SUBool::DnnfSmooth::VisitDecision(const std::shared_ptr<NnfGateDecision> &node)
{
   MakeNodeSmooth(node, node->DecisionVar());
}

void
SUBool::DnnfSmooth::VisitTrue(const std::shared_ptr<NnfTrue> & /*node*/)
{
}

void
SUBool::DnnfSmooth::VisitFalse(const std::shared_ptr<NnfFalse> & /*node*/)
{
}

SUBool::NnfNode::tPointer
SUBool::DnnfSmooth::LiteralNode(const Literal &lit)
{
   if (mLiteralNodes.count(lit) > 0)
   {
      return mLiteralNodes[lit];
   }
   return mLiteralNodes[lit] =
              std::dynamic_pointer_cast<NnfLiteral>(NnfNode::MakeLiteral(lit));
}

SUBool::NnfNode::tPointer
SUBool::DnnfSmooth::DummyOr(unsigned var)
{
   if (mDummyOrNodes.count(var) > 0)
   {
      return mDummyOrNodes[var];
   }
   return mDummyOrNodes[var] =
              NnfNode::MakeOr(var, LiteralNode(Literal(var, true)),
                              LiteralNode(Literal(var, false)));
}

SUBool::NnfNode::tPointer
SUBool::DnnfSmooth::AddInputsToNode(const NnfNode::tPointer &node,
                                    const std::vector<unsigned> &inputs,
                                    unsigned skip_var)
{
   std::vector<NnfNode::tPointer> and_inputs{node};
   for (auto var : inputs)
   {
      if (skip_var != var)
      {
         and_inputs.push_back(DummyOr(var));
      }
   }
   return NnfNode::MakeAnd(std::move(and_inputs));
}

void
SUBool::DnnfSmooth::MakeNodeSmooth(const NnfNode::tPointer &node,
                                   unsigned skip_var)
{
   const auto &inputs = node->CachedValueConstRef(mNodeVariables);
   std::vector<unsigned> diff;
   for (auto i_child = node->begin(); i_child != node->end(); ++i_child)
   {
      const auto &child_inputs =
          (*i_child)->CachedValueConstRef(mNodeVariables);
      diff.resize(inputs.size());
      auto end = std::set_difference(inputs.begin(), inputs.end(),
                                     child_inputs.begin(), child_inputs.end(),
                                     diff.begin());
      diff.resize(end - diff.begin());
      if (!diff.empty())
      {
         auto new_child = AddInputsToNode(*i_child, diff, skip_var);
         *i_child = new_child;
      }
   }
}

void
SUBool::DnnfSmooth::MakeSmooth()
{
   mNnf.AcceptPostorderMut(*this);
}

SUBool::Nnf
SUBool::nnf_smooth(Nnf nnf)
{
   NnfNodeInputsVisitorMut inputs(true);
   nnf.AcceptPostorderMut(inputs);
   DnnfSmooth smooth(std::move(nnf), std::move(inputs).NodeVariables(),
                     std::move(inputs).LiteralNodes());
   smooth.MakeSmooth();
   return std::move(smooth).GetNnf();
}

std::optional<SUBool::PartialAssignment>
SUBool::smooth_dnnf_propagate(const Nnf & /*nnf*/, PartialAssignment /*alpha*/)
{
   throw NotImplementedException("smooth_dnnf_propagate",
                                 "Not yet implemented");
   return std::optional<SUBool::PartialAssignment>();
}

std::optional<SUBool::PartialAssignment>
SUBool::dnnf_propagate(const Nnf & /*nnf*/, PartialAssignment /*alpha*/)
{
   throw NotImplementedException("dnnf_propagate", "Not yet implemented");
   return std::optional<SUBool::PartialAssignment>();
}
