/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file dnnf.h
 * DNNF traits checking, manipulation, and propagation.
 */

#ifndef __DNNF_H
#define __DNNF_H

#include "nnf.h"

namespace SUBool
{

   struct NnfTraits
   {
      bool decomposable;
      bool smooth;
      bool decision_nnf; // only decision nodes and ands
   };

   NnfTraits nnf_traits(const Nnf &nnf);
   Nnf nnf_smooth(Nnf nnf);

   std::optional<PartialAssignment>
   smooth_dnnf_propagate(const Nnf &nnf, PartialAssignment alpha);

   std::optional<PartialAssignment> dnnf_propagate(const Nnf &nnf,
                                                   PartialAssignment alpha);

   template <typename T> class NodeInputsVisitor : public NnfVisitorTemplate<T>
   {
      using Base = NnfVisitorTemplate<T>;

      NnfNode::tValueCache<std::vector<unsigned>> mNodeVariables;
      NnfNode::tValueCache<NnfTraits> mNodeTraits;
      std::unordered_map<Literal, typename Base::LiteralPtr> mLiteralNodes;
      bool mCollectLiteralNodes{false};

    public:
      ~NodeInputsVisitor() = default;
      NodeInputsVisitor(bool collect_literal_nodes = false)
          : Base(), mNodeVariables(), mNodeTraits(), mLiteralNodes(),
            mCollectLiteralNodes(collect_literal_nodes)
      {
      }

      virtual void VisitLiteral(const typename Base::LiteralPtr &node) override;
      virtual void VisitAnd(const typename Base::GateAndPtr &node) override;
      virtual void VisitOr(const typename Base::GateOrPtr &node) override;
      virtual void
      VisitDecision(const typename Base::GateDecisionPtr &node) override;
      virtual void VisitTrue(const typename Base::TruePtr &node) override;
      virtual void VisitFalse(const typename Base::FalsePtr &node) override;
      virtual void VisitLeafCNF(const typename Base::CnfPtr &node) override;

      NnfNode::tValueCache<std::vector<unsigned>> &&NodeVariables() &&;
      std::unordered_map<Literal, typename Base::LiteralPtr> &&
      LiteralNodes() &&;

      NnfTraits NodeTraits(const NnfNode::tConstPointer &node) const;
   };

   using NnfNodeInputsVisitor = NodeInputsVisitor<const NnfNode>;
   using NnfNodeInputsVisitorMut = NodeInputsVisitor<NnfNode>;
} // namespace SUBool

// Implementation of NodeInputsVisitor<T>

template <typename T>
void
SUBool::NodeInputsVisitor<T>::VisitLiteral(
    const typename Base::LiteralPtr &node)
{
   std::vector<unsigned> inputs({node->GetLiteral().Variable()});
   node->CacheValue(mNodeVariables, std::move(inputs));
   node->CacheValue(mNodeTraits, {true, true, true});
   if (mCollectLiteralNodes)
   {
      mLiteralNodes[node->GetLiteral()] = node;
   }
}

template <typename T>
void
SUBool::NodeInputsVisitor<T>::VisitAnd(const typename Base::GateAndPtr &node)
{
   std::vector<unsigned> inputs;
   std::vector<unsigned> new_inputs;
   NnfTraits traits{true, true, true};
   for (const auto &child : *node)
   {
      const auto &child_inputs = child->CachedValueConstRef(mNodeVariables);
      const auto &child_traits = child->CachedValueConstRef(mNodeTraits);
      new_inputs.resize(inputs.size() + child_inputs.size());
      auto end =
          std::set_union(inputs.begin(), inputs.end(), child_inputs.begin(),
                         child_inputs.end(), new_inputs.begin());
      new_inputs.resize(end - new_inputs.begin());
      /*std::cout << new_inputs.size() << ", " << inputs.size() << ", "
                << child_inputs.size() << std::endl;*/
      traits.decomposable &=
          child_traits.decomposable
          && (new_inputs.size() == inputs.size() + child_inputs.size());
      traits.decision_nnf &= child_traits.decision_nnf;
      traits.smooth &= child_traits.smooth;
      std::swap(inputs, new_inputs);
   }
   node->CacheValue(mNodeVariables, std::move(inputs));
   node->CacheValue(mNodeTraits, std::move(traits));
}

template <typename T>
void
SUBool::NodeInputsVisitor<T>::VisitDecision(
    const typename Base::GateDecisionPtr &node)
{
   const auto &true_inputs =
       node->TrueInput()->CachedValueConstRef(mNodeVariables);
   const auto &false_inputs =
       node->FalseInput()->CachedValueConstRef(mNodeVariables);
   const auto &true_traits =
       node->TrueInput()->CachedValueConstRef(mNodeTraits);
   const auto &false_traits =
       node->FalseInput()->CachedValueConstRef(mNodeTraits);
   std::vector<unsigned> inputs(true_inputs.size() + false_inputs.size() + 1);
   auto end =
       std::set_union(true_inputs.begin(), true_inputs.end(),
                      false_inputs.begin(), false_inputs.end(), inputs.begin());
   NnfTraits traits{true_traits.decomposable && false_traits.decomposable,
                    true_traits.smooth && false_traits.smooth
                        && end - inputs.begin()
                               == static_cast<long int>(true_inputs.size())
                        && true_inputs.size() == false_inputs.size(),
                    true_traits.decision_nnf && false_traits.decision_nnf};
   if (std::binary_search(inputs.begin(), end, node->DecisionVar()))
   {
      inputs.resize(end - inputs.begin());
   }
   else
   {
      inputs.resize(end - inputs.begin() + 1);
      while (end != inputs.begin() && node->DecisionVar() < *(end - 1))
      {
         *end = *(end - 1);
         --end;
      }
      *end = node->DecisionVar();
   }
   node->CacheValue(mNodeVariables, std::move(inputs));
   node->CacheValue(mNodeTraits, std::move(traits));
}

template <typename T>
void
SUBool::NodeInputsVisitor<T>::VisitOr(const typename Base::GateOrPtr &node)
{
   std::vector<unsigned> inputs;
   std::vector<unsigned> new_inputs;
   NnfTraits traits{true, true, false};
   bool not_first = false;
   for (const auto &child : *node)
   {
      const auto &child_inputs = child->CachedValueConstRef(mNodeVariables);
      const auto &child_traits = child->CachedValueConstRef(mNodeTraits);
      new_inputs.resize(inputs.size() + child_inputs.size());
      auto end =
          std::set_union(inputs.begin(), inputs.end(), child_inputs.begin(),
                         child_inputs.end(), new_inputs.begin());
      new_inputs.resize(end - new_inputs.begin());
      traits.decomposable &= child_traits.decomposable;
      traits.smooth &= child_traits.smooth;
      if (not_first)
      {
         traits.smooth &= new_inputs.size() == inputs.size();
         traits.smooth &= new_inputs.size() == child_inputs.size();
      }
      else
      {
         not_first = true;
      }
      std::swap(inputs, new_inputs);
   }
   node->CacheValue(mNodeVariables, std::move(inputs));
   node->CacheValue(mNodeTraits, std::move(traits));
}

template <typename T>
void
SUBool::NodeInputsVisitor<T>::VisitTrue(const typename Base::TruePtr &node)
{
   node->CacheValue(mNodeVariables, {});
   node->CacheValue(mNodeTraits, {true, true, true});
}

template <typename T>
void
SUBool::NodeInputsVisitor<T>::VisitFalse(const typename Base::FalsePtr &node)
{
   node->CacheValue(mNodeVariables, {});
   node->CacheValue(mNodeTraits, {true, true, true});
}

template <typename T>
SUBool::NnfTraits
SUBool::NodeInputsVisitor<T>::NodeTraits(
    const NnfNode::tConstPointer &node) const
{
   return *node->CachedValue(mNodeTraits);
}

template <typename T>
SUBool::NnfNode::tValueCache<std::vector<unsigned>> &&
SUBool::NodeInputsVisitor<T>::NodeVariables() &&
{
   return std::move(mNodeVariables);
}

template <typename T>
std::unordered_map<SUBool::Literal,
                   typename SUBool::NodeInputsVisitor<T>::Base::LiteralPtr> &&
SUBool::NodeInputsVisitor<T>::LiteralNodes() &&
{
   return std::move(mLiteralNodes);
}

template <typename T>
void
SUBool::NodeInputsVisitor<T>::VisitLeafCNF(const typename Base::CnfPtr &node)
{
   auto vars = node->Variables();
   node->CacheValue(mNodeVariables,
                    std::vector<unsigned>(vars.begin(), vars.end()));
   node->CacheValue(mNodeTraits, {true, true, true});
}
#endif
