/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file dnnf_pce.h
 * DNNF to PCE translation.
 */

#include <stack>

#include "cardinal.h"
#include "dnnf_enc.h"
#include "logstream.h"
#include "nnfvisit.h"

namespace SUBool
{
   NnfNode::tValueCache<unsigned> nnf_levels(const Nnf &nnf);

   using Separator = std::vector<NnfNode::tConstPointer>;
   using InputSeparators = std::vector<Separator>;

   struct NodeInfo
   {
      unsigned level{0};
      std::unordered_set<unsigned> inputs;

      NodeInfo() : inputs() {}
      NodeInfo(unsigned l, std::initializer_list<unsigned> inp)
          : level(l), inputs(inp)
      {
      }
   };

   using NodeInfoCache = NnfNode::tValueCache<NodeInfo>;
   struct NnfWithSeparators
   {
      Nnf nnf;
      std::vector<InputSeparators> separators;

      NnfWithSeparators(Nnf nnf) : nnf(std::move(nnf)), separators() {}
      void InitSeparators(const std::vector<unsigned> max_levels);
   };

   NnfWithSeparators nnf_with_separators(Nnf nnf, LEVEL_FUNC level_func);

   void add_separator_clauses(const NnfWithSeparators &nnf,
       EncodingPart &enc_part, ConstraintEncoder &card_encoder);
   EncodingPart encode_smooth_dnnf(Nnf nnf, LEVEL_FUNC level_func,
       ConstraintEncoder &card_encoder, VarStore &var_store,
       unsigned log_level);

   class NodeInfoVisitor : public NnfVisitor
   {
      std::stack<NnfNode::tConstPointer> mNodeStack{};
      NnfNode::tValueCache<NodeInfo> mNodeInfo{};
      std::vector<unsigned> mMaxLevels{};

      void ComputeLevelsFromRoot();
      void ComputeLevelsFromLeaves(const Nnf &nnf);
      void ComputeLevels(const Nnf &nnf, LEVEL_FUNC level_func);

    public:
      virtual void VisitLiteral(const std::shared_ptr<const NnfLiteral> &node);
      virtual void VisitAnd(const std::shared_ptr<const NnfGateAnd> &node);
      virtual void VisitOr(const std::shared_ptr<const NnfGateOr> &node);
      virtual void VisitDecision(
          const std::shared_ptr<const NnfGateDecision> &node);
      virtual void VisitTrue(const std::shared_ptr<const NnfTrue> &node);
      virtual void VisitFalse(const std::shared_ptr<const NnfFalse> &node);

      void VisitGate(NnfNode::tConstPointer node);

      void ComputeNodeInfo(const Nnf &nnf, LEVEL_FUNC level_func);

      const NnfNode::tValueCache<NodeInfo> &
      NodeInfoCache() &
      {
         return mNodeInfo;
      }
      NnfNode::tValueCache<NodeInfo>
      NodeInfoCache() &&
      {
         return std::move(mNodeInfo);
      }
      const std::vector<unsigned> &
      MaxLevels() &
      {
         return mMaxLevels;
      }
      std::vector<unsigned>
      MaxLevels() &&
      {
         return std::move(mMaxLevels);
      }
   };

   class MakeNnfSeparableVisitor : public NnfVisitorMut
   {
      NnfWithSeparators mNnf;
      NnfNode::tValueCache<NodeInfo> mNodeInfo;
      std::vector<unsigned> mMaxLevels;

    public:
      virtual void VisitLiteral(
          const std::shared_ptr<NnfLiteral> &node) override;
      virtual void VisitAnd(const std::shared_ptr<NnfGateAnd> &node) override;
      virtual void VisitOr(const std::shared_ptr<NnfGateOr> &node) override;
      virtual void VisitDecision(
          const std::shared_ptr<NnfGateDecision> &node) override;
      virtual void VisitTrue(const std::shared_ptr<NnfTrue> &node) override;
      virtual void VisitFalse(const std::shared_ptr<NnfFalse> &node) override;

      void VisitGate(NnfNode::tPointer node);

      MakeNnfSeparableVisitor(Nnf nnf, NnfNode::tValueCache<NodeInfo> node_info,
          std::vector<unsigned> max_levels);
      const NnfWithSeparators &
      Get() &
      {
         return mNnf;
      }
      NnfWithSeparators
      Get() &&
      {
         return std::move(mNnf);
      }
      void MakeSeparable();
   };

} // namespace SUBool

const SUBool::EnumAnnotation<SUBool::LEVEL_FUNC> SUBool::kLevelFuncAnnotation(
    "LEVEL_FUNC",
    {                                                                         //
        {LEVEL_FUNC::FROM_ROOT, "from_root", "compute levels from the root"}, //
        {LEVEL_FUNC::FROM_LEAVES, "from_leaves",
            "compute levels from the leaves"}});

inline void
SUBool::DnnfInputUrcEncoder::AddClause(Clause clause)
{
   if (!clause.Trivial())
   {
      mClauses.push_back(std::move(clause));
   }
}

void
SUBool::DnnfInputUrcEncoder::AddRoot(const NnfNode::tConstPointer &root)
{
   AddClause(Clause(*(root->CachedValue(mGateLiterals))));
}

SUBool::URCEncoding
SUBool::DnnfInputUrcEncoder::GetURCEncoding() const
{
   CNF cnf(mClauses);
   return URCEncoding(std::move(cnf), mMaxInputVariable, false);
}

void
SUBool::DnnfInputUrcEncoder::VisitLiteral(
    const std::shared_ptr<const NnfLiteral> &node)
{
   node->CacheValue(mGateLiterals, node->GetLiteral());
}

void
SUBool::DnnfInputUrcEncoder::VisitTrue(
    const std::shared_ptr<const NnfTrue> &node)
{
   auto gate_literal = NewGateLiteral();
   node->CacheValue(mGateLiterals, gate_literal);
   AddClause({gate_literal});
}

void
SUBool::DnnfInputUrcEncoder::VisitFalse(
    const std::shared_ptr<const NnfFalse> &node)
{
   auto gate_literal = NewGateLiteral();
   node->CacheValue(mGateLiterals, gate_literal);
   AddClause({gate_literal.Negation()});
}

void
SUBool::DnnfInputUrcEncoder::VisitAnd(
    const std::shared_ptr<const NnfGateAnd> &node)
{
   auto gate_literal = NewGateLiteral();
   node->CacheValue(mGateLiterals, gate_literal);
   if (node->InputSize() > 0)
   {
      for (const auto &input : *node)
      {
         AddClause(
             {*(input->CachedValue(mGateLiterals)), gate_literal.Negation()});
      }
   }
   else
   {
      AddClause({gate_literal});
   }
}

void
SUBool::DnnfInputUrcEncoder::VisitOr(
    const std::shared_ptr<const NnfGateOr> &node)
{
   auto gate_literal = NewGateLiteral();
   node->CacheValue(mGateLiterals, gate_literal);
   std::vector<Literal> literals;
   literals.reserve(node->InputSize() + 1);
   for (const auto &input : *node)
   {
      literals.push_back(*(input->CachedValue(mGateLiterals)));
   }
   literals.push_back(gate_literal.Negation());
   AddClause(std::move(literals));
}

void
SUBool::DnnfInputUrcEncoder::VisitDecision(
    const std::shared_ptr<const NnfGateDecision> &node)
{
   auto gate_literal = NewGateLiteral();
   node->CacheValue(mGateLiterals, gate_literal);
   auto true_input_lit = node->TrueInput()->CachedValue(mGateLiterals);
   auto false_input_lit = node->FalseInput()->CachedValue(mGateLiterals);
   AddClause({Literal(node->DecisionVar(), true), *false_input_lit,
       gate_literal.Negation()});
   AddClause({Literal(node->DecisionVar(), false), *true_input_lit,
       gate_literal.Negation()});
   AddClause({*true_input_lit, *false_input_lit, gate_literal.Negation()});
}

SUBool::Literal
SUBool::DnnfInputUrcEncoder::NewGateLiteral()
{
   mVarStore.NameStream() << "v_{" << mVarStore.LastVariable() << "}"
                          << std::endl;
   return Literal(mVarStore.NewVariable(), true);
}

void
SUBool::DnnfInputUrcEncoder::Encode(const Nnf &nnf)
{
   mMaxInputVariable = nnf.MaxVariable();
   nnf.AcceptPostorder(*this);
   AddRoot(nnf.Root());
}

SUBool::URCEncoding
SUBool::DnnfInputUrcEncoder::URCEncode(const Nnf &nnf, VarStore &var_store)
{
   DnnfInputUrcEncoder encoder(var_store);
   encoder.Encode(nnf);
   return encoder.GetURCEncoding();
}

SUBool::URCEncoding
SUBool::DnnfInputUrcEncoder::URCEncode(const Nnf &nnf)
{
   VarStore var_store(false, nnf.MaxVariable());
   return URCEncode(nnf, var_store);
}

void
SUBool::SmoothDnnfInputPcEncoder::VisitLiteral(
    const std::shared_ptr<const NnfLiteral> &node)
{
   DnnfInputUrcEncoder::VisitLiteral(node);
   mLiteralNodes[node->GetLiteral()] = node;
   mParentClauses[node].push_back(node->GetLiteral().Negation());
}

void
SUBool::SmoothDnnfInputPcEncoder::VisitTrue(
    const std::shared_ptr<const NnfTrue> &node)
{
   DnnfInputUrcEncoder::VisitTrue(node);
   auto gate_literal = node->CachedValue(mGateLiterals);
   mParentClauses[node].push_back(*gate_literal);
}

void
SUBool::SmoothDnnfInputPcEncoder::VisitFalse(
    const std::shared_ptr<const NnfFalse> &node)
{
   DnnfInputUrcEncoder::VisitFalse(node);
   auto gate_literal = node->CachedValue(mGateLiterals);
   mParentClauses[node].push_back(gate_literal->Negation());
}

void
SUBool::SmoothDnnfInputPcEncoder::VisitAnd(
    const std::shared_ptr<const NnfGateAnd> &node)
{
   DnnfInputUrcEncoder::VisitAnd(node);
   auto gate_literal = node->CachedValue(mGateLiterals);
   mParentClauses[node].push_back(gate_literal->Negation());
   for (const auto &child : *node)
   {
      mParentClauses[child].push_back(*gate_literal);
   }
}

void
SUBool::SmoothDnnfInputPcEncoder::VisitOr(
    const std::shared_ptr<const NnfGateOr> &node)
{
   DnnfInputUrcEncoder::VisitOr(node);
   auto gate_literal = node->CachedValue(mGateLiterals);
   mParentClauses[node].push_back(gate_literal->Negation());
   for (const auto &child : *node)
   {
      mParentClauses[child].push_back(*gate_literal);
   }
}

void
SUBool::SmoothDnnfInputPcEncoder::VisitDecision(
    const std::shared_ptr<const NnfGateDecision> &node)
{
   DnnfInputUrcEncoder::VisitDecision(node);
   auto gate_literal = node->CachedValue(mGateLiterals);
   mLiteralDecisionParents[Literal(node->DecisionVar(), true)].push_back(
       *gate_literal);
   mLiteralDecisionParents[Literal(node->DecisionVar(), false)].push_back(
       *gate_literal);
   mParentClauses[node->TrueInput()].push_back(*gate_literal);
   mParentClauses[node->FalseInput()].push_back(*gate_literal);
}

void
SUBool::SmoothDnnfInputPcEncoder::Encode(const Nnf &nnf)
{
   DnnfInputUrcEncoder::Encode(nnf);
   for (unsigned var = 1; var <= nnf.MaxVariable(); ++var)
   {
      for (unsigned pos = 0; pos < 2; ++pos)
      {
         Literal lit(var, static_cast<bool>(pos));
         if (mLiteralNodes.count(lit) > 0)
         {
            auto &literals = mParentClauses[mLiteralNodes[lit]];
            auto &dec_parents = mLiteralDecisionParents[lit];
            literals.insert(
                literals.end(), dec_parents.begin(), dec_parents.end());
         }
         else
         {
            mLiteralDecisionParents[lit].push_back(lit.Negation());
            AddClause(std::move(mLiteralDecisionParents[lit]));
         }
      }
   }
   for (auto &literals : mParentClauses)
   {
      if (literals.second.size() > 1)
      {
         AddClause(std::move(literals.second));
      }
   }
}

SUBool::PCEncoding
SUBool::SmoothDnnfInputPcEncoder::GetPCEncoding() const
{
   CNF cnf(mClauses);
   return PCEncoding(std::move(cnf), mMaxInputVariable, false);
}

SUBool::PCEncoding
SUBool::SmoothDnnfInputPcEncoder::PCEncode(const Nnf &nnf)
{
   VarStore var_store(false, nnf.MaxVariable());
   SmoothDnnfInputPcEncoder encoder(var_store);
   encoder.Encode(nnf);
   return encoder.GetPCEncoding();
}

SUBool::PCEncoding
SUBool::SmoothDnnfInputPcEncoder::PCEncode(const Nnf &nnf, VarStore &var_store)
{
   SmoothDnnfInputPcEncoder encoder(var_store);
   encoder.Encode(nnf);
   return encoder.GetPCEncoding();
}

SUBool::EncodingPart
SUBool::SmoothDnnfInputPcEncoder::PartEncode(
    const Nnf &nnf, VarStore &var_store)
{
   SmoothDnnfInputPcEncoder encoder(var_store);
   encoder.Encode(nnf);
   return {encoder.mClauses, encoder.mGateLiterals, encoder.mMaxInputVariable};
}

SUBool::NnfNode::tValueCache<unsigned>
SUBool::nnf_levels(const Nnf &nnf)
{
   NnfNode::tValueCache<unsigned> levels;
   unsigned next_level = 0;
   nnf.AcceptTopologicalFunction([&](const NnfNode::tConstPointer &node)
       { node->CacheValue(levels, next_level++); });
   return levels;
}

SUBool::EncodingPart
SUBool::encode_smooth_dnnf(Nnf nnf, LEVEL_FUNC level_func,
    ConstraintEncoder &card_encoder, VarStore &var_store, unsigned log_level)
{
   logs.TLog(log_level) << "Creating NNF with separators" << std::endl;
   auto node_count = nnf.NodeCount();
   auto nnf_sep = nnf_with_separators(std::move(nnf), level_func);
   unsigned sep_count = 0;
   for (const auto &is : nnf_sep.separators)
   {
      for (const auto &sep : is)
      {
         sep_count += sep.size();
      }
   }
   logs.TLog(log_level) << "... Created " << sep_count << " separators"
                        << std::endl;
   logs.TLog(log_level) << "... Added " << nnf_sep.nnf.NodeCount() - node_count
                        << " edge subdividing nodes" << std::endl;
   logs.TLog(log_level) << "Creating Input PC encoding" << std::endl;
   auto enc_part = SmoothDnnfInputPcEncoder::PartEncode(nnf_sep.nnf, var_store);
   logs.TLog(log_level) << "Adding separator clauses" << std::endl;
   add_separator_clauses(nnf_sep, enc_part, card_encoder);
   logs.TLog(log_level) << "Finished" << std::endl;
   return enc_part;
}

SUBool::URCEncoding
SUBool::encode_smooth_dnnf_urc(
    Nnf nnf, LEVEL_FUNC level_func, unsigned log_level)
{
   logs.TLog(log_level) << "Compiling a smooth DNNF into a URC encoding"
                        << std::endl;
   VarStore var_store(false, nnf.MaxVariable());
   auto amo_encoder = AmoEncoder::NewGlobalEncoder(var_store);
   auto enc = encode_smooth_dnnf(
       std::move(nnf), level_func, *amo_encoder, var_store, log_level);
   return URCEncoding(enc.clauses, enc.max_input_var, true);
}

SUBool::PCEncoding
SUBool::encode_smooth_dnnf_pc(
    Nnf nnf, LEVEL_FUNC level_func, unsigned log_level)
{
   logs.TLog(log_level) << "Compiling a smooth DNNF into a PC encoding"
                        << std::endl;
   VarStore var_store(false, nnf.MaxVariable());
   auto exone_encoder = ExOneEncoder::NewGlobalEncoder(var_store);
   auto enc = encode_smooth_dnnf(
       std::move(nnf), level_func, *exone_encoder, var_store, log_level);
   return PCEncoding(enc.clauses, enc.max_input_var, true);
}

SUBool::NnfWithSeparators
SUBool::nnf_with_separators(Nnf nnf, LEVEL_FUNC level_func)
{
   // TODO
   /*
    * 1. Compute the set of inputs for each node and add the nodes to stack in
    *    topological order. (1st go through the nnf).
    * 2. Assign levels following the stack (Level(u) for a node u).
    * 3. Transform the nnf, for each node, level is an interval, for an original
    *    node u, the level is [Level(u), Level(u)], for an intermediate node u'
    *    for a long edge (v, u), the level is [Level(v)+1, Level(u)-1]. The pair
    *    is denoted as MinLevel(v) and MaxLevel(v). (2nd traverse of nnf)
    * 4. Go through the nodes of the new nnf in any order and add a node v into
    *    Separator S[i,j] (for x_i at level j), if:
    *    - v is an inner node with MinLevel(v)<=j<=MaxLevel(v), or
    *    - v is a leaf labeled with a literal on x_i and MinLevel(v)<=j. Note
    *      that for leaves we have MaxLevel(v)=MinLevel(v).
    *
    * Data structures:
    *    - We can modify NNF on the fly
    *    - For each node v of the new NNF: MinLevel(v), MaxLevel(v), inputs
    *    - For each input i: Separators (array 0 ... max_level-1, max_level is
    * its size).
    *
    */

   NodeInfoVisitor node_info;
   node_info.ComputeNodeInfo(nnf, level_func);
   MakeNnfSeparableVisitor sep_visitor(std::move(nnf),
       std::move(node_info).NodeInfoCache(), std::move(node_info).MaxLevels());
   sep_visitor.MakeSeparable();
   return std::move(sep_visitor).Get();
}

void
SUBool::add_separator_clauses(const NnfWithSeparators &nnf,
    EncodingPart &enc_part, ConstraintEncoder &card_encoder)
{
   for (const auto &is : nnf.separators)
   {
      for (const auto &sep : is)
      {
         std::vector<Literal> literals;
         for (const auto &node : sep)
         {
            literals.push_back(*(node->CachedValue(enc_part.gate_literals)));
         }
         card_encoder.EncodeAppend(literals, enc_part.clauses);
      }
   }
}

void
SUBool::NodeInfoVisitor::VisitLiteral(
    const std::shared_ptr<const NnfLiteral> &node)
{
   node->CacheValue(mNodeInfo, NodeInfo{0, {node->GetLiteral().Variable()}});
   mNodeStack.push(node);
}

void
SUBool::NodeInfoVisitor::VisitGate(NnfNode::tConstPointer node)
{
   NodeInfo node_info;
   for (const auto &child : *node)
   {
      const auto &ni = child->CachedValueConstRef(mNodeInfo);
      node_info.inputs.insert(ni.inputs.begin(), ni.inputs.end());
   }
   node->CacheValue(mNodeInfo, std::move(node_info));
   mNodeStack.push(node);
}

void
SUBool::NodeInfoVisitor::VisitAnd(const std::shared_ptr<const NnfGateAnd> &node)
{
   VisitGate(node);
}

void
SUBool::NodeInfoVisitor::VisitOr(const std::shared_ptr<const NnfGateOr> &node)
{
   VisitGate(node);
}

void
SUBool::NodeInfoVisitor::VisitDecision(
    const std::shared_ptr<const NnfGateDecision> & /*node*/)
{
   throw BadInputException("NodeInfoVisitor::VisitDecision",
       "Encoding of a smooth DNNF into a URC/PC encoding "
       "does not work with decision gates yet");
}

void
SUBool::NodeInfoVisitor::VisitTrue(
    const std::shared_ptr<const NnfTrue> & /*node*/)
{
   throw BadInputException("NodeInfoVisitor::VisitTrue",
       "Encoding of a smooth DNNF into a URC/PC encoding "
       "does not work with constants");
}

void
SUBool::NodeInfoVisitor::VisitFalse(
    const std::shared_ptr<const NnfFalse> & /*node*/)
{
   throw BadInputException("NodeInfoVisitor::VisitFalse",
       "Encoding of a smooth DNNF into a URC/PC encoding "
       "does not work with constants");
}

void
SUBool::NodeInfoVisitor::ComputeLevelsFromRoot()
{
   while (!mNodeStack.empty())
   {
      auto node = mNodeStack.top();
      mNodeStack.pop();
      const auto &node_info = node->CachedValueConstRef(mNodeInfo);
      auto next_level = node_info.level + 1;
      for (const auto &child : *node)
      {
         auto &child_node_info = child->CachedValueRef(mNodeInfo);
         child_node_info.level = std::max(child_node_info.level, next_level);
      }
      if (node->IsLeaf())
      {
         auto i = *(node_info.inputs.begin());
         mMaxLevels[i - 1] = std::max(mMaxLevels[i - 1], node_info.level);
      }
   }
}

void
SUBool::NodeInfoVisitor::ComputeLevelsFromLeaves(const Nnf &nnf)
{
   ComputeLevelsFromRoot();
   nnf.AcceptPostorderFunction(
       [this, &nnf](NnfNode::tConstPointer node)
       {
          if (node->IsLeaf())
          {
             return;
          }
          auto &node_info = node->CachedValueRef(this->mNodeInfo);
          node_info.level = nnf.NodeCount() + 1;
          for (const auto &child : *node)
          {
             const auto &child_node_info =
                 child->CachedValueConstRef(this->mNodeInfo);
             node_info.level = std::min(node_info.level, child_node_info.level);
          }
          if (node_info.level == 0)
          {
             throw InvariantFailedException(
                 "NodeInfoVisitor::ComputeLevelsFromLeaves", "negative level");
          }
          --node_info.level;
       });
}

void
SUBool::NodeInfoVisitor::ComputeLevels(const Nnf &nnf, LEVEL_FUNC level_func)
{
   switch (level_func)
   {
   case LEVEL_FUNC::FROM_ROOT:
      ComputeLevelsFromRoot();
      break;
   case LEVEL_FUNC::FROM_LEAVES:
      ComputeLevelsFromLeaves(nnf);
      break;
   default:
      throw BadParameterException("NodeInfoVisitor::ComputeLevels",
          "Unknown level_func value "
              + std::to_string(static_cast<unsigned>(level_func)));
   }
}

void
SUBool::NodeInfoVisitor::ComputeNodeInfo(const Nnf &nnf, LEVEL_FUNC level_func)
{
   mMaxLevels.resize(nnf.MaxVariable(), 0);
   nnf.AcceptPostorder(*this);
   ComputeLevels(nnf, level_func);
}

void
SUBool::NnfWithSeparators::InitSeparators(
    const std::vector<unsigned> max_levels)
{
   separators.resize(nnf.MaxVariable());
   for (unsigned i = 0; i < nnf.MaxVariable(); ++i)
   {
      separators[i].resize(max_levels[i]);
   }
}

SUBool::MakeNnfSeparableVisitor::MakeNnfSeparableVisitor(Nnf nnf,
    NnfNode::tValueCache<NodeInfo> node_info, std::vector<unsigned> max_levels)
    : mNnf(std::move(nnf)), mNodeInfo(std::move(node_info)),
      mMaxLevels(std::move(max_levels))
{
   mNnf.InitSeparators(mMaxLevels);
}

void
SUBool::MakeNnfSeparableVisitor::MakeSeparable()
{
   mNnf.nnf.AcceptPostorderMut(*this);
}

void
SUBool::MakeNnfSeparableVisitor::VisitLiteral(
    const std::shared_ptr<NnfLiteral> &node)
{
   const auto &node_info = node->CachedValueConstRef(mNodeInfo);
   for (unsigned l = node_info.level;
        l < mNnf.separators[node->MaxVariable() - 1].size(); ++l)
   {
      mNnf.separators[node->MaxVariable() - 1][l].push_back(node);
   }
}

void
SUBool::MakeNnfSeparableVisitor::VisitGate(NnfNode::tPointer node)
{
   const auto &node_info = node->CachedValueConstRef(mNodeInfo);
   for (const auto &i : node_info.inputs)
   {
      mNnf.separators[i - 1][node_info.level].push_back(node);
   }
   for (auto i_child = node->begin(); i_child != node->end(); ++i_child)
   {
      const auto &child_info = (*i_child)->CachedValueConstRef(mNodeInfo);
      if (node_info.level + 1 < child_info.level)
      {
         *i_child = NnfNode::MakeAnd({*i_child});
         for (const auto &i : child_info.inputs)
         {
            for (unsigned l = node_info.level + 1; l < child_info.level; ++l)
            {
               mNnf.separators[i - 1][l].push_back(*i_child);
            }
         }
      }
   }
}

void
SUBool::MakeNnfSeparableVisitor::VisitAnd(
    const std::shared_ptr<NnfGateAnd> &node)
{
   VisitGate(node);
}

void
SUBool::MakeNnfSeparableVisitor::VisitOr(const std::shared_ptr<NnfGateOr> &node)
{
   VisitGate(node);
}

void
SUBool::MakeNnfSeparableVisitor::VisitDecision(
    const std::shared_ptr<NnfGateDecision> & /*node*/)
{
   throw BadInputException("NodeInfoVisitor::VisitDecision",
       "Encoding of a smooth DNNF into a URC/PC encoding "
       "does not work with decision gates yet");
}

void
SUBool::MakeNnfSeparableVisitor::VisitTrue(
    const std::shared_ptr<NnfTrue> & /*node*/)
{
   throw BadInputException("NodeInfoVisitor::VisitTrue",
       "Encoding of a smooth DNNF into a URC/PC encoding "
       "does not work with constants");
}

void
SUBool::MakeNnfSeparableVisitor::VisitFalse(
    const std::shared_ptr<NnfFalse> & /*node*/)
{
   throw BadInputException("NodeInfoVisitor::VisitFalse",
       "Encoding of a smooth DNNF into a URC/PC encoding "
       "does not work with constants");
}
