/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file dnnf_pce.h
 * DNNF to PCE translation.
 */

#ifndef __DNNF_PCE_H
#define __DNNF_PCE_H

#include <unordered_map>

#include "cardinal.h"
#include "encoding.h"
#include "enum.h"
#include "nnf.h"
#include "varstore.h"

namespace SUBool
{
   class DnnfInputUrcEncoder : public NnfVisitor
   {
      VarStore &mVarStore;

    protected:
      unsigned mMaxInputVariable = 0;
      NnfNode::tValueCache<Literal> mGateLiterals;
      std::vector<Clause> mClauses;

      DnnfInputUrcEncoder(VarStore &var_store)
          : mVarStore(var_store), mGateLiterals(), mClauses()
      {
      }
      DnnfInputUrcEncoder(VarStore &&var_store) = delete;
      void AddRoot(const NnfNode::tConstPointer &root);
      void AddClause(Clause clause);
      Literal NewGateLiteral();

    public:
      virtual void VisitLiteral(
          const std::shared_ptr<const NnfLiteral> &node) override;
      virtual void VisitAnd(
          const std::shared_ptr<const NnfGateAnd> &node) override;
      virtual void VisitOr(
          const std::shared_ptr<const NnfGateOr> &node) override;
      virtual void VisitDecision(
          const std::shared_ptr<const NnfGateDecision> &node) override;
      virtual void VisitTrue(
          const std::shared_ptr<const NnfTrue> &node) override;
      virtual void VisitFalse(
          const std::shared_ptr<const NnfFalse> &node) override;

      virtual void Encode(const Nnf &nnf);

      URCEncoding GetURCEncoding() const;

      static URCEncoding URCEncode(const Nnf &nnf, VarStore &var_store);
      static URCEncoding URCEncode(const Nnf &nnf);
   };

   struct EncodingPart
   {
      std::vector<Clause> clauses;
      NnfNode::tValueCache<Literal> gate_literals;
      unsigned max_input_var;
   };

   class SmoothDnnfInputPcEncoder : public DnnfInputUrcEncoder
   {
      NnfNode::tValueCache<std::vector<Literal>> mParentClauses;

      std::unordered_map<Literal, NnfNode::tConstPointer> mLiteralNodes;
      std::unordered_map<Literal, std::vector<Literal>> mLiteralDecisionParents;

    protected:
      // unsigned mMaxVariable;
      SmoothDnnfInputPcEncoder(VarStore &var_store)
          : DnnfInputUrcEncoder(var_store), mParentClauses(), mLiteralNodes(),
            mLiteralDecisionParents()
      {
      }
      SmoothDnnfInputPcEncoder(VarStore &&var_store) = delete;

    public:
      virtual void VisitLiteral(
          const std::shared_ptr<const NnfLiteral> &node) override;
      virtual void VisitAnd(
          const std::shared_ptr<const NnfGateAnd> &node) override;
      virtual void VisitOr(
          const std::shared_ptr<const NnfGateOr> &node) override;
      virtual void VisitDecision(
          const std::shared_ptr<const NnfGateDecision> &node) override;
      virtual void VisitTrue(
          const std::shared_ptr<const NnfTrue> &node) override;
      virtual void VisitFalse(
          const std::shared_ptr<const NnfFalse> &node) override;

      virtual void Encode(const Nnf &nnf) override;
      PCEncoding GetPCEncoding() const;
      static PCEncoding PCEncode(const Nnf &nnf, VarStore &var_store);
      static PCEncoding PCEncode(const Nnf &nnf);
      static EncodingPart PartEncode(const Nnf &nnf, VarStore &var_store);
   };

   enum class LEVEL_FUNC
   {
      FROM_ROOT,
      FROM_LEAVES,
      Last = FROM_LEAVES
   };

   extern const EnumAnnotation<LEVEL_FUNC> kLevelFuncAnnotation;

   URCEncoding encode_smooth_dnnf_urc(
       Nnf nnf, LEVEL_FUNC level_func, unsigned log_level = 5);
   PCEncoding encode_smooth_dnnf_pc(
       Nnf nnf, LEVEL_FUNC level_func, unsigned log_level = 5);

} // namespace SUBool

#endif
