/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file dpelim.cpp
 * DP elimination of variables in a CNF. Especially elimination of auxiliary
 * variables in a CNF encoding.
 */

#include "dpelim.h"
#include "cnfutil.h"
#include "empower.h"
#include "resolve.h"
#include "util.h"

const SUBool::EnumAnnotation<SUBool::DPEliminator::IGNORE_POLICY>
    SUBool::DPEliminator::kIgnorePolicyAnnotation("IGNORE_POLICY",
        {{IGNORE_POLICY::NONE, "none", "do not discard any resolvents"},
            {IGNORE_POLICY::IMPLICATES, "implicates",
                "discard the clauses which are implicates of others"},
            {IGNORE_POLICY::ABSORBED, "absorbed",
                "discard the absorbed clauses (not empowering)"}});

const SUBool::EnumAnnotation<SUBool::DPEliminator::INPROCESSING>
    SUBool::DPEliminator::kInprocessingAnnotation("INPROCESSING",
        {{INPROCESSING::NONE, "none", "no inprocessing"},
            {INPROCESSING::PRIME, "prime", "find prime subimplicate"}});

const SUBool::EnumAnnotation<SUBool::DPEliminator::INCREASE_POLICY>
    SUBool::DPEliminator::kIncreasePolicyAnnotation("INCREASE_POLICY",
        {{INCREASE_POLICY::ESTIMATE, "estimate",
             "stop when the estimate of increase gets over the bound"},
            {INCREASE_POLICY::AFTER_BOUND, "after_bound",
                "stop when the size got over the bound"},
            {INCREASE_POLICY::NEW_CLAUSES, "new_clauses",
                "stop when the current number of new clauses gets over the "
                "bound"},
            {INCREASE_POLICY::NEW_CLAUSES_TOTAL, "new_clauses_total",
                "stop when the total number of new clauses gets over the "
                "bound"}});

SUBool::DPEliminator::DPEliminator(const CNF &cnf, unsigned max_fixed_variable,
    IGNORE_POLICY ignore_policy, INPROCESSING inprocessing,
    INCREASE_POLICY increase_policy)
    : mMaxFixedVariable(max_fixed_variable), mMaxVariable(cnf.MaxVariable()),
      mClauseInfo(), mInvalidClauseInfo(),
      mVarInfo(mMaxVariable > mMaxFixedVariable
                   ? mMaxVariable - mMaxFixedVariable
                   : 0),
      mVarInfoHeap(), mIgnorePolicy(ignore_policy), mInprocessing(inprocessing),
      mIncreasePolicy(increase_policy), mCnf(cnf)
{
   InitClauseInfo(cnf);
   InitVarInfo();
   InitVarHeap();
}

inline SUBool::DPEliminator::VarInfo &
SUBool::DPEliminator::GetVarInfo(unsigned var)
{
   return mVarInfo[var - mMaxFixedVariable - 1];
}

inline bool
SUBool::DPEliminator::IsFree(unsigned var)
{
   return var > mMaxFixedVariable;
}

void
SUBool::DPEliminator::InitClauseInfo(const CNF &cnf)
{
   for (const auto &clause : cnf)
   {
      mClauseInfo.emplace_back(clause, true);
   }
}

void
SUBool::DPEliminator::InitVarInfo()
{
   for (auto i_clause = mClauseInfo.begin(); i_clause != mClauseInfo.end();
        ++i_clause)
   {
      for (const auto &lit : i_clause->mClause)
      {
         if (!IsFree(lit.Variable()))
         {
            continue;
         }
         auto &vi = GetVarInfo(lit.Variable());
         auto &clauses = (lit.Positive() ? vi.mPosClauses : vi.mNegClauses);
         clauses.push_back(i_clause);
      }
   }
}

void
SUBool::DPEliminator::InitVarHeap()
{
   for (auto var = mMaxFixedVariable + 1; var <= mMaxVariable; ++var)
   {
      auto &vi = GetVarInfo(var);
      vi.mHandle = mVarInfoHeap.emplace(vi.ResolventCount(), var);
   }
}

size_t
SUBool::DPEliminator::VarInfo::ResolventCount() const
{
   size_t count = 0;
   for (const auto &pos_clause : mPosClauses)
   {
      for (const auto &neg_clause : mNegClauses)
      {
         if (resolvable(pos_clause->mClause, neg_clause->mClause))
         {
            ++count;
         }
      }
   }
   return count;
}

SUBool::CNF
SUBool::DPEliminator::Cnf() const &
{
   std::vector<Clause> clauses;
   clauses.reserve(mClauseInfo.size());
   for (const auto &ci : mClauseInfo)
   {
      clauses.push_back(ci.mClause);
   }
   return CNF(std::move(clauses));
}

SUBool::CNF
SUBool::DPEliminator::Cnf() &&
{
   std::vector<Clause> clauses;
   clauses.reserve(mClauseInfo.size());
   for (auto &ci : mClauseInfo)
   {
      clauses.push_back(std::move(ci.mClause));
   }
   return CNF(std::move(clauses));
}

std::pair<SUBool::CNF, std::vector<bool>>
SUBool::DPEliminator::CnfWithMask() const &
{
   std::vector<bool> new_clauses_mask;
   new_clauses_mask.reserve(mClauseInfo.size());
   CNF cnf;
   for (auto &ci : mClauseInfo)
   {
      cnf.PushBack(ci.mClause);
      new_clauses_mask.push_back(!ci.InputClause());
   }
   return std::make_pair(std::move(cnf), std::move(new_clauses_mask));
}

std::pair<SUBool::CNF, std::vector<bool>>
SUBool::DPEliminator::CnfWithMask() &&
{
   std::vector<bool> new_clauses_mask;
   new_clauses_mask.reserve(mClauseInfo.size());
   CNF cnf;
   for (auto &ci : mClauseInfo)
   {
      cnf.PushBack(std::move(ci.mClause));
      new_clauses_mask.push_back(!ci.InputClause());
   }
   return std::make_pair(std::move(cnf), std::move(new_clauses_mask));
}

SUBool::CNF
SUBool::DPEliminator::EliminateAuxVariables(const CNF &cnf,
    unsigned max_fixed_variable, IGNORE_POLICY ignore_policy,
    INPROCESSING inprocessing, INCREASE_POLICY increase_policy,
    size_t max_size_increase, bool at_least_one)
{
   logs.TLog(3) << "[DPEliminator] max_size_increase: " << max_size_increase
                << std::endl;
   if (cnf.MaxVariable() <= max_fixed_variable)
   {
      return cnf;
   }
   DPEliminator elim(
       cnf, max_fixed_variable, ignore_policy, inprocessing, increase_policy);
   elim.EliminateAuxVariables(max_size_increase, at_least_one);
   return std::move(elim).Cnf();
}

SUBool::CNF
SUBool::DPEliminator::EliminateAuxVariablesWithFactor(const CNF &cnf,
    unsigned max_fixed_variable, IGNORE_POLICY ignore_policy,
    INPROCESSING inprocessing, INCREASE_POLICY increase_policy,
    long double max_size_increase_factor, bool at_least_one)
{
   size_t max_size_increase = static_cast<size_t>(
       static_cast<long double>(cnf.Size()) * max_size_increase_factor);
   return EliminateAuxVariables(cnf, max_fixed_variable, ignore_policy,
       inprocessing, increase_policy, max_size_increase, at_least_one);
}

void
SUBool::DPEliminator::EliminateVariable(unsigned var)
{
   mVarInfoHeap.erase(GetVarInfo(var).mHandle);
   EliminateVarNotInHeap(var);
}

bool
SUBool::DPEliminator::CheckIncrease(size_t initial_size, size_t bound) const
{
   switch (mIncreasePolicy)
   {
   case INCREASE_POLICY::ESTIMATE:
      return mVarInfoHeap.top().first + Size() <= initial_size + bound;
   case INCREASE_POLICY::AFTER_BOUND:
      return Size() <= initial_size + bound;
   case INCREASE_POLICY::NEW_CLAUSES:
      return mCurrentNewClauses <= bound;
   case INCREASE_POLICY::NEW_CLAUSES_TOTAL:
      return mTotalNewClauses <= bound;
   default:
      throw BadParameterException(
          "DPEliminator::CheckIncrease", "Unknown increase policy");
   }
}

void
SUBool::DPEliminator::EliminateAuxVariables(
    size_t max_size_increase, bool at_least_one)
{
   if (mMaxVariable <= mMaxFixedVariable)
   {
      return;
   }
   const size_t initial_size = Size();
   logs.TLog(3) << "[DPEliminator] max_size_increase: " << max_size_increase
                << std::endl;

   unsigned vars = mCnf.MaxVariable();
   while (!mVarInfoHeap.empty()
          && (max_size_increase == kEliminateAll || at_least_one
              || CheckIncrease(initial_size, max_size_increase)))
   {
      at_least_one = false;
      unsigned var = mVarInfoHeap.top().second;
      mVarInfoHeap.pop();
      EliminateVarNotInHeap(var);
      mCnf = Cnf();
      --vars;
      logs.TLog(2) << "[DPEliminator] Current CNF (p cnf " << vars << " "
                   << mCnf.Size() << "), new clauses (" << mCurrentNewClauses
                   << "/" << mTotalNewClauses << ")" << std::endl;
   }
}

void
SUBool::DPEliminator::EliminateVarNotInHeap(unsigned var)
{
   if (!IsFree(var))
   {
      throw BadParameterException(
          "DPEliminator::EliminateVarNotInHeap", "Variable is fixed");
   }
   auto &vi = GetVarInfo(var);
   auto invalid = InvalidateClauses(vi.mPosClauses);
   invalid.splice(invalid.end(), InvalidateClauses(vi.mNegClauses));
   AddClauses(vi.Resolvents());
   std::unordered_set<unsigned> touched(mMaxVariable - mMaxFixedVariable);
   for (const auto &ci : invalid)
   {
      for (const auto &lit : ci.mClause)
      {
         if (IsFree(lit.Variable()))
         {
            touched.insert(lit.Variable());
         }
      }
   }
   touched.erase(var);
   UpdateVarInfo(touched);
   vi.ClearLists();
}

std::list<SUBool::DPEliminator::ClauseInfo>
SUBool::DPEliminator::InvalidateClauses(
    const std::list<ClauseInfo::tPointer> &clauses)
{
   std::list<ClauseInfo> invalid;
   for (auto &p : clauses)
   {
      if (p->Valid())
      {
         p->Invalidate();
         if (!p->InputClause())
         {
            --mCurrentNewClauses;
         }
         invalid.splice(invalid.end(), mClauseInfo, p);
      }
   }
   return invalid;
}

void
SUBool::DPEliminator::UpdateVarInfo(std::unordered_set<unsigned> variables)
{
   for (auto var : variables)
   {
      auto &vi = GetVarInfo(var);
      vi.RemoveInvalid();
      auto old_rc = (*vi.mHandle).first;
      (*vi.mHandle).first = vi.ResolventCount();
      if (old_rc > (*vi.mHandle).first)
      {
         mVarInfoHeap.increase(vi.mHandle);
      }
      else if (old_rc < (*vi.mHandle).first)
      {
         mVarInfoHeap.decrease(vi.mHandle);
      }
   }
}

void
SUBool::DPEliminator::AddClauses(std::vector<Clause> resolvents)
{
   logs.TLog(3) << "[DPEliminator] resolvents size: " << resolvents.size()
                << std::endl;
   // std::sort(resolvents.begin(), resolvents.end(), Clause::SizeCompare());
   // print_range(std::cout, resolvents.begin(), resolvents.end(), " ");
   // std::cout<<std::endl;

   sort_unique(resolvents);
   // auto end =
   // std::remove_if(resolvents.begin(), resolvents.end(),
   //[&cnf](const auto &c) { return cnf.SubsumedByAny(c); });
   // resolvents.erase(end, resolvents.end());
   // logs.TLog(3) << "[DPEliminator] after 1st round of subsumptions: "
   //             << resolvents.size() << std::endl;
   // CNF res(std::move(resolvents));
   // res.RemoveSubsumptions();
   // res.Sort();
   // resolvents = std::move(res).Elements();
   auto cnf = Cnf();
   // sort_vector(resolvents);
   // std::cout<<std::endl;
   logs.TLog(3) << "             ... sorted unique: " << resolvents.size()
                << std::endl;
   bool log_progress = (resolvents.size() > 10000);
   size_t i = 0;
   OneProve op(mIgnorePolicy == IGNORE_POLICY::ABSORBED ? &cnf : nullptr);
   for (auto &clause : resolvents)
   {
      if (log_progress && (++i) % 1000 == 0)
      {
         logs.TLog(4) << 100.0 * static_cast<double>(i)
                             / static_cast<double>(resolvents.size())
                      << "%" << std::endl;
      }
      if (cnf.SubsumedByAny(clause))
      {
         continue;
      }
      if (mInprocessing == INPROCESSING::PRIME)
      {
         clause = *prime_subimplicate(mCnf, clause);
      }
      // TODO: Test absorbptions incrementally
      if (IgnoreClause(cnf, clause, op))
      {
         continue;
      }
      cnf.PushBack(clause);
      op.AddClause(cnf.Size() - 1);
      ++mTotalNewClauses;
      ++mCurrentNewClauses;
      mClauseInfo.emplace_back(std::move(clause), false);
      auto last = mClauseInfo.end();
      --last;
      for (const auto &lit : last->mClause)
      {
         if (!IsFree(lit.Variable()))
         {
            continue;
         }
         auto &vi = GetVarInfo(lit.Variable());
         if (lit.Positive())
         {
            vi.mPosClauses.push_back(last);
         }
         else
         {
            vi.mNegClauses.push_back(last);
         }
      }
   }
}

std::vector<SUBool::Clause>
SUBool::DPEliminator::VarInfo::Resolvents() const
{
   std::vector<Clause> resolvents;
   for (const auto &pos_clause : mPosClauses)
   {
      for (const auto &neg_clause : mNegClauses)
      {
         auto res = resolve(pos_clause->mClause, neg_clause->mClause);
         if (res.second > 0)
         {
            resolvents.push_back(std::move(res.first));
         }
      }
   }
   return resolvents;
}

void
SUBool::DPEliminator::VarInfo::RemoveInvalid()
{
   auto pred = [](ClauseInfo::tPointer ci) -> bool { return !ci->Valid(); };
   mPosClauses.remove_if(pred);
   mNegClauses.remove_if(pred);
}

bool
SUBool::DPEliminator::IgnoreClause(
    const CNF &cnf, const Clause &clause, OneProve &op) const
{
   switch (mIgnorePolicy)
   {
   case IGNORE_POLICY::NONE:
      return false;
   case IGNORE_POLICY::IMPLICATES:
      return is_implicate(cnf, clause);
   case IGNORE_POLICY::ABSORBED:
      return !op.IsEmpowering(clause);
   default:
      throw BadParameterException(
          "DPEliminator::IgnoreClause", "Unknown ignore policy");
   }
}
