/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file dpelim.h
 * DP elimination of variables in a CNF. Especially elimination of auxiliary
 * variables in a CNF encoding.
 */

#ifndef __DPELIM_H
#define __DPELIM_H

#include <boost/heap/fibonacci_heap.hpp>
#include <list>
#include <unordered_set>
#include <vector>

#include "empower.h"
#include "enum.h"
#include "normalform.h"

namespace SUBool
{

   class DPEliminator
   {

    public:
      static const size_t kEliminateAll = std::numeric_limits<size_t>::max();

      enum class IGNORE_POLICY
      {
         NONE,
         IMPLICATES,
         ABSORBED,
         Last = ABSORBED
      };

      enum class INPROCESSING
      {
         NONE,
         PRIME,
         Last = PRIME
      };

      enum class INCREASE_POLICY
      {
         // Stop when the estimate of increase would get the size over the bound
         ESTIMATE,
         // Stop when the size gets over the bound
         AFTER_BOUND,
         // Stop when the current number of new clauses gets over the bound
         NEW_CLAUSES,
         // Stop when the total number of new clauses gets over the bound
         NEW_CLAUSES_TOTAL,
         Last = NEW_CLAUSES_TOTAL
      };

      static const EnumAnnotation<IGNORE_POLICY> kIgnorePolicyAnnotation;
      static const EnumAnnotation<INPROCESSING> kInprocessingAnnotation;
      static const EnumAnnotation<INCREASE_POLICY> kIncreasePolicyAnnotation;

    private:
      using VarHeapElement = std::pair<size_t, unsigned>;
      using VarHeap = boost::heap::fibonacci_heap<VarHeapElement,
          boost::heap::compare<std::greater<VarHeapElement>>>;

      class ClauseInfo
      {
         bool mValid;
         bool mInput;

       public:
         using tPointer = std::list<ClauseInfo>::iterator;
         Clause mClause;

         ClauseInfo(Clause clause, bool input)
             : mValid(true), mInput(input), mClause(std::move(clause))
         {
         }
         void
         Invalidate()
         {
            mValid = false;
         }
         bool
         Valid() const
         {
            return mValid;
         }
         bool
         InputClause() const
         {
            return mInput;
         }
      };

      struct VarInfo
      {
         VarHeap::handle_type mHandle;
         std::list<ClauseInfo::tPointer> mPosClauses;
         std::list<ClauseInfo::tPointer> mNegClauses;

         size_t ResolventCount() const;
         std::vector<Clause> Resolvents() const;
         void
         ClearLists()
         {
            mPosClauses.clear();
            mNegClauses.clear();
         }
         void RemoveInvalid();
         VarInfo() : mHandle(), mPosClauses(), mNegClauses() {}
      };

      unsigned mMaxFixedVariable;
      unsigned mMaxVariable;
      std::list<ClauseInfo> mClauseInfo;
      std::list<ClauseInfo> mInvalidClauseInfo;
      std::vector<VarInfo> mVarInfo; // only non fixed variables

      VarInfo &GetVarInfo(unsigned var);
      bool IsFree(unsigned var);
      void InitClauseInfo(const CNF &cnf);
      void InitVarInfo();
      void InitVarHeap();
      void EliminateVarNotInHeap(unsigned var);

      std::list<ClauseInfo> InvalidateClauses(
          const std::list<ClauseInfo::tPointer> &clauses);
      void AddClauses(std::vector<Clause> resolvents);
      void UpdateVarInfo(std::unordered_set<unsigned> variables);

      VarHeap mVarInfoHeap;
      IGNORE_POLICY mIgnorePolicy;
      INPROCESSING mInprocessing;
      INCREASE_POLICY mIncreasePolicy;

      bool IgnoreClause(
          const CNF &cnf, const Clause &clause, OneProve &op) const;
      bool CheckIncrease(size_t initial_size, size_t bound) const;

      CNF mCnf;
      size_t mTotalNewClauses{0};
      size_t mCurrentNewClauses{0};

    public:
      DPEliminator(const CNF &cnf, unsigned max_fixed_variable,
          IGNORE_POLICY ignore_policy, INPROCESSING inprocessing,
          INCREASE_POLICY increase_policy);

      void EliminateVariable(unsigned var); // only non fixed variable allowed
      void EliminateAuxVariables(
          size_t max_size_increase, bool at_least_one = false);

      size_t
      Size() const
      {
         return mClauseInfo.size();
      }

      CNF Cnf() const &;
      CNF Cnf() &&;

      std::pair<CNF, std::vector<bool>> CnfWithMask() const &;
      std::pair<CNF, std::vector<bool>> CnfWithMask() &&;

      static CNF EliminateAuxVariables(const CNF &cnf,
          unsigned max_fixed_variable, IGNORE_POLICY ignore_policy,
          INPROCESSING inprocessing, INCREASE_POLICY increase_policy,
          size_t max_size_increase = std::numeric_limits<size_t>::max() / 2,
          bool at_least_one = false);
      static CNF EliminateAuxVariablesWithFactor(const CNF &cnf,
          unsigned max_fixed_variable, IGNORE_POLICY ignore_policy,
          INPROCESSING inprocessing, INCREASE_POLICY increase_policy,
          long double max_size_increase_factor, bool at_least_one = false);
   };

} // namespace SUBool

#endif
