/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file dpelim_opt.cpp
 * Program options related to DP elimination of variables in a CNF.
 */

#include "dpelim_opt.h"

const SUBool::EnumAnnotation<SUBool::DPElimEliminationOptions::ELIMINATE_POLICY>
    SUBool::DPElimEliminationOptions::kEliminatePolicyAnnotation("ELIMINATE",
        {//
            {ELIMINATE_POLICY::ALL, "all",
                "Eliminate all auxiliary variables."},
            {ELIMINATE_POLICY::MAX_INCREASE, "max_increase",
                "Elimination limited with the absolute increase of "
                "the size of the CNF."},
            {ELIMINATE_POLICY::FACTOR, "factor",
                "Elimination limited with the factor by which the "
                "size of the CNF can increase."}});

void
SUBool::DPElimMinimizeOptions::Init()
{
   AddOptions() //
       ("minimize,m", po::bool_switch(&mFinalMinimize),
           "Perform minimization at the end (not needed with -M option).") //
       ("elim-minimize,M", po::bool_switch(&mElimMinimize),
           "Use minimization during elimination.") //
       ("elim-minimize-step",
           po::value<unsigned>(&mSizeStep)->default_value(kDefaultSizeStep),
           "Size increase step between two minimizations.") //
       ("elim-minimize-factor",
           po::value<double>(&mSizeFactor)
               ->default_value(kDefaultSizeFactor, kDefaultSizeFactorString),
           "Factor of the cnf size to be added to the minimize step.");
}

bool
SUBool::DPElimMinimizeOptions::GetValues()
{
   return true;
}

SUBool::DPElimClauseManipOptions::DPElimClauseManipOptions()
    : OptionsGroup("Clause manipulation"),
      mBlockPrimeOption(CNFEncoding::kBlockPrimeAnnotation, "block-prime,P",
          "Make new clauses prime after each block of DP elimination.",
          kDefaultBlockPrime),
      mInprocessingOption(DPEliminator::kInprocessingAnnotation, "process,p",
          "how to inprocess the resolvent during auxiliary variable "
          "elimination.",
          kDefaultInprocessing),
      mIgnorePolicyOption(DPEliminator::kIgnorePolicyAnnotation,
          "ignore-policy,g",
          "Which resolvents should be discarded during auxiliary variable "
          "elimination.",
          kDefaultIgnorePolicy)
{
   Add(mBlockPrimeOption);
   Add(mInprocessingOption);
   Add(mIgnorePolicyOption);
}

bool
SUBool::DPElimClauseManipOptions::GetValues()
{
   return mBlockPrimeOption.GetValue(mBlockPrime)
          && mInprocessingOption.GetValue(mInprocessing)
          && mIgnorePolicyOption.GetValue(mIgnorePolicy);
}

SUBool::DPElimEliminationOptions::DPElimEliminationOptions()
    : OptionsGroup("Elimination options"),
      mEliminatePolicyOption(kEliminatePolicyAnnotation, "eliminate,e",
          "When to stop elimination", kDefaultEliminatePolicy),
      mIncreasePolicyOption(DPEliminator::kIncreasePolicyAnnotation,
          "increase-policy,C", "How the size bound should be checked.",
          kDefaultIncreasePolicy)
{
   Add(mEliminatePolicyOption);
   AddOptions() //
       ("max-input-var,x", po::value<unsigned>(&mMaxInputVar)->default_value(0),
           "Maximum input variable. All variables with higher index are "
           "considered auxiliary and possibly eliminated.") //
       ("max-increase,s",
           po::value<unsigned>(&mMaxIncrease)
               ->default_value(kDefaultMaxIncrease),
           "Stop the elimination after the number of clauses increases by the "
           "given amount") //
       ("factor,f",
           po::value<long double>(&mFactor)->default_value(
               kDefaultFactor, kDefaultFactorString),
           "Stop the elimination when the number of clauses increases by the "
           "given factor");
   Add(mIncreasePolicyOption);
}

bool
SUBool::DPElimEliminationOptions::GetValues()
{
   return mEliminatePolicyOption.GetValue(mEliminatePolicy)
          && mIncreasePolicyOption.GetValue(mIncreasePolicy);
}
