/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file dpelim_opt.h
 * Program options related to DP elimination of variables in a CNF.
 */

#ifndef __DPELIM_OPT_H
#define __DPELIM_OPT_H

#include <boost/program_options.hpp>

#include "dpelim.h"
#include "encoding.h"
#include "enum_opt.h"
#include "prg_opt_group.h"

namespace po = boost::program_options;

namespace SUBool
{
   class DPElimMinimizeOptions : public OptionsGroup
   {
    public:
      static constexpr unsigned kDefaultSizeStep = 5000u;
      static constexpr double kDefaultSizeFactor = 0.6;
      inline static const std::string kDefaultSizeFactorString = "0.6";

    private:
      bool mElimMinimize{false};
      bool mFinalMinimize{false};
      unsigned mSizeStep{kDefaultSizeStep};
      double mSizeFactor{kDefaultSizeFactor};

      void Init();

    public:
      DPElimMinimizeOptions();

      virtual bool GetValues() override;

      bool ElimMinimize() const;
      bool FinalMinimize() const;
      unsigned SizeStep() const;
      double SizeFactor() const;
   };

   class DPElimClauseManipOptions : public OptionsGroup
   {
    public:
      static constexpr CNFEncoding::BLOCK_PRIME kDefaultBlockPrime =
          CNFEncoding::BLOCK_PRIME::NONE;
      static constexpr DPEliminator::INPROCESSING kDefaultInprocessing =
          DPEliminator::INPROCESSING::NONE;
      static constexpr DPEliminator::IGNORE_POLICY kDefaultIgnorePolicy =
          DPEliminator::IGNORE_POLICY::ABSORBED;

    private:
      EnumOption<CNFEncoding::BLOCK_PRIME> mBlockPrimeOption;
      EnumOption<DPEliminator::INPROCESSING> mInprocessingOption;
      EnumOption<DPEliminator::IGNORE_POLICY> mIgnorePolicyOption;

      CNFEncoding::BLOCK_PRIME mBlockPrime{kDefaultBlockPrime};
      DPEliminator::INPROCESSING mInprocessing{kDefaultInprocessing};
      DPEliminator::IGNORE_POLICY mIgnorePolicy{kDefaultIgnorePolicy};

    public:
      DPElimClauseManipOptions();

      virtual bool GetValues() override;

      CNFEncoding::BLOCK_PRIME BlockPrime() const;
      DPEliminator::INPROCESSING Inprocessing() const;
      DPEliminator::IGNORE_POLICY IgnorePolicy() const;
   };

   class DPElimEliminationOptions : public OptionsGroup
   {
    public:
      enum class ELIMINATE_POLICY
      {
         ALL,
         MAX_INCREASE,
         FACTOR,
         Last = FACTOR
      };

      static const EnumAnnotation<ELIMINATE_POLICY> kEliminatePolicyAnnotation;

      static constexpr ELIMINATE_POLICY kDefaultEliminatePolicy =
          ELIMINATE_POLICY::ALL;
      static constexpr DPEliminator::INCREASE_POLICY kDefaultIncreasePolicy =
          DPEliminator::INCREASE_POLICY::AFTER_BOUND;
      static constexpr unsigned kDefaultMaxIncrease = 1;
      static constexpr long double kDefaultFactor = 1.2l;
      inline static const std::string kDefaultFactorString{"1.2"};

    private:
      EnumOption<ELIMINATE_POLICY> mEliminatePolicyOption;
      EnumOption<DPEliminator::INCREASE_POLICY> mIncreasePolicyOption;

      unsigned mMaxInputVar{0};
      ELIMINATE_POLICY mEliminatePolicy{kDefaultEliminatePolicy};
      DPEliminator::INCREASE_POLICY mIncreasePolicy{kDefaultIncreasePolicy};
      unsigned mMaxIncrease{kDefaultMaxIncrease};
      long double mFactor{kDefaultFactor};

    public:
      DPElimEliminationOptions();

      virtual bool GetValues() override;

      ELIMINATE_POLICY EliminatePolicy() const;
      DPEliminator::INCREASE_POLICY IncreasePolicy() const;
      unsigned MaxIncrease() const;
      long double Factor() const;
      unsigned MaxInputVar() const;
   };

} // namespace SUBool

inline SUBool::DPElimMinimizeOptions::DPElimMinimizeOptions()
    : OptionsGroup("Minimize options")
{
   Init();
}

inline bool
SUBool::DPElimMinimizeOptions::ElimMinimize() const
{
   return mElimMinimize;
}

inline bool
SUBool::DPElimMinimizeOptions::FinalMinimize() const
{
   return mFinalMinimize;
}

inline unsigned
SUBool::DPElimMinimizeOptions::SizeStep() const
{
   return mSizeStep;
}

inline double
SUBool::DPElimMinimizeOptions::SizeFactor() const
{
   return mSizeFactor;
}

inline SUBool::CNFEncoding::BLOCK_PRIME
SUBool::DPElimClauseManipOptions::BlockPrime() const
{
   return mBlockPrime;
}

inline SUBool::DPEliminator::INPROCESSING
SUBool::DPElimClauseManipOptions::Inprocessing() const
{
   return mInprocessing;
}

inline SUBool::DPEliminator::IGNORE_POLICY
SUBool::DPElimClauseManipOptions::IgnorePolicy() const
{
   return mIgnorePolicy;
}

inline SUBool::DPElimEliminationOptions::ELIMINATE_POLICY
SUBool::DPElimEliminationOptions::EliminatePolicy() const
{
   return mEliminatePolicy;
}

inline SUBool::DPEliminator::INCREASE_POLICY
SUBool::DPElimEliminationOptions::IncreasePolicy() const
{
   return mIncreasePolicy;
}

inline unsigned
SUBool::DPElimEliminationOptions::MaxIncrease() const
{
   return mMaxIncrease;
}
inline unsigned
SUBool::DPElimEliminationOptions::MaxInputVar() const
{
   return mMaxInputVar;
}

inline long double
SUBool::DPElimEliminationOptions::Factor() const
{
   return mFactor;
}

#endif
