/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file drenc.h
 * Functions related to dual rail encoding.
 */
#include <algorithm>

#include "drenc.h"

const SUBool::EnumAnnotation<SUBool::LIT_IMPL_SYSTEM>
    SUBool::kLitImplSystemAnnotation("LIT_IMPL_SYSTEM",
        {                                                                  //
            {LIT_IMPL_SYSTEM::DRENC, "drenc", "plain dual rail encoding"}, //
            {LIT_IMPL_SYSTEM::IRREDUNDANT, "irredundant",
                "make the dual rail encoding irredundant (both body and right "
                "irredundant)"}, //
            {LIT_IMPL_SYSTEM::GDBASIS, "gdbasis",
                "use GD basis of the dual rail encoding."}, //
            {LIT_IMPL_SYSTEM::GDBASIS_RIGHT_IRRED, "gdbasis_right_irred",
                "construct the GD basis and make it right irredundant."}, //
            {LIT_IMPL_SYSTEM::BODY_MINIMAL, "body_minimal",
                "make the dual rail encoding body minimal (like GD basis but "
                "without left saturation)."}, //
            {LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED, "body_minimal_irred",
                "make the dual rail encoding body minimal and right "
                "irredundant."}});

SUBool::LitImplSystem
SUBool::init_lit_impl_system(const CNF &cnf, LIT_IMPL_SYSTEM is_kind)
{
   auto impl_system = dual_rail(cnf, true);
   if (is_kind == LIT_IMPL_SYSTEM::GDBASIS
       || is_kind == LIT_IMPL_SYSTEM::GDBASIS_RIGHT_IRRED)
   {
      impl_system.MakeGDBasis();
   }
   else if (is_kind == LIT_IMPL_SYSTEM::BODY_MINIMAL
            || is_kind == LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED)
   {
      impl_system.MakeBodyMinimal();
   }
   else if (is_kind == LIT_IMPL_SYSTEM::IRREDUNDANT)
   {
      impl_system.MakeBodyIrredundant();
   }
   if (is_kind == LIT_IMPL_SYSTEM::IRREDUNDANT
       || is_kind == LIT_IMPL_SYSTEM::GDBASIS_RIGHT_IRRED
       || is_kind == LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED)
   {
      impl_system.MakeRightIrredundant();
   }
   return impl_system;
}

void
SUBool::append_clause_dual_rail(
    std::vector<LitImpl> &impl, const Clause &clause)
{
   if (clause.Empty())
   {
      throw BadParameterException("append_clause_dual_rail", "empty clause");
   }
   std::vector<Literal> body(clause.Size() - 1);
   for (size_t i = 0; i < clause.Size(); ++i)
   {
      if (std::remove_copy(
              clause.cbegin(), clause.cend(), body.begin(), clause[i])
          != body.end())
      {
         throw BadParameterException(
             "append_clause_dual_rail", "two equal literals");
      }
      std::for_each(body.begin(), body.end(), [](auto &lit) { lit.Negate(); });
      impl.push_back(LitImpl(body, {clause[i]}, true));
   }
}

SUBool::LitImplSystem
SUBool::dual_rail(const CNF &cnf, bool merge_bodies)
{
   std::vector<LitImpl> impl;
   for (const auto &clause : cnf)
   {
      append_clause_dual_rail(impl, clause);
   }
   LitImplSystem is(std::move(impl));
   if (merge_bodies)
   {
      is.MergeBodies();
   }
   return is;
}

void
SUBool::append_clauses_of_impl(
    std::vector<Clause> &clauses, const LitImpl &impl)
{
   LitImpl simple_impl = impl;
   simple_impl.MakeConsWithoutBody();
   Clause base;
   for (const auto &lit : simple_impl.Body())
   {
      base.PushBack(lit.Negation());
   }
   for (const auto &lit : simple_impl.Cons())
   {
      clauses.push_back(base);
      clauses.back().PushBack(lit);
   }
}

SUBool::CNF
SUBool::dual_rail_to_cnf(const LitImplSystem &is)
{
   std::vector<Clause> clauses;
   for (const auto &impl : is)
   {
      append_clauses_of_impl(clauses, impl);
   }
   CNF cnf(std::move(clauses));
   cnf.RemoveDuplicities();
   return cnf;
}

std::vector<SUBool::Clause>
SUBool::dr_clauses_of_impl(const LitImpl &impl)
{
   std::vector<Clause> clauses;
   append_clauses_of_impl(clauses, impl);
   return clauses;
}
