/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file empower.cpp
 * Functions related to empowering implicates and 1-provability.
 * Making CNF prime just by using unit propagation.
 */

#include "empower.h"
#include "cnfutil.h"
#include "logstream.h"

const SUBool::EnumAnnotation<SUBool::MINIMIZE_METHOD>
    SUBool::kMinimizeMethodAnnotation("MINIMIZE_METHOD",
        {
            //
            {SUBool::MINIMIZE_METHOD::FULL, "full", "full minimization"}, //
            {SUBool::MINIMIZE_METHOD::SIMPLE_INCREMENTAL, "simple_incremental",
                "sort the clauses in increasing order by sizes and then add "
                "nonabsorbed clauses in this sequence, starting with the empty "
                "CNF, the result is not necessarily fully minimized."}, //
            {SUBool::MINIMIZE_METHOD::FULL_INCREMENTAL, "full_incremental",
                "same as simple_incremental, in addition, now after a specific "
                "length of clauses is processed, they are reprocessed again as "
                "in "
                "full minimization. Only after then longer clauses are "
                "considered. "
                "The result is not necessarily fully minimized, but it can be "
                "significantly smaller than in case of simple_incremental."} //
        });

bool
SUBool::OneProve::IsOneProvable(const Clause &clause)
{
   return (mUP.Propagate(clause_falsifying(clause, mUP.MaxVariable()), nullptr)
           >= 0);
}

bool
SUBool::OneProve::IsEmpowering(const Clause &clause)
{
   return static_cast<bool>(CollectEmpoweredLiterals(clause, false));
}

std::optional<SUBool::Literal>
SUBool::OneProve::EmpoweredLiteral(const Clause &clause)
{
   auto emp = CollectEmpoweredLiterals(clause, false);
   if (!emp || emp->empty())
   {
      return {};
   }
   return std::make_optional(emp->front());
}

std::optional<std::vector<SUBool::Literal>>
SUBool::OneProve::AllEmpoweredLiterals(const Clause &clause)
{
   return CollectEmpoweredLiterals(clause, true);
}

std::optional<std::vector<SUBool::Literal>>
SUBool::OneProve::CollectEmpoweredLiterals(
    const Clause &clause, bool collect_all)
{
   auto alpha = clause_falsifying(clause, mUP.MaxVariable());
   // Check if the clause is 1-provable
   if (mUP.Propagate(alpha, nullptr) < 0)
   {
      if (collect_all)
      {
         return clause.Literals();
      }
      if (clause.Empty())
      {
         return std::vector<SUBool::Literal>();
      }
      return std::vector<SUBool::Literal>({clause[0]});
   }
   std::vector<Literal> emp_literals;
   for (const auto &lit : clause)
   {
      if (lit.Variable() > mUP.MaxVariable())
      {
         // The clause is 1-provable by previous test, and
         // the literal is on a variable which is not in the CNF.
         // Therefore, the clause cannot be empowering with respect to this
         // literal
         // Moreover, the literals in the clause are in increasing order, thus
         // the clause cannot be empowering with respect to any subsequent
         // literal either
         break;
      }
      alpha[lit.Variable()] = PartialValue::UNDEF;
      auto beta = Propagate(alpha);
      if (beta && (*beta)[lit.Variable()] == PartialValue::UNDEF)
      {
         emp_literals.push_back(lit);
         if (!collect_all)
         {
            break;
         }
      }
      alpha[lit.Variable()] = lit.FalseValue();
   }
   if (emp_literals.empty())
   {
      return {};
   }
   return std::make_optional(emp_literals);
}

std::optional<SUBool::PartialAssignment>
SUBool::OneProve::Propagate(const PartialAssignment &alpha)
{
   PartialAssignment beta(0);
   if (mUP.Propagate(alpha, &beta) >= 0)
   {
      return {};
   }
   return std::make_optional(std::move(beta));
}

SUBool::Clause
SUBool::OneProve::MinOneProvableSubClause(
    const Clause &clause, unsigned ignore_var)
{
   auto alpha = clause_falsifying(clause, mUP.MaxVariable());
   if (mUP.Propagate(alpha, nullptr) < 0)
   {
      return clause;
   }
   // Try to remove single literal
   for (const auto &lit : clause)
   {
      unsigned var = lit.Variable();
      if (var == ignore_var)
      {
         continue;
      }
      alpha[var] = PartialValue::UNDEF;
      if (mUP.Propagate(alpha, nullptr) < 0)
      {
         alpha[var] = lit.FalseValue();
      }
   }
   return Clause(alpha);
}

bool
SUBool::op_is_empowering(const CNF &cnf, const Clause &clause)
{
   OneProve op(&cnf);
   return op.IsEmpowering(clause);
}

bool
SUBool::op_is_empowering_with_mask(
    const CNF &cnf, const Clause &clause, const std::vector<bool> &mask)
{
   OneProve op(&cnf, [&mask](unsigned i) { return mask[i]; });
   return op.IsEmpowering(clause);
}

bool
SUBool::is_one_provable(const CNF &cnf, const Clause &clause)
{
   OneProve op(&cnf);
   return op.IsOneProvable(clause);
}

SUBool::CNF
SUBool::empower_minimize_full(CNF cnf, size_t log_step)
{
   auto m = cnf.Size();
   OneProve op(&cnf);
   std::vector<bool> mask(cnf.Size(), true);
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      if (log_step > 0 && i > 0 && i % log_step == 0)
      {
         double p = (static_cast<double>(i) / static_cast<double>(m)) * 100.0;
         logs.TLog(4) << p << "%" << std::endl;
      }
      mask[i] = false;
      op.RemoveClause(i);
      if (op.IsEmpowering(cnf[i]))
      {
         mask[i] = true;
         op.AddClause(i);
      }
   }
   cnf.RemoveIndices([&mask](auto i) { return !mask[i]; });
   return cnf;
}

namespace SUBool
{
   unsigned
   reprocess(const CNF &cnf, OneProve &op, std::vector<bool> &mask,
       const unsigned size, const unsigned clause_size, std::queue<unsigned> &q)
   {
      unsigned new_size = size;
      logs.TLog(3) << "reprocessing clause size" << clause_size
                   << " (queue size: " << q.size() << ")" << std::endl;
      while (!q.empty())
      {
         auto j = q.front();
         q.pop();
         op.RemoveClause(j);
         if (op.IsEmpowering(cnf[j]))
         {
            op.AddClause(j);
         }
         else
         {
            --new_size;
            mask[j] = false;
         }
      }
      logs.TLog(3) << "reprocessing finished (saved: " << size - new_size << ")"
                   << std::endl;
      return new_size;
   }
} // namespace SUBool

SUBool::CNF
SUBool::empower_minimize_inc(CNF cnf, bool reprocessing, size_t log_step)
{
   logs.TLog(3) << "[empower_minimize_inc] sorting" << std::endl;
   cnf.Sort(Clause::SizeCompare());
   logs.TLog(3) << "[empower_minimize_inc] entering the main cycle"
                << std::endl;
   OneProve op(&cnf, false);
   std::vector<bool> mask(cnf.Size(), false);
   auto m = cnf.Size();
   unsigned clause_size = cnf.Front().Size();
   logs.TLog(3) << "[empower_minimize_inc] min clause size=" << clause_size
                << ", max clause size=" << cnf.Back().Size() << std::endl;
   std::queue<unsigned> q;
   unsigned size = 0;
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      if (clause_size < cnf[i].Size())
      {
         if (reprocessing)
         {
            size = reprocess(cnf, op, mask, size, clause_size, q);
         }
         clause_size = cnf[i].Size();
         logs.TLog(3) << "next clause size=" << clause_size << std::endl;
      }
      if (log_step > 0 && i > 0 && i % log_step == 0)
      {
         double p = (static_cast<double>(i) / static_cast<double>(m)) * 100.0;
         logs.TLog(4) << p << "%" << std::endl;
      }
      if (op.IsEmpowering(cnf[i]))
      {
         mask[i] = true;
         op.AddClause(i);
         if (reprocessing)
         {
            q.push(i);
            ++size;
         }
      }
   }
   if (reprocessing)
   {
      size = reprocess(cnf, op, mask, size, clause_size, q);
   }
   cnf.RemoveIndices([&mask](auto i) { return !mask[i]; });
   logs.TLog(3) << "[empower_minimize_inc] result (p cnf " << cnf.MaxVariable()
                << " " << cnf.Size() << ")" << std::endl;
   return cnf;
}

SUBool::CNF
SUBool::empower_minimize(CNF cnf, MINIMIZE_METHOD method, size_t log_step)
{
   switch (method)
   {
   case MINIMIZE_METHOD::FULL:
      return empower_minimize_full(std::move(cnf), log_step);
   case MINIMIZE_METHOD::SIMPLE_INCREMENTAL:
      return empower_minimize_inc(std::move(cnf), false, log_step);
   case MINIMIZE_METHOD::FULL_INCREMENTAL:
      return empower_minimize_inc(std::move(cnf), true, log_step);
   default:
      throw BadParameterException(
          "empower_minimize", "Bad minimization method");
   };
}
