/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file empower.h
 * Functions related to empowering implicates and 1-provability.
 * Making CNF prime just by using unit propagation.
 */

#ifndef __EMPOWER_H
#define __EMPOWER_H

#include <array>

#include "enum.h"
#include "logstream.h"
#include "unitprop.h"
#include "util.h"

namespace SUBool
{
   class OneProve
   {
      UnitPropagator mUP;

    public:
      OneProve(const CNF *cnf, bool rebuild = true) : mUP(cnf, rebuild) {}
      template <class Pred> OneProve(const CNF *cnf, Pred pred) : mUP(cnf, pred)
      {
      }

      template <class Pred>
      void
      Rebuild(Pred pred)
      {
         mUP.Rebuild(pred);
      }

      void
      Rebuild()
      {
         mUP.Rebuild();
      }

      // The clause must be in the CNF
      void
      AddClause(size_t clause_index)
      {
         mUP.AddClause(clause_index);
      }
      void
      RemoveClause(size_t clause_index)
      {
         mUP.RemoveClause(clause_index);
      }

      const UnitPropagator &
      UP() const
      {
         return mUP;
      }

      bool IsOneProvable(const Clause &clause);
      bool IsEmpowering(const Clause &clause);
      // Returns an empty clause if the clause is not 1-provable
      Clause MinOneProvableSubClause(
          const Clause &clause, unsigned ignore_var = 0);
      std::optional<std::vector<Literal>> AllEmpoweredLiterals(
          const Clause &clause);
      std::optional<Literal> EmpoweredLiteral(const Clause &clause);
      std::optional<std::vector<Literal>> CollectEmpoweredLiterals(
          const Clause &clause, bool collect_all);
      std::optional<PartialAssignment> Propagate(
          const PartialAssignment &alpha);
   };

   bool op_is_empowering(const CNF &cnf, const Clause &clause);
   bool op_is_empowering_with_mask(
       const CNF &cnf, const Clause &clause, const std::vector<bool> &mask);

   bool is_one_provable(const CNF &cnf, const Clause &clause);

   template <class Pred> CNF up_prime_cnf_with_ignore(CNF cnf, Pred pred);

   inline CNF
   up_prime_cnf(CNF cnf)
   {
      return up_prime_cnf_with_ignore(
          std::move(cnf), [](auto) { return false; });
   }

   enum class MINIMIZE_METHOD
   {
      FULL,
      SIMPLE_INCREMENTAL,
      FULL_INCREMENTAL
   };

   extern const EnumAnnotation<MINIMIZE_METHOD> kMinimizeMethodAnnotation;
   inline const MINIMIZE_METHOD kDefaultMinimizeMethod = MINIMIZE_METHOD::FULL;

   CNF empower_minimize_full(CNF cnf, size_t log_step = 0);
   CNF empower_minimize_inc(CNF cnf, bool reprocessing, size_t log_step = 0);

   CNF empower_minimize(CNF cnf, MINIMIZE_METHOD method, size_t log_step = 0);
} // namespace SUBool

template <class Pred>
SUBool::CNF
SUBool::up_prime_cnf_with_ignore(CNF cnf, Pred pred)
{
   // keep prime copy,
   // add nonignored clauses in a cycle, then pred(i) is checked before
   // subsumedbyany
   // subsumed clauses only remove from op, not add back
   std::unordered_set<size_t> prime;
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      if (pred(i))
      {
         prime.insert(i);
      }
   }
   OneProve op(&cnf);
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      if (pred(i))
      {
         continue;
      }
      bool subsumed = false;
      for (const auto &j : prime)
      {
         if (cnf[j].Subsumes(cnf[i]))
         {
            subsumed = true;
            break;
         }
      }
      if (subsumed)
      {
         op.RemoveClause(i);
         continue;
      }
      auto c = op.MinOneProvableSubClause(cnf[i]);
      if (c.Empty())
      {
         return CNF(Clause());
      }
      if (c.Size() < cnf[i].Size())
      {
         op.RemoveClause(i);
         cnf.Set(i, std::move(c));
         op.AddClause(i);
      }
      prime.insert(i);
   }
   cnf.RemoveIndices([&prime](auto i) { return prime.find(i) == prime.end(); });
   return cnf;
}

#endif
