/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file empower.cpp
 * Functions related to empowering implicates and 1-provability.
 * Using UnitPropagatorM from unitprop2.h
 */

#include "empower2.h"
#include "cnfutil.h"
#include "logstream.h"

bool
SUBool::OneProveM::IsOneProvable(const Clause &clause) const
{
   Clause c_neg(clause);
   c_neg.NegateLiterals();
   return !mUP.Consistent(c_neg.Literals());
}

bool
SUBool::OneProveM::IsEmpowering(const Clause &clause) const
{
   return static_cast<bool>(CollectEmpoweredLiterals(clause, false));
}

std::optional<SUBool::Literal>
SUBool::OneProveM::EmpoweredLiteral(const Clause &clause) const
{
   auto emp = CollectEmpoweredLiterals(clause, false);
   if (!emp || emp->empty())
   {
      return {};
   }
   return std::make_optional(emp->front());
}

std::optional<std::vector<SUBool::Literal>>
SUBool::OneProveM::AllEmpoweredLiterals(const Clause &clause) const
{
   return CollectEmpoweredLiterals(clause, true);
}

std::optional<std::vector<SUBool::Literal>>
SUBool::OneProveM::CollectEmpoweredLiterals(const Clause &clause,
                                            bool collect_all) const
{
   if (clause.Trivial())
   {
      return {};
   }
   auto alpha = clause_falsifying(clause, mUP.MaxVariable());
   // Check if the clause is 1-provable
   if (mUP.Consistent(alpha))
   {
      if (collect_all)
      {
         return clause.Literals();
      }
      if (clause.Empty())
      {
         return std::vector<SUBool::Literal>();
      }
      return std::vector<SUBool::Literal>({clause[0]});
   }
   std::vector<Literal> emp_literals;
   for (const auto &lit : clause)
   {
      if (lit.Variable() > mUP.MaxVariable())
      {
         // The clause is 1-provable by previous test, and
         // the literal is on a variable which is not in the CNF.
         // Therefore, the clause cannot be empowering with respect to this
         // literal
         // Moreover, the literals in the clause are in increasing order, thus
         // the clause cannot be empowering with respect to any subsequent
         // literal either
         break;
      }
      alpha[lit.Variable()] = PartialValue::UNDEF;
      auto beta = Propagate(alpha);
      if (beta && (*beta)[lit.Variable()] == PartialValue::UNDEF)
      {
         emp_literals.push_back(lit);
         if (!collect_all)
         {
            break;
         }
      }
      alpha[lit.Variable()] = lit.FalseValue();
   }
   if (emp_literals.empty())
   {
      return {};
   }
   return emp_literals;
}

SUBool::Clause
SUBool::OneProveM::MinOneProvableSubClause(const Clause &clause,
                                           unsigned ignore_var) const
{
   if (mUP.Contradictory())
   {
      return Clause::EmptyClause();
   }
   auto alpha = clause_falsifying(clause, mUP.MaxVariable());
   if (mUP.Consistent(alpha))
   {
      return clause;
   }
   // Try to remove single literal
   for (const auto &lit : clause)
   {
      unsigned var = lit.Variable();
      if (var == ignore_var)
      {
         continue;
      }
      alpha[var] = PartialValue::UNDEF;
      if (mUP.Consistent(alpha))
      {
         alpha[var] = lit.FalseValue();
      }
   }
   return Clause(alpha);
}

void
SUBool::OneProveM::MinimizeFull(size_t log_step)
{
   if (mUP.Contradictory())
   {
      return;
   }
   auto m = mUP.NumOfClauses();
   auto i_clause = mUP.begin();
   size_t cnt = 0;
   while (i_clause != mUP.end())
   {
      ++cnt;
      if (log_step > 0 && cnt % log_step == 0)
      {
         double p = (static_cast<double>(cnt) / static_cast<double>(m)) * 100.0;
         logs.TLog(4) << p << "%" << std::endl;
      }
      mUP.DisableClause(i_clause);
      if (IsEmpowering(i_clause->clause))
      {
         mUP.EnableClause(i_clause);
         ++i_clause;
      }
      else
      {
         i_clause = mUP.RemoveClause(i_clause);
      }
   }
}

void
SUBool::OneProveM::Reprocess(const unsigned clause_size,
                             std::queue<UnitPropagatorM::const_iterator> &q)
{
   unsigned old_size = mUP.NumOfClauses();
   logs.TLog(4) << "reprocessing clause size " << clause_size
                << " (queue size: " << q.size() << ")" << std::endl;
   while (!q.empty())
   {
      auto i_clause = q.front();
      q.pop();
      mUP.DisableClause(i_clause);
      if (IsEmpowering(i_clause->clause))
      {
         mUP.EnableClause(i_clause);
      }
      else
      {
         mUP.RemoveClause(i_clause);
      }
   }
   logs.TLog(4) << "reprocessing finished (saved: "
                << old_size - mUP.NumOfClauses() << ")" << std::endl;
}

void
SUBool::OneProveM::AddWithIncrementalMinimization(CNF cnf, bool reprocessing,
                                                  size_t log_step)
{
   logs.TLog(3)
       << "[OneProveM::AddWithIncrementalMinimization] minimizing (p cnf "
       << cnf.MaxVariable() << " " << cnf.Size() << ")" << std::endl;
   cnf.Sort(Clause::SizeCompare());
   logs.TLog(4)
       << "[OneProveM::AddWithIncrementalMinimization] entering the main cycle"
       << std::endl;
   auto m = cnf.Size();
   unsigned clause_size = cnf.Front().Size();
   logs.TLog(4)
       << "[OneProveM::AddWithIncrementalMinimization] min clause size="
       << clause_size << ", max clause size=" << cnf.Back().Size() << std::endl;
   std::queue<UnitPropagatorM::const_iterator> q;
   for (unsigned i = 0; i < cnf.Size() && !mUP.Contradictory(); ++i)
   {
      if (clause_size < cnf[i].Size())
      {
         if (reprocessing)
         {
            Reprocess(clause_size, q);
         }
         clause_size = cnf[i].Size();
         logs.TLog(4) << "next clause size=" << clause_size << std::endl;
      }
      if (log_step > 0 && i > 0 && i % log_step == 0)
      {
         double p = (static_cast<double>(i) / static_cast<double>(m)) * 100.0;
         logs.TLog(4) << p << "%" << std::endl;
      }
      if (IsEmpowering(cnf[i]))
      {
         auto i_clause = mUP.AddClause(std::move(cnf[i]));
         if (i_clause != mUP.end() && reprocessing)
         {
            q.push(i_clause);
         }
      }
   }
   if (reprocessing)
   {
      Reprocess(clause_size, q);
   }
   logs.TLog(3) << "[OneProveM::AddWithIncrementalMinimization] result (p cnf "
                << cnf.MaxVariable() << " " << cnf.Size() << ")" << std::endl;
}

std::vector<SUBool::ClauseWithEmpowered>
SUBool::OneProveM::MinimizeWithCollectingEmpowered(bool collect_all,
                                                   size_t log_step)
{
   if (mUP.Contradictory())
   {
      return {ClauseWithEmpowered{Clause::EmptyClause(), {}}};
   }
   std::vector<ClauseWithEmpowered> clauses_with_emp;
   auto m = mUP.NumOfClauses();
   auto i_clause = mUP.begin();
   size_t cnt = 0;
   while (i_clause != mUP.end())
   {
      ++cnt;
      if (log_step > 0 && cnt % log_step == 0)
      {
         double p = (static_cast<double>(cnt) / static_cast<double>(m)) * 100.0;
         logs.TLog(4) << p << "%" << std::endl;
      }
      mUP.DisableClause(i_clause);
      auto emp = CollectEmpoweredLiterals(i_clause->clause, collect_all);
      if (emp)
      {
         mUP.EnableClause(i_clause);
         clauses_with_emp.emplace_back(i_clause->clause, std::move(*emp));
         ++i_clause;
      }
      else
      {
         i_clause = mUP.RemoveClause(i_clause);
      }
   }
   return clauses_with_emp;
}
