/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file empower2.h
 * Functions related to empowering implicates and 1-provability.
 * Using UnitPropagatorM from unitprop2.h
 */

#ifndef __EMPOWER2_H
#define __EMPOWER2_H

#include <array>

#include "logstream.h"
#include "unitprop2.h"
#include "util.h"

namespace SUBool
{
   struct ClauseWithEmpowered
   {
      Clause clause;
      std::vector<Literal> empowered;

      ClauseWithEmpowered() : clause(), empowered() {}
      ClauseWithEmpowered(Clause c, std::vector<Literal> emp)
          : clause(std::move(c)), empowered(std::move(emp))
      {
      }
   };

   class OneProveM
   {
      UnitPropagatorM mUP;

      void Reprocess(const unsigned clause_size,
          std::queue<UnitPropagatorM::const_iterator> &q);

    public:
      OneProveM() : mUP() {}
      OneProveM(const CNF &cnf) : mUP(cnf) {}
      OneProveM(CNF &&cnf) : mUP(std::move(cnf)) {}

      // The clause must be in the CNF
      void
      AddClause(Clause clause)
      {
         mUP.AddClause(std::move(clause));
      }

      const UnitPropagatorM &
      UP() const
      {
         return mUP;
      }

      CNF
      ExtractCNF() const &
      {
         return mUP.ExtractCNF();
      }
      CNF
      ExtractCNF() &&
      {
         return std::move(mUP).ExtractCNF();
      }

      bool IsOneProvable(const Clause &clause) const;
      bool IsEmpowering(const Clause &clause) const;
      // Returns an empty clause if the clause is not 1-provable
      Clause MinOneProvableSubClause(
          const Clause &clause, unsigned ignore_var = 0) const;
      std::optional<std::vector<Literal>> AllEmpoweredLiterals(
          const Clause &clause) const;
      std::optional<Literal> EmpoweredLiteral(const Clause &clause) const;
      std::optional<std::vector<Literal>> CollectEmpoweredLiterals(
          const Clause &clause, bool collect_all) const;
      std::optional<PartialAssignment>
      Propagate(const PartialAssignment &alpha) const
      {
         return mUP.Propagate(alpha);
      }
      bool
      Closed(const PartialAssignment &alpha) const
      {
         return mUP.Closed(alpha);
      }

      void MinimizeFull(size_t log_step = 0);
      void AddWithIncrementalMinimization(
          CNF cnf, bool reprocessing, size_t log_step = 0);
      std::vector<ClauseWithEmpowered> MinimizeWithCollectingEmpowered(
          bool collect_all, size_t log_step = 0);
   };

} // namespace SUBool
#endif
