/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file comp_hyp_is_provider.cpp
 * Implication system provider for Encothesis.
 */

#include "enc_is_provider.h"

void
SUBool::add_clause_dr_to_impl_system(
    const Clause &clause, LitImplSystem &impl_system)
{
   std::vector<LitImpl> clause_dr;
   append_clause_dual_rail(clause_dr, clause);
   for (auto &impl : clause_dr)
   {
      impl_system.PushBack(std::move(impl));
   }
}

void
SUBool::EncISProvider::AddClause(const Clause &clause)
{
   if (mImplSystem)
   {
      add_clause_dr_to_impl_system(clause, *mImplSystem);
      mSCCGraphNeedsRebuild = true;
   }
}

const SUBool::LitImplSystem &
SUBool::EncISProvider::OneProveImplSystem() const
{
   return *mImplSystem;
}

const SUBool::LitImplSystem &
SUBool::EncISProvider::EmpowerImplSystem() const
{
   return *mImplSystem;
}

SUBool::LitImplSystem
SUBool::EncISProvider::ProcessedEmpowerImplSystem() const
{
   return EmpowerImplSystem();
}

void
SUBool::EncISProvider::RebuildSCCGraph()
{
   if (mImplSystem && mSCCGraphNeedsRebuild)
   {
      mSCCGraph = LitISGraph(*mImplSystem);
      mSCCGraphNeedsRebuild = false;
      logs.TLog(3)
          << "[EncISProvider::RebuildSCCGraph] Effective SCC var sizes:";
      for (unsigned i = 0; i < mSCCGraph->SCCCount(); ++i)
      {
         logs.Log(3) << " " << mSCCGraph->EffSCCVarSize(i);
      }
      logs.Log(3) << std::endl;
   }
}

bool
SUBool::EncISProvider::HasEmpowerImplSystem() const
{
   return mImplSystem.has_value();
}

void
SUBool::DummyISProvider::Cleanup(const CNF &)
{
   logs.TLog(3) << "[DummyISProvider::Cleanup] Nothing to do" << std::endl;
}

const SUBool::LitISGraph &
SUBool::EncISProvider::SCCGraph() const
{
   return *mSCCGraph;
}

void
SUBool::DREncISProvider::Cleanup(const CNF &cnf)
{
   mImplSystem = dual_rail(cnf, true);
}

SUBool::LitImplSystem
SUBool::DREncISProvider::ProcessedEmpowerImplSystem() const
{
   auto impl_system = EmpowerImplSystem();
   impl_system.MakeBodyMinimal();
   impl_system.MakeRightIrredundant();
   return impl_system;
}

void
SUBool::IrredundantISProvider::Cleanup(const CNF &cnf)
{
   if (!mImplSystem)
   {
      mImplSystem = dual_rail(cnf, true);
   }
   mImplSystem->MakeBodyIrredundant();
   mImplSystem->MakeRightIrredundant();
}

SUBool::LitImplSystem
SUBool::IrredundantISProvider::ProcessedEmpowerImplSystem() const
{
   auto impl_system = EmpowerImplSystem();
   impl_system.MakeBodyMinimal();
   impl_system.MakeRightIrredundant();
   return impl_system;
}

void
SUBool::GDBasisISProvider::Cleanup(const CNF &cnf)
{
   if (!mImplSystem)
   {
      mImplSystem = dual_rail(cnf, true);
   }
   mImplSystem->MakeGDBasis();
   mSCCGraphNeedsRebuild = true;
}

SUBool::LitImplSystem
SUBool::GDBasisISProvider::ProcessedEmpowerImplSystem() const
{
   auto impl_system = EmpowerImplSystem();
   impl_system.MakeBodyMinimal();
   impl_system.MakeRightIrredundant();
   return impl_system;
}

void
SUBool::GDBasisRightIrredISProvider::Cleanup(const CNF &cnf)
{
   if (!mImplSystem)
   {
      mImplSystem = dual_rail(cnf, true);
   }
   mImplSystem->MakeGDBasis();
   mImplSystem->MakeRightIrredundant();
   mSCCGraphNeedsRebuild = true;
}

SUBool::LitImplSystem
SUBool::GDBasisRightIrredISProvider::ProcessedEmpowerImplSystem() const
{
   auto impl_system = EmpowerImplSystem();
   impl_system.MakeBodyMinimal();
   impl_system.MakeRightIrredundant();
   return impl_system;
}

void
SUBool::BodyMinimalISProvider::Cleanup(const CNF &cnf)
{
   if (!mImplSystem)
   {
      mImplSystem = dual_rail(cnf, true);
   }
   mImplSystem->MakeBodyMinimal();
}

SUBool::LitImplSystem
SUBool::BodyMinimalISProvider::ProcessedEmpowerImplSystem() const
{
   auto impl_system = EmpowerImplSystem();
   impl_system.MakeRightIrredundant();
   return impl_system;
}

void
SUBool::BodyMinimalIrredISProvider::Cleanup(const CNF &cnf)
{
   if (!mImplSystem)
   {
      mImplSystem = dual_rail(cnf, true);
   }
   mImplSystem->MakeBodyMinimal();
   mImplSystem->MakeRightIrredundant();
}

std::shared_ptr<SUBool::EncISProvider>
SUBool::make_enc_is_provider(LIT_IMPL_SYSTEM kind)
{
   switch (kind)
   {
   case LIT_IMPL_SYSTEM::DRENC:
      return std::make_shared<DREncISProvider>();
   case LIT_IMPL_SYSTEM::IRREDUNDANT:
      return std::make_shared<IrredundantISProvider>();
   case LIT_IMPL_SYSTEM::GDBASIS:
      return std::make_shared<GDBasisISProvider>();
   case LIT_IMPL_SYSTEM::GDBASIS_RIGHT_IRRED:
      return std::make_shared<GDBasisRightIrredISProvider>();
   case LIT_IMPL_SYSTEM::BODY_MINIMAL:
      return std::make_shared<BodyMinimalISProvider>();
   case LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED:
      return std::make_shared<BodyMinimalIrredISProvider>();
   default:
      throw BadParameterException(
          "make_owning_is_provider", "Unknown implication system kind");
   };
}

std::shared_ptr<SUBool::EncISProvider>
SUBool::make_negatives_is_provider(
    LIT_IMPL_SYSTEM kind, const LitImplSystem &negatives)
{
   switch (kind)
   {
   case LIT_IMPL_SYSTEM::DRENC:
      return std::make_shared<NegDREncISProvider>(negatives);
   case LIT_IMPL_SYSTEM::IRREDUNDANT:
      return std::make_shared<NegIrredundantISProvider>(negatives);
   case LIT_IMPL_SYSTEM::GDBASIS:
      return std::make_shared<NegGDBasisISProvider>(negatives);
   case LIT_IMPL_SYSTEM::GDBASIS_RIGHT_IRRED:
      return std::make_shared<NegGDBasisRightIrredISProvider>(negatives);
   case LIT_IMPL_SYSTEM::BODY_MINIMAL:
      return std::make_shared<NegBodyMinimalISProvider>(negatives);
   case LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED:
      return std::make_shared<NegBodyMinimalIrredISProvider>(negatives);
   default:
      throw BadParameterException(
          "make_negatives_is_provider", "Unknown implication system kind");
   };
}

