/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file enc_is_provider.h
 * Implication system provider for the AbsEncoder.
 */

#ifndef __ENC_IS_PROVIDER
#define __ENC_IS_PROVIDER

#include "drenc.h"
#include "impl_system_graph.h"
#include "nfliteralset.h"
#include "normalform.h"

namespace SUBool
{
   class EncISProvider
   {
    protected:
      std::optional<LitImplSystem> mImplSystem{};
      std::optional<LitISGraph> mSCCGraph{};
      bool mSCCGraphNeedsRebuild{true};

    public:
      virtual ~EncISProvider() = default;
      virtual void Cleanup(const CNF &cnf) = 0;
      virtual void AddClause(const Clause &clause);
      virtual const LitImplSystem &OneProveImplSystem() const;
      virtual const LitImplSystem &EmpowerImplSystem() const;
      virtual LitImplSystem ProcessedEmpowerImplSystem() const;
      virtual void RebuildSCCGraph();
      virtual const LitISGraph &SCCGraph() const;
      virtual bool HasEmpowerImplSystem() const;
   };

   class DummyISProvider : public EncISProvider
   {
    public:
      virtual void Cleanup(const CNF &cnf);
   };

   class DREncISProvider : public EncISProvider
   {
    public:
      virtual void Cleanup(const CNF &cnf) override;
      virtual LitImplSystem ProcessedEmpowerImplSystem() const override;
   };

   class IrredundantISProvider : public EncISProvider
   {
    public:
      virtual void Cleanup(const CNF &cnf) override;
      virtual LitImplSystem ProcessedEmpowerImplSystem() const override;
   };

   class GDBasisISProvider : public EncISProvider
   {
    public:
      virtual void Cleanup(const CNF &cnf) override;
      virtual LitImplSystem ProcessedEmpowerImplSystem() const override;
   };

   class GDBasisRightIrredISProvider : public EncISProvider
   {
    public:
      virtual void Cleanup(const CNF &cnf) override;
      virtual LitImplSystem ProcessedEmpowerImplSystem() const override;
   };

   class BodyMinimalISProvider : public EncISProvider
   {
    public:
      virtual void Cleanup(const CNF &cnf) override;
      virtual LitImplSystem ProcessedEmpowerImplSystem() const override;
   };

   class BodyMinimalIrredISProvider : public EncISProvider
   {
    public:
      virtual void Cleanup(const CNF &cnf) override;
   };

   std::shared_ptr<EncISProvider> make_enc_is_provider(LIT_IMPL_SYSTEM kind);

   template <class ISP> class NegativesISProvider : public ISP
   {
      const LitImplSystem &mNegatives;

    public:
      NegativesISProvider(const LitImplSystem &negatives)
          : mNegatives(negatives)
      {
      }

      virtual const LitImplSystem &
      EmpowerImplSystem() const override
      {
         return mNegatives;
      }

      virtual bool
      HasEmpowerImplSystem() const override
      {
         return true;
      }

      virtual LitImplSystem
      ProcessedEmpowerImplSystem() const override
      {
         logs.TLog(3) << "[NegativesISProvider::ProcessedEmpowerImplSystem] "
                         "Preparing the implication system"
                      << std::endl;
         auto impl_system = EmpowerImplSystem();
         impl_system.MakeBodyMinimal();
         impl_system.MakeRightIrredundant();
         logs.TLog(3) << "[NegativesISProvider::ProcessedEmpowerImplSystem] "
                         "Implication system prepared"
                      << std::endl;
         return impl_system;
      }
   };

   using NegDummyISProvider = NegativesISProvider<DummyISProvider>;
   using NegGDBasisISProvider = NegativesISProvider<GDBasisISProvider>;
   using NegDREncISProvider = NegativesISProvider<DREncISProvider>;
   using NegIrredundantISProvider = NegativesISProvider<IrredundantISProvider>;
   using NegGDBasisRightIrredISProvider =
       NegativesISProvider<GDBasisRightIrredISProvider>;
   using NegBodyMinimalISProvider = NegativesISProvider<BodyMinimalISProvider>;
   using NegBodyMinimalIrredISProvider =
       NegativesISProvider<BodyMinimalIrredISProvider>;

   void add_clause_dr_to_impl_system(
       const Clause &clause, LitImplSystem &impl_system);

   std::shared_ptr<EncISProvider> make_negatives_is_provider(
       LIT_IMPL_SYSTEM kind, const LitImplSystem &negatives);

} // namespace SUBool

#endif
