/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file encoding.cpp
 * CNF encodings. URC and PC encodings implementation.
 */

#include "encoding.h"
#include "dimacs.h"
#include "dpelim.h"
#include "glucose_bind.h"
#include "pccomp.h"
#include "propagate.h"
#include "sbexception.h"

const size_t SUBool::CNFEncoding::kEliminateAll =
    SUBool::DPEliminator::kEliminateAll;
const std::vector<std::string> SUBool::CNFEncoding::kKindNames = {
    "sat", "i-urc", "urc", "i-pc", "pc"};
const std::regex SUBool::CNFEncoding::kHeaderRegex(
    "^[[:space:]]*c[[:space:]]+![[:space:]]*encoding[[:space:]]"
    "+(sat|i-urc|i-pc|urc|pc)[[:space:]]+([[:digit:]]+)[[:space:]]*$");
const SUBool::EnumAnnotation<SUBool::CNFEncoding::BLOCK_PRIME>
    SUBool::CNFEncoding::kBlockPrimeAnnotation("BLOCK_PRIME",
        {{BLOCK_PRIME::NONE, "none", "CNF is not made prime"},
            {BLOCK_PRIME::UP_PRIME, "up_prime",
                "use unit propagation for primality"},
            {BLOCK_PRIME::PRIME, "prime", "use SAT for primality"}});

SUBool::CNFEncoding::KIND
SUBool::CNFEncoding::KindOfName(const std::string &kind_name)
{
   for (unsigned k = static_cast<unsigned>(KIND::Begin);
        k < static_cast<unsigned>(KIND::End); ++k)
   {
      if (kKindNames[k] == kind_name)
      {
         return static_cast<KIND>(k);
      }
   }
   return KIND::UNKNOWN;
}

SUBool::CNFEncoding::~CNFEncoding() {}
SUBool::SATEncoding::~SATEncoding() {}
SUBool::URCEncoding::~URCEncoding() {}
SUBool::PCEncoding::~PCEncoding() {}

void
SUBool::CNFEncoding::CnfChanged()
{
}

void
SUBool::SATEncoding::CnfChanged()
{
   mSolver.Solver().Reset(mCnf);
   mUP = UnitPropagatorM(mCnf);
   mMinSubimplicant.Reset(&mCnf);
   mPropagator.Reset(mSolver.GetSolverPtr(), &mUP, &mMinSubimplicant);
}

void
SUBool::URCEncoding::CnfChanged()
{
   CPUTime cpu_time;
   auto &logs = SUBool::logs;
   logs.TLog(1) << "CNF structures update" << std::endl;
   mUP = UnitPropagatorM(mCnf);
   logs.TLog(1) << "CNF structures update finished, time=" << cpu_time
                << std::endl;
}

void
SUBool::CNFEncoding::EliminateAuxVariablesWithMaxSize(size_t max_size)
{
   if (max_size < mCnf.Size())
   {
      return;
   }
   CPUTime cpu_time;
   auto &logs = SUBool::logs;
   auto max_var = mCnf.MaxVariable();
   logs.TLog(1) << "Starting elimination (p cnf " << mCnf.MaxVariable() << " "
                << mCnf.Size() << "), max input variable: " << mMaxInputVariable
                << std::endl;
   mCnf = DPEliminator::EliminateAuxVariables(mCnf, mMaxInputVariable,
       mElimIgnorePolicy, mElimInprocessing, mElimIncreasePolicy,
       (max_size == kEliminateAll ? kEliminateAll : max_size - mCnf.Size()));
   mCnf.RemoveSubsumptions();
   CNFCompressor compressor(mCnf, mMaxInputVariable);
   mCnf = compressor.CompressCNF(std::move(mCnf));
   CnfChanged();
   logs.TLog(1) << "Elimination finished (p cnf " << mCnf.MaxVariable() << " "
                << mCnf.Size() << "), saved " << (max_var - mCnf.MaxVariable())
                << " variables, processor time=" << cpu_time << std::endl;
}

std::pair<SUBool::CNF, std::vector<bool>>
SUBool::CNFEncoding::CallEliminate(size_t size_inc, bool at_least_one)
{
   CNFCompressor compressor(mCnf, mMaxInputVariable);
   DPEliminator dpelim(compressor.CompressCNF(mCnf), mMaxInputVariable,
       mElimIgnorePolicy, mElimInprocessing, mElimIncreasePolicy);
   dpelim.EliminateAuxVariables(size_inc, at_least_one);
   auto res = std::move(dpelim).CnfWithMask();
   res.first = compressor.UncompressCNF(std::move(res.first));
   return res;
}

size_t
SUBool::CNFEncoding::SizeIncrease(
    size_t max_size, size_t min_steps, double min_factor) const
{
   size_t max_size_inc =
       (max_size == kEliminateAll ? kEliminateAll : max_size - mCnf.Size());
   double size_inc = static_cast<double>(min_steps)
                     + min_factor * static_cast<double>(mCnf.Size());
   if (size_inc > 0.0 && size_inc < static_cast<double>(max_size_inc))
   {
      return static_cast<size_t>(size_inc);
   }
   return max_size_inc;
}

void
SUBool::CNFEncoding::EliminateAuxVariablesWithMaxSizeMinimize(
    size_t max_size, size_t min_steps, double min_factor)
{
   if (max_size < mCnf.Size())
   {
      return;
   }
   CPUTime cpu_time;
   CNF ref_cnf;
   IncSolverWrapper ref_solver;
   if (mBlockPrime == BLOCK_PRIME::PRIME)
   {
      ref_cnf = mCnf;
      ref_solver.Solver().AddCnf(ref_cnf);
   }
   auto &logs = SUBool::logs;
   logs.TLog(1) << "Starting elimination with minimization (p cnf "
                << mCnf.MaxVariable() << " " << mCnf.Size()
                << "), max input variable: " << mMaxInputVariable << std::endl;
   // std::vector<bool> new_clauses_mask(mCnf.Size(), true);
   Minimize(kMinimizeLogStep);
   unsigned ref_max_var = mCnf.MaxVariable();
   unsigned curr_max_var = mCnf.MaxVariable();
   auto max_var = mCnf.MaxVariable();
   while (curr_max_var > mMaxInputVariable && max_size >= mCnf.Size())
   {
      auto old_max_var = mCnf.MaxVariable();
      auto size_inc = SizeIncrease(max_size, min_steps, min_factor);
      auto res = CallEliminate(size_inc, max_size == kEliminateAll);
      size_t max_size_inc =
          (max_size == kEliminateAll ? kEliminateAll : max_size - mCnf.Size());
      CNFCompressor compress(res.first, mMaxInputVariable);
      curr_max_var = compress.MaxVariable();
      if (curr_max_var < old_max_var)
      {
         switch (mBlockPrime)
         {
         case BLOCK_PRIME::UP_PRIME:
            mCnf = BlockPrimeUP(res.first, curr_max_var,
                [&res](auto i) { return !res.second[i]; });
            break;
         case BLOCK_PRIME::PRIME:
            mCnf = BlockPrime(res.first, curr_max_var, ref_solver.Solver(),
                [&res](auto i) { return !res.second[i]; });
            break;
         default:
            mCnf = std::move(res.first);
         }
         mCnf = compress.CompressCNF(std::move(mCnf));
         Minimize(kMinimizeLogStep);
         curr_max_var = mCnf.MaxVariable();
         mCnf = compress.UncompressCNF(std::move(mCnf));
      }
      else
      {
         if (max_size_inc == size_inc)
         {
            break;
         }
         double d_min_steps = static_cast<double>(min_steps) * kMinStepsFactor;
         if (d_min_steps < static_cast<double>(kEliminateAll))
         {
            min_steps = static_cast<size_t>(d_min_steps);
         }
         else
         {
            min_steps = kEliminateAll;
         }
      }
      if (mBlockPrime == BLOCK_PRIME::PRIME
          && curr_max_var + mCnf.Size() <= ref_max_var + ref_cnf.Size())
      {
         ref_max_var = curr_max_var;
         ref_cnf = mCnf;
         ref_solver.Solver().Reset(ref_cnf);
         logs.TLog(3) << "New ref_cnf (p cnf " << ref_max_var << " "
                      << ref_cnf.Size() << ")" << std::endl;
      }
   }
   CNFCompressor compress(mCnf, mMaxInputVariable);
   mCnf = compress.CompressCNF(std::move(mCnf));
   logs.TLog(1) << "Elimination finished (p cnf " << mCnf.MaxVariable() << " "
                << mCnf.Size() << "), saved " << (max_var - mCnf.MaxVariable())
                << " variables, processor time=" << cpu_time << std::endl;
}

void
SUBool::CNFEncoding::EliminateAuxVariablesWithMaxIncMinimize(
    size_t max_size_increase, size_t min_steps, double min_factor)
{
   EliminateAuxVariablesWithMaxSizeMinimize(
       (max_size_increase == kEliminateAll ? kEliminateAll
                                           : mCnf.Size() + max_size_increase),
       min_steps, min_factor);
}

void
SUBool::CNFEncoding::EliminateAuxVariablesWithMaxInc(size_t max_size_increase)
{
   EliminateAuxVariablesWithMaxSize(
       (max_size_increase == kEliminateAll ? kEliminateAll
                                           : mCnf.Size() + max_size_increase));
}

void
SUBool::CNFEncoding::EliminateAuxVariablesWithFactor(
    long double max_size_increase_factor)
{
   EliminateAuxVariablesWithMaxInc(
       static_cast<size_t>(max_size_increase_factor * mCnf.Size())
       - mCnf.Size());
}

void
SUBool::CNFEncoding::EliminateAuxVariablesWithFactorMinimize(
    long double max_size_increase_factor, size_t min_steps, double min_factor)
{
   EliminateAuxVariablesWithMaxIncMinimize(
       static_cast<size_t>(max_size_increase_factor * mCnf.Size())
           - mCnf.Size(),
       min_steps, min_factor);
}

void
SUBool::CNFEncoding::Minimize(size_t log_step)
{
   CPUTime cpu_time;
   logs.TLog(3) << TypeDesc() << " encoding minimization starts (p cnf "
                << mCnf.MaxVariable() << ' ' << mCnf.Size() << ")" << std::endl;
   mCnf = empower_minimize(std::move(mCnf), MINIMIZE_METHOD::FULL, log_step);
   CnfChanged();
   logs.TLog(3) << TypeDesc() << " encoding minimization finished (p cnf "
                << mCnf.MaxVariable() << ' ' << mCnf.Size()
                << "), time=" << cpu_time << std::endl;
}

bool
SUBool::SATEncoding::Consistent(const Assignment &alpha) const
{
   if (alpha.Size() != mMaxInputVariable)
   {
      throw BadParameterException(
          "CNFEncoding::Consistent", "alpha.Size() != mMaxInputVariable");
   }
   if (mMaxInputVariable == mCnf.MaxVariable())
   {
      return mCnf.Value(alpha);
   }
   return mSolver.Solver().Solve(alpha);
}

bool
SUBool::SATEncoding::Consistent(const PartialAssignment &alpha) const
{
   return mSolver.Solver().Solve(alpha);
}

std::optional<SUBool::PartialAssignment>
SUBool::SATEncoding::Propagate(const PartialAssignment &alpha) const
{
   if (alpha.Size() != mMaxInputVariable)
   {
      throw BadParameterException(
          "CNFEncoding::Propagate", "alpha.Size() != mMaxInputVariable");
   }
   auto beta = mPropagator.PropagateMaxVar(alpha, mMaxInputVariable);
   if (beta)
   {
      beta->Resize(mMaxInputVariable, PartialValue::UNDEF);
   }
   return beta;
}

bool
SUBool::URCEncoding::Consistent(const Assignment &alpha) const
{
   return mUP.Consistent(literals_of_assignment(alpha));
}

bool
SUBool::URCEncoding::Consistent(const PartialAssignment &alpha) const
{
   return mUP.Consistent(literals_of_assignment(alpha));
}

std::optional<SUBool::PartialAssignment>
SUBool::URCEncoding::Propagate(const PartialAssignment &alpha) const
{
   auto gamma = mUP.Propagate(alpha);
   if (!gamma)
   {
      return {};
   }
   for (unsigned i = 1; i <= mMaxInputVariable; ++i)
   {
      if ((*gamma)[i] == PartialValue::UNDEF)
      {
         (*gamma)[i] = PartialValue::TRUE;
         auto beta = mUP.Propagate(*gamma);
         if (!beta)
         {
            (*gamma)[i] = PartialValue::FALSE;
            beta = mUP.Propagate(*gamma);
            if (!beta)
            {
               throw InvariantFailedException(
                   "URCEncoding::Propagate", "Unexpected conflict");
            }
            gamma = std::move(beta);
            continue;
         }
         (*gamma)[i] = PartialValue::FALSE;
         beta = mUP.Propagate(*gamma);
         if (!beta)
         {
            (*gamma)[i] = PartialValue::TRUE;
            gamma = std::move(beta);
            continue;
         }
         (*gamma)[i] = PartialValue::UNDEF;
      }
   }
   gamma->Resize(mMaxInputVariable, PartialValue::UNDEF);
   return gamma;
}

std::optional<SUBool::PartialAssignment>
SUBool::PCEncoding::Propagate(const PartialAssignment &alpha) const
{
   auto beta = mUP.Propagate(literals_of_assignment(alpha));
   if (!beta)
   {
      return {};
   }
   beta->Resize(mMaxInputVariable, PartialValue::UNDEF);
   return beta;
}

std::ostream &
SUBool::CNFEncoding::WriteToStream(std::ostream &output) const
{
   output << "c ! encoding " << TypeDesc() << " " << mMaxInputVariable
          << std::endl;
   output << mCnf;
   return output;
}

std::string
SUBool::SATEncoding::TypeDesc() const
{
   return "sat";
}

std::string
SUBool::URCEncoding::TypeDesc() const
{
   return (mFullyComplete ? "urc" : "i-urc");
}

std::string
SUBool::PCEncoding::TypeDesc() const
{
   return (mFullyComplete ? "pc" : "i-pc");
}

SUBool::CNFEncoding::KIND
SUBool::SATEncoding::Kind() const
{
   return KIND::SAT;
}

SUBool::CNFEncoding::KIND
SUBool::URCEncoding::Kind() const
{
   return (mFullyComplete ? KIND::URC : KIND::I_URC);
}

SUBool::CNFEncoding::KIND
SUBool::PCEncoding::Kind() const
{
   return (mFullyComplete ? KIND::PC : KIND::I_PC);
}

std::ostream &
SUBool::operator<<(std::ostream &output, const CNFEncoding &enc)
{
   return enc.WriteToStream(output);
}

SUBool::CNFEncoding::Header
SUBool::CNFEncoding::ReadHeader(std::istream &input)
{
   std::string line;
   while (getline(input, line))
   {
      std::smatch sm;
      if (std::regex_match(line, sm, kHeaderRegex))
      {
         return Header{KindOfName(sm[1]),
             static_cast<unsigned int>(std::atoi(sm[2].str().c_str()))};
      }
   }
   return Header{KIND::UNKNOWN, 0};
}

std::unique_ptr<SUBool::CNFEncoding>
SUBool::CNFEncoding::ReadFromStream(std::istream &input)
{
   auto header = ReadHeader(input);
   auto cnf = Dimacs::parse_cnf(input);
   switch (header.kind)
   {
   case KIND::SAT:
      return std::unique_ptr<CNFEncoding>(
          new SATEncoding(cnf, header.max_input_var));
   case KIND::I_URC:
      return std::unique_ptr<CNFEncoding>(
          new URCEncoding(cnf, header.max_input_var, false));
   case KIND::URC:
      return std::unique_ptr<CNFEncoding>(
          new URCEncoding(cnf, header.max_input_var, true));
   case KIND::I_PC:
      return std::unique_ptr<CNFEncoding>(
          new PCEncoding(cnf, header.max_input_var, false));
   case KIND::PC:
      return std::unique_ptr<CNFEncoding>(
          new PCEncoding(cnf, header.max_input_var, true));
   default:
      break;
   }
   return std::unique_ptr<CNFEncoding>();
}

