/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file encoding.h
 * CNF encodings. URC and PC encodings implementation.
 */

#ifndef __ENCODING_H
#define __ENCODING_H

#include <memory>
#include <regex>

#include "assign.h"
#include "cnfutil.h"
#include "cpuutil.h"
#include "dpelim.h"
#include "empower.h"
#include "enum.h"
#include "logstream.h"
#include "normalform.h"
#include "propagate.h"
#include "unitprop.h"

namespace SUBool
{

   class CNFEncoding
   {
    public:
      static const size_t kEliminateAll;
      constexpr static const double kMinStepsFactor = 1.2;
      static const size_t kMinimizeLogStep = 1000;

      enum class KIND
      {
         SAT,
         I_URC,
         URC,
         I_PC,
         PC,
         UNKNOWN,
         Begin = SAT,
         End = UNKNOWN
      };

      static const std::vector<std::string> kKindNames;
      static const std::regex kHeaderRegex;

      static KIND KindOfName(const std::string &kind_name);

      struct Header
      {
         KIND kind;
         unsigned max_input_var;
      };

      enum class BLOCK_PRIME
      {
         NONE,
         UP_PRIME,
         PRIME,
         Last = PRIME
      };

      static const EnumAnnotation<BLOCK_PRIME> kBlockPrimeAnnotation;

    private:
      DPEliminator::IGNORE_POLICY mElimIgnorePolicy;
      DPEliminator::INPROCESSING mElimInprocessing;
      DPEliminator::INCREASE_POLICY mElimIncreasePolicy;
      BLOCK_PRIME mBlockPrime;

      size_t SizeIncrease(
          size_t max_size, size_t min_steps, double min_factor) const;

    protected:
      CNF mCnf;
      unsigned mMaxInputVariable;

      virtual void CnfChanged();
      static Header ReadHeader(std::istream &input);

      template <class IgnorePred>
      CNF BlockPrimeUP(const CNF &cnf, unsigned cnf_max_var, IgnorePred pred);

      template <class IgnorePred>
      CNF BlockPrime(const CNF &cnf, unsigned cnf_max_var,
          const SolverInterface &ref_solver, IgnorePred pred);
      std::pair<CNF, std::vector<bool>> CallEliminate(
          size_t size_inc, bool at_least_one);

    public:
      virtual ~CNFEncoding();
      // The input variables have to be 1 to max_input_var,
      // all remaining variables are auxiliary.
      CNFEncoding(CNF cnf, unsigned max_input_var,
          DPEliminator::IGNORE_POLICY elim_ignore_policy,
          DPEliminator::INPROCESSING elim_inprocessing,
          DPEliminator::INCREASE_POLICY elim_increase_policy,
          BLOCK_PRIME block_prime)
          : mElimIgnorePolicy(elim_ignore_policy),
            mElimInprocessing(elim_inprocessing),
            mElimIncreasePolicy(elim_increase_policy), mBlockPrime(block_prime),
            mCnf(std::move(cnf)), mMaxInputVariable(max_input_var)
      {
      }
      // The input variables are 1 to MaxInputVariable()
      unsigned
      MaxInputVariable() const
      {
         return mMaxInputVariable;
      }
      // alpha on input variables
      virtual bool Consistent(const Assignment &alpha) const = 0;
      // alpha on input variables
      virtual bool Consistent(const PartialAssignment &alpha) const = 0;
      // Returns none if not consistent
      virtual std::optional<PartialAssignment> Propagate(
          const PartialAssignment &alpha) const = 0;

      void
      SetEliminationPolicy(DPEliminator::IGNORE_POLICY ignore_policy,
          DPEliminator::INPROCESSING inprocessing,
          DPEliminator::INCREASE_POLICY increase_policy,
          BLOCK_PRIME block_prime)
      {
         mElimIgnorePolicy = ignore_policy;
         mElimInprocessing = inprocessing;
         mElimIncreasePolicy = increase_policy;
         mBlockPrime = block_prime;
      }

      virtual std::string TypeDesc() const = 0;
      virtual KIND Kind() const = 0;
      virtual std::ostream &WriteToStream(std::ostream &output) const;

      virtual void Minimize(size_t log_step = 0);

      CNF
      Cnf() const &&
      {
         return std::move(mCnf);
      }

      const CNF &
      Cnf() const &
      {
         return mCnf;
      }

      size_t
      Size() const
      {
         return mCnf.Size();
      }
      virtual void EliminateAuxVariablesWithMaxInc(size_t max_size_increase);
      // If min_steps==0, then the maximum possible size increase is used when
      // calling DPEliminator.
      virtual void EliminateAuxVariablesWithMaxIncMinimize(
          size_t max_size_increase, size_t min_steps, double min_factor);

      virtual void EliminateAuxVariablesWithMaxSize(size_t max_size);
      // If min_steps==0, then the maximum possible size increase is used when
      // calling DPEliminator.
      virtual void EliminateAuxVariablesWithMaxSizeMinimize(
          size_t max_size, size_t min_steps, double min_factor);

      virtual void EliminateAuxVariablesWithFactor(
          long double max_size_increase_factor);
      // If min_steps==0, then the maximum possible size increase is used when
      // calling DPEliminator.
      virtual void EliminateAuxVariablesWithFactorMinimize(
          long double max_size_increase_factor, size_t min_steps,
          double min_factor);

      static std::unique_ptr<CNFEncoding> ReadFromStream(std::istream &input);
   };

   std::ostream &operator<<(std::ostream &output, const CNFEncoding &enc);

   class SATEncoding : public CNFEncoding
   {
      IncSolverWrapper mSolver;
      UnitPropagatorM mUP;
      MinimalSubimplicant mMinSubimplicant;
      Propagator mPropagator;

    protected:
      virtual void CnfChanged() override;

    public:
      SATEncoding(CNF cnf, unsigned max_input_var)
          : CNFEncoding(std::move(cnf), max_input_var,
              DPEliminator::IGNORE_POLICY::IMPLICATES,
              DPEliminator::INPROCESSING::NONE,
              DPEliminator::INCREASE_POLICY::ESTIMATE, BLOCK_PRIME::NONE),
            mSolver(mCnf), mUP(mCnf), mMinSubimplicant(&mCnf),
            mPropagator(mSolver.GetSolverPtr(), &mUP, &mMinSubimplicant)
      {
      }
      virtual ~SATEncoding();
      // alpha on input variables, using SAT
      virtual bool Consistent(const Assignment &alpha) const override;
      // alpha on input variables, using SAT
      virtual bool Consistent(const PartialAssignment &alpha) const override;
      // Using failed literals
      virtual std::optional<PartialAssignment> Propagate(
          const PartialAssignment &alpha) const override;
      virtual std::string TypeDesc() const override;
      virtual KIND Kind() const override;
   };

   class URCEncoding : public CNFEncoding
   {
    protected:
      mutable UnitPropagatorM mUP;
      bool mFullyComplete;

      virtual void CnfChanged() override;

    public:
      URCEncoding(CNF cnf, unsigned max_input_var, bool fully_complete)
          : CNFEncoding(std::move(cnf), max_input_var,
              DPEliminator::IGNORE_POLICY::ABSORBED,
              DPEliminator::INPROCESSING::NONE,
              DPEliminator::INCREASE_POLICY::ESTIMATE, BLOCK_PRIME::NONE),
            mUP(mCnf), mFullyComplete(fully_complete)
      {
      }
      virtual ~URCEncoding();
      // alpha on input variables, using unit propagation
      virtual bool Consistent(const Assignment &alpha) const override;
      // alpha on input variables, using unit propagation
      virtual bool Consistent(const PartialAssignment &alpha) const override;
      // Using failed literals
      virtual std::optional<PartialAssignment> Propagate(
          const PartialAssignment &alpha) const override;
      virtual std::string TypeDesc() const override;
      virtual KIND Kind() const override;
   };

   class PCEncoding : public URCEncoding
   {
    public:
      PCEncoding(CNF cnf, unsigned max_input_var, bool fully_complete)
          : URCEncoding(std::move(cnf), max_input_var, fully_complete)
      {
      }
      virtual ~PCEncoding();
      // Returns none if not consistent
      // Using unit propagation
      virtual std::optional<PartialAssignment> Propagate(
          const PartialAssignment &alpha) const override;
      virtual std::string TypeDesc() const override;
      virtual KIND Kind() const override;
   };
} // namespace SUBool

template <class IgnorePred>
SUBool::CNF
SUBool::CNFEncoding::BlockPrimeUP(
    const CNF &cnf, unsigned cnf_max_var, IgnorePred pred)
{
   CPUTime cpu_time;
   CNF p_cnf;
   logs.TLog(3) << TypeDesc() << " up block prime starts (p cnf " << cnf_max_var
                << ' ' << cnf.Size() << ")" << std::endl;
   p_cnf = up_prime_cnf_with_ignore(cnf, pred);
   logs.TLog(3) << TypeDesc() << " up block_prime finished (p cnf "
                << cnf_max_var << ' ' << p_cnf.Size() << "), time=" << cpu_time
                << std::endl;
   return p_cnf;
}

template <class IgnorePred>
SUBool::CNF
SUBool::CNFEncoding::BlockPrime(const CNF &cnf, unsigned cnf_max_var,
    const SolverInterface &ref_solver, IgnorePred pred)
{
   CPUTime cpu_time;
   CNF p_cnf;
   logs.TLog(3) << TypeDesc() << " block prime starts (p cnf " << cnf_max_var
                << ' ' << cnf.Size() << ")" << std::endl;
   p_cnf = prime_cnf_with_ignore(ref_solver, cnf, pred);
   logs.TLog(3) << TypeDesc() << " block_prime finished (p cnf " << cnf_max_var
                << ' ' << p_cnf.Size() << "), time=" << cpu_time << std::endl;
   return p_cnf;
}

#endif
