/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file enum.h
 * Utility functions and structures related to annotated enums.
 */

#ifndef __ENUM_H
#define __ENUM_H

#include <algorithm>

#include "logstream.h"
#include "sbexception.h"

namespace SUBool
{

   template <class T>
   [[deprecated("Use EnumAnnotation")]] T
   enum_of_int(unsigned val)
   {
      if (val > static_cast<unsigned>(T::Last))
      {
         throw BadParameterException("enum_of_int", "val too high");
      }
      return static_cast<T>(val);
   }

   template <class E, size_t S>
   [[deprecated("Use EnumAnnotation")]] E
   enum_of_name(const std::string &val, const std::array<std::string, S> &names)
   {
      auto i = std::find(names.begin(), names.end(), val);
      if (i == names.end())
      {
         throw BadParameterException("enum_of_name", "name is not on the list");
      }
      return enum_of_int<E>(i - names.begin());
   }

   template <class Enum, class Names>
   [[deprecated("Use EnumAnnotation")]] void
   dump_enum_value(const std::string &line_prefix, unsigned level,
       const std::string &name, const Enum &enum_value, const Names &names)
   {
      logs.TLog(level) << line_prefix << name << ": "
                       << names.at(static_cast<unsigned>(enum_value))
                       << std::endl;
   }

   template <class T> class EnumAnnotation
   {
    public:
      using enum_t = T;
      struct Value
      {
         T value;
         std::string name;
         std::string desc;
      };
      using ValueList = std::vector<Value>;
      using iterator = typename ValueList::iterator;
      using const_iterator = typename ValueList::const_iterator;

    private:
      ValueList mValues;
      std::string mName;

      typename ValueList::const_iterator FindByName(
          const std::string &name) const;
      typename ValueList::const_iterator FindByEnumValue(
          const T &enum_val) const;

    public:
      template <class InputIterator>
      EnumAnnotation(
          std::string enum_name, InputIterator begin, InputIterator end);
      EnumAnnotation(std::string enum_name, ValueList va);
      EnumAnnotation(std::string enum_name, std::initializer_list<Value> va);

      T Parse(const std::string &name) const;
      const std::string &Name(const T &value) const;
      void DumpValue(const std::string &line_prefix, unsigned level,
          const std::string &label, const T &enum_value) const;

      iterator begin();
      iterator end();
      const_iterator begin() const;
      const_iterator end() const;
      const_iterator cbegin() const;
      const_iterator cend() const;
   };

} // namespace SUBool

// Template and inline functions implementation

template <class T>
template <class InputIterator>
SUBool::EnumAnnotation<T>::EnumAnnotation(
    std::string enum_name, InputIterator begin, InputIterator end)
    : mValues(begin, end), mName(std::move(enum_name))
{
}

template <class T>
SUBool::EnumAnnotation<T>::EnumAnnotation(
    std::string enum_name, std::vector<Value> va)
    : mValues(std::move(va)), mName(std::move(enum_name))
{
}

template <class T>
SUBool::EnumAnnotation<T>::EnumAnnotation(
    std::string enum_name, std::initializer_list<Value> va)
    : mValues(std::move(va)), mName(std::move(enum_name))
{
}

template <class T>
auto
SUBool::EnumAnnotation<T>::Parse(const std::string &name) const -> T
{
   auto i_val = FindByName(name);
   if (i_val == mValues.end())
   {
      std::string err = "Failed to find value " + name + " in enum " + mName;
      throw ParseException("EnumAnnotation::Parse", err);
   }
   return i_val->value;
}

template <class T>
const std::string &
SUBool::EnumAnnotation<T>::Name(const T &value) const
{
   auto i_name = FindByEnumValue(value);
   if (i_name == mValues.end())
   {
      std::string err = "Failed to find name for a value in enum " + mName;
      throw ParseException("EnumAnnotation::Parse", err);
   }
   return i_name->name;
}

template <class T>
auto
SUBool::EnumAnnotation<T>::FindByName(const std::string &name) const ->
    typename std::vector<Value>::const_iterator
{
   return std::find_if(mValues.begin(), mValues.end(),
       [&name](const auto &val) { return name == val.name; });
}

template <class T>
auto
SUBool::EnumAnnotation<T>::FindByEnumValue(const T &enum_val) const ->
    typename std::vector<Value>::const_iterator
{
   return std::find_if(mValues.begin(), mValues.end(),
       [&enum_val](const auto &val) { return enum_val == val.value; });
}

template <class T>
void
SUBool::EnumAnnotation<T>::DumpValue(const std::string &line_prefix,
    unsigned level, const std::string &label, const T &enum_value) const
{
   logs.DumpValue(line_prefix, level, label, Name(enum_value));
}

template <class T>
auto
SUBool::EnumAnnotation<T>::begin() -> iterator
{
   return mValues.begin();
}

template <class T>
auto
SUBool::EnumAnnotation<T>::end() -> iterator
{
   return mValues.end();
}

template <class T>
auto
SUBool::EnumAnnotation<T>::begin() const -> const_iterator
{
   return mValues.begin();
}

template <class T>
auto
SUBool::EnumAnnotation<T>::end() const -> const_iterator
{
   return mValues.end();
}

template <class T>
auto
SUBool::EnumAnnotation<T>::cbegin() const -> const_iterator
{
   return mValues.begin();
}

template <class T>
auto
SUBool::EnumAnnotation<T>::cend() const -> const_iterator
{
   return mValues.end();
}

#endif
