/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file enum_opt.h
 * Program option for selecting an enum value.
 */

#ifndef __ENUM_OPT_H
#define __ENUM_OPT_H

#include <boost/program_options.hpp>
#include <boost/smart_ptr/make_shared.hpp>

#include "enum.h"
#include "sbexception.h"

namespace po = boost::program_options;

namespace SUBool
{

   template <class T> class EnumOption
   {
    public:
      using enum_t = T;

    private:
      const EnumAnnotation<T> &mEnumAnnotation;
      std::string mOptionName;
      std::string mOptionHelp;
      T mDefaultValue;

      std::string mOptionValue{};
      boost::shared_ptr<po::option_description> mOptionDesc{};

    public:
      EnumOption(const EnumAnnotation<T> &enum_annotation, std::string opt_name,
          std::string opt_help, const T &default_value);

      boost::shared_ptr<po::option_description> Desc();

      operator boost::shared_ptr<po::option_description>();

      T Value() const;
      bool GetValue(T &val) const;
   };

   template <class T, typename... Args>
   EnumOption<T> make_enum_option(const EnumAnnotation<T> &ann, Args &&...arg);
} // namespace SUBool

// Template and inline functions implementation
//
template <class T>
SUBool::EnumOption<T>::EnumOption(const EnumAnnotation<T> &enum_annotation,
    std::string opt_name, std::string opt_help, const T &default_value)
    : mEnumAnnotation(enum_annotation), mOptionName(std::move(opt_name)),
      mOptionHelp(std::move(opt_help)), mDefaultValue(default_value)
{
}

template <class T>
boost::shared_ptr<po::option_description>
SUBool::EnumOption<T>::Desc()
{
   if (mOptionDesc)
   {
      return mOptionDesc;
   }

   std::string help = mOptionHelp + " Possible values:";
   for (const auto &val : mEnumAnnotation)
   {
      help += '\n';
      help += val.name + " = " + val.desc;
   }
   return mOptionDesc =
              boost::make_shared<po::option_description>(mOptionName.c_str(),
                  po::value<std::string>(&mOptionValue)
                      ->default_value(mEnumAnnotation.Name(mDefaultValue)),
                  help.c_str());
}

template <class T>
SUBool::EnumOption<T>::operator boost::shared_ptr<po::option_description>()
{
   return Desc();
}

template <class T>
T
SUBool::EnumOption<T>::Value() const
{
   return mEnumAnnotation.Parse(mOptionValue);
}

template <class T>
bool
SUBool::EnumOption<T>::GetValue(T &val) const
{
   try
   {
      val = mEnumAnnotation.Parse(mOptionValue);
   }
   catch (const ParseException &e)
   {
      logs.Log(0) << "Unknown value \"" << mOptionValue << "\" of option \""
                  << mOptionName << '"' << std::endl;
      return false;
   }
   return true;
}

template <class T, typename... Args>
auto
SUBool::make_enum_option(const EnumAnnotation<T> &ann, Args &&...args)
    -> EnumOption<T>
{
   return EnumOption<T>(ann, std::forward<Args>(args)...);
}

#endif
