/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file glucose_bind.h
 * Binding class to glucose solver library. */

#ifndef __GLUCOSE_BIND_H
#define __GLUCOSE_BIND_H

#include "random_utils.h"
#include "solver_bind.h"
#include <simp/SimpSolver.h>

namespace SUBool
{

   template <class Solver> class GlucoseSolver : public Solver
   {
    public:
      GlucoseSolver(const SolverInterface::ADAPTATION adapt =
                        SolverInterface::ADAPTATION::DEFAULT)
          : Solver()
      {
         this->random_seed = static_cast<double>(sub_rand() % 2147483647);
         Solver::verbosity = -1;
         Adapt(adapt);
      }

      GlucoseSolver(const GlucoseSolver &s) = default;
      GlucoseSolver(GlucoseSolver &&s) = default;
      GlucoseSolver &operator=(const GlucoseSolver &s) = delete;
      GlucoseSolver &operator=(GlucoseSolver &&s) = delete;

      virtual Glucose::Clone *
      clone() const
      {
         return new GlucoseSolver(*this);
      }

      unsigned
      LearntsSize() const
      {
         return this->learnts.size() + this->permanentLearnts.size();
      }

      const Glucose::Clause &
      LearnedClause(unsigned i) const
      {
         return this
             ->ca[i < static_cast<unsigned>(this->permanentLearnts.size())
                      ? this->permanentLearnts[i]
                      : this->learnts[i - this->permanentLearnts.size()]];
      }

      void
      Adapt(const SolverInterface::ADAPTATION adapt)
      {
         if (adapt == SolverInterface::ADAPTATION::LIKELY_SAT)
         {
            this->luby_restart = true;
            this->luby_restart_factor = 100;
            this->rnd_init_act = true;
            this->random_var_freq = 0.05;
            this->rnd_pol = true;
         }
         else
         {
            this->luby_restart = false;
            this->luby_restart_factor = 100;
            this->rnd_init_act = false;
            this->random_var_freq = 0.0;
            this->rnd_pol = false;
         }
      }
   };

   class GlucoseSolverInc : public GlucoseSolver<Glucose::Solver>
   {
    public:
      using Base = GlucoseSolver<Glucose::Solver>;
      GlucoseSolverInc(const SolverInterface::ADAPTATION adapt =
                           SolverInterface::ADAPTATION::DEFAULT)
          : Base(adapt)
      {
         setIncrementalMode();
      }
   };

   class GlucoseSimpSolver : public GlucoseSolver<Glucose::SimpSolver>
   {
    public:
      using Base = GlucoseSolver<Glucose::SimpSolver>;
      GlucoseSimpSolver(const SolverInterface::ADAPTATION adapt =
                            SolverInterface::ADAPTATION::DEFAULT)
          : Base(adapt)
      {
         parsing = 0;
      }
   };

   template <class Solver> class GlucoseBind : public SolverInterface
   {
    private:
      std::unique_ptr<Solver> mSolver;
      mutable Glucose::vec<Glucose::Lit> mLitBuffer;
      mutable PartialValue mResult;
      ADAPTATION mAdapt;

      bool SolveLitBuffer() const;

    public:
      GlucoseBind(const SolverInterface::ADAPTATION adapt =
                      SolverInterface::ADAPTATION::DEFAULT);
      GlucoseBind(const CNF &cnf, const SolverInterface::ADAPTATION adapt =
                                      SolverInterface::ADAPTATION::DEFAULT)
          : GlucoseBind(adapt)
      {
         AddCnf(cnf);
      }

      virtual bool Reset(const CNF &cnf) override;
      virtual bool AddCnf(const CNF &cnf) override;
      virtual bool AddClause(const Clause &clause) override;

      virtual void Interrupt() const override;

      virtual bool Solve(const std::vector<Literal> &assump) const override;
      virtual bool Solve() const override;
      virtual bool Solve(const PartialAssignment &assump) const override;
      virtual bool Solve(const Assignment &assump) const override;

      /* Throws NoModelException if no model is currently
       * available. */
      virtual PartialValue ValueOfVariable(unsigned var) const override;
      virtual PartialValue ValueOfLiteral(const Literal &lit) const override;
      virtual PartialAssignment Model() const override;
      virtual Clause ConflictClause() const override;

      virtual size_t LearntsSize() const override;
      virtual std::vector<Clause> Learnts() const override;

      virtual bool IsIncremental() const override;
   };

   using GlucoseBindSimp = GlucoseBind<GlucoseSimpSolver>;
   using GlucoseBindInc = GlucoseBind<GlucoseSolverInc>;

   Literal of_glucose_lit(const Glucose::Lit &glit);
   Clause of_glucose_clause(const Glucose::Clause &gclause);
   Glucose::Lit glucose_lit(const Literal &lit);
   PartialValue of_glucose_value(Glucose::lbool val);

   using SimpSolverWrapper = SolverWrapperTemplate<GlucoseBindSimp>;
   using IncSolverWrapper = SolverWrapperTemplate<GlucoseBindInc>;
} // namespace SUBool

template <class Solver>
SUBool::GlucoseBind<Solver>::GlucoseBind(
    const SolverInterface::ADAPTATION adapt)
    : mSolver(), mLitBuffer(), mResult(PartialValue::UNDEF), mAdapt(adapt)
{
   Reset(CNF());
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::Reset(const CNF &cnf)
{
   mSolver = std::make_unique<Solver>(mAdapt);
   mResult = PartialValue::UNDEF;
   return AddCnf(cnf);
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::AddCnf(const CNF &cnf)
{
   for (const auto &clause : cnf)
   {
      if (!AddClause(clause))
      {
         return false;
      }
   }
   return true;
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::AddClause(const Clause &clause)
{
   mLitBuffer.clear();
   mResult = PartialValue::UNDEF;
   for (const auto &lit : clause)
   {
      while (static_cast<int>(lit.Variable()) > mSolver->nVars())
      {
         mSolver->newVar();
      }
      mLitBuffer.push(glucose_lit(lit));
   }
   if (!mSolver->addClause(mLitBuffer))
   {
      mResult = PartialValue::FALSE;
   }
   return (mResult != PartialValue::FALSE);
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::Solve() const
{
   mLitBuffer.clear();
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::Solve(const std::vector<Literal> &assump) const
{
   mLitBuffer.clear();
   std::for_each(assump.begin(), assump.end(), [this](const auto &lit) {
      mLitBuffer.push(glucose_lit(lit));
   });
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::Solve(const PartialAssignment &assump) const
{
   mLitBuffer.clear();
   for (unsigned i = 1; i <= assump.Size(); ++i)
   {
      if (assump[i] != PartialValue::UNDEF)
      {
         mLitBuffer.push(
             glucose_lit(Literal(i, assump[i] == PartialValue::TRUE)));
      }
   }
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::Solve(const Assignment &assump) const
{
   mLitBuffer.clear();
   for (unsigned i = 1; i <= assump.Size(); ++i)
   {
      mLitBuffer.push(glucose_lit(Literal(i, assump[i])));
   }
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::SolveLitBuffer() const
{
   auto ret = mSolver->solveLimited(mLitBuffer);
   mResult = of_glucose_value(ret);
   IncreaseSolveCount(mResult);
   if (mResult == PartialValue::UNDEF)
   {
      throw SolverTerminatedException("GlucoseBind::Solve",
                                      "Solver finished without result");
   }
   return bool_of_partial_value(mResult);
}

template <class Solver>
void
SUBool::GlucoseBind<Solver>::Interrupt() const
{
   mSolver->interrupt();
}

template <class Solver>
SUBool::PartialValue
SUBool::GlucoseBind<Solver>::ValueOfVariable(unsigned var) const
{
   if (!bool_of_partial_value(mResult, false))
   {
      throw NoModelException("GlucoseBind::ValueOfVariable",
                             "No model available");
   }
   return of_glucose_value(mSolver->modelValue(var - 1));
}

template <class Solver>
SUBool::PartialValue
SUBool::GlucoseBind<Solver>::ValueOfLiteral(const Literal &lit) const
{
   if (!bool_of_partial_value(mResult, false))
   {
      throw NoModelException("GlucoseBind::ValueOfLiteral",
                             "No model available");
   }
   return of_glucose_value(mSolver->modelValue(glucose_lit(lit)));
}

template <class Solver>
SUBool::PartialAssignment
SUBool::GlucoseBind<Solver>::Model() const
{
   if (!bool_of_partial_value(mResult, false))
   {
      throw NoModelException("GlucoseBind::Model", "No model available");
   }
   PartialAssignment alpha(mSolver->nVars());
   for (int i = 1; i <= mSolver->nVars(); ++i)
   {
      alpha[i] = ValueOfVariable(i);
   }
   return alpha;
}

template <class Solver>
SUBool::Clause
SUBool::GlucoseBind<Solver>::ConflictClause() const
{
   std::vector<Literal> conflict_literals;
   conflict_literals.reserve(mSolver->conflict.size());
   for (unsigned i = 0; i < static_cast<unsigned>(mSolver->conflict.size());
        ++i)
   {
      conflict_literals.push_back(of_glucose_lit(mSolver->conflict[i]));
   }
   return conflict_literals;
}

template <class Solver>
std::vector<SUBool::Clause>
SUBool::GlucoseBind<Solver>::Learnts() const
{
   std::vector<SUBool::Clause> learnts;
   learnts.reserve(mSolver->LearntsSize());
   for (unsigned i = 0; i < mSolver->LearntsSize(); ++i)
   {
      auto c = of_glucose_clause(mSolver->LearnedClause(i));
      learnts.push_back(of_glucose_clause(mSolver->LearnedClause(i)));
   }
   return learnts;
}

template <class Solver>
size_t
SUBool::GlucoseBind<Solver>::LearntsSize() const
{
   return mSolver->LearntsSize();
}

template <class Solver>
bool
SUBool::GlucoseBind<Solver>::IsIncremental() const
{
   return mSolver->isIncremental();
}

#endif
