/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file horn.cpp
 * Functions related to Horn functions.
 */

#include "horn.h"

bool
SUBool::IsDefiniteHornClause(const Clause &clause)
{
   return (clause.CountVariablePolarities().first == 1);
}

bool
SUBool::IsDefiniteHornTerm(const Term &term)
{
   return (term.CountVariablePolarities().second == 1);
}

bool
SUBool::IsHornClause(const Clause &clause)
{
   return (clause.CountVariablePolarities().first <= 1);
}

bool
SUBool::IsHornTerm(const Term &term)
{
   return (term.CountVariablePolarities().second <= 1);
}

bool
SUBool::IsAntiHornClause(const Clause &clause)
{
   return (clause.CountVariablePolarities().second <= 1);
}

bool
SUBool::IsAntiHornTerm(const Term &term)
{
   return (term.CountVariablePolarities().first <= 1);
}

bool
SUBool::IsHornCNF(const CNF &cnf)
{
   for (const auto &clause : cnf)
   {
      if (!IsHornClause(clause))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::IsHornDNF(const DNF &dnf)
{
   for (const auto &term : dnf)
   {
      if (!IsHornTerm(term))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::IsDefiniteHornCNF(const CNF &cnf)
{
   for (const auto &clause : cnf)
   {
      if (!IsDefiniteHornClause(clause))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::IsDefiniteHornDNF(const DNF &dnf)
{
   for (const auto &term : dnf)
   {
      if (!IsDefiniteHornTerm(term))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::IsAntiHornCNF(const CNF &cnf)
{
   for (const auto &clause : cnf)
   {
      if (!IsAntiHornClause(clause))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::IsAntiHornDNF(const DNF &dnf)
{
   for (const auto &term : dnf)
   {
      if (!IsAntiHornTerm(term))
      {
         return false;
      }
   }
   return true;
}

SUBool::VarImpl
SUBool::horn_clause_to_impl(const Clause &clause)
{
   std::vector<unsigned> body;
   body.reserve(clause.Size());
   unsigned head = 0;
   for (const auto &lit : clause)
   {
      if (lit.Positive())
      {
         if (head > 0)
         {
            throw BadParameterException("horn_clause_to_implication",
                                        "Not a horn clause");
         }
         head = lit.Variable();
      }
      else
      {
         body.push_back(lit.Variable());
      }
   }
   return head > 0 ? VarImpl(body, {head}, true) : VarImpl(body, {}, true);
}

SUBool::VarImplSystem
SUBool::horn_cnf_to_implsys(const CNF &cnf, bool merge_bodies)
{
   VarImplSystem is;
   for (const auto &c : cnf)
   {
      is.PushBack(horn_clause_to_impl(c));
   }
   if (merge_bodies)
   {
      is.MergeBodies();
   }
   return is;
}

std::vector<SUBool::Clause>
SUBool::impl_to_horn_clauses(const VarImpl &imp)
{
   std::vector<Literal> literals;
   literals.reserve(imp.Body().size() + 1);
   for (auto &var : imp.Body())
   {
      literals.emplace_back(var, false);
   }
   if (imp.Cons().empty())
   {
      return {Clause(literals)};
   }
   std::vector<Clause> clauses;
   for (const auto &var : imp.Cons())
   {
      literals.push_back(Literal(var, true));
      clauses.emplace_back(literals);
      literals.pop_back();
   }
   return clauses;
}

SUBool::CNF
SUBool::implsys_to_horn_cnf(const VarImplSystem &is)
{
   CNF cnf;
   for (const auto &imp : is)
   {
      cnf.Append(impl_to_horn_clauses(imp));
   }
   return cnf;
}

SUBool::CNF
SUBool::horn_gd_basis(const CNF &cnf)
{
   auto is = horn_cnf_to_implsys(cnf, true);
   is.MakeGDBasis();
   return implsys_to_horn_cnf(is);
}
