/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file impl.h
 * Representation of an implication
 *
 * An implication is represented as a pair of sets of variables, one set
 * represents the body (or antecedent) and the other set represents
 * consequences.
 */

#ifndef __IMPL_H
#define __IMPL_H

#include <iostream>
#include <optional>
#include <set>

#include "literal.h"
#include "types.h"
#include "util.h"

namespace SUBool
{

   template <class Var> class Impl
   {
      std::vector<Var> mBody; // Assumptions/antecedent
      std::vector<Var> mCons; // Consequences

    public:
      Impl() : mBody(), mCons() {}
      Impl(std::vector<Var> body, std::vector<Var> cons, bool sorted);

      void UpdateBody(std::vector<Var> body, bool sorted);
      void UpdateCons(std::vector<Var> cons, bool sorted);
      void SubtractFromBody(std::vector<Var> vars);
      void SubtractFromCons(std::vector<Var> vars);

      std::ostream &Print(std::ostream &os) const;

      // Printer should convert a variable to a printable value
      // It is used as
      // os<<printer(v) where v is a variable.
      template <class VarPrinter>
      std::ostream &PrettyPrint(std::ostream &os, VarPrinter printer,
                                const std::string &sep = ":") const;

      const std::vector<Var> &Body() const &;
      std::vector<Var> Body() &&;
      const std::vector<Var> &Cons() const &;
      std::vector<Var> Cons() &&;
      std::vector<Var> BodyWithCons() const;

      bool EmptyBody() const;
      bool EmptyCons() const;
      size_t BodySize() const;
      size_t ConsSize() const;
      size_t Length() const;

      bool SmallerBodyThan(const Impl &a) const;
      bool EqualBody(const Impl &a) const;
      bool Trivial() const; // body==closure or closure is empty
      SetCompareResult CompareBody(const Impl &a) const;
      bool SubsumesBody(const Impl &a) const;

      void MergeCons(const Impl &a);
      void MakeConsWithoutBody();
      void AddBodyToCons();
      void AddToBody(const std::vector<Var> &body_add);
      void AddToCons(const std::vector<Var> &cons_add);

      bool operator==(const Impl &a) const;
      bool operator!=(const Impl &a) const;
   };

   template <class Var, class InputIterator, class VarPrinter>
   std::ostream &pretty_print_impl(std::ostream &os, InputIterator begin,
                                   InputIterator end, VarPrinter printer,
                                   const std::string &sep = ":");

   template <class Var> struct ImplBodyCompare
   {
      bool
      operator()(const Impl<Var> &a, const Impl<Var> &b) const
      {
         return a.SmallerBodyThan(b);
      }
   };

   template <class Var> struct ImplBodySizeCompare
   {
      bool
      operator()(const Impl<Var> &a, const Impl<Var> &b) const
      {
         return a.BodySize() < b.BodySize();
      }
   };

   template <class Var> struct ImplEqualBody
   {
      bool
      operator()(const Impl<Var> &a, const Impl<Var> &b) const
      {
         return a.EqualBody(b);
      }
   };

   template <class Var>
   std::ostream &
   operator<<(std::ostream &out, const Impl<Var> &impl)
   {
      return impl.Print(out);
   }

   using VarImpl = Impl<unsigned>;
   using LitImpl = Impl<Literal>;
} // namespace SUBool

// Implementation
template <class Var>
SUBool::Impl<Var>::Impl(std::vector<Var> body, std::vector<Var> cons,
                        bool sorted)
    : mBody(std::move(body)), mCons(std::move(cons))
{
   if (!sorted)
   {
      sort_unique(mBody);
      sort_unique(mCons);
   }
}

template <class Var>
void
SUBool::Impl<Var>::UpdateBody(std::vector<Var> body, bool sorted)
{
   mBody = std::move(body);
   if (!sorted)
   {
      sort_unique(mBody);
   }
}

template <class Var>
void
SUBool::Impl<Var>::UpdateCons(std::vector<Var> cons, bool sorted)
{
   mCons = std::move(cons);
   if (!sorted)
   {
      sort_unique(mCons);
   }
}

template <class Var>
std::ostream &
SUBool::Impl<Var>::Print(std::ostream &os) const
{
   return PrettyPrint(os, [](const Var &v) { return v; });
}

template <class Var>
template <class VarPrinter>
std::ostream &
SUBool::Impl<Var>::PrettyPrint(std::ostream &os, VarPrinter printer,
                               const std::string &sep) const
{
   for (const auto &v : mBody)
   {
      os << printer(v) << " ";
   }
   os << sep;
   for (const auto &v : mCons)
   {
      os << " " << printer(v);
   }
   return os;
}

template <class Var>
size_t
SUBool::Impl<Var>::Length() const
{
   return mBody.size() + mCons.size();
}

template <class Var>
bool
SUBool::Impl<Var>::EmptyBody() const
{
   return mBody.empty();
}

template <class Var>
bool
SUBool::Impl<Var>::EmptyCons() const
{
   return mCons.empty();
}

template <class Var>
size_t
SUBool::Impl<Var>::BodySize() const
{
   return mBody.size();
}

template <class Var>
size_t
SUBool::Impl<Var>::ConsSize() const
{
   return mCons.size();
}

template <class Var>
bool
SUBool::Impl<Var>::SmallerBodyThan(const Impl<Var> &a) const
{
   return mBody.size() < a.mBody.size()
          || (mBody.size() == a.mBody.size() && mBody < a.mBody);
}

template <class Var>
bool
SUBool::Impl<Var>::EqualBody(const Impl<Var> &a) const
{
   return mBody == a.mBody;
}

template <class Var>
void
SUBool::Impl<Var>::MergeCons(const Impl<Var> &a)
{
   std::vector<Var> merged;
   std::merge(mCons.begin(), mCons.end(), a.mCons.begin(), a.mCons.end(),
              std::back_inserter(merged));
   auto end = std::unique(merged.begin(), merged.end());
   merged.resize(end - merged.begin());
   mCons = std::move(merged);
}

template <class Var>
bool
SUBool::Impl<Var>::operator==(const Impl<Var> &a) const
{
   return mBody.size() == a.mBody.size() && mCons.size() == a.mCons.size()
          && mBody == a.mBody && mCons == a.mCons;
}

template <class Var>
bool
SUBool::Impl<Var>::operator!=(const Impl<Var> &a) const
{
   return !(*this == a);
}

template <class Var>
const std::vector<Var> &
SUBool::Impl<Var>::Body() const &
{
   return mBody;
}

template <class Var>
std::vector<Var>
SUBool::Impl<Var>::Body() &&
{
   return std::move(mBody);
}

template <class Var>
const std::vector<Var> &
SUBool::Impl<Var>::Cons() const &
{
   return mCons;
}

template <class Var>
typename std::vector<Var>
SUBool::Impl<Var>::Cons() &&
{
   return std::move(mCons);
}

template <class Var>
typename std::vector<Var>
SUBool::Impl<Var>::BodyWithCons() const
{
   return vector_union(mBody, mCons);
}

template <class Var>
bool
SUBool::Impl<Var>::Trivial() const
{
   return mCons.empty() || mBody == mCons;
}

template <class Var>
bool
SUBool::Impl<Var>::SubsumesBody(const Impl &a) const
{
   return std::includes(a.mBody.begin(), a.mBody.end(), mBody.begin(),
                        mBody.end());
}

template <class Var>
SUBool::SetCompareResult
SUBool::Impl<Var>::CompareBody(const Impl &a) const
{
   return compare_sets(mBody, a.mBody);
}

template <class Var>
void
SUBool::Impl<Var>::MakeConsWithoutBody()
{
   mCons = vector_subtract(mCons, mBody);
}

template <class Var>
void
SUBool::Impl<Var>::AddBodyToCons()
{
   mCons = vector_union(mCons, mBody);
}

template <class Var>
void
SUBool::Impl<Var>::AddToBody(const std::vector<Var> &body_add)
{
   mBody = vector_union(mBody, body_add);
}

template <class Var>
void
SUBool::Impl<Var>::AddToCons(const std::vector<Var> &cons_add)
{
   mCons = vector_union(mCons, cons_add);
}

template <class Var>
void
SUBool::Impl<Var>::SubtractFromBody(std::vector<Var> vars)
{
   mBody = vector_subtract(mBody, vars);
}

template <class Var>
void
SUBool::Impl<Var>::SubtractFromCons(std::vector<Var> vars)
{
   mCons = vector_subtract(mCons, vars);
}

template <class Var, class InputIterator, class VarPrinter>
std::ostream &
SUBool::pretty_print_impl(std::ostream &os, InputIterator begin,
                          InputIterator end, VarPrinter printer,
                          const std::string &sep)
{
   for (auto i_impl = begin; i_impl != end; ++i_impl)
   {
      i_impl->PrettyPrint(os, printer, sep);
      if (std::next(i_impl) != end)
      {
         os << std::endl;
      }
   }
   return os;
}

#endif
