/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file impl_system.h
 * Representation of a system of implications
 *
 * An implication system is a set of implications. The functions
 * and structures defined here allow manipulation with implications and
 * implication system (making it irredundant, computing closures, GD basis,
 * saturation, etc.)
 */

#ifndef __IMPL_SYSTEM_H
#define __IMPL_SYSTEM_H

#include <algorithm>
#include <iostream>
#include <list>
#include <map>
#include <queue>
#include <set>
#include <unordered_map>
#include <unordered_set>
#include <vector>

#include "impl.h"
#include "literal.h"
#include "logstream.h"
#include "util.h"

// TODO: Make adding/removing/replacing(refining) implications incremental such
// as in UnitPropagator

namespace SUBool
{

   template <class Var> class ImplSystem
   {
    public:
      using ImplType = Impl<Var>;
      using impl_list_t = std::vector<ImplType>;
      using const_iterator = typename impl_list_t::const_iterator;
      using closure_to_impl_t =
          std::map<std::vector<Var>, std::unordered_set<size_t>>;

      struct Watch
      {
         size_t impl_index{0};
         typename std::vector<Var>::const_iterator watched_var{};
      };

      using watch_ptr = typename std::list<Watch>::iterator;

      struct ClosureData
      {
         std::unordered_map<Var, std::list<Watch>> mWatched;
         std::unordered_set<size_t> mImplWithEmptyBody;
         std::vector<watch_ptr> mImplWatch;
         std::unordered_set<Var> mClosureBuffer;

         void AddImplication(const ImplType &impl, size_t i);
         void Clear(size_t reserve_impl_size);
         void SetWatch(const Watch &watch);
         void ClearWatch(size_t impl_index);
         void RemoveImplication(const ImplType &impl, size_t i);
         ClosureData()
             : mWatched(), mImplWithEmptyBody(), mImplWatch(), mClosureBuffer()
         {
         }
      };

    private:
      impl_list_t mImpl;
      mutable ClosureData mClosureData;

      void BuildClosureData() const;
      std::optional<Watch> UpdateWatch(const Watch &watch) const;

      template <class IgnorePredicate>
      void BufferClosure(bool include_empty_bodies,
                         IgnorePredicate ignore_pred) const;
      template <class IgnorePredicate>
      void ClosureIntoBuffer(const std::vector<Var> &assump,
                             bool include_empty_bodies,
                             IgnorePredicate ignore_pred) const;
      void EmptyBodiesIntoBuffer(std::queue<Var> &q) const;
      template <class IgnorePredicate>
      std::vector<Var>
      Closure(const std::vector<Var> &assump, bool include_assumptions,
              bool include_empty_bodies, IgnorePredicate ignore_pred) const;
      void RemoveDuplicateBodies();
      closure_to_impl_t ClosureToImplications();
      void MakeLeftSaturated(const closure_to_impl_t &closure_to_impl);
      closure_to_impl_t RemoveSubsumptions(closure_to_impl_t closure_to_impl);
      void
      UpdateImplWithClosureToImpl(const closure_to_impl_t &closure_to_impl);

      void RemoveImplications(const std::vector<bool> &remove_mask);
      void MergeBodiesNoRebuild(); // Also sorts them
      void RemoveTrivialNoRebuild();
      void PropagateEmptyBodyNoRebuild();
      void RemoveBodyClosureFromCons(size_t impl_index);
      void MakeRightIrredundantImpl(size_t impl_index);
      void MakeLeftIrredundantImpl(size_t impl_index);

    public:
      ImplSystem();
      template <class InputIterator>
      ImplSystem(InputIterator begin, InputIterator end);
      ImplSystem(std::initializer_list<ImplType> imp);
      ImplSystem(std::vector<ImplType> imp);
      ImplSystem(const ImplSystem &is);
      ImplSystem(ImplSystem &&is) noexcept;
      ~ImplSystem();

      ImplSystem &operator=(ImplSystem &&is) noexcept;
      ImplSystem &operator=(const ImplSystem &is);

      std::vector<Var> Closure(const std::vector<Var> &assump,
                               bool include_empty_bodies = true) const;
      std::vector<Var>
      ClosureWithoutAssumptions(const std::vector<Var> &assump,
                                bool include_empty_bodies = true) const;

      void PushBack(ImplType impl);
      template <class InputIterator>
      void Append(InputIterator begin, InputIterator end);
      void Refine(size_t i, ImplType impl);

      size_t Size() const;
      size_t Length() const; // sum of the lengths of bodies plus the lengths of
                             // consequences

      const ImplType &At(size_t index) const;

      std::vector<ImplType>
      Implications() const &
      {
         return mImpl;
      }
      std::vector<ImplType>
      Implementation() &&
      {
         return std::move(mImpl);
      }

      void MergeBodies(); // Also sorts them
      void RemoveTrivial();
      // Removes the closure of the empty set from the bodies and consequences
      // of all other implications. The empty body is unique at index 0 after
      // propagation
      void PropagateEmptyBody();
      void MakeBodyIrredundant();
      void MakeRightIrredundant();
      void MakeLeftIrredundant();
      void MakeRightSaturated(); // Right saturation without body in cons
      void MakeRightLeftSaturated();

      // Right saturated+BodyIrredundant
      void MakeBodyMinimal();
      // Right saturated+Left saturated+body irredundant
      void MakeGDBasis();
      void MakeConsWithoutBody();
      void AddBodyToCons();
      void AddEmptyClosureToImpl(bool add_to_body, bool add_to_cons);

      std::ostream &Print(std::ostream &os) const;

      // Printer should convert a variable to a printable value
      // It is used as
      // os<<printer(v) where v is a variable.
      template <class VarPrinter>
      std::ostream &PrettyPrint(std::ostream &os, VarPrinter printer,
                                const std::string &sep = ":") const;

      const ImplType &operator[](size_t index) const;
      const_iterator begin() const;
      const_iterator end() const;
      const_iterator cbegin() const;
      const_iterator cend() const;

      bool operator==(const ImplSystem &a) const;
      bool operator!=(const ImplSystem &a) const;
   };

   template <class Var, class VarPrinter> class AnnotatedImplSystem
   {
      ImplSystem<Var> mImplSystem;
      VarPrinter mVarPrinter;
      size_t mMaxVariable;
      std::string mDimacsAnnotation;
      std::string mSep;

    public:
      AnnotatedImplSystem(ImplSystem<Var> impl_system, VarPrinter var_printer,
                          size_t max_variable,
                          std::string dimacs_annotation = "",
                          std::string sep = "");
      std::ostream &Print(std::ostream &out) const;
   };

   template <class Var, class VarPrinter, class... Args>
   AnnotatedImplSystem<Var, VarPrinter>
   make_annotated_is(ImplSystem<Var> impl_system, VarPrinter var_printer,
                     Args &&... args);

   template <class Var>
   std::ostream &
   operator<<(std::ostream &out, const ImplSystem<Var> &is)
   {
      return is.Print(out);
   }

   template <class Var, class VarPrinter>
   std::ostream &operator<<(std::ostream &out,
                            const AnnotatedImplSystem<Var, VarPrinter> ais);

   using VarImplSystem = ImplSystem<unsigned>;
   using LitImplSystem = ImplSystem<Literal>;
} // namespace SUBool

// Implementation

template <class Var>
SUBool::ImplSystem<Var>::ImplSystem() : mImpl(), mClosureData()
{
}

template <class Var>
template <class InputIterator>
SUBool::ImplSystem<Var>::ImplSystem(InputIterator begin, InputIterator end)
    : mImpl(begin, end), mClosureData()
{
   BuildClosureData();
}

template <class Var>
SUBool::ImplSystem<Var>::ImplSystem(std::initializer_list<ImplType> imp)
    : mImpl(std::move(imp)), mClosureData()
{
   BuildClosureData();
}

template <class Var>
SUBool::ImplSystem<Var>::ImplSystem(std::vector<ImplType> imp)
    : mImpl(std::move(imp)), mClosureData()
{
   BuildClosureData();
}

template <class Var>
SUBool::ImplSystem<Var>::ImplSystem(const ImplSystem &is)
    : mImpl(is.mImpl), mClosureData()
{
   BuildClosureData();
}

template <class Var>
SUBool::ImplSystem<Var>::ImplSystem(ImplSystem &&is) noexcept
    : mImpl(std::move(is.mImpl)), mClosureData()
{
   BuildClosureData();
}

template <class Var> SUBool::ImplSystem<Var>::~ImplSystem() {}

template <class Var>
SUBool::ImplSystem<Var> &
SUBool::ImplSystem<Var>::operator=(ImplSystem &&is) noexcept
{
   mImpl = std::move(is.mImpl);
   BuildClosureData();
   return *this;
}

template <class Var>
SUBool::ImplSystem<Var> &
SUBool::ImplSystem<Var>::operator=(const ImplSystem &is)
{
   mImpl = is.mImpl;
   BuildClosureData();
   return *this;
}

template <class Var>
void
SUBool::ImplSystem<Var>::ClosureData::Clear(size_t reserve_impl_size)
{
   mImplWithEmptyBody.clear();
   mWatched.clear();
   mImplWatch.clear();
   mClosureBuffer.clear();
   mImplWatch.reserve(reserve_impl_size);
}

template <class Var>
void
SUBool::ImplSystem<Var>::ClosureData::SetWatch(const Watch &watch)
{
   ClearWatch(watch.impl_index);
   auto &new_watch_list = mWatched[*watch.watched_var];
   new_watch_list.push_back(watch);
   mImplWatch[watch.impl_index] = std::prev(new_watch_list.end());
}

template <class Var>
void
SUBool::ImplSystem<Var>::ClosureData::ClearWatch(size_t impl_index)
{
   auto i_old_watch = mImplWatch[impl_index];
   mWatched[*(i_old_watch->watched_var)].erase(i_old_watch);
}

template <class Var>
void
SUBool::ImplSystem<Var>::ClosureData::AddImplication(const ImplType &impl,
                                                     size_t i)
{
   if (impl.EmptyBody())
   {
      mImplWithEmptyBody.insert(i);
      mImplWatch.push_back(watch_ptr());
      return;
   }
   auto &watch_list = mWatched[impl.Body().front()];
   watch_list.push_back({i, impl.Body().begin()});
   if (i >= mImplWatch.size())
   {
      mImplWatch.resize(i + 1);
   }
   mImplWatch[i] = std::prev(watch_list.end());
}

template <class Var>
void
SUBool::ImplSystem<Var>::ClosureData::RemoveImplication(const ImplType &impl,
                                                        size_t i)
{
   if (impl.EmptyBody())
   {
      mImplWithEmptyBody.erase(i);
      return;
   }
   ClearWatch(i);
}

template <class Var>
void
SUBool::ImplSystem<Var>::BuildClosureData() const
{
   mClosureData.Clear(mImpl.size());
   size_t i = 0;
   for (const auto &impl : mImpl)
   {
      mClosureData.AddImplication(impl, i++);
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::PushBack(ImplType impl)
{
   mImpl.push_back(std::move(impl));
   // TODO: Insert incrementally
   // BuildClosureData();
   mClosureData.AddImplication(mImpl.back(), mImpl.size() - 1);
}

template <class Var>
template <class InputIterator>
void
SUBool::ImplSystem<Var>::Append(InputIterator begin, InputIterator end)
{
   // mImpl.insert(mImpl.end(), begin, end);
   // TODO: Insert incrementally
   // BuildClosureData();
   std::for_each(begin, end, [this](const ImplType &imp) { PushBack(imp); });
}

template <class Var>
std::ostream &
SUBool::ImplSystem<Var>::Print(std::ostream &os) const
{
   return PrettyPrint(os, [](const Var &v) { return v; });
}

template <class Var>
template <class VarPrinter>
std::ostream &
SUBool::ImplSystem<Var>::PrettyPrint(std::ostream &os, VarPrinter printer,
                                     const std::string &sep) const
{
   return pretty_print_impl<Var>(os, mImpl.begin(), mImpl.end(), printer, sep);
}

template <class Var>
size_t
SUBool::ImplSystem<Var>::Size() const
{
   return mImpl.size();
}

template <class Var>
size_t
SUBool::ImplSystem<Var>::Length() const
{
   size_t len = 0;
   for (const auto &imp : mImpl)
   {
      len += imp.Length();
   }
   return len;
}

template <class Var>
const typename SUBool::ImplSystem<Var>::ImplType &
SUBool::ImplSystem<Var>::At(size_t index) const
{
   return mImpl.at(index);
}

template <class Var>
std::optional<typename SUBool::ImplSystem<Var>::Watch>
SUBool::ImplSystem<Var>::UpdateWatch(const Watch &watch) const
{
   const auto &body = mImpl[watch.impl_index].Body();
   auto i_var = cyclic_next(watch.watched_var, body.begin(), body.end());
   while (i_var != watch.watched_var
          && mClosureData.mClosureBuffer.find(*i_var)
                 != mClosureData.mClosureBuffer.end())
   {
      cyclic_increment(i_var, body.begin(), body.end());
   }
   if (i_var == watch.watched_var)
   {
      return {};
   }
   return std::make_optional(Watch{watch.impl_index, i_var});
}

template <class Var>
void
SUBool::ImplSystem<Var>::EmptyBodiesIntoBuffer(std::queue<Var> &q) const
{
   for (const auto &i : mClosureData.mImplWithEmptyBody)
   {
      for (const auto &var : mImpl[i].Cons())
      {
         if (mClosureData.mClosureBuffer.find(var)
             == mClosureData.mClosureBuffer.end())
         {
            mClosureData.mClosureBuffer.insert(var);
            if (!mClosureData.mWatched[var].empty())
            {
               q.push(var);
            }
         }
      }
   }
}

template <class Var>
template <class IgnorePredicate>
void
SUBool::ImplSystem<Var>::BufferClosure(bool include_empty_bodies,
                                       IgnorePredicate ignore_pred) const
{
   std::queue<Var> q;
   for (const auto &v : mClosureData.mClosureBuffer)
   {
      q.push(v);
   }
   if (include_empty_bodies)
   {
      EmptyBodiesIntoBuffer(q);
   }
   while (!q.empty())
   {
      auto v = q.front();
      q.pop();
      for (auto i_impl = mClosureData.mWatched[v].begin();
           i_impl != mClosureData.mWatched[v].end();)
      {
         if (ignore_pred(i_impl->impl_index))
         {
            ++i_impl;
            continue;
         }
         auto new_watch = UpdateWatch(*i_impl);
         auto next = std::next(i_impl);
         if (new_watch)
         {
            mClosureData.SetWatch(*new_watch);
         }
         else
         {
            for (const auto &var : mImpl[i_impl->impl_index].Cons())
            {
               if (mClosureData.mClosureBuffer.find(var)
                   == mClosureData.mClosureBuffer.end())
               {
                  mClosureData.mClosureBuffer.insert(var);
                  q.push(var);
               }
            }
         }
         i_impl = next;
      }
   }
}

template <class Var>
template <class IgnorePredicate>
void
SUBool::ImplSystem<Var>::ClosureIntoBuffer(const std::vector<Var> &assump,
                                           bool include_empty_bodies,
                                           IgnorePredicate ignore_pred) const
{
   mClosureData.mClosureBuffer.clear();
   mClosureData.mClosureBuffer.insert(assump.begin(), assump.end());
   BufferClosure(include_empty_bodies, ignore_pred);
}

template <class Var>
template <class IgnorePredicate>
std::vector<Var>
SUBool::ImplSystem<Var>::Closure(const std::vector<Var> &assump,
                                 bool include_assumptions,
                                 bool include_empty_bodies,
                                 IgnorePredicate ignore_pred) const
{
   ClosureIntoBuffer(assump, include_empty_bodies, ignore_pred);
   if (!include_assumptions)
   {
      for (const auto &var : assump)
      {
         mClosureData.mClosureBuffer.erase(var);
      }
   }
   std::vector<Var> closure(mClosureData.mClosureBuffer.begin(),
                            mClosureData.mClosureBuffer.end());
   std::sort(closure.begin(), closure.end());
   return closure;
}

template <class Var>
std::vector<Var>
SUBool::ImplSystem<Var>::Closure(const std::vector<Var> &assump,
                                 bool include_empty_bodies) const
{
   return Closure(assump, true, include_empty_bodies,
                  [](size_t) { return false; });
}

template <class Var>
std::vector<Var>
SUBool::ImplSystem<Var>::ClosureWithoutAssumptions(
    const std::vector<Var> &assump, bool include_empty_bodies) const
{
   return Closure(assump, false, include_empty_bodies,
                  [](size_t) { return false; });
}

template <class Var>
void
SUBool::ImplSystem<Var>::MergeBodiesNoRebuild()
{
   if (mImpl.size() <= 1)
   {
      return;
   }
   std::sort(mImpl.begin(), mImpl.end(), ImplBodyCompare<Var>());
   auto first = mImpl.begin();
   for (auto i_impl = mImpl.begin(); i_impl != mImpl.end(); ++i_impl)
   {
      if (first->SmallerBodyThan(*i_impl))
      {
         first = i_impl;
         continue;
      }
      first->MergeCons(*i_impl);
   }
   auto new_end = std::unique(mImpl.begin(), mImpl.end(), ImplEqualBody<Var>());
   mImpl.resize(new_end - mImpl.begin());
}

template <class Var>
void
SUBool::ImplSystem<Var>::MergeBodies()
{
   MergeBodiesNoRebuild();
   BuildClosureData();
}

template <class Var>
void
SUBool::ImplSystem<Var>::RemoveTrivialNoRebuild()
{
   std::remove_if(mImpl.begin(), mImpl.end(),
                  [](const auto &impl) { return impl.Trivial(); });
}

template <class Var>
void
SUBool::ImplSystem<Var>::RemoveTrivial()
{
   RemoveTrivialNoRebuild();
   BuildClosureData();
}

template <class Var>
void
SUBool::ImplSystem<Var>::RemoveDuplicateBodies()
{
   std::sort(mImpl.begin(), mImpl.end(), ImplBodyCompare<Var>());
   auto new_end = std::unique(mImpl.begin(), mImpl.end(), ImplEqualBody<Var>());
   mImpl.resize(new_end - mImpl.begin());
}

template <class Var>
void
SUBool::ImplSystem<Var>::PropagateEmptyBodyNoRebuild()
{
   if (mClosureData.mImplWithEmptyBody.empty())
   {
      return;
   }
   auto closure = Closure({});
   size_t empty_count = 0;
   for (auto &impl : mImpl)
   {
      impl.SubtractFromBody(closure);
      if (impl.EmptyBody())
      {
         ++empty_count;
      }
      else
      {
         impl.SubtractFromCons(closure);
      }
   }
   if (empty_count > 1 || mImpl.front().EmptyBody())
   {
      auto empty_body_check = [](const auto &impl) { return impl.EmptyBody(); };
      auto first_empty =
          std::find_if(mImpl.begin(), mImpl.end(), empty_body_check);
      auto end =
          std::remove_if(std::next(first_empty), mImpl.end(), empty_body_check);
      mImpl.resize(end - mImpl.begin());
      first_empty->UpdateCons(closure, true);
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::PropagateEmptyBody()
{
   PropagateEmptyBodyNoRebuild();
   BuildClosureData(); // TODO: Should be just update of modified/removed
                       // implications
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeRightSaturated()
{
   if (mImpl.empty())
   {
      return;
   }
   logs.TLog(3) << "[ImplSystem::MakeRightSaturated] initial size: "
                << mImpl.size() << std::endl;
   PropagateEmptyBodyNoRebuild();
   logs.TLog(3)
       << "[ImplSystem::MakeRightSaturated] empty body propagated (size: "
       << mImpl.size() << ")" << std::endl;
   RemoveTrivialNoRebuild();
   logs.TLog(3) << "[ImplSystem::MakeRightSaturated] removed trivial "
                   "implications (size: "
                << mImpl.size() << ")" << std::endl;
   MergeBodies();
   logs.TLog(3) << "[ImplSystem::MakeRightSaturated] merged bodies (size: "
                << mImpl.size() << ")" << std::endl;
   const unsigned log_step = (mImpl.size() > 10000 ? 1000 : 0);
   std::vector<std::vector<Var>> new_cons(mImpl.size());
   for (size_t i = 0; i < mImpl.size(); ++i)
   {
      if (log_step > 0 && i > 0 && i % log_step == 0)
      {
         double p = (static_cast<double>(i) / static_cast<double>(mImpl.size()))
                    * 100.0;
         logs.TLog(4) << p << "%" << std::endl;
      }
      if (mImpl[i].EmptyBody())
      {
         continue;
      }
      new_cons[i] =
          Closure(mImpl[i].Body(), false, false, [](size_t) { return false; });
   }
   logs.TLog(3)
       << "[ImplSystem::MakeRightSaturated] updating cons in the implications"
       << std::endl;
   for (size_t i = 0; i < mImpl.size(); ++i)
   {
      if (mImpl[i].EmptyBody())
      {
         continue;
      }
      mImpl[i].UpdateCons(std::move(new_cons[i]), true);
   }
   logs.TLog(3) << "[ImplSystem::MakeRightSaturated] finished" << std::endl;
}

template <class Var>
typename SUBool::ImplSystem<Var>::closure_to_impl_t
SUBool::ImplSystem<Var>::ClosureToImplications()
{
   closure_to_impl_t closure_to_impl; // Trie would be better
   for (size_t i = 0; i < mImpl.size(); ++i)
   {
      closure_to_impl[mImpl[i].BodyWithCons()].insert(i);
   }
   return closure_to_impl;
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeLeftSaturated(
    const closure_to_impl_t &closure_to_impl)
{
   std::vector<std::vector<Var>> new_bodies;
   new_bodies.resize(mImpl.size());
   logs.TLog(3) << "[ImplSystem::MakeLeftSaturated] closure_to_impl.size: "
                << closure_to_impl.size() << std::endl;
   unsigned cnt = 0;
   const unsigned log_step = (mImpl.size() > 10000 ? 1000 : 0);
   for (const auto &equal_impl : closure_to_impl)
   {
      for (const auto &i : equal_impl.second)
      {
         ++cnt;
         if (log_step > 0 && cnt > 0 && cnt % log_step == 0)
         {
            double p =
                (static_cast<double>(cnt) / static_cast<double>(mImpl.size()))
                * 100.0;
            logs.TLog(4) << p << "%" << std::endl;
         }
         if (mImpl[i].EmptyBody())
         {
            continue;
         }
         new_bodies[i] =
             Closure(mImpl[i].Body(), true, false, [&equal_impl](size_t i) {
                return equal_impl.second.find(i) != equal_impl.second.end();
             });
      }
   }
   for (size_t i = 0; i < mImpl.size(); ++i)
   {
      mImpl[i] =
          Impl<Var>(std::move(new_bodies[i]), std::move(mImpl[i].Cons()), true);
      mImpl[i].MakeConsWithoutBody();
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeBodyIrredundant()
{
   bool remove_any = false;
   std::vector<bool> remove_mask(mImpl.size(), false);
   for (size_t i = 0; i < mImpl.size(); ++i)
   {
      if (mImpl[i].EmptyBody())
      {
         continue;
      }
      auto closure = Closure(mImpl[i].Body(), false, true,
                             [i](size_t j) { return j == i; });
      if (std::includes(closure.begin(), closure.end(), mImpl[i].Cons().begin(),
                        mImpl[i].Cons().end()))
      {
         remove_mask[i] = true;
         remove_any = true;
         mClosureData.RemoveImplication(mImpl[i], i);
         continue;
      }
   }
   if (remove_any)
   {
      RemoveImplications(remove_mask);
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::RemoveBodyClosureFromCons(size_t impl_index)
{
   auto body_closure =
       Closure(mImpl[impl_index].Body(), true, true,
               [impl_index](size_t i) { return i == impl_index; });
   mImpl[impl_index].SubtractFromCons(body_closure);
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeRightIrredundantImpl(size_t impl_index)
{
   RemoveBodyClosureFromCons(impl_index);
   std::unordered_set<Var> assump(mImpl[impl_index].Body().begin(),
                                  mImpl[impl_index].Body().end());
   assump.insert(mImpl[impl_index].Cons().begin(),
                 mImpl[impl_index].Cons().end());
   std::vector<Var> cons;
   for (const auto &var : mImpl[impl_index].Cons())
   {
      assump.erase(var);
      mClosureData.mClosureBuffer.clear();
      mClosureData.mClosureBuffer.insert(assump.begin(), assump.end());
      // TODO: It would be nice to avoid using empty bodies
      // Maybe as a parameter?
      BufferClosure(true, [impl_index](size_t i) { return i == impl_index; });
      if (mClosureData.mClosureBuffer.find(var)
          == mClosureData.mClosureBuffer.end())
      {
         cons.push_back(var);
         assump.insert(var);
      }
   }
   mImpl[impl_index].UpdateCons(std::move(cons), true);
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeRightIrredundant()
{
   bool remove_any = false;
   std::vector<bool> remove_mask(mImpl.size(), false);
   for (size_t i = 0; i < mImpl.size(); ++i)
   {
      if (!mImpl[i].EmptyBody())
      {
         MakeRightIrredundantImpl(i);
         if (mImpl[i].EmptyCons())
         {
            remove_mask[i] = true;
            remove_any = true;
         }
      }
   }
   if (remove_any)
   {
      RemoveImplications(remove_mask);
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::RemoveImplications(
    const std::vector<bool> &remove_mask)
{
   vector_erase_indices(mImpl,
                        [&remove_mask](size_t i) { return remove_mask[i]; });
   BuildClosureData();
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeLeftIrredundantImpl(size_t impl_index)
{
   std::unordered_set<Var> assump(mImpl[impl_index].Body().begin(),
                                  mImpl[impl_index].Body().end());
   for (const auto &var : mImpl[impl_index].Body())
   {
      assump.erase(var);
      mClosureData.mClosureBuffer.clear();
      mClosureData.mClosureBuffer.insert(assump.begin(), assump.end());
      BufferClosure(true, [](size_t) { return false; });
      if (mClosureData.mClosureBuffer.find(var)
          != mClosureData.mClosureBuffer.end())
      {
         continue;
      }
      bool subset = true;
      for (const auto cons_var : mImpl[impl_index].Cons())
      {
         if (mClosureData.mClosureBuffer.find(cons_var)
             == mClosureData.mClosureBuffer.end())
         {
            subset = false;
            break;
         }
      }
      if (!subset)
      {
         assump.insert(var);
      }
   }
   if (assump.size() < mImpl[impl_index].Body().size())
   {
      std::vector<Var> body(assump.begin(), assump.end());
      Refine(impl_index,
             ImplType(std::move(body), mImpl[impl_index].Cons(), true));
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeLeftIrredundant()
{
   for (size_t i = 0; i < mImpl.size(); ++i)
   {
      if (!mImpl[i].EmptyBody())
      {
         MakeLeftIrredundantImpl(i);
      }
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeBodyMinimal()
{
   MakeRightSaturated();
   MakeBodyIrredundant();
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeRightLeftSaturated()
{
   MakeRightSaturated();
   MakeLeftSaturated(ClosureToImplications());
   RemoveDuplicateBodies();
   BuildClosureData();
}

template <class Var>
typename SUBool::ImplSystem<Var>::closure_to_impl_t
SUBool::ImplSystem<Var>::RemoveSubsumptions(closure_to_impl_t closure_to_impl)
{
   for (auto &equal_impl : closure_to_impl)
   {
      auto i_impl = equal_impl.second.begin();
      while (i_impl != equal_impl.second.end())
      {
         auto &impl = mImpl[*i_impl];
         if (impl.Trivial())
         {
            i_impl = equal_impl.second.erase(i_impl);
            continue;
         }
         auto j_impl = std::next(i_impl);
         bool removed = false;
         while (!removed && j_impl != equal_impl.second.end())
         {
            auto res = impl.CompareBody(mImpl[*j_impl]);
            switch (res)
            {
            case SetCompareResult::SUBSET:
            case SetCompareResult::EQUAL:
               j_impl = equal_impl.second.erase(j_impl);
               break;
            case SetCompareResult::SUPERSET:
               i_impl = equal_impl.second.erase(i_impl);
               removed = true;
               break;
            case SetCompareResult::INCOMPARABLE:
               ++j_impl;
               break;
            }
         }
         if (!removed)
         {
            ++i_impl;
         }
      }
   }
   return closure_to_impl;
}

template <class Var>
void
SUBool::ImplSystem<Var>::UpdateImplWithClosureToImpl(
    const closure_to_impl_t &closure_to_impl)
{
   impl_list_t new_impl;
   for (const auto &equal_impl : closure_to_impl)
   {
      for (auto &i : equal_impl.second)
      {
         new_impl.push_back(std::move(mImpl[i]));
      }
   }
   mImpl = std::move(new_impl);
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeGDBasis()
{
   logs.TLog(3) << "Making GD basis\n";
   logs.TLog(3) << "Computing right saturation" << std::endl;
   MakeRightSaturated();
   logs.TLog(3) << "Computing closure to implication" << std::endl;
   auto closure_to_impl = ClosureToImplications();
   logs.TLog(3) << "Computing left saturation" << std::endl;
   MakeLeftSaturated(closure_to_impl);
   logs.TLog(3) << "Removing subsumptions" << std::endl;
   closure_to_impl = RemoveSubsumptions(std::move(closure_to_impl));
   logs.TLog(3) << "Forming new list of implications" << std::endl;
   UpdateImplWithClosureToImpl(closure_to_impl);
   logs.TLog(3) << "Building closure data" << std::endl;

   BuildClosureData();
   // Remove trivial implications and subsumptions
}

template <class Var>
void
SUBool::ImplSystem<Var>::MakeConsWithoutBody()
{
   for (auto &impl : mImpl)
   {
      impl.MakeConsWithoutBody();
   }
}

template <class Var>
void
SUBool::ImplSystem<Var>::AddBodyToCons()
{
   std::for_each(mImpl.begin(), mImpl.end(),
                 [](auto &impl) { impl.AddBodyToCons(); });
}

template <class Var>
void
SUBool::ImplSystem<Var>::AddEmptyClosureToImpl(bool add_to_body,
                                               bool add_to_cons)
{
   if (!add_to_body && !add_to_cons)
   {
      return;
   }
   auto empty_closure = Closure({});
   if (add_to_body)
   {
      std::for_each(mImpl.begin(), mImpl.end(), [&empty_closure](auto &impl) {
         if (!impl.EmptyBody())
         {
            impl.AddToBody(empty_closure);
         }
      });
   }
   if (add_to_cons)
   {
      std::for_each(mImpl.begin(), mImpl.end(), [&empty_closure](auto &impl) {
         impl.AddToCons(empty_closure);
      });
   }
}

template <class Var>
bool
SUBool::ImplSystem<Var>::operator==(const ImplSystem &a) const
{
   return mImpl == a.mImpl;
}

template <class Var>
bool
SUBool::ImplSystem<Var>::operator!=(const ImplSystem &a) const
{
   return mImpl != a.mImpl;
}

template <class Var>
const typename SUBool::Impl<Var> &
SUBool::ImplSystem<Var>::operator[](size_t index) const
{
   return mImpl[index];
}

template <class Var>
typename SUBool::ImplSystem<Var>::const_iterator
SUBool::ImplSystem<Var>::begin() const
{
   return mImpl.begin();
}

template <class Var>
typename SUBool::ImplSystem<Var>::const_iterator
SUBool::ImplSystem<Var>::end() const
{
   return mImpl.end();
}

template <class Var>
typename SUBool::ImplSystem<Var>::const_iterator
SUBool::ImplSystem<Var>::cbegin() const
{
   return mImpl.cbegin();
}

template <class Var>
typename SUBool::ImplSystem<Var>::const_iterator
SUBool::ImplSystem<Var>::cend() const
{
   return mImpl.cend();
}

template <class Var>
void
SUBool::ImplSystem<Var>::Refine(size_t i, ImplType impl)
{
   mClosureData.RemoveImplication(impl, i);
   mImpl[i] = std::move(impl);
   mClosureData.AddImplication(mImpl[i], i);
}

template <class Var, class VarPrinter>
SUBool::AnnotatedImplSystem<Var, VarPrinter>::AnnotatedImplSystem(
    ImplSystem<Var> impl_system, VarPrinter var_printer, size_t max_variable,
    std::string dimacs_annotation, std::string sep)
    : mImplSystem(std::move(impl_system)), mVarPrinter(std::move(var_printer)),
      mMaxVariable(max_variable),
      mDimacsAnnotation(std::move(dimacs_annotation)), mSep(std::move(sep))
{
}

template <class Var, class VarPrinter>
std::ostream &
SUBool::operator<<(std::ostream &out,
                   const AnnotatedImplSystem<Var, VarPrinter> ais)
{
   return ais.Print(out);
}

template <class Var, class VarPrinter>
std::ostream &
SUBool::AnnotatedImplSystem<Var, VarPrinter>::Print(std::ostream &out) const
{
   if (!mDimacsAnnotation.empty())
   {
      out << mDimacsAnnotation << std::endl;
   }
   out << "p impl " << mMaxVariable << " " << mImplSystem.Size() << std::endl;
   return mImplSystem.PrettyPrint(out, mVarPrinter, mSep);
}

template <class Var, class VarPrinter, class... Args>
SUBool::AnnotatedImplSystem<Var, VarPrinter>
SUBool::make_annotated_is(ImplSystem<Var> impl_system, VarPrinter var_printer,
                          Args &&... args)
{
   return AnnotatedImplSystem<Var, VarPrinter>(std::move(impl_system),
                                               std::move(var_printer),
                                               std::forward<Args>(args)...);
}

#endif
