/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file impl_system_graph.cpp
 * Creating and manipulating the graphs associated with implication systems.
 */

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/graph/strong_components.hpp>
#include <boost/property_map/dynamic_property_map.hpp>
#include <sstream>
#include <vector>

#include "impl_system_graph.h"

SUBool::LitISGraph::LitISGraph(const LitImplSystem &is)
{
   Build(is);
   FindStrongComponents();
   for (unsigned i = 0; i < mVarSizes.size(); ++i)
   {
      mMaxEffSCCVarSize = std::max(mMaxEffSCCVarSize, EffSCCVarSize(i));
   }
}

void
SUBool::LitISGraph::Build(const LitImplSystem &is)
{
   std::vector<std::pair<unsigned, unsigned>> edges;
   for (const auto &impl : is)
   {
      for (const auto &hl : impl.Cons())
      {
         mMaxVariable = std::max(mMaxVariable, hl.Variable());
      }
      for (const auto &bl : impl.Body())
      {
         mMaxVariable = std::max(mMaxVariable, bl.Variable());
         for (const auto &hl : impl.Cons())
         {
            edges.emplace_back(bl.Index(), hl.Index());
         }
      }
   }
   mGraph = GraphType(edges.begin(), edges.end(), 2 * mMaxVariable);
}

void
SUBool::LitISGraph::FindStrongComponents()
{
   mComponent.resize(num_vertices(mGraph), 0);
   int num =
       strong_components(mGraph, make_iterator_property_map(mComponent.begin(),
                                     get(boost::vertex_index, mGraph)));
   mSizes.resize(num, 0);
   mVarSizes.resize(num, 0);
   mConflict.resize(num, false);
   for (unsigned var = 1; var <= mMaxVariable; ++var)
   {
      auto pos_i = positive_literal_index(var);
      auto neg_i = negative_literal_index(var);
      ++mSizes[mComponent[pos_i]];
      ++mSizes[mComponent[neg_i]];
      ++mVarSizes[mComponent[pos_i]];
      if (mComponent[pos_i] == mComponent[neg_i])
      {
         mConflict[mComponent[pos_i]] = true;
      }
      else
      {
         ++mVarSizes[mComponent[neg_i]];
      }
   }
}

void
SUBool::report_scc(const LitImplSystem &is)
{
   LitISGraph g(is);

   std::string c = "(";
   std::cout << "Number of scc: " << g.SCCCount() << " ";
   for (unsigned i = 0; i < g.SCCCount(); ++i)
   {
      if (g.SCCSize(i) > 1)
      {
         std::cout << c << g.SCCVarSize(i) << '/' << g.SCCSize(i)
                   << (g.SCCConflict(i) ? "c" : "");
         c = ", ";
      }
   }
   if (c == "(")
   {
      std::cout << '(';
   }
   std::cout << ')' << std::endl;
   // std::cout << "Number of scc: " << num << " (" << VectorPrinter(sizes) <<
   // ")"
   //         << std::endl;
#if 0
   boost::dynamic_properties dp;

   auto index = boost::get(boost::vertex_index, g);
   auto index2name = [&](unsigned i) {
      std::stringstream ss;
      ss << Literal(i);
      return ss.str();
   };
   dp.property("name",
               boost::make_transform_value_property_map(index2name, index));

   boost::write_graphviz_dp(std::cout, g, dp, "name");
#endif
}

SUBool::LitISGraph::ImplSCCInfo
SUBool::LitISGraph::InspectImplSCC(const LitImpl &impl) const
{
   std::unordered_set<unsigned> body_scc;
   for (const auto &lit : impl.Body())
   {
      body_scc.insert(SCC(lit));
   }
   ImplSCCInfo scc_info{.main_scc = SCCCount(), .cons_out_main_scc_size = 0};
   for (const auto &lit : impl.Cons())
   {
      auto scc = SCC(lit);
      if (body_scc.find(scc) == body_scc.end())
      {
         ++scc_info.cons_out_main_scc_size;
         continue;
      }
      scc_info.main_scc = scc;
   }
   return scc_info;
}

void
SUBool::LitISGraph::DumpImplSystem(
    unsigned level, const LitImplSystem &is) const
{
   for (unsigned i = 0; i < is.Size(); ++i)
   {
      logs.Log(level) << std::setw(2) << i << " | " << is[i] << std::endl;
   }
   std::vector<std::vector<Literal>> scc_literals(SCCCount());
   logs.TLog(level) << "SCCs:";
   for (unsigned i = 0; i < mMaxVariable * 2; ++i)
   {
      Literal lit(i);
      logs.Log(level) << " " << lit << " (" << SCC(lit) << ")";
      scc_literals[SCC(lit)].push_back(lit);
   }
   logs.Log(level) << std::endl;
   for (unsigned i = 0; i < scc_literals.size(); ++i)
   {
      logs.Log(level) << "SCC " << i << ":";
      for (const auto &lit : scc_literals[i])
      {
         logs.Log(level) << " " << lit;
      }
      logs.Log(level) << std::endl;
   }
}
