/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file impl_system_graph.h
 * Creating and manipulating the graphs associated with implication systems.
 */

#ifndef __IMPL_SYSTEM_GRAPH_H
#define __IMPL_SYSTEM_GRAPH_H

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/strong_components.hpp>

#include "impl_system.h"
#include "nfliteralset.h"
#include "util.h"

namespace SUBool
{
   void report_scc(const LitImplSystem &is);

   // Mainly for finding SCCs
   class LitISGraph
   {
    public:
      using GraphType =
          boost::adjacency_list<boost::vecS, boost::vecS, boost::directedS>;

      struct ImplSCCInfo
      {
         unsigned main_scc;
         unsigned cons_out_main_scc_size;
      };

    private:
      GraphType mGraph{};
      unsigned mMaxVariable{0};
      std::vector<unsigned> mComponent{};
      std::vector<unsigned> mSizes{};
      std::vector<unsigned> mVarSizes{};
      std::vector<bool> mConflict{};
      unsigned mMaxEffSCCVarSize{0};

      void Build(const LitImplSystem &is);
      void FindStrongComponents();

    public:
      LitISGraph(const LitImplSystem &is);
      LitISGraph() = default;

      inline unsigned SCCCount() const;
      inline unsigned SCC(const Literal &lit) const;
      inline unsigned SCCSize(unsigned scc_index) const;
      inline unsigned SCCSize(const Literal &lit) const;
      inline bool SCCConflict(unsigned scc_index) const;
      inline bool SCCConflict(const Literal &lit) const;
      inline unsigned SCCVarSize(unsigned scc_index) const;
      inline unsigned SCCVarSize(const Literal &lit) const;
      inline unsigned EffSCCVarSize(unsigned scc_index) const;
      inline unsigned EffSCCVarSize(const Literal &lit) const;
      inline unsigned MaxEffSCCVarSize() const;

      void DumpImplSystem(unsigned level, const LitImplSystem &is) const;

      ImplSCCInfo InspectImplSCC(const LitImpl &impl) const;
   };
} // namespace SUBool

inline unsigned
SUBool::LitISGraph::SCCCount() const
{
   return mSizes.size();
}

inline unsigned
SUBool::LitISGraph::SCC(const Literal &lit) const
{
   return mComponent.at(lit.Index());
}

inline unsigned
SUBool::LitISGraph::SCCSize(unsigned scc_index) const
{
   return mSizes.at(scc_index);
}

inline unsigned
SUBool::LitISGraph::SCCSize(const Literal &lit) const
{
   return SCCSize(SCC(lit));
}

inline bool
SUBool::LitISGraph::SCCConflict(unsigned scc_index) const
{
   return mConflict.at(scc_index);
}

inline bool
SUBool::LitISGraph::SCCConflict(const Literal &lit) const
{
   return SCCConflict(SCC(lit));
}

inline unsigned
SUBool::LitISGraph::SCCVarSize(unsigned scc_index) const
{
   return mVarSizes.at(scc_index);
}

inline unsigned
SUBool::LitISGraph::SCCVarSize(const Literal &lit) const
{
   return SCCVarSize(SCC(lit));
}

inline unsigned
SUBool::LitISGraph::EffSCCVarSize(unsigned scc_index) const
{
   return SCCConflict(scc_index) ? SCCVarSize(scc_index)
                                 : SCCVarSize(scc_index) - 1;
}

inline unsigned
SUBool::LitISGraph::EffSCCVarSize(const Literal &lit) const
{
   return EffSCCVarSize(SCC(lit));
}

inline unsigned
SUBool::LitISGraph::MaxEffSCCVarSize() const
{
   return mMaxEffSCCVarSize;
}

#endif
