/*
 * Project SUBool
 * Petr Kucera, 2020
 * Jakub Rozlivek, 2020
 */
/**@file kissat_bind.h
 * Binding class to kissat solver library. */

#ifndef __KISSAT_BIND_H
#define __KISSAT_BIND_H

#if HAS_KISSAT

#include "solver_bind.h"
extern "C"
{
#define export expo // c++ keyword "export" used in kissat as variable name
#include KISSAT_INTERNAL_H
#include KISSAT_KISSAT_H
#undef export
}
#include "random_utils.h"
#include "sbexception.h"

namespace SUBool
{

   class KissatSolver
   {

    public:
      kissat *solver;

      KissatSolver(const SolverInterface::ADAPTATION adapt =
                       SolverInterface::ADAPTATION::DEFAULT)
          : solver(kissat_init())
      {
         kissat_set_option(solver, "seed", sub_rand() % 2147483647);
         kissat_set_option(solver, "verbose", -1);
         kissat_set_option(solver, "quiet", 1);
         Adapt(adapt);
      }

      ~KissatSolver() { kissat_release(solver); }

      KissatSolver(const KissatSolver &s) = default;
      KissatSolver(KissatSolver &&s) = default;
      KissatSolver &operator=(const KissatSolver &s) = delete;
      KissatSolver &operator=(KissatSolver &&s) = delete;

      /*
            unsigned
            LearntsSize() const
            {
            return GET(learned);
            }
       */
      void Adapt(const SolverInterface::ADAPTATION adapt);
   };

   template <class Solver> class KissatBind : public SolverInterface
   {

    private:
      std::unique_ptr<Solver> mSolver;
      mutable std::vector<int> mLitBuffer;
      mutable PartialValue mResult;
      ADAPTATION mAdapt;

      bool SolveLitBuffer() const;

    public:
      KissatBind(const SolverInterface::ADAPTATION adapt =
                     SolverInterface::ADAPTATION::DEFAULT);
      KissatBind(const CNF &cnf) : KissatBind() { AddCnf(cnf); }

      virtual bool Reset(const CNF &cnf) override;
      virtual bool AddCnf(const CNF &cnf) override;
      virtual bool AddClause(const Clause &clause) override;

      virtual void Interrupt() const override;

      virtual bool Solve(const std::vector<Literal> &assump) const override;
      virtual bool Solve() const override;
      virtual bool Solve(const PartialAssignment &assump) const override;
      virtual bool Solve(const Assignment &assump) const override;

      /* Throws NoModelException if no model is currently
       * available. */
      virtual PartialValue ValueOfVariable(unsigned var) const override;
      virtual PartialValue ValueOfLiteral(const Literal &lit) const override;
      virtual PartialAssignment Model() const override;
      virtual Clause ConflictClause() const override; // TODO: not working

      virtual size_t LearntsSize() const override;
      virtual std::vector<Clause> Learnts() const override; // TODO:

      virtual bool IsIncremental() const override;
   };

   using KissatBindSolver = KissatBind<KissatSolver>;

   Literal of_kissat_lit(int lit);
   PartialValue of_kissat_value(int val);

   using KissatSolverWrapper = SolverWrapperTemplate<KissatBindSolver>;
} // namespace SUBool

template <class Solver>
SUBool::KissatBind<Solver>::KissatBind(const SolverInterface::ADAPTATION adapt)
    : mSolver(), mLitBuffer(), mResult(PartialValue::UNDEF), mAdapt(adapt)
{
   Reset(CNF());
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::Reset(const CNF &cnf)
{
   mSolver = std::make_unique<Solver>(mAdapt);
   mResult = PartialValue::UNDEF;
   return AddCnf(cnf);
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::AddCnf(const CNF &cnf)
{
   for (const auto &clause : cnf)
   {
      if (!AddClause(clause))
      {
         return false;
      }
   }
   return true;
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::AddClause(const Clause &clause)
{
   mLitBuffer.clear();
   mResult = PartialValue::UNDEF;
   for (const auto &lit : clause)
   {
      kissat_add(mSolver->solver,
                 lit.Positive() ? lit.Variable() : -lit.Variable());
      mLitBuffer.push_back(lit.Positive() ? lit.Variable() : -lit.Variable());
   }
   kissat_add(mSolver->solver, 0);
   return (mResult != PartialValue::FALSE);
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::Solve() const
{
   mLitBuffer.clear();
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::Solve(const std::vector<Literal> &assump) const
{
   mLitBuffer.clear();
   std::for_each(assump.begin(), assump.end(), [this](const auto &lit) {
      mLitBuffer.push_back(lit.mPositive ? lit.Variable() : -lit.Variable());
   });
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::Solve(const PartialAssignment &assump) const
{
   mLitBuffer.clear();
   for (unsigned i = 1; i <= assump.Size(); ++i)
   {
      if (assump[i] != PartialValue::UNDEF)
      {
         mLitBuffer.push_back(assump[i] == PartialValue::TRUE ? i : -i);
      }
   }
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::Solve(const Assignment &assump) const
{
   mLitBuffer.clear();
   for (unsigned i = 1; i <= assump.Size(); ++i)
   {
      mLitBuffer.push_back(assump[i] ? i : -i);
   }
   return SolveLitBuffer();
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::SolveLitBuffer() const
{
   for (const auto &lit : mLitBuffer)
   {
      kissat_add(mSolver->solver, lit);
      kissat_add(mSolver->solver, 0);
   }
   auto ret = kissat_solve(mSolver->solver);
   if (ret == 20)
      mResult = PartialValue::FALSE;
   else if (ret == 10)
      mResult = PartialValue::TRUE;
   else
      mResult = PartialValue::UNDEF;
   IncreaseSolveCount(mResult);
   if (mResult == PartialValue::UNDEF)
   {
      throw SolverTerminatedException("KissatBind::Solve",
                                      "Solver finished without result");
   }
   return bool_of_partial_value(mResult);
}

template <class Solver>
void
SUBool::KissatBind<Solver>::Interrupt() const
{
   kissat_terminate(mSolver->solver);
}

template <class Solver>
SUBool::PartialValue
SUBool::KissatBind<Solver>::ValueOfVariable(unsigned var) const
{
   if (!bool_of_partial_value(mResult, false))
   {
      throw NoModelException("KissatBind::ValueOfVariable",
                             "No model available");
   }
   return of_kissat_value(kissat_value(mSolver->solver, var));
}

template <class Solver>
SUBool::PartialValue
SUBool::KissatBind<Solver>::ValueOfLiteral(const Literal &lit) const
{
   if (!bool_of_partial_value(mResult, false))
   {
      throw NoModelException("KissatBind::ValueOfLiteral",
                             "No model available");
   }
   return of_kissat_value(kissat_value(
       mSolver->solver, lit.mPositive ? lit.Variable() : -lit.Variable()));
}

template <class Solver>
SUBool::PartialAssignment
SUBool::KissatBind<Solver>::Model() const
{
   if (!bool_of_partial_value(mResult, false))
   {
      throw NoModelException("KissatBind::Model", "No model available");
   }
   PartialAssignment alpha(mSolver->solver->vars);
   for (unsigned i = 1; i <= mSolver->solver->vars; ++i)
   {
      alpha[i] = ValueOfVariable(i);
   }
   return alpha;
}

template <class Solver>
SUBool::Clause
SUBool::KissatBind<Solver>::ConflictClause() const
{
   throw NotImplementedException("KissatBind::ConflictClause",
                                 "Not implemented");
   /* // TODO: not working
   std::vector<Literal> conflict_literals;
   if (mSolver->solver->conflict.used)
   {
      conflict_literals.reserve(mSolver->solver->conflict.size);
      for (unsigned i = 0; i < mSolver->solver->conflict.size; ++i)
      {
         conflict_literals.push_back(
             of_kissat_lit(mSolver->solver->conflict.lits[i]));
      }
   }
   return conflict_literals;
   */
}

template <class Solver>
std::vector<SUBool::Clause>
SUBool::KissatBind<Solver>::Learnts() const // TODO:
{
   throw NotImplementedException("KissatBind::Learnts", "Not implemented");
   // std::vector<SUBool::Clause> learnts;
   // return learnts;
}

template <class Solver>
size_t
SUBool::KissatBind<Solver>::LearntsSize() const
{
   throw NotImplementedException("KissatBind::LearntsSize", "Not implemented");
   // return mSolver->LearntsSize();
}

template <class Solver>
bool
SUBool::KissatBind<Solver>::IsIncremental() const
{
   return false;
}

#endif
#endif
