/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file literal.cpp
 * Literal structure.
 */

#include <sstream>
#include <string>

#include "assign.h"
#include "literal.h"
#include "sbexception.h"

SUBool::Literal::Literal() noexcept : mVariable(1), mPositive(true) {}

SUBool::Literal::Literal(unsigned literal_index) noexcept
    : mVariable((literal_index >> 1) + 1), mPositive(!(literal_index & 1))
{
}

SUBool::Literal::Literal(unsigned variable, bool positive)
    : mVariable(variable), mPositive(positive)
{
   if (mVariable == 0)
   {
      throw BadParameterException("Literal variable must be positive.");
   }
}

void
SUBool::Literal::Negate() noexcept
{
   mPositive = !mPositive;
}

SUBool::Literal
SUBool::Literal::Negation() const
{
   return Literal(mVariable, !mPositive);
}

bool
SUBool::Literal::Value(const Assignment &assignment) const
{
   return Value(assignment[mVariable]);
}

bool
SUBool::Literal::Value(bool variable_value) const noexcept
{
   return (variable_value == mPositive);
}

SUBool::PartialValue
SUBool::Literal::Value(SUBool::PartialValue variable_value) const noexcept
{
   return (mPositive ? variable_value : !variable_value);
}

SUBool::PartialValue
SUBool::Literal::Value(const PartialAssignment &assignment) const
{
   if (mVariable > assignment.Size())
   {
      return PartialValue::UNDEF;
   }
   return Value(assignment[mVariable]);
}

SUBool::PartialValue
SUBool::Literal::FalseValue() const noexcept
{
   return (mPositive ? PartialValue::FALSE : PartialValue::TRUE);
}

SUBool::PartialValue
SUBool::Literal::TrueValue() const noexcept
{
   return (mPositive ? PartialValue::TRUE : PartialValue::FALSE);
}

bool
SUBool::Literal::operator==(const Literal &lit) const
{
   return (mVariable == lit.mVariable && mPositive == lit.mPositive);
}

bool
SUBool::Literal::operator!=(const Literal &lit) const
{
   return (mVariable != lit.mVariable || mPositive != lit.mPositive);
}

bool
SUBool::Literal::operator<(const Literal &lit) const
{
   if (mVariable < lit.mVariable)
   {
      return true;
   }
   if (mVariable > lit.mVariable)
   {
      return false;
   }
   return (mPositive > lit.mPositive);
}

bool
SUBool::Literal::operator<=(const Literal &lit) const
{
   if (mVariable < lit.mVariable)
   {
      return true;
   }
   if (mVariable > lit.mVariable)
   {
      return false;
   }
   return (mPositive >= lit.mPositive);
}

bool
SUBool::Literal::operator>(const Literal &lit) const
{
   if (mVariable > lit.mVariable)
   {
      return true;
   }
   if (mVariable < lit.mVariable)
   {
      return false;
   }
   return (mPositive < lit.mPositive);
}

bool
SUBool::Literal::operator>=(const Literal &lit) const
{
   if (mVariable > lit.mVariable)
   {
      return true;
   }
   if (mVariable < lit.mVariable)
   {
      return false;
   }
   return (mPositive <= lit.mPositive);
}

std::string
SUBool::Literal::ToString() const
{
   std::stringstream ss;
   ss << *this;
   return ss.str();
}

std::ostream &
SUBool::operator<<(std::ostream &output, const SUBool::Literal &lit)
{
   if (!lit.mPositive)
   {
      output << "-";
   }
   output << lit.mVariable;
   return output;
}

std::istream &
SUBool::operator>>(std::istream &input, SUBool::Literal &lit)
{
   lit.mVariable = 0;
   lit.mPositive = true;
   if (input.good())
   {
      int lit_int = 0;
      input >> lit_int;
      if (!input.fail())
      {
         lit.mVariable = abs(lit_int);
         lit.mPositive = (lit_int >= 0);
      }
   }
   if (lit.mVariable == 0)
   {
      throw BadInputException("");
   }
   return input;
}

std::vector<SUBool::Literal>
SUBool::literals_of_assignment(const Assignment &alpha)
{
   std::vector<Literal> literals;
   unsigned var = 0;
   for (auto val : alpha)
   {
      literals.emplace_back(++var, val);
   }
   return literals;
}

std::vector<SUBool::Literal>
SUBool::literals_of_assignment(const PartialAssignment &alpha)
{
   std::vector<Literal> literals;
   unsigned var = 1;
   for (auto val : alpha)
   {
      if (val != PartialValue::UNDEF)
      {
         literals.emplace_back(var, (val == PartialValue::TRUE));
      }
      ++var;
   }
   return literals;
}
