/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file literal.h
 * Literal structure.
 */

#ifndef __LITERAL_H
#define __LITERAL_H

#include <boost/container_hash/hash.hpp>
#include <iostream>
#include <vector>

namespace SUBool
{
   class Assignment;
   enum class PartialValue;
   class PartialAssignment;

   struct Literal
   {
      unsigned mVariable;
      bool mPositive;

      Literal() noexcept;
      explicit Literal(unsigned literal_index) noexcept;
      Literal(unsigned variable, bool positive);

      void Negate() noexcept;
      Literal Negation() const;
      bool Value(const Assignment &assignment) const;
      bool Value(bool variable_value) const noexcept;
      PartialValue Value(const PartialAssignment &assignment) const;
      PartialValue Value(PartialValue variable_value) const noexcept;

      PartialValue FalseValue() const noexcept;
      PartialValue TrueValue() const noexcept;

      constexpr unsigned
      Variable() const noexcept
      {
         return mVariable;
      }
      constexpr bool
      Positive() const noexcept
      {
         return mPositive;
      }
      constexpr bool
      Negative() const noexcept
      {
         return !mPositive;
      }

      bool operator==(const Literal &lit) const;
      bool operator!=(const Literal &lit) const;
      bool operator<(const Literal &lit) const;
      bool operator<=(const Literal &lit) const;
      bool operator>(const Literal &lit) const;
      bool operator>=(const Literal &lit) const;

      /* For using in arrays. */
      constexpr unsigned Index() const noexcept;

      std::string ToString() const;

      friend std::ostream &operator<<(std::ostream &output, const Literal &lit);
      friend std::istream &operator>>(std::istream &input, Literal &lit);
   };

   constexpr unsigned positive_literal_index(unsigned var) noexcept;
   constexpr unsigned negative_literal_index(unsigned var) noexcept;

   std::vector<Literal> literals_of_assignment(const Assignment &alpha);
   std::vector<Literal> literals_of_assignment(const PartialAssignment &alpha);

   std::ostream &operator<<(std::ostream &output, const Literal &lit);
   std::istream &operator>>(std::istream &input, Literal &lit);

   Literal literal_negate(Literal lit);

   inline Literal
   make_positive_literal(unsigned var)
   {
      return Literal(var, true);
   }

   inline Literal
   make_negative_literal(unsigned var)
   {
      return Literal(var, false);
   }

   inline size_t
   hash_value(const Literal &lit)
   {
      std::hash<unsigned> hasher;
      return hasher(lit.Index());
   }

   struct LiteralIndex
   {
      constexpr unsigned
      operator()(const Literal &lit) const noexcept
      {
         return lit.Index();
      }
   };

   struct VariableIndex
   {
      constexpr unsigned
      operator()(unsigned var) const noexcept
      {
         return var - 1;
      }
   };

}; // namespace SUBool

namespace std
{
   template <> struct hash<SUBool::Literal>
   {
      size_t
      operator()(const SUBool::Literal &x) const
      {
         return static_cast<size_t>(x.Index());
      }
   };
} // namespace std

inline SUBool::Literal
SUBool::literal_negate(Literal lit)
{
   lit.Negate();
   return lit;
}

constexpr unsigned
SUBool::positive_literal_index(unsigned var) noexcept
{
   return 2 * (var - 1);
}

constexpr unsigned
SUBool::negative_literal_index(unsigned var) noexcept
{
   return 2 * var - 1;
}

constexpr unsigned
SUBool::Literal::Index() const noexcept
{
   return mPositive ? positive_literal_index(mVariable)
                    : negative_literal_index(mVariable);
}

#endif
