/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file min_model.cpp
 * Finding a minimal submodel (subimplicant) of a model.
 */

#include <unordered_set>

#include "min_model.h"

SUBool::MinimalSubimplicant::MinimalSubimplicant(const CNF *cnf)
    : mCnf(cnf), mVarState(), mClauseState(), mUnits(), mLitClauses()
{
   Reset(cnf);
}

void
SUBool::MinimalSubimplicant::Reset(const CNF *cnf)

{
   mCnf = cnf;
   mVarState.clear();
   mClauseState.clear();
   while (!mUnits.empty())
   {
      mUnits.pop();
   }
   mLitClauses.clear();
   if (mCnf == nullptr)
   {
      return;
   }
   mLitClauses.resize(2 * mCnf->MaxVariable());
   mVarState.assign(mCnf->MaxVariable(), VarState::FREE);
   mClauseState.assign(mCnf->Size(), ClauseState::FREE);
   for (unsigned i = 0; i < mCnf->Size(); ++i)
   {
      for (const auto &lit : mCnf->At(i))
      {
         mLitClauses[lit.Index()].push_back(i);
      }
   }
}

void
SUBool::MinimalSubimplicant::Init(const PartialAssignment &alpha) const
{
   mFreedCount = 0;
   while (!mUnits.empty())
   {
      mUnits.pop();
   }
   for (unsigned var = 1; var <= alpha.Size(); ++var)
   {
      mVarState[var - 1] = (alpha[var] == PartialValue::UNDEF ? VarState::UNSET
                                                              : VarState::FREE);
   }
   mClauseState.assign(mCnf->Size(), ClauseState::FREE);
   for (unsigned cl = 0; cl < mCnf->Size(); ++cl)
   {
      CheckClauseState(alpha, cl);
   }
}

SUBool::MinimalSubimplicant::ClauseState
SUBool::MinimalSubimplicant::CheckClauseState(const PartialAssignment &alpha,
                                              unsigned cl) const
{
   if (mClauseState[cl] == ClauseState::FIXED)
   {
      return ClauseState::FIXED;
   }
   std::optional<Literal> sat_lit;
   unsigned count_free = 0;
   for (const auto &lit : mCnf->At(cl))
   {
      if (lit.Value(alpha) == PartialValue::TRUE)
      {
         switch (mVarState[lit.Variable() - 1])
         {
         case VarState::FREE:
            ++count_free;
            if (!sat_lit)
            {
               sat_lit.emplace(lit);
            }
            break;
         case VarState::UNIT:
         case VarState::FIXED:
            mClauseState[cl] = ClauseState::FIXED;
            return ClauseState::FIXED;
         case VarState::UNSET:
            break;
         default:
            throw InvariantFailedException(
                "MinimalSubimplicant::CheckClauseState", "Invalid var state");
         }
      }
   }
   if (count_free == 0)
   {
      throw SUBool::BadParameterException(
          "MinimalSubimplicant::CheckClauseState", "alpha is not a model");
   }
   if (count_free == 1)
   {
      AddUnit(sat_lit->Variable());
      mClauseState[cl] = ClauseState::FIXED;
   }
   return ClauseState::FREE;
}

void
SUBool::MinimalSubimplicant::AddUnit(unsigned var) const
{
   if (mVarState[var - 1] == VarState::FREE)
   {
      mUnits.push(var);
      mVarState[var - 1] = VarState::UNIT;
   }
}

void
SUBool::MinimalSubimplicant::PropagateUnits(
    const PartialAssignment &alpha) const
{
   while (!mUnits.empty())
   {
      auto var = mUnits.front();
      mUnits.pop();
      FixVariable(alpha, var);
   }
}

SUBool::PartialAssignment
SUBool::MinimalSubimplicant::InnerCompute(PartialAssignment alpha) const
{
   PropagateUnits(alpha);
   for (unsigned var = 1; var <= mCnf->MaxVariable(); ++var)
   {
      if (mVarState[var - 1] == VarState::FREE)
      {
         RemoveVariable(alpha, var);
         PropagateUnits(alpha);
      }
   }
   return alpha;
}

SUBool::PartialAssignment
SUBool::MinimalSubimplicant::Compute(PartialAssignment alpha) const
{
   if (mCnf == nullptr)
   {
      throw BadParameterException("MinimalSubimplicant::Compute",
                                  "mCnf is nullptr");
   }
   Init(alpha);
   return InnerCompute(std::move(alpha));
}

SUBool::PartialAssignment
SUBool::MinimalSubimplicant::ComputeWithFixedVars(
    PartialAssignment alpha, const std::vector<unsigned> &fixed_vars) const
{
   if (mCnf == nullptr)
   {
      throw BadParameterException("MinimalSubimplicant::Compute",
                                  "mCnf is nullptr");
   }
   Init(alpha);
   for (const auto var : fixed_vars)
   {
      if (alpha[var] != PartialValue::UNDEF)
      {
         FixVariable(alpha, var);
      }
   }
   return InnerCompute(std::move(alpha));
}

SUBool::PartialAssignment
SUBool::MinimalSubimplicant::ComputeWithFreeVars(
    PartialAssignment alpha, const std::vector<unsigned> &free_vars) const
{
   if (mCnf == nullptr)
   {
      throw BadParameterException("MinimalSubimplicant::Compute",
                                  "mCnf is nullptr");
   }
   std::unordered_set<unsigned> free_vars_set(free_vars.begin(),
                                              free_vars.end());
   for (unsigned var = 1; var <= mCnf->MaxVariable(); ++var)
   {
      if (free_vars_set.find(var) == free_vars_set.end())
      {
         FixVariable(alpha, var);
      }
   }
   PropagateUnits(alpha);
   for (const auto var : free_vars)
   {
      if (mVarState[var - 1] == VarState::FREE)
      {
         RemoveVariable(alpha, var);
         PropagateUnits(alpha);
      }
   }
   return alpha;
}

void
SUBool::MinimalSubimplicant::FixVariable(const PartialAssignment &alpha,
                                         unsigned var) const
{
   ChangeVarState(alpha, Literal(var, (alpha[var] == PartialValue::TRUE)),
                  VarState::FIXED);
}

void
SUBool::MinimalSubimplicant::RemoveVariable(PartialAssignment &alpha,
                                            unsigned var) const
{
   Literal lit(var, (alpha[var] == PartialValue::TRUE));
   alpha[var] = PartialValue::UNDEF;
   ++mFreedCount;
   ChangeVarState(alpha, lit, VarState::UNSET);
}

void
SUBool::MinimalSubimplicant::ChangeVarState(const PartialAssignment &alpha,
                                            Literal lit,
                                            VarState new_state) const
{
   mVarState[lit.Variable() - 1] = new_state;
   for (auto c : mLitClauses[lit.Index()])
   {
      CheckClauseState(alpha, c);
   }
}

SUBool::PartialAssignment
SUBool::minimal_subimplicant(const CNF &cnf, PartialAssignment alpha,
                             const std::vector<unsigned> &fixed_vars,
                             unsigned *freed_count)
{
   auto ms = MinimalSubimplicant(&cnf);
   alpha = ms.ComputeWithFixedVars(std::move(alpha), fixed_vars);
   if (freed_count)
   {
      *freed_count = ms.FreedCount();
   }
   return alpha;
}

