/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file min_model.h
 * Finding a minimal submodel (subimplicant) of a model.
 */

#ifndef __MIN_MODEL_H
#define __MIN_MODEL_H

#include <optional>
#include <queue>

#include "assign.h"
#include "normalform.h"

namespace SUBool
{
   class MinimalSubimplicant
   {
      enum class VarState
      {
         FREE,
         UNIT,
         FIXED,
         UNSET
      };

      enum class ClauseState
      {
         FIXED,
         FREE
      };

      const CNF *mCnf;
      mutable std::vector<VarState> mVarState;
      mutable std::vector<ClauseState> mClauseState;
      mutable std::queue<unsigned> mUnits;
      std::vector<std::vector<unsigned>> mLitClauses;

      void Init(const PartialAssignment &alpha) const;

      void AddUnit(unsigned var) const;
      void PropagateUnits(const PartialAssignment &alpha) const;
      void FixVariable(const PartialAssignment &alpha, unsigned var) const;
      void RemoveVariable(PartialAssignment &alpha, unsigned var) const;
      void ChangeVarState(const PartialAssignment &alpha, Literal lit,
                          VarState new_state) const;
      ClauseState CheckClauseState(const PartialAssignment &alpha,
                                   unsigned cl) const;

      mutable unsigned mFreedCount{0};
      MinimalSubimplicant(CNF &&cnf);

      PartialAssignment InnerCompute(PartialAssignment alpha) const;

    public:
      MinimalSubimplicant(const CNF *cnf);
      MinimalSubimplicant(const MinimalSubimplicant &other) = default;
      MinimalSubimplicant(MinimalSubimplicant &&other) = default;
      ~MinimalSubimplicant() = default;
      MinimalSubimplicant &
      operator=(const MinimalSubimplicant &other) = default;
      MinimalSubimplicant &operator=(MinimalSubimplicant &&other) = default;

      void Reset(const CNF *cnf);

      /**@brief All variables are free, taken in increasing order by their
       * indices.
       */
      PartialAssignment Compute(PartialAssignment alpha) const;
      /**@brief Given variables are fixed and unchanged in alpha, remaining
       * variables are free, taken in increasing order by their indices. */
      PartialAssignment
      ComputeWithFixedVars(PartialAssignment alpha,
                           const std::vector<unsigned> &fixed_vars) const;
      /**@brief Given variables are free, taken in given order. */
      PartialAssignment
      ComputeWithFreeVars(PartialAssignment alpha,
                          const std::vector<unsigned> &free_vars) const;

      unsigned
      FreedCount() const
      {
         return mFreedCount;
      }
   };

   PartialAssignment minimal_subimplicant(
       const CNF &cnf, PartialAssignment alpha,
       const std::vector<unsigned> &fixed_vars = std::vector<unsigned>(),
       unsigned *freed_count = nullptr);
} // namespace SUBool

#endif /* end of include guard */
