/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file mrh.h
 * Algorithms for finding a maximum renamable Horn sub-CNF of a CNF.
 */

#include "mrh.h"
#include "sbexception.h"
#include "horn.h"

#include <vector>

namespace SUBool
{
   double RoundQ(const CNF &cnf, const std::vector<double> &q,
         Assignment *renaming);
   double EvalHForClause(const Clause &clause, const std::vector<double> &q);
   double EvalHForCNF(const CNF &cnf, const std::vector<double> &q);
};

   double
SUBool::EvalHForClause(const Clause &clause, const std::vector<double> &q)
{
   double z=0;
   for(unsigned i=0; i<=clause.Size(); ++i)
   {
      double v=1.0;
      for(unsigned j=0; j<clause.Size(); ++j)
      {
         const Literal &lit=clause[j];
         bool pos=(i==j ? !lit.mPositive : lit.mPositive);
         v*=(pos ? q[lit.mVariable-1] : 1.0-q[lit.mVariable-1]);
      }
      z+=v;
   }
   return z;
}

   double
SUBool::EvalHForCNF(const CNF &cnf, const std::vector<double> &q)
{
   double z=0;
   for(const auto &clause : cnf)
   {
      z+=EvalHForClause(clause, q);
   }
   return z;
}

   double
SUBool::RoundQ(const CNF &cnf, const std::vector<double> &q,
      Assignment *renaming)
{
   if (renaming == nullptr)
   {
      throw NullPointerException("RoundQ", "renaming == nullptr");
   }
   renaming->Resize(cnf.MaxVariable(), false);
   double z=EvalHForCNF(cnf, q);
   std::vector<double> qq(q);
   for(unsigned i=0; i<qq.size(); ++i)
   {
      qq[i]=1.0;
      double zz=EvalHForCNF(cnf, qq);
      if(zz>=z)
      {
         z=zz;
         (*renaming)[i+1]=true;
      }
      else
      {
         qq[i]=0.0;
         (*renaming)[i+1]=false;
         z=EvalHForCNF(cnf, qq);
      }
   }
   return z;
}


   unsigned
SUBool::SimpleMRH(const CNF &cnf, Assignment *renaming)
{
   if (renaming == nullptr)
   {
      throw NullPointerException("SimpleMRH", "renaming == nullptr");
   }
   std::vector<double> q(cnf.MaxVariable(), 0.5);
   return (unsigned)RoundQ(cnf, q, renaming);
}

   void
SUBool::SplitToRenamableHornParts(const CNF &cnf,
      std::vector< std::pair<CNF, Assignment> > *part_list)
{
   if (part_list == nullptr)
   {
      throw NullPointerException("SplitToRenamableHornParts",
            "part_list == nullptr");
   }
   CNF part_cnf(cnf);
   while(part_cnf.Size()>0)
   {
      Assignment renaming(0);
      SimpleMRH(part_cnf, &renaming);
      std::vector<Clause> horn_clauses;
      std::vector<Clause> non_horn_clauses;
      for(const auto &clause : part_cnf)
      {
         Clause ren_clause(clause);
         ren_clause.ApplyRenaming(renaming);
         if(IsHornClause(ren_clause))
         {
            horn_clauses.push_back(clause);
         }
         else
         {
            non_horn_clauses.push_back(clause);
         }
      }
      part_list->push_back(std::make_pair(CNF(horn_clauses), renaming));
      part_cnf=CNF(non_horn_clauses);
   }
}
