/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file nfliteralset.h
 * Defines template NFLiteralSet which is a general structure
 * representing a set of literals.
 * Also defines types Clause and Term for a clause and term special
 * literal sets.
 */

#ifndef __NFLITERALSET_H
#define __NFLITERALSET_H

#include <boost/container_hash/hash.hpp>
#include <utility>

#include <iterator>

#include "assign.h"
#include "literal.h"
#include "logstream.h"
#include "types.h"

namespace SUBool
{
   template <bool DETERMINING_VALUE> class NFLiteralSet
   {
    public:
      using iterator = std::vector<Literal>::iterator;
      using const_iterator = std::vector<Literal>::const_iterator;
      using value_type = Literal;

      struct SizeCompare
      {
         bool
         operator()(const NFLiteralSet<DETERMINING_VALUE> &a,
             const NFLiteralSet<DETERMINING_VALUE> &b) const
         {
            return a.Size() < b.Size();
         }
      };

    private:
      std::vector<Literal> mLiterals;

      static const PartialValue cDeterminingPartialValue =
          partial_value_of_bool(DETERMINING_VALUE);

      std::ostream &WriteToStream(std::ostream &output) const;

    public:
      NFLiteralSet();
      NFLiteralSet(Literal lit);
      NFLiteralSet(Literal l1, Literal l2);
      NFLiteralSet(Literal l1, Literal l2, Literal l3);
      NFLiteralSet(std::initializer_list<Literal> il);
      NFLiteralSet(std::vector<Literal> literals);
      template <class InputIterator>
      NFLiteralSet(InputIterator begin, InputIterator end);
      explicit NFLiteralSet(const PartialAssignment &assignment);

      NFLiteralSet(const NFLiteralSet<DETERMINING_VALUE> &lset) = default;
      NFLiteralSet(NFLiteralSet<DETERMINING_VALUE> &&lset) = default;
      ~NFLiteralSet() = default;
      NFLiteralSet<DETERMINING_VALUE> &operator=(
          const NFLiteralSet<DETERMINING_VALUE> &lset) = default;
      NFLiteralSet<DETERMINING_VALUE> &operator=(
          NFLiteralSet<DETERMINING_VALUE> &&lset) = default;

      size_t Size() const;
      bool Empty() const;

      void Clear();

      static NFLiteralSet<DETERMINING_VALUE>
      EmptyClause()
      {
         return NFLiteralSet<DETERMINING_VALUE>();
      }

      bool Value(const Assignment &assignment) const;
      PartialValue Value(const PartialAssignment &assignment) const;
      // In case the value of element is not undefined, it is empty
      // after the application of the assignment
      PartialValue Apply(const PartialAssignment &assignment);
      PartialValue Apply(const PartialAssignment &assignment,
          NFLiteralSet<DETERMINING_VALUE> *new_element) const;

      const Literal &GetLiteral(size_t index) const;
      const Literal &operator[](size_t index) const;
      const Literal &Front() const;
      const Literal &Back() const;

      /* Throws NotFoundException if there is no literal on this
       * variable. */
      const Literal &LiteralOnVariable(unsigned var) const;

      unsigned MaxVariable() const;

      static bool DeterminingValue();
      bool
      Trivial() const; // Nonempty and contains complemented pair of literals.

      SetCompareResult SubsetCompare(
          const NFLiteralSet<DETERMINING_VALUE> &ls) const;

      friend std::ostream &
      operator<<(
          std::ostream &output, const NFLiteralSet<DETERMINING_VALUE> &ls)
      {
         return ls.WriteToStream(output);
      }

      std::pair<unsigned, unsigned> CountVariablePolarities() const;

      bool operator==(const NFLiteralSet<DETERMINING_VALUE> &ls) const;
      bool operator!=(const NFLiteralSet<DETERMINING_VALUE> &ls) const;
      bool operator<(const NFLiteralSet<DETERMINING_VALUE> &ls) const;
      bool operator<=(const NFLiteralSet<DETERMINING_VALUE> &ls) const;
      bool operator>(const NFLiteralSet<DETERMINING_VALUE> &ls) const;
      bool operator>=(const NFLiteralSet<DETERMINING_VALUE> &ls) const;

      bool Subsumes(const NFLiteralSet<DETERMINING_VALUE> &ls) const;

      const_iterator
      begin() const
      {
         return mLiterals.begin();
      }
      const_iterator
      end() const
      {
         return mLiterals.end();
      }
      const_iterator
      cbegin() const
      {
         return mLiterals.cbegin();
      }
      const_iterator
      cend() const
      {
         return mLiterals.cend();
      }
      const std::vector<Literal>
      Literals() const &
      {
         return mLiterals;
      }
      std::vector<Literal>
      Literals() &&
      {
         return std::move(mLiterals);
      }
      void NegateLiterals();

      void PushBack(Literal lit);
      void
      push_back(Literal lit)
      {
         PushBack(lit);
      }

      void Append(const NFLiteralSet<DETERMINING_VALUE> &ls);

      void ApplyRenaming(const Assignment &renaming);

    private:
      void Normalize();
   };

   using Clause = NFLiteralSet<true>;
   using Term = NFLiteralSet<false>;

}; // namespace SUBool

namespace std
{
   template <bool DETERMINING_VALUE>
   struct hash<SUBool::NFLiteralSet<DETERMINING_VALUE>>
   {
      size_t
      operator()(const SUBool::NFLiteralSet<DETERMINING_VALUE> &lset) const
      {
         return boost::hash_value(lset.Literals());
      }
   };
} // namespace std

#include "nfliteralset_impl.h"

#endif
