/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnf.cpp
 * NNF representation, a wrapper around NNF node which is the root.
 */
#include <sstream>
#include <stack>

#include "nnf.h"
#include "nnfvisit.h"
#include "sbexception.h"

namespace SUBool
{
   class StatsVisitor : public NnfVisitor
   {
    public:
      unsigned mMaxVariable{0};
      size_t mNodeCount{0};
      size_t mEdgeCount{0};

      virtual void
      VisitLiteral(const std::shared_ptr<const NnfLiteral> &node) override
      {
         ++mNodeCount;
         mMaxVariable = std::max(mMaxVariable, node->MaxVariable());
      }

      virtual void
      VisitAnd(const std::shared_ptr<const NnfGateAnd> &node) override
      {
         ++mNodeCount;
         mEdgeCount += node->InputSize();
      }

      virtual void
      VisitOr(const std::shared_ptr<const NnfGateOr> &node) override
      {
         ++mNodeCount;
         mEdgeCount += node->InputSize();
      }

      virtual void
      VisitDecision(
          const std::shared_ptr<const NnfGateDecision> & /*node*/) override
      {
         ++mNodeCount;
         mEdgeCount += 2;
      }

      virtual void
      VisitTrue(const std::shared_ptr<const NnfTrue> & /*node*/) override
      {
         ++mNodeCount;
      }

      virtual void
      VisitFalse(const std::shared_ptr<const NnfFalse> & /*node*/) override
      {
         ++mNodeCount;
      }

      virtual void
      VisitLeafCNF(const CnfPtr &node) override
      {
         ++mNodeCount;
         mMaxVariable = std::max(mMaxVariable, node->MaxVariable());
      }
   };

   void nnf_header(std::istream &input, unsigned *max_var, size_t *edge_count,
       size_t *node_count);
   NnfNode::tPointer nnf_parse_and(
       const std::string &line, const std::vector<NnfNode::tPointer> &nodes);
   NnfNode::tPointer nnf_parse_literal(const std::string &line,
       std::unordered_map<Literal, NnfNode::tPointer> &literal_nodes);
   NnfNode::tPointer nnf_parse_or(
       const std::string &line, const std::vector<NnfNode::tPointer> &nodes);
   NnfNode::tPointer nnf_parse_decision(
       const std::string &line, const std::vector<NnfNode::tPointer> &nodes);

   class NnfCopyVisitor : public NnfVisitor
   {
      NnfNode::tValueCache<NnfNode::tPointer> mNodeCopy{};

    public:
      virtual void
      VisitLiteral(const std::shared_ptr<const NnfLiteral> &node) override
      {
         node->CacheValue(mNodeCopy, NnfNode::MakeLiteral(node->GetLiteral()));
      }
      virtual void
      VisitAnd(const std::shared_ptr<const NnfGateAnd> &node) override
      {
         std::vector<NnfNode::tPointer> child_nodes;
         child_nodes.reserve(node->InputSize());
         for (const auto &child : *node)
         {
            child_nodes.push_back(*child->CachedValue(mNodeCopy));
         }
         node->CacheValue(mNodeCopy, NnfNode::MakeAnd(child_nodes));
      }

      virtual void
      VisitOr(const std::shared_ptr<const NnfGateOr> &node) override
      {
         std::vector<NnfNode::tPointer> child_nodes;
         child_nodes.reserve(node->InputSize());
         for (const auto &child : *node)
         {
            child_nodes.push_back(*child->CachedValue(mNodeCopy));
         }
         if (node->SplitVar() > 0)
         {
            node->CacheValue(mNodeCopy,
                NnfNode::MakeOr(node->SplitVar(), std::move(child_nodes[0]),
                    std::move(child_nodes[1])));
         }
         else
         {
            node->CacheValue(mNodeCopy, NnfNode::MakeOr(child_nodes));
         }
      }

      virtual void
      VisitDecision(const std::shared_ptr<const NnfGateDecision> &node) override
      {
         node->CacheValue(
             mNodeCopy, NnfNode::MakeDecision(node->DecisionVar(),
                            *node->TrueInput()->CachedValue(mNodeCopy),
                            *node->FalseInput()->CachedValue(mNodeCopy)));
      }

      virtual void
      VisitTrue(const std::shared_ptr<const NnfTrue> &node) override
      {
         node->CacheValue(mNodeCopy, NnfNode::MakeTrue());
      }

      virtual void
      VisitFalse(const std::shared_ptr<const NnfFalse> &node) override
      {
         node->CacheValue(mNodeCopy, NnfNode::MakeFalse());
      }

      virtual void
      VisitLeafCNF(const CnfPtr &node) override
      {
         node->CacheValue(mNodeCopy, NnfNode::MakeLeafCNF(node->Cnf()));
      }

      static NnfNode::tPointer
      MakeRootCopy(const Nnf &nnf)
      {
         NnfCopyVisitor visitor;
         nnf.AcceptPostorder(visitor);
         return *nnf.Root()->CachedValue(visitor.mNodeCopy);
      }
   };

   class NnfConsolidateLiteralsVisitor : public NnfVisitorMut
   {
    private:
      std::unordered_map<Literal, NnfNode::tPointer> mLiteralNodes{};

      void
      VisitGate(const std::shared_ptr<NnfGate> &node)
      {
         for (auto &input : *node)
         {
            if (input->NodeType() == NnfNode::Type::Literal)
            {
               auto lit =
                   std::dynamic_pointer_cast<NnfLiteral>(input)->GetLiteral();
               input = mLiteralNodes[lit];
            }
         }
      }

    public:
      virtual void
      VisitLiteral(const std::shared_ptr<NnfLiteral> &node)
      {
         if (mLiteralNodes.find(node->GetLiteral()) == mLiteralNodes.end())
         {
            mLiteralNodes[node->GetLiteral()] = node;
         }
      }

      virtual void
      VisitAnd(const std::shared_ptr<NnfGateAnd> &node)
      {
         VisitGate(node);
      }

      virtual void
      VisitOr(const std::shared_ptr<NnfGateOr> &node)
      {
         VisitGate(node);
      }

      virtual void
      VisitDecision(const std::shared_ptr<NnfGateDecision> &node)
      {
         VisitGate(node);
      }

      virtual void
      VisitTrue(const std::shared_ptr<NnfTrue> & /*node*/)
      {
      }

      virtual void
      VisitFalse(const std::shared_ptr<NnfFalse> & /*node*/)
      {
      }
   };

} // namespace SUBool

SUBool::Nnf::Nnf(NnfNode::tPointer root) : mRoot(root) {}
SUBool::Nnf::Nnf(const Nnf &nnf)
    : mRoot(NnfCopyVisitor::MakeRootCopy(nnf)), mMaxVariable(nnf.MaxVariable()),
      mEdgeCount(nnf.EdgeCount()), mNodeCount(nnf.NodeCount())
{
}

SUBool::Nnf &
SUBool::Nnf::operator=(const Nnf &nnf)
{
   return *this = Nnf(nnf);
}

unsigned
SUBool::Nnf::MaxVariable() const
{
   if (!mMaxVariable)
   {
      RecomputeStats();
   }
   return *mMaxVariable;
}

size_t
SUBool::Nnf::EdgeCount() const
{
   if (!mEdgeCount)
   {
      RecomputeStats();
   }
   return *mEdgeCount;
}

size_t
SUBool::Nnf::NodeCount() const
{
   if (!mNodeCount)
   {
      RecomputeStats();
   }
   return *mNodeCount;
}

bool
SUBool::Nnf::Value(const Assignment &assignment) const
{
   return mRoot->Value(assignment);
}

SUBool::PartialValue
SUBool::Nnf::Value(const PartialAssignment &assignment) const
{
   return mRoot->Value(assignment);
}

bool
SUBool::Nnf::operator()(const Assignment &assignment) const
{
   return Value(assignment);
}

SUBool::PartialValue
SUBool::Nnf::operator()(const PartialAssignment &assignment) const
{
   return Value(assignment);
}

std::ostream &
SUBool::Nnf::WriteToStream(std::ostream &output) const
{
   output << "nnf " << NodeCount() << " " << EdgeCount() << " " << MaxVariable()
          << std::endl;
   output << *mRoot;
   return output;
}

std::ostream &
SUBool::operator<<(std::ostream &output, const Nnf &nnf)
{
   return nnf.WriteToStream(output);
}

void
SUBool::Nnf::RecomputeStats() const
{
   StatsVisitor visitor;
   AcceptPostorder(visitor);
   mMaxVariable = visitor.mMaxVariable;
   mNodeCount = visitor.mNodeCount;
   mEdgeCount = visitor.mEdgeCount;
}

void
SUBool::Nnf::InvalidateStats()
{
   mMaxVariable.reset();
   mNodeCount.reset();
   mEdgeCount.reset();
}

SUBool::Nnf
SUBool::Nnf::ReadFromStream(std::istream &input)
{
   unsigned max_var = 0;
   size_t edge_count = 0;
   size_t node_count = 0;
   nnf_header(input, &max_var, &edge_count, &node_count);
   std::vector<NnfNode::tPointer> nodes;
   std::unordered_map<Literal, NnfNode::tPointer> literal_nodes;
   nodes.reserve(node_count);
   while (input.good())
   {
      std::string line;
      std::getline(input, line);
      if (!line.empty())
      {
         switch (line[0])
         {
         case 'A':
            nodes.push_back(nnf_parse_and(line, nodes));
            break;
         case 'L':
            nodes.push_back(nnf_parse_literal(line, literal_nodes));
            break;
         case 'O':
            nodes.push_back(nnf_parse_or(line, nodes));
            break;
         case 'D':
            nodes.push_back(nnf_parse_decision(line, nodes));
            break;
         case 'c':
            break;
         default:
            throw ParseException("Nnf::ReadFromStream", "Bad line: " + line);
         }
      }
   }
   if (nodes.empty())
   {
      throw BadInputException(
          "Nnf::ReadFromStream", "Stream did not contain any node");
   }
   Nnf nnf(nodes.back(), max_var, edge_count, node_count);
   return nnf;
}

void
SUBool::nnf_header(std::istream &input, unsigned *max_var, size_t *edge_count,
    size_t *node_count)
{
   while (input.good())
   {
      std::string line;
      std::getline(input, line);
      if (line.empty() || line[0] == 'c')
      {
         continue;
      }
      if (line.substr(0, 4) == "nnf ")
      {
         std::stringstream ss(line.substr(4));
         ss >> *node_count;
         ss >> *edge_count;
         ss >> *max_var;
         return;
      }
      throw ParseException("nnf_header", "Bad line " + line);
   }
   throw ParseException("nnf_header", "Header not found");
}

SUBool::NnfNode::tPointer
SUBool::nnf_parse_and(
    const std::string &line, const std::vector<NnfNode::tPointer> &nodes)
{
   std::stringstream ss(line);
   char c;
   ss >> c;
   std::vector<NnfNode::tPointer> inputs;
   unsigned n;
   ss >> n;
   for (unsigned i = 0; i < n; ++i)
   {
      size_t index;
      ss >> index;
      if (index >= nodes.size())
      {
         throw ParseException(
             "nnf_parse_and", "Bad input index " + std::to_string(i));
      }
      inputs.push_back(nodes[index]);
   }
   return NnfNode::MakeAnd(inputs);
}

SUBool::NnfNode::tPointer
SUBool::nnf_parse_literal(const std::string &line,
    std::unordered_map<Literal, NnfNode::tPointer> &literal_nodes)
{
   std::stringstream ss(line);
   char c;
   ss >> c;
   Literal lit;
   ss >> lit;
   auto lit_node = literal_nodes.find(lit);
   if (lit_node != literal_nodes.end())
   {
      return lit_node->second;
   }
   auto new_lit_node = NnfNode::MakeLiteral(lit);
   literal_nodes[lit] = new_lit_node;
   return new_lit_node;
}

SUBool::NnfNode::tPointer
SUBool::nnf_parse_or(
    const std::string &line, const std::vector<NnfNode::tPointer> &nodes)
{
   std::stringstream ss(line);
   char c;
   ss >> c;
   unsigned split_var;
   ss >> split_var;
   std::vector<NnfNode::tPointer> inputs;
   unsigned n;
   ss >> n;
   for (unsigned i = 0; i < n; ++i)
   {
      size_t index;
      ss >> index;
      if (index >= nodes.size())
      {
         throw ParseException("nnf_parse_or",
             "Bad input index (" + std::to_string(i) + "in " + line + ")");
      }
      inputs.push_back(nodes[index]);
   }
   if (split_var > 0)
   {
      if (n != 2 || inputs.size() != 2)
      {
         throw ParseException("nnf_parse_or",
             "Positive split variable with a wrong number of inputs (" + line
                 + ")");
      }
      return NnfNode::MakeOr(split_var, inputs[0], inputs[1]);
   }
   return NnfNode::MakeOr(inputs);
}

SUBool::NnfNode::tPointer
SUBool::nnf_parse_decision(
    const std::string &line, const std::vector<NnfNode::tPointer> &nodes)
{
   std::stringstream ss(line);
   char c;
   ss >> c;
   unsigned decision_var;
   ss >> decision_var;
   unsigned true_node;
   ss >> true_node;
   unsigned false_node;
   ss >> false_node;
   return NnfNode::MakeDecision(
       decision_var, nodes[true_node], nodes[false_node]);
}

void
SUBool::Nnf::AcceptTopological(NnfVisitor &visitor) const
{
   std::stack<NnfNode::tConstPointer> order_stack;
   AcceptPostorderFunction([&order_stack](const NnfNode::tConstPointer &node)
       { order_stack.push(node); });
   while (!order_stack.empty())
   {
      auto node = order_stack.top();
      order_stack.pop();
      node->CallVisit(visitor);
   }
}
void
SUBool::Nnf::AcceptTopologicalMut(NnfVisitorMut &visitor)
{
   InvalidateStats();
   std::stack<NnfNode::tPointer> order_stack;
   AcceptPostorderFunctionMut([&order_stack](const NnfNode::tPointer &node)
       { order_stack.push(node); });
   while (!order_stack.empty())
   {
      auto node = order_stack.top();
      order_stack.pop();
      node->CallVisitMut(visitor);
   }
}

void
SUBool::Nnf::ConsolidateLiterals()
{
   NnfConsolidateLiteralsVisitor visitor;
   AcceptPostorderMut(visitor);
}
