/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnf.h
 * NNF representation, a wrapper around NNF node which is the root.
 */

#ifndef __NNF_H
#define __NNF_H

#include "nnfgate.h"
#include "nnfleaf.h"
#include "nnfnode.h"
#include "nnfvisit.h"
#include "util.h"

namespace SUBool
{
   class Nnf
   {
      NnfNode::tPointer mRoot;
      mutable std::optional<unsigned> mMaxVariable{};
      mutable std::optional<size_t> mEdgeCount{};
      mutable std::optional<size_t> mNodeCount{};

      void RecomputeStats() const;

      Nnf(NnfNode::tPointer root, unsigned max_var, size_t edge_count,
          size_t node_count)
          : mRoot(root), mMaxVariable(max_var), mEdgeCount(edge_count),
            mNodeCount(node_count)
      {
      }

      void InvalidateStats();

    public:
      Nnf(Nnf &&nnf) noexcept = default;
      Nnf &operator=(Nnf &&nnf) noexcept = default;
      Nnf(const Nnf &nnf);            // deep copy
      Nnf &operator=(const Nnf &nnf); // deep copy
      ~Nnf() = default;

      Nnf(NnfNode::tPointer root);

      unsigned MaxVariable() const;
      size_t EdgeCount() const;
      size_t NodeCount() const;
      NnfNode::tConstPointer
      Root() const
      {
         return mRoot;
      }

      bool Value(const Assignment &assignment) const;
      PartialValue Value(const PartialAssignment &assignment) const;

      bool operator()(const Assignment &assignment) const;
      PartialValue operator()(const PartialAssignment &assignment) const;

      std::ostream &WriteToStream(std::ostream &output) const;
      static Nnf ReadFromStream(std::istream &input);

      // Consolidates literals so that each literal has at most one node.
      void ConsolidateLiterals();

      void
      AcceptPostorder(NnfVisitor &visitor) const
      {
         mRoot->AcceptPostorder(visitor);
      }

      void
      AcceptPostorderMut(NnfVisitorMut &visitor)
      {
         InvalidateStats();
         mRoot->AcceptPostorderMut(visitor);
      }
      void
      AcceptPreorder(NnfVisitor &visitor) const
      {
         mRoot->AcceptPreorder(visitor);
      }

      void
      AcceptPreorderMut(NnfVisitorMut &visitor)
      {
         InvalidateStats();
         mRoot->AcceptPreorderMut(visitor);
      }

      template <typename F>
      void
      AcceptPostorderFunction(F f) const
      {
         NnfFuncVisitor<F> visitor(f);
         AcceptPostorder(visitor);
      }
      template <typename F>
      void
      AcceptPostorderFunctionMut(F f)
      {
         NnfFuncVisitorMut<F> visitor(f);
         AcceptPostorderMut(visitor);
      }
      template <typename F>
      void
      AcceptPreorderFunction(F f) const
      {
         NnfFuncVisitor<F> visitor(f);
         AcceptPreorder(visitor);
      }
      template <typename F>
      void
      AcceptPreorderFunctionMut(F f)
      {
         NnfFuncVisitorMut<F> visitor(f);
         AcceptPreorderMut(visitor);
      }

      /**@brief Visits the nodes of a NNF in topological order starting with the
       * root.
       *
       * The function first determines a topological order, then it visits the
       * nodes in this order.
       **/
      void AcceptTopological(NnfVisitor &visitor) const;

      /**@brief Visits the nodes of a NNF in topological order starting with the
       * root.
       *
       * The function first determines a topological order, then it visits the
       * nodes in this order. The order is fixed after it is computed, so if the
       * NNF is modified during the visit, the order does not change.
       * Invalidates the statistics.
       **/
      void AcceptTopologicalMut(NnfVisitorMut &visitor);

      template <typename F>
      void
      AcceptTopologicalFunction(F f) const
      {
         NnfFuncVisitor<F> visitor(f);
         AcceptTopological(visitor);
      }
      template <typename F>
      void
      AcceptTopologicalFunctionMut(F f)
      {
         NnfFuncVisitorMut<F> visitor(f);
         AcceptTopologicalMut(visitor);
      }
   };

   std::ostream &operator<<(std::ostream &output, const Nnf &nnf);

   template <>
   inline Nnf
   read_from_stream(std::istream &is)
   {
      return Nnf::ReadFromStream(is);
   }
} // namespace SUBool

#endif
