/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfgate.cpp
 * NNF gates (various inner nodes)
 */

#include "nnfgate.h"
#include "nnfvisit.h"
#include "sbexception.h"

std::ostream &
SUBool::NnfGate::WriteToStream(std::ostream &output,
                               tValueCache<size_t> &node_id,
                               size_t &next_id) const
{
   if (CachedValue(node_id))
   {
      return output;
   }
   std::vector<size_t> input_id;
   for (const auto &input : mInputs)
   {
      auto last_id = next_id;
      input->WriteToStream(output, node_id, next_id);
      if (last_id != next_id)
      {
         output << std::endl;
      }
      input_id.push_back(node_id[input]);
   }
   CacheValue(node_id, next_id++);
   return WriteToStreamWithId(output, input_id);
}

SUBool::NnfNode::iterator
SUBool::NnfGate::begin()
{
   return mInputs.begin();
}

SUBool::NnfNode::iterator
SUBool::NnfGate::end()
{
   return mInputs.end();
}

SUBool::NnfNode::const_iterator
SUBool::NnfGate::begin() const
{
   return mInputs.begin();
}

SUBool::NnfNode::const_iterator
SUBool::NnfGate::end() const
{
   return mInputs.end();
}

SUBool::NnfNode::const_iterator
SUBool::NnfGate::cbegin() const
{
   return mInputs.begin();
}

SUBool::NnfNode::const_iterator
SUBool::NnfGate::cend() const
{
   return mInputs.end();
}

size_t
SUBool::NnfGate::InputSize() const
{
   return mInputs.size();
}

void
SUBool::NnfGate::AddChild(tPointer child)
{
   mInputs.push_back(child);
}

bool
SUBool::NnfGateAnd::Value(const Assignment &assignment,
                          tValueCache<bool> &cached_values) const
{
   auto val = CachedValue(cached_values);
   if (val)
   {
      return *val;
   }
   for (const auto &input : *this)
   {
      if (!input->Value(assignment, cached_values))
      {
         CacheValue(cached_values, false);
         return false;
      }
   }
   CacheValue(cached_values, true);
   return true;
}

SUBool::PartialValue
SUBool::NnfGateAnd::Value(const PartialAssignment &assignment,
                          tValueCache<PartialValue> &cached_values) const
{
   auto val = CachedValue(cached_values);
   if (val)
   {
      return *val;
   }
   PartialValue pv = PartialValue::TRUE;
   for (const auto &input : *this)
   {
      pv = pv & input->Value(assignment, cached_values);
      if (pv == PartialValue::FALSE)
      {
         CacheValue(cached_values, PartialValue::FALSE);
         return PartialValue::FALSE;
      }
   }
   CacheValue(cached_values, pv);
   return pv;
}

std::ostream &
SUBool::NnfGateAnd::WriteToStreamWithId(
    std::ostream &output, const std::vector<size_t> &input_id) const
{
   output << "A " << InputSize();
   for (auto id : input_id)
   {
      output << ' ' << id;
   }
   return output;
}

void
SUBool::NnfGateAnd::CallVisit(NnfVisitor &visitor) const
{
   visitor.VisitAnd(
       std::dynamic_pointer_cast<const NnfGateAnd>(shared_from_this()));
}

void
SUBool::NnfGateAnd::CallVisitMut(NnfVisitorMut &visitor)
{
   visitor.VisitAnd(std::dynamic_pointer_cast<NnfGateAnd>(shared_from_this()));
}

void
SUBool::NnfGateOr::AddChild(tPointer child)
{
   NnfGate::AddChild(child);
   if (mSplitVar > 0)
   {
      mSplitVar = 0;
   }
}

bool
SUBool::NnfGateOr::Value(const Assignment &assignment,
                         tValueCache<bool> &cached_values) const
{
   auto val = CachedValue(cached_values);
   if (val)
   {
      return *val;
   }
   for (const auto &input : *this)
   {
      if (input->Value(assignment, cached_values))
      {
         CacheValue(cached_values, true);
         return true;
      }
   }
   CacheValue(cached_values, false);
   return false;
}

SUBool::PartialValue
SUBool::NnfGateOr::Value(const PartialAssignment &assignment,
                         tValueCache<PartialValue> &cached_values) const
{
   auto val = CachedValue(cached_values);
   if (val)
   {
      return *val;
   }
   PartialValue pv = PartialValue::FALSE;
   for (const auto &input : *this)
   {
      pv = pv | input->Value(assignment, cached_values);
      if (pv == PartialValue::TRUE)
      {
         CacheValue(cached_values, PartialValue::TRUE);
         return PartialValue::TRUE;
      }
   }
   CacheValue(cached_values, pv);
   return pv;
}

std::ostream &
SUBool::NnfGateOr::WriteToStreamWithId(
    std::ostream &output, const std::vector<size_t> &input_id) const
{
   output << "O " << mSplitVar << " " << InputSize();
   for (auto id : input_id)
   {
      output << ' ' << id;
   }
   return output;
}

void
SUBool::NnfGateOr::CallVisit(NnfVisitor &visitor) const
{
   visitor.VisitOr(
       std::dynamic_pointer_cast<const NnfGateOr>(shared_from_this()));
}

void
SUBool::NnfGateOr::CallVisitMut(NnfVisitorMut &visitor)
{
   visitor.VisitOr(std::dynamic_pointer_cast<NnfGateOr>(shared_from_this()));
}

bool
SUBool::NnfGateDecision::Value(const Assignment &assignment,
                               tValueCache<bool> &cached_values) const
{
   auto val =
       Input(1 - assignment[mDecisionVar])->Value(assignment, cached_values);
   CacheValue(cached_values, val);
   return val;
}

SUBool::PartialValue
SUBool::NnfGateDecision::Value(const PartialAssignment &assignment,
                               tValueCache<PartialValue> &cached_values) const
{
   switch (assignment[mDecisionVar])
   {
   case PartialValue::FALSE:
      return Input(1)->Value(assignment, cached_values);
   case PartialValue::TRUE:
      return Input(0)->Value(assignment, cached_values);
   case PartialValue::UNDEF:
   {
      auto v0 = Input(0)->Value(assignment, cached_values);
      auto v1 = Input(1)->Value(assignment, cached_values);
      if (v0 == PartialValue::TRUE && v1 == PartialValue::TRUE)
      {
         return PartialValue::TRUE;
      }
      if (v0 == PartialValue::FALSE && v1 == PartialValue::FALSE)
      {
         return PartialValue::FALSE;
      }
      return PartialValue::UNDEF;
   }
   };
   throw BadParameterException("NnfGateDecision::Value",
                               "Unknown partial value");
}

void SUBool::NnfGateDecision::AddChild(tPointer /*child*/)
{
   throw InvariantFailedException("NnfGateDecision::AddChild",
                                  "Cannot add a child to the decision node.");
}

std::ostream &
SUBool::NnfGateDecision::WriteToStreamWithId(
    std::ostream &output, const std::vector<size_t> &input_id) const
{
   return output << "D " << mDecisionVar << ' ' << input_id[0] << ' '
                 << input_id[1];
}

void
SUBool::NnfGateDecision::CallVisit(NnfVisitor &visitor) const
{
   visitor.VisitDecision(
       std::dynamic_pointer_cast<const NnfGateDecision>(shared_from_this()));
}

void
SUBool::NnfGateDecision::CallVisitMut(NnfVisitorMut &visitor)
{
   visitor.VisitDecision(
       std::dynamic_pointer_cast<NnfGateDecision>(shared_from_this()));
}
