/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfgate.h
 * NNF gates (various inner nodes)
 */

#ifndef __NNFGATE_H
#define __NNFGATE_H

#include <functional>

#include "nnfnode.h"

namespace SUBool
{
   class NnfGate : public NnfNode
   {
      std::vector<tPointer> mInputs;

    protected:
      virtual std::ostream &WriteToStream(std::ostream &output,
                                          tValueCache<size_t> &node_id,
                                          size_t &next_id) const override;
      virtual std::ostream &
      WriteToStreamWithId(std::ostream &output,
                          const std::vector<size_t> &input_id) const = 0;

      NnfGate(tPointer a) : mInputs{std::move(a)} {}
      NnfGate(tPointer a, tPointer b) : mInputs{std::move(a), std::move(b)} {}
      NnfGate(std::vector<tPointer> inputs) : mInputs(std::move(inputs)) {}
      NnfGate(std::initializer_list<tPointer> inputs)
          : mInputs(std::move(inputs))
      {
      }

      const tPointer &
      Input(size_t i) const
      {
         return mInputs[i];
      }

      const tConstPointer
      ConstInput(size_t i) const
      {
         return mInputs[i];
      }

    public:
      virtual void AddChild(tPointer child);

      virtual ~NnfGate() = default;

      virtual iterator begin() override;
      virtual iterator end() override;
      virtual const_iterator begin() const override;
      virtual const_iterator end() const override;
      virtual const_iterator cbegin() const override;
      virtual const_iterator cend() const override;
      virtual size_t InputSize() const override;

      virtual bool
      IsLeaf() const override
      {
         return false;
      }
      virtual bool
      IsGate() const override
      {
         return true;
      }
   };

   class NnfGateAnd : public NnfGate
   {
    protected:
      virtual bool Value(const Assignment &assignment,
                         tValueCache<bool> &cached_values) const override;
      virtual PartialValue
      Value(const PartialAssignment &assignment,
            tValueCache<PartialValue> &cached_values) const override;

      virtual std::ostream &
      WriteToStreamWithId(std::ostream &output,
                          const std::vector<size_t> &input_id) const override;

    public:
      template <typename... Args> NnfGateAnd(Args... args) : NnfGate(args...) {}

      virtual ~NnfGateAnd() = default;

      virtual void CallVisit(NnfVisitor &visitor) const override;
      virtual void CallVisitMut(NnfVisitorMut &visitor) override;

      virtual Type
      NodeType() const override
      {
         return Type::And;
      }
   };

   class NnfGateOr : public NnfGate
   {
      unsigned mSplitVar;

    protected:
      virtual bool Value(const Assignment &assignment,
                         tValueCache<bool> &cached_values) const override;
      virtual PartialValue
      Value(const PartialAssignment &assignment,
            tValueCache<PartialValue> &cached_values) const override;

      virtual std::ostream &
      WriteToStreamWithId(std::ostream &output,
                          const std::vector<size_t> &input_id) const override;

    public:
      template <typename... Args>
      NnfGateOr(Args... args) : NnfGate(args...), mSplitVar(0)
      {
      }

      NnfGateOr(unsigned split_var, tPointer left, tPointer right)
          : NnfGate(left, right), mSplitVar(split_var)
      {
      }

      virtual ~NnfGateOr() = default;

      virtual void CallVisit(NnfVisitor &visitor) const override;
      virtual void CallVisitMut(NnfVisitorMut &visitor) override;

      unsigned
      SplitVar() const
      {
         return mSplitVar;
      }

      virtual Type
      NodeType() const override
      {
         return Type::Or;
      }

      virtual void AddChild(tPointer child) override;
   };

   class NnfGateDecision : public NnfGate
   {
      unsigned mDecisionVar;

    protected:
      virtual bool Value(const Assignment &assignment,
                         tValueCache<bool> &cached_values) const override;
      virtual PartialValue
      Value(const PartialAssignment &assignment,
            tValueCache<PartialValue> &cached_values) const override;

      virtual std::ostream &
      WriteToStreamWithId(std::ostream &output,
                          const std::vector<size_t> &input_id) const override;

      virtual void AddChild(tPointer child) override;

    public:
      NnfGateDecision(unsigned decision_var, tPointer true_input,
                      tPointer false_input)
          : NnfGate(true_input, false_input), mDecisionVar(decision_var)
      {
      }

      virtual void CallVisit(NnfVisitor &visitor) const override;
      virtual void CallVisitMut(NnfVisitorMut &visitor) override;

      unsigned
      DecisionVar() const
      {
         return mDecisionVar;
      }

      const tConstPointer
      TrueInput() const
      {
         return ConstInput(0);
      }

      const tConstPointer
      FalseInput() const
      {
         return ConstInput(1);
      }

      virtual ~NnfGateDecision() = default;

      virtual Type
      NodeType() const override
      {
         return Type::Decision;
      }
   };
} // namespace SUBool

#endif
