/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfleaf.cpp
 * NNF leaf node (various kinds)
 */

#include "nnfleaf.h"
#include "nnfvisit.h"

std::vector<SUBool::NnfNode::tPointer> SUBool::NnfLeaf::kEmptyInputs;

SUBool::NnfNode::iterator
SUBool::NnfLeaf::begin()
{
   return kEmptyInputs.begin();
}

SUBool::NnfNode::iterator
SUBool::NnfLeaf::end()
{
   return kEmptyInputs.end();
}

SUBool::NnfNode::const_iterator
SUBool::NnfLeaf::begin() const
{
   return kEmptyInputs.begin();
}

SUBool::NnfNode::const_iterator
SUBool::NnfLeaf::cbegin() const
{
   return kEmptyInputs.begin();
}

SUBool::NnfNode::const_iterator
SUBool::NnfLeaf::end() const
{
   return kEmptyInputs.begin();
}

SUBool::NnfNode::const_iterator
SUBool::NnfLeaf::cend() const
{
   return kEmptyInputs.begin();
}

size_t
SUBool::NnfLeaf::InputSize() const
{
   return 0;
}

bool
SUBool::NnfLeaf::Value(const Assignment &assignment,
                       tValueCache<bool> &node_values) const
{
   auto val = CachedValue(node_values);
   if (val)
   {
      return *val;
   }
   auto new_val = Eval(assignment);
   CacheValue(node_values, new_val);
   return new_val;
}

SUBool::PartialValue
SUBool::NnfLeaf::Value(const PartialAssignment &assignment,
                       tValueCache<PartialValue> &node_values) const
{
   auto val = CachedValue(node_values);
   if (val)
   {
      return *val;
   }
   auto new_val = Eval(assignment);
   CacheValue(node_values, new_val);
   return new_val;
}

std::ostream &
SUBool::NnfLeaf::WriteToStream(std::ostream &output,
                               tValueCache<size_t> &node_id,
                               size_t &next_id) const
{
   if (CachedValue(node_id))
   {
      return output;
   }
   CacheValue(node_id, next_id++);
   return WriteToStream(output);
}

bool
SUBool::NnfLiteral::Eval(const Assignment &assignment) const
{
   return mLiteral.Value(assignment);
}

SUBool::PartialValue
SUBool::NnfLiteral::Eval(const PartialAssignment &assignment) const
{
   return mLiteral.Value(assignment);
}

std::ostream &
SUBool::NnfLiteral::WriteToStream(std::ostream &output) const
{
   output << "L " << mLiteral;
   return output;
}

void
SUBool::NnfLiteral::CallVisit(NnfVisitor &visitor) const
{
   visitor.VisitLiteral(
       std::dynamic_pointer_cast<const NnfLiteral>(shared_from_this()));
}

void
SUBool::NnfLiteral::CallVisitMut(NnfVisitorMut &visitor)
{
   visitor.VisitLiteral(
       std::dynamic_pointer_cast<NnfLiteral>(shared_from_this()));
}

unsigned
SUBool::NnfLiteral::MaxVariable() const
{
   return mLiteral.Variable();
}

std::ostream &
SUBool::NnfTrue::WriteToStream(std::ostream &output) const
{
   output << "A 0";
   return output;
}

void
SUBool::NnfTrue::CallVisit(NnfVisitor &visitor) const
{
   visitor.VisitTrue(
       std::dynamic_pointer_cast<const NnfTrue>(shared_from_this()));
}

void
SUBool::NnfTrue::CallVisitMut(NnfVisitorMut &visitor)
{
   visitor.VisitTrue(std::dynamic_pointer_cast<NnfTrue>(shared_from_this()));
}

std::ostream &
SUBool::NnfFalse::WriteToStream(std::ostream &output) const
{
   output << "O 0 0";
   return output;
}

void
SUBool::NnfFalse::CallVisit(NnfVisitor &visitor) const
{
   visitor.VisitFalse(
       std::dynamic_pointer_cast<const NnfFalse>(shared_from_this()));
}

void
SUBool::NnfFalse::CallVisitMut(NnfVisitorMut &visitor)
{
   visitor.VisitFalse(std::dynamic_pointer_cast<NnfFalse>(shared_from_this()));
}

bool
SUBool::NnfLeafCNF::Eval(const Assignment &assignment) const
{
   return mCnf.Value(assignment);
}

SUBool::PartialValue
SUBool::NnfLeafCNF::Eval(const PartialAssignment &assignment) const
{
   return mCnf.Value(assignment);
}

std::ostream &
SUBool::NnfLeafCNF::WriteToStream(std::ostream &output) const
{
   output << mCnf;
   return output;
}

void
SUBool::NnfLeafCNF::CallVisit(NnfVisitor &visitor) const
{
   visitor.VisitLeafCNF(
       std::dynamic_pointer_cast<const NnfLeafCNF>(shared_from_this()));
}

void
SUBool::NnfLeafCNF::CallVisitMut(NnfVisitorMut &visitor)
{
   visitor.VisitLeafCNF(
       std::dynamic_pointer_cast<NnfLeafCNF>(shared_from_this()));
}

unsigned
SUBool::NnfLeafCNF::MaxVariable() const
{
   return mCnf.MaxVariable();
}

const std::vector<unsigned> &
SUBool::NnfLeafCNF::Variables() const
{
   if (!mVariables)
   {
      auto vars = mCnf.Variables();
      mVariables = std::vector<unsigned>(vars.begin(), vars.end());
   }
   return *mVariables;
}
