/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfleaf.h
 * NNF leaf node (various kinds)
 */

#ifndef __NNFLEAF_H
#define __NNFLEAF_H

#include <utility>

#include "literal.h"
#include "nnfnode.h"
#include "normalform.h"

namespace SUBool
{
   class NnfLeaf : public NnfNode
   {
      static std::vector<tPointer> kEmptyInputs;

    protected:
      virtual bool Eval(const Assignment &assignment) const = 0;
      virtual PartialValue Eval(const PartialAssignment &assignment) const = 0;

      virtual bool Value(const Assignment &assignment,
                         tValueCache<bool> &cached_values) const override;
      virtual PartialValue
      Value(const PartialAssignment &assignment,
            tValueCache<PartialValue> &cached_values) const override;

      virtual std::ostream &WriteToStream(std::ostream &output,
                                          tValueCache<size_t> &node_id,
                                          size_t &next_id) const override;
      virtual std::ostream &
      WriteToStream(std::ostream &output) const override = 0;

    public:
      virtual ~NnfLeaf() = default;

      virtual iterator begin() override;
      virtual iterator end() override;
      virtual const_iterator begin() const override;
      virtual const_iterator end() const override;
      virtual const_iterator cbegin() const override;
      virtual const_iterator cend() const override;
      virtual size_t InputSize() const override;

      virtual unsigned MaxVariable() const = 0;

      virtual bool
      IsLeaf() const override
      {
         return true;
      }
      virtual bool
      IsGate() const override
      {
         return false;
      }
   };

   class NnfLiteral : public NnfLeaf
   {
      Literal mLiteral;

    protected:
      virtual bool Eval(const Assignment &assignment) const override;
      virtual PartialValue
      Eval(const PartialAssignment &assignment) const override;
      virtual std::ostream &WriteToStream(std::ostream &output) const override;

    public:
      template <typename... Args>
      NnfLiteral(Args... args) : mLiteral(std::forward<Args>(args)...)
      {
      }

      virtual ~NnfLiteral() = default;
      virtual unsigned MaxVariable() const override;

      virtual void CallVisit(NnfVisitor &visitor) const override;
      virtual void CallVisitMut(NnfVisitorMut &visitor) override;

      const Literal &
      GetLiteral() const
      {
         return mLiteral;
      }

      virtual Type
      NodeType() const override
      {
         return Type::Literal;
      }
   };

   template <bool val> class NnfConstant : public NnfLeaf
   {
    protected:
      virtual bool Eval(const Assignment &assignment) const override;
      virtual PartialValue
      Eval(const PartialAssignment &assignment) const override;

    public:
      virtual ~NnfConstant() = default;
      virtual unsigned
      MaxVariable() const override
      {
         return 0;
      }
   };

   class NnfTrue : public NnfConstant<true>
   {
    protected:
      virtual std::ostream &WriteToStream(std::ostream &output) const override;

    public:
      virtual ~NnfTrue() = default;

      virtual void CallVisit(NnfVisitor &visitor) const override;
      virtual void CallVisitMut(NnfVisitorMut &visitor) override;

      virtual Type
      NodeType() const override
      {
         return Type::True;
      }
   };

   class NnfFalse : public NnfConstant<false>
   {
    protected:
      virtual std::ostream &WriteToStream(std::ostream &output) const override;

    public:
      virtual ~NnfFalse() = default;

      virtual void CallVisit(NnfVisitor &visitor) const override;
      virtual void CallVisitMut(NnfVisitorMut &visitor) override;

      virtual Type
      NodeType() const override
      {
         return Type::False;
      }
   };

   class NnfLeafCNF : public NnfLeaf
   {
      CNF mCnf;
      mutable std::optional<std::vector<unsigned>> mVariables;

    protected:
      virtual bool Eval(const Assignment &assignment) const override;
      virtual PartialValue
      Eval(const PartialAssignment &assignment) const override;
      virtual std::ostream &WriteToStream(std::ostream &output) const override;

    public:
      template <typename... Args>
      NnfLeafCNF(Args... args) : mCnf(std::forward<Args>(args)...), mVariables()
      {
      }
      virtual ~NnfLeafCNF() = default;
      virtual unsigned MaxVariable() const override;
      const std::vector<unsigned> &Variables() const;

      virtual void CallVisit(NnfVisitor &visitor) const override;
      virtual void CallVisitMut(NnfVisitorMut &visitor) override;

      virtual Type
      NodeType() const override
      {
         return Type::Cnf;
      }
      const CNF &
      Cnf() const
      {
         return mCnf;
      }
   };
} // namespace SUBool

// NnfConstant implementation
template <bool val>
bool
SUBool::NnfConstant<val>::Eval(const Assignment & /*assignment*/) const
{
   return val;
}

template <bool val>
SUBool::PartialValue
SUBool::NnfConstant<val>::Eval(const PartialAssignment & /*assignment*/) const
{
   return partial_value_of_bool(val);
}

#endif
