/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfnode.cpp
 * NNF node (various kinds)
 */

#include "nnfnode.h"
#include "literal.h"

#include "nnfgate.h"
#include "nnfleaf.h"
#include "nnfvisit.h"

SUBool::NnfNode::tPointer SUBool::NnfNode::sTrueNode;
SUBool::NnfNode::tPointer SUBool::NnfNode::sFalseNode;

bool
SUBool::NnfNode::Value(const Assignment &assignment) const
{
   tValueCache<bool> values;
   return Value(assignment, values);
}

SUBool::PartialValue
SUBool::NnfNode::Value(const PartialAssignment &assignment) const
{
   tValueCache<PartialValue> values;
   return Value(assignment, values);
}

std::ostream &
SUBool::NnfNode::WriteToStream(std::ostream &output) const
{
   tValueCache<size_t> node_id;
   size_t id = 0;
   return WriteToStream(output, node_id, id);
}

void
SUBool::NnfNode::AcceptPostorder(
    NnfVisitor &visitor, std::unordered_set<tConstPointer> &visited) const
{
   auto shared_this = shared_from_this();
   if (visited.count(shared_this) == 0)
   {
      for (auto &input : *this)
      {
         input->AcceptPostorder(visitor, visited);
      }
      CallVisit(visitor);
      visited.insert(shared_from_this());
   }
}

void
SUBool::NnfNode::AcceptPostorder(NnfVisitor &visitor) const
{
   std::unordered_set<tConstPointer> visited;
   AcceptPostorder(visitor, visited);
}

void
SUBool::NnfNode::AcceptPreorder(
    NnfVisitor &visitor, std::unordered_set<tConstPointer> &visited) const
{
   auto shared_this = shared_from_this();
   if (visited.count(shared_this) == 0)
   {
      CallVisit(visitor);
      for (auto &input : *this)
      {
         input->AcceptPreorder(visitor, visited);
      }
      visited.insert(shared_from_this());
   }
}

void
SUBool::NnfNode::AcceptPreorderMut(NnfVisitorMut &visitor,
                                   std::unordered_set<tConstPointer> &visited)
{
   auto shared_this = shared_from_this();
   if (visited.count(shared_this) == 0)
   {
      CallVisitMut(visitor);
      for (auto &input : *this)
      {
         input->AcceptPreorderMut(visitor, visited);
      }
      visited.insert(shared_from_this());
   }
}

void
SUBool::NnfNode::AcceptPostorderMut(NnfVisitorMut &visitor,
                                    std::unordered_set<tConstPointer> &visited)
{
   auto shared_this = shared_from_this();
   if (visited.count(shared_this) == 0)
   {
      for (auto &input : *this)
      {
         input->AcceptPostorderMut(visitor, visited);
      }
      CallVisitMut(visitor);
      visited.insert(shared_from_this());
   }
}

void
SUBool::NnfNode::AcceptPreorder(NnfVisitor &visitor) const
{
   std::unordered_set<tConstPointer> visited;
   AcceptPreorder(visitor, visited);
}

void
SUBool::NnfNode::AcceptPreorderMut(NnfVisitorMut &visitor)
{
   std::unordered_set<tConstPointer> visited;
   AcceptPreorderMut(visitor, visited);
}

void
SUBool::NnfNode::AcceptPostorderMut(NnfVisitorMut &visitor)
{
   std::unordered_set<tConstPointer> visited;
   AcceptPostorderMut(visitor, visited);
}

bool
SUBool::NnfNode::operator()(const Assignment &assignment) const
{
   return Value(assignment);
}

SUBool::PartialValue
SUBool::NnfNode::operator()(const PartialAssignment &assignment) const
{
   return Value(assignment);
}

std::ostream &
SUBool::operator<<(std::ostream &output, const NnfNode &node)
{
   return node.WriteToStream(output);
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeLiteral(Literal lit)
{
   return tPointer(new NnfLiteral(lit));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeLiteral(unsigned var, bool positive)
{
   return tPointer(new NnfLiteral(var, positive));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeAnd(tPointer a, tPointer b)
{
   return tPointer(new NnfGateAnd(a, b));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeAnd(std::vector<tPointer> inputs)
{
   return tPointer(new NnfGateAnd(inputs));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeOr(tPointer a, tPointer b)
{
   return tPointer(new NnfGateOr(a, b));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeOr(unsigned split_var, tPointer left, tPointer right)
{
   return tPointer(new NnfGateOr(split_var, left, right));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeOr(std::vector<tPointer> inputs)
{
   return tPointer(new NnfGateOr(inputs));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeDecision(unsigned var, tPointer true_input,
                              tPointer false_input)
{
   return tPointer(new NnfGateDecision(var, true_input, false_input));
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeTrue()
{
   if (sTrueNode == nullptr)
   {
      sTrueNode = tPointer(new NnfTrue());
   }
   return sTrueNode;
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeFalse()
{
   if (sFalseNode == nullptr)
   {
      sFalseNode = tPointer(new NnfFalse());
   }
   return sFalseNode;
}

SUBool::NnfNode::tPointer
SUBool::NnfNode::MakeLeafCNF(CNF cnf)
{
   return tPointer(new NnfLeafCNF(std::move(cnf)));
}
