/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfnode.h
 * NNF node (base class)
 */

#ifndef __NNFNODE_H
#define __NNFNODE_H

#include <iostream>
#include <memory>
#include <unordered_map>
#include <unordered_set>

#include "assign.h"
#include "literal.h"
#include "nnfvisit.h"
#include "normalform.h"

namespace SUBool
{
#pragma GCC diagnostic push
   // enable_shared_from_this has not virtual destructor by design
#pragma GCC diagnostic ignored "-Wnon-virtual-dtor"
   class NnfNode : public std::enable_shared_from_this<NnfNode>
   {
    public:
      using tPointer = std::shared_ptr<NnfNode>;
      using tConstPointer = std::shared_ptr<const NnfNode>;
      using tWeakPointer = std::weak_ptr<NnfNode>;
      using tConstWeakPointer = std::weak_ptr<const NnfNode>;
      using iterator = std::vector<tPointer>::iterator;
      using const_iterator = std::vector<tPointer>::const_iterator;

      template <class T>
      using tValueCache = typename std::unordered_map<tConstPointer, T>;

      enum class Type
      {
         Literal,
         And,
         Or,
         Decision,
         True,
         False,
         Cnf
      };

    private:
      static tPointer sTrueNode;
      static tPointer sFalseNode;

    protected:
      NnfNode() {}

    public:
      template <class T>
      const T &CachedValueConstRef(const tValueCache<T> &cached_values) const;
      template <class T>
      std::optional<T> CachedValue(const tValueCache<T> &cached_values) const;
      template <class T> T &CachedValueRef(tValueCache<T> &cached_values) const;
      template <class T>
      void CacheValue(tValueCache<T> &cached_values, T value) const;

      virtual bool Value(const Assignment &assignment,
                         tValueCache<bool> &cached_values) const = 0;
      virtual PartialValue
      Value(const PartialAssignment &assignment,
            tValueCache<PartialValue> &cached_values) const = 0;

      virtual std::ostream &WriteToStream(std::ostream &output,
                                          tValueCache<size_t> &node_id,
                                          size_t &next_id) const = 0;

      virtual void
      AcceptPostorder(NnfVisitor &visitor,
                      std::unordered_set<tConstPointer> &visited) const;
      virtual void
      AcceptPreorder(NnfVisitor &visitor,
                     std::unordered_set<tConstPointer> &visited) const;

      virtual void
      AcceptPreorderMut(NnfVisitorMut &visitor,
                        std::unordered_set<tConstPointer> &visited);

      virtual void
      AcceptPostorderMut(NnfVisitorMut &visitor,
                         std::unordered_set<tConstPointer> &visited);

      virtual void CallVisit(NnfVisitor &visitor) const = 0;
      virtual void CallVisitMut(NnfVisitorMut &visitor) = 0;

      virtual iterator begin() = 0;
      virtual iterator end() = 0;
      virtual const_iterator begin() const = 0;
      virtual const_iterator end() const = 0;
      virtual const_iterator cbegin() const = 0;
      virtual const_iterator cend() const = 0;

      virtual size_t InputSize() const = 0;

      virtual Type NodeType() const = 0;
      virtual bool IsLeaf() const = 0;
      virtual bool IsGate() const = 0;

      virtual ~NnfNode() = default;

      bool Value(const Assignment &assignment) const;
      PartialValue Value(const PartialAssignment &assignment) const;

      virtual std::ostream &WriteToStream(std::ostream &output) const;

      virtual void AcceptPostorder(NnfVisitor &visitor) const;
      virtual void AcceptPostorderMut(NnfVisitorMut &visitor);
      virtual void AcceptPreorder(NnfVisitor &visitor) const;
      virtual void AcceptPreorderMut(NnfVisitorMut &visitor);

      bool operator()(const Assignment &assignment) const;
      PartialValue operator()(const PartialAssignment &assignment) const;

      static tPointer MakeLiteral(Literal lit);
      static tPointer MakeLiteral(unsigned var, bool positive);

      static tPointer MakeAnd(tPointer a, tPointer b);
      static tPointer MakeAnd(std::vector<tPointer> inputs);
      static tPointer MakeOr(tPointer a, tPointer b);
      static tPointer MakeOr(unsigned split_var, tPointer left, tPointer right);
      static tPointer MakeOr(std::vector<tPointer> inputs);
      static tPointer MakeDecision(unsigned var, tPointer true_input,
                                   tPointer false_input);
      static tPointer MakeTrue();
      static tPointer MakeFalse();
      static tPointer MakeLeafCNF(CNF cnf);
   };
#pragma GCC diagnostic pop

   std::ostream &operator<<(std::ostream &output, const NnfNode &node);
} // namespace SUBool

template <class T>
std::optional<T>
SUBool::NnfNode::CachedValue(const tValueCache<T> &cached_values) const
{
   try
   {
      return cached_values.at(shared_from_this());
   }
   catch (std::out_of_range &e)
   {
      return {};
   }
}

template <class T>
const T &
SUBool::NnfNode::CachedValueConstRef(const tValueCache<T> &cached_values) const
{
   return cached_values.at(shared_from_this());
}

template <class T>
T &
SUBool::NnfNode::CachedValueRef(tValueCache<T> &cached_values) const
{
   return cached_values.at(shared_from_this());
}

template <class T>
void
SUBool::NnfNode::CacheValue(tValueCache<T> &cached_values, T value) const
{
   cached_values[shared_from_this()] = std::move(value);
}
#endif
