/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file nnfvisit.h
 * NnfVisitor
 */

#ifndef __NNFVISIT_H
#define __NNFVISIT_H

#include <memory>
#include <type_traits>

//#include "nnfgate.h"
//#include "nnfleaf.h"

namespace SUBool
{
   class NnfNode;
   class NnfLiteral;
   class NnfGateAnd;
   class NnfGateOr;
   class NnfGateDecision;
   class NnfTrue;
   class NnfFalse;
   class NnfLeafCNF;

   template <typename T, typename P> struct preserve_const_ptr
   {
      using type = std::shared_ptr<
          typename std::conditional<std::is_const<T>::value, const P, P>::type>;
   };

   template <typename T, typename P> struct preserve_const
   {
      using type =
          typename std::conditional<std::is_const<T>::value, const P, P>::type;
   };

   template <typename T> class NnfVisitorTemplate
   {
    public:
      using LiteralPtr = typename preserve_const_ptr<T, NnfLiteral>::type;
      using GateAndPtr = typename preserve_const_ptr<T, NnfGateAnd>::type;
      using GateOrPtr = typename preserve_const_ptr<T, NnfGateOr>::type;
      using GateDecisionPtr =
          typename preserve_const_ptr<T, NnfGateDecision>::type;
      using TruePtr = typename preserve_const_ptr<T, NnfTrue>::type;
      using FalsePtr = typename preserve_const_ptr<T, NnfFalse>::type;
      using CnfPtr = typename preserve_const_ptr<T, NnfLeafCNF>::type;

      virtual ~NnfVisitorTemplate() = default;

      virtual void VisitLiteral(const LiteralPtr &node) = 0;
      virtual void VisitAnd(const GateAndPtr &node) = 0;
      virtual void VisitOr(const GateOrPtr &node) = 0;
      virtual void VisitDecision(const GateDecisionPtr &node) = 0;
      virtual void VisitTrue(const TruePtr &node) = 0;
      virtual void VisitFalse(const FalsePtr &node) = 0;
      // VisitLeafCNF is empty, because in most cases this is not relevant
      virtual void VisitLeafCNF(const CnfPtr & /*node*/){};
   };

   template <typename T>
   class NnfSimpleVisitorTemplate : public NnfVisitorTemplate<T>
   {
    public:
      using Base = NnfVisitorTemplate<T>;
      using NodePtr = typename preserve_const_ptr<T, NnfNode>::type;

      virtual void VisitNode(const NodePtr &node) = 0;

      virtual void
      VisitLiteral(const typename Base::LiteralPtr &node) override
      {
         VisitNode(std::static_pointer_cast<
                   typename preserve_const<T, NnfNode>::type>(node));
      }
      virtual void
      VisitAnd(const typename Base::GateAndPtr &node) override
      {
         VisitNode(std::static_pointer_cast<
                   typename preserve_const<T, NnfNode>::type>(node));
      }
      virtual void
      VisitOr(const typename Base::GateOrPtr &node) override
      {
         VisitNode(std::static_pointer_cast<
                   typename preserve_const<T, NnfNode>::type>(node));
      }
      virtual void
      VisitDecision(const typename Base::GateDecisionPtr &node) override
      {
         VisitNode(std::static_pointer_cast<
                   typename preserve_const<T, NnfNode>::type>(node));
      }
      virtual void
      VisitTrue(const typename Base::TruePtr &node) override
      {
         VisitNode(std::static_pointer_cast<
                   typename preserve_const<T, NnfNode>::type>(node));
      }
      virtual void
      VisitFalse(const typename Base::FalsePtr &node) override
      {
         VisitNode(std::static_pointer_cast<
                   typename preserve_const<T, NnfNode>::type>(node));
      }
      virtual void
      VisitLeafCNF(const typename Base::CnfPtr &node) override
      {
         VisitNode(std::static_pointer_cast<
                   typename preserve_const<T, NnfNode>::type>(node));
      }
   };

   template <typename T, typename F>
   class NnfFuncVisitorTemplate : public NnfSimpleVisitorTemplate<T>
   {
      F mFunction;

    public:
      using Base = NnfSimpleVisitorTemplate<T>;

      NnfFuncVisitorTemplate(F f) : mFunction{f} {}

      virtual void
      VisitNode(const typename Base::NodePtr &node)
      {
         mFunction(node);
      }
   };

   using NnfVisitor = NnfVisitorTemplate<const NnfNode>;
   using NnfVisitorMut = NnfVisitorTemplate<NnfNode>;
   using NnfSimpleVisitor = NnfSimpleVisitorTemplate<const NnfNode>;
   using NnfSimpleVisitorMut = NnfSimpleVisitorTemplate<NnfNode>;

   template <typename F>
   class NnfFuncVisitor : public NnfFuncVisitorTemplate<const NnfNode, F>
   {
    public:
      NnfFuncVisitor(F f) : NnfFuncVisitorTemplate<const NnfNode, F>(f) {}
   };

   template <typename F>
   class NnfFuncVisitorMut : public NnfFuncVisitorTemplate<NnfNode, F>
   {
    public:
      NnfFuncVisitorMut(F f) : NnfFuncVisitorTemplate<NnfNode, F>(f) {}
   };
} // namespace SUBool

#endif
