/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file normalform.h
 * Defines template NormalForm which is a general structure
 * representing a set of NFLiteralSet objects.
 * Also defines types CNF and DNF (special normal forms).
 */

#ifndef __NORMALFORM_H
#define __NORMALFORM_H

#include <set>

#include "nfliteralset.h"
#include "util.h"

namespace SUBool
{
   struct NFSize;

   template <class ElementType, bool DETERMINING_VALUE> class NormalForm
   {
    public:
      using tElements = typename std::vector<ElementType>;
      using iterator = typename tElements::iterator;
      using const_iterator = typename tElements::const_iterator;

    private:
      unsigned mMaxVariable;
      tElements mElements;
      static const PartialValue cDeterminingPartialValue =
          partial_value_of_bool(DETERMINING_VALUE);
      NormalForm(unsigned max_variable, tElements elements);
      void UpdateMaxVariable();

    public:
      NormalForm();
      NormalForm(ElementType element);
      NormalForm(tElements elements);
      NormalForm(std::initializer_list<ElementType> elements);
      template <class InputIterator>
      NormalForm(InputIterator begin, InputIterator end);

      NormalForm(
          const NormalForm<ElementType, DETERMINING_VALUE> &lset) = default;
      NormalForm(NormalForm<ElementType, DETERMINING_VALUE> &&lset) = default;
      ~NormalForm() = default;
      NormalForm<ElementType, DETERMINING_VALUE> &operator=(
          const NormalForm<ElementType, DETERMINING_VALUE> &lset) = default;
      NormalForm<ElementType, DETERMINING_VALUE> &operator=(
          NormalForm<ElementType, DETERMINING_VALUE> &&lset) = default;

      static bool DeterminingValue();

      unsigned MaxVariable() const;
      size_t Size() const;
      size_t Length() const;
      NFSize GetNFSize() const;
      bool Empty() const;
      bool Value(const Assignment &assignment) const;
      PartialValue Value(const PartialAssignment &assignment) const;
      PartialValue Apply(const PartialAssignment &assignment);
      PartialValue Apply(const PartialAssignment &assignment,
          NormalForm<ElementType, DETERMINING_VALUE> *new_nf) const;

      std::set<unsigned> Variables() const;

      std::ostream &WriteToStream(std::ostream &output) const;
      void ApplyRenaming(const Assignment &renaming);

      friend std::ostream &
      operator<<(std::ostream &output,
          const NormalForm<ElementType, DETERMINING_VALUE> &nf)
      {
         return nf.WriteToStream(output);
      }

      const ElementType &operator[](size_t index) const;
      const ElementType &At(size_t index) const;
      const_iterator begin() const;
      const_iterator end() const;
      const_iterator cbegin() const;
      const_iterator cend() const;

      const tElements &Elements() const &;
      tElements Elements() &&;

      void Sort();
      void SortUnique();
      template <class Compare> void Sort(Compare comp);

      void Set(size_t index, ElementType element);
      void PushBack(ElementType element);
      void PopBack(); // Does not update maxvariable
      const ElementType &Front() const;
      const ElementType &Back() const;
      void Append(NormalForm<ElementType, DETERMINING_VALUE> nf);
      template <typename InputIterator>
      void Append(InputIterator begin, InputIterator end);
      bool AddWithoutSubsumptions(const ElementType &element);
      bool AddIfNotPresent(const ElementType &element);
      void RemoveDuplicities();
      void RemoveSubsumptions();
      bool SubsumedByAny(const ElementType &element) const;
      // Updates every clause c o f(c)
      template <class Func> void Transform(Func f);
      template <class UnaryPredicate> void RemoveIf(UnaryPredicate pred);
      /* Remove an element at given index. It is O(1) procedure which however
       * modifies the order of elements. In particular, all elements before the
       * ith ones are the same, all after ith except the last one are the same
       * and the last one is moved to ith position.
       */
      void RemoveAt(size_t i);
      template <class Pred> void RemoveIndices(Pred pred);
      void Clear();

      bool operator==(
          const NormalForm<ElementType, DETERMINING_VALUE> &nf) const;
      bool operator!=(
          const NormalForm<ElementType, DETERMINING_VALUE> &nf) const;
   };

   using CNF = NormalForm<Clause, false>;
   using DNF = NormalForm<Term, true>;

   template <class T> struct AnnotatedNF
   {
      T nf;
      std::string dimacs_annotation;

      AnnotatedNF() : nf(), dimacs_annotation() {}
      AnnotatedNF(T a_nf, std::string annotation)
          : nf(std::move(a_nf)), dimacs_annotation(std::move(annotation))
      {
      }
   };

   template <class T>
   AnnotatedNF<T>
   make_annotated_nf(T nf, std::string annotation)
   {
      return AnnotatedNF<T>(std::move(nf), std::move(annotation));
   }

   template <class T>
   std::ostream &operator<<(std::ostream &output, const AnnotatedNF<T> &nf);

   using AnnotatedCNF = AnnotatedNF<CNF>;
   using AnnotatedDNF = AnnotatedNF<DNF>;

   struct NFSize
   {
      unsigned max_variable{0};
      size_t size{0};
      size_t len{0};

      NFSize
      operator+(NFSize s) const
      {
         s.max_variable += max_variable;
         s.size += size;
         s.len += len;
         return s;
      }

      NFSize
      operator+=(NFSize s)
      {
         max_variable += s.max_variable;
         size += s.size;
         len += s.len;
         return *this;
      }

      NFSize
      operator*(unsigned l) const
      {
         return {max_variable * l, size * l, len * l};
      }

      bool
      operator==(const NFSize &b)
      {
         return max_variable == b.max_variable && size == b.size
                && len == b.len;
      }

      bool
      operator!=(const NFSize &b)
      {
         return !(*this == b);
      }
   };

   inline std::ostream &
   operator<<(std::ostream &os, const NFSize s)
   {
      return os << "{" << s.max_variable << ", " << s.size << ", " << s.len
                << "}";
   }
}; // namespace SUBool

#include "normalform_impl.h"

#endif
