/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file normalform_impl.h
 * Adds implementation of member functions of template NormalForm
 * which is a general structure representing a set of NFLiteralSet
 * objects. Also defines types CNF and DNF (special normal forms).
 */

#include <algorithm>

template <class ElementType, bool DETERMINING_VALUE>
const SUBool::PartialValue SUBool::NormalForm<ElementType,
    DETERMINING_VALUE>::cDeterminingPartialValue;

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::DeterminingValue()
{
   return DETERMINING_VALUE;
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::NormalForm()
    : mMaxVariable(0), mElements()
{
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::NormalForm(
    ElementType element)
    : mMaxVariable(element.MaxVariable()), mElements()
{
   mElements.push_back(std::move(element));
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::NormalForm(
    tElements elements)
    : mMaxVariable(0), mElements(std::move(elements))
{
   for (const auto &element : mElements)
   {
      mMaxVariable = std::max(mMaxVariable, element.MaxVariable());
   }
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::NormalForm(
    std::initializer_list<ElementType> elements)
    : mMaxVariable(0), mElements(std::move(elements))
{
   for (const auto &element : mElements)
   {
      mMaxVariable = std::max(mMaxVariable, element.MaxVariable());
   }
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::NormalForm(
    unsigned max_variable, tElements elements)
    : mMaxVariable(max_variable), mElements(std::move(elements))
{
}

template <class ElementType, bool DETERMINING_VALUE>
template <class InputIterator>
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::NormalForm(
    InputIterator begin, InputIterator end)
    : NormalForm()
{
   for (auto i = begin; i != end; ++i)
   {
      PushBack(*i);
   }
}

template <class ElementType, bool DETERMINING_VALUE>
unsigned
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::MaxVariable() const
{
   return mMaxVariable;
}

template <class ElementType, bool DETERMINING_VALUE>
size_t
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Size() const
{
   return mElements.size();
}

template <class ElementType, bool DETERMINING_VALUE>
size_t
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Length() const
{
   size_t len = 0;
   for (const auto &e : mElements)
   {
      len += e.Size();
   }
   return len;
}

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Empty() const
{
   return (Size() == 0);
}

template <class ElementType, bool DETERMINING_VALUE>
std::ostream &
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::WriteToStream(
    std::ostream &output) const
{
   output << "p cnf " << mMaxVariable << " " << mElements.size();
   for (const auto &element : mElements)
   {
      output << '\n' << element;
   }
   return output;
}

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Value(
    const Assignment &assignment) const
{
   for (const auto &element : mElements)
   {
      if (element.Value(assignment) == DETERMINING_VALUE)
      {
         return DETERMINING_VALUE;
      }
   }
   return !DETERMINING_VALUE;
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::PartialValue
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Value(
    const PartialAssignment &assignment) const
{
   bool has_undef = false;
   for (const auto &element : mElements)
   {
      switch (element.Value(assignment))
      {
      case cDeterminingPartialValue:
         return cDeterminingPartialValue;
      case PartialValue::UNDEF:
         has_undef = true;
         break;
      default:
         break;
      }
   }
   return (has_undef ? PartialValue::UNDEF : !cDeterminingPartialValue);
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::PartialValue
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Apply(
    const PartialAssignment &assignment,
    NormalForm<ElementType, DETERMINING_VALUE> *new_nf) const
{
   if (new_nf == nullptr)
   {
      throw NullPointerException("NormalForm::Apply", "new_nf == nullptr");
   }
   tElements new_elements;
   PartialValue val = !cDeterminingPartialValue;
   for (const auto &e : mElements)
   {
      ElementType new_e;
      switch (e.Apply(assignment, &new_e))
      {
      case cDeterminingPartialValue:
         val = cDeterminingPartialValue;
         new_elements.clear();
         new_elements.emplace_back();
         break;
      case PartialValue::UNDEF:
         val = PartialValue::UNDEF;
         new_elements.push_back(std::move(new_e));
         break;
      default:
         break;
      }
      if (val == cDeterminingPartialValue)
      {
         break;
      }
   }
   *new_nf =
       NormalForm<ElementType, DETERMINING_VALUE>(std::move(new_elements));
   return val;
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::PartialValue
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Apply(
    const PartialAssignment &assignment)
{
   NormalForm<ElementType, DETERMINING_VALUE> new_nf;
   PartialValue val = Apply(assignment, &new_nf);
   *this = std::move(new_nf);
   return val;
}

template <class ElementType, bool DETERMINING_VALUE>
const ElementType &
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::operator[](
    size_t index) const
{
   return mElements[index];
}

template <class ElementType, bool DETERMINING_VALUE>
const ElementType &
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::At(size_t index) const
{
   return mElements.at(index);
}

template <class ElementType, bool DETERMINING_VALUE>
typename SUBool::NormalForm<ElementType, DETERMINING_VALUE>::const_iterator
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::begin() const
{
   return mElements.begin();
}

template <class ElementType, bool DETERMINING_VALUE>
typename SUBool::NormalForm<ElementType, DETERMINING_VALUE>::const_iterator
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::end() const
{
   return mElements.end();
}

template <class ElementType, bool DETERMINING_VALUE>
typename SUBool::NormalForm<ElementType, DETERMINING_VALUE>::const_iterator
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::cbegin() const
{
   return mElements.cbegin();
}

template <class ElementType, bool DETERMINING_VALUE>
typename SUBool::NormalForm<ElementType, DETERMINING_VALUE>::const_iterator
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::cend() const
{
   return mElements.cend();
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Set(
    size_t index, ElementType element)
{
   if (index >= mElements.size())
   {
      throw SUBool::OutOfBoundsException(
          "NormalForm::Set", "index>=mElements.size()");
   }
   mElements[index] = std::move(element);
   if (mElements[index].MaxVariable() > mMaxVariable)
   {
      mMaxVariable = mElements[index].MaxVariable();
   }
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::PushBack(
    ElementType element)
{
   mMaxVariable = std::max(mMaxVariable, element.MaxVariable());
   mElements.push_back(std::move(element));
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::PopBack()
{
   mElements.pop_back();
}

template <class ElementType, bool DETERMINING_VALUE>
const ElementType &
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Front() const
{
   return mElements.front();
}

template <class ElementType, bool DETERMINING_VALUE>
const ElementType &
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Back() const
{
   return mElements.back();
}

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::AddWithoutSubsumptions(
    const ElementType &element)
{
   unsigned i = 0;
   while (i < mElements.size())
   {
      switch (element.SubsetCompare(mElements[i]))
      {
      case SetCompareResult::SUBSET:
         vector_erase(mElements, i);
         break;
      case SetCompareResult::EQUAL:
      case SetCompareResult::SUPERSET:
         return false;
      case SetCompareResult::INCOMPARABLE:
         ++i;
         break;
      }
   }
   PushBack(element);
   return true;
}

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::AddIfNotPresent(
    const ElementType &element)
{
   for (const auto &el : mElements)
   {
      if (el == element)
      {
         return false;
      }
   }
   PushBack(element);
   return true;
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::RemoveDuplicities()
{
   std::sort(mElements.begin(), mElements.end());
   auto it = std::unique(mElements.begin(), mElements.end());
   mElements.resize(std::distance(mElements.begin(), it));
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::RemoveSubsumptions()
{
   NormalForm<ElementType, DETERMINING_VALUE> nf;
   for (auto &element : *this)
   {
      nf.AddWithoutSubsumptions(std::move(element));
   }
   *this = std::move(nf);
}

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::SubsumedByAny(
    const ElementType &element) const
{
   return std::any_of(begin(), end(),
       [&element](const auto &a) { return a.Subsumes(element); });
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::RemoveAt(size_t i)
{
   vector_erase(mElements, i);
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Clear()
{
   mElements.clear();
   mMaxVariable = 0;
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Append(
    NormalForm<ElementType, DETERMINING_VALUE> nf)
{
   if (mElements.empty())
   {
      mElements = std::move(nf.mElements);
      mMaxVariable = nf.mMaxVariable;
   }
   else
   {
      mMaxVariable = std::max(mMaxVariable, nf.mMaxVariable);
      mElements.reserve(Size() + nf.Size());
      std::move(nf.mElements.begin(), nf.mElements.end(),
          std::back_inserter(mElements));
   }
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::ApplyRenaming(
    const Assignment &renaming)
{
   for (auto &element : mElements)
   {
      element.ApplyRenaming(renaming);
   }
}

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::operator==(
    const NormalForm<ElementType, DETERMINING_VALUE> &nf) const
{
   if (nf.MaxVariable() != mMaxVariable || nf.Size() != mElements.size())
   {
      return false;
   }
   for (unsigned i = 0; i < mElements.size(); ++i)
   {
      if (mElements[i] != nf[i])
      {
         return false;
      }
   }
   return true;
}

template <class ElementType, bool DETERMINING_VALUE>
bool
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::operator!=(
    const NormalForm<ElementType, DETERMINING_VALUE> &nf) const
{
   return !(*this == nf);
}

template <class ElementType, bool DETERMINING_VALUE>
std::set<unsigned>
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Variables() const
{
   std::set<unsigned> vars;
   for (const auto &elem : *this)
   {
      for (const auto &lit : elem)
      {
         vars.insert(lit.Variable());
      }
   }
   return vars;
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::UpdateMaxVariable()
{
   mMaxVariable = 0;
   for (const auto &elem : mElements)
   {
      mMaxVariable = std::max(mMaxVariable, elem.MaxVariable());
   }
}

template <class ElementType, bool DETERMINING_VALUE>
template <class UnaryPredicate>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::RemoveIf(
    UnaryPredicate pred)
{
   auto it = std::remove_if(mElements.begin(), mElements.end(), pred);
   mElements.resize(it - mElements.begin());
   UpdateMaxVariable();
}

template <class ElementType, bool DETERMINING_VALUE>
template <class Pred>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::RemoveIndices(Pred pred)
{
   vector_erase_indices(mElements, pred);
   UpdateMaxVariable();
}

template <class ElementType, bool DETERMINING_VALUE>
const typename SUBool::NormalForm<ElementType, DETERMINING_VALUE>::tElements &
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Elements() const &
{
   return mElements;
}

template <class ElementType, bool DETERMINING_VALUE>
typename SUBool::NormalForm<ElementType, DETERMINING_VALUE>::tElements
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Elements() &&
{
   return std::move(mElements);
}

template <class ElementType, bool DETERMINING_VALUE>
template <class Compare>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Sort(Compare comp)
{
   std::sort(mElements.begin(), mElements.end(), comp);
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Sort()
{
   std::sort(mElements.begin(), mElements.end());
}

template <class ElementType, bool DETERMINING_VALUE>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::SortUnique()
{
   sort_unique(mElements);
}

template <class T>
std::ostream &
SUBool::operator<<(std::ostream &output, const AnnotatedNF<T> &nf)
{
   if (!nf.dimacs_annotation.empty())
   {
      output << nf.dimacs_annotation << std::endl;
   }
   return output << nf.nf;
}

template <class ElementType, bool DETERMINING_VALUE>
template <class InputIterator>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Append(
    InputIterator begin, InputIterator end)
{
   for (auto i = begin; i != end; ++i)
   {
      PushBack(*i);
   }
}

template <class ElementType, bool DETERMINING_VALUE>
SUBool::NFSize
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::GetNFSize() const
{
   return {MaxVariable(), Size(), Length()};
}

template <class ElementType, bool DETERMINING_VALUE>
template <class Func>
void
SUBool::NormalForm<ElementType, DETERMINING_VALUE>::Transform(Func f)
{
   mMaxVariable = 0;
   std::transform(mElements.begin(), mElements.end(), mElements.begin(),
       [&f, this](const auto &e)
       {
          auto e2 = f(e);
          mMaxVariable = std::max(mMaxVariable, e2.MaxVariable());
          return e2;
       });
}
