/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file pc_opt.h
 * Common options of programs related to propagation completeness.
 */

#include "pc_opt.h"

const SUBool::EnumAnnotation<SUBool::PCCompAlgorithmOptions::ALGORITHM>
    SUBool::PCCompAlgorithmOptions::kAlgorithmAnnotation("ALGORITHM",
        {//
            {ALGORITHM::INCREMENT_PC, "incremental_pc",
                "heuristic algorithm based on incremental adding empowering "
                "implicates. The goal is an equivalent PC formula and SAT "
                "solver is used to "
                "check if the formula is PC."}, //
            {ALGORITHM::INCREMENT_URC, "incremental_urc",
                "heuristic algorithm based on incremental adding empowering "
                "implicates. The goal is an equivalent URC formula and SAT "
                "solver is used to "
                "check if the formula is URC."}, //
            {ALGORITHM::INCREMENT_URC_PC, "incremental_urc_pc",
                "heuristic algorithm based on incremental adding empowering "
                "implicates. The goal is URC formula and SAT solver is used to "
                "check if the formula is URC or PC in an interleaving way."}, //
            {ALGORITHM::LEARNING, "learning",
                "based on the algorithm for learning a Horn formula through "
                "equivalence and closure queries (Arias et al. 2015). The goal "
                "is an equivalent PC formula."}});

const SUBool::EnumAnnotation<SUBool::PCCheckGeneralOptions::GOAL>
    SUBool::PCCheckGeneralOptions::kGoalAnnotation(
        "GOAL", {
                    //
                    {GOAL::PC, "pc",
                        "check if the formula is propagation complete."}, //
                    {GOAL::URC, "urc",
                        "check if the formula is unit refutation complete."} //
                });

auto
SUBool::minimize_method_opt() -> EnumOption<MINIMIZE_METHOD>
{
   return make_enum_option(kMinimizeMethodAnnotation, "minimize-method",
       "Minimization method", kDefaultMinimizeMethod);
}

SUBool::PCCompAlgorithmOptions::PCCompAlgorithmOptions()
    : OptionsGroup("Algorithm Selection"),
      mAlgorithmOption(kAlgorithmAnnotation, "algorithm,a",
          "Algorithm used for compilation", kDefaultAlgorithm)
{
   Add(mAlgorithmOption);
}

bool
SUBool::PCCompAlgorithmOptions::GetValues()
{
   return mAlgorithmOption.GetValue(mAlgorithm);
}

SUBool::PCLearnCompOptions::PCLearnCompOptions()
    : OptionsGroup("Learning Algorithm Options"),
      mInitialNegativesOption(PCLearnComp::kInitialNegativesAnnotation,
          "initial-negatives",
          "How the initial negative examples in the learning algorithm "
          "should be produced. The option specifies how the negatives for each "
          "clause in the preprocessed CNF are created.",
          kDefaultInitialNegatives),
      mCandidatesFromNegativePolicyOption(
          PCLearnComp::kCandidatesFromNegativePolicyAnnotation,
          "candidates-from-negative",
          "Whenever a negative example is added to the hypothesis, the "
          "corresponding clauses are added to the CNF hypothesis if they are "
          "empowering. In this case, the clause may be used to obtain further "
          "negative examples to keep the CNF hypothesis synchronized with the "
          "implicational representation of the hypothesis. This option allows "
          "to set a policy of how this extraction should be done.",
          kDefaultCandidatesFromNegativePolicy)
{
   Add(mInitialNegativesOption);
   Add(mCandidatesFromNegativePolicyOption);
   AddOptions()("use-cnf-hypothesis-for-empowerment",
       po::bool_switch(&mUseCNFHypothesisForEmpowerment),
       "If present, then the CNF hypothesis is used to encode empowerment in "
       "the SAT based PC checks within the learning algorithm. If this option "
       "is not present, then the default approach is to use the implicational "
       "system based on the negative examples for encoding empowerment. (It is "
       "suggested to combine this option with '--candidates-from-negative "
       "none'.)");
}

bool
SUBool::PCLearnCompOptions::GetValues()
{
   return mInitialNegativesOption.GetValue(mInitialNegatives)
          && mCandidatesFromNegativePolicyOption.GetValue(
              mCandidatesFromNegativePolicy);
}

SUBool::PCCompPreprocessOptions::PCCompPreprocessOptions()
    : OptionsGroup("Preprocessing"),
      mPreprocessMethodOption(AbsComp::kPreprocessMethodAnnotation,
          "preprocess,r", "Method of preprocessing.", kDefaultPreprocessMethod)
{
   Add(mPreprocessMethodOption);
   AddOptions()("no-backbone-propagation",
       po::bool_switch(&mNoBackbonePropagation),
       "Do not perform backbone propagation before preprocessing.");
}

bool
SUBool::PCCompPreprocessOptions::GetValues()
{
   return mPreprocessMethodOption.GetValue(mPreprocessMethod);
}

SUBool::PCSolverOptions::PCSolverOptions()
    : OptionsGroup("Solvers"),
      mSolverTypeOption(kSolverTypeAnnotation, "pccheck-solver",
          "Solver to be used for the PC check. Only the available solvers are "
          "listed.",
          kDefaultSolverType)
{
   Add(mSolverTypeOption);
}

bool
SUBool::PCSolverOptions::GetValues()
{
   return mSolverTypeOption.GetValue(mSolverType);
}

SUBool::PCOneProveOptions::PCOneProveOptions(bool add_compile_opts)
    : OptionsGroup("1-Provability Encoding"),
      mOneProveOption(AbsEncoder::kUPOneProveAnnotation, "one-prove,1",
          "The kind of encoding used to encode 1-provability.",
          kDefaultOneProve),
      mImplSystemOption(kLitImplSystemAnnotation, "impl-system",
          "Determines the type of an implication system to be used if needed "
          "by encoding of empowerement or 1-provability.",
          kDefaultLitImplSystem),
      mImplSystemRebuildOption(kImplSystemRebuildAnnotation,
          "impl-system-rebuild",
          "The strategy for rebuilding the implication system",
          kDefaultImplSystemRebuild)
{
   Add(mOneProveOption);
   Add(mImplSystemOption);
   AddOptions()("level,l",
       po::value<unsigned>(&mUPLevelBound)->default_value(mUPLevelBound),
       "Maximum level of unit propagation for 1-provability. This is the "
       "maximum level used when checking propagation or unit refutation "
       "completeness by SAT. Value 0 means, the level is not bounded.") //
       ("log-enc-opts",
           po::value<std::string>(&mLogEncOpts)
               ->default_value(kDefaultLogEncOpts),
           "Options specifying some of the features of the logarithmic "
           "encoding. It is a string composed of comma separated option "
           "strings which specifies which additional parts should be added to "
           "the encoding. Possible values specifying the features are:\n"
           "amf -- AMO on fire variables for each derived literal will be part "
           "of the encoding.\n"
           "cex -- Comparison exclusivity clauses will be added to the "
           "encoding.\n"
           "flm -- Clauses forcing linear order when the number of levels is "
           "maximum will be added to the encoding.\n"
           "cml -- A variable is contradictory if and only its level has all "
           "ones.");
   if (add_compile_opts)
   {
      Add(mImplSystemRebuildOption);
   }
}

bool
SUBool::PCOneProveOptions::GetValues()
{
   return mOneProveOption.GetValue(mOneProve)
          && mImplSystemOption.GetValue(mImplSystem) && ParseLogEncOpts();
}

bool
SUBool::PCOneProveOptions::ParseLogEncOpts()
{
   std::vector<bool> opt_values(kLogEncOptsNames.size(), false);
   size_t start = 0;
   size_t end = 0;
   if (!mLogEncOpts.empty())
   {
      while (start != std::string::npos)
      {
         end = mLogEncOpts.find_first_of(',', start);
         auto i = std::find(kLogEncOptsNames.begin(), kLogEncOptsNames.end(),
             mLogEncOpts.substr(start, end - start));
         if (i == kLogEncOptsNames.end())
         {
            logs.Log(0) << "Error parsing the value of option --log-enc-opts"
                        << std::endl;
            return false;
         }
         opt_values[i - kLogEncOptsNames.begin()] = true;
         start = end + static_cast<size_t>(end < std::string::npos);
      }
   }
   mEncLogarithmicConf.add_amo_on_fire_vars = opt_values[0];
   mEncLogarithmicConf.add_compare_exclusivity = opt_values[1];
   mEncLogarithmicConf.add_force_linear_order_on_max_level = opt_values[2];
   mEncLogarithmicConf.contradiction_has_max_level = opt_values[3];
   return true;
}

SUBool::PCRandomPCChecksOptions::PCRandomPCChecksOptions()
    : OptionsGroup("Random PC Checks")
{
   AddOptions()("pccheck-random-rounds",
       po::value<unsigned>(&mRounds)->default_value(kDefaultRounds),
       "Number of rounds during one random check of propagation "
       "completeness.") //
       ("pccheck-random-budget",
           po::value<unsigned>(&mBudget)->default_value(kDefaultBudget),
           "How many times an unsucessful random PC check should be repeated. "
           "0 means no random check is ever performed.") //
       ("pccheck-random-budget-inc",
           po::value<unsigned>(&mBudgetInc)->default_value(kDefaultBudgetInc),
           "Increase of the current budget in case a random negative example "
           "is successfully found. 0 means no increase.");
}

bool
SUBool::PCRandomPCChecksOptions::GetValues()
{
   return true;
}

SUBool::PCMinimizationOptions::PCMinimizationOptions()
    : OptionsGroup("Minimization"),
      mMinimizeMethodOption(kMinimizeMethodAnnotation, "minimize-method",
          "Minimization method.", kDefaultMinimizeMethod)
{
   AddOptions()("no-final-minimize", po::bool_switch(&mNoFinalMinimize),
       "do not minimize the compiled CNF at the end of compilation.");
   Add(mMinimizeMethodOption);
   AddOptions() //
       ("minimize-steps,s", po::value<unsigned>(&mMinimizeSteps),
           "Number of empowering clauses after which a heuristic minimization "
           "of the current formula proceeds. Zero (0) means that no "
           "minimization is made during the compilation. However, final "
           "minimization shall proceed depending on the presence of -m option.") //
       ("level-minimize", po::bool_switch(&mLevelMinimize),
           "Perform minimization when the number of levels of unit "
           "propagation increases during the compilation. (Not needed for SCC "
           "based encodings of 1-provability when the implication system "
           "rebuild strategy is clause_added or always)");
}

bool
SUBool::PCMinimizationOptions::GetValues()
{
   return mMinimizeMethodOption.GetValue(mMinimizeMethod);
}

SUBool::PCTimeoutsOptions::PCTimeoutsOptions(bool include_timeout)
    : OptionsGroup("Timeouts")
{
   if (include_timeout)
   {
      AddOptions() //
          ("timeout,t",
              po::value<long double>(&mTimeout)->default_value(-1.0l, "-1.0"),
              "Total timeout for the computation in seconds as a fractional "
              "number. After this amount of seconds, the process finishes. "
              "Value t<=0.0 means that the computation is not bounded by a "
              "timeout. The timer uses CLOCK_REALTIME");
   }
   AddOptions() //
       ("emp-timeout",
           po::value<long double>(&mEmpTimeouts.timeout)
               ->default_value(-1.0l, "-1.0"),
           "Timeout for a single SAT call which looks for an empowering "
           "implicate. The timeout is in seconds as a fractional number.") //
       ("emp-timeout-level-base",
           po::value<long double>(&mEmpTimeouts.level_base)
               ->default_value(-1.0l, "-1.0"),
           "This parameter together with emp-timeout-level-factor allows to "
           "set the timeout for a single SAT which looks for an empowering "
           "implicate depending on the number of levels of unit propagation "
           "allowed in the encoding. Denote the value of "
           "emp-timeout-level-base as B and the value of "
           "emp-timeout-level-factor as F. For a level L which is neither 0 "
           "nor maximum (i.e. the number of variables), the timeout is set as "
           "B+F*L. If the value is higher than emp-timeout, then the value of "
           "emp-timeout is used instead. If the value is not positive, then it "
           "is interpreted as infinity. The parameter value is passed as a "
           "fractional number specifying the number of seconds.") //
       ("emp-timeout-level-factor",
           po::value<long double>(&mEmpTimeouts.level_factor)
               ->default_value(0.0l, "0.0"),
           "This parameter together with emp-timeout-level-base allows to set "
           "the timeout for a single SAT which looks for an empowering "
           "implicate depending on the number of levels of unit propagation "
           "allowed in the encoding (see the description of "
           "emp-timeout-level-base for more details).");
}

bool
SUBool::PCTimeoutsOptions::GetValues()
{
   return true;
}

SUBool::PCCheckGeneralOptions::PCCheckGeneralOptions()
    : OptionsGroup("General Options"),
      mGoalOption(
          kGoalAnnotation, "goal,g", "What should be checked", kDefaultGoal)
{
   Add(mGoalOption);
   AddOptions() //
       ("encode,e", po::value(&mEncodeFile),
           "Do not check, just write the encoding to the given file.") //
       ("dump,d", po::bool_switch(&mDump),
           "Dump the names of variables into the output file (given with -e "
           "option).") //
       ("no-up-backbones,n", po::bool_switch(&mNoUPBackbones),
           "Do NOT perform propagation of literals which can be derived by "
           "unit propagation from the input formula. This is useful for "
           "checking the construction of an encoding.") //
       ("initial-up-level",
           po::value(&mInitialUPLevel)->default_value(mInitialUPLevel),
           "UP level of the first call of the SAT solver with the encoding. If "
           "0, then there is only one call with the maximum level.");
}

bool
SUBool::PCCheckGeneralOptions::GetValues()
{
   return true;
}

