/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file pc_opt.h
 * Common options of programs related to propagation completeness.
 */

#ifndef __PC_OPT_H
#define __PC_OPT_H

#include "drenc.h"
#include "empower.h"
#include "enum_opt.h"
#include "pclearncomp.h"
#include "prg_opt_group.h"

namespace SUBool
{
   // "Minimization method", "minimize-method"
   EnumOption<MINIMIZE_METHOD> minimize_method_opt();

   // Algorithm selection in pccompile
   class PCCompAlgorithmOptions : public OptionsGroup
   {
    public:
      enum class ALGORITHM
      {
         INCREMENT_PC,
         INCREMENT_URC,
         INCREMENT_URC_PC,
         LEARNING
      };

      static const EnumAnnotation<ALGORITHM> kAlgorithmAnnotation;
      static constexpr ALGORITHM kDefaultAlgorithm = ALGORITHM::LEARNING;

    private:
      EnumOption<ALGORITHM> mAlgorithmOption;
      ALGORITHM mAlgorithm{kDefaultAlgorithm};

    public:
      PCCompAlgorithmOptions();

      virtual bool GetValues() override;

      ALGORITHM Algorithm() const;
   };

   // Learning Algorithm Options
   class PCLearnCompOptions : public OptionsGroup
   {
    public:
      PCLearnComp::INITIAL_NEGATIVES kDefaultInitialNegatives{
          PCLearnComp::INITIAL_NEGATIVES::BODY_MINIMAL};
      PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY
          kDefaultCandidatesFromNegativePolicy{
              PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY::SYNCHRONIZE};

    private:
      EnumOption<PCLearnComp::INITIAL_NEGATIVES> mInitialNegativesOption;
      PCLearnComp::INITIAL_NEGATIVES mInitialNegatives{
          kDefaultInitialNegatives};
      EnumOption<PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY>
          mCandidatesFromNegativePolicyOption;
      PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY
          mCandidatesFromNegativePolicy{kDefaultCandidatesFromNegativePolicy};
      bool mUseCNFHypothesisForEmpowerment{false};

    public:
      PCLearnCompOptions();

      virtual bool GetValues() override;

      PCLearnComp::INITIAL_NEGATIVES InitialNegatives() const;
      PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY
      CandidatesFromNegativePolicy() const;
      bool UseCNFHypothesisForEmpeworment() const;
   };

   // pccompile preprocessing
   class PCCompPreprocessOptions : public OptionsGroup
   {
    public:
      static constexpr AbsComp::PREPROCESS_METHOD kDefaultPreprocessMethod =
          AbsComp::PREPROCESS_METHOD::PRIME_MINIMIZE;

    private:
      bool mNoBackbonePropagation{false};
      EnumOption<AbsComp::PREPROCESS_METHOD> mPreprocessMethodOption;
      AbsComp::PREPROCESS_METHOD mPreprocessMethod{kDefaultPreprocessMethod};

    public:
      PCCompPreprocessOptions();

      virtual bool GetValues() override;

      bool NoBackbonePropagation() const;
      AbsComp::PREPROCESS_METHOD PreprocessMethod() const;
   };

   // Solver selection
   class PCSolverOptions : public OptionsGroup
   {
    public:
      static constexpr SOLVER_TYPE kDefaultSolverType = SOLVER_TYPE::GLUCOSE;

    private:
      EnumOption<SOLVER_TYPE> mSolverTypeOption;
      SOLVER_TYPE mSolverType{kDefaultSolverType};

    public:
      PCSolverOptions();

      virtual bool GetValues() override;

      SOLVER_TYPE SolverType() const;
   };

   class PCOneProveOptions : public OptionsGroup
   {
    public:
      static constexpr AbsEncoder::UP_ONE_PROVE kDefaultOneProve =
          AbsEncoder::UP_ONE_PROVE::SCC_UNARY;
      static constexpr LIT_IMPL_SYSTEM kDefaultLitImplSystem =
          LIT_IMPL_SYSTEM::Default;
      static constexpr std::array kLogEncOptsNames{"amf", "cex", "flm", "cml"};
      inline static const std::string kDefaultLogEncOpts{"amf,cex,flm,cml"};
      static constexpr IMPL_SYSTEM_REBUILD kDefaultImplSystemRebuild =
          IMPL_SYSTEM_REBUILD::Default;

    private:
      EnumOption<AbsEncoder::UP_ONE_PROVE> mOneProveOption;
      std::string mLogEncOpts{};
      EnumOption<LIT_IMPL_SYSTEM> mImplSystemOption;
      unsigned mUPLevelBound{0};
      EnumOption<IMPL_SYSTEM_REBUILD> mImplSystemRebuildOption;

      UPEncLogarithmic::Config mEncLogarithmicConf{};
      AbsEncoder::UP_ONE_PROVE mOneProve{kDefaultOneProve};
      LIT_IMPL_SYSTEM mImplSystem{kDefaultLitImplSystem};
      IMPL_SYSTEM_REBUILD mImplSystemRebuild{kDefaultImplSystemRebuild};

      bool ParseLogEncOpts();

    public:
      PCOneProveOptions(bool add_compile_opts);

      virtual bool GetValues() override;

      AbsEncoder::UP_ONE_PROVE OneProve() const;
      const UPEncLogarithmic::Config EncLogarithmicConf() const;
      LIT_IMPL_SYSTEM ImplSystem() const;
      unsigned UPLevelBound() const;
      IMPL_SYSTEM_REBUILD ImplSystemRebuild() const;
   };

   class PCRandomPCChecksOptions : public OptionsGroup
   {
    public:
      static constexpr unsigned kDefaultRounds = 100;
      static constexpr unsigned kDefaultBudget = 10;
      static constexpr unsigned kDefaultBudgetInc = 1;

    private:
      unsigned mRounds{kDefaultRounds};
      unsigned mBudget{kDefaultBudget};
      unsigned mBudgetInc{kDefaultBudgetInc};

    public:
      PCRandomPCChecksOptions();

      virtual bool GetValues() override;

      unsigned Rounds() const;
      unsigned Budget() const;
      unsigned BudgetInc() const;
   };

   class PCMinimizationOptions : public OptionsGroup
   {
    public:
      static constexpr unsigned kDefaultMinimizeSteps = 50;

    private:
      bool mNoFinalMinimize{false};
      EnumOption<MINIMIZE_METHOD> mMinimizeMethodOption;
      MINIMIZE_METHOD mMinimizeMethod{kDefaultMinimizeSteps};
      unsigned mMinimizeSteps{kDefaultMinimizeSteps};
      bool mLevelMinimize{false};

    public:
      PCMinimizationOptions();

      virtual bool GetValues() override;

      bool FinalMinimize() const;
      MINIMIZE_METHOD MinimizeMethod() const;
      unsigned MinimizeSteps() const;
      bool LevelMinimize() const;
   };

   class PCTimeoutsOptions : public OptionsGroup
   {
      AbsCheckerSAT::Timeouts mEmpTimeouts{};
      long double mTimeout{-1.0};

    public:
      PCTimeoutsOptions(bool include_timeout = true);

      virtual bool GetValues() override;

      const AbsCheckerSAT::Timeouts &EmpTimeouts() const;
      long double Timeout() const;
   };

   class PCCheckGeneralOptions : public OptionsGroup
   {
    public:
      enum class GOAL
      {
         PC,
         URC
      };

      static const EnumAnnotation<GOAL> kGoalAnnotation;
      static constexpr GOAL kDefaultGoal = GOAL::PC;

    private:
      EnumOption<GOAL> mGoalOption;
      GOAL mGoal{kDefaultGoal};
      bool mDump{false};
      boost::optional<std::string> mEncodeFile{};
      bool mNoUPBackbones{false};
      unsigned mInitialUPLevel{1};

    public:
      PCCheckGeneralOptions();

      virtual bool GetValues() override;

      GOAL Goal() const;
      bool Dump() const;
      const boost::optional<std::string> &EncodeFile() const;
      bool NoUPBackbones() const;
      unsigned InitialUPLevel() const;
   };
} // namespace SUBool

inline auto
SUBool::PCCompAlgorithmOptions::Algorithm() const -> ALGORITHM
{
   return mAlgorithm;
}

inline auto
SUBool::PCLearnCompOptions::InitialNegatives() const
    -> PCLearnComp::INITIAL_NEGATIVES
{
   return mInitialNegatives;
}

inline auto
SUBool::PCLearnCompOptions::CandidatesFromNegativePolicy() const
    -> PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY
{
   return mCandidatesFromNegativePolicy;
}

inline bool
SUBool::PCLearnCompOptions::UseCNFHypothesisForEmpeworment() const
{
   return mUseCNFHypothesisForEmpowerment;
}

inline bool
SUBool::PCCompPreprocessOptions::NoBackbonePropagation() const
{
   return mNoBackbonePropagation;
}

inline SUBool::AbsComp::PREPROCESS_METHOD
SUBool::PCCompPreprocessOptions::PreprocessMethod() const
{
   return mPreprocessMethod;
}

inline SUBool::SOLVER_TYPE
SUBool::PCSolverOptions::SolverType() const
{
   return mSolverType;
}

inline SUBool::AbsEncoder::UP_ONE_PROVE
SUBool::PCOneProveOptions::OneProve() const
{
   return mOneProve;
}

inline const SUBool::UPEncLogarithmic::Config
SUBool::PCOneProveOptions::EncLogarithmicConf() const
{
   return mEncLogarithmicConf;
}

inline SUBool::LIT_IMPL_SYSTEM
SUBool::PCOneProveOptions::ImplSystem() const
{
   return mImplSystem;
}

inline unsigned
SUBool::PCOneProveOptions::UPLevelBound() const
{
   return mUPLevelBound;
}

inline SUBool::IMPL_SYSTEM_REBUILD
SUBool::PCOneProveOptions::ImplSystemRebuild() const
{
   return mImplSystemRebuild;
}

inline unsigned
SUBool::PCRandomPCChecksOptions::Rounds() const
{
   return mRounds;
}

inline unsigned
SUBool::PCRandomPCChecksOptions::Budget() const
{
   return mBudget;
}

inline unsigned
SUBool::PCRandomPCChecksOptions::BudgetInc() const
{
   return mBudgetInc;
}

inline bool
SUBool::PCMinimizationOptions::FinalMinimize() const
{
   return !mNoFinalMinimize;
}

inline SUBool::MINIMIZE_METHOD
SUBool::PCMinimizationOptions::MinimizeMethod() const
{
   return mMinimizeMethod;
}

inline unsigned
SUBool::PCMinimizationOptions::MinimizeSteps() const
{
   return mMinimizeSteps;
}

inline bool
SUBool::PCMinimizationOptions::LevelMinimize() const
{
   return mLevelMinimize;
}

inline const SUBool::AbsCheckerSAT::Timeouts &
SUBool::PCTimeoutsOptions::EmpTimeouts() const
{
   return mEmpTimeouts;
}

inline long double
SUBool::PCTimeoutsOptions::Timeout() const
{
   return mTimeout;
}

inline SUBool::PCCheckGeneralOptions::GOAL
SUBool::PCCheckGeneralOptions::Goal() const
{
   return mGoal;
}

inline bool
SUBool::PCCheckGeneralOptions::Dump() const
{
   return mDump;
}

inline const boost::optional<std::string> &
SUBool::PCCheckGeneralOptions::EncodeFile() const
{
   return mEncodeFile;
}

inline bool
SUBool::PCCheckGeneralOptions::NoUPBackbones() const
{
   return mNoUPBackbones;
}

inline unsigned
SUBool::PCCheckGeneralOptions::InitialUPLevel() const
{
   return mInitialUPLevel;
}

#endif
