/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file pcchecker.cpp
 * Checking propagation or unit refutation completeness of a formula with SAT.
 */

#include "pcchecker.h"
#include "cpuutil.h"
#include "drenc.h"
#include "glucose_bind.h"
#include "kissat_bind.h"
#include "logstream.h"

void
SUBool::AbsCheckerSAT::InitBase(unsigned max_up_level_bound,
    unsigned current_up_level_bound, bool wait_before_max_level_check,
    const Timeouts &timeouts)
{
   mTimeouts = timeouts;
   mMaxUPLevelBound = max_up_level_bound;
   mMaxUPLevel = 0;
   mCurrentUPLevelBound = current_up_level_bound;
   mTotalTime = mTotalTimeSAT = mTotalTimeUNSAT = 0.0;
   mSolveCountSAT = mSolveCountUNSAT = 0;
   mLastWitnessVariable = 0;
   mState =
       (wait_before_max_level_check ? STATE::NO_INCREASE : STATE::NO_SPLIT);
}

long double
SUBool::AbsCheckerSAT::Timeout(unsigned level) const
{
   auto t = mTimeouts.timeout;
   if (level > 0 && level < mMaxUPLevel && mTimeouts.level_base > 0.0)
   {
      auto lt = mTimeouts.level_base
                + static_cast<double>(level - 1) * mTimeouts.level_factor;
      if (t < 0.0 || lt < t)
      {
         t = lt;
      }
   }
   return t;
};

std::pair<std::optional<SUBool::Clause>, unsigned>
SUBool::AbsCheckerSAT::FindEmpoweringWithLevel(const CNF &cnf,
    const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider> impl_system_provider, unsigned level,
    bool assume_sat)
{
   if (mState == STATE::INTERRUPT)
   {
      logs.TLog(3) << "[AbsCheckerSAT::FindEmpoweringWithLevel] Interrupted"
                   << std::endl;
      return std::make_pair(std::optional<SUBool::Clause>(), 0);
   }
   if (mMaxUPLevelBound > 0 && level > mMaxUPLevelBound)
   {
      logs.TLog(3) << "[AbsCheckerSAT::FindEmpoweringWithLevel] level bigger "
                      "than allowed"
                   << std::endl;
      return std::make_pair(std::optional<SUBool::Clause>(), 0);
   }
   CPUTime cpu_time;
   auto enc = Encoder(level);
   enc->BuildEncoding(cnf, cl2lit, impl_system_provider);
   auto max_level = enc->MaxUPLevelBound();
   if (mMaxUPLevelBound > 0 && max_level > mMaxUPLevelBound)
   {
      mMaxUPLevel = mMaxUPLevelBound;
   }
   else
   {
      mMaxUPLevel = max_level;
   }
   logs.TLog(3) << "[AbsCheckerSAT::FindEmpoweringWithLevel] level="
                << (level == 0 ? mMaxUPLevel : level) << ", input cnf "
                << cnf.MaxVariable() << " " << cnf.Size() << std::endl;
   auto wrap = solver_wrapper(
       mSolverType, assume_sat ? SolverInterface::ADAPTATION::LIKELY_SAT
                               : SolverInterface::ADAPTATION::LIKELY_UNSAT);
   auto &g = wrap.Solver();
   enc->PassToSolver(g);
   logs.TLog(3) << "... Calling SAT with encoding (p cnf " << enc->MaxVariable()
                << " " << enc->Size() << ")" << std::endl;
   // enc->WriteEncoding(std::cout);
   auto t = Timeout(level);
   logs.TLog(3) << "... timeout: " << t << "s" << std::endl;
   // logs.TLog(3) << "... cnf: " << cnf << std::endl;
   // logs.TLog(3) << "... enc: " << *enc << std::endl;
   auto res = CallSolve(g, t);
   mLastWitnessVariable = 0;
   mTotalTime += cpu_time.TimeSpan();
   if (res)
   {
      Clause emp;
      mLastWitnessVariable = enc->ExtractClause(g, &emp);
      mTotalTimeSAT += cpu_time.TimeSpan();
      logs.TLog(3) << "... Found empowering implicate, time=" << cpu_time
                   << std::endl;
      ++mSolveCountSAT;
      return std::make_pair(std::make_optional(emp), level);
   }
   mTotalTimeUNSAT += cpu_time.TimeSpan();
   ++mSolveCountUNSAT;
   if (TimedOut())
   {
      logs.TLog(3) << "... No empowering implicate found (timed out), time="
                   << cpu_time << std::endl;
   }
   else
   {
      logs.TLog(3) << "... No empowering implicate found, time=" << cpu_time
                   << std::endl;
   }
   return std::make_pair(std::optional<Clause>(),
       enc->NextUPLevelBound(cnf, cl2lit, impl_system_provider, level));
}

bool
SUBool::AbsCheckerSAT::CallSolve(
    SUBool::SolverInterface &g, long double timeout)
{
   if (mState == STATE::INTERRUPT)
   {
      return false;
   }
   mTimedOut = false;
   CPUTime cpu_time;
   bool result = false;
   if (timeout > 0.0)
   {
      auto res =
          g.SolveWithTimeout(std::chrono::duration<long double>(timeout));
      mTimedOut = (res == PartialValue::UNDEF);
      result = (res == PartialValue::TRUE);
   }
   else
   {
      result = g.Solve();
   }
   mLastSolveTime = cpu_time.TimeSpan();
   return result;
}

std::optional<SUBool::Clause>
SUBool::AbsCheckerSAT::FindEmpowering(const CNF &cnf,
    const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider> impl_system_provider)
{
   if (mState == STATE::NO_SPLIT || mState == STATE::NO_INCREASE)
   {
      auto emp = FindEmpoweringWithLevel(
          cnf, cl2lit, impl_system_provider, mCurrentUPLevelBound, true);
      if (emp.first)
      {
         return emp.first;
      }
      if (mCurrentUPLevelBound == 0 || mCurrentUPLevelBound >= mMaxUPLevel)
      {
         return {};
      }
      mCurrentUPLevelBound = emp.second;
      if (mState == STATE::NO_INCREASE)
      {
         mState = STATE::WILL_INCREASE;
         return {};
      }
   }
   if (mState == STATE::WILL_INCREASE)
   {
      mState = STATE::NO_INCREASE;
   }
   return FindEmpoweringWithLevel(cnf, cl2lit, impl_system_provider, 0, false)
       .first;
}

unsigned
SUBool::AbsCheckerSAT::CurrentUPLevelBound() const
{
   return mCurrentUPLevelBound;
}

long double
SUBool::AbsCheckerSAT::TotalTime() const
{
   return mTotalTime;
}

long double
SUBool::AbsCheckerSAT::TotalTimeSAT() const
{
   return mTotalTimeSAT;
}

long double
SUBool::AbsCheckerSAT::TotalTimeUNSAT() const
{
   return mTotalTimeUNSAT;
}

long double
SUBool::AbsCheckerSAT::LastSolveTime() const
{
   return mLastSolveTime;
}

unsigned
SUBool::AbsCheckerSAT::SolveCountSAT() const
{
   return mSolveCountSAT;
}

unsigned
SUBool::AbsCheckerSAT::SolveCountUNSAT() const
{
   return mSolveCountUNSAT;
}

bool
SUBool::AbsCheckerSAT::WillIncreaseLevel() const
{
   return mState == STATE::WILL_INCREASE;
}

unsigned
SUBool::AbsCheckerSAT::LastWitnessVariable() const
{
   return mLastWitnessVariable;
}

std::unique_ptr<SUBool::AbsEncoder>
SUBool::UCPCInterleaveCheckerSAT::NewPCEncoder(unsigned up_level_bound)
{
   logs.TLog(4) << "[UCPCInterleaveCheckerSAT::Encoder] Using PCEncoder"
                << std::endl;
   mUseUC = true;
   mPCEncConf.up_level_bound = up_level_bound;
   return std::unique_ptr<AbsEncoder>(new PCEncoder(mPCEncConf));
}

std::unique_ptr<SUBool::AbsEncoder>
SUBool::UCPCInterleaveCheckerSAT::NewUCEncoder(unsigned up_level_bound)
{
   mUseUC = false;
   logs.TLog(4) << "[UCPCInterleaveCheckerSAT::Encoder] Using UCEncoder"
                << std::endl;
   mEncConf.up_level_bound = up_level_bound;
   return std::unique_ptr<AbsEncoder>(new UCEncoder(mEncConf));
}

void
SUBool::UCPCInterleaveCheckerSAT::Init(UCEncoder::Config uc_enc_conf,
    PCEncoder::Config pc_enc_conf, bool wait_before_max_level_check,
    const Timeouts &timeouts)
{
   InitBase(
       uc_enc_conf.up_level_bound, 1, wait_before_max_level_check, timeouts);
   mEncConf = std::move(uc_enc_conf);
   mPCEncConf = std::move(pc_enc_conf);
   mUseUC = true;
}

std::unique_ptr<SUBool::AbsEncoder>
SUBool::UCPCInterleaveCheckerSAT::Encoder(unsigned up_level_bound)
{
   return (up_level_bound == 0 || mUseUC) ? NewUCEncoder(up_level_bound)
                                          : NewPCEncoder(up_level_bound);
}

bool
SUBool::UCPCInterleaveCheckerSAT::RequiresImplSystem() const
{
   return AbsEncoder::AbsConfig::OneProveRequiresImplSystem(
              mEncConf.up_one_prove)
          || AbsEncoder::AbsConfig::OneProveRequiresImplSystem(
              mPCEncConf.up_one_prove);
}

bool
SUBool::UCPCInterleaveCheckerSAT::RequiresSCCGraph() const
{
   return AbsEncoder::AbsConfig::OneProveRequiresSCCGraph(mEncConf.up_one_prove)
          || AbsEncoder::AbsConfig::OneProveRequiresSCCGraph(
              mPCEncConf.up_one_prove);
}

bool
SUBool::is_urc(const CNF &cnf)
{
   UCCheckerSAT checker;
   ClausesToLiterals cl2lit(cnf);
   std::shared_ptr<EncISProvider> impl_system_provider;
   if (checker.RequiresImplSystem())
   {
      impl_system_provider = make_enc_is_provider(LIT_IMPL_SYSTEM::Default);
      impl_system_provider->Cleanup(cnf);
   }
   checker.Init(UCEncoder::Config(), false, AbsCheckerSAT::Timeouts());
   return !checker.FindEmpowering(cnf, cl2lit, impl_system_provider);
}

bool
SUBool::is_pc(const CNF &cnf)
{
   PCCheckerSAT checker;
   ClausesToLiterals cl2lit(cnf);
   std::shared_ptr<EncISProvider> impl_system_provider;
   if (checker.RequiresImplSystem())
   {
      impl_system_provider = make_enc_is_provider(LIT_IMPL_SYSTEM::Default);
      impl_system_provider->Cleanup(cnf);
   }
   checker.Init(PCEncoder::Config(), false, AbsCheckerSAT::Timeouts());
   return !checker.FindEmpowering(cnf, cl2lit, impl_system_provider);
}
