/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file pcchecker.h
 * Checking propagation or unit refutation completeness of a formula with SAT.
 */

#ifndef __PCCHECKER_H
#define __PCCHECKER_H

#include <random>

#include <optional>

#include "empower.h"
#include "logstream.h"
#include "pcenc.h"
#include "ucenc.h"

namespace SUBool
{
   /* Checking propagation and unit refutation completeness based on SAT
    * encoding. */
   class AbsCheckerSAT
   {
    public:
      struct Timeouts
      {
         long double timeout{-1.0l};
         long double level_base{-1.0l};
         long double level_factor{0.0l};
      };

    private:
      unsigned mMaxUPLevelBound{0};
      unsigned mMaxUPLevel{0};
      unsigned mCurrentUPLevelBound{1};
      long double mTotalTime{0.0l};
      long double mTotalTimeSAT{0.0l};
      long double mTotalTimeUNSAT{0.0l};
      long double mLastSolveTime{0.0l};
      unsigned mSolveCountSAT{0};
      unsigned mSolveCountUNSAT{0};
      unsigned mLastWitnessVariable{0};
      SOLVER_TYPE mSolverType{SOLVER_TYPE::GLUCOSE};
      enum class STATE
      {
         NO_SPLIT,
         WILL_INCREASE,
         NO_INCREASE,
         INTERRUPT
      };
      std::atomic<STATE> mState{STATE::NO_SPLIT};
      Timeouts mTimeouts{};
      bool mTimedOut{false};

      long double Timeout(unsigned level) const;
      bool CallSolve(SolverInterface &g, long double timeout);

    protected:
      void InitBase(unsigned max_up_level_bound,
          unsigned current_up_level_bound, bool wait_before_max_level_check,
          const Timeouts &timeouts);

    public:
      virtual ~AbsCheckerSAT() = default;
      virtual std::unique_ptr<AbsEncoder> Encoder(unsigned up_level_bound) = 0;
      virtual std::optional<Clause> FindEmpowering(const CNF &cnf,
          const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider);
      std::pair<std::optional<Clause>, unsigned> FindEmpoweringWithLevel(
          const CNF &cnf, const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider,
          unsigned level, bool assume_sat);
      virtual bool RequiresImplSystem() const = 0;
      virtual bool RequiresSCCGraph() const = 0;
      unsigned CurrentUPLevelBound() const;
      /**@brief Total time of FindEmpoweringWithLevel. */
      long double TotalTime() const;
      /**@brief Total time of SAT solver calls with result SAT. */
      long double TotalTimeSAT() const;
      /**@brief Total time of SAT solver calls with result UNSAT. */
      long double TotalTimeUNSAT() const;
      /**@brief Time of the last SAT solver call. 0 if no call has been made
       * yet. */
      long double LastSolveTime() const;
      /**@brief The number of SAT solver calls with result SAT.
       */
      unsigned SolveCountSAT() const;
      /**@brief The number of SAT solver calls with result UNSAT.
       */
      unsigned SolveCountUNSAT() const;
      bool WillIncreaseLevel() const;
      unsigned LastWitnessVariable() const;
      bool
      TimedOut() const
      {
         return mState == STATE::INTERRUPT || mTimedOut;
      }

      // Does not interrupt the solver.
      void
      Interrupt()
      {
         mState = STATE::INTERRUPT;
      }

      void
      SetSolverType(SOLVER_TYPE solver_type)
      {
         mSolverType = solver_type;
      }
   };

   template <class T> class CheckerSATTemplate : public AbsCheckerSAT
   {
      typename T::Config mEncConf;

    public:
      CheckerSATTemplate() : AbsCheckerSAT(), mEncConf() {}

      virtual void
      Init(typename T::Config conf, bool wait_before_max_level_check,
          const Timeouts &timeouts, unsigned initial_up_level = 1)
      {
         mEncConf = std::move(conf);
         InitBase(conf.up_level_bound, initial_up_level,
             wait_before_max_level_check, timeouts);
      }

      virtual std::unique_ptr<AbsEncoder>
      Encoder(unsigned up_level_bound) override
      {
         // mEncConf.use_backmap = true;
         mEncConf.up_level_bound = up_level_bound;
         return std::unique_ptr<AbsEncoder>(new T(mEncConf));
      }

      virtual bool
      RequiresImplSystem() const override
      {
         return AbsEncoder::AbsConfig::OneProveRequiresImplSystem(
             mEncConf.up_one_prove);
      }

      virtual bool
      RequiresSCCGraph() const override
      {
         return AbsEncoder::AbsConfig::OneProveRequiresSCCGraph(
             mEncConf.up_one_prove);
      }
   };

   typedef CheckerSATTemplate<PCEncoder> PCCheckerSAT;
   typedef CheckerSATTemplate<UCEncoder> UCCheckerSAT;

   class UCPCInterleaveCheckerSAT : public AbsCheckerSAT
   {
      UCEncoder::Config mEncConf;
      PCEncoder::Config mPCEncConf;
      bool mUseUC{true};

      std::unique_ptr<SUBool::AbsEncoder> NewPCEncoder(unsigned up_level_bound);
      std::unique_ptr<SUBool::AbsEncoder> NewUCEncoder(unsigned up_level_bound);

    public:
      UCPCInterleaveCheckerSAT() : AbsCheckerSAT(), mEncConf(), mPCEncConf() {}
      virtual void Init(UCEncoder::Config uc_enc_conf,
          PCEncoder::Config pc_enc_conf, bool wait_before_max_level_check,
          const Timeouts &timeouts);
      virtual std::unique_ptr<AbsEncoder> Encoder(
          unsigned up_level_bound) override;

      virtual bool RequiresImplSystem() const override;
      virtual bool RequiresSCCGraph() const override;
   };

   bool is_urc(const CNF &cnf);
   bool is_pc(const CNF &cnf);
} // namespace SUBool

#endif
