/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file pcenc.cpp
 * Encoding of checking if the given CNF admits an empowering
 * implicate.
 */

#include <memory>
#include <sstream>

#include "cardinal.h"
#include "cnfutil.h"
#include "pcenc.h"
#include "up_log.h"
#include "up_max.h"
#include "up_onelevel.h"
#include "up_onelevel_is.h"
#include "up_quad.h"

// On only for debugging
//#define ONLY_UP_ONE_PROVE

SUBool::PCEncoder::PCEncoder(const Config &conf)
    : AbsEncoder(conf), mConf(conf), mOneProveInputs(), mOneProveOutputs(),
      mEmpowerInputs(), mEmpowerOutputs(), mSelectionLits(),
      mEmpEncoder(nullptr)
{
}

void
SUBool::PCEncoder::InitEmpEncoder(const CNF &cnf,
    const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider> impl_system_provider)
{
   if (impl_system_provider->HasEmpowerImplSystem())
   {
      mEmpEncoder = std::make_unique<UPEncOneLevelIS>(cnf, cl2lit, mVarStore,
          impl_system_provider->ProcessedEmpowerImplSystem());
      return;
   }
   mEmpEncoder = std::make_unique<UPEncOneLevel>(cnf, cl2lit, mVarStore);
}

void
SUBool::PCEncoder::AddClauses(const CNF &cnf, const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider> impl_system_provider)
{
   InitEmpEncoder(cnf, cl2lit, impl_system_provider);
   // std::cout << "Adding one provability" << std::endl;
   AddOneProvability();
#ifndef ONLY_UP_ONE_PROVE
   // std::cout << "Adding empowerement" << std::endl;
   AddEmpowerement();
   // std::cout << "AddClauses finished" << std::endl;
#endif
}

const SUBool::AbsEncoder::AbsConfig &
SUBool::PCEncoder::GetConfig() const
{
   return mConf;
}

unsigned
SUBool::PCEncoder::EmpowerOutput(unsigned i) const
{
   return (mEmpowerOutputs.empty() ? mEmpowerInputs[i] : mEmpowerOutputs[i]);
}

unsigned
SUBool::PCEncoder::ExtractClause(
    const SolverInterface &solver, Clause *clause) const
{

   if (clause == nullptr)
   {
      throw NullPointerException(
          "PCEncoder::ExtractClause", "clause == nullptr");
   }
   std::vector<Literal> literals;
   for (unsigned i = 0; i < mOneProveInputs.size(); ++i)
   {
      if (solver.ValueOfVariable(mOneProveInputs[i]) == PartialValue::TRUE)
      {
         Literal lit(i);
         literals.push_back(lit.Negation());
      }
   }
   if (mConf.use_backmap)
   {
      for (unsigned i = 0; i < mVarStore.LastVariable(); ++i)
      {
         std::cout << i + 1 << "=" << solver.ValueOfVariable(i + 1) << " ["
                   << mVarStore.Name(i + 1) << "]" << std::endl;
      }
   }
   *clause = Clause(literals);
   for (unsigned i = 0; i < mSelectionLits.size(); ++i)
   {
      if (solver.ValueOfLiteral(mSelectionLits[i]) == PartialValue::TRUE)
      {
         return i + 1;
      }
   }
   return 0;
}

void
SUBool::PCEncoder::AddOneProveNoInputConflict()
{
   for (unsigned i = 1; i <= mCnfMaxVariable; ++i)
   {
      Literal pos(i, true);
      Literal neg(i, false);
      mClauses.emplace_back(Literal(mOneProveInputs[pos.Index()], false),
          Literal(mOneProveInputs[neg.Index()], false));
   }
}

void
SUBool::PCEncoder::AddOneProveOutputConflict()
{
   std::vector<Literal> t_literals;
   // t_literals.reserve(cnf.MaxVariable());
   t_literals.reserve(mOneProveOutputs.size() / 2);
   for (unsigned i = 1; i <= mOneProveOutputs.size() / 2; ++i)
   // for(unsigned i=1; i<=cnf.MaxVariable(); ++i)
   {
      mVarStore.NameStream() << "t_" << i;
      unsigned t = mVarStore.NewVariable();
      t_literals.emplace_back(t, true);

      unsigned pos_var = mOneProveOutputs[Literal(i, true).Index()];
      unsigned neg_var = mOneProveOutputs[Literal(i, false).Index()];
      Literal(i, false).Index();
      mClauses.emplace_back(Literal(t, false), Literal(pos_var, true));
      mClauses.emplace_back(Literal(t, false), Literal(neg_var, true));
      if (!mConf.optimize)
      {
         mClauses.emplace_back(Literal(t, true), Literal(pos_var, false),
             Literal(neg_var, false));
      }
   }
   mClauses.emplace_back(t_literals);
}

void
SUBool::PCEncoder::AddOneProveBlock()
{
   mVarStore.SetPrefix("v");
   mOneProveEncoder->Encode(
       mConf.up_level_bound, &mOneProveInputs, &mOneProveOutputs, &mClauses);
   mVarStore.SetPrefix("");
}

void
SUBool::PCEncoder::AddSelection()
{
   mVarStore.InitLitArray(&mSelectionLits, mCnfMaxVariable, false, "s_{", "}");
   auto exone_encoder = ExOneEncoder::NewGlobalEncoder(mVarStore);
   exone_encoder->EncodeAppend(mSelectionLits, mClauses);
}

void
SUBool::PCEncoder::AddPassValuesToEmpower()
{
   for (unsigned i = 0; i < mOneProveInputs.size(); ++i)
   {
      mClauses.emplace_back(
          Literal(mEmpowerInputs[i], false), mSelectionLits[i / 2].Negation());
      mClauses.emplace_back(
          Literal(mEmpowerInputs[i], false), Literal(mOneProveInputs[i], true));
      mClauses.emplace_back(Literal(mEmpowerInputs[i], true),
          mSelectionLits[i / 2], Literal(mOneProveInputs[i], false));
      /* XXX */
      // mClauses.emplace_back(
      //     Literal(mEmpowerInputs[i], false),
      //    Literal(mSelectionVars[i/2], true),
      //   Literal(mOneProveInputs[i], true));
      /* */
   }
}

void
SUBool::PCEncoder::AddEmpowerBlock()
{
   mVarStore.SetPrefix("v");
   mEmpEncoder->Encode(0, &mEmpowerInputs, &mEmpowerOutputs, &mClauses);
   mVarStore.SetPrefix("");
}

void
SUBool::PCEncoder::AddEmpNoConflict()
{
   if (mEmpowerOutputs.empty())
   {
      return;
   }
   for (unsigned i = 1; i <= mCnfMaxVariable; ++i)
   {
      Literal pos(i, true);
      Literal neg(i, false);
      mClauses.emplace_back(Literal(EmpowerOutput(pos.Index()), false),
          Literal(EmpowerOutput(neg.Index()), false));
   }
}

void
SUBool::PCEncoder::AddEmpNotDerived()
{
   for (unsigned i = 1; i <= mCnfMaxVariable; ++i)
   {
      Literal pos(i, true);
      Literal neg(i, false);
      mClauses.emplace_back(mSelectionLits[i - 1].Negation(),
          Literal(EmpowerOutput(pos.Index()), false));
      mClauses.emplace_back(mSelectionLits[i - 1].Negation(),
          Literal(EmpowerOutput(neg.Index()), false));
   }
}

void
SUBool::PCEncoder::AddOneProvability()
{
   AddOneProveBlock();
#ifndef ONLY_UP_ONE_PROVE
   AddOneProveNoInputConflict();
   AddOneProveOutputConflict();
#endif
}

void
SUBool::PCEncoder::AddEmpowerement()
{
   AddEmpowerBlock();
   AddSelection();
   AddPassValuesToEmpower();
   AddEmpNoConflict();
   AddEmpNotDerived();
}

void
SUBool::PCEncoder::WriteMappingsOfVariable(std::ostream &out, unsigned i) const
{
   if (i == 0)
   {
      throw OutOfBoundsException(
          "PCEncoder::WriteMappingsOfVariable", "Variable i is 0");
   }
   out << "c Mappings of variable " << i << std::endl;
   // s variables
   if (mSelectionLits.size() < i)
   {
      out << "c No mappings of selection variables." << std::endl;
   }
   else
   {
      out << "c s_" << i << " = " << mSelectionLits[i - 1];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mSelectionLits[i - 1].Variable()) << "]";
      }
      out << std::endl;
   }
   // One provability input variables
   if (mOneProveInputs.size() < 2 * i)
   {
      out << "c No mappings of variables to inputs of 1-provability block."
          << std::endl;
   }
   else
   {
      out << "c input variable for literal " << Literal(2 * (i - 1)) << " = "
          << mOneProveInputs[2 * (i - 1)];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mOneProveInputs[2 * (i - 1)]) << "] ";
      }
      out << std::endl;
      out << "c input variable for literal " << Literal(2 * i - 1) << " = "
          << mOneProveInputs[2 * i - 1];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mOneProveInputs[2 * i - 1]) << "] ";
      }
      out << std::endl;
   }
   // One provability output variables
   if (mOneProveOutputs.size() < 2 * i)
   {
      out << "c No mappings of variables to outputs of 1-provability block."
          << std::endl;
   }
   else
   {
      out << "c output variable for literal " << Literal(2 * (i - 1)) << " = "
          << mOneProveOutputs[2 * (i - 1)];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mOneProveOutputs[2 * (i - 1)]) << "] ";
      }
      out << std::endl;
      out << "c output variable for literal " << Literal(2 * i - 1) << " = "
          << mOneProveOutputs[2 * i - 1];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mOneProveOutputs[2 * i - 1]) << "] ";
      }
      out << std::endl;
   }
   // Empowerment input variables
   if (mEmpowerInputs.size() < 2 * i)
   {
      out << "c No mappings of variables to inputs of empower block."
          << std::endl;
   }
   else
   {
      out << "c u_{" << Literal(2 * (i - 1))
          << "} = " << mEmpowerInputs[2 * (i - 1)];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mEmpowerInputs[2 * (i - 1)]) << "] ";
      }
      out << std::endl;
      out << "c u_{" << Literal(2 * i - 1)
          << "} = " << mEmpowerInputs[2 * i - 1];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mEmpowerInputs[2 * i - 1]) << "] ";
      }
      out << std::endl;
   }
   // Empowerement output variables
   if (mEmpowerOutputs.size() < 2 * i)
   {
      out << "c No mappings of variables to outputs of empower block."
          << std::endl;
      out << "c The outputs are the same as inputs." << std::endl;
   }
   else
   {
      out << "c u_{" << Literal(2 * i - 1)
          << "} = " << mEmpowerOutputs[2 * i - 1];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mEmpowerOutputs[2 * i - 1]) << "] ";
      }
      out << std::endl;
   }
}

std::string
SUBool::PCEncoder::Name() const
{
   return "PC encoding";
}

