/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file pcenc.h
 * Encoding of checking if the given CNF admits an empowering
 * implicate with respect to propagation completeness.
 */

#ifndef __PCENC_H
#define __PCENC_H

#include <iostream>
#include <map>
#include <set>

#include "absenc.h"
#include "cl2lit.h"
#include "enum.h"
#include "glucose_bind.h"
#include "normalform.h"
#include "upenc.h"
#include "varstore.h"

namespace SUBool
{
   class PCEncoder : public AbsEncoder
   {
    public:
      using Config = AbsConfig;

    private:
      Config mConf;

      std::vector<unsigned> mOneProveInputs;
      std::vector<unsigned> mOneProveOutputs;
      std::vector<unsigned> mEmpowerInputs;
      std::vector<unsigned> mEmpowerOutputs; // if empty, then outputs=inputs
      std::vector<Literal> mSelectionLits;

      virtual void AddClauses(const CNF &cnf, const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider) override;

      unsigned NewAuxVar(const std::string &name);
      unsigned NewAuxComparisonVar(
          const std::string &left, const std::string &right, unsigned bit);
      unsigned NewFireVar(
          unsigned c_index, unsigned c_subindex, const Literal &lit);
      unsigned NewFireVar(unsigned c_index, const Literal &lit);
      unsigned EmpowerOutput(unsigned i) const;

      std::unique_ptr<UPEncoder> mEmpEncoder;

      /* S1 (no conflict at the input). */
      void AddOneProveNoInputConflict();
      /* S3 (conflict at output). */
      void AddOneProveOutputConflict();
      /* Block for 1-provability */
      void AddOneProveBlock();

      /* S2 (ExactlyOne on selection variables) */
      void AddSelection();
      /* S4 (Passing values to empowerement block) */
      void AddPassValuesToEmpower();
      /* Block for empowerement. Fills in the arrays of input and
       * output variables for empowerement. Empty outputs mean that
       * inputs are the same as outputs. */
      void AddEmpowerBlock();
      /* S5 (No conflict at output of empowerement) */
      void AddEmpNoConflict();
      /* S6 (Empowered literal not derived) */
      void AddEmpNotDerived();

      /* S1, S3, block for 1-provability. */
      void AddOneProvability();
      /* S2, S4, S5, S6, block for empowerement. */
      void AddEmpowerement();

      void InitEmpEncoder(const CNF &cnf, const ClausesToLiterals &cl2lit,
          std::shared_ptr<const EncISProvider> impl_system_provider);

      virtual void WriteMappingsOfVariable(
          std::ostream &out, unsigned i) const override;

      virtual std::string Name() const override;

    protected:
      virtual const AbsConfig &GetConfig() const override;

    public:
      PCEncoder(const Config &conf = Config());

      // Returns empowering variable.
      virtual unsigned ExtractClause(
          const SolverInterface &solver, Clause *clause) const override;
   };
} // namespace SUBool

#endif
