/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file pcl_negative.h
 * Negative examples maintenance and storage.
 *
 * The compiler of a CNF into PC representation which is based on the
 * algorithm for learning of Horn formulas through equivalence and closure
 * queries (Arias et al. 2015).
 */

#ifndef __PCL_NEGATIVE_H
#define __PCL_NEGATIVE_H

#include <algorithm>
#include <map>
#include <optional>
#include <queue>
#include <set>
#include <unordered_map>
#include <vector>

#include "cnfcompress.h"
#include "impl.h"
#include "literal.h"

namespace SUBool
{

   using NEQueueCompare = ImplBodySizeCompare<Literal>;
   using NEQueue =
       std::priority_queue<LitImpl, std::vector<LitImpl>, NEQueueCompare>;

   LitImpl proto_negative_of_clause(const Clause &clause, const Literal &emp);
   void add_proto_negatives_of_clause(std::vector<LitImpl> &negatives,
       const Clause &clause, const std::vector<Literal> &empowered);

   // returns none if the intersection is equal to b, or it is empty
   std::optional<LitImpl> ne_proper_intersection(
       const LitImpl &a, const LitImpl &b);

   template <class InputIterator>
   std::ostream &print_negatives(std::ostream &out, InputIterator begin,
       InputIterator end, const CNFCompressor &compressor);
   template <class InputIterator>
   std::ostream &print_negatives_clauses(std::ostream &out, InputIterator begin,
       InputIterator end, const CNFCompressor &compressor);
   template <class InputIterator>
   CNF negatives_clauses_cnf(
       InputIterator begin, InputIterator end, const CNFCompressor &compressor);

   struct NEPrinter
   {
      const CNFCompressor &compressor;
      const LitImpl &neg;
      bool print_as_clauses{false};

      NEPrinter(const CNFCompressor &c, const LitImpl &n, bool as_clauses)
          : compressor(c), neg(n), print_as_clauses(as_clauses)
      {
      }
   };

   std::ostream &operator<<(std::ostream &out, const NEPrinter &printer);

} // namespace SUBool

// Template implementation
template <class InputIterator>
std::ostream &
SUBool::print_negatives(std::ostream &out, InputIterator begin,
    InputIterator end, const CNFCompressor &compressor)
{
   std::for_each(begin, end,
       [&out, &compressor](const LitImpl &neg)
       { out << NEPrinter(compressor, neg, false) << std::endl; });
   return out;
}

template <class InputIterator>
std::ostream &
SUBool::print_negatives_clauses(std::ostream &out, InputIterator begin,
    InputIterator end, const CNFCompressor &compressor)
{
   std::for_each(begin, end,
       [&out, &compressor](const LitImpl &neg)
       { out << NEPrinter(compressor, neg, true); });
   return out;
}

template <class InputIterator>
SUBool::CNF
SUBool::negatives_clauses_cnf(
    InputIterator begin, InputIterator end, const CNFCompressor &compressor)
{
   std::vector<Clause> clauses;
   for (auto i_impl = begin; i_impl != end; ++i_impl)
   {
      append_clauses_of_impl(clauses, *i_impl);
   }
   return compressor.UncompressCNF(CNF(std::move(clauses)));
}

#endif

