/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file pclearncomp.cpp
 * The compiler of a CNF into PC representation which is based on the
 * algorithm for learning of Horn formulas through equivalence and closure
 * queries (Arias et al. 2015).
 */

#include <algorithm>

#include "cnfutil.h"
#include "cpuutil.h"
#include "drenc.h"
#include "empower.h"
#include "logstream.h"
#include "pclearncomp.h"
#include "random_utils.h"
#include "sbexception.h"

// TODO: During refinement when a negative example B is replaced in mNegatives
//       with the intersection B', we could try another round of refinements
//       with B'. (Maybe not worth it, low priority)
// TODO: It would be possible to store also the hypothesis with removing the
//       refined negative examples and reentering the queue. We would store both
//       versions, the one without removing (phi_b) and with removing (GD-basis
//       phi_a). phi_b would be kept minimized and used in equivalence queries.
//       phi_a would be stored in form of unit propagator, it would be used to
//       determine negative examples, especially during refinements. See e-mails
//       of Petr Savicky.

// !!! The correctness of the code depends on that body is part of cons in the
// negative example !!!

const SUBool::AbsComp::StateValue SUBool::PCLearnComp::INITIAL_HYPOTHESIS{
    "INITIAL_HYPOTHESIS", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue
    SUBool::PCLearnComp::MINIMIZE_WITH_COLLECTING_EMPOWERED{
        "MINIMIZE_WITH_COLLECTING_EMPOWERED",
        SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue
    SUBool::PCLearnComp::MERGE_DUPLICITIES_IN_NEGATIVES{
        "MERGE_DUPLICITIES_IN_NEGATIVES", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue
    SUBool::PCLearnComp::COMPUTE_CLOSURES_OF_NEGATIVES{
        "COMPUTE_CLOSURES_OF_NEGATIVES", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::PCLearnComp::CONSTRUCT_HYPOTHESIS{
    "CONSTRUCT_HYPOTHESIS", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue SUBool::PCLearnComp::NEXT_NEGATIVE_EXAMPLE{
    "NEXT_NEGATIVE_EXAMPLE", SUBool::AbsComp::RESULT::UNFINISHED};
const SUBool::AbsComp::StateValue
    SUBool::PCLearnComp::REMOVE_DUPLICITIES_IN_NEGATIVES{
        "REMOVE_DUPLICITIES_IN_NEGATIVES", SUBool::AbsComp::RESULT::UNFINISHED};

const SUBool::EnumAnnotation<SUBool::PCLearnComp::INITIAL_NEGATIVES>
    SUBool::PCLearnComp::kInitialNegativesAnnotation("INITIAL_NEGATIVES",
        {//
            {PCLearnComp::INITIAL_NEGATIVES::SINGLE_EMPOWERING,
                "single_empowering",
                "one negative example is produced from every empowering clause "
                "(the clause C without an empowered literal)"}, //
            {PCLearnComp::INITIAL_NEGATIVES::ALL, "all",
                "all negative examples from all clauses are produced, i.e. for "
                "every clause C and a literal l in C, C without l is added as "
                "a negative example regardless on whether C is empowering with "
                "l being an empowered literal."}, //
            {PCLearnComp::INITIAL_NEGATIVES::ALL_EMPOWERING, "all_empowering",
                "for every empowering clause C and every empowered literal l "
                "in C, C without l is added as a negative example to the "
                "initial list."}, //
            {PCLearnComp::INITIAL_NEGATIVES::GD_BASIS, "gd_basis",
                "GD basis of the dual rail encoding of the preprocessed CNF is "
                "computed, the bodies in the GD basis form the initial list of "
                "negative examples."}, //
            {PCLearnComp::INITIAL_NEGATIVES::BODY_MINIMAL, "body_minimal",
                "body minimal representation of the dual rail encoding of the "
                "preprocessed CNF is computed. Its bodies are  used as an "
                "initial list of negative examples."}});

const SUBool::EnumAnnotation<
    SUBool::PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY>
    SUBool::PCLearnComp::kCandidatesFromNegativePolicyAnnotation(
        "CANDIDATES_FROM_ADDED_CLAUSE",
        {//
            {PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY::NONE, "none",
                "No candidates are extracted from the clause added to the "
                "hypothesis"}, //
            {PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY::SYNCHRONIZE,
                "synchronize",
                "Like none, but before the construction of the encoding of PC, "
                "consider the bodies in the implicational system associated "
                "with the CNF and try to reconsider them as candidates on "
                "negative examples."}, //
            {PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY::SEMANTIC, "semantic",
                "Semantic closure is computed and compared with the forward "
                "chaining closure with respect to the implicational system"}, //
            {PCLearnComp::CANDIDATES_FROM_NEGATIVE_POLICY::UNITPROP, "unitprop",
                "It is only checked if the set of literals is closed under "
                "unit propagation in the current CNF hypothesis"}});

SUBool::PCLearnComp::PCLearnComp(CNF cnf, Config conf)
    : AbsComp(cnf, conf.minimize_method, conf.impl_system_rebuild),
      mConf(std::move(conf)), mNegatives(), mCandidatesWithClosure(),
      mCandidatesWithoutClosure(), mClosureCache(), mPCCheckerSAT(),
      mPCCheckRandomBudget(mConf.pccheck_random_budget)
{
   mConf.preprocess = AbsComp::PMRemoveMinimize(mConf.preprocess);
   mPCCheckerSAT.Init(mConf.EncoderConfig(), true, mConf.timeouts);
   mPCCheckerSAT.SetSolverType(mConf.solver_type);
   if (mPCCheckerSAT.RequiresImplSystem()
       || mConf.candidates_from_negative_policy
              == CANDIDATES_FROM_NEGATIVE_POLICY::SYNCHRONIZE)
   {
      mHypothesis.SetImplSystemProvider(
          conf.use_cnf_hypothesis_for_empowerment
              ? make_enc_is_provider(conf.impl_system)
              : make_negatives_is_provider(conf.impl_system, mNegatives));
   }
   else
   {
      mHypothesis.SetImplSystemProvider(
          std::make_shared<NegDummyISProvider>(mNegatives));
   }
}

void
SUBool::PCLearnComp::CompilePreprocessed()
{
   logs.TLog(2) << "The main cycle of compilation starts" << std::endl;
   ComputeInitialCandidates();
   mHypothesis.SetHypothesisToRepresentation();
   logs.TLog(2) << "Initialization finished, entering main cycle" << std::endl;
   while (auto e = NextNegativeExample())
   {
      if (!Refine(*e))
      {
         AddNegativeExample(std::move(*e));
      }
      if (mCandidatesWithClosure.empty())
      {
         mHypothesis.SetSafe();
      }
      if (mConf.minimize_steps > 0
          && mAddedClausesSinceMinimization >= mConf.minimize_steps)
      {
         MinimizeHypothesis();
      }
   }
   if (!CheckState(TIMED_OUT))
   {
      mHypothesis.SetSafe();
      if (mConf.final_minimize)
      {
         MinimizeHypothesis();
      }
      State(FINISHED);
   }
   PrintStatistics(1);
}

std::string
SUBool::PCLearnComp::ResultDesc() const
{
   switch (StateResult())
   {
   case RESULT::NOT_RUN:
   case RESULT::COMPILED:
      return "Compilation finished successfully, formula is propagation "
             "complete";
      break;
   case RESULT::UNFINISHED:
      return "Compilation in progress";
      break;
   case RESULT::EMP_FOUND:
      return "An empowering implicate was found, formula is not propagation "
             "complete";
      break;
   case RESULT::EMP_NOT_FOUND:
      return "An empowering implicate was not found, formula is propagation "
             "complete";
      break;
   case RESULT::ERROR:
      return "Compilation failed due to an error";
      break;
   case RESULT::TIMED_OUT:
      return "Compilation timed out";
      break;
   case RESULT::TERMINATED:
      return "Compilation terminated";
      break;
   case RESULT::FAILED:
      return "Compilation failed";
      break;
   default:
      return "Unknown result";
   }
}

std::string
SUBool::PCLearnComp::Annotation() const
{
   // TODO: Add meaningful annotation, don't use mCurrent
   std::string s = "c Using the learning algorithm\n";
   if (CheckState(UNSAT))
   {
      s += "c The formula is UNSATISFIABLE";
      return s;
   }
   if (!mHypothesis.Safe())
   {
      s += "c The hypothesis is not equivalent to the input CNF\n";
      s += "c Returning the empty formula";
      return s;
   }
   s += "c " + ResultDesc();
   return s;
}

void
SUBool::PCLearnComp::PrintStatistics(unsigned level) const
{
   if (logs.CheckLevel(4))
   {
      logs.TLog(4) << "Final negatives list:" << '\n';
      const auto &comp = mHypothesis.Compressor();
      mNegatives.PrettyPrint(std::cout,
          [&comp](const Literal &lit) { return comp.UncompressLiteral(lit); });
      logs.TLog(4) << "Final negatives clauses:" << '\n';
      std::cout << comp.UncompressCNF(dual_rail_to_cnf(mNegatives)) << '\n';
   }
   logs.TLog(level) << ResultDesc() << '\n';
   logs.TLog(level) << "Total time: " << mRTime << '\n';
   logs.TLog(level) << "Total processor time: " << mCPUTime << '\n';
   logs.TLog(level) << "Processor time until the last SAT based EQ check: "
                    << mTimeTillTheLastSATEqCheck << '\n';
   logs.TLog(level) << "Processor time of the last SAT based EQ check: "
                    << mTimeOfLastSATEqCheck << '\n';
   logs.TLog(level)
       << "Processor time of the last SAT solver call within an EQ check: "
       << mPCCheckerSAT.LastSolveTime() << '\n';
   logs.TLog(level) << "Negatives added to the hypothesis: "
                    << mNegativesAddedToHypothesis << '\n';
   logs.TLog(level) << "Clauses added to the hypothesis: "
                    << mClausesAddedToHypothesis << '\n';
   logs.TLog(level) << "Number of successful refinements: "
                    << mSuccessfulRefinements << '\n';
   logs.TLog(level) << "Number of initial candidates: "
                    << mInitialCandidatesListSize << '\n';
   logs.TLog(level) << "Total number of candidates with closure: "
                    << mCandidatesWithClosureCount << '\n';
   logs.TLog(level) << "Total number of candidates without closure: "
                    << mCandidatesWithoutClosureCount << '\n';
   logs.TLog(level) << "Total number of learned clauses considered: "
                    << mLearnedClausesConsideredCount << '\n';
   logs.TLog(level) << "Total number of negative examples extracted from the "
                       "learned clauses: "
                    << mLearnedClausesCandidatesCount << '\n';
   logs.TLog(level) << "Total number of random empowering implicates: "
                    << mRandomEmpoweringCount << '\n';
   logs.TLog(level) << "Total number of random bodies: " << mRandomBodyCount
                    << '\n';
   logs.TLog(level) << "Total number of learned clauses from random queries: "
                    << mRandomLearnedCount << '\n';
   logs.TLog(level) << "Total number of random queries: " << mRandomQueriesCount
                    << '\n';
   logs.TLog(level) << "Candidates extracted from new negatives: "
                    << mCandidatesFromNegativesCount << '\n';
   logs.TLog(level) << "Total number of empowering implicates found by SAT: "
                    << mSATEmpoweringCount << '\n';
   logs.TLog(level) << "Total time of SAT based equivalence queries: "
                    << mPCCheckerSAT.TotalTime() << "s" << '\n';
   logs.TLog(level) << "SAT based equivalence with result SAT (time/count): "
                    << mPCCheckerSAT.TotalTimeSAT() << "s / "
                    << mPCCheckerSAT.SolveCountSAT() << '\n';
   logs.TLog(level) << "SAT based equivalence with result UNSAT (time/count): "
                    << mPCCheckerSAT.TotalTimeUNSAT() << "s / "
                    << mPCCheckerSAT.SolveCountUNSAT() << '\n';
   logs.TLog(level) << "Maximum UP level: "
                    << mPCCheckerSAT.CurrentUPLevelBound() << '\n';
   logs.TLog(level) << "Closure cache hit count: " << mCacheHitCount << '\n';
   logs.TLog(level) << "Closure cache miss count: " << mCacheMissCount << '\n';
   logs.TLog(level) << "Closure cache find total time: " << mCacheFindTotalTime
                    << '\n';
   logs.TLog(level) << "Number of intersections: " << mIntersectionCount
                    << '\n';
   logs.TLog(level) << "Total time spent with intersections: "
                    << mIntersectionTotalTime << '\n';
   logs.TLog(level) << "Total time spent in RefinedNegative: "
                    << mRefinedNegativeTotalTime << '\n';
   logs.TLog(level) << "Total time of semantic closure computation: "
                    << mSemanticClosureTotalTime << '\n';
   logs.TLog(level)
       << "Total number of SAT calls in semantic closure computation: "
       << mSemanticClosureSolveCount << '\n';
   logs.TLog(level) << "Number of semantic closure computations: "
                    << mSemanticClosureCount << '\n';
   double avg_semantic_closure_solve_count =
       (mSemanticClosureCount > 0
               ? static_cast<double>(mSemanticClosureSolveCount)
                     / static_cast<double>(mSemanticClosureCount)
               : 0.0);
   logs.TLog(level)
       << "Average number of SAT calls in semantic closure computations: "
       << avg_semantic_closure_solve_count << '\n';
   logs.TLog(level)
       << "CSV " << mRTime.Duration() << ';' << mCPUTime.TimeSpan() << ';'
       << mTimeTillTheLastSATEqCheck << ';' << mTimeOfLastSATEqCheck << ';'
       << mPCCheckerSAT.LastSolveTime() << ';' << mNegativesAddedToHypothesis
       << ';' << mClausesAddedToHypothesis << ';' << mSuccessfulRefinements
       << ';' << mCandidatesWithClosureCount << ';'
       << mCandidatesWithoutClosureCount << ';'
       << mLearnedClausesConsideredCount << ';'
       << mLearnedClausesCandidatesCount << ';' << mRandomEmpoweringCount << ';'
       << mRandomBodyCount << ';' << mRandomLearnedCount << ';'
       << mRandomQueriesCount << ';' << mCandidatesFromNegativesCount << ';'
       << mSATEmpoweringCount << ';' << mPCCheckerSAT.TotalTime() << ';'
       << mPCCheckerSAT.TotalTimeSAT() << ';' << mPCCheckerSAT.SolveCountSAT()
       << ';' << mPCCheckerSAT.TotalTimeUNSAT() << ';'
       << mPCCheckerSAT.SolveCountUNSAT() << ';' << ';'
       << mPCCheckerSAT.CurrentUPLevelBound() << ';' << mCacheHitCount << ';'
       << mCacheMissCount << ';' << mCacheFindTotalTime << ';'
       << mIntersectionCount << ';' << mIntersectionTotalTime << ';'
       << mRefinedNegativeTotalTime << ';' << mSemanticClosureTotalTime << ';'
       << mSemanticClosureSolveCount << ';' << mSemanticClosureCount << ';'
       << avg_semantic_closure_solve_count << std::endl;
}

void
SUBool::PCLearnComp::ComputeInitialCandidates()
{
   if (!State(INITIAL_HYPOTHESIS))
   {
      return;
   }
   logs.TLog(2) << "Constructing initial hypothesis (p cnf "
                << mHypothesis.ReprMaxVariable() << " "
                << mHypothesis.ReprSize() << ")" << std::endl;
   CPUTime cpu_time;
   MoveNegativesToCandidates(InitialCandidatesList());
   logs.TLog(2) << "Finished initial candidates list, time=" << cpu_time
                << std::endl;
}

void
SUBool::PCLearnComp::MoveNegativesToCandidates(std::vector<LitImpl> negatives)
{
   logs.TLog(2) << "Moving " << negatives.size()
                << " initial candidates to the queue" << std::endl;
   mInitialCandidatesListSize = negatives.size();
   if (logs.CheckLevel(4))
   {
      logs.TLog(4) << "Initial negatives list:" << std::endl;
      const auto &comp = mHypothesis.Compressor();
      LitImplSystem negatives_is(negatives);
      negatives_is.PrettyPrint(std::cout,
          [&comp](const Literal &lit) { return comp.UncompressLiteral(lit); });
      logs.TLog(4) << "Initial negatives clauses:" << std::endl;
      std::cout << comp.UncompressCNF(dual_rail_to_cnf(negatives_is))
                << std::endl;
   }
   mCandidatesWithoutClosure = NEQueue(NEQueueCompare(), std::move(negatives));
}

std::vector<SUBool::LitImpl>
SUBool::PCLearnComp::InitialCandidatesList()
{
   switch (mConf.initial_negatives)
   {
   case INITIAL_NEGATIVES::SINGLE_EMPOWERING:
      return InitialCandidatesEmpowering(false);
   case INITIAL_NEGATIVES::ALL:
      return InitialCandidatesAll();
   case INITIAL_NEGATIVES::ALL_EMPOWERING:
      return InitialCandidatesEmpowering(true);
   case INITIAL_NEGATIVES::GD_BASIS:
      return InitialCandidatesGDBasis();
   case INITIAL_NEGATIVES::BODY_MINIMAL:
      return InitialCandidatesBodyMinimal();
   default:
      throw BadParameterException("PCLearnComp::InitialCandidatesList",
          "Unknown INITIAL_NEGATIVES value");
   }
}

std::vector<SUBool::LitImpl>
SUBool::PCLearnComp::InitialCandidatesAll()
{
   if (!State(MINIMIZE_WITH_COLLECTING_EMPOWERED))
   {
      return {};
   }
   std::vector<LitImpl> negatives;
   mHypothesis.MinimizeRepresentation();
   for (const auto &clause : mHypothesis.GetRepresentation(false))
   {
      add_proto_negatives_of_clause(negatives, clause, clause.Literals());
   }
   return RemoveDuplicitiesInNegatives(std::move(negatives));
}

std::vector<SUBool::LitImpl>
SUBool::PCLearnComp::InitialCandidatesEmpowering(bool collect_all)
{
   if (!State(MINIMIZE_WITH_COLLECTING_EMPOWERED))
   {
      return {};
   }
   std::vector<LitImpl> negatives;
   for (const auto &clause_we :
       mHypothesis.MinimizeReprWithCollectingEmpowered(collect_all))
   {
      add_proto_negatives_of_clause(
          negatives, clause_we.clause, clause_we.empowered);
   }
   return RemoveDuplicitiesInNegatives(std::move(negatives));
}

std::vector<SUBool::LitImpl>
SUBool::PCLearnComp::InitialCandidatesGDBasis()
{
   auto is = dual_rail(mHypothesis.GetRepresentation(false), true);
   is.MakeGDBasis();
   return std::move(is).Implications();
}

std::vector<SUBool::LitImpl>
SUBool::PCLearnComp::InitialCandidatesBodyMinimal()
{
   auto is = dual_rail(mHypothesis.GetRepresentation(false), true);
   is.MakeBodyMinimal();
   return std::move(is).Implications();
}

std::vector<SUBool::LitImpl>
SUBool::PCLearnComp::RemoveDuplicitiesInNegatives(
    std::vector<LitImpl> negatives)
{
   if (!State(REMOVE_DUPLICITIES_IN_NEGATIVES))
   {
      return negatives;
   }
   CPUTime cpu_time;
   logs.TLog(3) << "Removing duplicities in negatives" << std::endl;
   std::sort(negatives.begin(), negatives.end(), ImplBodyCompare<Literal>());
   auto end = std::unique(
       negatives.begin(), negatives.end(), ImplEqualBody<Literal>());
   logs.TLog(3) << "Resizing negatives from " << negatives.size() << " to "
                << end - negatives.begin() << std::endl;
   negatives.resize(end - negatives.begin());
   logs.TLog(3) << "Finished removing duplicities in negatives, time="
                << cpu_time << std::endl;
   return negatives;
}

std::optional<std::vector<SUBool::Literal>>
SUBool::PCLearnComp::SemanticClosure(const std::vector<Literal> &body)
{
   CPUTime cpu_time;
   PartialAssignment alpha(mHypothesis.ReprMaxVariable());
   alpha.SetLiterals(body.begin(), body.end());
   auto solve_cnt = SolverInterface::SolveCount();
   auto beta = mHypothesis.PropagateAllSAT(alpha);
   solve_cnt = SolverInterface::SolveCount() - solve_cnt;
   mSemanticClosureSolveCount += solve_cnt;
   ++mSemanticClosureCount;
   if (!beta)
   {
      mSemanticClosureTotalTime += cpu_time.TimeSpan();
      logs.TLog(4) << "[SemanticClosure] UNSAT, sat count=" << solve_cnt
                   << ", time=" << cpu_time << std::endl;
      return {};
   }
   std::vector<Literal> closure(literals_of_assignment(*beta));
   mSemanticClosureTotalTime += cpu_time.TimeSpan();
   logs.TLog(4) << "[SemanticClosure] closure, sat count=" << solve_cnt
                << ", time=" << cpu_time << std::endl;
   return closure;
}

std::optional<std::vector<SUBool::Literal>>
SUBool::PCLearnComp::SemanticClosureWithCaching(
    const std::vector<Literal> &body)
{
   auto i_cached = CachedClosure(body);
   if (i_cached != mClosureCache.end())
   {
      return i_cached->second; //(*i_cached)->first.Cons();
   }
   auto closure = SemanticClosure(body);
   mClosureCache.insert({body, *closure});
   return closure;
}

SUBool::PCLearnComp::NECache::const_iterator
SUBool::PCLearnComp::CachedClosure(const std::vector<Literal> &body) const
{
   CPUTime cpu_time;
   auto i_cached_closure = mClosureCache.find(body);
   mCacheFindTotalTime += cpu_time.TimeSpan();
   if (i_cached_closure != mClosureCache.end())
   {
      ++mCacheHitCount;
      if (cpu_time.TimeSpan() > mConf.cache_time_logging_threshold)
      {
         logs.TLog(5) << "[CachedClosure] Cache hit, size="
                      << mClosureCache.size() << ", time=" << cpu_time
                      << std::endl;
      }
      return i_cached_closure;
   }
   ++mCacheMissCount;
   if (cpu_time.TimeSpan() > mConf.cache_time_logging_threshold)
   {
      logs.TLog(5) << "[CachedClosure] Cache miss, size="
                   << mClosureCache.size() << ", time=" << cpu_time
                   << std::endl;
   }
   return i_cached_closure;
}

bool
SUBool::PCLearnComp::UpdateClosureWithMinimizing(LitImpl &neg)
{
   // logs.TLog(3) << "UpdateClosureWithMinimizing, negative="
   //            << neg.Printer(mHypothesis.Compressor()) << std::endl;
   bool nonprime = false;
   auto body = std::move(neg).Body();
   // Should do it differently, if the negation of the body is an implicate,
   // then we should find a prime subimplicate
   // Actually, this is fine, it finds a subset which is not an implicate (when
   // negated), but it implies a value from the original body (because with it
   // it was an implicate).
   auto beta = SemanticClosureWithCaching(body);
   while (!beta && !body.empty())
   {
      std::cout << "ssss" << std::endl;
      nonprime = true;
      body.pop_back();
      beta = SemanticClosureWithCaching(body);
   }
   if (body.empty())
   {
      throw InvariantFailedException("PCLearnComp::UpdateClosureWithMinimizing",
          "Unit or empty implicate found, should not happen at this stage");
   }
   neg = LitImpl(body, *beta, true);
   return nonprime;
}

std::optional<SUBool::LitImpl>
SUBool::PCLearnComp::UpdateClosureWithCaching(const LitImpl &neg, bool minimize)
{
   if (minimize)
   {
      LitImpl new_neg(neg);
      UpdateClosureWithMinimizing(new_neg);
      return new_neg;
   }
   if (auto cl = SemanticClosureWithCaching(neg.Body()))
   {
      return LitImpl(neg.Body(), *cl, true);
   }
   return {};
}

std::optional<SUBool::LitImpl>
SUBool::PCLearnComp::NextNegativeExample()
{
   if (!State(NEXT_NEGATIVE_EXAMPLE))
   {
      return {};
   }
   logs.TLog(3) << "Getting next negative" << std::endl;
   std::optional<LitImpl> neg;
   while (!neg)
   {
      if (!mCandidatesWithClosure.empty())
      {
         logs.TLog(3) << "Have a candidate with closure" << std::endl;
         ++mCandidatesWithClosureCount;
         logs.TLog(3) << "... "
                      << NEPrinter(mHypothesis.Compressor(),
                             mCandidatesWithClosure.top(), false)
                      << std::endl;
         neg = CheckNegativeWithBodyUpdate(mCandidatesWithClosure.top());
         if (neg)
         {
            logs.TLog(3) << "... "
                         << NEPrinter(mHypothesis.Compressor(), *neg, false)
                         << std::endl;
         }
         else
         {
            logs.TLog(3) << "... (none)" << std::endl;
         }
         mCandidatesWithClosure.pop();
      }
      else if (!mCandidatesWithoutClosure.empty())
      {
         logs.TLog(3) << "Have a candidate without closure "
                      << NEPrinter(mHypothesis.Compressor(),
                             mCandidatesWithoutClosure.top(), false)
                      << std::endl;
         ++mCandidatesWithoutClosureCount;
         neg = UpdateCandidate(mCandidatesWithoutClosure.top());
         if (neg)
         {
            logs.TLog(3) << "... "
                         << NEPrinter(mHypothesis.Compressor(), *neg, false)
                         << std::endl;
         }
         else
         {
            logs.TLog(3) << "... (none)" << std::endl;
         }
         mCandidatesWithoutClosure.pop();
      }
      else if (auto c = mHypothesis.NextLearnedClause())
      {
         logs.TLog(3) << "Considering a learned clause ("
                      << mHypothesis.LearnedSize() << " remaining)"
                      << std::endl;
         ++mLearnedClausesConsideredCount;
         mLearnedClausesCandidatesCount += AddNegativesForImplicate(*c, true);
      }
      else
      {
         logs.TLog(3) << "No candidate, looking for an empowering implicate"
                      << std::endl;
         if (CheckPC())
         {
            break;
         }
      }
   }
   if (!neg)
   {
      if (mPCCheckerSAT.TimedOut())
      {
         logs.TLog(3) << "PC Check with SAT timed out." << std::endl;
         State(TIMED_OUT);
      }
      else
      {
         logs.TLog(3) << "No candidate, hypothesis is PC" << std::endl;
      }
      return {};
   }
   return neg;
}

bool
SUBool::PCLearnComp::CheckNegativeExample(const LitImpl &neg)
{
   if (neg.ConsSize() > neg.BodySize())
   {
      return true;
   }
   return false;
}

std::optional<SUBool::LitImpl>
SUBool::PCLearnComp::CheckNegativeWithBodyUpdate(const LitImpl &neg)
{
   auto closure = mNegatives.Closure(neg.Body());
   assert(!Clause(closure).Trivial());
   LitImpl saturated(std::move(closure), neg.Cons(), true);
   if (CheckNegativeExample(saturated))
   // if (neg.ConsSize() > neg.BodySize())
   {
      return saturated;
   }
   return {};
}

std::optional<SUBool::LitImpl>
SUBool::PCLearnComp::UpdateCandidate(const LitImpl &neg)
{
   auto e = UpdateClosureWithCaching(neg, true);
   if (!e)
   {
      return {};
   }
   return CheckNegativeWithBodyUpdate(*e);
}

bool
SUBool::PCLearnComp::CheckPC()
{
   return CheckPCRandomly() && !SynchronizeHypothesis() && CheckPCWithSAT();
}

bool
SUBool::PCLearnComp::SynchronizeHypothesis()
{
   if (!mHypothesisSyncNeeded)
   {
      return false;
   }
   mHypothesisSyncNeeded = false;
   const auto &is_provider = mHypothesis.ImplSystemProvider();
   const auto &is = is_provider.OneProveImplSystem();
   bool found_candidate{false};
   for (const auto &impl : is)
   {
      auto i_cached_closure = CachedClosure(impl.Body());
      if (i_cached_closure != mClosureCache.end())
      {
         continue;
      }
      found_candidate = true;
      ++mCandidatesFromNegativesCount;
      mCandidatesWithoutClosure.push(LitImpl{impl.Body(), {}, true});
   }
   return found_candidate;
}

bool
SUBool::PCLearnComp::CheckPCRandomly()
{
   if (mPCCheckRandomBudget == 0)
   {
      //   logs.TLog(2) << "... budget empty, no random set generated" <<
      //   std::endl;
      return true;
   }
   logs.TLog(3) << "Checking PC with random sets of literals" << std::endl;
   CPUTime cpu_time;
   std::geometric_distribution<unsigned> gd(
       mConf.pccheck_random_geometry_mult
       / static_cast<double>(mHypothesis.MaxVariable() - 2));
   unsigned round = 0;
   for (; round < mConf.pccheck_random_rounds && !HasNegative(); ++round)
   {
      // Use geometric distribution
      unsigned var_size = mHypothesis.MaxVariable() + 1;
      while (var_size > mHypothesis.MaxVariable())
      {
         var_size = 2 + gd(sub_rand);
      }
      auto literals =
          random_literal_vector_with_size(var_size, mHypothesis.MaxVariable());
      ++mRandomQueriesCount;
      Clause c(literals);
      logs.TLog(3) << "Trying random clause "
                   << mHypothesis.Compressor().UncompressClause(c) << std::endl;
      if (mHypothesis.IsImplicate(c))
      {
         CheckRandomImplicate(c);
         continue;
      }
      CheckRandomBody(std::move(literals));
   }
   if (HasNegative())
   {
      logs.TLog(3) << "... found a candidate by random, round=" << round
                   << std::endl;
      mRandomLearnedCount += mHypothesis.LearnedSize();
      mPCCheckRandomBudget += mConf.pccheck_random_budget_inc;
      return false;
   }
   logs.TLog(3) << "No empowering implicate found by random" << std::endl;
   --mPCCheckRandomBudget;
   return true;
}

void
SUBool::PCLearnComp::CheckRandomImplicate(const Clause &c)
{
   logs.TLog(4) << "Checking random implicate" << std::endl;
   if (AddNegativesForImplicate(c, true) > 0)
   {
      logs.TLog(3) << "... found candidates for a random implicate"
                   << std::endl;
      ++mRandomEmpoweringCount;
      return;
   }
   logs.TLog(4) << "No candidates for the implicate" << std::endl;
}

void
SUBool::PCLearnComp::CheckRandomBody(std::vector<Literal> literals)
{
   logs.TLog(4) << "Checking random body" << std::endl;
   for (auto &lit : literals)
   {
      lit.Negate();
   }
   if (AddCandidateIfNegative(LitImpl(std::move(literals), {}, true)))
   {
      logs.TLog(3) << "... found a random body, round=" << std::endl;
      ++mRandomBodyCount;
   }
}

bool
SUBool::PCLearnComp::HasNegative()
{
   return !mCandidatesWithClosure.empty() || !mCandidatesWithoutClosure.empty()
          || mHypothesis.HasLearnedClause();
}

bool
SUBool::PCLearnComp::CheckPCWithSAT()
{
   logs.TLog(2) << "Checking PC with SAT" << std::endl;
   CPUTime cpu_time;
   mTimeTillTheLastSATEqCheck = mCPUTime.TimeSpan();
   auto emp = mHypothesis.FindEmpowering(mPCCheckerSAT, mConf.level_minimize);
   if (mHypothesis.HypMinimized())
   {
      mAddedClausesSinceMinimization = 0;
   }
   mTimeOfLastSATEqCheck = cpu_time.TimeSpan();
   logs.TLog(2) << "PC check finished, time=" << mTimeOfLastSATEqCheck << 's'
                << std::endl;
   if (!emp)
   {
      logs.TLog(2) << "No empowering implicate found" << std::endl;
      return true;
   }
   ++mSATEmpoweringCount;
   AddNegativesForImplicate(*emp, true);
   return false;
}

unsigned
SUBool::PCLearnComp::AddNegativesForImplicate(
    const Clause &clause, bool find_prime_subimplicate)
{
   logs.TLog(3) << "Considering implicate "
                << mHypothesis.Compressor().UncompressClause(clause)
                << std::endl;
   Clause c;
   if (find_prime_subimplicate)
   {
      c = *mHypothesis.PrimeSubimplicate(clause, 0, true);
      logs.TLog(3) << "Found prime subimplicate "
                   << mHypothesis.Compressor().UncompressClause(c) << std::endl;
   }
   else
   {
      c = clause;
   }
   c.NegateLiterals();
   unsigned candidates_cnt = 0;
   for (const auto &lit : c)
   {
      std::vector<Literal> body;
      std::remove_copy(c.begin(), c.end(), std::back_inserter(body), lit);
      if (AddCandidateIfNegative(LitImpl(std::move(body), {}, true)))
      {
         ++candidates_cnt;
      }
   }
   logs.TLog(3) << "Added " << candidates_cnt << " candidates for the clause"
                << std::endl;
   return candidates_cnt;
}

bool
SUBool::PCLearnComp::AddCandidateIfNegative(const LitImpl &candidate)
{
   auto neg = UpdateCandidate(candidate);
   if (!neg)
   {
      return false;
   }
   mCandidatesWithClosure.push(*std::move(neg));
   return true;
}

bool
SUBool::PCLearnComp::AddCandidateIfNotUPClosed(const std::vector<Literal> &body)
{
   auto closure = mNegatives.Closure(body);
   assert(!Clause(closure).Trivial());
   if (mHypothesis.UPClosed(
           PartialAssignment(mHypothesis.MaxVariable(), closure)))
   {
      return false;
   }
   mCandidatesWithoutClosure.push(LitImpl(body, {}, true));
   return true;
}

bool
SUBool::PCLearnComp::Refine(const LitImpl &e)
{
   logs.TLog(3) << "Refining negative "
                << NEPrinter(mHypothesis.Compressor(), e, false) << std::endl;
   logs.TLog(3) << "... number of negatives=" << mNegatives.Size() << std::endl;
   CPUTime cpu_time;
   auto intersect_count = mIntersectionCount;
   auto intersect_time = mIntersectionTotalTime;
   auto cache_time = mCacheFindTotalTime;
   auto cache_hit = mCacheHitCount;
   auto cache_miss = mCacheMissCount;
   auto closure_time = mSemanticClosureTotalTime;
   auto refined_negative_time = mRefinedNegativeTotalTime;
   for (size_t i = 0; i < mNegatives.Size(); ++i)
   {
      auto neg = TryToRefine(e, mNegatives.At(i));
      if (neg)
      {
         ++mSuccessfulRefinements;
         mCandidatesWithClosure.push(mNegatives.At(i));
         mCandidatesWithClosure.push(e);
         AddNegativeToHypothesis(*neg);
         logs.TLog(3) << "Refining "
                      << NEPrinter(
                             mHypothesis.Compressor(), mNegatives.At(i), false)
                      << " to "
                      << NEPrinter(mHypothesis.Compressor(), *neg, false)
                      << std::endl;
         mNegatives.Refine(i, std::move(*neg));
         logs.TLog(3) << "Refinement finished (refined), round=" << i
                      << ", time=" << cpu_time << std::endl;
         logs.TLog(3) << "... " << mIntersectionCount - intersect_count
                      << " intersections, time="
                      << mIntersectionTotalTime - intersect_time << "s"
                      << std::endl;
         logs.TLog(3) << "... cache: " << mCacheHitCount - cache_hit
                      << " hits, " << mCacheMissCount - cache_miss
                      << " misses, time=" << mCacheFindTotalTime - cache_time
                      << "s" << std::endl;
         logs.TLog(3) << "... semantic closure time="
                      << mSemanticClosureTotalTime - closure_time << "s"
                      << std::endl;
         logs.TLog(3) << "... refined negative time="
                      << mRefinedNegativeTotalTime - refined_negative_time
                      << "s" << std::endl;
         return true;
      }
   }
   logs.TLog(3) << "Refinement finished (still negative), time=" << cpu_time
                << std::endl;
   logs.TLog(3) << "... " << mIntersectionCount - intersect_count
                << " intersections, time="
                << mIntersectionTotalTime - intersect_time << "s" << std::endl;
   logs.TLog(3) << "... cache: " << mCacheHitCount - cache_hit << " hits, "
                << mCacheMissCount - cache_miss
                << " misses, time=" << mCacheFindTotalTime - cache_time << "s"
                << std::endl;
   logs.TLog(3) << "... semantic closure time="
                << mSemanticClosureTotalTime - closure_time << "s" << std::endl;
   logs.TLog(3) << "... refined negative time="
                << mRefinedNegativeTotalTime - refined_negative_time << "s"
                << std::endl;
   return false;
}

std::optional<SUBool::LitImpl>
SUBool::PCLearnComp::TryToRefine(const LitImpl &e, const LitImpl &neg)
{
   CPUTime cpu_time;
   auto intersect = ne_proper_intersection(e, neg);
   mIntersectionTotalTime += cpu_time.TimeSpan();
   ++mIntersectionCount;
   if (!intersect)
   {
      return {};
   }
   auto closed_intersect = UpdateClosureWithCaching(*intersect, false);
   if (!closed_intersect
       || closed_intersect->BodySize() == closed_intersect->ConsSize())
   {
      return {};
   }
   return closed_intersect;
}

void
SUBool::PCLearnComp::AddNegativeExample(LitImpl e)
{
   AddNegativeToHypothesis(e);
   mNegatives.PushBack(std::move(e));
}

// TODO: It would be faster to use mNegatives as an implication system for
// checking PC (if using implication system for the check)
// Also, hypothesis minimization is not essential if using implication system,
// the implication system can be used also for checking empowerement. mNegatives
// is reduced as it is a GD basis anyway.
void
SUBool::PCLearnComp::AddNegativeToHypothesis(const LitImpl &e)
{
   CPUTime cpu_time;
   logs.TLog(3) << "Adding a negative example to hypothesis ("
                << NEPrinter(mHypothesis.Compressor(), e, false) << ")"
                << std::endl;
   ++mNegativesAddedToHypothesis;
   auto clauses = dr_clauses_of_impl(e);
   for (const auto &clause : clauses)
   {
      auto c = mHypothesis.PrimeSubimplicate(clause, 0, true);
      if (mHypothesis.PushIfEmpowering(*c, false))
      {
         ++mClausesAddedToHypothesis;
         ++mAddedClausesSinceMinimization;
         mCandidatesFromNegativesCount += AddCandidatesForAddedClause(*c);
      }
   }

   logs.TLog(3) << "Negative example added to hypothesis, time=" << cpu_time
                << std::endl;
}

unsigned
SUBool::PCLearnComp::AddCandidatesForAddedClause(const Clause &clause)
{
   switch (mConf.candidates_from_negative_policy)
   {
   case CANDIDATES_FROM_NEGATIVE_POLICY::NONE:
      return 0;
   case CANDIDATES_FROM_NEGATIVE_POLICY::SYNCHRONIZE:
      mHypothesisSyncNeeded = true;
      return 0;
   case CANDIDATES_FROM_NEGATIVE_POLICY::SEMANTIC:
      return AddNegativesForImplicate(clause, false);
   case CANDIDATES_FROM_NEGATIVE_POLICY::UNITPROP:
      return AddCandidatesForAddedClauseUnitProp(clause);
   }
   throw BadParameterException("PCLearnComp::AddCandidatesForAddedClause",
       "Unknown candidates_from_negative_policy");
}

unsigned
SUBool::PCLearnComp::AddCandidatesForAddedClauseUnitProp(
    const Clause &clause [[maybe_unused]])
{
   logs.TLog(3) << "Adding candidates using unit propagation for clause "
                << mHypothesis.Compressor().UncompressClause(clause)
                << std::endl;
   unsigned cnt = 0;
   for (const auto &lit : clause)
   {
      std::vector<Literal> body;
      std::remove_copy(
          clause.begin(), clause.end(), std::back_inserter(body), lit);
      std::for_each(
          body.begin(), body.end(), [](Literal &lit) { lit.Negate(); });
      if (AddCandidateIfNotUPClosed(body))
      {
         ++cnt;
      }
   }
   return cnt;
}

void
SUBool::PCLearnComp::MinimizeHypothesis()
{
   if (!State(MINIMIZE))
   {
      return;
   }
   CPUTime cpu_time;
   logs.TLog(2) << "Minimizing hypothesis" << std::endl;
   mHypothesis.MinimizeHypothesis();
   mAddedClausesSinceMinimization = 0;
   logs.TLog(2) << "Hypothesis minimized (p cnf " << mHypothesis.MaxVariable()
                << " " << mHypothesis.Size() << "), time=" << cpu_time
                << std::endl;
}

SUBool::PCEncoder::Config
SUBool::PCLearnComp::Config::EncoderConfig()
{
   return enc_conf;
}

const SUBool::AbsComp::BaseConfig &
SUBool::PCLearnComp::GetBaseConfig() const
{
   return mConf;
}
SUBool::AbsCheckerSAT &
SUBool::PCLearnComp::CheckerSAT()
{
   return mPCCheckerSAT;
}

const SUBool::AbsCheckerSAT &
SUBool::PCLearnComp::CheckerSAT() const
{
   return mPCCheckerSAT;
}
