/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file pclearncomp.h
 * The compiler of a CNF into PC representation which is based on the
 * algorithm for learning of Horn formulas through equivalence and closure
 * queries (Arias et al. 2015).
 */

#ifndef __PCLEARNCOMP_H
#define __PCLEARNCOMP_H

#include <array>
#include <atomic>
#include <queue>
#include <random>

#include "abscomp.h"
#include "absenc.h"
#include "cnfcompress.h"
#include "comp_hypothesis.h"
#include "cpuutil.h"
#include "empower.h"
#include "enum.h"
#include "impl_system.h"
#include "implic_learn.h"
#include "pcchecker.h"
#include "pcenc.h"
#include "pcl_negative.h"

namespace SUBool
{
   class PCLearnComp : public AbsComp
   {
    public:
      enum class INITIAL_NEGATIVES
      {
         SINGLE_EMPOWERING,
         ALL,
         ALL_EMPOWERING,
         GD_BASIS,
         BODY_MINIMAL,
         Last = BODY_MINIMAL

      };
      static const EnumAnnotation<INITIAL_NEGATIVES>
          kInitialNegativesAnnotation;

      enum class CANDIDATES_FROM_NEGATIVE_POLICY
      {
         NONE,        // No candidates extracted from the clause added to the
                      // hypothesis
         SYNCHRONIZE, // Regularly synchronize the CNF hypothesis with the IS
                      // hypothesis by reconsidering the bodies of the
                      // implicational system associated with the CNF hypothesis
         SEMANTIC,    // Semantic closure is used to determine if the set of
                      // literals is a negative example
         UNITPROP     // Only UP closures of the hypothesis with the implication
                      // system are compared
      };
      static const EnumAnnotation<CANDIDATES_FROM_NEGATIVE_POLICY>
          kCandidatesFromNegativePolicyAnnotation;

      struct Config : public BaseConfig
      {
         bool hypothesis_minimize{true};
         // whether to compute and use mPrimeClosure in negatives.
         unsigned pccheck_random_rounds{1000};
         unsigned pccheck_random_budget{5};
         double pccheck_random_geometry_mult{5.0};
         unsigned pccheck_random_budget_inc{1};
         double cache_time_logging_threshold{0.001};
         INITIAL_NEGATIVES initial_negatives{INITIAL_NEGATIVES::BODY_MINIMAL};
         CANDIDATES_FROM_NEGATIVE_POLICY candidates_from_negative_policy{
             CANDIDATES_FROM_NEGATIVE_POLICY::SYNCHRONIZE};
         bool use_cnf_hypothesis_for_empowerment{false};

         Config() noexcept {}

         PCEncoder::Config EncoderConfig();
      };

      static const StateValue INITIAL_HYPOTHESIS;
      static const StateValue MINIMIZE_WITH_COLLECTING_EMPOWERED;
      static const StateValue MERGE_DUPLICITIES_IN_NEGATIVES;
      static const StateValue COMPUTE_CLOSURES_OF_NEGATIVES;
      static const StateValue CONSTRUCT_HYPOTHESIS;
      static const StateValue NEXT_NEGATIVE_EXAMPLE;
      static const StateValue REMOVE_DUPLICITIES_IN_NEGATIVES;

    private:
      Config mConf;

      LitImplSystem mNegatives;

      // TODO: Better to have it with moveable top, but the standard
      // priority_queue does not allow it. In future we might think about using
      // a different structure, if copying out of the queue becomes a
      // bottleneck. I do not expect this to be the case.
      NEQueue mCandidatesWithClosure;
      NEQueue mCandidatesWithoutClosure;

      using NECache = std::map<std::vector<Literal>, std::vector<Literal>>;
      NECache mClosureCache;

      unsigned mAddedClausesSinceMinimization{0};

      PCCheckerSAT mPCCheckerSAT;
      unsigned mPCCheckRandomBudget;

      bool mHypothesisSyncNeeded{false};

      // Statistical data
      double mTimeTillTheLastSATEqCheck{0.0};
      double mTimeOfLastSATEqCheck{0.0};
      size_t mNegativesAddedToHypothesis{0};
      size_t mClausesAddedToHypothesis{0};
      size_t mSuccessfulRefinements{0};
      size_t mInitialCandidatesListSize{0};
      size_t mCandidatesWithClosureCount{0};
      size_t mCandidatesWithoutClosureCount{0};
      size_t mLearnedClausesConsideredCount{0};
      size_t mLearnedClausesCandidatesCount{0};
      size_t mRandomEmpoweringCount{0};
      size_t mRandomBodyCount{0};
      size_t mRandomLearnedCount{0};
      size_t mRandomQueriesCount{0};
      size_t mSATEmpoweringCount{0};
      size_t mIntersectionCount{0};
      double mIntersectionTotalTime{0.0};
      mutable size_t mCacheHitCount{0};
      mutable size_t mCacheMissCount{0};
      mutable double mCacheFindTotalTime{0.0};
      double mSemanticClosureTotalTime{0.0};
      double mRefinedNegativeTotalTime{0.0};
      long unsigned mSemanticClosureSolveCount{0};
      unsigned mSemanticClosureCount{0};
      size_t mCandidatesFromNegativesCount{0};

      void MinimizeHypothesis();
      void ComputeInitialCandidates();
      std::vector<LitImpl> InitialCandidatesList();
      std::vector<LitImpl> InitialCandidatesEmpowering(bool collect_all);
      std::vector<LitImpl> InitialCandidatesAll();
      std::vector<LitImpl> InitialCandidatesGDBasis();
      std::vector<LitImpl> InitialCandidatesBodyMinimal();

      void SortNegatives();
      std::optional<LitImpl> NextNegativeExample();
      bool Refine(const LitImpl &e);
      std::optional<LitImpl> TryToRefine(const LitImpl &e, const LitImpl &neg);
      std::optional<LitImpl> RefinedNegative(
          std::vector<Literal> hyp_positive, std::vector<Literal> closure);
      void AddNegativeExample(LitImpl e);
      void AddNegativeToHypothesis(const LitImpl &e);
      void ConstructHypothesis();
      bool UpdateClosureWithMinimizing(LitImpl &neg);
      // Returns none if the assumption given by the body is unsatisfiable.
      std::optional<std::vector<Literal>> SemanticClosure(
          const std::vector<Literal> &body);
      std::optional<std::vector<Literal>> SemanticClosureWithCaching(
          const std::vector<Literal> &body);
      std::optional<std::vector<Literal>> NegClosure(
          const std::vector<Literal> &body);
      std::optional<LitImpl> UpdateClosureWithCaching(
          const LitImpl &neg, bool minimize);
      std::vector<LitImpl> RemoveDuplicitiesInNegatives(
          std::vector<LitImpl> negatives);
      // Check whether it is still negative
      bool CheckNegativeExample(const LitImpl &neg);
      std::optional<LitImpl> CheckNegativeWithBodyUpdate(const LitImpl &neg);
      bool CheckPC(); // stores the result in mCandidatesWithoutClosure
      void MoveNegativesToCandidates(std::vector<LitImpl> negatives);
      bool CheckPCRandomly();
      bool CheckPCWithSAT();
      void CheckRandomImplicate(const Clause &c);
      void CheckRandomBody(std::vector<Literal> literals);
      bool SynchronizeHypothesis(); // true, if any example added to any queue

      // Returns the number of added negatives
      unsigned AddNegativesForImplicate(
          const Clause &clause, bool find_prime_subimplicate);
      // Computes closure, updates body and checks, returns an implication if
      // still negative, otherwise none.
      std::optional<LitImpl> UpdateCandidate(const LitImpl &neg);
      bool AddCandidateIfNegative(const LitImpl &candidate);
      unsigned AddCandidatesForAddedClause(const Clause &clause);
      unsigned AddCandidatesForAddedClauseUnitProp(const Clause &clause);
      bool AddCandidateIfNotUPClosed(const std::vector<Literal> &body);

      NECache::const_iterator CachedClosure(
          const std::vector<Literal> &body) const;

      bool HasNegative();

      void PrintStatistics(unsigned level) const override;
      virtual void CompilePreprocessed() override;

    protected:
      virtual const BaseConfig &GetBaseConfig() const override;
      virtual AbsCheckerSAT &CheckerSAT() override;
      virtual const AbsCheckerSAT &CheckerSAT() const override;

    public:
      PCLearnComp(CNF cnf, Config conf);

      virtual std::string ResultDesc() const override;
      virtual std::string Annotation() const override;
   };
} // namespace SUBool

#endif
