/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file prgutil.h
 * Utility functions and structures used in programs.
 */

#include "prgutil.h"
#include "random_utils.h"
#include "version.h"

const std::string SUBool::kVersionInfo = SUBOOL_VERSION;
const std::string SUBool::kAuthorYearInfo = SUBOOL_AUTHORYEAR;

SUBool::OutputStream::OutputStream(const std::string &fname)
    : mFile(), mOk(true), mOut(&std::cout)
{
   if (!fname.empty() && fname != "-")
   {
      mFile.open(fname);
      mOk = static_cast<bool>(mFile);
      if (mOk)
      {
         mOut = &mFile;
      }
   }
}

void
SUBool::print_version(std::ostream &os, const std::string &prg_name,
    const std::string &prg_desc, bool comment)
{
   std::string c(comment ? "c " : "");
   os << c << prg_name << " - " << prg_desc << '\n';
   os << c << kAuthorYearInfo << '\n';
   os << c << kVersionInfo;
   if (!comment)
   {
      os << std::endl;
   }
}

void
SUBool::PrgOptions::InitCommon()
{
   mCommon.add_options()("help,h", "Print help")(
       "version,v", "Print version info");
   auto opt_verbosity =
       boost::make_shared<po::option_description>("verbosity,b",
           po::value<unsigned>(&mVerbosity)->default_value(mVerbosity),
           "Verbosity level of the log which is written to the standard output."
           " 0 means no log is written, however, messages of glucose can still "
           "appear.");
   mCommon.add(opt_verbosity);
   if (mUseConfigFile)
   {
      mConfigFile = mPrgName + kCfgSuffix;
      mCommon.add_options()("config,c",
          po::value<std::string>(&mConfigFile)->default_value(mConfigFile),
          "Config file")("no-config", "Do not attempt to read a config file.");
      mConfigOnly.add(std::move(opt_verbosity));
   }
   mCommon.add_options()("seed,S", po::value<unsigned long long>(&mSeed),
       "Initial random seed. Current time is used if the seed is not provided "
       "explicitly.");
}

void
SUBool::PrgOptions::InputOutput(
    std::string *input_file, std::string *output_file)
{
   mHidden.add_options()("input", po::value<std::string>(input_file))(
       "output", po::value<std::string>(output_file));
   mPositional.add("input", 1);
   mPositional.add("output", 1);
   mPositionalHelp += "[input [output]]";
}

void
SUBool::PrgOptions::InputPair(
    const PositFileDesc &first, const PositFileDesc &second)
{
   mHidden.add_options()(
       first.desc.c_str(), po::value<std::string>(first.fname))(
       second.desc.c_str(), po::value<std::string>(second.fname));
   mPositional.add(first.desc.c_str(), 1);
   mPositional.add(second.desc.c_str(), 1);
   mPositionalHelp += "[" + first.desc + " [" + second.desc + "]]";
}

void
SUBool::PrgOptions::SinglePositional(const PositFileDesc &file_desc)
{
   mHidden.add_options()(
       file_desc.desc.c_str(), po::value<std::string>(file_desc.fname));
   mPositional.add(file_desc.desc.c_str(), 1);
   mPositionalHelp += "[" + file_desc.desc + "]";
}

bool
SUBool::PrgOptions::ProcessCommon(PrgConfigBase &conf) const
{
   if (mVM.count("help"))
   {
      Help();
      return false;
   }
   if (mVM.count("version"))
   {
      Version();
      return false;
   }
   conf.Init(mVerbosity, mSeed);
   return true;
}

void
SUBool::PrgOptions::Help() const
{
   Info();
   if (!mPrgHelpDesc.empty())
   {
      std::cout << '\n' << mPrgHelpDesc << '\n';
   }
   std::cout << "\nUsage:\n\t" << mPrgName << " [OPTION...] " << mPositionalHelp
             << "\n\n";
   std::cout << mVisible << std::endl;
}

void
SUBool::PrgOptions::Version() const
{
   print_version(std::cout, mPrgName, mPrgDesc);
}

void
SUBool::PrgOptions::AddOptionsGroup(std::shared_ptr<OptionsGroup> opt_group)
{
   Add(opt_group->Desc());
   mOptGroups.push_back(std::move(opt_group));
}

bool
SUBool::PrgOptions::Parse(int argc, char *argv[], PrgConfigBase &conf)
{
   po::options_description cmdline_opts;
   mVisible.add(mCommon).add(mConfig);
   cmdline_opts.add(mCommon).add(mConfig).add(mHidden);

   try
   {
      po::store(po::command_line_parser(argc, argv)
                    .options(cmdline_opts)
                    .positional(mPositional)
                    .run(),
          mVM);
      po::notify(mVM);
   }
   catch (const std::exception &e)
   {
      std::cerr << "Error parsing command line options: " << e.what()
                << std::endl;
      return false;
   }
   if (!ProcessCommon(conf) || !ProcessConfig())
   {
      return false;
   }
   for (auto &opt_group : mOptGroups)
   {
      if (!opt_group->GetValues())
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::PrgOptions::ProcessConfig()
{
   if (!mUseConfigFile || mVM.count("no-config") || mConfigFile.empty())
   {
      return true;
   }
   logs.TLog(4) << "Trying to read config file: " << mConfigFile << std::endl;
   std::ifstream ifs(mConfigFile);
   if (!ifs)
   {
      if (mVM["config"].defaulted())
      {
         logs.TLog(4) << "Failed to open the default config file ("
                      << mConfigFile << ")" << std::endl;
         return true;
      }
      std::cerr << "Failed to open config file: " << mConfigFile << std::endl;
      return false;
   }
   po::options_description config_file_options;
   config_file_options.add(mConfigOnly).add(mConfig).add(mHidden);
   try
   {
      po::store(po::parse_config_file(ifs, config_file_options), mVM);
      po::notify(mVM);
   }
   catch (const std::exception &e)
   {
      std::cerr << "Error parsing config file: " << e.what() << std::endl;
      return false;
   }
   return true;
}

void
SUBool::PrgConfigBase::Init(unsigned verbosity, unsigned long long seed)
{
   logs.mVerbLevel = verbosity;
   mSeed = seed;
   SUBool::sub_rand.seed(seed);
}

void
SUBool::PrgConfigBase::Dump(unsigned level) const
{
   if (!SUBool::logs.CheckLevel(level))
   {
      return;
   }
   SUBool::logs.TLog(level) << "Configuration:" << std::endl;
   SUBool::logs.DumpVerbosity("\t", level);
   SUBool::logs.DumpValue("\t", level, "seed", mSeed);
   InnerDump(level);
   SUBool::logs.TLog(level) << "END" << std::endl;
}
