/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file prgutil.h
 * Utility functions and structures used in programs.
 */

#ifndef __PRGUTIL_H
#define __PRGUTIL_H

#include <algorithm>
#include <array>
#include <chrono>
#include <fstream>
#include <iostream>
#include <sstream>

#include <boost/program_options.hpp>
#include <boost/smart_ptr/make_shared.hpp>

#include "enum.h"
#include "logstream.h"
#include "prg_opt_group.h"
#include "sbexception.h"
#include "util.h"

namespace po = boost::program_options;

namespace SUBool
{
   class OutputStream
   {
      std::ofstream mFile;
      bool mOk;
      std::ostream *mOut;

    public:
      OutputStream(const std::string &fname);

      ~OutputStream() = default;
      OutputStream(const OutputStream &other) = delete;
      OutputStream(OutputStream &&other) = default;
      OutputStream &operator=(const OutputStream &other) = delete;
      OutputStream &operator=(OutputStream &&other) = default;

      std::ostream &
      Out()
      {
         return *mOut;
      }
      operator std::ostream &() { return Out(); }
      bool
      Ok() const
      {
         return mOk;
      }
      operator bool() const { return Ok(); }
   };

   template <class T>
   T
   input(const std::string &fname)
   {
      if (fname.empty() || fname == "-")
      {
         return read_from_stream<T>(std::cin);
      }
      std::ifstream in(fname);
      if (!in)
      {
         std::stringstream ss;
         ss << "Failed to open file '" << fname << "' for reading" << std::endl;
         throw IOException(ss.str());
      }
      return read_from_stream<T>(in);
   }

   template <class T>
   T
   input_or_die(const std::string &fname, unsigned level = 0, int errcode = 2,
       const std::string &msg = "")
   {
      try
      {
         return input<T>(fname);
      }
      catch (SUBool::SBException &e)
      {
         logs.TLog(level) << msg << (msg.empty() ? "" : ": ") << e.what()
                          << std::endl;
         exit(errcode);
      }
   }

   template <class T>
   void
   output(
       const std::string &fname, const T &t, const std::string &preamble = "")
   {
      OutputStream os(fname);
      if (!os)
      {
         std::stringstream ss;
         ss << "Failed to open file '" << fname << "' for writing" << std::endl;
         throw IOException(ss.str());
      }
      if (!preamble.empty())
      {
         os.Out() << preamble << std::endl;
      }
      os.Out() << t << std::endl;
   }

   extern const std::string kAuthorYearInfo; // Defined in prgutil.cpp
   extern const std::string kVersionInfo;    // Defined in prgutil.cpp
   extern const std::string kPrgName; // Need to be defined by the program
   extern const std::string kPrgDesc; // Need to be defined by the program

   void print_version(std::ostream &os, const std::string &prg_name,
       const std::string &prg_desc, bool comment = false);
   inline void
   print_version(std::ostream &os, bool comment = false)
   {
      print_version(os, kPrgName, kPrgDesc, comment);
   }

   class PrgConfigBase
   {
      unsigned long long mSeed{0};

    public:
      PrgConfigBase() = default;

      virtual ~PrgConfigBase() = default;

      virtual void Dump(unsigned level = LogStream::kDumpConfigLevel) const;
      virtual void
      InnerDump(unsigned level [[maybe_unused]]) const
      {
      }

      void Init(unsigned verbosity, unsigned long long seed);
   };

   class PrgOptions
   {
    public:
      struct PositFileDesc
      {
         std::string *fname;
         std::string desc;
      };

    private:
      std::string mPrgName{};
      std::string mPrgDesc{};
      std::string mPrgHelpDesc{};
      std::string mPositionalHelp{};
      unsigned long long mSeed{0};
      unsigned mVerbosity{LogStream::kDefaultLogLevel};
      bool mUseConfigFile{false};
      std::string mConfigFile{};

      po::options_description mCommon{};
      po::options_description mConfig{};
      po::options_description mConfigOnly{};
      po::options_description mHidden{};
      po::options_description mVisible{};
      po::positional_options_description mPositional{};
      po::variables_map mVM{};
      std::vector<std::shared_ptr<OptionsGroup>> mOptGroups{};

      void Help() const;
      void Version() const;
      void Info() const;
      bool ProcessCommon(PrgConfigBase &conf) const;
      bool ProcessConfig();

      void InitCommon();

    public:
      inline static const std::string kCommonOptsLabel = "Common options";
      inline static const std::string kInputOutputHelp = "[input [output]]";
      inline static const std::string kCfgSuffix = ".cfg";

      PrgOptions(bool use_config_file = false,
          const std::string &common_opts_label = kCommonOptsLabel)
          : mPrgName(kPrgName), mPrgDesc(kPrgDesc),
            mSeed(std::chrono::system_clock::now().time_since_epoch().count()),
            mUseConfigFile(use_config_file), mCommon(common_opts_label)
      {
         InitCommon();
      }

      // Empty prg_help_desc means using mPrgDesc
      void SetPrgName(std::string prg_name);
      void SetPrgDesc(std::string prg_desc);
      void SetHelpDesc(std::string prg_help_desc);
      void InputOutput(std::string *input_file, std::string *output_file);
      void InputPair(const PositFileDesc &first, const PositFileDesc &second);
      void SinglePositional(const PositFileDesc &file_desc);
      bool Parse(int argc, char *argv[], PrgConfigBase &conf);

      const po::variables_map &VarMap() const;
      po::options_description_easy_init AddOptions();
      void Add(const po::options_description &desc);
      void AddOptionsGroup(std::shared_ptr<OptionsGroup> opt_group);
      template <typename T, typename... Args>
      std::shared_ptr<T> MakeOptionsGroup(Args &&...args);
      bool HasValue(const std::string &opt_name) const;
   };

} // namespace SUBool

// Template and inline functions implementation
inline void
SUBool::PrgOptions::SetHelpDesc(std::string prg_help_desc)
{
   mPrgHelpDesc = std::move(prg_help_desc);
}

inline void
SUBool::PrgOptions::SetPrgName(std::string prg_name)
{
   mPrgName = std::move(prg_name);
}

inline void
SUBool::PrgOptions::SetPrgDesc(std::string prg_desc)
{
   mPrgDesc = std::move(prg_desc);
}

inline po::options_description_easy_init
SUBool::PrgOptions::AddOptions()
{
   return mConfig.add_options();
}

inline void
SUBool::PrgOptions::Add(const po::options_description &desc)
{
   mConfig.add(desc);
}

inline const po::variables_map &
SUBool::PrgOptions::VarMap() const
{
   return mVM;
}

inline void
SUBool::PrgOptions::Info() const
{
   std::cout << mPrgName << " - " << mPrgDesc << std::endl;
}

inline bool
SUBool::PrgOptions::HasValue(const std::string &opt_name) const
{
   return mVM.count(opt_name);
}

template <typename T, typename... Args>
std::shared_ptr<T>
SUBool::PrgOptions::MakeOptionsGroup(Args &&...args)
{
   auto opt_group = std::make_shared<T>(std::forward<Args>(args)...);
   AddOptionsGroup(opt_group);
   return opt_group;
}

#endif
