/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file propagate.cpp
 * Semantic propagation and backbone finding.
 */

#include "propagate.h"

namespace SUBool
{
   std::unordered_set<Literal>
   intersect_lits_with_assign(std::unordered_set<Literal> lits,
                              const PartialAssignment &alpha);

}; // namespace SUBool

std::optional<SUBool::PartialAssignment>
SUBool::AbsPropagator::Propagate(const SUBool::PartialAssignment &alpha) const
{
   if (!valid_solver_ptr(mSolverPtr) || mUP == nullptr
       || mMinSubimplicant == nullptr)
   {
      throw BadParameterException(
          "AbsPropagator::Propagate(alpha)",
          "No solver, unit propagator, or minimal subimplicant ");
   }
   auto gamma = mUP->Propagate(alpha);
   if (!gamma)
   {
      return {};
   }
   ClearFixedVars();
   return InnerPropagate(*gamma);
}

std::optional<std::vector<SUBool::Literal>>
SUBool::AbsPropagator::BackboneLiterals() const
{
   auto backbone = Backbones();
   if (backbone)
   {
      return literals_of_assignment(*backbone);
   }
   return {};
}

std::optional<SUBool::PartialAssignment>
SUBool::AbsPropagator::ConsistentModel(const PartialAssignment &alpha) const
{
   if (!mSolverPtr->solver->Solve(alpha))
   {
      return {};
   }
   return mSolverPtr->solver->Model();
}

std::optional<SUBool::PartialAssignment>
SUBool::AbsPropagator::MinimalConsistentModelFixedVars(
    const PartialAssignment &alpha) const
{
   auto model = ConsistentModel(alpha);
   if (!model)
   {
      return {};
   }
   return std::make_optional(
       mMinSubimplicant->ComputeWithFixedVars(*model, mFixedVars));
}

SUBool::PartialAssignment
SUBool::AbsPropagator::MinimalSubmodelWithFreeVars(
    PartialAssignment model, const std::vector<unsigned> &free_vars) const
{
   return mMinSubimplicant->ComputeWithFreeVars(std::move(model), free_vars);
}

std::optional<SUBool::PartialAssignment>
SUBool::Propagator::InnerPropagate(const PartialAssignment &alpha) const
{
   auto min_model = MinimalConsistentModelFixedVars(alpha);
   if (!min_model)
   {
      return {};
   }
   InitFreeVars(alpha, *min_model);
   return std::make_optional(ProcessMinModel(alpha, *min_model));
}

SUBool::PartialAssignment
SUBool::Propagator::ProcessMinModel(PartialAssignment alpha,
                                    PartialAssignment beta) const
{
   size_t fixed_var_index = 0;
   for (unsigned i = 1; i <= alpha.Size(); ++i)
   {
      if (FixedVarAt(fixed_var_index) == i)
      {
         ++fixed_var_index;
         continue;
      }
      if (alpha[i] != PartialValue::UNDEF || beta[i] == PartialValue::UNDEF)
      {
         continue;
      }
      alpha[i] = !(beta[i]);
      auto model = ConsistentModel(alpha);
      if (!model)
      {
         alpha[i] = beta[i];
         alpha = UpdateFreeVarsNoModel(i, std::move(alpha), beta);
         continue;
      }
      beta = UpdateFreeVarsWithModel(
          std::move(beta), MinimalModelWithFreeVars(*std::move(model)));
      alpha[i] = PartialValue::UNDEF;
   }
   return alpha;
}

void
SUBool::Propagator::InitFreeVars(const PartialAssignment &alpha,
                                 const PartialAssignment &min_model) const
{
   mFreeSetVars.clear();
   mFreeUnsetVars.clear();
   size_t fixed_var_index = 0;
   for (unsigned i = 1; i <= alpha.Size(); ++i)
   {
      if (FixedVarAt(fixed_var_index) == i)
      {
         ++fixed_var_index;
         continue;
      }
      if (alpha[i] == PartialValue::UNDEF)
      {
         if (min_model[i] == PartialValue::UNDEF)
         {
            mFreeUnsetVars.insert(i);
         }
         else
         {
            mFreeSetVars.insert(i);
         }
      }
   }
}

SUBool::PartialAssignment
SUBool::Propagator::MinimalModelWithFreeVars(PartialAssignment model) const
{
   mFreeVars.assign(mFreeSetVars.begin(), mFreeSetVars.end());
   std::shuffle(mFreeVars.begin(), mFreeVars.end(), sub_rand);
   mFreeVars.insert(mFreeVars.end(), mFreeUnsetVars.begin(),
                    mFreeUnsetVars.end());
   std::shuffle(mFreeVars.begin() + mFreeSetVars.size(), mFreeVars.end(),
                sub_rand);
   return MinimalSubmodelWithFreeVars(std::move(model), mFreeVars);
}

SUBool::PartialAssignment
SUBool::Propagator::UpdateFreeVarsNoModel(unsigned var, PartialAssignment alpha,
                                          const PartialAssignment &beta) const
{
   auto closed_alpha = mUP->Propagate(alpha);
   assert(closed_alpha);
   assert(var > 0 && var <= closed_alpha->Size());
   mFreeSetVars.erase(var);
   for (unsigned i = 1; i <= closed_alpha->Size(); ++i)
   {
      if (alpha[i] != (*closed_alpha)[i])
      {
         if (beta[i] == PartialValue::UNDEF)
         {
            mFreeUnsetVars.erase(i);
         }
         else
         {
            mFreeSetVars.erase(i);
         }
      }
   }
   return *closed_alpha;
}

SUBool::PartialAssignment
SUBool::Propagator::UpdateFreeVarsWithModel(
    PartialAssignment beta, const PartialAssignment &min_model) const
{
   for (unsigned i = 1; i <= beta.Size(); ++i)
   {
      if (beta[i] != min_model[i])
      {
         beta[i] = PartialValue::UNDEF;
         auto i_it = mFreeSetVars.find(i);
         if (i_it != mFreeSetVars.end())
         {
            mFreeSetVars.erase(i_it);
            mFreeUnsetVars.insert(i);
         }
      }
   }
   return beta;
}

SUBool::PartialAssignment
SUBool::propagate_core_chunk(const CNF &cnf, PartialAssignment alpha,
                             std::unordered_set<Literal> free_lits,
                             size_t chunk_size)
{
   std::unordered_set<Literal> chunk_literals;
   std::unordered_set<Literal> omega;
   std::vector<Literal> omega_lits;
   // Should be optimized so that we do not need to make a copy of cnf here
   std::vector<Clause> units;
   for (unsigned var = 1; var <= alpha.Size(); ++var)
   {
      if (alpha[var] != PartialValue::UNDEF)
      {
         units.emplace_back(Literal(var, alpha[var] == PartialValue::TRUE));
      }
   }
   while (!free_lits.empty())
   {
      chunk_literals.clear();
      std::copy_n(free_lits.begin(), std::min(free_lits.size(), chunk_size),
                  std::inserter(chunk_literals, chunk_literals.end()));
      omega.clear();
      for (auto &lit : chunk_literals)
      {
         omega.insert(lit.Negation());
      }
      while (true)
      {
         SimpSolverWrapper wrap;
         auto &g = wrap.Solver();
         g.AddCnf(cnf);
         g.AddClauses(units);
         omega_lits.assign(omega.begin(), omega.end());
         auto result = g.Solve(omega_lits);
         if (result)
         {
            free_lits =
                intersect_lits_with_assign(std::move(free_lits), g.Model());
            break;
         }
         auto c = g.ConflictClause();
         size_t intersect_size = 0;
         Literal intersect_lit;
         for (const auto &lit : c)
         {
            auto ln = lit.Negation();
            if (omega.find(ln) != omega.end())
            {
               ++intersect_size;
               if (intersect_size == 1)
               {
                  intersect_lit = lit;
               }
               omega.erase(ln);
            }
         }
         if (intersect_size == 0)
         {
            throw InvariantFailedException(
                "propagate_core_chunk",
                "Intersection of core with omega is empty");
         }
         if (intersect_size == 1)
         {
            chunk_literals.erase(intersect_lit);
            free_lits.erase(intersect_lit);
            units.emplace_back(Literal(intersect_lit));
            alpha.SetLiteral(intersect_lit);
         }
         if (omega.empty())
         {
            for (const auto &lit : chunk_literals)
            {
               free_lits.erase(lit);
            }
            auto cl_size = chunk_literals.size();
            alpha = propagate_iterative(cnf, alpha, std::move(chunk_literals),
                                        iterative_chunk_size(cl_size), units);
            break;
         }
      }
   }
   return alpha;
}

SUBool::PartialAssignment
SUBool::propagate_iterative(const CNF &cnf, PartialAssignment alpha,
                            std::unordered_set<Literal> free_lits,
                            size_t chunk_size, const std::vector<Clause> &units)
{
   Clause c;
   while (!free_lits.empty())
   {
      c.Clear();
      std::copy_n(free_lits.begin(), std::min(free_lits.size(), chunk_size),
                  std::back_inserter(c));
      c.NegateLiterals();
      SimpSolverWrapper wrap;
      auto &g = wrap.Solver();
      if (g.AddCnf(cnf) && g.AddClauses(units) && g.AddClause(c)
          && g.Solve(alpha))
      {
         free_lits =
             intersect_lits_with_assign(std::move(free_lits), g.Model());
         continue;
      }
      c.NegateLiterals();
      alpha.SetLiterals(c.begin(), c.end());
      std::for_each(c.begin(), c.end(),
                    [&free_lits](const Literal &lit) { free_lits.erase(lit); });
   }
   return alpha;
}

std::unordered_set<SUBool::Literal>
SUBool::intersect_lits_with_assign(std::unordered_set<Literal> lits,
                                   const PartialAssignment &alpha)
{
   auto i_lit = lits.begin();
   while (i_lit != lits.end())
   {
      if (alpha[i_lit->Variable()] != i_lit->TrueValue())
      {
         lits.erase(i_lit++);
      }
      else
      {
         ++i_lit;
      }
   }
   return lits;
}

size_t
SUBool::chunk_size(size_t free_lits_size, size_t min_chunk_size,
                   size_t max_chunk_size, size_t min_chunk_count)
{
   size_t k = std::min(max_chunk_size, free_lits_size / min_chunk_count);
   return std::max<size_t>(k, min_chunk_size);
}
