/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file propagate.h
 * Semantic propagation and backbone finding.
 */

#ifndef __PROPAGATE_H
#define __PROPAGATE_H

#include <unordered_set>

#include "glucose_bind.h"
#include "logstream.h"
#include "min_model.h"
#include "unitprop.h"
#include "unitprop2.h"
#include "util.h"

namespace SUBool
{
   class AbsPropagator
   {
    protected:
      SolverConstPtr
          mSolverPtr; // Must be incremental and equivalent to the CNF mUP is
                      // initialized with managed from the outside
      const UnitPropagatorM *mUP;
      const MinimalSubimplicant *mMinSubimplicant;

    private:
      mutable std::vector<unsigned>
          mFixedVars; // not to be used for propagation SAT calls

      void
      ClearFixedVars() const
      {
         mFixedVars.clear();
      }

      template <class FreePred> void InitFixedVars(FreePred free_pred) const;

    protected:
      // alpha is already closed under unit propagation when passed to
      // InnerPropagate
      virtual std::optional<PartialAssignment>
      InnerPropagate(const PartialAssignment &alpha) const = 0;

      std::optional<PartialAssignment>
      ConsistentModel(const PartialAssignment &alpha) const;
      std::optional<PartialAssignment>
      MinimalConsistentModelFixedVars(const PartialAssignment &alpha) const;
      PartialAssignment
      MinimalSubmodelWithFreeVars(PartialAssignment model,
                                  const std::vector<unsigned> &free_vars) const;
      unsigned
      FixedVarAt(size_t i) const
      {
         return (i < mFixedVars.size() ? mFixedVars[i] : 0);
      }

    public:
      AbsPropagator(SolverConstPtr solver_ptr, const UnitPropagatorM *up,
                    const MinimalSubimplicant *min_subimplicant) noexcept
          : mSolverPtr(solver_ptr), mUP(up), mMinSubimplicant(min_subimplicant),
            mFixedVars()
      {
         assert(!valid_solver_ptr(mSolverPtr)
                || mSolverPtr->solver->IsIncremental());
      }
      AbsPropagator(const AbsPropagator &other) = default;
      AbsPropagator(AbsPropagator &&other) = default;
      virtual ~AbsPropagator() = default;
      AbsPropagator &operator=(const AbsPropagator &other) = default;
      AbsPropagator &operator=(AbsPropagator &&other) = default;

      virtual void
      Reset(SolverConstPtr solver_ptr, const UnitPropagatorM *up,
            const MinimalSubimplicant *min_subimplicant)
      {
         mSolverPtr = solver_ptr;
         mUP = up;
         mMinSubimplicant = min_subimplicant;
         assert(!valid_solver_ptr(mSolverPtr)
                || mSolverPtr->solver->IsIncremental());
      }

      // @param pred specifies predicate for checking if a variable should be
      // considered for propagation (with a SAT call).
      template <class FreePred>
      std::optional<PartialAssignment> Propagate(const PartialAssignment &alpha,
                                                 FreePred free_pred) const;

      // propagates all
      std::optional<PartialAssignment>
      Propagate(const PartialAssignment &alpha) const;

      std::optional<PartialAssignment>
      PropagateMaxVar(const PartialAssignment &alpha, unsigned max_var) const
      {
         return Propagate(alpha,
                          [max_var](unsigned i) { return i <= max_var; });
      }

      std::optional<PartialAssignment>
      Backbones() const
      {
         return Propagate(PartialAssignment(0));
      }

      std::optional<std::vector<Literal>> BackboneLiterals() const;
   };

   class Propagator : public AbsPropagator
   {
      mutable std::unordered_set<unsigned> mFreeSetVars;
      mutable std::unordered_set<unsigned> mFreeUnsetVars;
      mutable std::vector<unsigned> mFreeVars;

      void InitFreeVars(const PartialAssignment &alpha,
                        const PartialAssignment &min_model) const;
      PartialAssignment ProcessMinModel(PartialAssignment alpha,
                                        PartialAssignment beta) const;
      PartialAssignment MinimalModelWithFreeVars(PartialAssignment model) const;
      PartialAssignment
      UpdateFreeVarsNoModel(unsigned var, PartialAssignment alpha,
                            const PartialAssignment &beta) const;
      PartialAssignment
      UpdateFreeVarsWithModel(PartialAssignment beta,
                              const PartialAssignment &min_model) const;

    protected:
      virtual std::optional<PartialAssignment>
      InnerPropagate(const PartialAssignment &alpha) const override;

    public:
      Propagator(SolverConstPtr solver, const UnitPropagatorM *up,
                 const MinimalSubimplicant *min_subimplicant)
          : AbsPropagator(solver, up, min_subimplicant), mFreeSetVars(),
            mFreeUnsetVars(), mFreeVars()
      {
      }
   };

   // A second propagation algorithm, right now it uses propagate_iterative
   template <class Pred>
   std::optional<PartialAssignment>
   propagate2(const CNF &cnf, const PartialAssignment &alpha, Pred pred);

   PartialAssignment propagate_core_chunk(const CNF &cnf,
                                          PartialAssignment alpha,
                                          std::unordered_set<Literal> free_lits,
                                          size_t chunk_size);
   size_t chunk_size(size_t free_lits_size, size_t min_chunk_size,
                     size_t max_chunk_size, size_t min_chunk_count);
   inline size_t
   core_chunk_size(size_t free_lits_size)
   {
      // return 30;
      return chunk_size(free_lits_size, 4, 50, 10);
   }
   inline size_t iterative_chunk_size(size_t /*free_lits_size*/)
   {
      return 1;
      // return chunk_size(free_lits_size, 1, 1, 1);
   }
   PartialAssignment propagate_iterative(const CNF &cnf,
                                         PartialAssignment alpha,
                                         std::unordered_set<Literal> free_lits,
                                         size_t chunk_size,
                                         const std::vector<Clause> &units);
} // namespace SUBool

// Template definition
template <class Pred>
std::optional<SUBool::PartialAssignment>
SUBool::propagate2(const CNF &cnf, const PartialAssignment &alpha, Pred pred)
{
   PartialAssignment gamma(0);
   UnitPropagator up(&cnf);
   if (up.Propagate(alpha, &gamma) >= 0)
   {
      return {};
   }
   SimpSolverWrapper wrap(SolverInterface::ADAPTATION::LIKELY_SAT);
   auto &g = wrap.Solver();
   g.AddCnf(cnf);
   if (!g.Solve(alpha))
   {
      return {};
   }
   std::unordered_set<Literal> free_lits;
   for (unsigned i = 1; i <= gamma.Size(); ++i)
   {
      if (pred(i) && gamma[i] == PartialValue::UNDEF
          && g.ValueOfVariable(i) != PartialValue::UNDEF)
      {
         free_lits.emplace(i, g.ValueOfVariable(i) == PartialValue::TRUE);
      }
   }
   // return propagate_core_chunk(cnf, gamma, free_lits,
   // core_chunk_size(free_lits.size()));
   return propagate_iterative(cnf, gamma, free_lits,
                              iterative_chunk_size(free_lits.size()),
                              std::vector<Clause>());
}

template <class FreePred>
void
SUBool::AbsPropagator::InitFixedVars(FreePred free_pred) const
{
   mFixedVars.clear();
   for (unsigned var = 1; var <= mUP->MaxVariable(); ++var)
   {
      if (!free_pred(var))
      {
         mFixedVars.push_back(var);
      }
   }
}

template <class FreePred>
std::optional<SUBool::PartialAssignment>
SUBool::AbsPropagator::Propagate(const PartialAssignment &alpha,
                                 FreePred free_pred) const
{
   if (!valid_solver_ptr(mSolverPtr) || mUP == nullptr
       || mMinSubimplicant == nullptr)
   {
      throw BadParameterException(
          "AbsPropagator::Propagate(alpha, free_pred)",
          "No solver, unit propagator, or minimal subimplicant ");
   }
   auto gamma = mUP->Propagate(alpha);
   if (!gamma)
   {
      return {};
   }
   InitFixedVars(free_pred);
   return InnerPropagate(*gamma);
}

#endif /* end of include guard */
