/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file random_utils.cpp
 * Randomly generating literals, clauses, normal forms etc.
 */

#include "random_utils.h"

std::mt19937_64 SUBool::sub_rand(
    std::chrono::system_clock::now().time_since_epoch().count());

std::vector<SUBool::Literal>
SUBool::random_literal_vector(const unsigned n_vars)
{
   std::vector<Literal> literals;
   for (unsigned var = 1; var <= n_vars; ++var)
   {
      switch (sub_rand() % 3)
      {
      case 0:
         literals.emplace_back(var, true);
         break;
      case 1:
         literals.emplace_back(var, false);
         break;
      default:
         break;
      };
   }
   return literals;
}

std::vector<SUBool::Literal>
SUBool::random_literal_vector_with_size(const unsigned size,
                                        const unsigned n_vars)
{
   std::vector<Literal> literals;
   std::vector<bool> variables(n_vars, false);
   for (unsigned count = 0; count < std::min(size, n_vars); ++count)
   {
      unsigned var_index = sub_rand() % (n_vars - count);
      unsigned var_count = 0;
      for (unsigned i = 0; i < variables.size(); ++i)
      {
         if (variables[i])
         {
            continue;
         }
         if (!variables[i] && var_count == var_index)
         {
            variables[i] = true;
            literals.emplace_back(i + 1, sub_rand() % 2 == 0);
            break;
         }
         ++var_count;
      }
   }
   return literals;
}

SUBool::CNF
SUBool::random_horn_cnf(const unsigned size, const unsigned max_clause_size,
                        const unsigned n_vars, bool pure)
{
   std::vector<Clause> clauses;
   while (clauses.size() < size)
   {
      clauses.push_back(
          random_horn_clause(sub_rand() % max_clause_size + 1, n_vars, pure));
   }
   return CNF(clauses);
}

SUBool::Clause
SUBool::random_horn_clause(const unsigned size, const unsigned n_vars,
                           bool pure)
{
   const unsigned base = (pure ? 0 : 1);
   std::vector<Literal> literals =
       random_literal_vector_with_size(size, n_vars);
   unsigned i = sub_rand() % (size + base);
   for (unsigned j = 0; j < literals.size(); ++j)
   {
      literals[j].mPositive = (j == i);
   }
   return Clause(literals);
}

SUBool::CNF
SUBool::random_k_cnf(const unsigned int size, const unsigned int clause_size,
                     const unsigned int n_vars)
{
   std::vector<Clause> clauses;
   while (clauses.size() < size)
   {
      clauses.emplace_back(
          random_literal_vector_with_size(clause_size, n_vars));
   }
   return CNF(clauses);
}

SUBool::CNF
SUBool::random_cnf(const unsigned int size, const unsigned int n_vars)
{
   std::vector<Clause> clauses;
   while (clauses.size() < size)
   {
      unsigned clause_size = 1 + sub_rand() % n_vars;
      std::vector<Literal> literals =
          random_literal_vector_with_size(clause_size, n_vars);
      clauses.emplace_back(literals);
   }
   return CNF(clauses);
}

std::pair<SUBool::CNF, SUBool::Assignment>
SUBool::random_sat_cnf(const unsigned int size, const unsigned n_vars)
{
   // Guess random satisfying assignment
   Assignment alpha(n_vars, 0.5);
   std::vector<Clause> clauses;
   while (clauses.size() < size)
   {
      unsigned clause_size = 1 + sub_rand() % n_vars;
      std::vector<Literal> literals =
          random_literal_vector_with_size(clause_size, n_vars);
      // Check if it is satisfied
      bool satisfied = false;
      for (const auto &lit : literals)
      {
         if (lit.Value(alpha))
         {
            satisfied = true;
            break;
         }
      }
      if (!satisfied)
      {
         unsigned i = sub_rand() % literals.size();
         literals[i].Negate();
      }
      clauses.emplace_back(literals);
   }
   return std::make_pair(CNF(clauses), std::move(alpha));
}
