/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file rh-bdmc.cpp
 * Renamable Horn BDMC compilation (based on looking for a backdoor set).
 */

#include "rh_bdmc.h"
#include "cnfcompare.h"
#include "cnfcompress.h"
#include "cnfgraph.h"
#include "cnfutil.h"
#include "empower.h"
#include "glucose_bind.h"
#include "logstream.h"
#include "rhorn.h"
#include "sbexception.h"
#include "unitprop.h"

#include "pcchecker.h"

SUBool::RHBackdoorTree::RHBackdoorTree(CNF cnf)
    : mCnf(std::move(cnf)), mWorkCnf(), mLeafSize(0), mCurrentNnf(nullptr),
      mLiteralNodes(), mCNFLeaves(), mDecisionStack(), mSplitStack(),
      mValues(mCnf.MaxVariable()), mShouldBacktrack(false)
{
}

void
SUBool::RHBackdoorTree::Compute()
{
   // bool after_backtrack = true;
   if (!Preprocess())
   {
      return;
   }
   bool not_finished = true;
   while (not_finished)
   // Invariant, at the beginning, work cnf is updated
   // that is:
   // it is up prime
   // it is within current component
   // it is after mValues are set
   // it is satisfiable
   {
      // std::cout << "values: " << mValues << std::endl;
      switch (DetermineNextEvent())
      {
      case EventType::FINISHED:
         // std::cout << "finished" << std::endl;
         not_finished = false;
         break;
      case EventType::BACKTRACK:
         // std::cout << "backtrack" << std::endl;
         Backtrack();
         break;
      case EventType::DECISION:
         // std::cout << "decide" << std::endl;
         Decide();
         break;
      case EventType::COMPONENT_SPLIT:
         // std::cout << "component split" << std::endl;
         NextComponentSplit();
         break;
      }
   }
   // the result should be in current NNF node.
}

SUBool::RHBackdoorTree::EventType
SUBool::RHBackdoorTree::DetermineNextEvent()
{
   if (mShouldBacktrack)
   {
      return EventType::BACKTRACK;
   }
   auto level_info = CurrentLevel();
   switch (level_info.stack_type)
   {
   case StackType::NONE:
      return EventType::FINISHED;
   case StackType::DECISION:
      CanSplit();
      // if (CanSplit())
      //{
      //   return EventType::COMPONENT_SPLIT;
      //}
      return EventType::DECISION;
   case StackType::SPLIT:
      // This is after backtrack which takes care of changing the current
      // component.
      return EventType::DECISION;
      // return EventType::COMPONENT_SPLIT;
   }
   throw BadParameterException("RHBackdoorTree::DetermineNextEvent",
                               "Unknown stack type");
}

void
SUBool::RHBackdoorTree::UpdateWorkCNF(bool down)
{
   // TODO: needs update
   // 1. select the right base cnf (if down, then mWorkCNF, if not down, then
   // mCnf).
   // 2. Apply mValues
   if (down)
   {
      mWorkCnf.Apply(mValues);
   }
   else
   {
      mCnf.Apply(mValues, &mWorkCnf);
   }
   mWorkCnf = up_prime_cnf(std::move(mWorkCnf));
   // Making it up prime should not be necessary even if not down
}

void
SUBool::RHBackdoorTree::RestrictComponent()
{
   mWorkCnf.RemoveIf(
       [this](const Clause &clause) { return ClauseNotInComponent(clause); });
}

bool
SUBool::RHBackdoorTree::Preprocess()
{
   logs.TLog(1) << "Start preprocessing" << std::endl;
   if (!is_satisfiable(mCnf))
   {
      mCurrentNnf = NnfNode::MakeFalse();
      logs.TLog(1) << "CNF is unsatisfiable" << std::endl;
      return false;
   }
   mWorkCnf = up_prime_cnf(mCnf);
   DecisionInfo dec_info(Literal(mCnf.MaxVariable(), true), 0);
   for (const auto &clause : mWorkCnf)
   {
      if (clause.Size() == 1)
      {
         dec_info.implied.push_back(clause[0]);
         mValues.SetLiteral(clause[0]);
      }
   }
   mDecisionStack.push(std::move(dec_info));
   UpdateWorkCNF(true);
   logs.TLog(1) << "Preprocessing finished" << std::endl;
   return true;
}

void
SUBool::RHBackdoorTree::Decide()
{
   auto lit = PickLiteral();
   if (!lit)
   {
      // renamable horn leaf
      // TODO: move work cnf? depends on backtrack implementation!
      mCurrentNnf = CNFLeafNode(mWorkCnf);
      mShouldBacktrack = true;
      return;
   }
   mDecisionStack.emplace(*lit, CurrentLevel().level + 1);
   mValues.SetLiteral(*lit);
   UpdateWorkCNF(true);
   if (!AddImplied())
   {
      mShouldBacktrack = true;
   }
}

void
SUBool::RHBackdoorTree::NextComponentSplit()
{
   // check if there is next component
   // if so, set it to be the current component
   if (mSplitStack.top().current_component >= mSplitStack.top().component_size)
   {
      // ???
      mShouldBacktrack = true;
      return;
   }
   // recompute work cnf
   UpdateWorkCNF(false);
   Decide();
}

bool
SUBool::RHBackdoorTree::AddImplied()
{
   if (mWorkCnf.Empty())
   {
      mCurrentNnf = NnfNode::MakeTrue();
      // std::cout << "AddImplied: make true" << std::endl;
      return false;
   }
   if (!is_satisfiable(mWorkCnf))
   {
      mCurrentNnf = NnfNode::MakeFalse();
      // std::cout << "AddImplied: make false" << std::endl;
      return false;
   }
   for (const auto &clause : mWorkCnf)
   {
      if (clause.Size() == 1)
      {
         mDecisionStack.top().implied.push_back(clause[0]);
         mValues.SetLiteral(clause[0]);
      }
   }
   mWorkCnf.Apply(mValues);
   return true;
}

bool
SUBool::RHBackdoorTree::CanSplit()
{
   CNFCompressor compress(mWorkCnf);
   auto compressed_work_cnf = compress.CompressCNF(mWorkCnf);
   IncidenceGraph g(compressed_work_cnf);
   auto comp = g.ConnectedComponents();
   if (comp.count <= 1)
   {
      return false;
   }
   // std::cout << "Split happens" << std::endl;
   // std::cout << "Work cnf: " << mWorkCnf << std::endl;
   // Initialize a new component split event.
   ComponentSplit comp_split(CurrentLevel().level + 1, comp.count);
   for (unsigned var = 1; var <= comp.comp_variables.size(); ++var)
   {
      comp_split.components[compress.UncompressVar(var)] =
          comp.comp_variables[var - 1];
      // std::cout << "Component of variable " << compressed.UncompressVar(var)
      //<< " : " << comp.comp_variables[var - 1] << std::endl;
   }
   mSplitStack.push(std::move(comp_split));
   UpdateWorkCNF(true);
   RestrictComponent();
   return true;
}

void
SUBool::RHBackdoorTree::Backtrack()
{
   // Just do one step and keep mShouldBacktrack true if further steps are
   // required.
   auto level_info = CurrentLevel();
   switch (level_info.stack_type)
   {
   case StackType::NONE:
      mShouldBacktrack = false;
      break;
   case StackType::DECISION:
      // std::cout << "backtrack decision" << std::endl;
      BacktrackDecision();
      break;
   case StackType::SPLIT:
      // std::cout << "backtrack component split" << std::endl;
      BacktrackComponentSplit();
      break;
   }
}

void
SUBool::RHBackdoorTree::BacktrackDecision()
{
   AddImpliedToCurrentNnf(mDecisionStack.top().implied);
   for (const auto &lit : mDecisionStack.top().implied)
   {
      mValues[lit.Variable()] = PartialValue::UNDEF;
   }
   mDecisionStack.top().implied.clear();
   if (mDecisionStack.top().level == 0)
   {
      // std::cout << "Level 0" << std::endl;
      mShouldBacktrack = false;
      mDecisionStack.pop();
      // std::cout << "current level: " << CurrentLevel().level << std::endl;
      return;
   }
   if (mDecisionStack.top().first_branch == nullptr)
   {
      // std::cout << "first branch is null" << std::endl;
      mShouldBacktrack = false;
      mDecisionStack.top().first_branch = std::move(mCurrentNnf);
      mDecisionStack.top().literal.Negate();
      mValues.SetLiteral(mDecisionStack.top().literal);
      UpdateWorkCNF(false);
      RestrictComponent();
      if (!AddImplied())
      {
         mShouldBacktrack = true;
      }
      return;
   }
   if (mDecisionStack.top().first_branch != mCurrentNnf)
   {
      if (mDecisionStack.top().literal.Positive())
      {
         std::swap(mCurrentNnf, mDecisionStack.top().first_branch);
      }
      mCurrentNnf =
          NnfNode::MakeDecision(mDecisionStack.top().literal.Variable(),
                                mDecisionStack.top().first_branch, mCurrentNnf);
   }
   // std::cout << "here, " << mShouldBacktrack << std::endl;
   mValues[mDecisionStack.top().literal.Variable()] = PartialValue::UNDEF;
   mDecisionStack.pop();
}

void
SUBool::RHBackdoorTree::BacktrackComponentSplit()
{
   auto &cs = mSplitStack.top();
   auto &node = cs.node;
   if (node == nullptr)
   {
      node = std::move(mCurrentNnf);
   }
   else if (mSplitStack.top().node->NodeType() == NnfNode::Type::And)
   {
      // mCurrentNnf is new child of node
      std::dynamic_pointer_cast<NnfGateAnd>(node)->AddChild(
          std::move(mCurrentNnf));
   }
   else
   {
      // new and node with two child nodes
      node = NnfNode::MakeAnd(std::move(node), std::move(mCurrentNnf));
   }
   // Increase current component
   if (++cs.current_component >= cs.component_size)
   {
      // If there is no other, backtrack further.
      // -- set mCurrentNnf to the node nnf and pop
      mCurrentNnf = std::move(node);
      mSplitStack.pop();
      return;
   }
   // Else UpdateWorkCnf (the next step will be decision)
   UpdateWorkCNF(false);
   RestrictComponent();
   mShouldBacktrack = false;
}

void
SUBool::RHBackdoorTree::AddImpliedToCurrentNnf(
    const std::vector<Literal> &literals)
{
   if (literals.empty())
   {
      return;
   }
   unsigned skip_shift = 0;
   if (mCurrentNnf->NodeType() != NnfNode::Type::And)
   {
      mCurrentNnf =
          NnfNode::MakeAnd(mCurrentNnf, LiteralNode(literals.front()));
      skip_shift = 1;
   }
   auto node = std::dynamic_pointer_cast<NnfGateAnd>(mCurrentNnf);
   for (auto i_lit = literals.begin() + skip_shift; i_lit != literals.end();
        ++i_lit)
   {
      node->AddChild(LiteralNode(*i_lit));
   }
}

std::optional<SUBool::Literal>
SUBool::RHBackdoorTree::PickLiteral()
{
   // If the RH check succeeds, set current BDMC as NnfLeafCNF with the given
   // cnf.
   // CNFCompress compressed(mWorkCnf);
   // IncidenceGraph g(compressed.CompressedCNF());
   //   auto comp = g.ConnectedComponents();
   //  logs.TLog(1)
   // << "Number of connected components: " << comp.count << std::endl;
   // std::cout << mWorkCnf << std::endl;
   // mCurrentNnf = CNFLeafNode(std::move(mWorkCnf));
   if (is_urc(mWorkCnf))
   {
      return {};
   }
   return RenamableHornChecker::CheckCNFConflict(mWorkCnf);
}

unsigned
SUBool::RHBackdoorTree::LeafSize() const
{
   return mLeafSize;
}

SUBool::NnfNode::tPointer
SUBool::RHBackdoorTree::LiteralNode(Literal lit)
{
   if (mLiteralNodes.count(lit) == 0)
   {
      mLiteralNodes[lit] = NnfNode::MakeLiteral(lit);
   }
   return mLiteralNodes[lit];
}

SUBool::Nnf
SUBool::RHBackdoorTree::CurrentNnf() &&
{
   return Nnf(std::move(mCurrentNnf));
}

const SUBool::Nnf
SUBool::RHBackdoorTree::CurrentNnf() const &
{
   return Nnf(mCurrentNnf);
}

SUBool::NnfNode::tPointer
SUBool::RHBackdoorTree::CNFLeafNode(CNF cnf)
{
   if (cnf.Empty())
   {
      // logs.TLog(1)
      //<< "CNFLeafNode, CNF empty, returning true node." << std::endl;
      return NnfNode::MakeTrue();
   }
   // logs.TLog(1)
   //<< "CNFLeafNode, looking for a cached value." << std::endl;
   logs.TLog(1) << cnf << std::endl;
   auto vars = cnf.Variables();
   std::vector<unsigned> vars_vec(vars.begin(), vars.end());
   auto key = cnf_leaf_key{cnf.MaxVariable(), vars_vec.size()};
   for (auto i_node = mCNFLeaves[key].crbegin();
        i_node != mCNFLeaves[key].crend(); ++i_node)
   {
      auto leaf_node = std::dynamic_pointer_cast<NnfLeafCNF>(*i_node);
      if (vars_vec == leaf_node->Variables()
          && cnf_equal_up(cnf, leaf_node->Cnf()))
      {
         // logs.TLog(1) << "Cache hit" << std::endl;
         return *i_node;
      }
   }
   // logs.TLog(1) << "New CNF leaf" << std::endl;
   auto node = NnfNode::MakeLeafCNF(std::move(cnf));
   mCNFLeaves[key].push_back(node);
   return node;
}

SUBool::RHBackdoorTree::LevelInfo
SUBool::RHBackdoorTree::CurrentLevel() const
{
   int decision_level =
       (mDecisionStack.empty() ? -1 : mDecisionStack.top().level);
   int split_level = (mSplitStack.empty() ? -1 : mSplitStack.top().level);
   int level = std::max(decision_level, split_level);
   // std::cout << "level: " << level << std::endl;
   // std::cout << "decision_level: " << decision_level << std::endl;
   // std::cout << "split_level: " << split_level << std::endl;
   if (level < 0)
   {
      return LevelInfo{StackType::NONE, -1};
   }
   return LevelInfo{
       (level == decision_level ? StackType::DECISION : StackType::SPLIT),
       level};
}

bool
SUBool::RHBackdoorTree::ClauseNotInComponent(const Clause &clause)
{
   if (mSplitStack.empty() || clause.Empty())
   {
      return false;
   }
   return mSplitStack.top().components[clause[0].Variable()]
          != mSplitStack.top().current_component;
}
