/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file rh-bdmc.h
 * Renamable Horn BDMC compilation (based on looking for a backdoor set).
 */

#ifndef __RH_BDMC_H
#define __RH_BDMC_H

#include <stack>

#include <optional>

#include "nnf.h"
#include "normalform.h"

namespace SUBool
{

   // TODO: Rename it to something related to BDMC, such as RenamableHornBDMC,
   // it does not compile to backdoor tree.
   // Backdoor Tree with respect to renamable Horn formulas
   // It is a special case of RH-BDMC.
   class RHBackdoorTree
   {
      struct cnf_leaf_key
      {
         unsigned max_variable;
         std::size_t var_size;

         bool
         operator==(const cnf_leaf_key &rhs) const
         {
            return rhs.max_variable == max_variable && rhs.var_size == var_size;
         }
      };

      struct cnf_leaf_key_hash
      {
         std::size_t
         operator()(const cnf_leaf_key &key) const
         {
            return std::hash<unsigned>()(key.max_variable)
                   ^ std::hash<std::size_t>()(key.var_size);
         }
      };

      enum class DecisionState
      {
         FIRST_DECISION,
         SECOND_DECISION,
         IMPLIED
      };

      enum class StackType
      {
         NONE,
         DECISION,
         SPLIT
      };

      struct LevelInfo
      {
         StackType stack_type;
         int level;
      };

      enum class EventType
      {
         FINISHED,
         BACKTRACK,
         DECISION,
         COMPONENT_SPLIT
      };

      struct DecisionInfo
      {
         int level;
         // variable=0 means no literal, the initial „decision“ which represents
         // only unit clauses in the initial CNF after preprocessing.
         Literal literal;
         NnfNode::tPointer first_branch;
         std::vector<Literal> implied;

         DecisionInfo(Literal lit, int l)
             : level(l), literal(lit), first_branch(nullptr), implied()
         {
         }
      };

      struct ComponentSplit
      {
         int level;
         std::unordered_map<unsigned, unsigned> components;
         unsigned current_component;
         unsigned component_size;
         NnfNode::tPointer node;

         ComponentSplit(int level, unsigned n_comp)
             : level(level), components(), current_component(0),
               component_size(n_comp), node(nullptr)
         {
         }
      };

      CNF mCnf;
      CNF mWorkCnf;
      unsigned mLeafSize;
      NnfNode::tPointer mCurrentNnf;
      std::unordered_map<Literal, NnfNode::tPointer> mLiteralNodes;
      // indexed by max variable and number of variables.
      std::unordered_map<cnf_leaf_key, std::vector<NnfNode::tPointer>,
                         cnf_leaf_key_hash>
          mCNFLeaves;
      std::stack<DecisionInfo> mDecisionStack;
      std::stack<ComponentSplit> mSplitStack;
      PartialAssignment mValues;
      // int mLevel;
      bool mShouldBacktrack;

      EventType DetermineNextEvent();
      bool CanSplit(); // Checks if the current work cnf can be split and
                       // prepares a split event if this is the case.
      bool Preprocess();
      void Backtrack();
      void BacktrackDecision();
      void BacktrackComponentSplit();
      void Decide();
      void NextComponentSplit();
      bool ClauseNotInComponent(const Clause &clause);

      void
      UpdateWorkCNF(bool down); // relative to mValues and mCurrentComponent,
                                // different when going down or up.
      void RestrictComponent(); // in work cnf
      bool AddImplied();        // true success, false=contradiction
      void AddImpliedToCurrentNnf(const std::vector<Literal> &literals);
      std::optional<Literal> PickLiteral();
      NnfNode::tPointer LiteralNode(Literal lit);
      NnfNode::tPointer CNFLeafNode(CNF cnf);
      LevelInfo CurrentLevel() const;

    public:
      RHBackdoorTree(CNF cnf);

      void Compute();

      // valid after computation
      unsigned LeafSize() const;
      Nnf CurrentNnf() &&;
      const Nnf CurrentNnf() const &;
   };

} // namespace SUBool

#endif
