/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file rhorn.cpp
 * Class implementing an algorithm for testing if a given CNF is
 * renamable Horn.
 */

#include "rhorn.h"

SUBool::RenamableHornChecker::RenamableHornChecker(const CNF &cnf)
    : mCNF(cnf), mSetH(2 * cnf.MaxVariable(), false),
      mSetPositiveT(2 * cnf.MaxVariable(), false), mLitPositiveT(),
      mSetNegativeT(2 * cnf.MaxVariable(), false), mLitNegativeT(),
      mClausesToLiterals(2 * cnf.MaxVariable()),
      mLiteralAddress(2 * cnf.MaxVariable()), mPositiveQ(), mNegativeQ(),
      mCoherentPositiveT(true), mCoherentNegativeT(true)
{
}

bool
SUBool::RenamableHornChecker::Check(Assignment *renaming)
{
   ComputeClausesToLiterals();
   for (unsigned i = 1; i <= mCNF.MaxVariable(); ++i)
   {
      if (!VariableInH(i))
      {
         if (!ProcessVariable(i))
         {
            return false;
         }
      }
   }
   FillRenamingWithH(renaming);
   return true;
}

bool
SUBool::RenamableHornChecker::CheckCNF(const CNF &cnf, Assignment *renaming)
{
   RenamableHornChecker checker(cnf);
   return checker.Check(renaming);
}

std::optional<SUBool::Literal>
SUBool::RenamableHornChecker::CheckConflict()
{
   Check(nullptr);
   return mConflictLiteral;
}

std::optional<SUBool::Literal>
SUBool::RenamableHornChecker::CheckCNFConflict(const CNF &cnf)
{
   RenamableHornChecker checker(cnf);
   return checker.CheckConflict();
}

bool
SUBool::RenamableHornChecker::VariableInH(unsigned var) const
{
   return mSetH[Literal(var, true).Index()]
          || mSetH[Literal(var, false).Index()];
}

void
SUBool::RenamableHornChecker::ComputeClausesToLiterals()
{
   for (CNF::const_iterator i_clause = mCNF.begin(); i_clause != mCNF.end();
        ++i_clause)
   {
      for (const auto &lit : *i_clause)
      {
         if (i_clause->Size() > 1)
         {
            mClausesToLiterals[lit.Index()].push_back(i_clause);
         }
      }
   }
   for (unsigned lit = 0; lit < mLiteralAddress.size(); ++lit)
   {
      if (!mClausesToLiterals[lit].empty())
      {
         mLiteralAddress[lit].first = mClausesToLiterals[lit].begin();
         mLiteralAddress[lit].second = mClausesToLiterals[lit].front()->begin();
      }
   }
}

bool
SUBool::RenamableHornChecker::ProcessVariable(unsigned var)
{
   CleanUp();
   unsigned pos_index = Literal(var, true).Index();
   unsigned neg_index = Literal(var, false).Index();
   mSetPositiveT[pos_index] = true;
   mLitPositiveT.push(Literal(var, true));
   mSetNegativeT[neg_index] = true;
   mLitNegativeT.push(Literal(var, false));
   if (!mClausesToLiterals[pos_index].empty())
   {
      mPositiveQ.push(Literal(var, true));
   }
   if (!mClausesToLiterals[neg_index].empty())
   {
      mNegativeQ.push(Literal(var, false));
   }
   std::optional<Literal> positive_conflict;
   std::optional<Literal> negative_conflict;
   while (!mPositiveQ.empty() && !mNegativeQ.empty() && mCoherentPositiveT
          && mCoherentNegativeT)
   {
      positive_conflict = OneStep(mSetPositiveT, mLitPositiveT, mPositiveQ);
      mCoherentPositiveT = !positive_conflict;
      negative_conflict = OneStep(mSetNegativeT, mLitNegativeT, mNegativeQ);
      mCoherentNegativeT = !negative_conflict;
   }
   if (mPositiveQ.empty() && mCoherentPositiveT)
   {
      AddTToH(mSetPositiveT, mLitPositiveT);
   }
   else if (mNegativeQ.empty() && mCoherentNegativeT)
   {
      AddTToH(mSetNegativeT, mLitNegativeT);
   }
   else if (!mPositiveQ.empty() && !mCoherentNegativeT)
   {
      while (!mPositiveQ.empty() && mCoherentPositiveT)
      {
         positive_conflict = OneStep(mSetPositiveT, mLitPositiveT, mPositiveQ);
         mCoherentPositiveT = !positive_conflict;
      }
      if (mCoherentPositiveT)
      {
         AddTToH(mSetPositiveT, mLitPositiveT);
      }
      else
      {
         mConflictLiteral = positive_conflict;
         return false;
      }
   }
   else if (!mNegativeQ.empty() && !mCoherentPositiveT)
   {
      while (!mNegativeQ.empty() && mCoherentNegativeT)
      {
         negative_conflict = OneStep(mSetNegativeT, mLitNegativeT, mNegativeQ);
         mCoherentNegativeT = !negative_conflict;
      }
      if (mCoherentNegativeT)
      {
         AddTToH(mSetNegativeT, mLitNegativeT);
      }
      else
      {
         mConflictLiteral = negative_conflict;
         return false;
      }
   }
   else
   {
      mConflictLiteral = positive_conflict;
      return false;
   }
   return true;
}

void
SUBool::RenamableHornChecker::CleanUp()
{
   while (!mLitPositiveT.empty())
   {
      mSetPositiveT[mLitPositiveT.front().Index()] = false;
      mLitPositiveT.pop();
   }
   while (!mLitNegativeT.empty())
   {
      mSetNegativeT[mLitNegativeT.front().Index()] = false;
      mLitNegativeT.pop();
   }
   while (!mPositiveQ.empty())
   {
      mPositiveQ.pop();
   }
   while (!mNegativeQ.empty())
   {
      mNegativeQ.pop();
   }
   mCoherentPositiveT = mCoherentNegativeT = true;
}

std::optional<SUBool::Literal>
SUBool::RenamableHornChecker::OneStep(std::vector<bool> &setT,
                                      std::queue<Literal> &litT,
                                      std::queue<Literal> &Q)
{
   const Literal &lit = Q.front();
   if (!IsValidAddress(lit))
   {
      Q.pop();
      return {};
   }
   Literal succ_lit = GetLiteralForAddress(lit);
   if (succ_lit != lit)
   {
      succ_lit.Negate();
      unsigned succ_index = succ_lit.Index();
      if (!mSetH[succ_index] && !setT[succ_index])
      {
         if (setT[Literal(succ_lit.mVariable, !succ_lit.mPositive).Index()])
         {
            return succ_lit;
         }
         setT[succ_index] = true;
         litT.push(succ_lit);
         if (!mClausesToLiterals[succ_index].empty())
         {
            Q.push(succ_lit);
         }
      }
   }
   AdvanceAddress(lit);
   return {};
}

void
SUBool::RenamableHornChecker::FillRenamingWithH(Assignment *renaming) const
{
   if (renaming != nullptr)
   {
      renaming->Resize(mCNF.MaxVariable(), false);
      for (unsigned var = 1; var <= mCNF.MaxVariable(); ++var)
      {
         (*renaming)[var] = mSetH[Literal(var, false).Index()];
      }
   }
}

void
SUBool::RenamableHornChecker::AddTToH(std::vector<bool> &mSetT,
                                      std::queue<Literal> &mLitT)
{
   while (!mLitT.empty())
   {
      unsigned lit_index = mLitT.front().Index();
      mSetT[lit_index] = false;
      mSetH[lit_index] = true;
      mLitT.pop();
   }
}

bool
SUBool::RenamableHornChecker::IsValidAddress(const Literal &lit) const
{
   unsigned lit_index = lit.Index();
   const address_t &lit_addr = mLiteralAddress[lit_index];
   return (lit_addr.first != mClausesToLiterals[lit_index].end());
}

void
SUBool::RenamableHornChecker::AdvanceAddress(const Literal &lit)
{
   unsigned lit_index = lit.Index();
   address_t &lit_addr = mLiteralAddress[lit_index];
   if (lit_addr.first != mClausesToLiterals[lit_index].end())
   {
      ++lit_addr.second;
      if (lit_addr.second == (*lit_addr.first)->end())
      {
         ++lit_addr.first;
         if (lit_addr.first != mClausesToLiterals[lit_index].end())
         {
            lit_addr.second = (*lit_addr.first)->begin();
         }
      }
   }
}

const SUBool::Literal &
SUBool::RenamableHornChecker::GetLiteralForAddress(const Literal &lit) const
{
   const address_t &lit_addr = mLiteralAddress[lit.Index()];
   return *(lit_addr.second);
}
