/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file rhorn.h
 * Class implementing an algorithm for testing if a given CNF is
 * renamable Horn.
 */

#ifndef __RHORN_H
#define __RHORN_H

#include "normalform.h"

#include <list>
#include <optional>
#include <queue>

namespace SUBool
{
   /**@brief Class for checking if a given CNF is renamable Horn.
    * Implements the algorithm from "A linear algorithm for renaming a set of
    * clauses as a Horn set" by J.-J. Hérbrard.
    */
   class RenamableHornChecker
   {
      using clause_list_t = std::list<CNF::const_iterator>;
      using address_t =
          std::pair<clause_list_t::iterator, Clause::const_iterator>;

      const CNF &mCNF;

      std::vector<bool> mSetH;
      std::vector<bool> mSetPositiveT;
      std::queue<Literal> mLitPositiveT;
      std::vector<bool> mSetNegativeT;
      std::queue<Literal> mLitNegativeT;
      std::vector<clause_list_t> mClausesToLiterals;
      std::vector<address_t> mLiteralAddress;
      std::queue<Literal> mPositiveQ;
      std::queue<Literal> mNegativeQ;
      bool mCoherentPositiveT;
      bool mCoherentNegativeT;
      std::optional<Literal> mConflictLiteral{};

      RenamableHornChecker(const CNF &cnf);
      bool Check(Assignment *renaming);
      std::optional<Literal> CheckConflict();

    public:
      static bool CheckCNF(const CNF &cnf, Assignment *renaming);
      static std::optional<Literal> CheckCNFConflict(const CNF &cnf);

    private:
      bool VariableInH(unsigned var) const;
      void ComputeClausesToLiterals();
      bool ProcessVariable(unsigned var);
      void CleanUp();
      std::optional<Literal> OneStep(std::vector<bool> &setT,
                                     std::queue<Literal> &litT,
                                     std::queue<Literal> &Q);
      void FillRenamingWithH(Assignment *renaming) const;
      void AddTToH(std::vector<bool> &mSetT, std::queue<Literal> &mLitT);
      bool IsValidAddress(const Literal &lit) const;
      void AdvanceAddress(const Literal &lit);
      const Literal &GetLiteralForAddress(const Literal &lit) const;
   };
}; // namespace SUBool

#endif
