/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file schoning.cpp
 * Class implementing the algorithm of Uwe Schoning described in "A
 * Probabilistic Algorithm for k-SAT and Constraint Satisfaction
 * Problems". That is the randomized SAT algorithm.
 */

#include <algorithm>
#include <cmath>

#include "random_utils.h"
#include "schoning.h"

SUBool::Schoning::Schoning(const CNF &cnf)
    : mCNF(cnf), mWidth(0), mClausesToLiterals(2 * cnf.MaxVariable()),
      mSatisfiedClauses(), mUnsatisfiedClauses(), mClauseInList(),
      mClauseStatus(cnf.Size())
{
   ComputeClausesToLiterals();
}

void
SUBool::Schoning::InitWalkWithAssignment(const Assignment &assignment)
{
   mSatisfiedClauses.clear();
   mUnsatisfiedClauses.clear();
   mClauseInList.resize(mCNF.Size());
   for (size_t i = 0; i < mCNF.Size(); ++i)
   {
      if (mCNF[i].Value(assignment))
      {
         mSatisfiedClauses.push_front(i);
         mClauseInList[i] = mSatisfiedClauses.begin();
         mClauseStatus[i] = true;
      }
      else
      {
         mUnsatisfiedClauses.push_front(i);
         mClauseInList[i] = mUnsatisfiedClauses.begin();
         mClauseStatus[i] = false;
      }
   }
}

bool
SUBool::Schoning::RandomWalk(Assignment *satisfying_assignment)
{
   // std::cout<<"Next round"<<std::endl;
   Assignment random_assignment(mCNF.MaxVariable(), 0.5);
   InitWalkWithAssignment(random_assignment);
   for (unsigned i = 0; i < 3 * (mCNF.MaxVariable() + 1); ++i)
   {
      // std::cout<<random_assignment<<std::endl;
      /*
      std::cout<<"Satisfied:";
      for(const auto &clause : mSatisfiedClauses)
      {
         std::cout<<" "<<clause;
      }
      std::cout<<std::endl;
      std::cout<<"Unsatisfied:";
      for(const auto &clause : mUnsatisfiedClauses)
      {
         std::cout<<" "<<clause;
      }
      std::cout<<std::endl;
      */
      if (mUnsatisfiedClauses.empty())
      {
         if (satisfying_assignment != nullptr)
         {
            *satisfying_assignment = random_assignment;
         }
         return true;
      }
      size_t clause_index = mUnsatisfiedClauses.front();
      size_t literal_in_clause =
          (size_t)(sub_rand()) % mCNF[clause_index].Size();
      size_t var = mCNF[clause_index][literal_in_clause].mVariable;
      random_assignment[var] = !random_assignment[var];
      CheckClauseList(Literal(var, true), random_assignment);
      CheckClauseList(Literal(var, false), random_assignment);
   }
   return false;
}

void
SUBool::Schoning::CheckClauseList(const Literal &lit,
                                  const Assignment &assignment)
{
   for (auto clause_index : mClausesToLiterals[lit.Index()])
   {
      if (mClauseStatus[clause_index])
      {
         // std::cout<<"+Clause: "<<mCNF[clause_index]<<std::endl;
         if (!mCNF[clause_index].Value(assignment))
         {
            mUnsatisfiedClauses.splice(mUnsatisfiedClauses.end(),
                                       mSatisfiedClauses,
                                       mClauseInList[clause_index]);
            mClauseStatus[clause_index] = false;
         }
      }
      else
      {
         // std::cout<<"-Clause: "<<mCNF[clause_index]<<std::endl;
         mClauseStatus[clause_index] = true;
         mSatisfiedClauses.splice(mSatisfiedClauses.end(), mUnsatisfiedClauses,
                                  mClauseInList[clause_index]);
      }
   }
}

bool
SUBool::Schoning::Run(Assignment *satisfying_assignment, unsigned long rounds)
{
   long double repetitions =
       static_cast<long double>(rounds)
       * pow(static_cast<long double>(2 * (mWidth - 1))
                 / static_cast<long double>(mWidth),
             static_cast<long double>(mCNF.MaxVariable()));
   long double remaining_repetitions = repetitions;
   long double percent = 0.0;
   std::cout << "0.0% Remaining repetitions: " << repetitions << std::endl;
   while (remaining_repetitions > 0.0)
   {
      if ((long double)100.0
              - (remaining_repetitions / repetitions) * (long double)100.0
          >= percent + 1.0)
      {
         percent =
             (long double)100.0 - (remaining_repetitions / repetitions) * 100.0;
         std::cout << percent << "% ";
         std::cout << "Remaining repetitions: " << remaining_repetitions
                   << std::endl;
      }
      if (RandomWalk(satisfying_assignment))
      {
         return true;
      }
      remaining_repetitions -= 1.0L;
   }
   return false;
}

void
SUBool::Schoning::ComputeClausesToLiterals()
{
   for (size_t i = 0; i < mCNF.Size(); ++i)
   {
      mWidth = std::max(mWidth, mCNF[i].Size());
      for (const auto &lit : mCNF[i])
      {
         // if(mCNF[i].Size()>1)
         //{
         mClausesToLiterals[lit.Index()].push_back(i);
         //}
      }
   }
}

bool
SUBool::Schoning::SAT(const CNF &cnf, Assignment *satisfying_assignment,
                      unsigned long rounds)
{
   Schoning sat(cnf);
   return sat.Run(satisfying_assignment, rounds);
}

