/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file schoning.h
 * Class implementing the algorithm of Uwe Schoning described in "A
 * Probabilistic Algorithm for k-SAT and Constraint Satisfaction
 * Problems". That is the randomized SAT algorithm.
 */

#ifndef __SCHONING_H
#define __SCHONING_H

#include <list>
#include <vector>

#include "normalform.h"

namespace SUBool
{
   /**@brief Class implementing a randomized SAT algorithm of Uwe
    * Schoning. The algorithm was presented in "A Probabilistic
    * Algorithm for k-SAT and Constraint Satisfaction Problems" by Uwe
    * Schoning in 1999.
    */
   class Schoning
   {
      using tClauseList = std::list<size_t>;

      const CNF &mCNF;
      unsigned long mWidth;
      std::vector<tClauseList> mClausesToLiterals;
      tClauseList mSatisfiedClauses;
      tClauseList mUnsatisfiedClauses;
      std::vector<tClauseList::iterator> mClauseInList;
      std::vector<bool> mClauseStatus;

      Schoning(const CNF &cnf);

      void ComputeClausesToLiterals();

      bool Run(Assignment *satisfying_assignment, unsigned long rounds);
      void InitWalkWithAssignment(const Assignment &assignment);
      bool RandomWalk(Assignment *satisfying_assignment);
      void CheckClauseList(const Literal &lit, const Assignment &assignment);

    public:
      /**@brief Executes the Schoning SAT algorithm.
       *
       * @param cnf Input CNF.
       * @param satisfying_assignment In here a satisfying
       * assignment is returned if it was found.
       * @param rounds Number of repetitions of the basic
       * procedure. The basic random walk is repeated
       * rounds*(2*(k-1)/k)^n times where n is the number of variables in
       * @p cnf and k is the width of @p cnf that is the maximum
       * number of literals in a clause of @p cnf.
       */
      static bool SAT(const CNF &cnf, Assignment *satisfying_assignment,
                      unsigned long rounds = 1);
   };
}; // namespace SUBool

#endif
