/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file solver_bind.cpp
 * Binding class to solver library. */

#include "solver_bind.h"
#include "glucose_bind.h"
#include "kissat_bind.h"
#include "util.h"

std::atomic<long unsigned> SUBool::SolverInterface::sSolveCountSAT;
std::atomic<long unsigned> SUBool::SolverInterface::sSolveCountUNSAT;
std::atomic<long unsigned> SUBool::SolverInterface::sSolveCountUNDEF;
SUBool::SolverPool SUBool::SolverPool::sDefaultPool;

const SUBool::EnumAnnotation<SUBool::SOLVER_TYPE> SUBool::kSolverTypeAnnotation(
    "SOLVER_TYPE", { //
       {
          SOLVER_TYPE::GLUCOSE, "glucose", "use glucose."
       }
#if HAS_KISSAT
       , { SOLVER_TYPE::KISSAT, "kissat", "use kissat." }
#endif
    });

bool
SUBool::SolverInterface::AddClauses(const std::vector<Clause> &clauses)
{
   for (const auto &clause : clauses)
   {
      if (!AddClause(clause))
      {
         return false;
      }
   }
   return true;
}

void
SUBool::SolverPool::InterruptAll()
{
   std::lock_guard<std::mutex> lock(mMutex);
   std::for_each(mSolverList.begin(), mSolverList.end(),
       [](auto &h) { h.solver->Interrupt(); });
}

void
SUBool::SolverPool::InterruptInThread(std::thread::id thread_id)
{
   std::lock_guard<std::mutex> lock(mMutex);
   std::for_each(mSolverList.begin(), mSolverList.end(),
       [thread_id](auto &h)
       {
          if (h.thread_id == thread_id)
          {
             h.solver->Interrupt();
          }
       });
}

SUBool::AbsSolverWrapper
SUBool::solver_wrapper(
    SOLVER_TYPE solver_type, SolverInterface::ADAPTATION adapt)
{
   switch (solver_type)
   {
   case SOLVER_TYPE::GLUCOSE:
      return SimpSolverWrapper(adapt);
   case SOLVER_TYPE::KISSAT:
#if HAS_KISSAT
      return KissatSolverWrapper(adapt);
#else
      throw BadParameterException(
          "AbsCheckerSAT::NewSolverWrapper", "Kissat support not available");
#endif
   default:
      throw BadParameterException(
          "AbsCheckerSAT::NewSolverWrapper", "Unknown solver type");
   }
};

