/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file solver_bind.h
 * Binding class to solver library. */

#ifndef __SOLVER_BIND_H
#define __SOLVER_BIND_H

#include <chrono>
#include <future>
#include <list>
#include <memory>
#include <optional>
#include <thread>

#include "enum.h"
#include "normalform.h"

namespace SUBool
{
   class SolverInterface
   {
      static std::atomic<long unsigned> sSolveCountUNSAT;
      static std::atomic<long unsigned> sSolveCountSAT;
      static std::atomic<long unsigned> sSolveCountUNDEF;

    protected:
      static void
      IncreaseSolveCount(PartialValue result)
      {
         switch (result)
         {
         case PartialValue::TRUE:
            ++sSolveCountSAT;
            break;
         case PartialValue::FALSE:
            ++sSolveCountUNSAT;
            break;
         case PartialValue::UNDEF:
            ++sSolveCountUNDEF;
            break;
         }
      }

    public:
      enum class ADAPTATION
      {
         DEFAULT,
         LIKELY_SAT,
         LIKELY_UNSAT
      };

      virtual ~SolverInterface() = default;
      virtual bool Reset(const CNF &cnf) = 0;
      virtual bool AddCnf(const CNF &cnf) = 0;
      virtual bool AddClause(const Clause &clause) = 0;
      virtual bool AddClauses(const std::vector<Clause> &clause);
      static long unsigned
      SolveCountSAT()
      {
         return sSolveCountSAT;
      }
      static long unsigned
      SolveCountUNSAT()
      {
         return sSolveCountUNSAT;
      }
      static long unsigned
      SolveCountUNDEF()
      {
         return sSolveCountUNDEF;
      }
      static long unsigned
      SolveCount()
      {
         return sSolveCountSAT + sSolveCountUNSAT + sSolveCountUNDEF;
      }

      virtual bool Solve(const std::vector<Literal> &assump) const = 0;
      virtual bool Solve() const = 0;
      virtual bool Solve(const PartialAssignment &assump) const = 0;
      virtual bool Solve(const Assignment &assump) const = 0;

      template <class Rep, class Period>
      PartialValue SolveWithTimeout(
          const std::chrono::duration<Rep, Period> &timeout_duration) const;
      template <class Rep, class Period>
      PartialValue SolveWithTimeout(const std::vector<Literal> &assump,
          const std::chrono::duration<Rep, Period> &timeout_duration) const;

      virtual PartialValue ValueOfVariable(unsigned var) const = 0;
      virtual PartialValue ValueOfLiteral(const Literal &lit) const = 0;
      virtual PartialAssignment Model() const = 0;
      virtual Clause ConflictClause() const = 0;

      virtual void Interrupt() const = 0;

      virtual size_t LearntsSize() const = 0;
      virtual std::vector<Clause> Learnts() const = 0;

      virtual bool IsIncremental() const = 0;
   };

   struct SolverHandle
   {
      std::thread::id thread_id;
      std::unique_ptr<SolverInterface> solver;

      SolverHandle() : thread_id(), solver() {}
      SolverHandle(std::thread::id id, std::unique_ptr<SolverInterface> s)
          : thread_id(id), solver(std::move(s))
      {
      }
   };
   using SolverList = std::list<SolverHandle>;
   using SolverPtr = SolverList::iterator;
   using SolverConstPtr = SolverList::const_iterator;

   class SolverPool
   {
      SolverList mSolverList;

      static SolverPool sDefaultPool;

      std::mutex mMutex;

    public:
      SolverPool() : mSolverList(), mMutex() {}
      ~SolverPool() = default;

      static SolverPool &
      DefaultPool()
      {
         return sDefaultPool;
      }

      // template <class S> SolverPtr NewSolver();
      template <class S, class... Args> SolverPtr NewSolver(Args &&...arg);

      void
      Release(SolverPtr &ptr)
      {
         std::lock_guard<std::mutex> lock(mMutex);
         if (ptr != NullPtr())
         {
            mSolverList.erase(ptr);
            ptr = NullPtr();
         }
      }

      void InterruptAll();
      void InterruptInThread(std::thread::id thread_id);
      SolverPtr
      NullPtr()
      {
         return mSolverList.end();
      }

      SolverConstPtr
      NullPtr() const
      {
         return mSolverList.end();
      }
   };

   inline bool
   valid_solver_ptr(SolverConstPtr solver_ptr)
   {
      return solver_ptr != SolverPool::DefaultPool().NullPtr();
   }

   class AbsSolverWrapper
   {
    protected:
      SolverPtr mSolverPtr;

    public:
      AbsSolverWrapper(SolverPtr ptr) : mSolverPtr(ptr) {}
      AbsSolverWrapper(const AbsSolverWrapper &s) = delete;
      AbsSolverWrapper(AbsSolverWrapper &&s) : mSolverPtr(s.mSolverPtr)
      {
         s.mSolverPtr = SolverPool::DefaultPool().NullPtr();
      }
      AbsSolverWrapper &operator=(const AbsSolverWrapper &s) = delete;
      AbsSolverWrapper &
      operator=(AbsSolverWrapper &&s)
      {
         SolverPool::DefaultPool().Release(mSolverPtr);
         mSolverPtr = s.mSolverPtr;
         s.mSolverPtr = SolverPool::DefaultPool().NullPtr();
         return *this;
      }

      virtual ~AbsSolverWrapper()
      {
         SolverPool::DefaultPool().Release(mSolverPtr);
      }

      SolverInterface &
      Solver()
      {
         return *(mSolverPtr->solver);
      }

      const SolverInterface &
      Solver() const
      {
         return *(mSolverPtr->solver);
      }

      SolverPtr
      GetSolverPtr()
      {
         return mSolverPtr;
      }
      SolverConstPtr
      GetSolverPtr() const
      {
         return mSolverPtr;
      }
   }; // namespace SUBool

   enum class SOLVER_TYPE
   {
      GLUCOSE,
      KISSAT
   };

   extern const EnumAnnotation<SOLVER_TYPE> kSolverTypeAnnotation;

   AbsSolverWrapper solver_wrapper(
       SOLVER_TYPE solver_type, SolverInterface::ADAPTATION adapt);

   template <class S> class SolverWrapperTemplate : public AbsSolverWrapper
   {
    public:
      template <class... Args>
      SolverWrapperTemplate(Args &&...arg)
          : AbsSolverWrapper(SolverPool::DefaultPool().NewSolver<S>(
              std::forward<Args>(arg)...))
      {
      }

      SolverWrapperTemplate(const SolverWrapperTemplate &s) = delete;
      SolverWrapperTemplate(SolverWrapperTemplate &&s) = default;
      SolverWrapperTemplate &operator=(const SolverWrapperTemplate &s) = delete;
      SolverWrapperTemplate &operator=(SolverWrapperTemplate &&s) = default;
   };

} // namespace SUBool

// Template functions implementation
template <class Rep, class Period>
SUBool::PartialValue
SUBool::SolverInterface::SolveWithTimeout(
    const std::chrono::duration<Rep, Period> &timeout_duration) const
{
   return SolveWithTimeout({}, timeout_duration);
}

template <class Rep, class Period>
SUBool::PartialValue
SUBool::SolverInterface::SolveWithTimeout(const std::vector<Literal> &assump,
    const std::chrono::duration<Rep, Period> &timeout_duration) const
{
   auto fut = std::async(
       std::launch::async, [this, &assump]() { return Solve(assump); });
   if (fut.wait_for(timeout_duration) == std::future_status::timeout)
   {
      Interrupt();
      fut.wait();
      return PartialValue::UNDEF;
   }
   return partial_value_of_bool(fut.get());
}

template <class S, class... Args>
SUBool::SolverPtr
SUBool::SolverPool::NewSolver(Args &&...arg)
{
   std::lock_guard<std::mutex> lock(mMutex);
   mSolverList.emplace_back(std::this_thread::get_id(),
       std::make_unique<S>(std::forward<Args>(arg)...));
   return std::prev(mSolverList.end());
}
#endif
