/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file task.h
 * Class representing a task which provides access to the timeout value, abort
 * status and possibly various budget types.
 */
#ifndef __TASK_H
#define __TASK_H

#include <atomic>
#include <memory>
#include <optional>

#include "cpuutil.h"

namespace SUBool
{
   class Task
   {
      CPUTime mCPUTime{};
      std::atomic<bool> mAborted{false};
      std::atomic<long double> mTimeout{-1.0l};

    public:
      // TODO: Constructors with other budgets
      Task(long double timeout) : mTimeout(timeout) {}

      std::optional<long double>
      TimeoutRemains() const
      {
         return (mTimeout >= 0.0l ? mTimeout - mCPUTime.TimeSpan()
                                  : std::optional<long double>());
      }
      bool
      Aborted() const
      {
         return mAborted;
      }
      bool
      CheckTimeout() const
      {
         auto t_rest = TimeoutRemains();
         return !t_rest || *t_rest > 0.0;
      }
      // not aborted, not timed out, does not check other budgets
      bool
      Ok() const
      {
         return !Aborted() && CheckTimeout();
      }

      void
      ResetStart()
      {
         mCPUTime.Reset();
      }

      void
      Abort()
      {
         mAborted = true;
      }
   };

   using TaskPtr = std::shared_ptr<Task>;

   template <typename... Args>
   TaskPtr
   new_task(Args &&...args)
   {
      return std::make_shared<Task>(std::forward<Args>(args)...);
   };

   inline bool
   task_ok(const TaskPtr &task)
   {
      return task == nullptr || task->Ok();
   }

} // namespace SUBool

#endif

