/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file ucenc.h
 * Encoding of checking if the given CNF admits an empowering
 * implicate with respect to unit refutation completeness.
 */

#include "ucenc.h"
#include "cardinal.h"
#include "up_onelevel.h"

SUBool::UCEncoder::UCEncoder(const Config &conf)
    : AbsEncoder(conf), mPartAssignVars(), mSelectLits(), mConf(conf)
{
}

const SUBool::AbsEncoder::AbsConfig &
SUBool::UCEncoder::GetConfig() const
{
   return mConf;
}

void
SUBool::UCEncoder::WriteMappingsOfVariable(std::ostream &out, unsigned i) const
{
   if (mSelectLits.size() < i)
   {
      out << "c No mappings of selection variables." << std::endl;
   }
   else
   {
      out << "c s_" << i << " = " << mSelectLits[i - 1];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mSelectLits[i - 1].Variable()) << "]";
      }
      out << std::endl;
   }
   if (mPartAssignVars.size() < 2 * i)
   {
      out << "c No mappings of variable in partial assignment specifying the "
             "clause."
          << std::endl;
   }
   else
   {
      out << "c partial assign variable for literal " << Literal(2 * (i - 1))
          << " = " << mPartAssignVars[2 * (i - 1)];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mPartAssignVars[2 * (i - 1)]) << "] ";
      }
      out << std::endl;
      out << "c partial assign variable for literal " << Literal(2 * i - 1)
          << " = " << mPartAssignVars[2 * i - 1];
      if (mConf.use_backmap)
      {
         out << " [" << mVarStore.Name(mPartAssignVars[2 * i - 1]) << "] ";
      }
      out << std::endl;
   }
}

std::string
SUBool::UCEncoder::Name() const
{
   return "UC encoding";
}

unsigned
SUBool::UCEncoder::ExtractClause(
    const SolverInterface &solver, Clause *clause) const
{
   if (clause == nullptr)
   {
      throw NullPointerException(
          "UCEncoder::ExtractClause", "clause == nullptr");
   }
   std::vector<Literal> literals;
   for (unsigned i = 0; i < mPartAssignVars.size(); ++i)
   {
      if (solver.ValueOfVariable(mPartAssignVars[i]) == PartialValue::TRUE)
      {
         Literal lit(i);
         literals.push_back(lit.Negation());
      }
   }
   if (mConf.use_backmap)
   {
      for (unsigned i = 0; i < mVarStore.LastVariable(); ++i)
      {
         std::cout << i + 1 << "=" << solver.ValueOfVariable(i + 1) << " ["
                   << mVarStore.Name(i + 1) << "]" << std::endl;
      }
   }
   Clause c(literals);
   *clause = std::move(c);
   for (unsigned i = 0; i < mSelectLits.size(); ++i)
   {
      if (solver.ValueOfLiteral(mSelectLits[i]) == PartialValue::TRUE)
      {
         return i + 1;
      }
   }
   return 0;
}

void
SUBool::UCEncoder::AddClauses(const CNF &cnf, const ClausesToLiterals &cl2lit,
    std::shared_ptr<const EncISProvider>)
{
   PartAssignClosedUnderUP(cnf, cl2lit);
   PartAssignNoConflict();
   InitSelectLits();
   SelectedNotInPA();
   OneProve(true);
   OneProve(false);
}

void
SUBool::UCEncoder::PartAssignClosedUnderUP(
    const CNF &cnf, const ClausesToLiterals &cl2lit)
{
   UPEncOneLevel enc(cnf, cl2lit, mVarStore);
   mVarStore.SetPrefix("u");
   enc.Encode(0, &mPartAssignVars, nullptr, &mClauses);
   mVarStore.SetPrefix("");
}

void
SUBool::UCEncoder::PartAssignNoConflict()
{
   for (unsigned i = 1; i <= mCnfMaxVariable; ++i)
   {
      Literal pos(i, true);
      Literal neg(i, false);
      mClauses.emplace_back(Literal(mPartAssignVars[pos.Index()], false),
          Literal(mPartAssignVars[neg.Index()], false));
   }
}

void
SUBool::UCEncoder::InitSelectLits()
{
   mVarStore.InitLitArray(&mSelectLits, mCnfMaxVariable, false, "s_{", "}");
   auto exone_encoder = ExOneEncoder::NewGlobalEncoder(mVarStore);
   exone_encoder->EncodeAppend(mSelectLits, mClauses);
}

void
SUBool::UCEncoder::SelectedNotInPA()
{
   for (unsigned i = 0; i < mSelectLits.size(); ++i)
   {
      Literal pos(i + 1, true);
      Literal neg(i + 1, false);
      mClauses.emplace_back(mSelectLits[i].Negation(),
          Literal(mPartAssignVars[pos.Index()], false));
      mClauses.emplace_back(mSelectLits[i].Negation(),
          Literal(mPartAssignVars[neg.Index()], false));
   }
}

void
SUBool::UCEncoder::OneProve(bool positive)
{
   mVarStore.SetPrefix(positive ? "pv" : "nv");
   std::vector<unsigned> in;
   std::vector<unsigned> out;
   mOneProveEncoder->Encode(mConf.up_level_bound, &in, &out, &mClauses);
   PassInputsToOneProve(positive, in);
   ConflictInOneProve(out);
   mVarStore.SetPrefix("");
}

void
SUBool::UCEncoder::PassInputsToOneProve(
    bool positive, const std::vector<unsigned> &in)
{
   for (unsigned i = 0; i < mCnfMaxVariable; ++i)
   {
      Literal pos(i + 1, positive);
      Literal neg(i + 1, !positive);
      mClauses.emplace_back(Literal(in[pos.Index()], false),
          Literal(mPartAssignVars[pos.Index()], true), mSelectLits[i]);
      mClauses.emplace_back(Literal(mPartAssignVars[pos.Index()], false),
          Literal(in[pos.Index()], true));
      mClauses.emplace_back(
          mSelectLits[i].Negation(), Literal(in[pos.Index()], true));
      mClauses.emplace_back(
          Literal(in[neg.Index()], false), mSelectLits[i].Negation());
      mClauses.emplace_back(Literal(in[neg.Index()], false),
          Literal(mPartAssignVars[neg.Index()], true));
      mClauses.emplace_back(mSelectLits[i],
          Literal(mPartAssignVars[neg.Index()], false),
          Literal(in[neg.Index()], true));
   }
}

void
SUBool::UCEncoder::ConflictInOneProve(const std::vector<unsigned> &out)
{
   std::vector<Literal> t_literals;
   for (unsigned i = 0; i < out.size() / 2; ++i)
   {
      mVarStore.NameStream() << "t_{" << i + 1 << "}";
      unsigned t = mVarStore.NewVariable();
      Literal pos(i + 1, true);
      Literal neg(i + 1, false);
      Literal pos_t(t, true);
      Literal neg_t(t, false);
      t_literals.push_back(pos_t);
      mClauses.emplace_back(neg_t, Literal(out[pos.Index()], true));
      mClauses.emplace_back(neg_t, Literal(out[neg.Index()], true));
      mClauses.emplace_back(Literal(out[pos.Index()], false),
          Literal(out[neg.Index()], false), pos_t);
   }
   mClauses.emplace_back(t_literals);
}

