/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file unitprop.cpp
 * Linear time unit propagation.
 */

#include "unitprop.h"
#include "sbexception.h"
#include <sstream>

SUBool::UnitPropagator::UnitPropagator(const CNF *cnf, bool rebuild)
    : mCnf(cnf), mWatchLiterals(), mWatchedClauses(), mUnitClauses(),
      mEmptyClauses()
{
   if (rebuild)
   {
      Rebuild();
   }
}

void
SUBool::UnitPropagator::AddClauseInner(size_t clause_index)
{
   const Clause &clause = (*mCnf)[clause_index];
   switch (clause.Size())
   {
   case 0:
      mEmptyClauses.insert(clause_index);
      mWatchLiterals[clause_index] = std::make_pair(Watch(), Watch());
      break;
   case 1:
   {
      mUnitClauses.insert(clause_index);
      mWatchedClauses[clause[0].Index()].push_back(clause_index);
      Watch w(0, std::prev(mWatchedClauses[clause[0].Index()].end()));
      mWatchLiterals[clause_index] = std::make_pair(w, w);
   }
   break;
   default:
   {
      mWatchedClauses[clause[0].Index()].push_back(clause_index);
      auto w1 = Watch(0, std::prev(mWatchedClauses[clause[0].Index()].end()));
      mWatchedClauses[clause[1].Index()].push_back(clause_index);
      auto w2 = Watch(1, std::prev(mWatchedClauses[clause[1].Index()].end()));
      mWatchLiterals[clause_index] = std::make_pair(w1, w2);
   }
   }
}

void
SUBool::UnitPropagator::RemoveClause(size_t clause_index)
{
   const auto &clause = (*mCnf)[clause_index];
   auto &w = mWatchLiterals[clause_index];
   switch (clause.Size())
   {
   case 0:
      mEmptyClauses.erase(clause_index);
      break;
   case 1:
      mUnitClauses.erase(clause_index);
      mWatchedClauses[clause[0].Index()].erase(w.first.watch_list_ptr);
      break;
   default:
      mWatchedClauses[clause[w.first.index].Index()].erase(
          w.first.watch_list_ptr);
      mWatchedClauses[clause[w.second.index].Index()].erase(
          w.second.watch_list_ptr);
   }
   w.first.index = w.second.index = -1;
}

void
SUBool::UnitPropagator::AddClause(size_t clause_index)
{
   if (clause_index >= mWatchLiterals.size())
   {
      mWatchLiterals.resize(clause_index + 1);
   }
   if (2 * MaxVariable() > mWatchedClauses.size())
   {
      mWatchedClauses.resize(2 * MaxVariable());
   }
   AddClauseInner(clause_index);
}

void
SUBool::UnitPropagator::UpdateWatch(const Literal &lit,
                                    std::list<size_t>::iterator &it,
                                    const PartialAssignment &beta) const
{
   int i = *it;
   auto &w1 = mWatchLiterals[i].first;
   auto &w2 = mWatchLiterals[i].second;

   const Clause &clause = (*mCnf)[i];
   bool watch_updated = false;
   if (clause.Size() > 2)
   {
      if (clause[w2.index].Variable() == lit.Variable())
      {
         std::swap(w1, w2);
      }
      if (clause[w1.index].Value(beta) == PartialValue::FALSE)
      {
         for (int j = (w1.index + 1) % clause.Size(); j != w1.index;
              j = (j + 1) % clause.Size())
         {
            if (j != w2.index && clause[j].Value(beta) != PartialValue::FALSE)
            {
               watch_updated = true;
               auto neg = clause[w1.index];
               it = mWatchedClauses[neg.Index()].erase(it);
               mWatchedClauses[clause[j].Index()].push_back(i);
               w1.index = j;
               w1.watch_list_ptr =
                   std::prev(mWatchedClauses[clause[j].Index()].end());
               break;
            }
         }
      }
   }
   if (!watch_updated)
   {
      ++it;
   }

   if (clause[w1.index].Value(beta) == PartialValue::FALSE)
   {
      std::swap(w1, w2);
   }
}

SUBool::UnitPropagator::CLAUSE_STATE
SUBool::UnitPropagator::ClauseStatus(int i, const PartialAssignment &beta) const
{
   PartialValue v1 = (*mCnf)[i][mWatchLiterals[i].first.index].Value(beta);
   PartialValue v2 = (*mCnf)[i][mWatchLiterals[i].second.index].Value(beta);
   if (v1 == PartialValue::TRUE || v2 == PartialValue::TRUE)
   {
      return SATISFIED;
   }
   if (v1 == PartialValue::UNDEF && v2 == PartialValue::UNDEF)
   {
      return OTHER;
   }
   if (v1 == PartialValue::UNDEF || v2 == PartialValue::UNDEF)
   {
      return UNIT;
   }
   return UNSATISFIED;
}

int
SUBool::UnitPropagator::UpdateWatches(const Literal &lit,
                                      std::queue<Literal> &lq,
                                      const PartialAssignment &beta) const
{
   Literal neg(lit);
   neg.Negate();
   auto it = mWatchedClauses[neg.Index()].begin();
   while (it != mWatchedClauses[neg.Index()].end())
   {
      int i = *it;
      UpdateWatch(lit, it, beta);
      switch (ClauseStatus(i, beta))
      {
      case UNSATISFIED:
         return i;
      case UNIT:
      {
         lq.push((*mCnf)[i][mWatchLiterals[i].first.index]);
         break;
      }
      default:
         break;
      }
   }
   return -1;
}

int
SUBool::UnitPropagator::AddLiteralValue(const Literal &lit,
                                        std::queue<Literal> &lq,
                                        PartialAssignment &beta) const
{
   switch (lit.Value(beta))
   {
   case PartialValue::FALSE:
      return NumOfClauses();
      // throw InvariantFailedException(
      //"UnitPropagator::AddLiteralValue",
      //"Empty clause without a reason");
   case PartialValue::TRUE:
      break;
   case PartialValue::UNDEF:
      beta[lit.Variable()] = lit.Value(PartialValue::TRUE);
      return UpdateWatches(lit, lq, beta);
   }
   return -1;
}

void
SUBool::UnitPropagator::PropInit(const PartialAssignment &assump,
                                 std::queue<Literal> &lq) const
{
   for (auto i : mUnitClauses)
   {
      lq.push((*mCnf)[i][0]);
   }
   for (unsigned i = 1; i <= assump.Size(); ++i)
   {
      if (assump[i] != PartialValue::UNDEF)
      {
         lq.emplace(i, (assump[i] == PartialValue::TRUE));
      }
   }
}

int
SUBool::UnitPropagator::Propagate(const PartialAssignment &assump,
                                  PartialAssignment *conseq) const
{
   if (!mEmptyClauses.empty())
   {
      return *mEmptyClauses.begin();
   }

   SUBool::PartialAssignment beta(MaxVariable(), PartialValue::UNDEF);
   if (mWatchedClauses.size() < 2 * MaxVariable())
   {
      mWatchedClauses.resize(2 * MaxVariable());
   }
   std::queue<Literal> lq;
   PropInit(assump, lq);

   while (!lq.empty())
   {
      if (lq.front().Variable() <= MaxVariable())
      {
         int conflict = AddLiteralValue(lq.front(), lq, beta);
         if (conflict >= 0)
         {
            return conflict;
         }
      }
      lq.pop();
   }

   if (conseq != nullptr)
   {
      *conseq = std::move(beta);
      if (assump.Size() > conseq->Size())
      {
         auto s = conseq->Size();
         conseq->Resize(assump.Size(), PartialValue::UNDEF);
         for (unsigned var = s + 1; var <= assump.Size(); ++var)
         {
            (*conseq)[var] = assump[var];
         }
      }
   }
   return -1;
}
