/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file unitprop.h
 * Linear time unit propagation.
 */
#ifndef __UNITPROP_H
#define __UNITPROP_H

#include <list>
#include <queue>
#include <unordered_set>
#include <vector>

#include "assign.h"
#include "normalform.h"

namespace SUBool
{
   class UnitPropagator
   {
    private:
      const CNF *mCnf;

      struct Watch
      {
         int index;
         // the pointer is only valid if index is within the bounds in the
         // clause
         std::list<size_t>::iterator watch_list_ptr;

         Watch() : index(-1), watch_list_ptr() {}
         Watch(int i, std::list<size_t>::iterator ptr)
             : index(i), watch_list_ptr(ptr)
         {
         }
      };

      mutable std::vector<std::pair<Watch, Watch>> mWatchLiterals;
      mutable std::vector<std::list<size_t>> mWatchedClauses;
      std::unordered_set<int> mUnitClauses;
      std::unordered_set<int> mEmptyClauses;

      void AddClauseInner(size_t clause_index);

      int AddLiteralValue(const Literal &lit, std::queue<Literal> &lq,
                          PartialAssignment &beta) const;

      void PropInit(const PartialAssignment &assump,
                    std::queue<Literal> &lq) const;

      enum CLAUSE_STATE
      {
         UNSATISFIED,
         SATISFIED,
         UNIT,
         OTHER
      };

      int UpdateWatches(const Literal &lit, std::queue<Literal> &lq,
                        const PartialAssignment &beta) const;
      void UpdateWatch(const Literal &lit, std::list<size_t>::iterator &it,
                       const PartialAssignment &beta) const;
      CLAUSE_STATE ClauseStatus(int i, const PartialAssignment &beta) const;

    public:
      UnitPropagator(const CNF *cnf, bool rebuild = true);

      UnitPropagator(const SUBool::UnitPropagator &up) = default;
      UnitPropagator(SUBool::UnitPropagator &&up) = default;
      ~UnitPropagator() = default;
      UnitPropagator &operator=(const UnitPropagator &up) = default;
      UnitPropagator &operator=(UnitPropagator &&up) = default;

      // Only includes clauses for which the predicate is satisfied.
      template <class Pred> UnitPropagator(const CNF *cnf, Pred pred);
      template <class Pred> void Rebuild(Pred pred);
      void
      Rebuild()
      {
         Rebuild([](unsigned) { return true; });
      }

      // In case the empty clause is generated, an index of a
      // conflict clause within mCnf is returned. Otherwise the
      // function returns -1.
      int Propagate(const PartialAssignment &assump,
                    PartialAssignment *conseq) const;
      unsigned
      MaxVariable() const
      {
         return (mCnf == nullptr ? 0 : mCnf->MaxVariable());
      }
      size_t
      NumOfClauses() const
      {
         return (mCnf == nullptr ? 0 : mCnf->Size());
      }

      // The clause must be in the CNF
      void AddClause(size_t clause_index);
      void RemoveClause(size_t clause_index);
   };
} // namespace SUBool

// Implementation of templates

template <class Pred>
SUBool::UnitPropagator::UnitPropagator(const CNF *cnf, Pred pred)
    : mCnf(cnf), mWatchLiterals(), mWatchedClauses(), mUnitClauses(),
      mEmptyClauses()
{
   Rebuild(pred);
}

template <class Pred>
void
SUBool::UnitPropagator::Rebuild(Pred pred)
{
   mWatchLiterals.resize(NumOfClauses());
   mWatchedClauses.clear();
   mWatchedClauses.resize(2 * MaxVariable());
   mUnitClauses.clear();
   mEmptyClauses.clear();

   for (unsigned i = 0; i < NumOfClauses(); ++i)
   {
      if (pred(i))
      {
         AddClauseInner(i);
      }
   }
}

#endif
