/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file unitprop2.h
 * Linear time unit propagation (with CNF managed by the propagator).
 */
#include <sstream>

#include "sbexception.h"
#include "unitprop2.h"
#include "util.h"

SUBool::UnitPropagatorM::UnitPropagatorM(const CNF &cnf) : UnitPropagatorM()
{
   UpdateMaxVariable(cnf.MaxVariable());
   for (const auto &clause : cnf)
   {
      // std::cout << "adding " << clause << std::endl;
      AddClause(clause);
      // std::cout << "added " << clause << std::endl;
      if (mContradictory)
      {
         break;
      }
   }
}

SUBool::UnitPropagatorM::UnitPropagatorM(CNF &&cnf) : UnitPropagatorM()
{
   UpdateMaxVariable(cnf.MaxVariable());
   for (const auto &clause : cnf)
   {
      // std::cout << "adding " << clause << std::endl;
      AddClause(std::move(clause));
      // std::cout << "added " << clause << std::endl;
      if (mContradictory)
      {
         break;
      }
   }
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddClause(const Clause &clause)
{
   UpdateMaxVariable(clause.MaxVariable());
   switch (clause.Size())
   {
   case 0:
      mContradictory = true;
      return end();
   case 1:
      return AddUnitClause(clause);
   case 2:
      return AddBinaryClause(clause);
   default:
      return AddBigClause(clause);
   }
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddClause(Clause &&clause)
{
   UpdateMaxVariable(clause.MaxVariable());
   switch (clause.Size())
   {
   case 0:
      mContradictory = true;
      return end();
   case 1:
      return AddUnitClause(std::move(clause));
   case 2:
      return AddBinaryClause(std::move(clause));
   default:
      return AddBigClause(std::move(clause));
   }
}

bool
SUBool::UnitPropagatorM::ShouldAddUnitClause(const Clause &clause) const
{
   return mUnitClauses.find(clause.Front()) == mUnitClauses.end();
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddUnitClause(const Clause &clause)
{
   if (!ShouldAddUnitClause(clause))
   {
      return end();
   }
   mClauses.push_back(clause);
   auto it = std::prev(mClauses.end());
   EnableUnitClause(it);
   return it;
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddUnitClause(Clause &&clause)
{
   if (!ShouldAddUnitClause(clause))
   {
      return end();
   }
   mClauses.push_back(std::move(clause));
   auto it = std::prev(mClauses.end());
   EnableUnitClause(it);
   return it;
}

bool
SUBool::UnitPropagatorM::ShouldAddBinaryClause(const Clause &clause) const
{
   auto &neigh = mBinaryNeighbours[clause.Front().Negation().Index()];
   return neigh.find(clause.Back()) == neigh.end();
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddBinaryClause(const Clause &clause)
{
   if (!ShouldAddBinaryClause(clause))
   {
      return end();
   }
   mClauses.push_back(clause);
   auto it = std::prev(mClauses.end());
   EnableBinaryClause(it);
   return it;
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddBinaryClause(Clause &&clause)
{
   if (!ShouldAddBinaryClause(clause))
   {
      return end();
   }
   mClauses.push_back(std::move(clause));
   auto it = std::prev(mClauses.end());
   EnableBinaryClause(it);
   return it;
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddBigClause(const Clause &clause)
{
   mClauses.push_back(clause);
   auto it = std::prev(mClauses.end());
   EnableBigClause(it);
   return it;
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::AddBigClause(Clause &&clause)
{
   mClauses.push_back(std::move(clause));
   auto it = std::prev(mClauses.end());
   EnableBigClause(it);
   return it;
}

bool
SUBool::UnitPropagatorM::EnableClause(const_iterator i_clause)
{
   if (mContradictory)
   {
      return false;
   }
   UpdateMaxVariable(i_clause->clause.MaxVariable());
   switch (i_clause->clause.Size())
   {
   case 0:
      mContradictory = true;
      return false;
   case 1:
      return EnableUnitClause(i_clause);
   case 2:
      return EnableBinaryClause(i_clause);
   default:
      return EnableBigClause(i_clause);
   }
}

bool
SUBool::UnitPropagatorM::EnableUnitClause(const_iterator i_clause)
{
   const auto &lit = i_clause->clause.Front();
   if (mUnitClauses.find(lit.Negation()) != mUnitClauses.end())
   {
      mContradictory = true;
      return false;
   }
   mUnitClauses.insert(lit);
   return true;
}

bool
SUBool::UnitPropagatorM::EnableBinaryClause(const_iterator i_clause)
{
   const auto &front = i_clause->clause.Front();
   const auto &back = i_clause->clause.Back();
   mBinaryNeighbours[front.Negation().Index()].insert(back);
   mBinaryNeighbours[back.Negation().Index()].insert(front);
   return true;
}

void
SUBool::UnitPropagatorM::AddWatchToList(const_iterator i_clause, Watch &w) const
{
   auto &lit_watched = mWatchedClauses[i_clause->clause[w.index].Index()];
   lit_watched.push_back(i_clause);
   w.watch_list_index = lit_watched.size() - 1;
}

void
SUBool::UnitPropagatorM::RemoveWatchFromList(const_iterator i_clause, Watch &w)
{
   if (w.index < 0)
   {
      return;
   }
   auto &lit = i_clause->clause[w.index];
   auto &lit_watched = mWatchedClauses[lit.Index()];
   // std::cout << "Removing " << *i_clause << " from the list for " << lit
   //         << std::endl;
   // std::cout << "List before:";
   // std::for_each(lit_watched.begin(), lit_watched.end(),
   //[](auto x) { std::cout << " " << *x; });
   // std::cout << std::endl;
   // std::cout << "w.watch_list_index: " << w.watch_list_index << std::endl;
   // std::cout << "lit_watched.size: " << lit_watched.size() << std::endl;
   if (w.watch_list_index + 1 < lit_watched.size())
   {
      MoveWatch(lit, lit_watched, lit_watched.size() - 1, w.watch_list_index);
   }
   lit_watched.pop_back();
   // std::cout << "List after:";
   // std::for_each(lit_watched.begin(), lit_watched.end(),
   //[](auto x) { std::cout << " " << *x; });
   // std::cout << std::endl;
   w.index = -1;
}

void
SUBool::UnitPropagatorM::UpdateMaxVariable(unsigned max_var)
{
   if (max_var > mMaxVariable)
   {
      mMaxVariable = max_var;
      mValues.Resize(mMaxVariable, PartialValue::UNDEF);
      mWatchedClauses.resize(2 * mMaxVariable);
      mBinaryNeighbours.resize(2 * mMaxVariable);
   }
}

bool
SUBool::UnitPropagatorM::EnableBigClause(const_iterator i_clause)
{
   i_clause->w1.index = 0;
   i_clause->w2.index = 1;
   AddWatchToList(i_clause, i_clause->w1);
   AddWatchToList(i_clause, i_clause->w2);
   // std::cout << "Enabling: " << *i_clause << std::endl;
   return true;
}

SUBool::UnitPropagatorM::const_iterator
SUBool::UnitPropagatorM::RemoveClause(const_iterator i_clause)
{
   DisableClause(i_clause);
   return mClauses.erase(i_clause);
}

void
SUBool::UnitPropagatorM::DisableBinaryClause(const_iterator i_clause)
{
   const auto &front = i_clause->clause.Front();
   const auto &back = i_clause->clause.Back();
   mBinaryNeighbours[front.Negation().Index()].erase(back);
   mBinaryNeighbours[back.Negation().Index()].erase(front);
}

void
SUBool::UnitPropagatorM::DisableClause(const_iterator i_clause)
{
   const auto &clause = i_clause->clause;
   auto &w1 = i_clause->w1;
   auto &w2 = i_clause->w2;
   switch (clause.Size())
   {
   case 0:
      // mContradictory status cannot change once set
      break;
   case 1:
      mUnitClauses.erase(clause.Front());
      break;
   case 2:
      DisableBinaryClause(i_clause);
      break;
   default:
      // std::cout << "Disabling: " << *i_clause << std::endl;
      RemoveWatchFromList(i_clause, w1);
      RemoveWatchFromList(i_clause, w2);
   }
}

bool
SUBool::UnitPropagatorM::UpdateWatch(const Literal &lit,
                                     const_iterator i_clause) const
{
   auto &w1 = i_clause->w1;
   auto &w2 = i_clause->w2;
   const Clause &clause = i_clause->clause;
   bool watch_updated = false;
   // std::cout << "Updating due to " << lit << ": " << *i_clause << std::endl;
   assert(clause.Size() > 2);
   assert(w1.index >= 0);
   assert(w2.index >= 0);
   if (clause[w2.index].Variable() == lit.Variable())
   {
      std::swap(w1, w2);
   }
   if (clause[w1.index].Value(mValues) == PartialValue::FALSE)
   {
      for (int j = (w1.index + 1) % clause.Size(); j != w1.index;
           j = (j + 1) % clause.Size())
      {
         if (j != w2.index && clause[j].Value(mValues) != PartialValue::FALSE)
         {
            watch_updated = true;
            w1.index = j;
            AddWatchToList(i_clause, w1);
            break;
         }
      }
   }
   if (clause[w1.index].Value(mValues) == PartialValue::FALSE)
   {
      std::swap(w1, w2);
   }
   // if (watch_updated)
   //{
   // std::cout << "Updated to " << *i_clause << std::endl;
   //}
   // else
   //{
   // std::cout << "No update" << std::endl;
   //}
   return watch_updated;
}

SUBool::UnitPropagatorM::CLAUSE_STATE
SUBool::UnitPropagatorM::ClauseStatus(const_iterator i_clause) const
{
   // std::cout << "Checking status of clause " << *i_clause << std::endl;
   // std::cout << "with values: " << mValues << std::endl;
   PartialValue v1 = i_clause->FirstWatchedLiteral().Value(mValues);
   PartialValue v2 = i_clause->SecondWatchedLiteral().Value(mValues);
   if (v1 == PartialValue::TRUE || v2 == PartialValue::TRUE)
   {
      // std::cout << "SATISFIED" << std::endl;
      return SATISFIED;
   }
   if (v1 == PartialValue::UNDEF && v2 == PartialValue::UNDEF)
   {
      // std::cout << "OTHER" << std::endl;
      return OTHER;
   }
   if (v1 == PartialValue::UNDEF || v2 == PartialValue::UNDEF)
   {
      // std::cout << "UNIT" << std::endl;
      return UNIT;
   }
   // std::cout << "UNSATISFIED" << std::endl;
   return UNSATISFIED;
}

void
SUBool::UnitPropagatorM::MoveWatch(const Literal &lit,
                                   const_iterator_list &watch_list, size_t src,
                                   size_t target)
{
   watch_list[target] = std::move(watch_list[src]);
   auto i_moved_clause = watch_list[target];
   // std::cout << "moved: " << *i_moved_clause << std::endl;
   if (i_moved_clause->clause[i_moved_clause->w1.index] == lit)
   {
      i_moved_clause->w1.watch_list_index = target;
   }
   else
   {
      assert(i_moved_clause->clause[i_moved_clause->w2.index] == lit);
      i_moved_clause->w2.watch_list_index = target;
   }
}

bool
SUBool::UnitPropagatorM::UpdateWatches(const Literal &lit) const
{
   auto lit_neg = lit.Negation();
   auto &watched_list = mWatchedClauses[lit_neg.Index()];
   // std::cout << "Updating watches in list for " << lit_neg << std::endl;
   size_t new_size = 0;
   bool unsat = false;
   for (size_t watch_index = 0; watch_index < watched_list.size();
        ++watch_index)
   {
      auto i_clause = watched_list[watch_index];
      if (unsat || !UpdateWatch(lit, i_clause))
      {
         if (watch_index > new_size)
         {
            MoveWatch(lit_neg, watched_list, watch_index, new_size);
         }
         ++new_size;
      }
      if (unsat)
      {
         continue;
      }
      switch (ClauseStatus(i_clause))
      {
      case UNSATISFIED:
         unsat = true;
         {
            if (watch_index + 1 == new_size)
            {
               // the list is unchanged
               return false;
            }
            /*std::cout << removed << " ?=? " << watch_index - new_size + 1
                      << std::endl;*/
            size_t removed = watch_index - new_size + 1;
            size_t rest = watched_list.size() - watch_index - 1;
            if (rest > removed)
            {
               for (unsigned i = 0; i < removed; ++i)
               {
                  MoveWatch(lit_neg, watched_list, watched_list.size() - 1,
                            new_size);
                  ++new_size;
                  watched_list.pop_back();
               }
               return false;
            }
         }
         break;
      case UNIT:
      {
         // std::cout << "Adding " << i_clause->FirstWatchedLiteral()
         // << " to queue" << std::endl;
         mLiteralQueue.push(i_clause->FirstWatchedLiteral());
         break;
      }
      default:
         break;
      }
   }
   watched_list.resize(new_size);
   return !unsat;
}

bool
SUBool::UnitPropagatorM::PushBinaryNeighbours(const Literal &lit) const
{
   for (const auto &nb : mBinaryNeighbours[lit.Index()])
   {
      mLiteralQueue.push(nb);
   }
   return true;
}

bool
SUBool::UnitPropagatorM::AddLiteralValue(const Literal &lit) const
{
   // std::cout << "Adding literal value " << lit << std::endl;
   switch (lit.Value(mValues))
   {
   case PartialValue::FALSE:
      return false;
   case PartialValue::TRUE:
      return true;
   case PartialValue::UNDEF:
   {
      mValues[lit.Variable()] = lit.Value(PartialValue::TRUE);
      return PushBinaryNeighbours(lit) && UpdateWatches(lit);
   }
   default:
      throw InvariantFailedException("UnitPropagator::AddLiteralValue",
                                     "Unknown literal value");
   }
}

void
SUBool::UnitPropagatorM::PropInit() const
{
   mLiteralQueue = std::queue<Literal>();
   mValues.Reset(PartialValue::UNDEF);
   for (const auto &lit : mUnitClauses)
   {
      mLiteralQueue.push(lit);
   }
}

bool
SUBool::UnitPropagatorM::PropagateQueue() const
{
   bool no_conflict = true;
   while (!mLiteralQueue.empty() && no_conflict)
   {
      auto lit = mLiteralQueue.front();
      // std::cout << "queue literal " << lit << std::endl;
      mLiteralQueue.pop();
      no_conflict = AddLiteralValue(lit);
   }
   // std::cout << "returning " << bool_to_string(no_conflict) << std::endl;
   return no_conflict;
}

bool
SUBool::UnitPropagatorM::InnerPropagate(
    const std::vector<Literal> &assump) const
{
   if (mContradictory)
   {
      return false;
   }
   PropInit();
   if (!PropagateQueue())
   {
      mContradictory = true;
      return false;
   }
   // std::cout << "New assumptions" << std::endl;
   for (const auto &lit : assump)
   {
      // std::cout << "Assumption literal: " << lit << std::endl;
      if (lit.Variable() <= MaxVariable()
          && (!AddLiteralValue(lit) || !PropagateQueue()))
      {
         // std::cout << "false" << std::endl;
         return false;
      }
   }
   // std::cout << "true" << std::endl;
   return true;
}

SUBool::CNF
SUBool::UnitPropagatorM::ExtractCNF() const &
{
   if (mContradictory)
   {
      return CNF(Clause::EmptyClause());
   }
   CNF cnf;
   for (const auto &i_clause : mClauses)
   {
      cnf.PushBack(i_clause.clause);
   }
   return cnf;
}

SUBool::CNF
SUBool::UnitPropagatorM::ExtractCNF() &&
{
   if (mContradictory)
   {
      return CNF(Clause::EmptyClause());
   }
   CNF cnf;
   for (auto &i_clause : mClauses)
   {
      cnf.PushBack(std::move(i_clause.clause));
   }
   return cnf;
}

std::ostream &
SUBool::operator<<(std::ostream &out,
                   const SUBool::UnitPropagatorM::ClauseInfo &ci)
{
   out << ci.clause << " (" << ci.FirstWatchedLiteral() << ", "
       << ci.SecondWatchedLiteral() << ")";
   return out;
}
