/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file unitprop2.h
 * Linear time unit propagation (with CNF managed by the propagator).
 */
#ifndef __UNITPROP2_H
#define __UNITPROP2_H

#include <list>
#include <queue>
#include <unordered_map>
#include <unordered_set>
#include <vector>

#include "assign.h"
#include "normalform.h"

namespace SUBool
{
   class UnitPropagatorM
   {
    public:
      struct ClauseInfo;
      using clause_list = std::list<ClauseInfo>;
      using iterator = clause_list::iterator;
      using const_iterator = clause_list::const_iterator;
      using const_iterator_list = std::vector<const_iterator>;

      struct Watch
      {
         int index;
         // the index is only valid if index is within the bounds of the
         // clause
         size_t watch_list_index;

         Watch() : index(-1), watch_list_index(0) {}
         Watch(int i) : index(i), watch_list_index(0) {}
         Watch(int i, size_t watch_index)
             : index(i), watch_list_index(watch_index)
         {
         }
      };

      struct ClauseInfo
      {
         Clause clause;
         mutable Watch w1, w2;

         ClauseInfo(Clause c, int w1_index = -1, int w2_index = -1)
             : clause(std::move(c)), w1(w1_index), w2(w2_index)
         {
         }
         Literal
         FirstWatchedLiteral() const
         {
            return clause[w1.index];
         }
         Literal
         SecondWatchedLiteral() const
         {
            return clause[w2.index];
         }
      };

    private:
      clause_list mClauses;
      mutable std::vector<const_iterator_list> mWatchedClauses;
      std::unordered_set<Literal> mUnitClauses;
      std::vector<std::unordered_set<Literal>> mBinaryNeighbours{};
      mutable bool mContradictory; // cannot be unset once set to true
      unsigned mMaxVariable{0};
      mutable PartialAssignment mValues;
      mutable std::queue<Literal> mLiteralQueue{};

      bool AddLiteralValue(const Literal &lit) const;

      void PropInit() const;

      enum CLAUSE_STATE
      {
         UNSATISFIED,
         SATISFIED,
         UNIT,
         OTHER
      };

      bool PushBinaryNeighbours(const Literal &lit) const;
      bool UpdateWatches(const Literal &lit) const;
      // Returns if the watch was updated and moved to another list
      bool UpdateWatch(const Literal &lit, const_iterator i_clause) const;
      CLAUSE_STATE ClauseStatus(const_iterator i_clause) const;

      bool ShouldAddUnitClause(const Clause &clause) const;
      bool ShouldAddBinaryClause(const Clause &clause) const;

      bool EnableUnitClause(const_iterator i_clause);
      bool EnableBinaryClause(const_iterator i_clause);
      bool EnableBigClause(const_iterator i_clause);
      void AddWatchToList(const_iterator i_clause, Watch &w) const;
      void RemoveWatchFromList(const_iterator i_clause, Watch &w);

      void DisableBinaryClause(const_iterator i_clause);

      const_iterator AddUnitClause(const Clause &clause);
      const_iterator AddUnitClause(Clause &&clause);
      const_iterator AddBinaryClause(const Clause &clause);
      const_iterator AddBinaryClause(Clause &&clause);
      const_iterator AddBigClause(const Clause &clause);
      const_iterator AddBigClause(Clause &&clause);

      bool PropagateQueue() const;
      bool InnerPropagate(const std::vector<Literal> &assump) const;

      void UpdateMaxVariable(unsigned max_var);
      static void MoveWatch(const Literal &lit, const_iterator_list &watch_list,
          size_t src, size_t target);

    public:
      UnitPropagatorM()
          : mClauses(), mWatchedClauses(), mUnitClauses(),
            mContradictory(false), mMaxVariable(0), mValues(0)
      {
      }
      UnitPropagatorM(const CNF &cnf);
      UnitPropagatorM(CNF &&cnf);

      UnitPropagatorM(const SUBool::UnitPropagatorM &up) = default;
      UnitPropagatorM(SUBool::UnitPropagatorM &&up) = default;
      ~UnitPropagatorM() = default;
      UnitPropagatorM &operator=(const UnitPropagatorM &up) = default;
      UnitPropagatorM &operator=(UnitPropagatorM &&up) = default;

      // Only includes clauses for which the predicate is satisfied.
      template <class Pred> UnitPropagatorM(const CNF &cnf, Pred pred);

      std::optional<PartialAssignment>
      Propagate(const PartialAssignment &assump) const
      {
         return Propagate(literals_of_assignment(assump));
      }

      bool
      Consistent(const PartialAssignment &assump) const
      {
         return Consistent(literals_of_assignment(assump));
      }

      std::optional<PartialAssignment>
      Propagate(const std::vector<Literal> &assump) const
      {
         return Consistent(assump) ? std::make_optional(mValues)
                                   : std::optional<PartialAssignment>();
      }

      bool
      Consistent(const std::vector<Literal> &assump) const
      {
         return InnerPropagate(assump);
      }

      bool
      Contradictory() const
      {
         return mContradictory;
      }

      bool
      Closed(const PartialAssignment &alpha) const
      {
         return Consistent(alpha) && pa_equal(mValues, alpha);
      }

      unsigned
      MaxVariable() const
      {
         return mMaxVariable;
      }
      size_t
      NumOfClauses() const
      {
         return mClauses.size();
      }

      CNF ExtractCNF() const &;
      CNF ExtractCNF() &&;

      const_iterator
      begin() const
      {
         return mClauses.begin();
      }
      const_iterator
      end() const
      {
         return mClauses.end();
      }
      const_iterator
      cbegin() const
      {
         return mClauses.cbegin();
      }
      const_iterator
      cend() const
      {
         return mClauses.cend();
      }

      // Returns iterator to the added clause info, or end if the clause was not
      // actually added
      const_iterator AddClause(const Clause &clause);
      const_iterator AddClause(Clause &&clause);
      const_iterator RemoveClause(const_iterator i_clause);
      void DisableClause(const_iterator i_clause);
      // Returns the contradictory status after enabling the clause
      bool EnableClause(const_iterator i_clause);
   };

   std::ostream &operator<<(
       std::ostream &out, const UnitPropagatorM::ClauseInfo &ci);
} // namespace SUBool

#endif
