/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file up_impl_level.cpp
 * Encoding of unit propagation based on assigning k-bit levels to each variable
 * and implication.
 */

#include "up_impl_level.h"
#include "up_levels.h"

SUBool::UPEncImplLevel::UPEncImplLevel(const Config &conf, const CNF &cnf,
    const ClausesToLiterals &cl2lit, VarStore &var_store,
    const LitImplSystem &is)
    : UPEncoder(cnf, cl2lit, var_store), mImplSystem(is), mConf(conf)
{
}

unsigned
SUBool::UPEncImplLevel::MaxLevelBound() const
{
   return (mConf.use_bit_level ? BitSizeForLevel(0) : mCnf.MaxVariable());
}

unsigned
SUBool::UPEncImplLevel::NextLevelBound(unsigned level) const
{
   if (level == 0 || level >= MaxLevelBound())
   {
      return 0;
   }
   if (mConf.use_bit_level)
   {
      return level + 1;
   }
   // For simplicity, we compute the next level by subsequent increasing the
   // current level.
   // It can get slow for a high value of level.
   // We should give an explicit formula (close to doubling the level).
   unsigned bits = BitSizeForLevel(level++);
   while (BitSizeForLevel(level) == bits)
   {
      ++level;
   }
   return (level >= MaxLevelBound() ? 0 : level);
}

void
SUBool::UPEncImplLevel::Encode(unsigned max_level,
    std::vector<unsigned> *inputs, std::vector<unsigned> *outputs,
    std::vector<Clause> *clauses) const
{
   if (inputs == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::Encode", "inputs == nullptr");
   }
   if (outputs == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::Encoder", "outputs == nullptr");
   }
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   unsigned bit_size =
       (mConf.use_bit_level ? max_level : BitSizeForLevel(max_level));
   mVarStore.InitVarArray(inputs, mCnf.MaxVariable(), true, "i_{", "}");
   mVarStore.InitVarArray(outputs, mCnf.MaxVariable(), true, "o_{", "}");
   UPLevels level_vars(mCnf, mVarStore, true, bit_size, "");
   // lit_clauses could be directly added to clauses and we could only remember
   // the offset of the start of lit_clauses within the clauses array
   auto lit_clauses = InitLitClauses(*inputs, *outputs);
   AddUnitsForEmptyBodies(*inputs, clauses);
   bool new_var = true;
   for (unsigned i = 0; i < mImplSystem.Size(); ++i)
   {
      AddImplication(i, /* *inputs, */ *outputs, level_vars, lit_clauses,
          new_var, clauses);
   }
   for (auto &lit_clause : lit_clauses)
   {
      clauses->push_back(std::move(lit_clause));
   }
   AssumptionsImplyConsequences(*inputs, *outputs, clauses);
   AssumptionsHaveZeroLevel(*inputs, level_vars, clauses);
}

void
SUBool::UPEncImplLevel::AddUnitsForEmptyBodies(
    const std::vector<unsigned> inputs, std::vector<Clause> *clauses) const
{
   for (const auto &impl : mImplSystem)
   {
      if (impl.EmptyBody())
      {
         for (const auto &lit : impl.Cons())
         {
            clauses->emplace_back(Literal(inputs[lit.Index()], true));
         }
      }
   }
}

SUBool::NFSize
SUBool::UPEncImplLevel::CountAddUnitsForEmptyBodies() const
{
   NFSize s;
   for (const auto &impl : mImplSystem)
   {
      if (impl.EmptyBody())
      {
         s.size += impl.ConsSize();
         s.len += impl.ConsSize();
      }
   }
   return s;
}

std::vector<SUBool::Clause>
SUBool::UPEncImplLevel::InitLitClauses(const std::vector<unsigned> &inputs,
    const std::vector<unsigned> &outputs) const
{
   std::vector<Clause> lit_clauses(2 * mCnf.MaxVariable());
   for (unsigned lit_index = 0; lit_index < lit_clauses.size(); ++lit_index)
   {
      lit_clauses[lit_index].PushBack(Literal(outputs[lit_index], false));
      lit_clauses[lit_index].PushBack(Literal(inputs[lit_index], true));
   }
   return lit_clauses;
}

void
SUBool::UPEncImplLevel::AddImplication(unsigned impl_index,
    // const std::vector<unsigned> &inputs,
    const std::vector<unsigned> &outputs, UPLevels &level_vars,
    std::vector<Clause> &lit_clauses, bool &new_var,
    std::vector<Clause> *clauses) const
{
   const auto &impl = mImplSystem[impl_index];
   if (impl.EmptyBody())
   {
      return;
   }
   // logs.TLog(3) << "Adding implication " << mImplSystem[impl_index]
   //<< std::endl;
   auto impl_var = InitImplVar(impl_index, outputs, level_vars, new_var);
   // Determine implvar
   for (const auto &lit : impl.Cons())
   {
      AddImplToLitClause(lit, impl_index, impl_var, level_vars,
          lit_clauses[lit.Index()], clauses);
   }
   if (impl.BodySize() > 1)
   {
      for (const auto &lit : impl.Body())
      {
         AddBodyLitClauses(impl_var, lit, level_vars, outputs, clauses);
      }
   }
}

SUBool::UPEncImplLevel::impl_var_t
SUBool::UPEncImplLevel::InitImplVar(unsigned impl_index,
    const std::vector<unsigned> &outputs, UPLevels &level_vars,
    bool &new_var) const
{
   const auto &impl = mImplSystem[impl_index];
   if (impl.Body().size() == 1)
   {
      return {outputs[impl.Body().front().Index()], impl.Body().front()};
   }
   // Add variable only if the last used literal is negative, if
   // positive, use the last variable in a negative literal
   // This way the level variables for literals do not contain gaps
   impl_var_t impl_var;
   mVarStore.NameStream() << "f_{" << impl_index << "="
                          << level_vars.MaxVariable() << "}";
   impl_var.aux_var = mVarStore.NewVariable();
   if (new_var)
   {
      level_vars.AddVariable();
   }
   impl_var.level_literal = Literal(level_vars.MaxVariable(), new_var);
   new_var = !new_var;
   return impl_var;
   // return {mVarStore.NewVariable(), Literal(level_vars.MaxVariable(), true)};
}

void
SUBool::UPEncImplLevel::AddImplToLitClause(const Literal &cons_lit,
    unsigned impl_index, impl_var_t impl_var, UPLevels &level_vars,
    Clause &lit_clause, std::vector<Clause> *clauses) const
{
   // In case of single literal in a body when the bodies are not merged and the
   // implications are not right irredundant, this could lead to repeating
   // comparisons, for the same pairs of variables. We ignore this fact, since
   // it is unlikely this would happen given the fact how the input implication
   // system is created in pccompile.
   //
   // TODO: - Need to use literal levels or assume that the CNF is UP prime.
   //       - Or detect contradictory bodies and for them, directly add
   //         implication from the body to all output literals
   //       - We will need to
   //         1. check the cons if there is a conflict
   //         2. compute the closure and check the closure for a conflict
   //         3. Or directly compute the closure? It could be slower in case of
   //            right saturated systems
   //       - We could have two options in the configuration
   //         check_contradictory_bodies - set if there is no preprocessing
   //         is_right_saturated - set when implication system is right
   //         saturated
   //       - The easiest option still is to enforce that the preprocessing is
   //         at least up_prime, the problem is that up_prime is not performed
   //         until fixed point, but only one round. It means that in the end,
   //         clauses may not necessarily be minimal 1-provable.
   //       - Literal levels are necessary, because situations such as
   //         ... x ... -> ... -x ... cannot be avoided. The difference from the
   //         logarithmic case is that here we do not have redundancy in
   //         general.
   auto out_var =
       level_vars.EncodeGT(cons_lit, impl_var.level_literal, clauses);
   mVarStore.NameStream() << "z_{" << impl_index << "}^{" << cons_lit << "}";
   auto and_var = mVarStore.NewVariable();
   clauses->emplace_back(
       Literal(and_var, false), Literal(impl_var.aux_var, true));
   clauses->emplace_back(Literal(and_var, false), Literal(out_var, true));
   lit_clause.PushBack(Literal(and_var, true));
}

void
SUBool::UPEncImplLevel::AddBodyLitClauses(impl_var_t impl_var,
    const Literal &body_literal, UPLevels &level_vars,
    const std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const
{
   clauses->emplace_back(Literal(impl_var.aux_var, false),
       Literal(outputs[body_literal.Index()], true));
   // TODO: Encode directly LE
   auto out_var =
       level_vars.EncodeGT(body_literal, impl_var.level_literal, clauses);
   clauses->emplace_back(
       Literal(impl_var.aux_var, false), Literal(out_var, false));
}

void
SUBool::UPEncImplLevel::AssumptionsImplyConsequences(
    const std::vector<unsigned> &inputs, const std::vector<unsigned> &outputs,
    std::vector<Clause> *clauses) const
{
   for (unsigned lit_index = 0; lit_index < inputs.size(); ++lit_index)
   {
      clauses->emplace_back(
          Literal(inputs[lit_index], false), Literal(outputs[lit_index], true));
   }
}

void
SUBool::UPEncImplLevel::AssumptionsHaveZeroLevel(
    const std::vector<unsigned> &inputs, const UPLevels &level_vars,
    std::vector<Clause> *clauses) const
{
   unsigned c = 0;
   for (unsigned i = 0; i < inputs.size(); ++i)
   {
      Literal in(inputs[i], false);
      const auto &lvars = level_vars.LevelVars(Literal(i));
      for (unsigned var : lvars)
      {
         ++c;
         clauses->emplace_back(in, Literal(var, false));
      }
   }
}

SUBool::NFSize
SUBool::UPEncImplLevel::CountAssumptionsImplyConsequences() const
{
   return {0, 2 * mCnf.MaxVariable(), 4 * mCnf.MaxVariable()};
}

SUBool::NFSize
SUBool::UPEncImplLevel::CountAssumptionsHaveZeroLevel(unsigned bit_size) const
{
   return {
       0, 2 * mCnf.MaxVariable() * bit_size, 4 * bit_size * mCnf.MaxVariable()};
}

SUBool::NFSize
SUBool::UPEncImplLevel::EstimateSize(unsigned max_level) const
{
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   unsigned bit_size =
       (mConf.use_bit_level ? max_level : BitSizeForLevel(max_level));
   // logs.TLog(1) << "bit_size: " << bit_size << std::endl;
   // Inputs and outputs for each literal
   NFSize s{4 * mCnf.MaxVariable(), 0, 0};
   s += UPLevels::CountInitialSize(mCnf, true, bit_size);
   s += CountAddUnitsForEmptyBodies();
   // InitLitClauses
   s.size += 2 * mCnf.MaxVariable();
   s.len += 4 * mCnf.MaxVariable();
   bool new_var = true;
   for (unsigned i = 0; i < mImplSystem.Size(); ++i)
   {
      s += CountAddImplication(i, bit_size, new_var);
   }
   s += CountAssumptionsImplyConsequences();
   s += CountAssumptionsHaveZeroLevel(bit_size);
   return s;
}

SUBool::NFSize
SUBool::UPEncImplLevel::CountAddImplication(
    unsigned impl_index, unsigned bit_size, bool &new_var) const
{
   NFSize s;
   const auto &impl = mImplSystem[impl_index];
   if (impl.EmptyBody())
   {
      return s;
   }
   s += CountInitImplVar(impl_index, bit_size, new_var);
   s += UPLevels::CountEncodeGTSize(true, bit_size) * impl.ConsSize();
   s += NFSize{1, 2, 5} * impl.ConsSize();
   if (impl.BodySize() > 1)
   {
      s += UPLevels::CountEncodeGTSize(true, bit_size) * impl.BodySize();
      s += NFSize{0, 2, 4} * impl.BodySize();
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncImplLevel::CountInitImplVar(
    unsigned impl_index, unsigned bit_size, bool &new_var) const
{
   NFSize s;
   const auto &impl = mImplSystem[impl_index];
   if (impl.BodySize() > 1)
   {
      ++s.max_variable;
      if (new_var)
      {
         s.max_variable += 2 * bit_size;
      }
      new_var = !new_var;
   }
   return s;
}
