/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file up_levels.cpp
 * Functions for encoding levels of unit propagations encoded in
 * binary and their comparisons.
 */

#include "up_levels.h"

SUBool::UPLevels::UPLevels(const CNF &cnf, VarStore &var_store,
    bool literal_levels, unsigned level_bound, const std::string &name_base)
    : mCnf(cnf), mVarStore(var_store), mLiteralLevels(literal_levels),
      mLevelBound(level_bound), mLevelVars(), mNameBase(name_base),
      mMaxVariable(0)
{
   for (unsigned i = 0; i < mCnf.MaxVariable(); ++i)
   {
      AddVariable();
   }
}

const std::vector<unsigned> &
SUBool::UPLevels::LevelVars(unsigned var) const
{
   if (mLiteralLevels)
   {
      var += var;
   }
   return mLevelVars[var];
}

const std::vector<unsigned> &
SUBool::UPLevels::LevelVars(const Literal &lit) const
{
   const unsigned index = (mLiteralLevels ? lit.Index() : lit.Variable() - 1);
   return mLevelVars[index];
}

void
SUBool::UPLevels::AddLiteral(const Literal &lit)
{
   mLevelVars.emplace_back(mLevelBound);
   std::stringstream ss;
   for (unsigned j = 0; j < mLevelBound; ++j)
   {
      mVarStore.NameStream() << mNameBase << "_{" << lit << "}^{" << j << "}";
      mLevelVars.back()[j] = mVarStore.NewVariable();
   }
   /*
      ss << "c ! order";
      for (unsigned j = mLevelBound; j > 0; --j)
      {
         ss << ' ' << mLevelVars.back()[j - 1];
      }
      ss << " 0" << std::endl;
      std::cout << ss.str();
      */
}

void
SUBool::UPLevels::AddVariable()
{
   AddLiteral(Literal(++mMaxVariable, true));
   if (mLiteralLevels)
   {
      AddLiteral(Literal(mMaxVariable, false));
   }
}

unsigned
SUBool::UPLevels::EncodeGT(const Literal &left, const Literal &right,
    std::vector<Clause> *clauses) const
{
   // auto max_var = mVarStore.LastVariable();
   // auto clauses_size = clauses->size();
   return encode_bitvector_gt(LevelVars(left), LevelVars(right), mVarStore,
       clauses, mNameBase + "g", left.ToString(), right.ToString());
   // std::cout << "comparison size: " << mVarStore.LastVariable() - max_var <<
   // " "
   //<< clauses->size() - clauses_size << std::endl;
}

unsigned
SUBool::UPLevels::MaxVariable() const
{
   return mMaxVariable;
}

unsigned
SUBool::new_aux_comparison_var(const std::string &name_base,
    const std::string &left_desc, const std::string &right_desc, unsigned bit,
    VarStore &var_store)
{
   var_store.NameStream() << name_base << "_{" << left_desc << ", "
                          << right_desc << "}^{" << bit << "}";
   return var_store.NewVariable();
}

unsigned
SUBool::encode_bitvector_gt(const std::vector<unsigned> &left_bv,
    const std::vector<unsigned> &right_bv, VarStore &var_store,
    std::vector<Clause> *clauses, const std::string &name_base,
    const std::string &left_desc, const std::string &right_desc)
{
   if (clauses == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::AddLevelGT", "clauses == nullptr");
   }
   if (left_bv.size() == 0)
   {
      throw BadParameterException("[PCEncoder::AddLevelGT] left.size()==0");
   }
   unsigned b = left_bv.size() - 1;
   unsigned g =
       new_aux_comparison_var(name_base, left_desc, right_desc, b, var_store);
   clauses->emplace_back(Literal(g, false), Literal(left_bv[b], true));
   clauses->emplace_back(Literal(g, false), Literal(right_bv[b], false));
   /* XXX */
   // The opposite implication for G var in comparison.
   clauses->emplace_back(Literal(g, true), Literal(left_bv[b], false),
       Literal(right_bv[b], true));
   /* */
   while (b > 0)
   {
      --b;
      unsigned ng = new_aux_comparison_var(
          name_base, left_desc, right_desc, b, var_store);
      clauses->emplace_back(Literal(ng, false), Literal(left_bv[b], true),
          Literal(right_bv[b], false));
      clauses->emplace_back(
          Literal(ng, false), Literal(left_bv[b], true), Literal(g, true));
      clauses->emplace_back(
          Literal(ng, false), Literal(right_bv[b], false), Literal(g, true));
      /* XXX */
      // The opposite implications for G var in comparison.
      clauses->emplace_back(Literal(ng, true), Literal(left_bv[b], false),
          Literal(right_bv[b], true));
      clauses->emplace_back(
          Literal(ng, true), Literal(left_bv[b], false), Literal(g, false));
      clauses->emplace_back(
          Literal(ng, true), Literal(right_bv[b], true), Literal(g, false));
      /* */
      g = ng;
   }
   return g;
}

SUBool::NFSize
SUBool::UPLevels::CountInitialSize(
    const CNF &cnf, bool literal_levels, unsigned level_bound)
{
   return CountAddVariable(literal_levels, level_bound) * cnf.MaxVariable();
}

SUBool::NFSize
SUBool::UPLevels::CountAddVariable(bool literal_levels, unsigned level_bound)
{
   NFSize s{/*.max_variable = */ level_bound, /*.size = */ 0, /*.len = */ 0};
   return (literal_levels ? s * 2 : s);
}

SUBool::NFSize
SUBool::UPLevels::CountEncodeGTSize(
    bool /*literal_levels*/, unsigned level_bound)
{
   // std::cout << "est comparison size: " << level_bound << " "
   //<< 3 + 6 * (level_bound - 1) << std::endl;
   return {/*.max_variable = */ level_bound,
       /*.size = */ 3 + 6 * (level_bound - 1),
       /*.len = */ 7 + 18 * (level_bound - 1)};
}
