/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file up_log.cpp
 * Logarithmic size encoding of unit propagation constraint. Allows the
 * two implications separately or together. Splits CNF into a k-CNF
 * in the process.
 */

#include <cmath>

#include "cardinal.h"
#include "logstream.h"
#include "prgutil.h"
#include "simple_constraints.h"
#include "up_log.h"

SUBool::UPEncLogarithmic::UPEncLogarithmic(const Config &conf, const CNF &cnf,
    const ClausesToLiterals &cl2lit, VarStore &var_store)
    : UPEncoder(cnf, cl2lit, var_store), mConf(conf)
{
}

void
SUBool::UPEncLogarithmic::Encode(unsigned max_level,
    std::vector<unsigned> *inputs, std::vector<unsigned> *outputs,
    std::vector<Clause> *clauses) const
{
   if (inputs == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::Encode", "inputs == nullptr");
   }
   if (outputs == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::Encoder", "outputs == nullptr");
   }
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   unsigned bit_size =
       (mConf.use_bit_level ? max_level : BitSizeForLevel(max_level));
   mVarStore.InitVarArray(inputs, mCnf.MaxVariable(), true, "i_{", "}");
   mVarStore.InitVarArray(outputs, mCnf.MaxVariable(), true, "o_{", "}");
   // logs.TLog(3) << "[UPEncLogarithmic::Encode] After init vars "
   //<< mVarStore.LastVariable() << " " << clauses->size()
   //<< std::endl;
   UPLevels level_vars(mCnf, mVarStore, mConf.literal_levels, bit_size, "");
   // We cannot force level 0 on input variables which are set, because we need
   // to consider the derivation of the opposite literal as well.
   // InputsHaveZeroLevels(level_vars, *inputs, clauses);
   // logs.TLog(3) << "[UPEncLogarithmic::Encode] After level vars "
   //<< mVarStore.LastVariable() << " " << clauses->size()
   //<< std::endl;
   std::map<std::pair<unsigned, unsigned>, unsigned> comp;
   std::vector<std::vector<Literal>> fire_clauses(outputs->size());
   for (unsigned i = 0; i < mCnf.Size(); ++i)
   {
      AddClause(mCnf[i], i, level_vars, comp, (max_level == max_level_bound),
          fire_clauses, *outputs, clauses);
      /*logs.TLog(3) << "[UPEncLogarithmic::Encode] Clause added "
                   << mVarStore.LastVariable() << " " << clauses->size() << " "
                   << CNF(*clauses).Length() << std::endl;*/
   }
   /*logs.TLog(3) << "[UPEncLogarithmic::Encode] outputs size: "
                << outputs->size() << std::endl;*/
   AddFireClauses(fire_clauses, *inputs, *outputs, clauses);
   if (mConf.contradiction_has_max_level)
   {
      AddContradictionHasMaxLevel(*outputs, level_vars, clauses);
   }
   /*logs.TLog(3) << "[UPEncLogarithmic::Encode] comp.size=" << comp.size()
                << std::endl;*/
   /*logs.TLog(3) << "[UPEncLogarithmic::Encode] Fire clauses added "
                << mVarStore.LastVariable() << " " << clauses->size()
                << std::endl;*/
}

void
SUBool::UPEncLogarithmic::InputsHaveZeroLevels(const UPLevels &level_vars,
    const std::vector<unsigned> &inputs, std::vector<Clause> *clauses) const
{
   unsigned c = 0;
   for (unsigned i = 0; i < inputs.size(); ++i)
   {
      Literal in(inputs[i], false);
      const auto &lvars =
          (mConf.literal_levels ? level_vars.LevelVars(Literal(i))
                                : level_vars.LevelVars(i / 2));
      for (unsigned var : lvars)
      {
         ++c;
         clauses->emplace_back(in, Literal(var, false));
      }
   }
   // std::cout << "Added " << c << " clauses" << std::endl;
}

unsigned
SUBool::UPEncLogarithmic::MaxLevelBound() const
{
   return (mConf.use_bit_level ? BitSizeForLevel(0) : mCnf.MaxVariable());
}

unsigned
SUBool::UPEncLogarithmic::BitSizeForLevel(unsigned level) const
{
   if (level == 0)
   {
      level = mCnf.MaxVariable();
   }
   // unsigned max_value = (mConf.literal_levels ? 2 * level + 1 : level + 1);
   return static_cast<unsigned>(ceil(log2(static_cast<double>(level + 1))));
}

unsigned
SUBool::UPEncLogarithmic::NextLevelBound(unsigned level) const
{
   if (level == 0 || level >= MaxLevelBound())
   {
      return 0;
   }
   if (mConf.use_bit_level)
   {
      return level + 1;
   }
   // For simplicity, we compute the next level by subsequent increasing the
   // current level.
   // It can get slow for a high value of level.
   // We should give an explicit formula (close to doubling the level).
   unsigned bits = BitSizeForLevel(level++);
   while (BitSizeForLevel(level) == bits)
   {
      ++level;
   }
   return (level >= MaxLevelBound() ? 0 : level);
}

void
SUBool::UPEncLogarithmic::AddClause(const Clause &clause, unsigned index,
    UPLevels &level_vars,
    std::map<std::pair<unsigned, unsigned>, unsigned> &comp, bool is_max_level,
    std::vector<std::vector<Literal>> &fire_clauses,
    std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const
{
   if (clause.Size() == 0)
   {
      throw BadInputException("[PCEncoder::AddOPLogClause] empty clause");
   }
   if (clause.Size() == 1)
   {
      return;
   }
   if (clause.Size() <= mConf.split_size)
   {
      AddClauseSmall(clause, index, 0, level_vars, comp, is_max_level,
          fire_clauses, outputs, clauses);
      return;
   }
   AddClauseBig(clause, index, level_vars, comp, is_max_level, fire_clauses,
       outputs, clauses);
}

void
SUBool::UPEncLogarithmic::AddClauseBig(const Clause &clause, unsigned index,
    UPLevels &level_vars,
    std::map<std::pair<unsigned, unsigned>, unsigned> &comp, bool is_max_level,
    std::vector<std::vector<Literal>> &fire_clauses,
    std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const
{
   // std::cout << "big clause: " << clause << std::endl;
   unsigned link_var = 0;
   unsigned i = 0;
   unsigned subindex = 0;
   // std::cout<<"Splitting clause: "<<clause<<std::endl;
   std::vector<Literal> clause_literals;
   clause_literals.reserve(mConf.split_size);
   while (i < clause.Size())
   {
      clause_literals.push_back(clause[i++]);
      if (clause_literals.size() + 1 < mConf.split_size && i < clause.Size())
      {
         continue;
      }
      if (i + 1 == clause.Size())
      {
         clause_literals.push_back(clause[i++]);
      }
      else if (i + 1 < clause.Size())
      {
         link_var = NewLinkVar(level_vars, fire_clauses, outputs);
         clause_literals.emplace_back(link_var, true);
      }
      Clause subclause(clause_literals);
      AddClauseSmall(subclause, index, subindex, level_vars, comp, is_max_level,
          fire_clauses, outputs, clauses);
      // std::cout<<"Link var "<<link_var<<std::endl;
      // std::cout<<"Subclause "<<subindex<<": "<<subclause<<std::endl;
      ++subindex;
      clause_literals.clear();
      clause_literals.emplace_back(link_var, false);
   }
}

void
SUBool::UPEncLogarithmic::AddClauseSmall(const Clause &clause, unsigned index,
    unsigned subindex, const UPLevels &level_vars,
    std::map<std::pair<unsigned, unsigned>, unsigned> &comp, bool is_max_level,
    std::vector<std::vector<Literal>> &fire_clauses,
    const std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const
{
   if (clauses == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::AddClauseSmall", "clauses == nullptr");
   }
   for (const auto &lit : clause)
   {
      unsigned fire = NewFireVar(index, subindex, lit);
      fire_clauses[lit.Index()].emplace_back(fire, true);
      Literal fire_neg(fire, false);
      for (const auto &f : clause)
      {
         if (f == lit)
         {
            continue;
         }
         clauses->emplace_back(
             fire_neg, Literal(outputs[f.Negation().Index()], true));
         unsigned comp_var = Comparison(
             lit, f.Negation(), level_vars, comp, is_max_level, clauses);
         clauses->emplace_back(fire_neg, Literal(comp_var, true));
         /* XXX */
         // If a literal f is satisfied at the end of derivation,
         // then the clause cannot fire for a literal lit which
         // is different from f.
         // -- This really helps
         clauses->emplace_back(fire_neg, Literal(outputs[f.Index()], false));
         /* */
      }
   }
}

unsigned
SUBool::UPEncLogarithmic::NewFireVar(
    unsigned c_index, unsigned c_subindex, const Literal &lit) const
{
   mVarStore.NameStream() << "f_{" << c_index << ", " << lit << "}^{"
                          << c_subindex << "}";
   return mVarStore.NewVariable();
}

unsigned
SUBool::UPEncLogarithmic::Comparison(const Literal &left, const Literal &right,
    const UPLevels &level_vars,
    std::map<std::pair<unsigned, unsigned>, unsigned> &comp, bool is_max_level,
    std::vector<Clause> *clauses) const
{
   assert(left != right);
   /*if (left == right)
   {
      std::cout << "left==right (" << left << "==" << right << ")" << std::endl;
   }*/
   auto key = (mConf.literal_levels
                   ? std::make_pair(left.Index(), right.Index())
                   : std::make_pair(left.Variable(), right.Variable()));
   auto it = comp.find(key);
   if (it != comp.end())
   {
      return it->second;
   }
   unsigned out_var = level_vars.EncodeGT(left, right, clauses);
   comp[key] = out_var;
   /* either left<right or right>left but not both) */
   if (mConf.add_compare_exclusivity
       || (mConf.add_force_linear_order_on_max_level && is_max_level))
   {
      auto rev_it = comp.find(std::make_pair(key.second, key.first));
      if (rev_it != comp.end())
      {
         if (mConf.add_compare_exclusivity)
         {
            clauses->emplace_back(
                Literal(out_var, false), Literal(rev_it->second, false));
         }
         if (mConf.add_force_linear_order_on_max_level && is_max_level)
         {
            clauses->emplace_back(
                Literal(out_var, true), Literal(rev_it->second, true));
         }
      }
   }
   return out_var;
}

void
SUBool::UPEncLogarithmic::AddFireClauses(
    std::vector<std::vector<Literal>> &fire_clauses,
    const std::vector<unsigned> &inputs, const std::vector<unsigned> &outputs,
    std::vector<Clause> *clauses) const
{
   if (clauses == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::AddFireClauses", "clauses == nullptr");
   }
   auto amo_encoder = AmoEncoder::NewGlobalEncoder(mVarStore);
   if (mConf.add_amo_on_fire_vars)
   {
      for (auto fc : fire_clauses)
      {
         std::vector<Literal> fire_vars;
         for (const auto &fire : fc)
         {
            fire_vars.emplace_back(fire.Variable(), true);
         }
         amo_encoder->EncodeAppend(fire_vars, *clauses);
      }
   }
   for (unsigned i = 0; i < inputs.size(); ++i)
   {
      /* XXX */
      // Direct implication saying that a literal e is always derived,
      // if it is among the assumptions.
      clauses->emplace_back(
          Literal(inputs[i], false), Literal(outputs[i], true));
      if (!mCl2Lit.IsUnit(Literal(i)))
      {
         for (const auto &fire : fire_clauses[i])
         {
            clauses->emplace_back(fire.Negation(), Literal(outputs[i], true));
         }
         fire_clauses[i].emplace_back(inputs[i], true);
         fire_clauses[i].emplace_back(outputs[i], false);
         clauses->emplace_back(fire_clauses[i]);
      }
   }
   for (unsigned i = inputs.size(); i < fire_clauses.size(); ++i)
   {
      fire_clauses[i].emplace_back(outputs[i], false);
      clauses->emplace_back(fire_clauses[i]);
   }
}

void
SUBool::UPEncLogarithmic::AddContradictionHasMaxLevel(
    const std::vector<unsigned> &outputs, const UPLevels &level_vars,
    std::vector<Clause> *clauses) const
{
   if (clauses == nullptr)
   {
      throw NullPointerException(
          "UPEncLogarithmic::AddContradictionHasMaxLevel",
          "clauses == nullptr");
   }
   for (unsigned i = 0; i < outputs.size(); i += 2)
   {
      std::vector<Literal> out{
          Literal(outputs[i], true), Literal(outputs[i + 1], true)};
      std::vector<Literal> level_literals;
      level_literals.resize(level_vars.BitSize());
      if (mConf.literal_levels)
      {
         for (unsigned j = i; j < i + 2; ++j)
         {
            const auto &lvars = level_vars.LevelVars(Literal(j));
            std::transform(lvars.begin(), lvars.end(), level_literals.begin(),
                [](unsigned var) { return Literal(var, true); });
            encode_equivalence(out, level_literals, *clauses);
         }
      }
      else
      {
         const auto &lvars = level_vars.LevelVars(i / 2);
         std::transform(lvars.begin(), lvars.end(), level_literals.begin(),
             [](unsigned var) { return Literal(var, true); });
         encode_equivalence(out, level_literals, *clauses);
      }
   }
}

unsigned
SUBool::UPEncLogarithmic::NewLinkVar(UPLevels &level_vars,
    std::vector<std::vector<Literal>> &fire_clauses,
    std::vector<unsigned> &outputs) const
{
   for (unsigned i = 0; i < 2; ++i)
   {
      fire_clauses.emplace_back();
      mVarStore.NameStream()
          << "q_{" << Literal(fire_clauses.size() - 1) << "}";
      outputs.push_back(mVarStore.NewVariable());
   }
   level_vars.AddVariable();
   return level_vars.MaxVariable();
}

SUBool::NFSize
SUBool::UPEncLogarithmic::EstimateSize(unsigned max_level) const
{
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   unsigned bit_size =
       (mConf.use_bit_level ? max_level : BitSizeForLevel(max_level));
   NFSize s{/*.max_variable = */ 4 * mCnf.MaxVariable(), /*.size = */ 0,
       /*.len = */ 0};
   unsigned max_link_var{mCnf.MaxVariable()};
   // logs.TLog(3) << "[UPEncLogarithmic::EstimateSize] After init vars "
   //<< s.max_variable << " " << s.size << std::endl;
   s += UPLevels::CountInitialSize(mCnf, mConf.literal_levels, bit_size);
   // s += CountInputsHaveZeroLevels(bit_size);
   // logs.TLog(3) << "[UPEncLogarithmic::EstimateSize] After level vars "
   //<< s.max_variable << " " << s.size << std::endl;
   std::set<std::pair<unsigned, unsigned>> comp;
   std::vector<unsigned> fire_clauses(2 * mCnf.MaxVariable());
   for (unsigned i = 0; i < mCnf.Size(); ++i)
   {
      s += CountAddClause(mCnf[i], comp, bit_size, max_link_var, fire_clauses);
      // logs.TLog(3) << "[UPEncLogarithmic::EstimateSize] Clause added "
      //<< s.max_variable << " " << s.size << " " << s.len
      //<< std::endl;
   }
   s += CountComparison(comp, bit_size, max_level == max_level_bound);
   // logs.TLog(3) << "[UPEncLogarithmic::EstimateSize] comp.size=" <<
   // comp.size()
   //<< std::endl;
   // logs.TLog(3) << "[UPEncLogarithmic::EstimateSize] Comparisons "
   //<< s.max_variable << " " << s.size << " " << s.len << std::endl;
   s += CountAddFireClauses(fire_clauses);
   if (mConf.contradiction_has_max_level)
   {
      s += CountAddContradictionHasMaxLevel(max_link_var, bit_size);
   }
   // logs.TLog(3) << "[UPEncLogarithmic::EstimateSize] outputs size: "
   //<< max_link_var * 2 << std::endl;
   // logs.TLog(3) << "[UPEncLogarithmic::EstimateSize] Fire clauses added "
   //<< s.max_variable << " " << s.size << std::endl;
   return s;
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountComparison(
    const std::set<std::pair<unsigned, unsigned>> &comp, unsigned bit_size,
    bool is_max_level) const
{
   auto s = UPLevels::CountEncodeGTSize(mConf.literal_levels, bit_size)
            * comp.size();
   if (mConf.add_compare_exclusivity
       || (mConf.add_force_linear_order_on_max_level && is_max_level))
   {
      for (const auto &p : comp)
      {
         if (p.second < p.first
             && comp.find(std::make_pair(p.second, p.first)) != comp.end())
         {
            if (mConf.add_compare_exclusivity)
            {
               ++s.size;
               s.len += 2;
            }
            if (mConf.add_force_linear_order_on_max_level && is_max_level)
            {
               ++s.size;
               s.len += 2;
            }
         }
      }
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountAddClause(const Clause &clause,
    std::set<std::pair<unsigned, unsigned>> &comp, unsigned max_level,
    unsigned &max_link_var, std::vector<unsigned> &fire_clauses) const
{
   if (clause.Size() == 0)
   {
      throw BadInputException("[PCEncoder::AddOPLogClause] empty clause");
   }
   if (clause.Size() == 1)
   {
      return {0, 0, 0};
   }
   if (clause.Size() <= mConf.split_size)
   {
      return CountAddClauseSmall(clause, comp, fire_clauses);
   }
   return CountAddClauseBig(
       clause, comp, max_level, max_link_var, fire_clauses);
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountAddClauseSmall(const Clause &clause,
    std::set<std::pair<unsigned, unsigned>> &comp,
    std::vector<unsigned> &fire_clauses) const
{
   // std::cout << "small clause: " << clause << std::endl;
   NFSize s{0, 0, 0};
   for (const auto &lit : clause)
   {
      ++s.max_variable;
      //++s.size;   // CountAddFireClauses
      // s.len += 2; // CountAddFireClauses
      ++fire_clauses[lit.Index()];
      // ++s.len;
      for (const auto &f : clause)
      {
         if (f != lit)
         {
            s.size += 3; //+ 1; // +1 is for fire clause (not counted in
                         // CountAddFireClauses
            s.len += 6;  // + 2;  // +2 is for fire clause (not counted in
                         // CountAddFireClauses
            if (mConf.literal_levels)
            {
               comp.insert(std::make_pair(lit.Index(), f.Negation().Index()));
            }
            else
            {
               comp.insert(std::make_pair(lit.Variable(), f.Variable()));
            }
         }
      }
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountAddClauseBig(const Clause &clause,
    std::set<std::pair<unsigned, unsigned>> &comp, unsigned max_level,
    unsigned &max_link_var, std::vector<unsigned> &fire_clauses) const
{
   NFSize s{0, 0, 0};

   unsigned i = 0;
   // std::cout<<"Splitting clause: "<<clause<<std::endl;
   std::vector<Literal> clause_literals;
   clause_literals.reserve(mConf.split_size);
   while (i < clause.Size())
   {
      clause_literals.push_back(clause[i++]);
      if (clause_literals.size() + 1 < mConf.split_size && i < clause.Size())
      {
         continue;
      }
      if (i + 1 == clause.Size())
      {
         clause_literals.push_back(clause[i++]);
      }
      else if (i + 1 < clause.Size())
      {
         s += CountNewLinkVar(max_level);
         fire_clauses.emplace_back();
         fire_clauses.emplace_back();
         clause_literals.emplace_back(++max_link_var, true);
      }
      Clause subclause(clause_literals);
      s += CountAddClauseSmall(subclause, comp, fire_clauses);
      clause_literals.clear();
      clause_literals.emplace_back(max_link_var, false);
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountNewLinkVar(unsigned max_level) const
{
   NFSize s{/*.max_variable = */ 2, /*.size = */ 0 /* 1 -- fire clause */,
       /*.len = */ 0};
   return s + UPLevels::CountAddVariable(mConf.literal_levels, max_level);
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountAddFireClauses(
    std::vector<unsigned> &fire_clauses) const
{
   NFSize s{/*.max_variable = */ 0,
       /*.size = */ 2 * mCnf.MaxVariable(),
       /*.len = */ 4 * mCnf.MaxVariable()};
   auto amo_encoder = AmoEncoder::NewGlobalEncoder(mVarStore);
   if (mConf.add_amo_on_fire_vars)
   {
      for (auto fc : fire_clauses)
      {
         s += amo_encoder->CountClauses(fc);
      }
   }
   for (unsigned i = 0; i < 2 * mCnf.MaxVariable(); ++i)
   {
      if (!mCl2Lit.IsUnit(Literal(i)))
      {
         s.size += 1 + fire_clauses[i];
         s.len += 2 + 3 * fire_clauses[i];
      }
   }
   for (unsigned i = 2 * mCnf.MaxVariable(); i < fire_clauses.size(); ++i)
   {
      ++s.size;
      s.len += 1 + fire_clauses[i];
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountAddContradictionHasMaxLevel(
    unsigned max_link_var, unsigned bit_size) const
{
   NFSize s{/* .max_variable = */ 0,
       /* .size = */ max_link_var * (bit_size + 2),
       /* .len = */ max_link_var * (3 * bit_size + bit_size * 2 + 2)};
   if (mConf.literal_levels)
   {
      s += s;
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncLogarithmic::CountInputsHaveZeroLevels(unsigned max_level) const
{
   // std::cout << "Estimating " << m * mCnf.MaxVariable() << " clauses"
   //<< std::endl;
   return NFSize{/* .max_variable = */ 0,
       /* .size = */ 2 * max_level * mCnf.MaxVariable(),
       /* .len = */ 4 * max_level * mCnf.MaxVariable()};
}

void
SUBool::UPEncLogarithmic::Config::Dump(
    const std::string &line_prefix, unsigned level) const
{
   logs.DumpValue(line_prefix, level, "literal_levels", literal_levels);
   logs.DumpValue(line_prefix, level, "split_size", split_size);
   logs.DumpValue(line_prefix, level, "use_bit_level", use_bit_level);
   logs.DumpValue(
       line_prefix, level, "add_amo_on_fire_vars (amf)", add_amo_on_fire_vars);
   logs.DumpValue(line_prefix, level, "add_compare_exclusivity (cex)",
       add_compare_exclusivity);
   logs.DumpValue(line_prefix, level,
       "add_force_linear_order_on_max_level (flm)",
       add_force_linear_order_on_max_level);
   logs.DumpValue(line_prefix, level, "contradiction_has_max_level (cml)",
       contradiction_has_max_level);
}
