/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file up_log.h
 * Logarithmic factor size encoding of unit propagation constraint. Splits CNF
 * into a k-CNF in the process.
 */

#ifndef __UP_LOG_H
#define __UP_LOG_H

#include <map>

#include "cardinal.h"
#include "up_levels.h"
#include "upenc.h"

namespace SUBool
{

   class UPEncLogarithmic : public UPEncoder
   {
    public:
      struct Config
      {
         bool literal_levels{false};
         unsigned split_size{4};
         bool use_bit_level{true};
         bool add_amo_on_fire_vars{false};
         bool add_compare_exclusivity{true};
         bool add_force_linear_order_on_max_level{true};
         // A variable is contradictory if and only if its level has all ones.
         bool contradiction_has_max_level{true};

         Config() {}

         Config(bool ll, unsigned ss, bool ubl, bool amo_fire, bool comp_ex,
             bool force_lin, bool cml)
             : literal_levels(ll), split_size(ss), use_bit_level(ubl),
               add_amo_on_fire_vars(amo_fire), add_compare_exclusivity(comp_ex),
               add_force_linear_order_on_max_level(force_lin),
               contradiction_has_max_level(cml)
         {
         }

         void Dump(const std::string &line_prefix, unsigned level) const;
      };

    private:
      Config mConf;

      void InputsHaveZeroLevels(const UPLevels &level_vars,
          const std::vector<unsigned> &inputs,
          std::vector<Clause> *clauses) const;
      void AddClause(const Clause &clause, unsigned index, UPLevels &level_vars,
          std::map<std::pair<unsigned, unsigned>, unsigned> &comp,
          bool is_max_level, std::vector<std::vector<Literal>> &fire_clauses,
          std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const;
      void AddClauseSmall(const Clause &clause, unsigned index,
          unsigned subindex, const UPLevels &level_vars,
          std::map<std::pair<unsigned, unsigned>, unsigned> &comp,
          bool is_max_level, std::vector<std::vector<Literal>> &fire_clauses,
          const std::vector<unsigned> &outputs,
          std::vector<Clause> *clauses) const;
      void AddClauseBig(const Clause &clause, unsigned index,
          UPLevels &level_vars,
          std::map<std::pair<unsigned, unsigned>, unsigned> &comp,
          bool is_max_level, std::vector<std::vector<Literal>> &fire_clauses,
          std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const;
      void AddFireClauses(std::vector<std::vector<Literal>> &fire_clauses,
          const std::vector<unsigned> &inputs,
          const std::vector<unsigned> &outputs,
          std::vector<Clause> *clauses) const;
      void AddContradictionHasMaxLevel(const std::vector<unsigned> &outputs,
          const UPLevels &level_vars, std::vector<Clause> *clauses) const;
      unsigned NewLinkVar(UPLevels &level_vars,
          std::vector<std::vector<Literal>> &fire_clauses,
          std::vector<unsigned> &outputs) const;
      unsigned NewFireVar(
          unsigned c_index, unsigned c_subindex, const Literal &lit) const;
      unsigned Comparison(const Literal &left, const Literal &right,
          const UPLevels &level_vars,
          std::map<std::pair<unsigned, unsigned>, unsigned> &comp,
          bool is_max_level, std::vector<Clause> *clauses) const;

      NFSize CountInputsHaveZeroLevels(unsigned max_level) const;
      NFSize CountAddClause(const Clause &clause,
          std::set<std::pair<unsigned, unsigned>> &comp, unsigned max_level,
          unsigned &max_link_var, std::vector<unsigned> &fire_clauses) const;
      NFSize CountAddClauseSmall(const Clause &clause,
          std::set<std::pair<unsigned, unsigned>> &comp,
          std::vector<unsigned> &fire_clauses) const;
      NFSize CountAddClauseBig(const Clause &clause,
          std::set<std::pair<unsigned, unsigned>> &comp, unsigned max_level,
          unsigned &max_link_var, std::vector<unsigned> &fire_clauses) const;
      NFSize CountAddFireClauses(std::vector<unsigned> &fire_clauses) const;
      NFSize CountAddContradictionHasMaxLevel(
          unsigned max_link_var, unsigned bit_size) const;
      NFSize CountNewLinkVar(unsigned max_level) const;
      NFSize CountComparison(
          const std::set<std::pair<unsigned, unsigned>> &comp,
          unsigned bit_size, bool is_max_level) const;

      unsigned BitSizeForLevel(unsigned level) const;

    public:
      UPEncLogarithmic(const Config &conf, const CNF &cnf,
          const ClausesToLiterals &cl2lit, VarStore &var_store);
      virtual void Encode(unsigned max_level, std::vector<unsigned> *inputs,
          std::vector<unsigned> *outputs,
          std::vector<Clause> *clauses) const override;
      virtual NFSize EstimateSize(unsigned max_level) const override;
      virtual unsigned MaxLevelBound() const override;
      virtual unsigned NextLevelBound(unsigned level) const override;
   };
} // namespace SUBool

#endif
