/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file up_max.cpp
 * Encoding of unit propagation constraint based on selecting maximum level of a
 * literal.
 */

#include <cmath>

#include "cardinal.h"
#include "up_max.h"

SUBool::UPEncMaxLevel::UPEncMaxLevel(const CNF &cnf,
                                     const ClausesToLiterals &cl2lit,
                                     VarStore &var_store)
    : UPEncoder(cnf, cl2lit, var_store)
{
}

void
SUBool::UPEncMaxLevel::Encode(unsigned max_level, std::vector<unsigned> *inputs,
                              std::vector<unsigned> *outputs,
                              std::vector<Clause> *clauses) const
{
   if (inputs == nullptr)
   {
      throw NullPointerException("UPEncLogarithmic::Encode",
                                 "inputs == nullptr");
   }
   if (outputs == nullptr)
   {
      throw NullPointerException("UPEncLogarithmic::Encoder",
                                 "outputs == nullptr");
   }
   if (clauses == nullptr)
   {
      throw NullPointerException("UPEncMaxLevel::Encode", "clauses == nullptr");
   }
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   mVarStore.InitVarArray(inputs, mCnf.MaxVariable(), true, "i_{", "}");
   mVarStore.InitVarArray(outputs, mCnf.MaxVariable(), true, "o_{", "}");
   UPLevels level_vars(mCnf, mVarStore, false, max_level, "");
   std::vector<std::vector<Literal>> fire_clauses(outputs->size());
   for (unsigned i = 0; i < mCnf.Size(); ++i)
   {
      AddClause(mCnf[i], i, level_vars, fire_clauses, *inputs, *outputs,
                clauses);
   }
   AddFireClauses(fire_clauses, *inputs, *outputs, clauses);
}

unsigned
SUBool::UPEncMaxLevel::MaxLevelBound() const
{
   return static_cast<unsigned>(
       ceil(log2(static_cast<double>(mCnf.MaxVariable() + 1))));
}

unsigned
SUBool::UPEncMaxLevel::NextLevelBound(unsigned level) const
{
   return (level > 0 && level < MaxLevelBound() ? level + 1 : 0);
}

void
SUBool::UPEncMaxLevel::AddClause(
    const Clause &clause, unsigned index, const UPLevels &level_vars,
    std::vector<std::vector<Literal>> &fire_clauses,
    const std::vector<unsigned> &inputs, const std::vector<unsigned> &outputs,
    std::vector<Clause> *clauses) const
{
   switch (clause.Size())
   {
   case 0:
      throw BadInputException("[UPEncMaxLevel::AddClause] empty clause");
   case 1:
      AddUnitClause(clause, index, level_vars, fire_clauses, inputs, outputs,
                    clauses);
      break;
   case 2:
      AddBinaryClause(clause, index, level_vars, fire_clauses, inputs, outputs,
                      clauses);
      break;
   default:
      AddBigClause(clause, index, level_vars, fire_clauses, inputs, outputs,
                   clauses);
   }
}

void
SUBool::UPEncMaxLevel::AddUnitClause(
    const Clause &clause, unsigned /*index*/, const UPLevels &level_vars,
    std::vector<std::vector<Literal>> & /*fire_clauses*/,
    const std::vector<unsigned> &inputs, const std::vector<unsigned> &outputs,
    std::vector<Clause> *clauses) const
{
   // In case of unit clause we can always assume that the literal is
   // in the assumptions.
   clauses->emplace_back(Literal(inputs[clause[0].Index()], true));
   clauses->emplace_back(Literal(outputs[clause[0].Index()], true));
   // Level is set to 0.
   for (const auto &i : level_vars.LevelVars(clause[0]))
   {
      clauses->emplace_back(Literal(i, false));
   }
}

void
SUBool::UPEncMaxLevel::AddBinaryClause(
    const Clause &clause, unsigned index, const UPLevels &level_vars,
    std::vector<std::vector<Literal>> &fire_clauses,
    const std::vector<unsigned> & /*inputs*/,
    const std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const
{
   unsigned f0 = NewFireVar(index, clause[0]);
   unsigned f1 = NewFireVar(index, clause[1]);
   unsigned l0 = clause[0].Index();
   unsigned l1 = clause[1].Index();
   fire_clauses[l0].emplace_back(f0, true);
   fire_clauses[l1].emplace_back(f1, true);
   // AMO on fire literals
   clauses->emplace_back(Literal(f0, false), Literal(f1, false));
   // clause max level (D)
   std::vector<unsigned> d;
   InitClauseVars(level_vars.LevelBound(), &d, index);
   // f0=>D=B0
   unsigned v0 = clause[0].Variable();
   unsigned v1 = clause[1].Variable();
   for (unsigned i = 0; i < d.size(); ++i)
   {
      clauses->emplace_back(Literal(f0, false), Literal(d[i], false),
                            Literal(level_vars.LevelVars(clause[1])[i], true));
      clauses->emplace_back(Literal(f0, false), Literal(d[i], true),
                            Literal(level_vars.LevelVars(clause[1])[i], false));
      clauses->emplace_back(Literal(f1, false), Literal(d[i], false),
                            Literal(level_vars.LevelVars(clause[0])[i], true));
      clauses->emplace_back(Literal(f1, false), Literal(d[i], true),
                            Literal(level_vars.LevelVars(clause[0])[i], false));
   }
   std::stringstream b0_desc;
   b0_desc << "B" << v0;
   std::stringstream b1_desc;
   b1_desc << "B" << v1;
   std::stringstream d_desc;
   d_desc << "D" << index;
   unsigned c0 =
       encode_bitvector_gt(level_vars.LevelVars(clause[0]), d, mVarStore,
                           clauses, "g", b0_desc.str(), d_desc.str());
   // unsigned c0=AddLevelGT(level_vars[v0-1], d, b0_desc.str(), d_desc.str());
   unsigned c1 =
       encode_bitvector_gt(level_vars.LevelVars(clause[1]), d, mVarStore,
                           clauses, "g", b1_desc.str(), d_desc.str());
   // unsigned c1=AddLevelGT(level_vars[v1-1], d, b1_desc.str(), d_desc.str());
   clauses->emplace_back(Literal(f0, false), Literal(c0, true));
   clauses->emplace_back(Literal(f0, false),
                         Literal(outputs[clause[1].Negation().Index()], true));
   clauses->emplace_back(Literal(f1, false), Literal(c1, true));
   clauses->emplace_back(Literal(f1, false),
                         Literal(outputs[clause[0].Negation().Index()], true));
}
void
SUBool::UPEncMaxLevel::AddBigClause(
    const Clause &clause, unsigned index, const UPLevels &level_vars,
    std::vector<std::vector<Literal>> &fire_clauses,
    const std::vector<unsigned> & /*inputs*/,
    const std::vector<unsigned> &outputs, std::vector<Clause> *clauses) const
{
   std::vector<Literal> fire_literals;
   for (const auto &lit : clause)
   {
      fire_literals.emplace_back(NewFireVar(index, lit), true);
      fire_clauses[lit.Index()].emplace_back(fire_literals.back());
   }
   amo_representation(fire_literals, *clauses);
   // Initialize R1
   std::vector<unsigned> r;
   const std::vector<unsigned> &lv0 = level_vars.LevelVars(clause[0]);
   InitClauseIntermedVars(level_vars.LevelBound(), &r, index,
                          clause[0].Variable());
   for (unsigned i = 0; i < r.size(); ++i)
   {
      clauses->emplace_back(Literal(r[i], false), fire_literals[0].Negation());
      clauses->emplace_back(Literal(r[i], false), Literal(lv0[i], true));
      clauses->emplace_back(Literal(r[i], true), fire_literals[0],
                            Literal(lv0[i], false));
   }

   for (unsigned i = 1; i < clause.Size(); ++i)
   {
      std::vector<unsigned> old_r = std::move(r);
      if (i + 1 < clause.Size())
      {
         InitClauseIntermedVars(level_vars.LevelBound(), &r, index,
                                clause[i].Variable());
      }
      else
      {
         InitClauseVars(level_vars.LevelBound(), &r, index);
      }
      const std::vector<unsigned> &lv = level_vars.LevelVars(clause[i]);
      std::stringstream lv_desc;
      lv_desc << "B" << clause[i].Variable();
      std::stringstream old_r_desc;
      old_r_desc << "R" << clause[i - 1].Variable();
      unsigned c = encode_bitvector_gt(lv, old_r, mVarStore, clauses, "g",
                                       lv_desc.str(), old_r_desc.str());
      for (unsigned j = 0; j < r.size(); ++j)
      {
         clauses->emplace_back(fire_literals[i].Negation(),
                               Literal(old_r[j], false), Literal(r[j], true));
         clauses->emplace_back(fire_literals[i].Negation(),
                               Literal(old_r[j], true), Literal(r[j], false));
         clauses->emplace_back(Literal(c, true), Literal(old_r[j], false),
                               Literal(r[j], true));
         clauses->emplace_back(Literal(c, true), Literal(old_r[j], true),
                               Literal(r[j], false));
         std::vector<Literal> rb;
         rb.push_back(fire_literals[i]);
         rb.emplace_back(c, false);
         rb.emplace_back(r[j], true);
         rb.emplace_back(lv[j], false);
         clauses->emplace_back(rb);
         rb[2].Negate();
         rb[3].Negate();
         clauses->emplace_back(rb);
      }
   }
   // Compare D with the fired level
   std::stringstream d_desc;
   d_desc << "D" << index;
   for (unsigned i = 0; i < clause.Size(); ++i)
   {
      std::stringstream b_desc;
      unsigned v = clause[i].Variable();
      b_desc << "B" << v;
      unsigned c =
          encode_bitvector_gt(level_vars.LevelVars(clause[i]), r, mVarStore,
                              clauses, "g", b_desc.str(), d_desc.str());
      clauses->emplace_back(fire_literals[i].Negation(), Literal(c, true));
      for (unsigned j = 0; j < clause.Size(); ++j)
      {
         if (j == i)
         {
            continue;
         }
         clauses->emplace_back(
             fire_literals[i].Negation(),
             Literal(outputs[clause[j].Negation().Index()], true));
      }
   }
}

void
SUBool::UPEncMaxLevel::AddFireClauses(
    std::vector<std::vector<Literal>> &fire_clauses,
    const std::vector<unsigned> &inputs, const std::vector<unsigned> &outputs,
    std::vector<Clause> *clauses) const
{
   for (unsigned i = 0; i < inputs.size(); ++i)
   {
      // Unit clauses are always in the assumptions and in the
      // conclusions, they are treated separately.
      if (!mCl2Lit.IsUnit(Literal(i)))
      {
         // Direct implication saying that a literal e is always derived,
         // if it is among the assumptions.
         clauses->emplace_back(Literal(inputs[i], false),
                               Literal(outputs[i], true));
         for (const auto &fire : fire_clauses[i])
         {
            clauses->emplace_back(fire.Negation(), Literal(outputs[i], true));
         }
         fire_clauses[i].emplace_back(inputs[i], true);
         fire_clauses[i].emplace_back(outputs[i], false);
         clauses->emplace_back(fire_clauses[i]);
      }
   }
}

void
SUBool::UPEncMaxLevel::InitClauseVars(unsigned max_level,
                                      std::vector<unsigned> *vars,
                                      unsigned c_index) const
{
   vars->resize(max_level);
   for (unsigned j = 0; j < max_level; ++j)
   {
      mVarStore.NameStream() << "d_{" << c_index << "}^{" << j << "}";
      (*vars)[j] = mVarStore.NewVariable();
   }
}

void
SUBool::UPEncMaxLevel::InitClauseIntermedVars(unsigned max_level,
                                              std::vector<unsigned> *vars,
                                              unsigned c_index,
                                              unsigned var) const
{
   vars->resize(max_level);
   for (unsigned j = 0; j < max_level; ++j)
   {
      mVarStore.NameStream()
          << "r_{" << c_index << ", " << var << "}^{" << j << "}";
      (*vars)[j] = mVarStore.NewVariable();
   }
}

unsigned
SUBool::UPEncMaxLevel::NewFireVar(unsigned c_index, const Literal &lit) const
{
   mVarStore.NameStream() << "f_{" << c_index << ", " << lit << "}";
   return mVarStore.NewVariable();
}

SUBool::NFSize
SUBool::UPEncMaxLevel::EstimateSize(unsigned /*max_level*/) const
{
   throw NotImplementedException("UPEncMaxLevel::EstimateSize",
                                 "Not implemented");
}

