/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file up_min_lq.cpp
 * An encoding of unit propagation constraint. It chooses either quadratic or
 * logarithmic encoding based on their sizes.
 */

#include "up_min_lq.h"
#include "logstream.h"

SUBool::UPEncMinLQ::UPEncMinLQ(const CNF &cnf, const ClausesToLiterals &cl2lit,
                               VarStore &var_store,
                               const UPEncLogarithmic::Config &log_conf,
                               SizeWeights log_weights,
                               SizeWeights quad_weights)
    : UPEncoder(cnf, cl2lit, var_store), mLogWeights(log_weights),
      mQuadWeights(quad_weights),
      mEncQuad(UPEncQuad::IF, cnf, mCl2Lit, mVarStore),
      mEncLog(
          UPEncLogarithmic::Config(log_conf.literal_levels, log_conf.split_size,
                                   false, log_conf.add_amo_on_fire_vars,
                                   log_conf.add_compare_exclusivity,
                                   log_conf.add_force_linear_order_on_max_level,
                                   log_conf.contradiction_has_max_level),
          cnf, cl2lit, var_store)
{
}

void
SUBool::UPEncMinLQ::Encode(unsigned max_level, std::vector<unsigned> *inputs,
                           std::vector<unsigned> *outputs,
                           std::vector<Clause> *clauses) const
{
   if (PickQuad(max_level))
   {
      mEncQuad.Encode(max_level, inputs, outputs, clauses);
   }
   else
   {
      mEncLog.Encode(max_level, inputs, outputs, clauses);
   }
}

SUBool::NFSize
SUBool::UPEncMinLQ::EstimateSize(unsigned max_level) const
{
   NFSize size;
   PickQuad(max_level, &size);
   return size;
}

unsigned
SUBool::UPEncMinLQ::MaxLevelBound() const
{
   return mCnf.MaxVariable();
}

unsigned
SUBool::UPEncMinLQ::NextLevelBound(unsigned level) const
{
   return (PickQuad(level) ? mEncQuad.NextLevelBound(level)
                           : mEncLog.NextLevelBound(level));
}

bool
SUBool::UPEncMinLQ::PickQuad(unsigned level, NFSize *size) const
{
   auto log_weight = WeightedSize(mEncLog, level, mLogWeights);
   logs.TLog(4) << "[UPEncMinLQ::PickQuad] L: max_variable="
                << log_weight.first.max_variable
                << ", size=" << log_weight.first.size
                << ", len=" << log_weight.first.len
                << ", weight=" << log_weight.second << std::endl;
   auto quad_weight = WeightedSize(mEncQuad, level, mQuadWeights);
   logs.TLog(4) << "[UPEncMinLQ::PickQuad] Q: max_variable="
                << quad_weight.first.max_variable
                << ", size=" << quad_weight.first.size
                << ", len=" << quad_weight.first.len
                << ", weight=" << quad_weight.second << std::endl;
   const bool pick_quad = (quad_weight.second <= log_weight.second);
   logs.TLog(4) << "[UPEncMinLQ::PickQuad] Level " << level << ", picking "
                << (pick_quad ? "quadratic" : "logarithmic") << std::endl;
   if (size != nullptr)
   {
      *size = (pick_quad ? quad_weight.first : log_weight.first);
   }
   return pick_quad;
}

std::pair<SUBool::NFSize, double>
SUBool::UPEncMinLQ::WeightedSize(const UPEncoder &enc, unsigned level,
                                 const SizeWeights &weights) const
{
   std::pair<SUBool::NFSize, double> ws;
   ws.first = enc.EstimateSize(level);
   ws.second = weights.max_variable * static_cast<double>(ws.first.max_variable)
               + weights.size * static_cast<double>(ws.first.size)
               + weights.len * static_cast<double>(ws.first.len);
   return ws;
}
