/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file up_quad.cpp
 * Quadratic size encoding of unit propagation constraint. Allows the
 * two implications separately or together.
 */

#include "up_quad.h"

SUBool::UPEncQuad::UPEncQuad(DIRECTION direction, const CNF &cnf,
    const ClausesToLiterals &cl2lit, VarStore &var_store)
    : UPEncoder(cnf, cl2lit, var_store), mDirection(direction)
{
}

void
SUBool::UPEncQuad::InitVars(
    std::vector<std::vector<unsigned>> *vars, unsigned max_level) const
{
   vars->resize(max_level + 1);
   for (unsigned i = 0; i < vars->size(); ++i)
   {
      (*vars)[i].resize(2 * mCnf.MaxVariable());
      for (unsigned j = 0; j < 2 * mCnf.MaxVariable(); ++j)
      {
         mVarStore.NameStream() << "_{" << Literal(j) << "}^{" << i << "}";
         (*vars)[i][j] = mVarStore.NewVariable();
      }
   }
}

void
SUBool::UPEncQuad::Encode(unsigned max_level, std::vector<unsigned> *inputs,
    std::vector<unsigned> *outputs, std::vector<Clause> *clauses) const
{
   if (inputs == nullptr)
   {
      throw NullPointerException("UPEncQuad::Encode", "inputs == nullptr");
   }
   if (outputs == nullptr)
   {
      throw NullPointerException("UPEncQuad::Encode", "outputs == nullptr");
   }

   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   std::vector<std::vector<unsigned>> vars;
   InitVars(&vars, max_level);
   AddClauses(vars, clauses);
   *inputs = std::move(vars.front());
   *outputs = std::move(vars.back());
}

unsigned
SUBool::UPEncQuad::MaxLevelBound() const
{
   return mCnf.MaxVariable();
}

unsigned
SUBool::UPEncQuad::NextLevelBound(unsigned level) const
{
   return (level > 0 && level < MaxLevelBound() ? level + 1 : 0);
}

void
SUBool::UPEncQuad::AddClauses(const std::vector<std::vector<unsigned>> &vars,
    std::vector<Clause> *clauses) const
{
   for (unsigned i = 1; i < vars.size(); ++i)
   {
      AddClausesAtLevel(vars[i - 1], vars[i], i, clauses);
   }
}

void
SUBool::UPEncQuad::AddClausesAtLevel(const std::vector<unsigned> &prev_vars,
    const std::vector<unsigned> &vars, unsigned level,
    std::vector<Clause> *clauses) const
{
   for (unsigned i = 1; i <= mCnf.MaxVariable(); ++i)
   {
      Literal pos = Literal(i, true);
      Literal neg = Literal(i, false);
      AddClausesForLiteralAtLevel(
          pos, vars[pos.Index()], prev_vars, level, clauses);
      AddClausesForLiteralAtLevel(
          neg, vars[neg.Index()], prev_vars, level, clauses);
   }
}

void
SUBool::UPEncQuad::AddClausesForLiteralAtLevel(const Literal &lit,
    const unsigned var_out, const std::vector<unsigned> &prev_vars,
    unsigned level, std::vector<Clause> *clauses) const
{
   switch (mDirection)
   {
   case IF:
      AddClausesForLiteralAtLevelIF(lit, var_out, prev_vars, level, clauses);
      break;
   case ONLY_IF:
      AddClausesForLiteralAtLevelONLYIF(
          lit, var_out, prev_vars, level, clauses);
      break;
   case BOTH:
      AddClausesForLiteralAtLevelIF(lit, var_out, prev_vars, level, clauses);
      AddClausesForLiteralAtLevelONLYIF(
          lit, var_out, prev_vars, level, clauses);
      break;
   default:
      throw BadParameterException(
          "UPEncQuad::AddClausesForLiteralAtLevel", "Bad direction");
   }
}

void
SUBool::UPEncQuad::AddClausesForLiteralAtLevelIF(const Literal &lit,
    const unsigned var_out, const std::vector<unsigned> &prev_vars,
    unsigned level, std::vector<Clause> *clauses) const
{
   if (clauses == nullptr)
   {
      throw NullPointerException(
          "UPEncQuad::AddClausesForLiteralAtLevelIF", "clauses == nullptr");
   }
   const unsigned var_in = prev_vars[lit.Index()];
   std::vector<Literal> a_clause_lit;

   for (const auto &clause_index : mCl2Lit.Clauses(lit))
   {
      const Clause &c = mCnf[clause_index];
      /*if(c.Size()==1)
      {
         // Should be optimized -- in this case no a_clause is
         // necessary and no equivalence is needed.
         // Unit clauses should be considered separately.
         // Actually they even should not be here, they should be
         // propagated before the encoding is even invoked.
         continue;
      }*/
      if (c.Size() == 2)
      {
         const Literal &a = c[0];
         const Literal &b = c[1];
         if (b == lit)
         {
            // prev_vars[a.Negation().Index()];
            a_clause_lit.emplace_back(prev_vars[a.Negation().Index()], true);
         }
         else
         {
            // prev_vars[b.Negation().Index()];
            a_clause_lit.emplace_back(prev_vars[b.Negation().Index()], true);
         }
      }
      else
      {
         mVarStore.NameStream()
             << "z_{" << lit << ", " << clause_index << "}^{" << level << "}";
         const unsigned z = mVarStore.NewVariable();
         a_clause_lit.emplace_back(z, true);
         Literal neg_z(z, false);
         std::vector<Literal> b_clause_lit;
         for (const auto &f : c)
         {
            if (f != lit)
            {
               const unsigned v = prev_vars[f.Negation().Index()];
               b_clause_lit.emplace_back(v, false);
               clauses->emplace_back(neg_z, Literal(v, true));
            }
         }
         b_clause_lit.emplace_back(z, true);
         clauses->emplace_back(b_clause_lit);
      }
   }
   a_clause_lit.emplace_back(var_in, true);
   a_clause_lit.emplace_back(var_out, false);
   clauses->emplace_back(a_clause_lit);
}

void
SUBool::UPEncQuad::AddClausesForLiteralAtLevelONLYIF(const Literal &lit,
    const unsigned var_out, const std::vector<unsigned> &prev_vars,
    unsigned /*level*/, std::vector<Clause> *clauses) const
{
   if (clauses == nullptr)
   {
      throw NullPointerException(
          "UPEncQuad::AddClausesForLiteralAtLevelONLYIF", "clauses == nullptr");
   }
   const unsigned var_in = prev_vars[lit.Index()];
   clauses->emplace_back(Literal(var_in, false), Literal(var_out, true));
   for (const auto &clause_index : mCl2Lit.Clauses(lit))
   {
      std::vector<Literal> clause_lit;
      for (const auto &f : mCnf[clause_index])
      {
         if (f != lit)
         {
            clause_lit.emplace_back(prev_vars[f.Negation().Index()], false);
         }
      }
      clause_lit.emplace_back(var_out, true);
      clauses->emplace_back(clause_lit);
   }
}

SUBool::NFSize
SUBool::UPEncQuad::CountInitVars(unsigned max_level) const
{
   return {(max_level + 1) * 2 * mCnf.MaxVariable(), 0, 0};
}

SUBool::NFSize
SUBool::UPEncQuad::CountAddClauses(unsigned max_level, NFSize s) const
{
   s += CountAddClausesAtLevel() * max_level;
   return s;
}

SUBool::NFSize
SUBool::UPEncQuad::CountAddClausesAtLevel() const
{
   NFSize s{0, 0, 0};
   for (unsigned i = 1; i <= mCnf.MaxVariable(); ++i)
   {
      s += CountAddClausesForLiteralAtLevel(Literal(i, true))
           + CountAddClausesForLiteralAtLevel(Literal(i, false));
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncQuad::CountAddClausesForLiteralAtLevel(const Literal &lit) const
{
   switch (mDirection)
   {
   case IF:
      return CountAddClausesForLiteralAtLevelIF(lit);
   case ONLY_IF:
      return CountAddClausesForLiteralAtLevelONLYIF(lit);
   case BOTH:
      return CountAddClausesForLiteralAtLevelIF(lit)
             + CountAddClausesForLiteralAtLevelONLYIF(lit);
   default:
      throw BadParameterException(
          "UPEncQuad::CountAddClausesForLiteralAtLevel", "Bad direction");
   }
}

SUBool::NFSize
SUBool::UPEncQuad::CountAddClausesForLiteralAtLevelIF(const Literal &lit) const
{
   // NOTE: The estimate is slightly off (higher) if the CNF contains duplicate
   //       binary clauses. Should not be a problem in the application of the
   //       encoding.
   NFSize s{/* .max_variable = */ 0, /* .size = */ 1, /* .len = */ 2};
   for (const auto &clause_index : mCl2Lit.Clauses(lit))
   {
      const Clause &c = mCnf[clause_index];
      if (c.Size() == 2)
      {
         ++s.len;
      }
      else
      {
         ++s.max_variable;
         s.len += c.Size() * 3 - 1;
         s.size += c.Size();
      }
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncQuad::CountAddClausesForLiteralAtLevelONLYIF(
    const Literal &lit) const
{
   NFSize s{/* .max_variable = */ 0, /* .size = */ 1, /* .len = */ 2};
   for (const auto &clause_index : mCl2Lit.Clauses(lit))
   {
      ++s.size;
      s.len += mCnf[clause_index].Size();
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncQuad::EstimateSize(unsigned max_level) const
{
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   return CountAddClauses(max_level, CountInitVars(max_level));
}
