/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file up_quad_impl.cpp
 * Quadratic size encoding of unit propagation constraint which uses a given
 * implication system equivalent to the dual rail encoding of the CNF to be
 * encoded.
 */

#include "up_quad_impl.h"
#include "impl_system_graph.h"

SUBool::UPEncQuadImpl::UPEncQuadImpl(const CNF &cnf,
    const ClausesToLiterals &cl2lit, VarStore &var_store,
    const LitImplSystem &is)
    : UPEncoder(cnf, cl2lit, var_store), mImplSystem(is)
{
}

unsigned
SUBool::UPEncQuadImpl::MaxLevelBound() const
{
   return mCnf.MaxVariable();
}

unsigned
SUBool::UPEncQuadImpl::NextLevelBound(unsigned level) const
{
   return (level > 0 && level < MaxLevelBound() ? level + 1 : 0);
}

void
SUBool::UPEncQuadImpl::InitVars(
    std::vector<std::vector<unsigned>> *vars, unsigned max_level) const
{
   vars->resize(max_level + 1);
   for (unsigned i = 0; i < vars->size(); ++i)
   {
      (*vars)[i].resize(2 * mCnf.MaxVariable());
      for (unsigned j = 0; j < 2 * mCnf.MaxVariable(); ++j)
      {
         mVarStore.NameStream() << "_{" << Literal(j) << "}^{" << i << "}";
         (*vars)[i][j] = mVarStore.NewVariable();
      }
   }
   // Also initialize variables for implications at levels 1...max_level
   // or not --- they are needed only for bodies of size bigger than 1.
}

void
SUBool::UPEncQuadImpl::Encode(unsigned max_level, std::vector<unsigned> *inputs,
    std::vector<unsigned> *outputs, std::vector<Clause> *clauses) const
{
   if (inputs == nullptr)
   {
      throw NullPointerException("UPEncQuadImpl::Encode", "inputs == nullptr");
   }
   if (outputs == nullptr)
   {
      throw NullPointerException("UPEncQuadImpl::Encode", "outputs == nullptr");
   }

   // logs.TLog(1) << "Using implication system:\n" << mImplSystem << std::endl;
   // logs.TLog(3) << "[UPEncQuadImpl::Encode] Using implication system: "
   //<< std::endl;
   // logs.Log(3) << mImplSystem << std::endl;

   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   std::vector<std::vector<unsigned>> vars;
   InitVars(&vars, max_level);
   AddClauses(vars, clauses);
   *inputs = std::move(vars.front());
   *outputs = std::move(vars.back());
}

void
SUBool::UPEncQuadImpl::AddClauses(
    const std::vector<std::vector<unsigned>> &vars,
    std::vector<Clause> *clauses) const
{
   AddUnitsForEmptyBodies(vars[0], clauses);
   for (unsigned i = 1; i < vars.size(); ++i)
   {
      AddClausesAtLevel(vars[i - 1], vars[i], i, clauses);
   }
}

void
SUBool::UPEncQuadImpl::AddUnitsForEmptyBodies(
    const std::vector<unsigned> input_vars, std::vector<Clause> *clauses) const
{
   for (const auto &impl : mImplSystem)
   {
      if (impl.EmptyBody())
      {
         for (const auto &lit : impl.Cons())
         {
            clauses->emplace_back(Literal(input_vars[lit.Index()], true));
         }
      }
   }
}

std::vector<SUBool::Clause>
SUBool::UPEncQuadImpl::InitLitClauses(const std::vector<unsigned> &prev_vars,
    const std::vector<unsigned> &vars) const
{
   std::vector<Clause> lit_clauses(2 * mCnf.MaxVariable());
   for (unsigned lit_index = 0; lit_index < lit_clauses.size(); ++lit_index)
   {
      lit_clauses[lit_index].PushBack(Literal(vars[lit_index], false));
      lit_clauses[lit_index].PushBack(Literal(prev_vars[lit_index], true));
   }
   return lit_clauses;
}

void
SUBool::UPEncQuadImpl::AddClausesAtLevel(const std::vector<unsigned> &prev_vars,
    const std::vector<unsigned> &vars, unsigned level,
    std::vector<Clause> *clauses) const
{
   auto lit_clauses = InitLitClauses(prev_vars, vars);
   size_t impl_index = 0;
   for (const auto &impl : mImplSystem)
   {
      AddClausesForImplAtLevel(
          impl, impl_index++, prev_vars, level, &lit_clauses, clauses);
   }
   for (auto &lit_clause : lit_clauses)
   {
      clauses->push_back(std::move(lit_clause));
   }
}

void
SUBool::UPEncQuadImpl::AddClausesForImplAtLevel(const LitImpl &impl,
    size_t impl_index, const std::vector<unsigned> &prev_vars, unsigned level,
    std::vector<Clause> *lit_clauses, std::vector<Clause> *clauses) const
{
   if (impl.EmptyBody())
   {
      return;
   }
   auto body_literal = BodyLiteral(impl, impl_index, prev_vars, level);
   // add body_literal to lit_clauses for literals in cons
   for (const auto &lit : impl.Cons())
   {
      (*lit_clauses)[lit.Index()].push_back(body_literal);
   }
   // add clauses representing body_literal->body
   if (impl.Body().size() > 1)
   {
      for (const auto &lit : impl.Body())
      {
         const unsigned v = prev_vars[lit.Index()];
         clauses->emplace_back(body_literal.Negation(), Literal(v, true));
      }
   }
}

SUBool::Literal
SUBool::UPEncQuadImpl::BodyLiteral(const LitImpl &impl, size_t impl_index,
    const std::vector<unsigned> &prev_vars, unsigned level) const
{
   if (impl.Body().size() == 1)
   {
      return Literal(prev_vars[impl.Body().front().Index()], true);
   }
   assert(impl.Body().size() > 1);
   mVarStore.NameStream() << "z_{" << impl_index << "}^{" << level << "}";
   return Literal(mVarStore.NewVariable(), true);
}

SUBool::NFSize
SUBool::UPEncQuadImpl::EstimateSize(unsigned max_level) const
{
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   return CountAddClauses(max_level, CountInitVars(max_level));
}

SUBool::NFSize
SUBool::UPEncQuadImpl::CountInitVars(unsigned max_level) const
{
   return {(max_level + 1) * 2 * mCnf.MaxVariable(), 0, 0};
}

SUBool::NFSize
SUBool::UPEncQuadImpl::CountAddUnitsForEmptyBodies() const
{
   NFSize s;
   for (const auto &impl : mImplSystem)
   {
      if (impl.EmptyBody())
      {
         s += {0, impl.Cons().size(), impl.Cons().size()};
      }
   }
   return s;
}

SUBool::NFSize
SUBool::UPEncQuadImpl::CountAddClauses(unsigned max_level, NFSize s) const
{
   s += CountAddUnitsForEmptyBodies();
   s += CountAddClausesAtLevel() * max_level;
   return s;
}

SUBool::NFSize
SUBool::UPEncQuadImpl::CountAddClausesAtLevel() const
{
   NFSize s;
   std::vector<size_t> lit_clauses(2 * mCnf.MaxVariable(), 2);
   for (const auto &impl : mImplSystem)
   {
      if (impl.EmptyBody())
      {
         continue;
      }
      for (const auto &lit : impl.Cons())
      {
         ++lit_clauses[lit.Index()];
      }
      if (impl.Body().size() > 1)
      {
         s += {1, impl.Body().size(), 2 * impl.Body().size()};
      }
   }
   s += {0, 2 * mCnf.MaxVariable(), 0};
   for (const auto &clause_cnt : lit_clauses)
   {
      s.len += clause_cnt;
   }
   return s;
}
