/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file up_scc_bin_direct.cpp
 * Encoding based on SCCs of the graph of the implication system with levels
 * encoded in binary. This variant uses direct inequalities between the levels
 * of literals in the body and tail of each implication.
 */
#include "up_scc_bin_direct.h"
#include "up_levels.h"

void
SUBool::UPSCCBinaryDirect::Encode(unsigned max_level,
    std::vector<unsigned> *inputs, std::vector<unsigned> *outputs,
    std::vector<Clause> *clauses) const
{
   assert(inputs != nullptr);
   assert(outputs != nullptr);
   assert(clauses != nullptr);

   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }

#if 0
   if (logs.CheckLevel(3))
   {
      logs.TLog(3) << "[UPSCCBinaryDirect::Encode] Using implication system:"
                   << std::endl;
      mSCCGraph.DumpImplSystem(3, mImplSystem);
   }
#endif

   Context ctx{max_level, VarInfoList(2 * mCnf.MaxVariable()),
       mVarStore.VarArray(mCnf.MaxVariable(), true, ":in(", ")"),
       mVarStore.VarArray(mCnf.MaxVariable(), true, ":out(", ")"), {}};
   InitOutClauses(ctx);

   AddClauses(ctx, *clauses);
   *inputs = std::move(ctx.inputs);
   *outputs = std::move(ctx.outputs);
}

auto
SUBool::UPSCCBinaryDirect::EstimateSize(
    unsigned max_level [[maybe_unused]]) const -> NFSize
{
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }

   NFSize s{
       4 * mCnf.MaxVariable(), 2 * mCnf.MaxVariable(), 4 * mCnf.MaxVariable()};

   for (unsigned i = 0; i < mSCCGraph.SCCCount(); ++i)
   {
      if (mSCCGraph.SCCSize(i) <= 1)
      {
         continue;
      }
      unsigned bit_width = BitWidthForSCC(max_level, i);
      s.max_variable += bit_width * mSCCGraph.SCCSize(i);
   }

   std::set<std::pair<Literal, Literal>> comparisons;
   for (const auto &impl : mImplSystem)
   {
      s += FireClausesSize(impl, max_level, comparisons);
   }
   return s;
}

auto
SUBool::UPSCCBinaryDirect::FireClausesSize(const LitImpl &impl,
    unsigned max_level,
    std::set<std::pair<Literal, Literal>> &comparisons) const -> NFSize
{
   NFSize s;
   auto scc_info = mSCCGraph.InspectImplSCC(impl);
   if (scc_info.cons_out_main_scc_size > 0)
   {
      if (impl.EmptyBody())
      {
         s += NFSize{1, 1, 1};
      }
      else if (impl.BodySize() > 1)
      {
         s += NFSize{1, impl.BodySize(), 2 * impl.BodySize()};
      }
   }
   if (scc_info.main_scc < mSCCGraph.SCCCount())
   {
      unsigned bit_width = BitWidthForSCC(max_level, scc_info.main_scc);

      for (const auto &lit : impl.Cons())
      {
         if (scc_info.main_scc == mSCCGraph.SCC(lit))
         {
            s += FireVariableInMainSCCSize(impl, bit_width, lit, comparisons);
         }
      }
   }
   s.len += impl.ConsSize();
   return s;
}

SUBool::NFSize
SUBool::UPSCCBinaryDirect::FireVariableInMainSCCSize(const LitImpl &impl,
    unsigned bit_width, const Literal &lit,
    std::set<std::pair<Literal, Literal>> &comparisons) const
{
   NFSize s{1, impl.BodySize(), 2 * impl.BodySize()};
   for (const auto &b : impl.Body())
   {
      if (mSCCGraph.SCC(lit) != mSCCGraph.SCC(b))
      {
         continue;
      }
      if (comparisons.find({lit, b}) == comparisons.end())
      {
         comparisons.insert({lit, b});
         s += UPLevels::CountEncodeGTSize(true, bit_width);
      }
      s.size += 1;
      s.len += 2;
   }
   return s;
}

unsigned
SUBool::UPSCCBinaryDirect::MaxLevelBound() const
{
   return mConf.use_bit_level
              ? bit_size_for_level(0, mSCCGraph.MaxEffSCCVarSize())
              : mSCCGraph.MaxEffSCCVarSize();
}

unsigned
SUBool::UPSCCBinaryDirect::NextLevelBound(unsigned level) const
{
   const auto max_level_bound = MaxLevelBound();
   if (level == 0 || level >= max_level_bound)
   {
      return 0;
   }

   NFSize s = EstimateSize(level);
   unsigned max_var = s.max_variable;
   do
   {
      ++level;
      s = EstimateSize(level);
   } while (level < max_level_bound && s.max_variable == max_var);
   return (level >= max_level_bound ? max_level_bound : level);
}

void
SUBool::UPSCCBinaryDirect::AddClauses(
    Context &ctx, std::vector<Clause> &clauses) const
{
   for (unsigned impl_index = 0; impl_index < mImplSystem.Size(); ++impl_index)
   {
      FireClauses(ctx, mImplSystem[impl_index], impl_index, clauses);
   }
   MoveOutClauses(ctx, clauses);
}

void
SUBool::UPSCCBinaryDirect::InitOutClauses(Context &ctx) const
{
   for (unsigned i = 0; i < ctx.vars.size(); ++i)
   {
      auto &out_clause = ctx.vars[i].out_clause;
      out_clause.push_back(make_negative_literal(ctx.outputs[i]));
      out_clause.push_back(make_positive_literal(ctx.inputs[i]));
   }
}

void
SUBool::UPSCCBinaryDirect::FireClauses(Context &ctx, const LitImpl &impl,
    unsigned impl_index, std::vector<Clause> &clauses) const
{
   unsigned simple_fire_var = 0;
   auto scc_info = mSCCGraph.InspectImplSCC(impl);
   if (scc_info.cons_out_main_scc_size > 0)
   {
      simple_fire_var = SimpleFireVar(ctx, impl, impl_index, clauses);
   }
   for (const auto &lit : impl.Cons())
   {
      if (scc_info.main_scc == mSCCGraph.SCC(lit))
      {
         FireVariableInMainSCC(ctx, impl, impl_index, lit, clauses);
         continue;
      }
      FireVariableNotInMainSCC(ctx, lit, simple_fire_var);
   }
}

void
SUBool::UPSCCBinaryDirect::MoveOutClauses(
    Context &ctx, std::vector<Clause> &clauses) const
{
   for (auto &var_info : ctx.vars)
   {
      clauses.emplace_back(std::move(var_info.out_clause));
   }
}

void
SUBool::UPSCCBinaryDirect::FireVariableInMainSCC(Context &ctx,
    const LitImpl &impl, unsigned impl_index, const Literal &lit,
    std::vector<Clause> &clauses) const
{
   // Literal in the main SCC
   // - Init fire variable f[impl_index, lit]
   mVarStore.NameStream() << "f[" << impl_index << ", " << lit << "]";
   unsigned fv = mVarStore.NewVariable();
   auto neg_fv = make_negative_literal(fv);
   // - Add implications to the output variables of the body
   FireVariableImpliesBody(ctx, impl, fv, clauses);
   // - Add implications to the inequalities
   for (const auto &b : impl.Body())
   {
      if (mSCCGraph.SCC(lit) != mSCCGraph.SCC(b))
      {
         continue;
      }
      // c represents B[lit]>B[b]
      auto c = Comparison(ctx, lit, b, clauses);
      // fv->B[lit]>B[b]
      clauses.emplace_back(neg_fv, make_positive_literal(c));
   }
   // - Add to the out clause
   ctx.vars[lit.Index()].out_clause.push_back(make_positive_literal(fv));
}

void
SUBool::UPSCCBinaryDirect::FireVariableNotInMainSCC(
    Context &ctx, const Literal &lit, unsigned simple_fire_var) const
{
   ctx.vars[lit.Index()].out_clause.push_back(
       make_positive_literal(simple_fire_var));
}

unsigned
SUBool::UPSCCBinaryDirect::SimpleFireVar(const Context &ctx,
    const LitImpl &impl, unsigned impl_index,
    std::vector<Clause> &clauses) const
{
   if (impl.BodySize() == 1)
   {
      return ctx.outputs[impl.Body().front().Index()];
   }
   mVarStore.NameStream() << "f[" << impl_index << "]";
   unsigned fv = mVarStore.NewVariable();
   FireVariableImpliesBody(ctx, impl, fv, clauses);
   return fv;
}

void
SUBool::UPSCCBinaryDirect::FireVariableImpliesBody(const Context &ctx,
    const LitImpl &impl, unsigned fv, std::vector<Clause> &clauses) const
{
   if (impl.EmptyBody())
   {
      clauses.emplace_back(make_positive_literal(fv));
      return;
   }
   auto neg_fv = make_negative_literal(fv);
   for (const auto &lit : impl.Body())
   {
      clauses.emplace_back(
          neg_fv, make_positive_literal(ctx.outputs[lit.Index()]));
   }
}

unsigned
SUBool::UPSCCBinaryDirect::Comparison(Context &ctx, const Literal &cons_literal,
    const Literal &body_literal, std::vector<Clause> &clauses) const
{
   auto cached_comp = ctx.comparisons.find({cons_literal, body_literal});
   if (cached_comp != ctx.comparisons.end())
   {
      return cached_comp->second;
   }
   return NewComparison(ctx, cons_literal, body_literal, clauses);
}

unsigned
SUBool::UPSCCBinaryDirect::NewComparison(Context &ctx,
    const Literal &cons_literal, const Literal &body_literal,
    std::vector<Clause> &clauses) const
{
   InitLevelVars(ctx, cons_literal);
   InitLevelVars(ctx, body_literal);
   unsigned new_comp =
       encode_bitvector_gt(ctx.vars[cons_literal.Index()].level_vars,
           ctx.vars[body_literal.Index()].level_vars, mVarStore, &clauses, "g",
           cons_literal.ToString(), body_literal.ToString());
   return ctx.comparisons[{cons_literal, body_literal}] = new_comp;
}

void
SUBool::UPSCCBinaryDirect::InitLevelVars(Context &ctx, const Literal &lit) const
{
   auto &level_vars = ctx.vars[lit.Index()].level_vars;
   if (!level_vars.empty())
   {
      return;
   }
   unsigned bit_width = BitWidthForSCC(ctx.max_level, mSCCGraph.SCC(lit));
   level_vars.reserve(bit_width);
   for (unsigned bit = 0; bit < bit_width; ++bit)
   {
      mVarStore.NameStream() << "B[" << lit << ", " << bit << "]";
      level_vars.push_back(mVarStore.NewVariable());
   }
}

unsigned
SUBool::UPSCCBinaryDirect::BitWidthForSCC(
    unsigned max_level, unsigned scc_index) const
{
   if (mConf.use_bit_level)
   {
      return std::min(
          max_level, bit_size_for_level(mSCCGraph.EffSCCVarSize(scc_index),
                         mSCCGraph.EffSCCVarSize(scc_index)));
   }
   return bit_size_for_level(max_level, mSCCGraph.EffSCCVarSize(scc_index));
}

