/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file up_scc_binary.cpp
 * Encoding based on SCCs of the graph of the implication system with levels
 * encoded in binary. This variant uses level variables for each implication.
 */
#include "up_scc_binary.h"
#include "up_levels.h"

//#define PRINT_IMPL_SYSTEM

void
SUBool::UPSCCBinary::Encode(unsigned max_level, std::vector<unsigned> *inputs,
    std::vector<unsigned> *outputs, std::vector<Clause> *clauses) const
{
   assert(inputs != nullptr);
   assert(outputs != nullptr);
   assert(clauses != nullptr);

   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }

#ifdef PRINT_IMPL_SYSTEM
   if (logs.CheckLevel(3))
   {
      logs.TLog(3) << "[UPSCCBinary::Encode] Using implication system:"
                   << std::endl;
      mSCCGraph.DumpImplSystem(3, mImplSystem);
   }
#endif

   Context ctx{max_level, VarInfoList(2 * mCnf.MaxVariable()),
       mVarStore.VarArray(mCnf.MaxVariable(), true, ":in(", ")"),
       mVarStore.VarArray(mCnf.MaxVariable(), true, ":out(", ")")};
   InitOutClauses(ctx);

   AddClauses(ctx, *clauses);
   *inputs = std::move(ctx.inputs);
   *outputs = std::move(ctx.outputs);
}

auto
SUBool::UPSCCBinary::EstimateSize(unsigned max_level) const -> NFSize
{
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }

   NFSize s{
       4 * mCnf.MaxVariable(), 2 * mCnf.MaxVariable(), 4 * mCnf.MaxVariable()};

   for (unsigned i = 0; i < mSCCGraph.SCCCount(); ++i)
   {
      if (mSCCGraph.SCCSize(i) <= 1)
      {
         continue;
      }
      unsigned bit_width = BitWidthForSCC(max_level, i);
      s.max_variable += bit_width * mSCCGraph.SCCSize(i);
   }

   for (const auto &impl : mImplSystem)
   {
      s += FireClausesSize(impl, max_level);
   }
   return s;
}

unsigned
SUBool::UPSCCBinary::BitWidthForSCC(
    unsigned max_level, unsigned scc_index) const
{
   if (mConf.use_bit_level)
   {
      return std::min(
          max_level, bit_size_for_level(mSCCGraph.EffSCCVarSize(scc_index),
                         mSCCGraph.EffSCCVarSize(scc_index)));
   }
   return bit_size_for_level(max_level, mSCCGraph.EffSCCVarSize(scc_index));
}

auto
SUBool::UPSCCBinary::FireClausesSize(
    const LitImpl &impl, unsigned max_level) const -> NFSize
{
   NFSize s;
   auto scc_info = mSCCGraph.InspectImplSCC(impl);
   if (scc_info.cons_out_main_scc_size > 0)
   {
      if (impl.EmptyBody())
      {
         s += NFSize{1, 1, 1};
      }
      else if (impl.BodySize() > 1)
      {
         s += NFSize{1, impl.BodySize(), 2 * impl.BodySize()};
      }
   }
   if (scc_info.main_scc < mSCCGraph.SCCCount())
   {
      unsigned bit_width = BitWidthForSCC(max_level, scc_info.main_scc);

      if (!SingleBodyLiteralInMainSCC(scc_info, impl))
      {
         s += NFSize{bit_width, 0, 0};
         for (const auto &lit : impl.Body())
         {
            if (mSCCGraph.SCC(lit) == scc_info.main_scc)
            {
               s += UPLevels::CountEncodeGTSize(true, bit_width);
               s += NFSize{0, 1, 1};
            }
         }
      }
      unsigned cons_in_main_scc_size =
          impl.ConsSize() - scc_info.cons_out_main_scc_size;
      s += NFSize{cons_in_main_scc_size,
          cons_in_main_scc_size * (impl.BodySize() + 1),
          2 * cons_in_main_scc_size * (impl.BodySize() + 1)};
      s += UPLevels::CountEncodeGTSize(true, bit_width) * cons_in_main_scc_size;
   }
   s.len += impl.ConsSize();
   return s;
}

unsigned
SUBool::UPSCCBinary::MaxLevelBound() const
{
   return mConf.use_bit_level
              ? bit_size_for_level(0, mSCCGraph.MaxEffSCCVarSize())
              : mSCCGraph.MaxEffSCCVarSize();
}

unsigned
SUBool::UPSCCBinary::NextLevelBound(unsigned level) const
{
   const auto max_level_bound = MaxLevelBound();
   if (level == 0 || level >= max_level_bound)
   {
      return 0;
   }
   /*if (mConf.use_bit_level)
   {
      return level + 1;
   }*/

   NFSize s = EstimateSize(level);
   unsigned max_var = s.max_variable;
   do
   {
      ++level;
      s = EstimateSize(level);
   } while (level < max_level_bound && s.max_variable == max_var);
   return (level >= max_level_bound ? max_level_bound : level);
}

void
SUBool::UPSCCBinary::AddClauses(
    Context &ctx, std::vector<Clause> &clauses) const
{
   for (unsigned impl_index = 0; impl_index < mImplSystem.Size(); ++impl_index)
   {
      FireClauses(ctx, mImplSystem[impl_index], impl_index, clauses);
   }
   MoveOutClauses(ctx, clauses);
}

void
SUBool::UPSCCBinary::InitOutClauses(Context &ctx) const
{
   for (unsigned i = 0; i < ctx.vars.size(); ++i)
   {
      auto &out_clause = ctx.vars[i].out_clause;
      out_clause.push_back(make_negative_literal(ctx.outputs[i]));
      out_clause.push_back(make_positive_literal(ctx.inputs[i]));
   }
}

void
SUBool::UPSCCBinary::FireClauses(Context &ctx, const LitImpl &impl,
    unsigned impl_index, std::vector<Clause> &clauses) const
{
   unsigned simple_fire_var = 0;
   auto scc_info = mSCCGraph.InspectImplSCC(impl);
   if (scc_info.cons_out_main_scc_size > 0)
   {
      simple_fire_var = SimpleFireVar(ctx, impl, impl_index, clauses);
   }
   std::vector<unsigned> impl_level_vars;
   if (scc_info.main_scc < mSCCGraph.SCCCount())
   {
      impl_level_vars =
          InitImplLevelVars(ctx, scc_info, impl, impl_index, clauses);
   }

   for (const auto &lit : impl.Cons())
   {
      if (scc_info.main_scc == mSCCGraph.SCC(lit))
      {
         FireVariableInMainSCC(
             ctx, impl, impl_index, lit, impl_level_vars, clauses);
         continue;
      }
      FireVariableNotInMainSCC(ctx, lit, simple_fire_var);
   }
}

void
SUBool::UPSCCBinary::MoveOutClauses(
    Context &ctx, std::vector<Clause> &clauses) const
{
   for (auto &var_info : ctx.vars)
   {
      clauses.emplace_back(std::move(var_info.out_clause));
   }
}

void
SUBool::UPSCCBinary::FireVariableInMainSCC(Context &ctx, const LitImpl &impl,
    unsigned impl_index, const Literal &lit,
    const std::vector<unsigned> &impl_level_vars,
    std::vector<Clause> &clauses) const
{
   // Literal in the main SCC
   // - Init fire variable f[impl_index, lit]
   mVarStore.NameStream() << "f[" << impl_index << ", " << lit << "]";
   unsigned fv = mVarStore.NewVariable();
   // auto neg_fv = make_negative_literal(fv);
   //  - Add implications to the output variables of the body
   FireVariableImpliesBody(ctx, impl, fv, clauses);
   // - Add implications to the inequalities
   InitLevelVars(ctx, lit);
   unsigned comp_var = encode_bitvector_gt(ctx.vars[lit.Index()].level_vars,
       impl_level_vars, mVarStore, &clauses, "g", lit.ToString(),
       std::to_string(impl_index));
   clauses.emplace_back(
       make_negative_literal(fv), make_positive_literal(comp_var));

   // - Add to the out clause
   ctx.vars[lit.Index()].out_clause.push_back(make_positive_literal(fv));
}

void
SUBool::UPSCCBinary::FireVariableNotInMainSCC(
    Context &ctx, const Literal &lit, unsigned simple_fire_var) const
{
   ctx.vars[lit.Index()].out_clause.push_back(
       make_positive_literal(simple_fire_var));
}

unsigned
SUBool::UPSCCBinary::SimpleFireVar(const Context &ctx, const LitImpl &impl,
    unsigned impl_index, std::vector<Clause> &clauses) const
{
   if (impl.BodySize() == 1)
   {
      return ctx.outputs[impl.Body().front().Index()];
   }
   mVarStore.NameStream() << "f[" << impl_index << "]";
   unsigned fv = mVarStore.NewVariable();
   FireVariableImpliesBody(ctx, impl, fv, clauses);
   return fv;
}

void
SUBool::UPSCCBinary::FireVariableImpliesBody(const Context &ctx,
    const LitImpl &impl, unsigned fv, std::vector<Clause> &clauses) const
{
   if (impl.EmptyBody())
   {
      clauses.emplace_back(make_positive_literal(fv));
      return;
   }
   auto neg_fv = make_negative_literal(fv);
   for (const auto &lit : impl.Body())
   {
      clauses.emplace_back(
          neg_fv, make_positive_literal(ctx.outputs[lit.Index()]));
   }
}

void
SUBool::UPSCCBinary::InitLevelVars(Context &ctx, const Literal &lit) const
{
   auto &level_vars = ctx.vars[lit.Index()].level_vars;
   if (!level_vars.empty())
   {
      return;
   }
   unsigned bit_width = BitWidthForSCC(ctx.max_level, mSCCGraph.SCC(lit));
   level_vars.reserve(bit_width);
   for (unsigned bit = 0; bit < bit_width; ++bit)
   {
      mVarStore.NameStream() << "B[" << lit << ", " << bit << "]";
      level_vars.push_back(mVarStore.NewVariable());
   }
}

std::optional<SUBool::Literal>
SUBool::UPSCCBinary::SingleBodyLiteralInMainSCC(
    const LitISGraph::ImplSCCInfo &scc_info, const LitImpl &impl) const
{
   unsigned cnt = 0;
   std::optional<Literal> single_literal;
   for (const auto &lit : impl.Body())
   {
      if (mSCCGraph.SCC(lit) == scc_info.main_scc)
      {
         ++cnt;
         single_literal = lit;
      }
      if (cnt > 1)
      {
         break;
      }
   }
   if (cnt == 1)
   {
      return single_literal;
   }
   return {};
}

std::vector<unsigned>
SUBool::UPSCCBinary::InitImplLevelVars(Context &ctx,
    const LitISGraph::ImplSCCInfo &scc_info, const LitImpl &impl,
    unsigned impl_index, std::vector<Clause> &clauses) const
{
   auto single_literal = SingleBodyLiteralInMainSCC(scc_info, impl);
   if (single_literal)
   {
      InitLevelVars(ctx, *single_literal);
      return ctx.vars[single_literal->Index()].level_vars;
   }

   unsigned bit_width = BitWidthForSCC(ctx.max_level, scc_info.main_scc);
   std::vector<unsigned> level_vars(bit_width);
   for (unsigned bit = 0; bit < level_vars.size(); ++bit)
   {
      mVarStore.NameStream() << "BI[" << impl_index << ", " << bit << "]";
      level_vars[bit] = mVarStore.NewVariable();
   }
   for (const auto &lit : impl.Body())
   {
      if (mSCCGraph.SCC(lit) == scc_info.main_scc)
      {
         InitLevelVars(ctx, lit);
         unsigned comp_var = encode_bitvector_gt(
             ctx.vars[lit.Index()].level_vars, level_vars, mVarStore, &clauses,
             "gi", lit.ToString(), std::to_string(impl_index));
         clauses.emplace_back(make_negative_literal(comp_var));
      }
   }
   return level_vars;
}
