/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file up_scc_binary.h
 * Encoding based on SCCs of the graph of the implication system with levels
 * encoded in binary. This variant uses level variables for each implication.
 */
#ifndef __UP_SCC_BINARY_H
#define __UP_SCC_BINARY_H

#include <map>
#include <vector>

#include "impl_system.h"
#include "impl_system_graph.h"
#include "upenc.h"

namespace SUBool
{
   class UPSCCBinary : public UPEncoder
   {
    public:
      struct Config
      {
         bool use_bit_level{true};

         Config() {}

         Config(bool ubl) : use_bit_level(ubl) {}

         void Dump(const std::string &line_prefix, unsigned level) const;
      };

      struct VarInfo
      {
         std::vector<unsigned> level_vars{};
         std::vector<Literal> out_clause{};
      };

      using VarInfoList = std::vector<VarInfo>;

      struct Context
      {
         unsigned max_level;
         VarInfoList vars;
         std::vector<unsigned> inputs;
         std::vector<unsigned> outputs;
      };

    private:
      const LitImplSystem &mImplSystem;
      const LitISGraph &mSCCGraph;
      Config mConf;

      void InitOutClauses(Context &ctx) const;
      void AddClauses(Context &ctx, std::vector<Clause> &clauses) const;
      void FireClauses(Context &ctx, const LitImpl &impl, unsigned impl_index,
          std::vector<Clause> &clauses) const;
      void MoveOutClauses(Context &ctx, std::vector<Clause> &clauses) const;
      void FireVariableInMainSCC(Context &ctx, const LitImpl &impl,
          unsigned impl_index, const Literal &lit,
          const std::vector<unsigned> &impl_level_vars,
          std::vector<Clause> &clauses) const;
      void FireVariableNotInMainSCC(
          Context &ctx, const Literal &lit, unsigned simple_fire_var) const;
      unsigned SimpleFireVar(const Context &ctx, const LitImpl &impl,
          unsigned impl_index, std::vector<Clause> &clauses) const;
      void FireVariableImpliesBody(const Context &ctx, const LitImpl &impl,
          unsigned fv, std::vector<Clause> &clauses) const;
      void InitLevelVars(Context &ctx, const Literal &lit) const;
      std::vector<unsigned> InitImplLevelVars(Context &ctx,
          const LitISGraph::ImplSCCInfo &scc_info, const LitImpl &impl,
          unsigned impl_index, std::vector<Clause> &clauses) const;
      std::optional<Literal> SingleBodyLiteralInMainSCC(
          const LitISGraph::ImplSCCInfo &scc_info, const LitImpl &impl) const;

      unsigned BitWidthForSCC(unsigned max_level, unsigned scc_index) const;

      NFSize FireClausesSize(const LitImpl &impl, unsigned max_level) const;

    public:
      UPSCCBinary(const Config &conf, const CNF &cnf,
          const ClausesToLiterals &cl2lit, VarStore &var_store,
          const LitImplSystem &impl_system, const LitISGraph &scc_graph)
          : UPEncoder(cnf, cl2lit, var_store), mImplSystem(impl_system),
            mSCCGraph(scc_graph), mConf(conf)
      {
      }

      virtual void Encode(unsigned max_level, std::vector<unsigned> *inputs,
          std::vector<unsigned> *outputs,
          std::vector<Clause> *clauses) const override;
      virtual NFSize EstimateSize(unsigned max_level) const override;
      virtual unsigned MaxLevelBound() const override;
      virtual unsigned NextLevelBound(unsigned level) const override;
   };
} // namespace SUBool

#endif
