/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file up_scc_unary.cpp
 * Encoding based on SCCs of the graph of the implication system with levels
 * encoded in unary.
 */

#include "up_scc_unary.h"

void
SUBool::UPSCCUnary::Encode(unsigned max_level, std::vector<unsigned> *inputs,
    std::vector<unsigned> *outputs, std::vector<Clause> *clauses) const
{
   assert(inputs != nullptr);
   assert(outputs != nullptr);
   assert(clauses != nullptr);

   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
#if 0
   if (logs.CheckLevel(3))
   {
      logs.TLog(3) << "[UPSCCUnary::Encode] Using implication system:"
                   << std::endl;
      mSCCGraph.DumpImplSystem(3, mImplSystem);
   }
#endif

   Context ctx = InitialContext(max_level);
   InitLevelClauses(ctx);
   AddClauses(ctx, *clauses);
   *inputs = std::move(ctx.inputs);
   *outputs = std::move(ctx.outputs);
}

unsigned
SUBool::UPSCCUnary::LPrime(const Literal &lit) const
{
   auto scc_index = mSCCGraph.SCC(lit);
   return scc_index == mSCCGraph.SCC(lit.Negation())
              ? mSCCGraph.SCCVarSize(scc_index)
              : mSCCGraph.SCCVarSize(scc_index) - 1;
}

auto
SUBool::UPSCCUnary::InitVarInfo(unsigned max_level) const -> VarInfoList
{
   // vars are indexed by literals (their indices), each literal has variables
   // at levels 0, ..., l'(e).
   // std::cout << "max_level: " << max_level << std::endl;
   VarInfoList vars(mCnf.MaxVariable() * 2);
   for (unsigned i = 0; i < mCnf.MaxVariable() * 2; ++i)
   {
      Literal lit(i);
      /*std::cout << "lit: " << lit
                << ", eff scc var size: " << mSCCGraph.EffSCCVarSize(lit)
                << std::endl;*/
      auto lprime = std::min(LPrime(lit), max_level);
      // auto lprime=std::min(mSCCGraph.EffSCCVarSize(lit), max_level);
      vars[i].level_vars.reserve(lprime + 1);
      for (unsigned l = 0; l <= lprime; ++l)
      {
         if (l == lprime)
         {
            mVarStore.NameStream() << ":out";
         }
         mVarStore.NameStream() << "[" << lit << ", " << l << "]";
         vars[i].level_vars.push_back(mVarStore.NewVariable());
      }
   }
   return vars;
}

auto
SUBool::UPSCCUnary::InitialContext(unsigned max_level) const -> Context
{
   Context ctx(max_level);
   ctx.vars = InitVarInfo(max_level);
   ctx.inputs = mVarStore.VarArray(mCnf.MaxVariable(), true, ":in(", ")");
   for (const auto &var_info : ctx.vars)
   {
      ctx.outputs.push_back(var_info.level_vars.back());
   }
   return ctx;
}

auto
SUBool::UPSCCUnary::InitialContextSize(unsigned max_level) const -> NFSize
{
   NFSize s{mCnf.MaxVariable() * 2, 0, 0};
   for (unsigned i = 0; i < mCnf.MaxVariable() * 2; ++i)
   {
      Literal lit(i);
      auto lprime_plus_one = std::min(LPrime(lit), max_level) + 1;
      // auto lprime_plus_one =
      //    std::min(mSCCGraph.EffSCCVarSize(lit), max_level) + 1;
      s += NFSize{lprime_plus_one, lprime_plus_one, 2 * lprime_plus_one};
   }
   return s;
}

void
SUBool::UPSCCUnary::InitLevelClauses(Context &ctx) const
{
   for (unsigned i = 0; i < ctx.vars.size(); ++i)
   {
      auto &lclauses = ctx.vars[i].level_clauses;
      auto &lvars = ctx.vars[i].level_vars;
      lclauses.reserve(lvars.size());
      lclauses.push_back(
          {Literal(lvars[0], false), Literal(ctx.inputs[i], true)});
      for (unsigned l = 1; l < lvars.size(); ++l)
      {
         lclauses.push_back(
             {Literal(lvars[l], false), Literal(lvars[l - 1], true)});
      }
   }
}

auto
SUBool::UPSCCUnary::EstimateSize(unsigned max_level) const -> NFSize
{
   unsigned max_level_bound = MaxLevelBound();
   if (max_level == 0 || max_level > max_level_bound)
   {
      max_level = max_level_bound;
   }
   auto s = InitialContextSize(max_level);
   // std::cout << "After InitialContextSize: " << s << std::endl;
   s += FireClausesSize(max_level);
   // std::cout << "After FireClausesSize: " << s << std::endl;
   return s;
}

unsigned
SUBool::UPSCCUnary::MaxLevelBound() const
{
   return mSCCGraph.MaxEffSCCVarSize();
}

unsigned
SUBool::UPSCCUnary::NextLevelBound(unsigned level) const
{
   if (level == 0 || level >= MaxLevelBound())
   {
      return 0;
   }
   return level + 1;
}

void
SUBool::UPSCCUnary::AddClauses(Context &ctx, std::vector<Clause> &clauses) const
{
   unsigned impl_index = 0;
   for (const auto &impl : mImplSystem)
   {
      FireClauses(ctx, impl, impl_index++, clauses);
   }
   // std::cout << "After fire clauses" << std::endl;
   MoveLevelClauses(ctx, clauses);
   // std::cout << "Moved level clauses" << std::endl;
}

void
SUBool::UPSCCUnary::MoveLevelClauses(
    Context &ctx, std::vector<Clause> &clauses) const
{
   for (auto &var_info : ctx.vars)
   {
      for (auto &lclause : var_info.level_clauses)
      {
         clauses.emplace_back(std::move(lclause));
      }
      var_info.level_clauses.clear();
   }
}

void
SUBool::UPSCCUnary::FireClauses(Context &ctx, const LitImpl &impl,
    unsigned impl_index, std::vector<Clause> &clauses) const
{
   auto fv = InitFireVariables(ctx, impl, impl_index);
   // logs.TLog(3) << "Main scc index of implication " << impl << " is "
   //<< fv.main_scc << std::endl;
   if (impl.Body().size() > 1)
   {
      FireVarDefinition(ctx, impl, fv, clauses);
   }
   AddFireVarsToLevelClauses(ctx, impl, fv);
}

SUBool::NFSize
SUBool::UPSCCUnary::FireClausesSize(unsigned max_level) const
{
   NFSize s;
   for (const auto &impl : mImplSystem)
   {
      auto scc_info = mSCCGraph.InspectImplSCC(impl);
      if (impl.Body().size() != 1)
      {
         unsigned fv_count = 0;
         if (scc_info.cons_out_main_scc_size > 0)
         {
            ++fv_count;
         }
         if (scc_info.main_scc < mSCCGraph.SCCCount())
         {
            fv_count +=
                std::min(mSCCGraph.EffSCCVarSize(scc_info.main_scc), max_level);
         }
         s += NFSize{fv_count, fv_count * impl.Body().size(),
             fv_count * impl.Body().size() * 2};
      }
      for (const auto &lit : impl.Cons())
      {
         s.len += std::min(LPrime(lit), max_level);
      }
      s.len += scc_info.cons_out_main_scc_size;
   }
   return s;
}

void
SUBool::UPSCCUnary::FireVarDefinition(Context &ctx, const LitImpl &impl,
    const FireVariables &fv, std::vector<Clause> &clauses) const
{
   if (fv.simple_fire_var > 0)
   {
      SimpleFireVarDefinition(ctx, impl, fv.simple_fire_var, clauses);
   }
   for (unsigned l = 0; l < fv.level_fire_vars.size(); ++l)
   {
      LevelFireVarDefinition(
          ctx, impl, fv.level_fire_vars[l], l, fv.main_scc, clauses);
   }
}

void
SUBool::UPSCCUnary::SimpleFireVarDefinition(Context &ctx, const LitImpl &impl,
    unsigned fire_var, std::vector<Clause> &clauses) const
{
   auto fire_lit = make_negative_literal(fire_var);
   for (const auto &lit : impl.Body())
   {
      clauses.emplace_back(
          fire_lit, make_positive_literal(ctx.outputs[lit.Index()]));
   }
}

void
SUBool::UPSCCUnary::LevelFireVarDefinition(Context &ctx, const LitImpl &impl,
    unsigned fire_var, unsigned level, unsigned main_scc,
    std::vector<Clause> &clauses) const
{
   auto fire_lit = make_negative_literal(fire_var);
   for (const auto &lit : impl.Body())
   {
      if (mSCCGraph.SCC(lit) == main_scc)
      {
         clauses.emplace_back(fire_lit,
             make_positive_literal(ctx.vars[lit.Index()].level_vars[level]));
         continue;
      }
      clauses.emplace_back(
          fire_lit, make_positive_literal(ctx.outputs[lit.Index()]));
   }
}

void
SUBool::UPSCCUnary::AddFireVarsToLevelClauses(
    Context &ctx, const LitImpl &impl, const FireVariables &fv) const
{
   for (const auto &lit : impl.Cons())
   {
      auto &level_clauses = ctx.vars[lit.Index()].level_clauses;
      if (mSCCGraph.SCC(lit) == fv.main_scc)
      {
         for (unsigned l = 1; l < level_clauses.size(); ++l)
         {
            level_clauses[l].push_back(
                make_positive_literal(fv.level_fire_vars[l - 1]));
         }
         continue;
      }
      // for (unsigned l = 0; l < level_clauses.size(); ++l)
      //{
      // simple_fire_var is only needed at level 0
      level_clauses.front().push_back(
          make_positive_literal(fv.simple_fire_var));
      //}
   }
}

auto
SUBool::UPSCCUnary::InitFireVariables(Context &ctx, const LitImpl &impl,
    unsigned impl_index) const -> FireVariables
{
   auto scc_info = mSCCGraph.InspectImplSCC(impl);
   if (impl.Body().size() == 1)
   {
      return UnifiedFireVariables(ctx, impl, scc_info);
   }
   return NewFireVariables(impl_index, scc_info, ctx.max_level);
}

auto
SUBool::UPSCCUnary::NewFireVariables(unsigned impl_index,
    const LitISGraph::ImplSCCInfo &scc_info, unsigned max_level) const
    -> FireVariables
{
   FireVariables fv{scc_info.main_scc, 0, {}};
   if (scc_info.cons_out_main_scc_size > 0)
   {
      mVarStore.NameStream() << "f[" << impl_index << "]";
      fv.simple_fire_var = mVarStore.NewVariable();
   }
   if (scc_info.main_scc < mSCCGraph.SCCCount())
   {
      for (unsigned l = 0;
           l < std::min(mSCCGraph.EffSCCVarSize(scc_info.main_scc), max_level);
           ++l)
      {
         mVarStore.NameStream() << "f[" << impl_index << ", " << l << "]";
         fv.level_fire_vars.push_back(mVarStore.NewVariable());
      }
   }
   return fv;
}

auto
SUBool::UPSCCUnary::UnifiedFireVariables(Context &ctx, const LitImpl &impl,
    const LitISGraph::ImplSCCInfo &scc_info) const -> FireVariables
{
   FireVariables fv{scc_info.main_scc, 0, {}};
   unsigned lit_index = impl.Body().front().Index();
   if (scc_info.cons_out_main_scc_size > 0)
   {
      fv.simple_fire_var = ctx.outputs[lit_index];
   }
   if (scc_info.main_scc < mSCCGraph.SCCCount())
   {
      // for (unsigned l = 0; l < mSCCGraph.EffSCCVarSize(scc_info.main_scc);
      // ++l)
      for (unsigned l = 0; l < ctx.vars[lit_index].level_vars.size(); ++l)
      {
         fv.level_fire_vars.push_back(ctx.vars[lit_index].level_vars[l]);
      }
   }
   return fv;
}
