/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file up_scc_unary.h
 * Encoding based on SCCs of the graph of the implication system with levels
 * encoded in unary.
 */
#ifndef __UP_SCC_UNARY_H
#define __UP_SCC_UNARY_H

#include "impl_system.h"
#include "impl_system_graph.h"
#include "upenc.h"

namespace SUBool
{
   class UPSCCUnary : public UPEncoder
   {
    public:
      struct VarInfo
      {
         std::vector<unsigned> level_vars{};
         std::vector<std::vector<Literal>> level_clauses{};
      };

      using VarInfoList = std::vector<VarInfo>;

      struct Context
      {
         unsigned max_level{0};
         VarInfoList vars{};
         std::vector<unsigned> inputs{};
         std::vector<unsigned> outputs{};

         Context(unsigned ml) : max_level(ml) {}
      };

      struct FireVariables
      {
         unsigned main_scc;
         unsigned simple_fire_var;
         std::vector<unsigned> level_fire_vars;
      };

    private:
      const LitImplSystem &mImplSystem;
      const LitISGraph &mSCCGraph;

      Context InitialContext(unsigned max_level) const;
      VarInfoList InitVarInfo(unsigned max_level) const;
      FireVariables InitFireVariables(
          Context &ctx, const LitImpl &impl, unsigned impl_index) const;
      FireVariables NewFireVariables(unsigned impl_index,
          const LitISGraph::ImplSCCInfo &scc_info, unsigned max_level) const;
      FireVariables UnifiedFireVariables(Context &ctx, const LitImpl &impl,
          const LitISGraph::ImplSCCInfo &scc_info) const;
      void InitLevelClauses(Context &ctx) const;
      void AddClauses(Context &ctx, std::vector<Clause> &clauses) const;
      unsigned LPrime(const Literal &lit) const;

      void FireClauses(Context &ctx, const LitImpl &impl, unsigned impl_index,
          std::vector<Clause> &clauses) const;
      void SimpleFireVarDefinition(Context &ctx, const LitImpl &impl,
          unsigned fire_var, std::vector<Clause> &clauses) const;
      void LevelFireVarDefinition(Context &ctx, const LitImpl &impl,
          unsigned fire_var, unsigned level, unsigned main_scc,
          std::vector<Clause> &clauses) const;
      void FireVarDefinition(Context &ctx, const LitImpl &impl,
          const FireVariables &fv, std::vector<Clause> &clauses) const;
      void AddFireVarsToLevelClauses(
          Context &ctx, const LitImpl &impl, const FireVariables &fv) const;
      void MoveLevelClauses(Context &ctx, std::vector<Clause> &clauses) const;

      NFSize InitialContextSize(unsigned max_level) const;
      NFSize FireClausesSize(unsigned max_level) const;

    public:
      UPSCCUnary(const CNF &cnf, const ClausesToLiterals &cl2lit,
          VarStore &var_store, const LitImplSystem &impl_system,
          const LitISGraph &scc_graph)
          : UPEncoder(cnf, cl2lit, var_store), mImplSystem(impl_system),
            mSCCGraph(scc_graph)
      {
      }

      virtual void Encode(unsigned max_level, std::vector<unsigned> *inputs,
          std::vector<unsigned> *outputs,
          std::vector<Clause> *clauses) const override;
      virtual NFSize EstimateSize(unsigned max_level) const override;
      virtual unsigned MaxLevelBound() const override;
      virtual unsigned NextLevelBound(unsigned level) const override;
   };
} // namespace SUBool

#endif
