/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file util.h
 * Various utility functions and structures.
 */

#ifndef __UTIL_H
#define __UTIL_H

#include <algorithm>
#include <iostream>
#include <optional>
#include <string>
#include <vector>

#include "sbexception.h"

namespace SUBool
{
   template <class T> T read_from_stream(std::istream &s);

   template <class InputIterator>
   void cyclic_increment(
       InputIterator &it, InputIterator begin, InputIterator end);
   template <class InputIterator>
   InputIterator cyclic_next(
       InputIterator it, InputIterator begin, InputIterator end);

   template <class E> void sort_unique(std::vector<E> &vec);
   template <class E>
   std::vector<E> vector_subtract(
       const std::vector<E> &a, const std::vector<E> &b);
   template <class E>
   std::vector<E> vector_union(
       const std::vector<E> &a, const std::vector<E> &b);
   template <class E, class Pred>
   void vector_erase_indices(std::vector<E> &a, Pred pred);

   template <class T> void vector_erase(std::vector<T> &vec, size_t i);

   template <class InputIterator>
   std::ostream &print_range(std::ostream &os, InputIterator begin,
       InputIterator end, const std::string &sep = " ");

   template <class InputIterator> struct RangePrinter
   {
      InputIterator mBegin;
      InputIterator mEnd;
      std::string mSep;

      RangePrinter(
          InputIterator begin, InputIterator end, std::string sep = " ")
          : mBegin(std::move(begin)), mEnd(std::move(end)), mSep(std::move(sep))
      {
      }
   };

   template <class InputIterator>
   std::ostream &operator<<(
       std::ostream &os, const RangePrinter<InputIterator> &printer);

   template <class E>
   struct VectorPrinter : RangePrinter<typename std::vector<E>::const_iterator>
   {
      using base_class = RangePrinter<typename std::vector<E>::const_iterator>;
      VectorPrinter(const std::vector<E> &vec, std::string sep = " ")
          : base_class(vec.begin(), vec.end(), std::move(sep))
      {
      }
   };

   template <class InputIterator>
   std::ostream &operator<<(
       std::ostream &os, const RangePrinter<InputIterator> &printer);

   template <class T>
   bool
   true_pred(const T &)
   {
      return true;
   }
   template <class T>
   bool
   false_pred(const T &)
   {
      return false;
   }

   template <class T>
   std::ostream &
   operator<<(std::ostream &os, std::optional<T> e)
   {
      if (e)
      {
         os << *e;
      }
      else
      {
         os << "(none)";
      }
      return os;
   }

   constexpr const char *
   bool_to_string(bool a)
   {
      return (a ? "true" : "false");
   }

   template <class T> struct Identity
   {
      constexpr T &&
      operator()(T &&t) const noexcept
      {
         return t;
      }
   };
} // namespace SUBool

// Implementation
template <class InputIterator>
void
SUBool::cyclic_increment(
    InputIterator &it, InputIterator begin, InputIterator end)
{
   if (it != end)
   {
      ++it;
   }
   if (it == end)
   {
      it = begin;
   }
}

template <class InputIterator>
InputIterator
SUBool::cyclic_next(InputIterator it, InputIterator begin, InputIterator end)
{
   cyclic_increment(it, begin, end);
   return it;
}

template <class E>
void
SUBool::sort_unique(std::vector<E> &vec)
{
   std::sort(vec.begin(), vec.end());
   auto end = std::unique(vec.begin(), vec.end());
   vec.erase(end, vec.end());
}

template <class E>
std::vector<E>
SUBool::vector_subtract(const std::vector<E> &a, const std::vector<E> &b)
{
   std::vector<E> c;
   std::set_difference(
       a.begin(), a.end(), b.begin(), b.end(), std::back_inserter(c));
   return c;
}

template <class E>
std::vector<E>
SUBool::vector_union(const std::vector<E> &a, const std::vector<E> &b)
{
   std::vector<E> c;
   std::merge(a.begin(), a.end(), b.begin(), b.end(), std::back_inserter(c));
   return c;
}

template <class InputIterator>
std::ostream &
SUBool::print_range(std::ostream &os, InputIterator begin, InputIterator end,
    const std::string &sep)
{
   if (begin == end)
   {
      return os;
   }
   os << *begin;
   for (++begin; begin != end; ++begin)
   {
      os << sep << *begin;
   }
   return os;
}

template <class InputIterator>
std::ostream &
SUBool::operator<<(std::ostream &os, const RangePrinter<InputIterator> &printer)
{
   return print_range(os, printer.mBegin, printer.mEnd, printer.mSep);
}

template <class E, class Pred>
void
SUBool::vector_erase_indices(std::vector<E> &a, Pred pred)
{
   size_t i = 0;
   for (size_t j = 0; j < a.size(); ++j)
   {
      if (!pred(j))
      {
         if (j > i)
         {
            a[i] = std::move(a[j]);
         }
         ++i;
      }
   }
   a.resize(i);
}

template <class T>
void
SUBool::vector_erase(std::vector<T> &vec, size_t i)
{
   if (i >= vec.size())
   {
      throw SUBool::OutOfBoundsException("vector_erase", "i is out of bounds");
   }
   if (i + 1 < vec.size())
   {
      vec[i] = std::move(vec.back());
   }
   vec.pop_back();
}

#endif
