/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file varstore.cpp
 * Class for storing and creating new variables.
 */

#include "varstore.h"
#include "literal.h"
#include "sbexception.h"

SUBool::VarStore::VarStore(bool use_backmap, unsigned last_variable)
    : mNullBuffer(), mUseBackMap(use_backmap), mLastVariable(last_variable),
      mNames(), mEmptyName(), mNameStream(), mPrefix()
{
   if (!use_backmap)
   {
      mNameStream.std::iostream::rdbuf(&mNullBuffer);
   }
}

SUBool::VarStore::VarStore(const VarStore &vs)
    : mNullBuffer(), mUseBackMap(vs.mUseBackMap),
      mLastVariable(vs.mLastVariable), mNames(vs.mNames),
      mEmptyName(vs.mEmptyName), mNameStream(), mPrefix(vs.mPrefix)
{
   mNameStream << vs.mNameStream.str();
}

SUBool::VarStore::VarStore(VarStore &&vs) noexcept
    : mNullBuffer(), mUseBackMap(vs.mUseBackMap),
      mLastVariable(vs.mLastVariable), mNames(std::move(vs.mNames)),
      mEmptyName(std::move(vs.mEmptyName)), mNameStream(),
      mPrefix(std::move(vs.mPrefix))
{
   mNameStream << vs.mNameStream.str();
}

SUBool::VarStore &
SUBool::VarStore::operator=(const VarStore &vs)
{
   mUseBackMap = vs.mUseBackMap;
   mLastVariable = vs.mLastVariable;
   mNames = vs.mNames;
   mEmptyName = vs.mEmptyName;
   mNameStream.str(vs.mNameStream.str());
   mPrefix = vs.mPrefix;
   return *this;
}

SUBool::VarStore &
SUBool::VarStore::operator=(VarStore &&vs) noexcept
{
   mUseBackMap = std::move(vs.mUseBackMap);
   mLastVariable = std::move(vs.mLastVariable);
   mNames = std::move(vs.mNames);
   mEmptyName = std::move(vs.mEmptyName);
   mNameStream.str(vs.mNameStream.str());
   mPrefix = std::move(vs.mPrefix);
   return *this;
}

void
SUBool::VarStore::Clear(unsigned last_variable)
{
   mLastVariable = last_variable;
   mNames.clear();
   mNameStream.clear();
   mPrefix.clear();
}

unsigned
SUBool::VarStore::NewVariable(const std::string &name)
{
   if (mUseBackMap)
   {
      mNames.push_back(mPrefix + name);
      // std::cout<<"last name: "<<mNames.back()<<" (prefix:
      // "<<mPrefix<<")"<<std::endl;
   }
   return ++mLastVariable;
}

unsigned
SUBool::VarStore::NewVariable()
{
   unsigned new_var = NewVariable(mNameStream.str());
   mNameStream.str("");
   return new_var;
}

std::vector<unsigned>
SUBool::VarStore::VarArray(unsigned n, bool dual_rail,
                           const std::string &prefix, const std::string &suffix)
{
   std::vector<unsigned> vars(dual_rail ? 2 * n : n);
   for (unsigned i = 0; i < vars.size(); ++i)
   {
      NameStream() << prefix;
      if (dual_rail)
      {
         NameStream() << Literal(i);
      }
      else
      {
         NameStream() << i + 1;
      }
      NameStream() << suffix;
      vars[i] = NewVariable();
   }
   return vars;
}

std::vector<SUBool::Literal>
SUBool::VarStore::LitArray(unsigned n, bool dual_rail,
                           const std::string &prefix, const std::string &suffix)
{
   unsigned limit = (dual_rail ? 2 * n : n);
   std::vector<Literal> lits;
   lits.reserve(limit);
   for (unsigned i = 0; i < limit; ++i)
   {
      NameStream() << prefix;
      if (dual_rail)
      {
         NameStream() << Literal(i);
      }
      else
      {
         NameStream() << i + 1;
      }
      NameStream() << suffix;
      lits.emplace_back(NewVariable(), true);
   }
   return lits;
}

void
SUBool::VarStore::InitVarArray(std::vector<unsigned> *vars, unsigned n,
                               bool dual_rail, const std::string &prefix,
                               const std::string &suffix)
{
   if (vars == nullptr)
   {
      throw NullPointerException("VarStore::InitVarArray", "vars == nullptr");
   }
   *vars = VarArray(n, dual_rail, prefix, suffix);
}

void
SUBool::VarStore::InitLitArray(std::vector<Literal> *lits, unsigned n,
                               bool dual_rail, const std::string &prefix,
                               const std::string &suffix)
{
   if (lits == nullptr)
   {
      throw NullPointerException("VarStore::InitLitArray", "lits == nullptr");
   }
   *lits = LitArray(n, dual_rail, prefix, suffix);
}

const std::string &
SUBool::VarStore::Name(unsigned var) const
{
   if (var == 0)
   {
      throw BadParameterException("VarStore::Name", "var==0");
   }
   if (mUseBackMap && var <= mNames.size())
   {
      return mNames[var - 1];
   }
   return mEmptyName;
}

unsigned
SUBool::VarStore::LastVariable() const
{
   return mLastVariable;
}

void
SUBool::VarStore::WriteClauseComment(std::ostream &out,
                                     const Clause &clause) const
{
   if (mUseBackMap)
   {
      out << 'c';
      for (const auto &lit : clause)
      {
         out << ' ';
         if (!lit.Positive())
         {
            out << '-';
         }
         out << Name(lit.Variable());
      }
      out << std::endl;
   }
}
std::ostream &
SUBool::VarStore::NameStream()
{
   return mNameStream;
}
