# Project description

The project consists of a library and programs used to manipulate boolean
functions and their representations, mainly CNFs and NNFs. The project is
highly experimental.

# Compilation

## Dependendencies

For the compilation the following prerequisities are needed.

- C++17 compiler and a POSIX shell.
- [cmake](https://cmake.org/) of version at least 3.0.2
- [glucose](https://www.labri.fr/perso/lsimon/glucose/) compiled as a library
  (`lib_release.a`) with incremental capabilities. For more details see [glucose](#glucose) subsection below.
- [Boost libraries](https://www.boost.org) of version at least 1.64.0. Some of the
  boost libraries have their own dependencies within the boost library.
    - [Boost Graph Library](https://www.boost.org/doc/libs/1_71_0/libs/graph/doc/index.html),
    in this library, only the header files are needed.
    - [Boost.Program_options](https://www.boost.org/doc/libs/1_77_0/doc/html/program_options.html)
    is needed as a compiled library.
- (OPTIONAL) [kissat](https://github.com/arminbiere/kissat) -- Kissat solver
  for checking propagation completeness as an option

### Options parsing

Until version 2.4, the programs needed library
[cxxopts](https://github.com/jarro2783/cxxopts) to parse command line
parameters. From version 2.5, this dependence was removed and instead, 
[Boost.Program_options](https://www.boost.org/doc/libs/1_77_0/doc/html/program_options.html)
is used to parse command line parameters. The disadvantage is that it is needed
as a compiled library, on the other hand, it allows an easy processing of
config files and makes thus easy to use different sets of parameters.

## Compilation

### Glucose

Glucose solver can be downloaded at https://www.labri.fr/perso/lsimon/glucose
If you want to make the compilation a bit simpler for yourself, then see below a
[note on the patched version](#patched-version).

The default path to glucose is `solvers/glucose` relative to the main project
directory. A different path can be specified using `GLUCOSE_HOME` shell
variable during configuration using cmake. The path should point to the
directory with the sources (i.e. where subdirectories such as `core`, `mtl` and
`simp` are present).

Glucose is expected to be compiled as a library `lib_release.a` which can be
achieved by calling `make libr` within `simp` subdirectory of the glucose home
directory.

Moreover, glucose has to be compiled with incremental solving capabilities. In
particular in the `simp` subdirectory of the glucose root directory, you can
call

```bash
CFLAGS="-Wall -Wno-parentheses -std=c++11 -DINCREMENTAL -DINCREMNENTAL" make libr 
```

Let us add a few notes to the above statement.
- CFLAGS is initialized as
    ```make
    CFLAGS    ?= -Wall -Wno-parentheses -std=c++11
    ```
  since we need to add macro definitions of `INCREMENTAL` and `INCREMNENTAL` and
  we want to keep the other flags, the only way (except editing the
  `template.mk` file) is to use these flags at command line.
- Macro `INCREMENTAL` needs to be defined when compiling glucose with
  incremental solver capabilities. The source code of glucose 4.1 (available at
  the [glucose web page](https://www.labri.fr/perso/lsimon/glucose)) contains a
  typo and we thus need to add the definition of macro `INCREMNENTAL` as well.
  This is not needed if we use a newer version such as [glucose 4.2.1](http://sat-race-2019.ciirc.cvut.cz/solvers/glucose-4.2.1.zip) available
  at [SAT Race 2019](http://sat-race-2019.ciirc.cvut.cz) pages.
- One further note is that the compilation may fail if the path to the glucose
  home directory contains spaces, in that case variable `PWD` has to be set to
  `..` in `simp/Makefile` (relative to the glucose top directory) and to `.` in
  `mtl/template.mk`.
 
#### Patched version
 
To simplify the matter, a patched version is
[available](https://gitlab.mff.cuni.cz/kucep6am/glucosemod) which is already
configured for the incremental solving. In this version only `make libr` in the
`simp` subdirectory should be sufficient.

### Kissat (OPTIONAL)

Project can be optionally compiled with Kissat support. In order to do that, it
is enough to have correctly set path to the Kissat home directory and have it
compiled (i.e. `libkissat.a` must exist in the `build` subdirectory).

- The Kissat home directory is the one with subdirectories `src` and `build`.
- The default path to the Kissat home directory is set to `solvers/kissat`
  relative to the main project directory. A different path can be specified
  using `KISSAT_HOME` shell variable during the configuration using cmake.
- It is possible to avoid compilation of Kissat support even if `libkissat.a`
  is compiled and prace at the correct location, just use parameter
  `-DUSE_KISSAT=OFF` when invoking cmake. Alternatively, call `make cmake
  USE_KISSAT=OFF` in the main project directory to call cmake with the right
  parameters.
- If `pccompile` or `pccheck` are compiled with Kissat support, then Kissat can
  be selected using option `--pccheck-solver`.

#### Kissat on macOS

Currently, Kissat calls may fail (with `bus error`) under macOS and thus using
Kissat with pccompile on macOS is not recommended, it should be fine under
Linux.

### The subool library and programs

Once glucose is compiled and all other prerequisities are available, the
library and binaries can be compiled using

```bash
make cmake
make
```

During configuration (i.e. during `make cmake`), cmake looks for Boost library
of version at least 1.64.0. A path to a specific installation of the Boost
library can be passed to cmake using `BOOST_ROOT` shell variable, e.g. with

```bash
BOOST_ROOT=path/to/boost make cmake
```

To compile a single program (e.g. `pccompile`), run `make cmake` if necessary
and after that run

```bash
cd build
make pccompile
```

The compiled programs are then present in `build/bin` subdirectory of the
main project directory

# Programs

The program main files are in the `bin` subdirectory. All programs should admit
options `--version` giving a one line program description and `--help` giving
the description of available parameters.

In all utilities a CNF is expected in the DIMACS format, a DNNF is expected in
the format used by [c2d](http://reasoning.cs.ucla.edu/c2d/) compiler.

## pccompile, pccheck, and pcminimize

The program `pccompile` implements a compiler of a CNF into a PC (or URC)
formula. The program implements to algorithms for compilation into a PC
formula, an _iterative_ algorithm based on iterative adding empowering
implicates to the input formula, and a _learning_ algorithm which is based on
the algorithm of (Angluin et al, 1992; Arias et al. 2015) for learning a Horn
formula using equivalence and membership or closure queries. The latter
algorithm is based on the fact that a formula $\varphi$ is propagation complete
if its dual rail encoding represents a specific Horn function which models
semantic implications and which depends only on the function represented by
$\varphi$.

Related to it are programs `pcmiminize` and `pccheck`. Program `pcminimize` makes a
given CNF formula irredundant with respect to propagation completeness by
iteratively removing absorptions. Program `pccheck` checks if a given CNF is
propagation complete (or unit refutation complete) using SAT.

## Other programs for CNF manipulation

`cnfcanon`: Computing the canonical CNF for a given CNF. The canonical CNF
consists of all prime implicates of the function it represents.

`cnfcomp`: Comparing two CNFs whether they are equal or one implies the other.

`cnfgraphx`: Generates a graph for a given CNF (primal, dual, or incidence).

`cnfprime`: Makes a given CNF prime.

`cnfdiff`: Compares two CNF formulas as sets of clauses (so only syntactically,
not semantically). The output is similar to a UNIX command `comm`.

`dpelim`: Performs DP elimination of auxiliary variables in a CNF encoding.

`gdbasis`: Computes the GD basis of a definite Horn formula or of a
implicational dual rail encoding of an arbitrary CNF.

`rhbdmc_compile`: An experimental compiler of a CNF into a Renamable Horn BDMC
(=Backdoor Decomposable Monotone Circuit).

The other programs (`rhornsplit`, `triple_cnf`) are highly experimental, they either solve a very
specialized problem, or dey do not work properly if they work at all.

## Programs for NNF manipulation

`nnfencode`: Encodes a given DNNF into a CNF, allows several encodings of
different level of propagation strength (URC/PC with respect to input/all
variables).

`nnfsmooth`: Checks if a given DNNF is smooth and can make it smooth if
necessary.
