#!/bin/sh

# Project SUBool
# Petr Kucera, 2018
#
# Script for generating file version.h with version info
# The script generates version.h based on calls to git log and git describe

scriptname="$0"

usage() {
   echo 'Project SUBool'
   echo 'genver -- script for generating the version.h header file containing the version info.'
   echo 'Petr Kucera, 2018'
   echo
   printf 'Usage: %s [-q|-h]\n\n' "$scriptname"
   echo "Uses variable SUBOOL_TOP to determine the top directory "\
      "of the SUBool project. If this variable is empty, the script "\
      "attempts to determine the top directory. If this fails, you "\
      "can try to set the value of SUBOOL_TOP to the correct value "\
      "before calling the script."
   echo
   echo "   -h - print the help screen and finish"
   echo "   -q - do not write anything to the output, exit code still "\
      "determines if generating of the version header file was "\
      "successful."
   echo "   -f - update fallback_version.h as well"
}

quiet=false
update_fallback=false
while getopts fhq optname
do
   case $optname in
      f) update_fallback=true ;;
      h) usage ; exit 1 ;;
      q) quiet=true ;;
      *) usage ; exit 1 ;;
   esac
done

if [ -z "$SUBOOL_TOP" ]
then
   mydir=$(dirname "$0")
   SUBOOL_TOP=$(cd "$mydir" && cd .. && pwd)
fi

TEMPLATE="$SUBOOL_TOP/script/version_template.h"
FALLBACK_VERSION_H="$SUBOOL_TOP/lib/fallback_version.h"
VERSION_H="$SUBOOL_TOP/lib/version.h"
TEMPFILE=/tmp/genver_version_h.$$

if [ ! -f "$TEMPLATE" ]
then
   $quiet || echo "Could not find version template file, did"\
      "you specify the correct path to subool top directory?"
   exit 1
fi

if [ ! -d "$SUBOOL_TOP/.git" ]
then
   $quiet || echo "Git repository not present, using fallback version.h."
   cp "$FALLBACK_VERSION_H" "$VERSION_H"
   exit 0
fi

if ! which git >/dev/null 2>&1
then
   $quiet || echo "Git not available."
   exit 0
fi

desc=$(git describe --long)
tag=$(echo "$desc" | cut -d- -f1)
revision=$(echo "$desc" | cut -d- -f2)
hash=$(echo "$desc" | cut -d- -f3 | cut -c2-)
date=$(git show -s --format=%ci "$hash")

sed "s/==DATE==/$date/
s/==TAG==/$tag/
s/==REV==/$revision/
s/==HASH==/$hash/" "$TEMPLATE" >"$TEMPFILE"

if [ ! -f "$VERSION_H" ] || ! diff "$TEMPFILE" "$VERSION_H">/dev/null 2>&1
then
   $quiet || echo "Updating version.h"
   mv "$TEMPFILE" "$VERSION_H"
else
   rm "$TEMPFILE"
fi

if $update_fallback
then
   $quiet || echo "Updating fallback_version.h"
   sed 's/^\(#define SUBOOL_FALLBACK\).*$/\1 " (fallback)"/' "$VERSION_H" >"$FALLBACK_VERSION_H"
fi
