/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file test_canon.h
 * Contains tests for canonical CNF computation.
 */

#include "test_canon.h"
#include "cnfcompare.h"
#include "empower.h"
#include "random_utils.h"
#include "resolve.h"

std::unique_ptr<SUBool::AbsCanon>
SUBool::TestCanon::CanonProcedure(CNF cnf, CanonType canon_type) const
{
   switch (canon_type)
   {
   case CanonType::BASE:
      return std::unique_ptr<SUBool::AbsCanon>(new CanonBase(cnf));
   case CanonType::SAT:
      return std::unique_ptr<SUBool::AbsCanon>(new CanonSAT(cnf));
   }
   throw BadParameterException("TestCanon::CanonProcedure",
                               "Bad canon type specification");
}

bool
SUBool::TestCanon::IsCanonCNF(const CNF &cnf, const CNF &canonical) const
{
   if (cnf_compare(cnf, canonical) != CompareResult::EQUAL)
   {
      std::cout << "cnf is not equal to canonical ";
      return false;
   }
   for (unsigned i = 0; i < cnf.Size(); ++i)
   {
      for (unsigned j = i + 1; j < cnf.Size(); ++j)
      {
         auto res = resolve(cnf[i], cnf[j]);
         if (res.second == 0)
         {
            continue;
         }
         bool subsumed = false;
         for (const auto &cl : canonical)
         {
            if (res.second == 0)
            {
               continue;
            }
            auto comp = res.first.SubsetCompare(cl);
            if (comp == SetCompareResult::SUPERSET
                || comp == SetCompareResult::EQUAL)
            {
               subsumed = true;
               break;
            }
         }
         if (!subsumed)
         {
            std::cout << "Clause " << res.first << " which is a resolvent of "
                      << cnf[i] << " and " << cnf[j]
                      << " not subsumed by canonical. ";
            return false;
         }
      }
   }
   return true;
}

bool
SUBool::TestCanon::TestCanonCNF(CNF cnf, CanonType canon_type) const
{
   std::cout << "Testing canonical CNF, type="
             << static_cast<unsigned>(canon_type) << std::endl;
   auto canon_proc = CanonProcedure(cnf, canon_type);
   canon_proc->Compute();
   bool success = IsCanonCNF(cnf, canon_proc->Cnf());
   if (success)
   {
      std::cout << "[OK]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
   }
   return success;
}

bool
SUBool::TestCanon::TestCanonRandom(unsigned n_vars, unsigned n_clauses,
                                   unsigned n_tests, CanonType canon_type) const
{
   bool success = true;
   while (success && n_tests > 0)
   {
      success = TestCanonCNF(random_cnf(n_clauses, n_vars), canon_type);
      --n_tests;
   }
   return success;
}

bool
SUBool::TestCanon::RunTests() const
{
   bool success = true;
   success = success && TestCanonRandom(10, 20, 20, CanonType::BASE);
   success = success && TestCanonRandom(10, 20, 20, CanonType::SAT);
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::TestCanon test;
   return !test.RunTests();
}
