/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file test_cardinal.cpp
 * Contains tests for encodings of cardinality constraints.
 */

#include "test_cardinal.h"
#include "cardinal.h"
#include "logstream.h"
#include "varstore.h"

bool
SUBool::TestCardinal::TestSinglePositive(
    const CNFEncoding &enc, const std::vector<Literal> &inputs) const
{
   logs.TLog(1) << "Testing assumptions with a single positive literal"
                << std::endl;
   for (const auto &lit : inputs)
   {
      PartialAssignment alpha(enc.MaxInputVariable());
      alpha.SetLiteral(lit);
      PartialAssignment expected(enc.MaxInputVariable());
      expected.SetLiteralsNegated(inputs.begin(), inputs.end());
      expected.SetLiteral(lit);
      if (!enc.Consistent(alpha))
      {
         logs.TLog(1) << " [FAILED] consistency check" << std::endl;
         return false;
      }
      auto beta = enc.Propagate(alpha);
      if (!beta)
      {
         logs.TLog(1) << " [FAILED] consistency of propagation" << std::endl;
         return false;
      }
      if (*beta != expected)
      {
         logs.TLog(1) << " [FAILED] propagated values" << std::endl;
         logs.TLog(2) << "... expected: " << expected << std::endl;
         logs.TLog(2) << "... obtained: " << *beta << std::endl;
         return false;
      }
   }
   logs.TLog(1) << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCardinal::TestTwoPositive(const CNFEncoding &enc,
                                      const std::vector<Literal> &inputs) const
{
   logs.TLog(1) << "Testing assumptions with two positive literals"
                << std::endl;
   for (size_t i = 0; i < inputs.size(); ++i)
   {
      for (size_t j = i + 1; j < inputs.size(); ++j)
      {
         PartialAssignment alpha(enc.MaxInputVariable());
         alpha.SetLiteral(inputs[i]);
         alpha.SetLiteral(inputs[j]);
         if (enc.Consistent(alpha))
         {
            logs.TLog(1) << " [FAILED] consistency check" << std::endl;
            return false;
         }
      }
   }
   logs.TLog(1) << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCardinal::TestAmoMostNegatives(
    const CNFEncoding &enc, const std::vector<Literal> &inputs) const
{
   logs.TLog(1) << "Testing AMO with most negatives" << std::endl;
   for (size_t i = 0; i <= inputs.size(); ++i)
   {
      PartialAssignment alpha(enc.MaxInputVariable());
      alpha.SetLiteralsNegated(inputs.begin(), inputs.end());
      if (i < inputs.size())
      {
         alpha.UnsetLiteral(inputs[i]);
      }
      if (!enc.Consistent(alpha))
      {
         logs.TLog(1) << " [FAILED] consistency check" << std::endl;
         return false;
      }
   }
   logs.TLog(1) << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCardinal::TestExOneMostNegatives(
    const CNFEncoding &enc, const std::vector<Literal> &inputs) const
{
   logs.TLog(1) << "Testing Exactly-1 with most negatives" << std::endl;
   for (size_t i = 0; i < inputs.size(); ++i)
   {
      PartialAssignment alpha(enc.MaxInputVariable());
      alpha.SetLiteralsNegated(inputs.begin(), inputs.end());
      alpha.UnsetLiteral(inputs[i]);
      if (!enc.Consistent(alpha))
      {
         logs.TLog(1) << " [FAILED] consistency check" << std::endl;
         return false;
      }
      auto beta = enc.Propagate(alpha);
      alpha.SetLiteral(inputs[i]);
      if (!beta)
      {
         logs.TLog(1) << " [FAILED] consistency of propagation" << std::endl;
         return false;
      }
      if (*beta != alpha)
      {
         logs.TLog(1) << " [FAILED] propagated values" << std::endl;
         logs.TLog(2) << "... expected: " << alpha << std::endl;
         logs.TLog(2) << "... obtained: " << *beta << std::endl;
         return false;
      }
   }
   logs.TLog(1) << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCardinal::TestFewNegatives(const CNFEncoding &enc,
                                       const std::vector<Literal> &inputs) const
{
   logs.TLog(1) << "Testing assumptions with few negative literals"
                << std::endl;
   for (size_t i = 0; i < inputs.size(); ++i)
   {
      for (size_t j = i + 1; j < inputs.size(); ++j)
      {
         PartialAssignment alpha(enc.MaxInputVariable());
         alpha.SetLiteralsNegated(inputs.begin(), inputs.end());
         alpha.UnsetLiteral(inputs[i]);
         alpha.UnsetLiteral(inputs[j]);
         if (!enc.Consistent(alpha))
         {
            logs.TLog(1) << " [FAILED] consistency check" << std::endl;
            return false;
         }
         auto beta = enc.Propagate(alpha);
         if (!beta)
         {
            logs.TLog(1) << " [FAILED] consistency of propagation" << std::endl;
            return false;
         }
         if (*beta != alpha)
         {
            logs.TLog(1) << " [FAILED] propagated values" << std::endl;
            logs.TLog(2) << "... expected: " << alpha << std::endl;
            logs.TLog(2) << "... obtained: " << *beta << std::endl;
            return false;
         }
      }
   }
   logs.TLog(1) << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCardinal::TestExOneEncoding(
    const CNFEncoding &enc, const std::vector<Literal> &inputs) const
{
   return TestSinglePositive(enc, inputs) && TestTwoPositive(enc, inputs)
          && TestFewNegatives(enc, inputs)
          && TestExOneMostNegatives(enc, inputs);
}

bool
SUBool::TestCardinal::TestAmoEncoding(const CNFEncoding &enc,
                                      const std::vector<Literal> &inputs) const
{
   return TestSinglePositive(enc, inputs) && TestTwoPositive(enc, inputs)
          && TestFewNegatives(enc, inputs) && TestAmoMostNegatives(enc, inputs);
}

bool
SUBool::TestCardinal::TestSize(unsigned n, const CNFEncoding &enc,
                               const ConstraintEncoder &encoder) const
{
   logs.TLog(1) << "Testing size" << std::endl;
   auto s_enc = enc.Cnf().GetNFSize();
   auto s_encoder = encoder.CountClauses(n, enc.MaxInputVariable());
   if (s_enc != s_encoder)
   {
      logs.TLog(1) << " [FAILED]" << std::endl;
      logs.TLog(1) << "Size mismatch:" << std::endl;
      logs.TLog(1) << "........ real: " << s_enc << std::endl;
      logs.TLog(1) << "... estimated: " << s_encoder << std::endl;
      return false;
   }
   logs.TLog(1) << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCardinal::TestAmoRepresentation(unsigned n) const
{
   AmoRepresentation encoder;
   logs.TLog(1) << "Testing " << encoder.Description() << " on " << n
                << " variables" << std::endl;
   auto inputs = random_literal_vector_with_size(n, 2 * n);
   auto max_var = std::max_element(inputs.begin(), inputs.end())->Variable();
   logs.TLog(3) << "Encoding: " << std::endl;
   PCEncoding enc(encoder.Encode(inputs), max_var, true);
   logs.TLog(3) << enc << std::endl;
   return TestSize(n, enc, encoder) && TestAmoEncoding(enc, inputs);
}

bool
SUBool::TestCardinal::TestExOneRepresentation(unsigned n) const
{
   ExOneRepresentation encoder;
   logs.TLog(1) << "Testing " << encoder.Description() << " on " << n
                << " variables" << std::endl;
   auto inputs = random_literal_vector_with_size(n, 2 * n);
   auto max_var = std::max_element(inputs.begin(), inputs.end())->Variable();
   logs.TLog(3) << "Encoding: " << std::endl;
   PCEncoding enc(encoder.Encode(inputs), max_var, true);
   logs.TLog(3) << enc << std::endl;
   return TestSize(n, enc, encoder) && TestExOneEncoding(enc, inputs);
}

bool
SUBool::TestCardinal::TestAmoSequential(unsigned n) const
{
   auto inputs = random_literal_vector_with_size(n, 2 * n);
   auto max_var = std::max_element(inputs.begin(), inputs.end())->Variable();
   VarStore var_store(true, max_var);
   AmoSequential encoder(var_store);
   logs.TLog(1) << "Testing " << encoder.Description() << " on " << n
                << " variables" << std::endl;
   logs.TLog(3) << "Encoding: " << std::endl;
   PCEncoding enc(encoder.Encode(inputs), max_var, true);
   logs.TLog(3) << enc << std::endl;
   return TestSize(n, enc, encoder) && TestAmoEncoding(enc, inputs);
}

bool
SUBool::TestCardinal::TestExOneSplitting(unsigned n) const
{
   auto inputs = random_literal_vector_with_size(n, 2 * n);
   auto max_var = std::max_element(inputs.begin(), inputs.end())->Variable();
   VarStore var_store(true, max_var);
   ExOneSplitting encoder(var_store);
   logs.TLog(1) << "Testing " << encoder.Description() << " on " << n
                << " variables" << std::endl;
   logs.TLog(3) << "Encoding: " << std::endl;
   PCEncoding enc(encoder.Encode(inputs), max_var, true);
   logs.TLog(3) << enc << std::endl;
   return TestSize(n, enc, encoder) && TestExOneEncoding(enc, inputs);
}

bool
SUBool::TestCardinal::RunTests() const
{
   return TestAmoRepresentation(20) && TestAmoSequential(20)
          && TestExOneRepresentation(20) && TestExOneSplitting(20);
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::logs.mVerbLevel = 3;
   return !SUBool::TestCardinal().RunTests();
}
