/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file test_cnfutil.cpp
 * Contains tests for various functions defined in cnfutil.
 */

#include "test_cnfutil.h"
#include "cnfcompare.h"
#include "min_model.h"
#include "random_utils.h"

bool
SUBool::TestCNFUtil::TestMinimalSubimplicantCNF(const CNF &cnf,
                                                const Assignment &model) const
{
   std::cout << "Testing minimal SubimplicantCNF" << std::endl;
   // std::cout << "cnf: " << std::endl;
   // std::cout << cnf << std::endl;
   std::cout << "model: " << model << std::endl;
   PartialAssignment pv_model(model);
   auto val_in = cnf.Value(pv_model);
   std::cout << "value: " << val_in << std::endl;
   unsigned freed_count = 0;
   PartialAssignment alpha =
       minimal_subimplicant(cnf, pv_model, {}, &freed_count);
   std::cout << "Minimal subimplicant: " << alpha << std::endl;
   std::cout << "... freed count: " << freed_count << std::endl;
   auto val_out = cnf.Value(alpha);
   std::cout << "value on subimplicant: " << val_out;
   if (val_in != val_out)
   {
      std::cout << " [FAILED]" << std::endl;
      return false;
   }
   std::cout << std::endl;
   // Check minimality
   for (auto &v : alpha)
   {
      if (v != PartialValue::UNDEF)
      {
         auto vp = v;
         v = PartialValue::UNDEF;
         if (cnf.Value(alpha) != PartialValue::UNDEF)
         {
            std::cout << "implicant not minimal [FAILED]" << std::endl;
            std::cout << "... alpha:" << alpha << std::endl;
            std::cout << "... value:" << cnf.Value(alpha) << std::endl;
            return false;
         }
         v = vp;
      }
   }
   std::cout << "implicant is minimal [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCNFUtil::TestMinimalSubimplicantCNFFixed(
    const CNF &cnf, const Assignment &model) const
{
   std::cout << "Testing minimal SubimplicantCNF with fixed variables"
             << std::endl;
   // std::cout<<"cnf: "<<std::endl;
   // std::cout<<cnf<<std::endl;
   std::cout << "model: " << model << std::endl;
   PartialAssignment pv_model(model);
   auto val_in = cnf.Value(pv_model);
   std::cout << "value: " << val_in << std::endl;
   std::vector<bool> fixed(cnf.MaxVariable());
   std::vector<unsigned> fixed_vars;
   std::cout << "fixed variables:";
   for (unsigned i = 0; i < fixed.size(); ++i)
   {
      fixed[i] = (sub_rand() % 3 == 0);
      if (fixed[i])
      {
         fixed_vars.push_back(i + 1);
         std::cout << ' ' << i + 1;
      }
   }
   std::cout << std::endl;
   unsigned freed_count = 0;
   PartialAssignment alpha =
       minimal_subimplicant(cnf, pv_model, fixed_vars, &freed_count);
   std::cout << "Minimal subimplicant: " << alpha << std::endl;
   std::cout << "... freed count: " << freed_count << std::endl;
   auto val_out = cnf.Value(alpha);

   std::cout << "value on subimplicant: " << val_out;
   if (val_in != val_out)
   {
      std::cout << " [FAILED]";
      return false;
   }
   std::cout << std::endl;
   // Check minimality
   for (unsigned var = 1; var <= alpha.Size(); ++var)
   {
      if (alpha[var] == PartialValue::UNDEF)
      {
         if (fixed[var - 1])
         {
            std::cout << "Fixed variable is unset [FAILED]" << std::endl;
            return false;
         }
      }
      else if (!fixed[var - 1])
      {
         auto vp = alpha[var];
         alpha[var] = PartialValue::UNDEF;
         if (cnf.Value(alpha) != PartialValue::UNDEF)
         {
            std::cout << "implicant not minimal [FAILED]" << std::endl;
            std::cout << "... alpha:" << alpha << std::endl;
            std::cout << "... value:" << cnf.Value(alpha) << std::endl;
            return false;
         }
         alpha[var] = vp;
      }
   }
   std::cout << "implicant is minimal [OK]" << std::endl;
   return true;
}

bool
SUBool::TestCNFUtil::TestMinimalSubimplicant(unsigned n_vars,
                                             unsigned n_clauses,
                                             unsigned n_tests) const
{
   bool success = true;
   while (success && n_tests > 0)
   {
      auto cnf_model = random_sat_cnf(n_clauses, n_vars);
      success = TestMinimalSubimplicantCNF(cnf_model.first, cnf_model.second);
      --n_tests;
   }
   return success;
}

bool
SUBool::TestCNFUtil::TestMinimalSubimplicantFixed(unsigned n_vars,
                                                  unsigned n_clauses,
                                                  unsigned n_tests) const
{
   bool success = true;
   while (success && n_tests > 0)
   {
      auto cnf_model = random_sat_cnf(n_clauses, n_vars);
      success =
          TestMinimalSubimplicantCNFFixed(cnf_model.first, cnf_model.second);
      --n_tests;
   }
   return success;
}

bool
SUBool::TestCNFUtil::TestPrimeSubimplicate(const CNF &cnf,
                                           const Clause &clause) const
{
   std::cout << "Testing clause: " << clause << std::endl;
   auto prime = prime_subimplicate(cnf, clause);
   auto impl = is_implicate(cnf, clause);
   if (!impl)
   {
      if (prime)
      {
         std::cout << "... none was not returned although the clause is not an "
                      "implicate [FAILED]"
                   << std::endl;
         return false;
      }
      std::cout << "... clause is not implicate and none was returned [OK]"
                << std::endl;
      return true;
   }
   if (!prime)
   {
      std::cout << "... none was returned althought the clause is an implicate "
                   "[FAILED]"
                << std::endl;
      return false;
   }
   std::cout << "... prime: " << *prime << std::endl;
   auto sc = prime->SubsetCompare(clause);
   if (sc != SetCompareResult::SUBSET && sc != SetCompareResult::EQUAL)
   {
      std::cout << "... not subclause [FAILED]" << std::endl;
      return false;
   }
   if (!is_implicate(cnf, *prime))
   {
      std::cout << "... prime is not an implicate [FAILED]" << std::endl;
      return false;
   }
   for (unsigned i = 0; i < prime->Size(); ++i)
   {
      Clause p;
      for (unsigned j = 0; j < prime->Size(); ++j)
      {
         if (i != j)
         {
            p.PushBack((*prime)[j]);
         }
      }
      if (is_implicate(cnf, p))
      {
         std::cout << "... it is not prime [FAILED]" << std::endl;
         std::cout << cnf << std::endl;
         std::cout << "clause: " << clause << std::endl;
         std::cout << "prime: " << *prime << std::endl;
         std::cout << "subimplicate of prime: " << p << std::endl;
         return false;
      }
   }
   return true;
}

bool
SUBool::TestCNFUtil::TestPrimeSubimplicates(unsigned n_rounds,
                                            unsigned n_checks_per_round,
                                            unsigned n_vars,
                                            unsigned n_clauses) const
{
   std::cout << "Testing prime subimplicate" << std::endl;
   for (unsigned i = 0; i < n_rounds; ++i)
   {
      auto cnf = random_k_cnf(n_clauses, 4, n_vars);
      for (unsigned j = 0; j < n_checks_per_round; ++j)
      {
         Clause clause(random_literal_vector_with_size(n_vars / 3, n_vars));
         while (!is_implicate(cnf, clause))
         {
            clause = random_literal_vector_with_size(n_vars / 3, n_vars);
         }
         if (!TestPrimeSubimplicate(cnf, clause))
         {
            std::cout << "[FAILED]" << std::endl;
            return false;
         }
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestCNFUtil::TestPrimeCNF(const CNF &cnf) const
{
   std::cout << "Testing a prime CNF computation (p cnf " << cnf.MaxVariable()
             << " " << cnf.Size() << ")" << std::endl;
   auto prime = prime_cnf(cnf);
   std::cout << "Prime cnf (p cnf " << prime.MaxVariable() << " "
             << prime.Size() << ")" << std::endl;
   if (!cnf_equal(cnf, prime))
   {
      std::cout << "[FAILED] CNFs are not equal" << std::endl;
      return false;
   }
   for (const auto &clause : prime)
   {
      for (unsigned i = 0; i < clause.Size(); ++i)
      {
         Clause c;
         for (unsigned j = 0; j < clause.Size(); ++j)
         {
            if (i != j)
            {
               c.PushBack(clause[j]);
            }
         }
         if (is_implicate(prime, c))
         {
            std::cout << "[FAILED] clause not prime" << std::endl;
            return false;
         }
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestCNFUtil::TestPrimeRandomCNFs(unsigned n_rounds, unsigned n_vars,
                                         unsigned n_clauses) const
{
   bool success = true;
   for (; success && n_rounds > 0; --n_rounds)
   {
      success = TestPrimeCNF(random_cnf(n_clauses, n_vars));
   }
   return success;
}

bool
SUBool::TestCNFUtil::RunTests() const
{
   bool success = true;
   success = TestMinimalSubimplicant(10, 20, 20) && success;
   success = TestMinimalSubimplicantFixed(10, 20, 20) && success;
   success = TestPrimeSubimplicates(100, 10, 60, 200) && success;
   success = TestPrimeRandomCNFs(100, 40, 80) && success;
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   SUBool::logs.mVerbLevel = 4;
   srand(time(NULL));
   SUBool::TestCNFUtil test;
   return !test.RunTests();
}
