/*
 * Project SUBool
 * Petr Kucera, 2021
 */
/**@file test_dimacs.cpp
 * Contains tests for DIMACS file parsing (defined in dimacs.h).
 */

#include "test_dimacs.h"
#include "dimacs.h"
#include "logstream.h"

const char *dimacs_simple = "c\n"
                            "c start with comments\n"
                            "c\n"
                            "c\n"
                            "p cnf 5 3\n"
                            "1 -5 4 0\n"
                            "c comment in the middle\n"
                            "-1 5 3 4 0 "
                            "-3 -4 0\n"
                            "c at the end and there is an empty line below\n";

const SUBool::CNF cnf_simple({SUBool::Clause({SUBool::Literal(1, true),
                                  SUBool::Literal(5, false),
                                  SUBool::Literal(4, true)}),
    SUBool::Clause({SUBool::Literal(1, false), SUBool::Literal(5, true),
        SUBool::Literal(3, true), SUBool::Literal(4, true)}),
    SUBool::Clause({SUBool::Literal(3, false), SUBool::Literal(4, false)})});

bool
SUBool::TestDIMACS::TestSkipBlanks() const
{
   std::cout << "Testing Dimacs::skip_blanks" << std::endl;
   auto skip_blanks = [](auto begin, auto end)
   { return Dimacs::skip_blanks(begin, end); };
   return TestSkip("   \t  space    ", std::make_optional('s'), skip_blanks)
          && TestSkip("   \n  space    ", std::make_optional('\n'), skip_blanks)
          && TestSkip("   \t      ", {}, skip_blanks);
}

bool
SUBool::TestDIMACS::TestSkipSpaces() const
{
   std::cout << "Testing Dimacs::skip_spaces" << std::endl;
   auto skip_spaces = [](auto begin, auto end)
   { return Dimacs::skip_spaces(begin, end); };
   return TestSkip("   \t  space    ", std::make_optional('s'), skip_spaces)
          && TestSkip("   \n  space    ", std::make_optional('s'), skip_spaces)
          && TestSkip("   \t  \n    ", {}, skip_spaces);
}

bool
SUBool::TestDIMACS::TestSkipCommentLines() const
{
   std::cout << "Testing Dimacs::skip_comment_lines" << std::endl;
   auto skip_spaces = [](auto begin, auto end)
   { return Dimacs::skip_comment_lines(begin, end); };
   return TestSkip(
              "   \n c space \n  X    ", std::make_optional('X'), skip_spaces)
          && TestSkip(" c  \n c  space \n   ", {}, skip_spaces);
   return true;
}

bool
SUBool::TestDIMACS::TestParseLiteralSingle(const std::string &s, bool zero,
    std::optional<Literal> lit, bool ok, std::optional<char> c) const
{
   std::cout << "Testing parse_literal with string \"" << s << "\""
             << std::endl;
   std::stringstream ss(s);
   std::istreambuf_iterator<char> begin(ss), end;
   try
   {
      auto parsed_literal = Dimacs::parse_literal(begin, end);
      if (!ok)
      {
         std::cout << "[FAILED] Parse should have failed, but did not"
                   << std::endl;
         return false;
      }
      if (parsed_literal.lit != lit)
      {
         std::cout << "[FAILED] Different parsed literal (expected " << lit
                   << ", received " << parsed_literal.lit << ")" << std::endl;
         return false;
      }
      if (parsed_literal.zero != zero)
      {
         std::cout << "[FAILED] Zero status is different (expected "
                   << bool_to_string(zero) << ", received "
                   << bool_to_string(parsed_literal.zero) << ")" << std::endl;
      }
      if (!c)
      {
         if (parsed_literal.next_pos == end)
         {
            std::cout << "[OK]" << std::endl;
            return true;
         }
         std::cout << "[FAILED] Unexpected characters remain in the input"
                   << std::endl;
      }
      if (parsed_literal.next_pos == end)
      {
         std::cout << "[FAILED] Unexpected end of the input" << std::endl;
         return false;
      }
      if (*c != *parsed_literal.next_pos)
      {
         std::cout << "[FAILED] Unexpected character (expected " << *c
                   << ", received " << *parsed_literal.next_pos << ")"
                   << std::endl;
         return false;
      }
   }
   catch (const ParseException &e)
   {
      if (ok)
      {
         std::cout << "[FAILED] Unexpected error: " << e.what() << std::endl;
         return false;
      }
      std::cout << "[OK] Correctly failed: " << e.what() << std::endl;
      return true;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestDIMACS::TestParseLiteral() const
{
   bool succ = TestParseLiteralSingle(
       "  \t \n  -20X", false, Literal(20, false), false, 'X');
   succ = succ && TestParseLiteralSingle("  \t \n X", false, {}, false, 'X');
   succ = succ && TestParseLiteralSingle("   0 X ", true, {}, true, ' ');
   succ = succ && TestParseLiteralSingle("   -0 X ", true, {}, false, ' ');
   succ = succ
          && TestParseLiteralSingle(
              "   10 X ", false, Literal(10, true), true, ' ');
   succ = succ
          && TestParseLiteralSingle(
              "   -10 X ", false, Literal(10, false), true, ' ');
   return succ;
}

bool
SUBool::TestDIMACS::TestParseClauseSingle(const std::string &s,
    std::optional<Clause> clause, bool zero, bool ok,
    std::optional<char> c) const
{
   std::cout << "Testing parse_clause with string \"" << s << "\"" << std::endl;
   std::stringstream ss(s);
   std::istreambuf_iterator<char> begin(ss), end;
   try
   {
      auto parsed_clause = Dimacs::parse_clause(begin, end);
      if (!ok)
      {
         std::cout << "[FAILED] Parse should have failed, but did not"
                   << std::endl;
         return false;
      }
      if (parsed_clause.clause != clause)
      {
         std::cout << "[FAILED] Different parsed clause (expected " << clause
                   << ", received " << parsed_clause.clause << ")" << std::endl;
         return false;
      }
      if (parsed_clause.zero != zero)
      {
         std::cout << "[FAILED] Zero status is different (expected "
                   << bool_to_string(zero) << ", received "
                   << bool_to_string(parsed_clause.zero) << ")" << std::endl;
      }
      if (!c)
      {
         if (parsed_clause.next_pos == end)
         {
            std::cout << "[OK]" << std::endl;
            return true;
         }
         std::cout << "[FAILED] Unexpected characters remain in the input"
                   << std::endl;
      }
      if (parsed_clause.next_pos == end)
      {
         std::cout << "[FAILED] Unexpected end of the input" << std::endl;
         return false;
      }
      if (*c != *parsed_clause.next_pos)
      {
         std::cout << "[FAILED] Unexpected character (expected " << *c
                   << ", received " << *parsed_clause.next_pos << ")"
                   << std::endl;
         return false;
      }
   }
   catch (const ParseException &e)
   {
      if (ok)
      {
         std::cout << "[FAILED] Unexpected error: " << e.what() << std::endl;
         return false;
      }
      std::cout << "[OK] Correctly failed: " << e.what() << std::endl;
      return true;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestDIMACS::TestParseClause() const
{
   return TestParseClauseSingle("14 16 -7 11 0",
              Clause({Literal(14, true), Literal(16, true), Literal(7, false),
                  Literal(11, true)}),
              true, true, {})
          && TestParseClauseSingle("14 16 -7 11 0\n XX",
              Clause({Literal(14, true), Literal(16, true), Literal(7, false),
                  Literal(11, true)}),
              true, true, '\n')
          && TestParseClauseSingle("14 16 -7 11 \n ",
              Clause({Literal(14, true), Literal(16, true), Literal(7, false),
                  Literal(11, true)}),
              false, true, {})
          && TestParseClauseSingle("14 16 -0 11 \n ",
              Clause({Literal(14, true), Literal(16, true), Literal(7, false),
                  Literal(11, true)}),
              false, false, {})
          && TestParseClauseSingle("   \n 14 16\n  -16 11 \n ",
              Clause({Literal(14, true), Literal(16, true), Literal(16, false),
                  Literal(11, true)}),
              false, true, {});
}

bool
SUBool::TestDIMACS::TestParseCNFSingle(
    const std::string &s, std::optional<CNF> cnf) const
{
   std::cout << "Testing single parse CNF" << std::endl;
   auto parsed_cnf = Dimacs::parse_cnf(s.begin(), s.end());
   if (parsed_cnf != cnf)
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestDIMACS::TestParseCNF() const
{
   return TestParseCNFSingle(dimacs_simple, cnf_simple);
}

bool
SUBool::TestDIMACS::RunTests() const
{
   return TestSkipBlanks() && TestSkipSpaces() && TestSkipCommentLines()
          && TestParseLiteral() && TestParseClause() && TestParseCNF();
}

int
main(int /*argc*/, char * /*argv*/[])
{
   SUBool::logs.mVerbLevel = 4;
   srand(time(NULL));
   SUBool::TestDIMACS test;
   return !test.RunTests();
}
