/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file test_dnnf.cpp
 * Contains tests for DNNF encoding and other related stuff.
 */

#include <optional>
#include <sstream>

#include "dimacs.h"
#include "dnnf_enc.h"
#include "encoding.h"
#include "logstream.h"
#include "pcchecker.h"
#include "test_common.h"
#include "test_dnnf.h"
#include "util.h"

const char *example_nnf_str = "nnf 21 28 4\n"
                              "L 1\n"         // 0
                              "L 2\n"         // 1
                              "L 3\n"         // 2
                              "L 4\n"         // 3
                              "L -2\n"        // 4
                              "L -3\n"        // 5
                              "L -4\n"        // 6
                              "L -1\n"        // 7
                              "O 1 2 0 7\n"   // 8
                              "O 2 2 1 4\n"   // 9
                              "O 3 2 2 5\n"   // 10
                              "A 2 4 2\n"     // 11
                              "A 2 1 10\n"    // 12
                              "O 2 2 12 11\n" // 13
                              "A 2 0 3\n"     // 14
                              "A 2 13 14\n"   // 15
                              "A 3 0 4 5\n"   // 16
                              "A 3 2 9 8\n"   // 17
                              "O 3 2 17 16\n" // 18
                              "A 2 6 18\n"    // 19
                              "O 4 2 15 19";  // 20
const char *example_cnf_str = "p cnf 4 4\n"
                              "1 3 0\n"
                              "1 -4 0\n"
                              "2 3 -4 0\n"
                              "-2 3 4 0";
const char *example_decision_nnf_str = "nnf 19 26 4\n"
                                       "L 1\n"         // 0
                                       "L 2\n"         // 1
                                       "L 3\n"         // 2
                                       "L -1\n"        // 3
                                       "L -2\n"        // 4
                                       "L -3\n"        // 5
                                       "O 1 2 0 3\n"   // 6 | 1 -1
                                       "O 2 2 1 4\n"   // 7 | 2 -2
                                       "O 3 2 2 5\n"   // 8 | 3 -3
                                       "A 2 1 8\n"     // 9
                                       "A 2 2 7\n"     // 10
                                       "O 0 2 9 10\n"  // 11
                                       "A 2 0 11\n"    // 12
                                       "A 2 2 6\n"     // 13
                                       "A 2 0 8\n"     // 14
                                       "O 0 2 14 13\n" // 15
                                       "A 2 2 6\n"     // 16
                                       "D 2 16 15\n"   // 17
                                       "D 4 12 17";    // 18

const char *example_cnf_monotone_x = "p cnf 3 2\n"
                                     "1 0\n"
                                     "2 3 0";

const char *example_dnnf_monotone_x = "nnf 11 11 3\n"
                                      "L 1\n"       // 0
                                      "L 2\n"       // 1
                                      "L 3\n"       // 2
                                      "L -1\n"      // 3
                                      "L -2\n"      // 4
                                      "L -3\n"      // 5
                                      "O 2 2 1 4\n" // 6
                                      "O 2 2 2 5\n" // 7
                                      "A 3 0 1 7\n" // 8
                                      "A 3 0 2 6\n" // 9
                                      "O 0 2 8 9";  // 10

const char *horn_triangle_03_cnf_str = "p cnf 10 6\n"
                                       "1 -2 -3 0\n"
                                       "2 -4 -5 0\n"
                                       "3 -5 -6 0\n"
                                       "4 -7 -8 0\n"
                                       "5 -8 -9 0\n"
                                       "6 -9 -10 0\n";
const char *horn_triangle_03_nnf_str = "nnf 82 66 10\n"
                                       "L -5\n"
                                       "L -1\n"
                                       "L 3\n"
                                       "L -3\n"
                                       "O 0 2 2 3\n"
                                       "L -2\n"
                                       "A 2 4 5\n"
                                       "L 2\n"
                                       "A 2 7 3\n"
                                       "O 0 2 6 8\n"
                                       "A 2 1 9\n"
                                       "L 1\n"
                                       "O 0 2 2 3\n"
                                       "O 0 2 7 5\n"
                                       "A 3 11 12 13\n"
                                       "O 1 2 10 14\n"
                                       "L -8\n"
                                       "L -6\n"
                                       "L 9\n"
                                       "L -10\n"
                                       "A 2 18 19\n"
                                       "L -9\n"
                                       "L 10\n"
                                       "O 0 2 22 19\n"
                                       "A 2 21 23\n"
                                       "O 0 2 20 24\n"
                                       "A 2 17 25\n"
                                       "L 6\n"
                                       "O 0 2 18 21\n"
                                       "O 0 2 22 19\n"
                                       "A 3 27 28 29\n"
                                       "O 6 2 26 30\n"
                                       "A 2 16 31\n"
                                       "L 4\n"
                                       "L -4\n"
                                       "O 0 2 33 34\n"
                                       "L 7\n"
                                       "L -7\n"
                                       "O 0 2 36 37\n"
                                       "A 3 32 35 38\n"
                                       "L 8\n"
                                       "A 2 34 37\n"
                                       "O 0 2 36 37\n"
                                       "A 2 33 42\n"
                                       "O 4 2 41 43\n"
                                       "A 3 40 21 44\n"
                                       "O 0 2 27 17\n"
                                       "O 0 2 22 19\n"
                                       "A 3 45 46 47\n"
                                       "O 0 2 39 48\n"
                                       "A 3 0 15 49\n"
                                       "L 5\n"
                                       "A 2 16 36\n"
                                       "O 0 2 16 40\n"
                                       "A 2 37 53\n"
                                       "O 0 2 52 54\n"
                                       "A 2 34 55\n"
                                       "O 0 2 36 37\n"
                                       "O 0 2 16 40\n"
                                       "A 3 33 57 58\n"
                                       "O 4 2 56 59\n"
                                       "A 3 11 2 31\n"
                                       "A 3 3 17 25\n"
                                       "L -1\n"
                                       "O 0 2 63 11\n"
                                       "A 2 62 64\n"
                                       "O 0 2 61 65\n"
                                       "A 3 7 60 66\n"
                                       "A 2 17 25\n"
                                       "O 0 2 2 3\n"
                                       "A 2 68 69\n"
                                       "A 2 2 27\n"
                                       "O 0 2 18 21\n"
                                       "O 0 2 22 19\n"
                                       "A 3 71 72 73\n"
                                       "O 0 2 70 74\n"
                                       "A 4 5 34 55 75\n"
                                       "O 0 2 63 11\n"
                                       "A 2 76 77\n"
                                       "O 0 2 67 78\n"
                                       "A 2 51 79\n"
                                       "O 0 2 50 80\n";

bool
SUBool::TestDnnf::TestDnnfInputUrcEncoder() const
{
   std::cout << "Testing DNNF I-URC encoder." << std::endl;
   std::stringstream nnf_stream(example_nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto cnf = Dimacs::parse_cnf(example_cnf_str);
   SATEncoding sat_enc(cnf, cnf.MaxVariable());
   auto enc = DnnfInputUrcEncoder::URCEncode(nnf);
   std::cout << "Constructed encoding:" << std::endl;
   std::cout << enc << std::endl;
   return test_encoding_consistency_full(enc, sat_enc)
          && test_encoding_consistency_partial(enc, sat_enc)
          && test_encoding_propagation(enc, sat_enc);
}

bool
SUBool::TestDnnf::TestSmoothDnnfInputPcEncoder() const
{
   std::cout << "Testing smooth DNNF PC encoder." << std::endl;
   std::stringstream nnf_stream(example_nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto cnf = Dimacs::parse_cnf(example_cnf_str);
   SATEncoding sat_enc(cnf, cnf.MaxVariable());
   auto enc = SmoothDnnfInputPcEncoder::PCEncode(nnf);
   std::cout << "Constructed encoding:" << std::endl;
   std::cout << enc << std::endl;
   return test_encoding_consistency_full(enc, sat_enc)
          && test_encoding_consistency_partial(enc, sat_enc)
          && test_encoding_propagation(enc, sat_enc);
}

bool
SUBool::TestDnnf::TestSmoothDnnfPcEncoder(LEVEL_FUNC level_func) const
{
   std::cout << "Testing smooth DNNF PC encoder with level_func:"
             << static_cast<unsigned>(level_func) << std::endl;
   std::stringstream nnf_stream(example_nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto cnf = Dimacs::parse_cnf(example_cnf_str);
   SATEncoding sat_enc(cnf, cnf.MaxVariable());
   auto enc = encode_smooth_dnnf_pc(nnf, level_func, 3);
   std::cout << "Constructed encoding:" << std::endl;
   std::cout << enc << std::endl;
   if (!test_encoding_consistency_full(enc, sat_enc)
       || !test_encoding_consistency_partial(enc, sat_enc)
       || !test_encoding_propagation(enc, sat_enc))
   {
      return false;
   }
   std::cout << "Checking if the encoding is PC" << std::endl;
   if (is_pc(enc.Cnf()))
   {
      std::cout << "... [OK]" << std::endl;
      return true;
   }
   std::cout << "... [FAILED]" << std::endl;
   return false;
}

bool
SUBool::TestDnnf::TestSmoothDnnfUrcEncoder(LEVEL_FUNC level_func) const
{
   std::cout << "Testing DNNF URC encoder with level_func:"
             << static_cast<unsigned>(level_func) << std::endl;
   std::stringstream nnf_stream(example_nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto cnf = Dimacs::parse_cnf(example_cnf_str);
   SATEncoding sat_enc(cnf, cnf.MaxVariable());
   auto enc = encode_smooth_dnnf_urc(nnf, level_func, 3);
   std::cout << "Constructed encoding:" << std::endl;
   std::cout << enc << std::endl;
   if (!test_encoding_consistency_full(enc, sat_enc)
       || !test_encoding_consistency_partial(enc, sat_enc)
       || !test_encoding_propagation(enc, sat_enc))
   {
      return false;
   }
   std::cout << "Checking if the encoding is URC" << std::endl;
   if (is_urc(enc.Cnf()))
   {
      std::cout << "... [OK]" << std::endl;
      return true;
   }
   std::cout << "... [FAILED]" << std::endl;
   return false;
}

bool
SUBool::TestDnnf::TestEliminationOnPCE(
    const char *nnf_str, const char *cnf_str) const
{
   std::cout << "Testing smooth DNNF PC encoder with DP elimination."
             << std::endl;
   std::stringstream nnf_stream(nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto cnf = Dimacs::parse_cnf(cnf_str);
   auto enc = SmoothDnnfInputPcEncoder::PCEncode(nnf);
   auto sat_enc = SATEncoding(cnf, cnf.MaxVariable());
   std::cout << "Constructed encoding:" << std::endl;
   std::cout << enc << std::endl;
   // enc.EliminateAuxVariablesWithMaxInc(1000,
   //                                   false); /*CNFEncoding::kEliminateAll);*/
   std::cout << "Encoding after elimination:" << std::endl;
   std::cout << enc << std::endl;
   return test_encoding_consistency_full(enc, sat_enc)
          && test_encoding_consistency_partial(enc, sat_enc)
          && test_encoding_propagation(enc, sat_enc);
}

bool
SUBool::TestDnnf::TestDnnfUrcWithDecisions() const
{
   std::cout << "Testing DNNF URC encoder with decisions." << std::endl;
   std::stringstream nnf_stream(example_decision_nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto cnf = Dimacs::parse_cnf(example_cnf_str);
   std::cout << "Checking equivalence";
   {
      Assignment alpha(4);
      do
      {
         if (nnf.Value(alpha) != cnf.Value(alpha))
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << ".. mismatch on " << alpha << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   auto enc = DnnfInputUrcEncoder::URCEncode(nnf);
   std::cout << "Constructed encoding:" << std::endl;
   std::cout << enc << std::endl;
   std::cout << "Testing consistency on full assignments";
   {
      Assignment alpha(4);
      do
      {
         if (cnf.Value(alpha) != enc.Consistent(alpha))
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... alpha: " << alpha << ", cnf: " << cnf.Value(alpha)
                      << ", nnf: " << nnf.Value(alpha)
                      << ", enc: " << enc.Consistent(alpha) << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   std::cout << "Testing consistency on partial assignments";
   {
      PartialAssignment alpha(4);
      do
      {
         if (enc.Consistent(alpha) != (cnf.Value(alpha) != PartialValue::FALSE))
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... alpha: " << alpha << ", cnf: " << cnf.Value(alpha)
                      << ", nnf: " << nnf.Value(alpha)
                      << ", enc: " << enc.Consistent(alpha) << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   std::cout << "Testing propagation";
   PCEncoding pcenc(cnf, cnf.MaxVariable(), false);
   {
      PartialAssignment alpha(4);
      do
      {
         auto beta_enc = enc.Propagate(alpha);
         auto beta_pcenc = pcenc.Propagate(alpha);
         if (beta_enc != beta_pcenc)
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... propagate " << alpha << ", enc: " << beta_enc
                      << ", pcenc: " << beta_pcenc << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   return true;
}

bool
SUBool::TestDnnf::TestSmoothDnnfPc(
    const char *dnnf_string, const char *cnf_string) const
{
   std::cout << "Testing Smooth DNNF PC encoder." << std::endl;
   std::stringstream nnf_stream(dnnf_string);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   std::cout << "NNF:" << std::endl;
   std::cout << nnf << std::endl;
   auto cnf = Dimacs::parse_cnf(cnf_string);
   std::cout << "Equivalent CNF:" << std::endl;
   std::cout << cnf << std::endl;
   std::cout << "Checking equivalence";
   {
      Assignment alpha(cnf.MaxVariable());
      do
      {
         if (nnf.Value(alpha) != cnf.Value(alpha))
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << ".. mismatch on " << alpha << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   auto enc = SmoothDnnfInputPcEncoder::PCEncode(nnf);
   std::cout << "Constructed encoding:" << std::endl;
   std::cout << enc << std::endl;
   std::cout << "Testing consistency on full assignments";
   {
      Assignment alpha(cnf.MaxVariable());
      do
      {
         if (cnf.Value(alpha) != enc.Consistent(alpha))
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... alpha: " << alpha << ", cnf: " << cnf.Value(alpha)
                      << ", nnf: " << nnf.Value(alpha)
                      << ", enc: " << enc.Consistent(alpha) << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   std::cout << "Testing consistency on partial assignments";
   {
      PartialAssignment alpha(cnf.MaxVariable());
      do
      {
         if (enc.Consistent(alpha) != (cnf.Value(alpha) != PartialValue::FALSE))
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... alpha: " << alpha << ", cnf: " << cnf.Value(alpha)
                      << ", nnf: " << nnf.Value(alpha)
                      << ", enc: " << enc.Consistent(alpha) << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   std::cout << "Testing propagation";
   PCEncoding pcenc(cnf, cnf.MaxVariable(), false);
   {
      PartialAssignment alpha(cnf.MaxVariable());
      do
      {
         auto beta_enc = enc.Propagate(alpha);
         auto beta_pcenc = pcenc.Propagate(alpha);
         if (beta_enc != beta_pcenc)
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... propagate " << alpha << ", enc: " << beta_enc
                      << ", pcenc: " << beta_pcenc << std::endl;
            return false;
         }
      } while (alpha.Inc());
      std::cout << " [OK]" << std::endl;
   }
   return true;
}

bool
SUBool::TestDnnf::TestSmoothDnnfPcWithDecisions() const
{
   return TestSmoothDnnfPc(example_decision_nnf_str, example_cnf_str);
}

bool
SUBool::TestDnnf::TestMonotoneX() const
{
   return TestSmoothDnnfPc(example_dnnf_monotone_x, example_cnf_monotone_x);
}

bool
SUBool::TestDnnf::TestTopologicalOrderVisitor(const char *dnnf_string) const
{
   std::cout << "Testing topological order visitor.";
   std::stringstream nnf_stream(dnnf_string);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   NnfNode::tValueCache<bool> topo;
   try
   {
      nnf.AcceptTopologicalFunction(
          [&topo](const NnfNode::tConstPointer &node)
          {
             if (node->CachedValue(topo))
             {
                throw FailureException("Second visit to a single node");
             }
             node->CacheValue(topo, true);
             for (auto child : *node)
             {
                if (child->CachedValue(topo))
                {
                   throw FailureException("Not in topological order");
                }
             }
          });
   }
   catch (FailureException &e)
   {
      std::cout << std::endl;
      std::cout << e.what() << " [FAILED]" << std::endl;
      return false;
   }
   if (topo.size() != nnf.NodeCount())
   {
      std::cout << std::endl;
      std::cout << "Number of visited nodes does not match the number of nodes "
                   "in the NNF. [FAILED]"
                << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   return true;
}
#if 0
bool
SUBool::TestDnnf::TestEncodingConsistencyFull(const CNFEncoding &enc_test,
                                              const CNFEncoding &enc_ref) const
{
   std::cout << "Checking consistency of an encoding on full assignments";
   Assignment alpha(enc_ref.MaxInputVariable());
   do
   {
      if (enc_ref.Consistent(alpha) != enc_test.Consistent(alpha))
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestDnnf::TestEncodingConsistencyPartial(
    const CNFEncoding &enc_test, const CNFEncoding &enc_ref) const
{
   std::cout << "Checking consistency of a na encoding on partial assignments";
   PartialAssignment alpha(enc_ref.MaxInputVariable());
   do
   {
      if (enc_ref.Consistent(alpha) != enc_test.Consistent(alpha))
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestDnnf::TestEncodingPropagation(const CNFEncoding &enc_test,
                                          const CNFEncoding &enc_ref) const
{
   std::cout << "Checking propagation of an encoding on partial assignments";
   PartialAssignment alpha(enc_ref.MaxInputVariable());
   do
   {
      auto beta_ref = enc_ref.Propagate(alpha);
      auto beta_test = enc_test.Propagate(alpha);
      if (!beta_ref)
      {
         if (beta_test)
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... Propagation returns an assignment on an "
                         "unsatisfiable formula"
                      << std::endl;
            return false;
         }
         continue;
      }
      if (!beta_test)
      {
         std::cout << " [FAILED]" << std::endl;
         std::cout << "... Propagation is inconsistent on a satisfiable formula"
                   << std::endl;
         return false;
      }
      for (unsigned var = 1; var <= beta_ref->Size(); ++var)
      {
         if ((*beta_ref)[var] != (*beta_test)[var])
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... Propagated literals mismatch" << std::endl;
         }
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}
#endif

bool
SUBool::TestDnnf::TestDnnfEquivalence(
    const char *dnnf_str, const char *cnf_str) const
{
   std::cout << "Testing if a given DNNF (" << dnnf_str
             << ") represents a given CNF (" << cnf_str << ")." << std::endl;
   std::stringstream nnf_stream(dnnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto cnf = Dimacs::parse_cnf(cnf_str);
   std::cout << "NNF: " << nnf << std::endl;
   Assignment alpha(cnf.MaxVariable());
   do
   {
      if (cnf.Value(alpha) != nnf.Value(alpha))
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestDnnf::RunTests() const
{
   bool success = true;

   success &= TestDnnfInputUrcEncoder();
   success &= TestSmoothDnnfInputPcEncoder();
   success &= TestSmoothDnnfUrcEncoder(LEVEL_FUNC::FROM_ROOT);
   success &= TestSmoothDnnfPcEncoder(LEVEL_FUNC::FROM_ROOT);
   success &= TestSmoothDnnfUrcEncoder(LEVEL_FUNC::FROM_LEAVES);
   success &= TestSmoothDnnfPcEncoder(LEVEL_FUNC::FROM_LEAVES);
   success &= TestEliminationOnPCE(example_nnf_str, example_cnf_str);
   success &=
       TestDnnfEquivalence(horn_triangle_03_nnf_str, horn_triangle_03_cnf_str);

   success &=
       TestEliminationOnPCE(horn_triangle_03_nnf_str, horn_triangle_03_cnf_str);
   success &= TestDnnfUrcWithDecisions();
   success &= TestSmoothDnnfPcWithDecisions();
   success &= TestMonotoneX();
   success &= TestTopologicalOrderVisitor(example_nnf_str);
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::logs.mVerbLevel = 3;
   return !SUBool::TestDnnf().RunTests();
}
