/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file test_empower2.cpp
 * Contains tests of OneProveM.
 */

#include <chrono>
#include <sstream>

#include "cnfcompare.h"
#include "empower.h"
#include "empower2.h"
#include "random_utils.h"
#include "test_empower2.h"
#include "util.h"

bool
SUBool::TestOneProveM::TestSingleRandomCNF(const CNF &cnf,
                                           const unsigned clause_rounds,
                                           const unsigned clause_size) const
{
   std::cout << "Testing " << cnf << std::endl;
   OneProve op(&cnf);
   OneProveM op_m(cnf);
   for (unsigned round = 0; round < clause_rounds; ++round)
   {
      Clause clause(
          random_literal_vector_with_size(clause_size, cnf.MaxVariable()));
      if (op.IsOneProvable(clause) != op_m.IsOneProvable(clause))
      {
         return false;
      }
      if (op.IsEmpowering(clause) != op_m.IsEmpowering(clause))
      {
         return false;
      }
   }
   return TestMinimizeFull(cnf) && TestMinimizeIncremental(cnf, false)
          && TestMinimizeIncremental(cnf, true);
}

bool
SUBool::TestOneProveM::TestMinimizeIncremental(const CNF &cnf,
                                               bool reprocess) const
{
   OneProveM op_m;
   op_m.AddWithIncrementalMinimization(cnf, reprocess);
   auto min_cnf = op_m.ExtractCNF();
   if (!cnf_equal(cnf, min_cnf))
   {
      std::cout << "[FAILED] not equal" << std::endl;
      return false;
   }
   if (!cnf_equal_absorb(cnf, min_cnf))
   {
      std::cout << "[FAILED] TestMinimizeIncremental: not equal with respect "
                   "to absorbtions"
                << std::endl;
      std::cout << "cnf: " << cnf << std::endl;
      std::cout << "min_cnf: " << min_cnf << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestOneProveM::TestMinimizeFull(const CNF &cnf) const
{
   OneProveM op_m(cnf);
   op_m.MinimizeFull();
   auto min_cnf = op_m.ExtractCNF();
   if (!cnf_equal(cnf, min_cnf))
   {
      std::cout << "[FAILED] not equal" << std::endl;
      return false;
   }
   auto a_min_cnf = empower_minimize_full(cnf);
   if (!cnf_absorbed(min_cnf, cnf))
   {
      std::cout << "[FAILED] TestMinimizeFull: cnf not absorbed by min_cnf"
                << std::endl;
      std::cout << "cnf: " << cnf << std::endl;
      std::cout << "min_cnf: " << min_cnf << std::endl;
      std::cout << "a_min_cnf: " << a_min_cnf << std::endl;
      return false;
   }
   auto m_min_cnf = empower_minimize_full(min_cnf);
   if (m_min_cnf.Size() != min_cnf.Size())
   {
      std::cout << "[FAILED] not minimal" << std::endl;
      std::cout << "min_cnf: " << min_cnf << std::endl;
      std::cout << "m_min_cnf: " << m_min_cnf << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestOneProveM::TestRandomCNF(const unsigned rounds,
                                     const unsigned n_clauses,
                                     const unsigned n_vars,
                                     const unsigned clause_rounds,
                                     const unsigned clause_size) const
{
   std::cout << "Testing empowerement on random formulas" << std::endl;
   for (unsigned round = 0; round < rounds; ++round)
   {
      auto cnf = random_sat_cnf(n_clauses, n_vars);
      if (!TestSingleRandomCNF(cnf.first, clause_rounds, clause_size))
      {
         std::cout << "[FAILED] round " << round << std::endl;
         return false;
      }
   }
   std::cout << "[ALL OK]" << std::endl;
   return true;
}

bool
SUBool::TestOneProveM::RunTests() const
{
   bool success = TestRandomCNF(100, 50, 20, 100, 5);
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   SUBool::logs.mVerbLevel = 4;
   srand(time(NULL));
   SUBool::TestOneProveM test;
   return !test.RunTests();
}
