/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file test_encoding.cpp
 * Contains tests for CNF encoding representation.
 */

#include <ctime>
#include <sstream>

#include "dimacs.h"
#include "encoding.h"
#include "glucose_bind.h"
#include "logstream.h"
#include "random_utils.h"
#include "test_common.h"
#include "test_encoding.h"

const char *dimacs = "p cnf 8 8\n"
                     "1 2 5 0\n"
                     "-5 3 4 0\n"
                     "1 2 6 0\n"
                     "-6 -3 4 0\n"
                     "1 2 7 0\n"
                     "-7 3 -4 0\n"
                     "1 2 8 0\n"
                     "-8 -3 -4 0\n";

const char *encoding = "c ! encoding pc 4\n"
                       "p cnf 17 40\n"
                       "2 -6 0\n"
                       "5 -6 0\n"
                       "-2 -7 0\n"
                       "3 -7 0\n"
                       "6 7 -8 0\n"
                       "1 -9 0\n"
                       "4 -9 0\n"
                       "8 -10 0\n"
                       "9 -10 0\n"
                       "3 -13 0\n"
                       "11 -13 0\n"
                       "12 -13 0\n"
                       "1 -14 0\n"
                       "-2 -14 0\n"
                       "-3 -14 0\n"
                       "13 14 -15 0\n"
                       "-4 -16 0\n"
                       "15 -16 0\n"
                       "10 16 -17 0\n"
                       "17 0\n"
                       "-16 17 0\n"
                       "-14 15 0\n"
                       "-15 16 0\n"
                       "-13 15 0\n"
                       "-12 13 0\n"
                       "-10 17 0\n"
                       "-9 10 0\n"
                       "-4 9 0\n"
                       "-1 9 12 14 0\n"
                       "-8 10 0\n"
                       "-11 13 0\n"
                       "-2 6 11 0\n"
                       "-7 8 0\n"
                       "2 7 11 14 0\n"
                       "1 12 0\n"
                       "-5 6 0\n"
                       "3 5 14 0\n"
                       "4 16 0\n"
                       "-3 5 7 13 0\n"
                       "-6 8 0";

bool
SUBool::TestEncoding::TestSimpleSATEncoding() const
{
   std::cout << "Testing SAT encoding" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs);
   std::cout << "Formula: " << cnf << std::endl;
   SATEncoding enc(cnf, 4);
   Assignment alpha(4, true);
   std::cout << "Checking consistency on all true assignment" << std::endl;
   if (!enc.Consistent(alpha))
   {
      std::cout << " [FAILED]" << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   alpha = Assignment(4, false);
   std::cout << "Checking consistency on all false assignment" << std::endl;
   if (enc.Consistent(alpha))
   {
      std::cout << " [FAILED]" << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   PartialAssignment beta(4);
   std::cout << "Checking consistency of empty partial assignment" << std::endl;
   if (!enc.Consistent(beta))
   {
      std::cout << " [FAILED]" << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   std::cout << "Checking propagation of empty partial assignment" << std::endl;
   auto gamma = enc.Propagate(beta);
   if (!gamma)
   {
      std::cout << " - inconsistent [FAILED]" << std::endl;
      return false;
   }
   for (auto val : *gamma)
   {
      if (val != PartialValue::UNDEF)
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
   }
   std::cout << " [OK]" << std::endl;
   std::cout << "Checking propagation of nonempty partial assignment"
             << std::endl;
   beta[2] = PartialValue::FALSE;
   gamma = enc.Propagate(beta);
   if (!gamma)
   {
      std::cout << " - inconsistent [FAILED]" << std::endl;
      return false;
   }
   if ((*gamma)[1] != PartialValue::TRUE || (*gamma)[2] != PartialValue::FALSE
       || (*gamma)[3] != PartialValue::UNDEF
       || (*gamma)[4] != PartialValue::UNDEF)
   {
      std::cout << " [FAILED]" << std::endl;
      std::cout << "propagated: " << *gamma << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestEncoding::TestSATEncodingConsistencyFull(
    const CNFEncoding &enc, const CNF &cnf) const
{
   std::cout << "Checking consistency of a SAT encoding on full assignments";
   Assignment alpha(cnf.MaxVariable());
   do
   {
      if (enc.Consistent(alpha) != cnf.Value(alpha))
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestEncoding::TestSATEncodingConsistencyPartial(
    const CNFEncoding &enc, const CNF &cnf) const
{
   std::cout << "Checking consistency of a SAT encoding on partial assignments";
   PartialAssignment alpha(cnf.MaxVariable());
   do
   {
      if (enc.Consistent(alpha) != is_consistent(cnf, alpha))
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestEncoding::TestSATEncodingPropagation(
    const CNFEncoding &enc, const CNF &cnf) const
{
   std::cout << "Checking propagation of a SAT encoding on partial assignments";
   PartialAssignment alpha(cnf.MaxVariable());
   do
   {
      auto beta = enc.Propagate(alpha);
      auto is_sat = is_consistent(cnf, alpha);
      if (!is_sat)
      {
         if (beta)
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... Propagation returns an assignment on an "
                         "unsatisfiable formula"
                      << std::endl;
            return false;
         }
         continue;
      }
      if (!beta)
      {
         std::cout << " [FAILED]" << std::endl;
         std::cout << "... Propagation is inconsistent on a satisfiable formula"
                   << std::endl;
         return false;
      }
      for (unsigned var = 1; var <= cnf.MaxVariable(); ++var)
      {
         if (alpha[var] == PartialValue::UNDEF)
         {
            alpha[var] = PartialValue::TRUE;
            bool pos = is_consistent(cnf, alpha);
            alpha[var] = PartialValue::FALSE;
            bool neg = is_consistent(cnf, alpha);
            alpha[var] = PartialValue::UNDEF;
            if (!pos && (*beta)[var] != PartialValue::FALSE)
            {
               std::cout << " [FAILED]" << std::endl;
               std::cout << "... Failed propagation of negative literal -"
                         << var << std::endl;
               return false;
            }
            if (!neg && (*beta)[var] != PartialValue::TRUE)
            {
               std::cout << " [FAILED]" << std::endl;
               std::cout << "... Failed propagation of positive literal " << var
                         << std::endl;
               return false;
            }
            if (pos && neg && (*beta)[var] != PartialValue::UNDEF)
            {
               std::cout << " [FAILED]" << std::endl;
               std::cout << "... Propagated literal "
                         << ((*beta)[var] == PartialValue::FALSE ? "-" : "")
                         << var << " which is not implied" << std::endl;
               return false;
            }
         }
         else if (alpha[var] != (*beta)[var])
         {
            std::cout << " [FAILED]" << std::endl;
            std::cout << "... The propagated partial assignment is "
                         "inconsistent with the input one."
                      << std::endl;
         }
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestEncoding::TestSATEncoding(
    unsigned n_rounds, unsigned n_clauses, unsigned n_vars) const
{
   std::cout << "Checking SAT encoding on random formulas" << std::endl;
   bool success = true;
   for (unsigned i = 0; success && i < n_rounds; ++i)
   {
      auto cnf = random_cnf(n_clauses, n_vars);
      std::cout << "...Checking formula" << std::endl;
      std::cout << cnf << std::endl;
      SATEncoding enc(cnf, cnf.MaxVariable());
      success = TestSATEncodingConsistencyFull(enc, cnf)
                && TestSATEncodingConsistencyPartial(enc, cnf)
                && TestSATEncodingPropagation(enc, cnf);
   }
   std::cout << "Checked SAT encoding on random formulas "
             << (success ? "[OK]" : "[FAILED]") << std::endl;
   return success;
}

bool
SUBool::TestEncoding::TestAssignment(const Assignment &alpha,
    const CNFEncoding &ref_encoding, const CNFEncoding &check_encoding) const
{
   if (ref_encoding.Consistent(alpha) != check_encoding.Consistent(alpha))
   {
      std::cout << "Consistency on a full assignment failed [FAILED]"
                << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestEncoding::TestPartialAssignment(const PartialAssignment &alpha,
    const CNFEncoding &ref_encoding, const CNFEncoding &check_encoding) const
{
   bool cons = ref_encoding.Consistent(alpha);
   if (cons != check_encoding.Consistent(alpha))
   {
      std::cout << "Consistency on a partial assignment failed [FAILED]"
                << std::endl;
      return false;
   }
   auto r_beta = ref_encoding.Propagate(alpha);
   auto c_beta = ref_encoding.Propagate(alpha);
   if (!cons)
   {
      if (c_beta)
      {
         std::cout << "Propagation on an inconsistent partial assignment "
                      "failed [FAILED]"
                   << std::endl;
         return false;
      }
   }
   else if (*r_beta != *c_beta)
   {
      std::cout
          << "Propagation on a consistent partial assignment failed [FAILED]"
          << std::endl;
   }
   return true;
}

bool
SUBool::TestEncoding::TestURCEncoding(unsigned n_rounds) const
{
   std::cout << "Checking URC encoding on random Horn CNFs" << std::endl;
   const auto n_vars = 40;
   const auto n_clauses = 2 * n_vars;
   const auto max_clause_size = n_vars / 4;
   const auto n_input_vars = n_vars / 2;
   bool success = true;
   for (unsigned i = 0; success && i < n_rounds; ++i)
   {
      auto cnf = random_horn_cnf(n_clauses, max_clause_size, n_vars, false);
      URCEncoding u_enc(cnf, n_input_vars, false);
      SATEncoding s_enc(cnf, n_input_vars);
      success &= TestAssignment(Assignment(n_input_vars, false), s_enc, u_enc);
      success &= TestAssignment(Assignment(n_input_vars, true), s_enc, u_enc);
      success &=
          TestPartialAssignment(PartialAssignment(n_input_vars), s_enc, u_enc);
      for (unsigned j = 0; success && j < 100; ++j)
      {
         success &= TestAssignment(Assignment(n_input_vars, 0.5), s_enc, u_enc);
         success &= TestPartialAssignment(
             PartialAssignment(n_input_vars, 0.2, 0.2), s_enc, u_enc);
      }
   }
   if (success)
   {
      std::cout << "All [OK]" << std::endl;
   }
   return success;
}

bool
SUBool::TestEncoding::TestEliminateAuxVariable(unsigned n_rounds) const
{
   std::cout << "Checking elimination of auxiliary variable" << std::endl;
   const auto n_vars = 20;
   const auto n_clauses = 2 * n_vars;
   const auto max_clause_size = n_vars / 4;
   const auto n_input_vars = n_vars / 2;
   bool success = true;
   for (unsigned i = 0; success && i < n_rounds; ++i)
   {
      auto cnf = random_horn_cnf(n_clauses, max_clause_size, n_vars, false);
      URCEncoding u_enc(cnf, n_input_vars, false);
      u_enc.EliminateAuxVariablesWithMaxIncMinimize(2000, 1000, 0.5);
      std::cout << "Elimination finished" << std::endl;
      SATEncoding s_enc(cnf, n_input_vars);
      success = test_encoding_consistency_full(u_enc, s_enc)
                && test_encoding_consistency_partial(u_enc, s_enc)
                && test_encoding_propagation(u_enc, s_enc);
      std::cout << "Tested " << (success ? "[OK]" : "[FAILED]") << std::endl;
   }
   if (success)
   {
      std::cout << "All [OK]" << std::endl;
   }
   return true;
}

bool
SUBool::TestEncoding::TestRead() const
{
   std::cout << "Test CNF encoding read from stream." << std::endl;
   std::stringstream ss(encoding);
   auto enc = CNFEncoding::ReadFromStream(ss);
   if (enc)
   {
      std::cout << *enc << std::endl;
   }
   else
   {
      std::cout << "(null) [FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestEncoding::RunTests() const
{
   return TestSimpleSATEncoding() && TestSATEncoding(10, 20, 10)
          && TestURCEncoding(10) && TestEliminateAuxVariable(10) && TestRead();
}

int
main(int /*argc*/, char * /*argv*/[])
{
   SUBool::logs.mVerbLevel = 4;
   srand(time(NULL));
   return !SUBool::TestEncoding().RunTests();
}
