/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file test_glucose.cpp
 * Test connection to glucose library.
 */

#include <chrono>
#include <sstream>

#include "cnfutil.h"
#include "dimacs.h"
#include "glucose_bind.h"
#include "random_utils.h"
#include "test_glucose.h"
#include "util.h"

const char *dimacs = "p cnf 8 8\n"
                     "1 2 5 0\n"
                     "-5 3 4 0\n"
                     "1 2 6 0\n"
                     "-6 -3 4 0\n"
                     "1 2 7 0\n"
                     "-7 3 -4 0\n"
                     "1 2 8 0\n"
                     "-8 -3 -4 0\n";

const char *dimacs_1 = "c 1 clause\n"
                       "p cnf 3 1\n"
                       "-1 2 -3 0";

const char *dimacs_quinn = "c  quinn.cnf\n"
                           "c\n"
                           "p cnf 16 18\n"
                           "  1    2  0\n"
                           " -2   -4  0\n"
                           "  3    4  0\n"
                           " -4   -5  0\n"
                           "  5   -6  0\n"
                           "  6   -7  0\n"
                           "  6    7  0\n"
                           "  7  -16  0\n"
                           "  8   -9  0\n"
                           " -8  -14  0\n"
                           "  9   10  0\n"
                           "  9  -10  0\n"
                           "-10  -11  0\n"
                           " 10   12  0\n"
                           " 11   12  0\n"
                           " 13   14  0\n"
                           " 14  -15  0\n"
                           " 15   16  0";

bool
SUBool::TestGlucose::TestSimple() const
{
   std::cout << "Testing Glucose binding on a simple CNF" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs_quinn);
   std::cout << "Input: " << cnf << std::endl;
   SimpSolverWrapper wrap;
   auto &g = wrap.Solver();
   if (!g.AddCnf(cnf))
   {
      std::cout << "Glucose returned false when adding CNF [FAILED]"
                << std::endl;
      return false;
   }
   bool result = g.Solve();
   std::cout << "Glucose solve result: " << result
             << (result ? " [PASSED]" : " [FAILED]") << std::endl;
   PartialAssignment alpha = g.Model();
   std::cout << "Model: " << alpha << std::endl;
   auto val = cnf.Value(alpha);
   std::cout << "Value on model: " << val << std::endl;
   return (val == PartialValue::TRUE);
}

bool
SUBool::TestGlucose::TestLearntsRandom(const unsigned nrounds,
    const unsigned nclauses, const unsigned clause_size,
    const unsigned nvars) const
{
   std::cout << "Testing glucose learnts" << std::endl;
   for (unsigned round = 0; round < nrounds; ++round)
   {
      CNF cnf = random_k_cnf(nclauses, clause_size, nvars);
      std::cout << "Round " << round + 1 << std::endl;
      //<<", cnf:"<<std::endl; std::cout<<cnf;
      SimpSolverWrapper wrap;
      auto &g = wrap.Solver();
      if (!g.AddCnf(cnf))
      {
         std::cout << "Glucose returned false when adding CNF [FAILED]"
                   << std::endl;
         return false;
      }
      std::vector<Literal> assump(5);
      std::cout << "Using assumptions:";
      for (unsigned i = 0; i < assump.size(); ++i)
      {
         assump[i] = Literal(i + 1, (sub_rand() % 2) == 0);
         std::cout << " " << assump[i];
      }
      std::cout << std::endl;
      bool result = g.Solve(assump);
      std::cout << "Glucose solve result: " << (result ? "SAT" : "UNSAT")
                << std::endl;
      auto learnts = g.Learnts();
      std::cout << "...nLearnts=" << learnts.size() << std::endl;
      bool all_implicates = true;
      for (const auto &clause : learnts)
      {
         std::cout << clause;
         if (!is_implicate(cnf, clause))
         {
            std::cout << " !not implicate!";
            all_implicates = false;
         }
         std::cout << std::endl;
      }
      if (all_implicates)
      {
         std::cout << "All learnts are implicates [PASSED]" << std::endl;
      }
      else
      {
         std::cout << "Not all learnts are implicates [FAILED]" << std::endl;
         return false;
      }
   }
   return true;
}

bool
SUBool::TestGlucose::TestConflict(
    const CNF &cnf, const std::vector<Literal> &assumptions) const
{
   std::cout << "Testing glucose conflict clause with assumptions: ";
   for (unsigned i = 0; i < assumptions.size(); ++i)
   {
      std::cout << " " << assumptions[i];
   }
   std::cout << std::endl;
   SimpSolverWrapper wrap;
   auto &g = wrap.Solver();
   g.AddCnf(cnf);
   if (g.Solve(assumptions))
   {
      std::cout << "no conflict [OK]" << std::endl;
      return true;
   }
   auto c = g.ConflictClause();
   std::cout << "... conflict clause: " << c << std::endl;
   if (!is_implicate(cnf, c))
   {
      return false;
   }
   std::unordered_set<Literal> assump_set(
       assumptions.begin(), assumptions.end());
   for (const auto &lit : c)
   {
      if (assump_set.find(lit.Negation()) == assump_set.end())
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::TestGlucose::TestConflicts(const unsigned nrounds,
    const unsigned nconflicts_per_round, const unsigned nvars,
    const unsigned nclauses) const
{
   std::cout << "Testing glucose conflict clause" << std::endl;
   for (unsigned round = 0; round < nrounds; ++round)
   {
      auto sat_cnf = random_sat_cnf(nclauses, nvars);
      for (unsigned confl_round = 0; confl_round < nconflicts_per_round;
           ++confl_round)
      {
         if (!TestConflict(sat_cnf.first,
                 random_literal_vector_with_size(nvars / 3, nvars)))
         {
            std::cout << "[FAILED]" << std::endl;
            return false;
         }
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestGlucose::TestIncrementalSimple() const
{
   std::cout << "Testing glucose incremental capabilities on a simple CNF"
             << std::endl;
   IncSolverWrapper wrap;
   auto &g = wrap.Solver();
   auto cnf = Dimacs::parse_cnf(dimacs);
   std::cout << cnf << std::endl;
   g.AddCnf(cnf);
   PartialAssignment gamma(8);
   gamma[2] = PartialValue::FALSE;
   auto b = g.Solve(gamma);
   std::cout << "Solving " << gamma << " with result " << (b ? "SAT" : "UNSAT")
             << std::endl;
   gamma[1] = PartialValue::FALSE;
   b = g.Solve(gamma);
   std::cout << "Solving " << gamma << " with result " << (b ? "SAT" : "UNSAT")
             << std::endl;
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestGlucose::TestIncremental(const unsigned n_rounds,
    const unsigned n_clauses, const unsigned n_vars,
    const unsigned assump_length) const
{
   std::cout << "Testing glucose incremental capabilities" << std::endl;
   IncSolverWrapper wrap;
   auto &g = wrap.Solver();
   auto sat_cnf = random_sat_cnf(n_clauses, n_vars);
   g.AddCnf(sat_cnf.first);
   if (!g.Solve())
   {
      std::cout << "The CNF is declared UNSATISFIABLE [FAILED]" << std::endl;
   }
   for (unsigned round = 0; round < n_rounds; ++round)
   {
      auto assump = random_literal_vector_with_size(assump_length, n_vars);
      std::cout << "Trying assumptions: " << VectorPrinter(assump) << std::endl;
      auto b = g.Solve(assump);
      std::cout << "Result: " << (b ? "SAT" : "UNSAT") << std::endl;
      SimpSolverWrapper wrap2;
      auto &g2 = wrap2.Solver();
      g2.AddCnf(sat_cnf.first);
      auto b2 = g2.Solve(assump);
      std::cout << "Result 2: " << (b2 ? "SAT" : "UNSAT") << std::endl;
      if (b != b2)
      {
         std::cout << "[FAILED] mismatch" << std::endl;
      }
      std::cout << g.LearntsSize() << std::endl;
   }
   g.Solve();
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestGlucose::RunTests() const
{
   bool success = TestSimple();
   success = TestLearntsRandom(100, 150, 3, 30) && success;
   success = TestConflicts(10, 10, 30, 50) && success;
   success = TestIncrementalSimple() && success;
   success = TestIncremental(100, 1000, 500, 50) && success;
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::TestGlucose test;
   return !test.RunTests();
}
