/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file test_horn.cpp
 * Contains tests for functions related to Horn functions.
 */

#include <cstdlib>
#include <ctime>
#include <iostream>
#include <sstream>

#include "dimacs.h"
#include "horn.h"
#include "mrh.h"
#include "random_utils.h"
#include "rhorn.h"
#include "test_horn.h"

const char *dimacs_simple = "c\n"
                            "c start with comments\n"
                            "c\n"
                            "c\n"
                            "p cnf 5 3\n"
                            "1 -5 4 0\n"
                            "-1 5 3 4 0\n"
                            "-3 4 0";

const char *dimacs_horn = "c\n"
                          "c A horn formula\n"
                          "c\n"
                          "p cnf 7 8\n"
                          "2 0\n"
                          "-4 0\n"
                          "5 -7 -2 0\n"
                          "-1 3 -4 0\n"
                          "-5 -4 -2 1 0\n"
                          "3 -1 -2 0\n"
                          "5 -2 0\n"
                          "1 -2 -3 -4 0";

const char *dimacs_renhorn = "c\n"
                             "p cnf 7 8\n"
                             "2 0\n"
                             "-4 0\n"
                             "-2 5 -7 0\n"
                             "-1 -3 -4 0\n"
                             "1 -2 -4 -5 0\n"
                             "-1 -2 -3 0\n"
                             "-2 5 0\n"
                             "1 -2 3 -4 0";

const char *dimacs_nonrenhorn = "c\n"
                                "c A formula which is not renamable Horn\n"
                                "c\n"
                                "p cnf 8 8\n"
                                "1 -2 0\n"
                                "2 -3 0\n"
                                "3 1 -2 0\n"
                                "-1 4 0\n"
                                "-4 5 0\n"
                                "-5 7 3 0\n"
                                "-7 -1 5 4 0\n"
                                "8 -1 7 0";

SUBool::TestHorn::TestHorn() {}

bool
SUBool::TestHorn::TestHornCNFCheck(const char *dimacs, bool is_horn) const
{
   std::cout << "Test Horn check of a CNF" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs);
   std::cout << "Input: " << std::endl;
   std::cout << dimacs << std::endl;
   if (IsHornCNF(cnf) != is_horn)
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   else
   {
      std::cout << "[PASSED]" << std::endl;
   }
   return true;
}

#if 0
bool
SUBool::TestHorn::TestHornDNFCheck(const char *dimacs, bool is_horn) const
{
   std::cout << "Test Horn check of a DNF" << std::endl;
   std::stringstream ss;
   ss << dimacs;
   DNF dnf;
   ss >> dnf;
   std::cout << "Input: " << std::endl;
   std::cout << dimacs << std::endl;
   if (IsHornDNF(dnf) != is_horn)
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   else
   {
      std::cout << "[PASSED]" << std::endl;
   }
   return true;
}
#endif

bool
SUBool::TestHorn::TestAntiHornCNFCheck(
    const char *dimacs, bool is_antihorn) const
{
   std::cout << "Test AntiHorn check of a CNF" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs);
   std::cout << "Input: " << std::endl;
   std::cout << dimacs << std::endl;
   if (IsAntiHornCNF(cnf) != is_antihorn)
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   else
   {
      std::cout << "[PASSED]" << std::endl;
   }
   return true;
}

#if 0
bool
SUBool::TestHorn::TestAntiHornDNFCheck(
    const char *dimacs, bool is_antihorn) const
{
   std::cout << "Test AntiHorn check of a DNF" << std::endl;
   std::stringstream ss;
   ss << dimacs;
   DNF dnf;
   ss >> dnf;
   std::cout << "Input: " << std::endl;
   std::cout << dimacs << std::endl;
   if (IsAntiHornDNF(dnf) != is_antihorn)
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   else
   {
      std::cout << "[PASSED]" << std::endl;
   }
   return true;
}
#endif

bool
SUBool::TestHorn::TestRenamableHornCheck() const
{
   const unsigned n_rounds = 100;
   const unsigned n_vars = 50;
   const unsigned size = 100;
   const unsigned clause_size = 10;

   std::cout << "Test horn renamability check" << std::endl;
   /*   std::stringstream ss;
      ss<<dimacs_horn;
      CNF cnf;
      ss>>cnf;
      Assignment renaming(7);
      */
   for (unsigned round = 0; round < n_rounds; ++round)
   {
      Assignment renaming(n_vars);
      CNF cnf = random_horn_cnf(size, clause_size, n_vars, false);
      for (auto &&ren : renaming)
      {
         ren = (sub_rand() % 2 == 0);
      }
      cnf.ApplyRenaming(renaming);
      // std::cout<<"Input CNF: "<<std::endl;
      // std::cout<<cnf<<std::endl;
      Assignment out_renaming(7);
      if (!RenamableHornChecker::CheckCNF(cnf, &out_renaming))
      {
         std::cout << "CheckRenamableHornCNF returned false. [FAILED]"
                   << std::endl;
         return false;
      }
      // std::cout<<"Renaming: "<<out_renaming<<std::endl;
      cnf.ApplyRenaming(out_renaming);
      // std::cout<<"Renamed CNF: "<<std::endl;
      // std::cout<<cnf<<std::endl;
      if (!IsHornCNF(cnf))
      {
         std::cout << "Renamed CNF is not Horn. [FAILED]" << std::endl;
         return false;
      }
   }

   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestHorn::TestRenamableHornCheckNegative() const
{
   std::cout << "Test renamable Horn check with a non renamable Horn formula ";
   auto cnf = Dimacs::parse_cnf(dimacs_nonrenhorn);
   if (RenamableHornChecker::CheckCNF(cnf, nullptr))
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }

   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestHorn::TestSimpleMRH() const
{
   unsigned n_vars = 10;
   unsigned size = 50;
   unsigned clause_size = 3;
   std::cout << "Test a simple algorithm for MRH" << std::endl;

   CNF cnf = random_k_cnf(size, clause_size, n_vars);

#if 0
   std::stringstream ss;
   ss<<dimacs_nonrenhorn;
   CNF cnf;
   ss>>cnf;
#endif

   std::cout << "Generated " << clause_size << "-CNF:" << std::endl;
   std::cout << cnf.Size() << " clauses, " << cnf.MaxVariable() << " variables"
             << std::endl;
   // std::cout<<cnf;
   if (RenamableHornChecker::CheckCNF(cnf, nullptr))
   {
      std::cout << "The CNF is actually renamable Horn." << std::endl;
   }
   else
   {
      std::cout << "The CNF is not renamable Horn." << std::endl;
   }

   Assignment renaming(cnf.MaxVariable());
   unsigned mrh_size = SimpleMRH(cnf, &renaming);
   std::cout << "mrh size is: " << mrh_size << std::endl;
   if (mrh_size * 2 < cnf.Size())
   {
      std::cout << "not a half of clauses [FAILED]" << std::endl;
      return false;
   }
   CNF ren_cnf(cnf);
   ren_cnf.ApplyRenaming(renaming);
   unsigned horn_count = 0;
   for (const auto &clause : ren_cnf)
   {
      if (IsHornClause(clause))
      {
         ++horn_count;
      }
   }
   if (horn_count != mrh_size)
   {
      std::cout << "Bad size " << mrh_size << " vs " << horn_count
                << " [FAILED]" << std::endl;
      return false;
   }

   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestHorn::TestRHornSplit() const
{
   const unsigned n_rounds = 1;
   const unsigned n_vars = 1800;
   const unsigned size = 2000;
   const unsigned clause_size = 3;
   std::cout << "Test splitting to renamable horn parts" << std::endl;

   for (unsigned round = 0; round < n_rounds; ++round)
   {
      CNF cnf = random_k_cnf(size, clause_size, n_vars);
      std::vector<std::pair<CNF, Assignment>> parts;
      SplitToRenamableHornParts(cnf, &parts);
      std::cout << "Given CNF of size " << cnf.Size();
      std::cout << " I found split into " << parts.size() << " parts"
                << std::endl;
      unsigned size = 0;
      for (unsigned i = 0; i < parts.size(); ++i)
      {
         std::cout << "Part " << i << ": size " << parts[i].first.Size()
                   << std::endl;
         size += parts[i].first.Size();
         CNF part_cnf(parts[i].first);
         part_cnf.ApplyRenaming(parts[i].second);
         if (!IsHornCNF(part_cnf))
         {
            std::cout << "Part is not Horn [FAILED]" << std::endl;
            return false;
         }
      }
      if (size != cnf.Size())
      {
         std::cout << "Total size differs [FAILED]" << std::endl;
         return false;
      }
   }
   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestHorn::RunTests() const
{
   bool success = TestHornCNFCheck(dimacs_simple, false)
                  && TestHornCNFCheck(dimacs_horn, true);
   /*   success = TestHornDNFCheck(dimacs_simple, true)
                && TestHornDNFCheck(dimacs_horn, false) && success;*/
   success = TestAntiHornCNFCheck(dimacs_simple, true)
             && TestAntiHornCNFCheck(dimacs_horn, false) && success;
   /*success = TestAntiHornDNFCheck(dimacs_simple, false)
             && TestAntiHornDNFCheck(dimacs_horn, true) && success;*/
   success = TestRenamableHornCheck() && success;
   success = TestRenamableHornCheckNegative() && success;
   success = TestSimpleMRH() && success;
   success = TestRHornSplit() && success;
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand((unsigned)time(nullptr));
   SUBool::TestHorn test;
   return !test.RunTests();
}
