/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file test_imp_system.cpp
 * Contains tests of ImpSystem (see imp_system.h)
 */

#include "test_imp_system.h"
#include "cnfcompare.h"
#include "dimacs.h"
#include "drenc.h"
#include "horn.h"
#include "logstream.h"
#include "normalform.h"
#include "random_utils.h"
#include "sbexception.h"

const std::vector<SUBool::VarImpl> simple_implications(
    {SUBool::VarImpl({1}, {3, 4}, true), SUBool::VarImpl({2}, {3, 4}, true),
        SUBool::VarImpl({3, 4}, {5}, true)});
const size_t simple_implications_length = 9;

const std::vector<SUBool::VarImpl> merge_implications({SUBool::VarImpl(
                                                           {1, 2}, {6}, false),
    SUBool::VarImpl({3}, {1, 2}, false), SUBool::VarImpl({3, 5}, {3, 7}, false),
    SUBool::VarImpl({1, 2}, {4}, false), SUBool::VarImpl({3, 5}, {1, 8}, false),
    SUBool::VarImpl({1, 2}, {5}, false), SUBool::VarImpl({3, 5}, {4, 7}, false),
    SUBool::VarImpl({1, 2}, {7}, false), SUBool::VarImpl({3, 5}, {3, 2}, false),
    SUBool::VarImpl({1, 2}, {5}, false),
    SUBool::VarImpl({3, 5}, {1, 3}, false)});

const std::vector<SUBool::VarImpl> propagate_empty_body({SUBool::VarImpl({1, 2},
                                                             {6}, false),
    SUBool::VarImpl({3}, {1, 2}, false), SUBool::VarImpl({3, 5}, {3, 7}, false),
    SUBool::VarImpl({1, 2}, {4}, false), SUBool::VarImpl({3, 5}, {1, 8}, false),
    SUBool::VarImpl({1, 2}, {5}, false), SUBool::VarImpl({}, {1}, false),
    SUBool::VarImpl({3, 5}, {4, 7}, false), SUBool::VarImpl({1, 2}, {7}, false),
    SUBool::VarImpl({3, 5}, {3, 2}, false), SUBool::VarImpl({1, 2}, {5}, false),
    SUBool::VarImpl({3, 5}, {1, 3}, false), SUBool::VarImpl({}, {2}, false)});

const std::vector<SUBool::VarImpl> merged_implications(
    {SUBool::VarImpl({3}, {1, 2}, true),
        SUBool::VarImpl({1, 2}, {4, 5, 6, 7}, true),
        SUBool::VarImpl({3, 5}, {1, 2, 3, 4, 7, 8}, true)});

const std::vector<SUBool::VarImpl> rl_saturation(
    {SUBool::VarImpl({1}, {2}, true), SUBool::VarImpl({2}, {1}, true),
        SUBool::VarImpl({1, 3}, {2, 4}, true),
        SUBool::VarImpl({2, 3}, {1, 4}, true)});
const std::vector<SUBool::VarImpl> gd_basis({SUBool::VarImpl({1}, {2}, true),
    SUBool::VarImpl({1}, {3}, true), SUBool::VarImpl({1, 4}, {5}, true),
    SUBool::VarImpl({1, 2}, {5}, true)});

bool
SUBool::TestImpSystem::TestClosureSimple() const
{
   std::cout << "Testing simple closure" << std::endl;
   ImplSystem<unsigned> is;
   is.Append(simple_implications.begin(), simple_implications.end());
   auto bc = is.Size();
   std::cout << "body count: " << bc << std::endl;
   if (bc != simple_implications.size())
   {
      logs.TLog(1) << " [FAILED] Body count should be "
                   << simple_implications.size() << std::endl;
      return false;
   }
   auto length = is.Length();
   std::cout << "    length: " << length << std::endl;
   if (bc != simple_implications.size())
   {
      logs.TLog(1) << " [FAILED] Length should be "
                   << simple_implications_length << std::endl;
      return false;
   }
   std::cout << "Implications:" << std::endl;
   is.Print(std::cout);
   std::cout << std::endl;
   std::cout << "Computing closure of {1, 2}" << std::endl;
   auto closure = is.Closure({1, 2});
   if (closure.size() != 5)
   {
      std::cout << "[FAILED] Closure too small (" << closure.size() << ")"
                << std::endl;
      std::copy(closure.begin(), closure.end(),
          std::ostream_iterator<unsigned>(std::cout, " "));
      return false;
   }
   std::cout << "[OK]";
   for (const auto &v : closure)
   {
      std::cout << " " << v;
   }
   std::cout << std::endl;

   std::cout << "Computing closure of {1}" << std::endl;
   closure = is.Closure({1});
   if (closure.size() != 4)
   {
      std::cout << "[FAILED] Closure too small (" << closure.size() << ")"
                << std::endl;
      return false;
   }
   std::cout << "[OK]";
   for (const auto &v : closure)
   {
      std::cout << " " << v;
   }
   std::cout << std::endl;

   std::cout << "Computing closure of {2}" << std::endl;
   closure = is.Closure({2});
   if (closure.size() != 4)
   {
      std::cout << "[FAILED] Closure too small (" << closure.size() << ")"
                << std::endl;
      return false;
   }
   std::cout << "[OK]";
   for (const auto &v : closure)
   {
      std::cout << " " << v;
   }
   std::cout << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestMergeBodies() const
{
   std::cout << "Testing merging bodies" << std::endl;
   ImplSystem<unsigned> is(
       merge_implications.begin(), merge_implications.end());
   std::cout << "Implications:" << std::endl;
   std::cout << is << std::endl;
   is.MergeBodies();
   std::cout << "Merged implications:" << std::endl;
   std::cout << is << std::endl;
   ImplSystem<unsigned> is_merged(
       merged_implications.begin(), merged_implications.end());
   std::cout << "Assumed merged implications:" << std::endl;
   std::cout << is_merged;
   if (!(is == is_merged))
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

const char *dimacs_horn = "c\n"
                          "c A horn formula\n"
                          "c\n"
                          "p cnf 7 10\n"
                          "2 0\n"
                          "-4 0\n"
                          "7 0\n"
                          "5 -7 -2 0\n"
                          "-1 3 -4 0\n"
                          "3 -1 -2 0\n"
                          "-1 -2 5 0\n"
                          "-4 -2 -5 1 0\n"
                          "5 -2 0\n"
                          "1 -2 -3 -4 0\n"
                          "-1 -3 -5 -7 0";

bool
SUBool::TestImpSystem::TestHorn() const
{
   auto cnf = Dimacs::parse_cnf(dimacs_horn);
   cnf.Sort();
   std::cout << "Testing Horn CNF" << std::endl;
   std::cout << cnf << std::endl;
   auto is = horn_cnf_to_implsys(cnf, true);
   std::cout << "Implicational system" << std::endl;
   std::cout << is << std::endl;
   auto back_cnf = implsys_to_horn_cnf(is);
   back_cnf.Sort();
   std::cout << "Clauses back" << std::endl;
   std::cout << back_cnf << std::endl;
   if (cnf != back_cnf)
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   auto cl = is.Closure({});
   std::cout << "Closure of the empty set:";
   for (unsigned var : cl)
   {
      std::cout << " " << var;
   }
   std::cout << std::endl;

   cl = is.Closure({1});
   std::cout << "Closure of {1}:";
   for (unsigned var : cl)
   {
      std::cout << " " << var;
   }
   std::cout << std::endl;

   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestRandomHornSingleClosure(const CNF &cnf,
    UnitPropagator &up, const VarImplSystem &is,
    const std::vector<unsigned> &body) const
{
   std::cout << "Testing random closure with body:";
   for (const auto &var : body)
   {
      std::cout << " " << var;
   }
   std::cout << std::endl;
   PartialAssignment up_assump(cnf.MaxVariable(), PartialValue::UNDEF);
   for (const auto &var : body)
   {
      up_assump[var] = PartialValue::TRUE;
   }
   PartialAssignment up_closure(0);
   up.Propagate(up_assump, &up_closure);
   auto up_literals = literals_of_assignment(up_closure);
   std::cout << "UP closure:";
   for (const auto &lit : up_literals)
   {
      if (lit.Negative())
      {
         throw BadInputException("TestImpSystem::TestRandomHornSingleClosure",
             "negative literal derived by unit propagation");
      }
      std::cout << " " << lit.Variable();
   }
   std::cout << std::endl;
   auto imp_closure = is.Closure(body);
   std::cout << "Imp closure:";
   for (const auto &var : imp_closure)
   {
      std::cout << " " << var;
   }
   std::cout << std::endl;
   if (up_literals.size() < imp_closure.size())
   {
      std::cout << "[FAILED] UP closure is smaller" << std::endl;
      return false;
   }
   if (imp_closure.size() < up_literals.size())
   {
      std::cout << "[FAILED] Imp closure is smaller" << std::endl;
      return false;
   }
   if (!std::equal(up_literals.begin(), up_literals.end(), imp_closure.begin(),
           [](Literal lit, unsigned var) { return lit.Variable() == var; }))
   {
      std::cout << "[FAILED] The closures are not equal" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestRandomHornClosures(const CNF &cnf,
    const VarImplSystem &is, const unsigned rounds,
    const unsigned body_size) const
{
   std::cout << "Testing random horn closures" << std::endl;
   auto up = UnitPropagator(&cnf);
   std::cout << "cnf: " << std::endl;
   std::cout << cnf << std::endl;
   std::cout << "is: " << std::endl;
   std::cout << is << std::endl;
   for (unsigned i = 0; i < rounds; ++i)
   {
      std::vector<unsigned> all_vars(cnf.MaxVariable());
      for (unsigned j = 0; j < all_vars.size(); ++j)
      {
         all_vars[j] = j + 1;
      }
      std::shuffle(all_vars.begin(), all_vars.end(), sub_rand);
      std::vector<unsigned> body(
          all_vars.begin(), all_vars.begin() + body_size);
      if (!TestRandomHornSingleClosure(cnf, up, is, body))
      {
         std::cout << "[FAILED]" << std::endl;
         return false;
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestRandomHornBodyMinimal(
    const CNF &cnf, const VarImplSystem &is) const
{
   auto gd_is = is;
   auto bm_is = is;
   gd_is.MakeGDBasis();
   bm_is.MakeBodyMinimal();
   if (gd_is.Size() != bm_is.Size())
   {
      std::cout << "[FAILED] Body minimal and GD basis size mismatch"
                << std::endl;
      return false;
   }
   auto gd_cnf = implsys_to_horn_cnf(gd_is);
   auto bm_cnf = implsys_to_horn_cnf(bm_is);
   if (!cnf_equal(gd_cnf, cnf))
   {
      std::cout << "[FAILED] GD basis not equivalent" << std::endl;
      return false;
   }
   if (!cnf_equal(bm_cnf, cnf))
   {
      std::cout << "[FAILED] body minimal not equivalent" << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestImpSystem::TestRandomHorn(unsigned cnf_rounds,
    unsigned closure_rounds, const unsigned size,
    const unsigned max_clause_size, const unsigned n_vars,
    const unsigned body_size) const
{
   for (unsigned i = 0; i < cnf_rounds; ++i)
   {
      auto cnf = random_horn_cnf(size, max_clause_size, n_vars, true);
      auto is = horn_cnf_to_implsys(cnf, true);
      if (!TestRandomHornClosures(cnf, is, closure_rounds, body_size))
      {
         return false;
      }
      if (!TestRightSaturate(is))
      {
         return false;
      }
      if (!TestRandomHornBodyMinimal(cnf, is))
      {
         return false;
      }
      if (!TestHornRightIrredundant(cnf, is))
      {
         return false;
      }
      if (!TestHornLeftIrredundant(cnf, is))
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::TestImpSystem::TestRightSaturate(const VarImplSystem &is) const
{
   std::cout << "Testing right saturation" << std::endl;
   std::cout << "Input is:\n" << is << std::endl;
   auto is_rs(is);
   is_rs.MakeRightSaturated();
   std::cout << "Right saturated is:\n" << is_rs << std::endl;
   auto i_is_rs = is_rs.begin();
   std::vector<unsigned> empty_closure = is.Closure({});
   if (i_is_rs->EmptyBody())
   {
      if (i_is_rs->Cons() != empty_closure)
      {
         std::cout << "[FAILED] Empty closure mismatch" << std::endl;
         std::cout << i_is_rs->Cons().size() << std::endl;
         std::cout << empty_closure.size() << std::endl;
         std::cout << "   empty cons: "
                   << VectorPrinter<unsigned>(i_is_rs->Cons()) << std::endl;
         std::cout << "empty closure: "
                   << RangePrinter<std::vector<unsigned>::iterator>(
                          empty_closure.begin(), empty_closure.end())
                   << std::endl;
         return false;
      }
      ++i_is_rs;
   }
   else if (!empty_closure.empty())
   {
      std::cout << "[FAILED] Empty cons but not the closure" << std::endl;
      std::cout << "empty closure: " << VectorPrinter<unsigned>(empty_closure)
                << std::endl;
      return false;
   }
   for (; i_is_rs != is_rs.end(); ++i_is_rs)
   {
      auto closure = is.Closure(i_is_rs->Body());
      auto min_closure = vector_subtract(
          vector_subtract(closure, empty_closure), i_is_rs->Body());
      // if (impl.Cons() != closure)
      if (i_is_rs->Cons() != min_closure)
      {
         std::cout << "[FAILED]" << std::endl;
         std::cout << "       impl: " << *i_is_rs << std::endl;
         std::cout << "min closure: " << VectorPrinter<unsigned>(min_closure)
                   << std::endl;
         return false;
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestRightLeftSaturate() const
{
   std::cout << "Testing right and left saturation" << std::endl;
   VarImplSystem is(gd_basis.begin(), gd_basis.end());
   std::cout << "Input is:\n" << is << std::endl;
   is.MakeRightLeftSaturated();
   std::cout << "Output is:\n" << is << std::endl;

   std::cout << "Testing right and left saturation" << std::endl;
   VarImplSystem is2(rl_saturation.begin(), rl_saturation.end());
   std::cout << "Input is:\n" << is2 << std::endl;
   is2.MakeRightLeftSaturated();
   std::cout << "Output is:\n" << is2 << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestGDBasis() const
{
   std::cout << "Testing GD basis" << std::endl;
   VarImplSystem is(gd_basis.begin(), gd_basis.end());
   std::cout << "Input is:\n" << is << std::endl;
   is.MakeGDBasis();
   std::cout << "Output is:\n" << is << std::endl;

   std::cout << "Testing GD basis" << std::endl;
   VarImplSystem is2(rl_saturation.begin(), rl_saturation.end());
   std::cout << "Input is:\n" << is2 << std::endl;
   is2.MakeGDBasis();
   std::cout << "Output is:\n" << is2 << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestBodyMinimal() const
{
   std::cout << "Testing body minimal" << std::endl;
   VarImplSystem is(gd_basis.begin(), gd_basis.end());
   std::cout << "Input is:\n" << is << std::endl;
   is.MakeBodyMinimal();
   std::cout << "Output is:\n" << is << std::endl;

   std::cout << "Testing GD basis saturation" << std::endl;
   VarImplSystem is2(rl_saturation.begin(), rl_saturation.end());
   std::cout << "Input is:\n" << is2 << std::endl;
   is2.MakeBodyMinimal();
   std::cout << "Output is:\n" << is2 << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestDualRail() const
{
   std::cout << "Testing dual rail encoding" << std::endl;
   auto cnf = random_horn_cnf(50, 5, 30, true);
   auto is = dual_rail(cnf, true);

   std::cout << "Input is:\n" << cnf << std::endl;
   std::cout << "Output is:\n" << is << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestPropagateEmptyBody() const
{
   std::cout << "Testing empty body propagation" << std::endl;
   VarImplSystem is(propagate_empty_body);
   std::cout << "Input: " << std::endl;
   std::cout << is << std::endl;
   is.PropagateEmptyBody();
   std::cout << "Output: " << std::endl;
   std::cout << is << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestHornRightIrredundant(
    const CNF &cnf, const VarImplSystem &is) const
{
   std::cout << "Testing horn right irredundancy";
   auto irred_is = is;
   irred_is.MakeRightIrredundant();
   auto irred_cnf = implsys_to_horn_cnf(irred_is);
   if (!cnf_equal(irred_cnf, cnf))
   {
      std::cout << "[FAILED]" << std::endl;
      std::cout << "Input IS: " << is << std::endl;
      std::cout << "Irred IS: " << irred_is << std::endl;
      std::cout << "Input CNF: " << cnf << std::endl;
      std::cout << "Irred CNF: " << irred_cnf << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::TestHornLeftIrredundant(
    const CNF &cnf, const VarImplSystem &is) const
{
   std::cout << "Testing horn left irredundancy";
   auto irred_is = is;
   irred_is.MakeLeftIrredundant();
   auto irred_cnf = implsys_to_horn_cnf(irred_is);
   if (!cnf_equal(irred_cnf, cnf))
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestImpSystem::RunTests() const
{
   return TestClosureSimple() && TestMergeBodies() && TestHorn()
          && TestRandomHorn(20, 20, 20, 5, 10, 5) && TestRightLeftSaturate()
          && TestDualRail() && TestPropagateEmptyBody() && TestGDBasis()
          && TestBodyMinimal();
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::logs.mVerbLevel = 3;
   return !SUBool::TestImpSystem().RunTests();
}
