/*
 * Project SUBool
 * Petr Kucera, 2017
 */
/**@file test_kissat.cpp
 * Test connection to kissat library.
 */

#include <chrono>
#include <sstream>
#include <unordered_set>

#include "dimacs.h"
#include "kissat_bind.h"
#include "random_utils.h"
#include "test_kissat.h"
#include "util.h"

const char *dimacs = "p cnf 8 8\n"
                     "1 2 5 0\n"
                     "-5 3 4 0\n"
                     "1 2 6 0\n"
                     "-6 -3 4 0\n"
                     "1 2 7 0\n"
                     "-7 3 -4 0\n"
                     "1 2 8 0\n"
                     "-8 -3 -4 0\n";

const char *dimacs_1 = "c 1 clause\n"
                       "p cnf 3 1\n"
                       "-1 2 -3 0";

const char *dimacs_quinn = "c  quinn.cnf\n"
                           "c\n"
                           "p cnf 16 18\n"
                           "  1    2  0\n"
                           " -2   -4  0\n"
                           "  3    4  0\n"
                           " -4   -5  0\n"
                           "  5   -6  0\n"
                           "  6   -7  0\n"
                           "  6    7  0\n"
                           "  7  -16  0\n"
                           "  8   -9  0\n"
                           " -8  -14  0\n"
                           "  9   10  0\n"
                           "  9  -10  0\n"
                           "-10  -11  0\n"
                           " 10   12  0\n"
                           " 11   12  0\n"
                           " 13   14  0\n"
                           " 14  -15  0\n"
                           " 15   16  0";

bool
SUBool::TestKissat::TestSimple() const
{
   std::cout << "Testing Kissat binding on a simple CNF" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs_quinn);
   std::cout << "Input: " << cnf << std::endl;
   KissatSolverWrapper wrap;
   auto &g = wrap.Solver();
   if (!g.AddCnf(cnf))
   {
      std::cout << "Kissat returned false when adding CNF [FAILED]"
                << std::endl;
      return false;
   }
   bool result = g.Solve();
   std::cout << "Kissat solve result: " << result
             << (result ? " [PASSED]" : " [FAILED]") << std::endl;
   PartialAssignment alpha = g.Model();
   std::cout << "Model: " << alpha << std::endl;
   auto val = cnf.Value(alpha);
   std::cout << "Value on model: " << val << std::endl;
   return (val == PartialValue::TRUE);
}

bool
SUBool::TestKissat::TestLearntsRandom(const unsigned nrounds,
    const unsigned nclauses, const unsigned clause_size,
    const unsigned nvars) const // TODO: not working - g.Learnts()
{
   std::cout << "Testing kissat learnts" << std::endl;
   for (unsigned round = 0; round < nrounds; ++round)
   {
      CNF cnf = random_k_cnf(nclauses, clause_size, nvars);
      std::cout << "Round " << round + 1 << std::endl;
      //<<", cnf:"<<std::endl; std::cout<<cnf;
      KissatSolverWrapper wrap;
      auto &g = wrap.Solver();
      if (!g.AddCnf(cnf))
      {
         std::cout << "Kissat returned false when adding CNF [FAILED]"
                   << std::endl;
         return false;
      }
      std::vector<Literal> assump(5);
      std::cout << "Using assumptions:";
      for (unsigned i = 0; i < assump.size(); ++i)
      {
         assump[i] = Literal(i + 1, (sub_rand() % 2) == 0);
         std::cout << " " << assump[i];
      }
      std::cout << std::endl;
      bool result = g.Solve(assump);
      std::cout << "Kissat solve result: " << (result ? "SAT" : "UNSAT")
                << std::endl;
      auto learnts = g.Learnts();
      std::cout << "...nLearnts=" << learnts.size() << " (" << g.LearntsSize()
                << ")" << std::endl;
      bool all_implicates = true;
      for (const auto &clause : learnts)
      {
         std::cout << clause;
         KissatSolverWrapper wrap2;
         auto &g2 = wrap2.Solver();
         g2.AddCnf(cnf);
         std::vector<Literal> neg_literals;
         neg_literals.reserve(clause.Size());
         for (const auto &lit : clause)
         {
            neg_literals.push_back(lit.Negation());
         }
         if (g2.Solve(neg_literals))
         {
            std::cout << " !not implicate!";
            all_implicates = false;
         }
         std::cout << std::endl;
      }
      if (all_implicates)
      {
         std::cout << "All learnts are implicates [PASSED]" << std::endl;
      }
      else
      {
         std::cout << "Not all learnts are implicates [FAILED]" << std::endl;
         return false;
      }
   }
   return true;
}

bool
SUBool::TestKissat::TestConflict(
    const CNF &cnf, // TODO: not working - g.ConflictClause()
    const std::vector<Literal> &assumptions) const
{
   std::cout << "Testing kissat conflict clause with assumptions: ";
   for (unsigned i = 0; i < assumptions.size(); ++i)
   {
      std::cout << " " << assumptions[i];
   }
   std::cout << std::endl;
   KissatSolverWrapper wrap;
   auto &g = wrap.Solver();
   g.AddCnf(cnf);
   if (g.Solve(assumptions))
   {
      std::cout << "no conflict [OK]" << std::endl;
      return true;
   }
   auto c = g.ConflictClause();
   std::cout << "... conflict clause: " << c << std::endl;

   KissatSolverWrapper wrap2;
   auto &g2 = wrap2.Solver();
   g2.AddCnf(cnf);
   std::vector<Literal> neg_literals;
   neg_literals.reserve(c.Size());
   for (const auto &lit : c)
   {
      neg_literals.push_back(lit.Negation());
   }

   if (g2.Solve(neg_literals))
   {
      return false;
   }
   std::unordered_set<Literal> assump_set(
       assumptions.begin(), assumptions.end());
   for (const auto &lit : c)
   {
      if (assump_set.find(lit.Negation()) == assump_set.end())
      {
         return false;
      }
   }
   return true;
}

bool
SUBool::TestKissat::TestConflicts(const unsigned nrounds,
    const unsigned nconflicts_per_round, const unsigned nvars,
    const unsigned nclauses) const // TODO: not working
{
   std::cout << "Testing kissat conflict clause" << std::endl;
   for (unsigned round = 0; round < nrounds; ++round)
   {
      auto sat_cnf = random_sat_cnf(nclauses, nvars);
      for (unsigned confl_round = 0; confl_round < nconflicts_per_round;
           ++confl_round)
      {
         if (!TestConflict(sat_cnf.first,
                 random_literal_vector_with_size(nvars / 3, nvars)))
         {
            std::cout << "[FAILED]" << std::endl;
            return false;
         }
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestKissat::RunTests() const
{
   bool success = TestSimple();
   // success = TestLearntsRandom(100, 150, 3, 30) && success;
   // success = TestConflicts(10, 10, 30, 50) && success;
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::TestKissat test;
   return !test.RunTests();
}
