/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file test_nfliteralset.cpp
 * Contains tests for NFLiteralSet.
 */

#include <iostream>
#include <sstream>
#include <vector>

#include "assign.h"
#include "dimacs.h"
#include "random_utils.h"
#include "test_nfliteralset.h"

bool
SUBool::TestNFLiteralSet::TestDeterminingValue(bool value) const
{
   bool success = true;

   bool determining_value = false;
   if (value)
   {
      determining_value = NFLiteralSet<true>::DeterminingValue();
   }
   else
   {
      determining_value = NFLiteralSet<false>::DeterminingValue();
   }

   std::cout << "Test determining value '";
   std::cout << (value ? "true" : "false") << "' of a literal set";
   if (determining_value == value)
   {
      std::cout << "[PASSED]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
      success = false;
   }
   return success;
}

bool
SUBool::TestNFLiteralSet::TestClauseDeterminingValue() const
{
   std::cout << "Test determining value of a clause ";
   bool success = true;
   if (Clause::DeterminingValue())
   {
      std::cout << "[PASSED]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
      success = false;
   }
   return success;
}

bool
SUBool::TestNFLiteralSet::TestTermDeterminingValue() const
{
   std::cout << "Test determining value of a term ";
   bool success = true;
   if (!Term::DeterminingValue())
   {
      std::cout << "[PASSED]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
      success = false;
   }
   return success;
}

bool
SUBool::TestNFLiteralSet::TestClauseSize() const
{
   std::cout << "Test clause size" << std::endl;
   size_t clause_size = sub_rand() % 100;
   std::vector<Literal> literals =
       random_literal_vector_with_size(clause_size, 200);

   Clause clause(literals);
   std::cout << "... Initialize a clause of size " << clause_size;
   std::cout << ", got clause of size " << clause.Size();
   if (clause.Size() == literals.size())
   {
      std::cout << " [PASSED]" << std::endl;
   }
   else
   {
      std::cout << " [FAILED]" << std::endl;
      return false;
   }

   return true;
}

bool
SUBool::TestNFLiteralSet::TestLiteralAccess() const
{
   std::cout << "Test literal access" << std::endl;
   size_t clause_size = sub_rand() % 100;
   std::vector<Literal> literals =
       random_literal_vector_with_size(clause_size, 200);

   Clause clause(literals);
   std::sort(literals.begin(), literals.end());
   if (literals.size() != clause.Size())
   {
      std::cout << "... Failed to generate a random clause [FAILED]"
                << std::endl;
      return false;
   }
   std::cout << "... Generated a random clause of size " << clause_size
             << std::endl;
   std::cout << "... Comparing the literals with the input literal vector ";

   bool success = true;
   for (size_t i = 0; i < clause.Size(); ++i)
   {
      //   std::cout<<clause[i].mVariable<<" "<<clause[i].mPositive<<" ";
      //  std::cout<<literals[i].mVariable<<"
      //  "<<literals[i].mPositive<<std::endl;
      if (clause[i] != literals[i])
      {
         success = false;
      }
   }
   std::cout << (success ? "[PASSED]" : "[FAILED]") << std::endl;
   return success;
}

bool
SUBool::TestNFLiteralSet::TestEvaluationSimple() const
{
   Assignment assignment(9, false);
   std::cout << "Test simple evaluation ";

   std::vector<Literal> literals({Literal(3, false), Literal(5, true),
       Literal(4, true), Literal(2, false), Literal(7, false)});

   Clause clause(literals);
   for (auto &lit : literals)
   {
      lit.Negate();
   }
   Term term(literals);
   if (!clause.Value(assignment) || term.Value(assignment))
   {
      std::cout << "first round [FAILED]" << std::endl;
      return false;
   }
   assignment[3] = true;
   assignment[2] = true;
   assignment[7] = true;
   if (clause.Value(assignment) || !term.Value(assignment))
   {
      std::cout << "second round [FAILED]" << std::endl;
      return false;
   }
   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestNFLiteralSet::TestEvaluationRandom() const
{
   const unsigned max_literals = 20;
   const unsigned n_vars = 40;
   const unsigned n_assign = 100;
   std::cout << "Test random evaluation" << std::endl;
   const size_t clause_size = sub_rand() % max_literals;
   std::vector<Literal> literals =
       random_literal_vector_with_size(clause_size, n_vars);

   const Clause clause(literals);
   if (literals.size() != clause.Size())
   {
      std::cout << "... Failed to generate a random clause [FAILED]"
                << std::endl;
      return false;
   }
   std::cout << "... Generated a random clause of size " << clause_size
             << std::endl;
   std::cout << "... Trying " << n_assign << " random assignments";
   bool success = true;
   for (unsigned i = 0; success && i < n_assign; ++i)
   {
      Assignment assignment(n_vars);
      for (auto &&val : assignment)
      {
         val = (sub_rand() % 2 == 0 ? true : false);
      }
      bool clause_value = false;
      for (unsigned lit = 0; !clause_value && lit < literals.size(); ++lit)
      {
         if (literals[lit].Value(assignment[literals[lit].mVariable]))
         {
            clause_value = true;
         }
      }
      if (clause_value != clause.Value(assignment))
      {
         success = false;
      }
   }
   std::cout << (success ? "[PASSED]" : "[FAILED]") << std::endl;
   return success;
}

bool
SUBool::TestNFLiteralSet::TestDIMACS() const
{
   std::cout << "Test construct from DIMACS string ";
   std::vector<Literal> literals = random_literal_vector_with_size(7, 20);

   Clause clause(literals);
   std::stringstream dimacs;
   dimacs << clause;
   auto clause_dimacs = Dimacs::parse_clause(dimacs);
   std::cout << "clause_dimacs       =" << clause_dimacs << std::endl;
   std::cout << "clause              =" << clause << std::endl;
   if (clause_dimacs == clause)
   {
      std::cout << "[PASSED]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }

   std::cout << "Test construct from DIMACS stream ";
   auto clause_dimacs_stream = Dimacs::parse_clause(dimacs);
   std::cout << "clause_dimacs_stream=" << clause_dimacs_stream << std::endl;
   if (clause_dimacs_stream == clause)
   {
      std::cout << "[PASSED]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestNFLiteralSet::TestLiteralSetPolarities() const
{
   const unsigned n_rounds = 50;
   const unsigned size = 5;
   const unsigned n_vars = 15;

   std::cout << "Test literal set polarity counting." << std::endl;
   std::cout << "Trying " << n_rounds << " random terms. ";
   for (unsigned i = 0; i < n_rounds; ++i)
   {
      std::vector<Literal> literals =
          random_literal_vector_with_size(size, n_vars);
      Term term(literals);
      unsigned positive = 0;
      unsigned negative = 0;
      for (const auto &lit : literals)
      {
         if (lit.mPositive)
         {
            ++positive;
         }
         else
         {
            ++negative;
         }
      }
      std::pair<unsigned, unsigned> polarities = term.CountVariablePolarities();
      if (polarities.first != positive || polarities.second != negative)
      {
         std::cout << "[FAILED]" << std::endl;
         return false;
      }
   }
   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestNFLiteralSet::TestApplyRenamingToLiteralSet() const
{
   std::cout << "Test apply renaming to literal set" << std::endl;
   const unsigned n_rounds = 100;
   const unsigned n_vars = 20;
   const unsigned size = 10;
   for (unsigned i = 0; i < n_rounds; ++i)
   {
      Assignment alpha(size);
      for (unsigned i = 1; i <= 7; ++i)
      {
         alpha[i] = (sub_rand() % 2 == 0);
      }
      std::vector<Literal> literals =
          random_literal_vector_with_size(size, n_vars);
      Clause clause(literals);
      Clause clause_out(literals);
      clause_out.ApplyRenaming(alpha);
      for (unsigned j = 0; j < clause.Size(); ++j)
      {
         const Literal &lit = clause[j];
         const Literal &lit_out = clause_out[j];
         if (alpha[lit.mVariable])
         {
            if (lit.mPositive == lit_out.mPositive)
            {
               std::cout << "[FAILED]" << std::endl;
               return false;
            }
         }
         else
         {
            if (lit.mPositive != lit_out.mPositive)
            {
               std::cout << "[FAILED]" << std::endl;
               return false;
            }
         }
      }
   }
   std::cout << "[TRUE]" << std::endl;
   return true;
}

bool
SUBool::TestNFLiteralSet::TestApply(const PartialAssignment &alpha,
    const Clause &a, const Clause &b, PartialValue res) const
{
   std::cout << "Test application of a partial assignment to a clause"
             << std::endl;
   std::cout << "Input clause: " << a << std::endl;
   std::cout << "alpha: " << alpha << std::endl;
   Clause c(a);
   PartialValue r = c.Apply(alpha);
   bool cmp = (c == b);
   std::cout << "Output clause: " << c << (cmp ? " [OK]" : " [FAILED]")
             << std::endl;
   std::cout << "Result: " << r << ((r == res) ? " [OK]" : " [FAILED]")
             << std::endl;
   return cmp && (r == res);
}

bool
SUBool::TestNFLiteralSet::TestApply() const
{
   bool success = TestApply(PartialAssignment("******"),
       Dimacs::parse_clause("1 2 3 4 5 0"), Dimacs::parse_clause("1 2 3 4 5 0"),
       PartialValue::UNDEF);
   success =
       TestApply(PartialAssignment("01100"),
           Dimacs::parse_clause("1 -2 -3 4 5 0"), Clause(), PartialValue::FALSE)
       && success;
   success = TestApply(PartialAssignment("0*1*0"),
                 Dimacs::parse_clause("1 -2 -3 4 5 0"),
                 Dimacs::parse_clause("-2 4 0"), PartialValue::UNDEF)
             && success;
   success =
       TestApply(PartialAssignment("0*0*0"),
           Dimacs::parse_clause("1 -2 -3 4 5 0"), Clause(), PartialValue::TRUE)
       && success;
   success =
       TestApply(PartialAssignment("0*1*1"),
           Dimacs::parse_clause("1 -2 -3 4 5 0"), Clause(), PartialValue::TRUE)
       && success;
   return success;
}

bool
SUBool::TestNFLiteralSet::TestSubsetComparison() const
{
   const unsigned n_vars = 20;
   const unsigned size = 10;
   std::vector<Literal> literals =
       random_literal_vector_with_size(size, n_vars);
   Clause clause1(literals);
   literals.pop_back();
   std::swap(literals[0], literals.back());
   literals.pop_back();
   std::swap(literals[5], literals.back());
   literals.pop_back();
   Clause clause2(literals);
   literals.emplace_back(21, false);
   Clause clause3(literals);
   std::cout << "Comparing clauses" << std::endl;
   std::cout << "Clause 1: " << clause1 << std::endl;
   std::cout << "Clause 2: " << clause2 << std::endl;
   std::cout << "Clause 3: " << clause3 << std::endl;
   bool success = true;
   if (clause1.SubsetCompare(clause2) == SetCompareResult::SUPERSET)
   {
      std::cout << "1 SUPERSET of 2 [OK]" << std::endl;
   }
   else
   {
      std::cout << "1 SUPERSET of 2 [FAILED] "
                << static_cast<unsigned>(clause1.SubsetCompare(clause2))
                << std::endl;
      success = false;
   }
   if (clause2.SubsetCompare(clause1) == SetCompareResult::SUBSET)
   {
      std::cout << "2 SUBSET of 1 [OK]" << std::endl;
   }
   else
   {
      std::cout << "2 SUBSET of 1 [FAILED] "
                << static_cast<unsigned>(clause2.SubsetCompare(clause1))
                << std::endl;
      success = false;
   }
   if (clause1.SubsetCompare(clause3) == SetCompareResult::INCOMPARABLE)
   {
      std::cout << "1 INCOMPARABLE to 3 [OK]" << std::endl;
   }
   else
   {
      std::cout << "1 INCOMPARABLE to 3 [FAILED] "
                << static_cast<unsigned>(clause1.SubsetCompare(clause3))
                << std::endl;
      success = false;
   }
   if (clause2.SubsetCompare(clause2) == SetCompareResult::EQUAL)
   {
      std::cout << "2 EQUAL to 2 [OK]" << std::endl;
   }
   else
   {
      std::cout << "2 EQUAL to 2 [FAILED] "
                << static_cast<unsigned>(clause2.SubsetCompare(clause2))
                << std::endl;
      success = false;
   }
   return success;
}

bool
SUBool::TestNFLiteralSet::TestLT(
    const Clause &a, const Clause &b, bool result) const
{
   std::cout << "Checking if (" << a << ")<(" << b << ")";
   if ((a < b) == result)
   {
      std::cout << " [OK]" << std::endl;
      return true;
   }
   std::cout << " [FAILED]" << std::endl;
   return false;
}

bool
SUBool::TestNFLiteralSet::TestGT(
    const Clause &a, const Clause &b, bool result) const
{
   std::cout << "Checking if (" << a << ")>(" << b << ")";
   if ((a > b) == result)
   {
      std::cout << " [OK]" << std::endl;
      return true;
   }
   std::cout << " [FAILED]" << std::endl;
   return false;
}

bool
SUBool::TestNFLiteralSet::TestLE(
    const Clause &a, const Clause &b, bool result) const
{
   std::cout << "Checking if (" << a << ")<=(" << b << ")";
   if ((a <= b) == result)
   {
      std::cout << " [OK]" << std::endl;
      return true;
   }
   std::cout << " [FAILED]" << std::endl;
   return false;
}

bool
SUBool::TestNFLiteralSet::TestGE(
    const Clause &a, const Clause &b, bool result) const
{
   std::cout << "Checking if (" << a << ")>=(" << b << ")";
   if ((a >= b) == result)
   {
      std::cout << " [OK]" << std::endl;
      return true;
   }
   std::cout << " [FAILED]" << std::endl;
   return false;
}

bool
SUBool::TestNFLiteralSet::TestCompare(
    const Clause &a, const Clause &b, int result) const
{
   return TestGT(a, b, (result > 0)) && TestLT(a, b, (result < 0))
          && TestGE(a, b, (result >= 0)) && TestLE(a, b, (result <= 0));
}

bool
SUBool::TestNFLiteralSet::TestComparison() const
{
   return TestCompare(Dimacs::parse_clause("1 2 3 0"),
              Dimacs::parse_clause("1 2 0"), +1)
          && TestCompare(
              Dimacs::parse_clause("-1 3 7 0"), Dimacs::parse_clause("-3"), -1)
          && TestCompare(Dimacs::parse_clause("3 4 7 0"),
              Dimacs::parse_clause("7 4 3 0"), 0)
          && TestCompare(Dimacs::parse_clause("1 -5 6 0"),
              Dimacs::parse_clause("1 5 6 0"), +1)
          && TestCompare(Dimacs::parse_clause("1 5 6 0"),
              Dimacs::parse_clause("1 -5 6 0"), -1)
          && TestCompare(Dimacs::parse_clause("3 5 -7 0"),
              Dimacs::parse_clause("4 7 -8 0"), -1);
}

bool
SUBool::TestNFLiteralSet::TestNegateLiterals() const
{
   std::cout << "Test NegateLiterals ";
   for (unsigned count = 0; count < 100; ++count)
   {
      Clause clause(random_literal_vector_with_size(20, 50));
      //   std::cout<<"input: "<<clause<<std::endl;
      auto nclause = clause;
      nclause.NegateLiterals();
      //  std::cout<<"output: "<<nclause<<std::endl;
      if (nclause.Size() != clause.Size())
      {
         std::cout << "-- Size mismatch [FAILED]" << std::endl;
         return false;
      }
      for (unsigned i = 0; i < clause.Size(); ++i)
      {
         if (nclause[i].Variable() != clause[i].Variable())
         {
            std::cout << "-- Variable mismatch [FAILED]" << std::endl;
            return false;
         }
         if (nclause[i].Positive() == clause[i].Positive())
         {
            std::cout << "-- Polarity mismatch [FAILED]" << std::endl;
            return false;
         }
      }
   }
   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestNFLiteralSet::TestPushback() const
{
   auto literals = random_literal_vector_with_size(12, 24);
   std::cout << "Testing NFLiteralSet::PushBack" << std::endl;
   std::cout << "Literals:";
   Clause c1;
   for (auto lit : literals)
   {
      std::cout << " " << lit;
      c1.PushBack(lit);
   }
   for (auto lit : literals)
   {
      c1.PushBack(lit);
   }
   Clause c2(literals);
   std::cout << std::endl;
   std::cout << "Clause:     " << c2 << std::endl;
   std::cout << "ClausePush: " << c2 << std::endl;
   if (c1 != c2)
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestNFLiteralSet::RunTests() const
{
   bool success = true;
   success = TestDeterminingValue(false) && success;
   success = TestDeterminingValue(true) && success;
   success = TestClauseDeterminingValue() && success;
   success = TestTermDeterminingValue() && success;
   success = TestClauseSize() && success;
   success = TestLiteralAccess() && success;
   success = TestEvaluationSimple() && TestEvaluationRandom() && success;
   success = TestDIMACS() && success;
   success = TestLiteralSetPolarities() && success;
   success = TestApplyRenamingToLiteralSet() && success;
   success = TestSubsetComparison() && success;
   success = TestComparison() && success;
   success = TestApply() && success;
   success = TestNegateLiterals() && success;
   success = TestPushback() && success;
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::TestNFLiteralSet test;
   return !test.RunTests();
}
