/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file test_nnfnode.cpp
 * Contains tests for Nnf and NnfNode classes.
 */

#include <sstream>

#include "dimacs.h"
#include "dnnf.h"
#include "dnnf_enc.h"
#include "nnf.h"
#include "nnfnode.h"
#include "normalform.h"
#include "test_nnfnode.h"

bool
SUBool::TestNnfNode::TestNnfNodeLiteral() const
{
   auto l = NnfNode::MakeLiteral(4, false);
   std::cout << *l << std::endl;
   Assignment alpha(6);
   alpha[4] = false;
   std::cout << "Value on " << alpha[4] << "=" << (*l)(alpha) << std::endl;
   alpha[4] = true;
   std::cout << "Value on " << alpha[4] << "=" << (*l)(alpha) << std::endl;
   PartialAssignment beta(6);
   std::cout << "Value on " << beta[4] << "=" << (*l)(beta) << std::endl;
   beta[4] = PartialValue::TRUE;
   std::cout << "Value on " << beta[4] << "=" << (*l)(beta) << std::endl;
   beta[4] = PartialValue::FALSE;
   std::cout << "Value on " << beta[4] << "=" << (*l)(beta) << std::endl;
   return true;
}

const char *example_nnfnode_str = "L 1\n"
                                  "L -2\n"
                                  "L -3\n"
                                  "A 3 0 1 2\n"
                                  "L 3\n"
                                  "O 3 2 3 4\n"
                                  "L -4\n"
                                  "A 2 5 6\n"
                                  "L 4\n"
                                  "A 2 0 8\n"
                                  "A 2 1 4\n"
                                  "L 2\n"
                                  "O 2 2 10 11\n"
                                  "A 2 9 12\n"
                                  "O 4 2 7 13";

const char *example_nnf_str = "nnf 15 17 4\n"
                              "L 1\n"
                              "L -2\n"
                              "L -3\n"
                              "A 3 0 1 2\n"
                              "L 3\n"
                              "O 3 2 3 4\n"
                              "L -4\n"
                              "A 2 5 6\n"
                              "L 4\n"
                              "A 2 0 8\n"
                              "A 2 1 4\n"
                              "L 2\n"
                              "O 2 2 10 11\n"
                              "A 2 9 12\n"
                              "O 4 2 7 13";

const char *example_dec_nnf_str = "nnf 9 10 4\n"
                                  "L 1\n"
                                  "A 0\n"
                                  "L 3\n"
                                  "D 2 1 2\n"
                                  "A 2 0 3\n"
                                  "L -2\n"
                                  "A 2 0 5\n"
                                  "D 3 1 6\n"
                                  "D 4 4 7";

const char *example_decision_str = "nnf 10 12 4\n"
                                   "L 4\n"
                                   "L 3\n"
                                   "O 0 2 0 1\n"
                                   "L -3\n"
                                   "D 2 1 3\n"
                                   "A 2 2 4\n"
                                   "L -4\n"
                                   "O 0 2 6 3\n"
                                   "A 2 4 7\n"
                                   "D 1 5 8";

SUBool::NnfNode::tPointer
SUBool::TestNnfNode::ExampleNnf() const
{
   auto l3n = NnfNode::MakeLiteral(3, false);
   auto l2n = NnfNode::MakeLiteral(2, false);
   auto l1p = NnfNode::MakeLiteral(1, true);
   auto a3 = NnfNode::MakeAnd({l1p, l2n, l3n});
   auto l3p = NnfNode::MakeLiteral(3, true);
   auto o5 = NnfNode::MakeOr(3, a3, l3p);
   auto l4n = NnfNode::MakeLiteral(4, false);
   auto a7 = NnfNode::MakeAnd(o5, l4n);
   auto l4p = NnfNode::MakeLiteral(4, true);
   auto a9 = NnfNode::MakeAnd(l1p, l4p);
   auto a10 = NnfNode::MakeAnd(l2n, l3p);
   auto l2p = NnfNode::MakeLiteral(2, true);
   auto o12 = NnfNode::MakeOr(2, a10, l2p);
   auto a13 = NnfNode::MakeAnd(a9, o12);
   auto o14 = NnfNode::MakeOr(4, a7, a13);
   return o14;
}

const char *example_cnf_str = "p cnf 4 4\n"
                              "1 2 3 0\n"
                              "-2 3 4 0\n"
                              "1 -4 0\n"
                              "2 3 -4 0";

bool
SUBool::TestNnfNode::TestNnfNodeGates() const
{
   std::cout << "Testing NnfNode Gates" << std::endl;
   auto nnf = ExampleNnf();
   std::stringstream ss;
   ss << *nnf;
   std::string expected_out = example_nnfnode_str;
   std::cout << ss.str() << std::endl;
   if (expected_out != ss.str())
   {
      std::cout << "Construction failed [FAILED]" << std::endl;
      return false;
   }
   std::cout << "Construction succeeded [OK]" << std::endl;
   std::cout << "Checking equivalence ";
   auto cnf = Dimacs::parse_cnf(example_cnf_str);
   for (auto i = 0; i < 16; ++i)
   {
      Assignment alpha(4);
      alpha[1] = (i % 2 == 1);
      alpha[2] = ((i >> 1) % 2 == 1);
      alpha[3] = ((i >> 2) % 2 == 1);
      alpha[4] = ((i >> 3) % 2 == 1);
      PartialAssignment beta(alpha);
      if (cnf.Value(alpha) != nnf->Value(alpha)
          || cnf.Value(beta) != nnf->Value(beta))
      {
         std::cout << "[FAILED]" << std::endl;
         return false;
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestNnfNode::TestNnfWrite() const
{
   std::cout << "Testing Nnf write capabilities";
   auto nnf = Nnf(ExampleNnf());
   std::stringstream ss;
   ss << nnf;
   std::string expected_out = "nnf 15 17 4\n";
   expected_out += example_nnfnode_str;
   if (expected_out != ss.str())
   {
      std::cout << " [FAILED]" << std::endl;
      std::cout << ss.str() << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   std::cout << ss.str() << std::endl;
   return true;
}

bool
SUBool::TestNnfNode::TestNnfRead() const
{
   std::cout << "Testing Nnf read capabilities";
   std::stringstream nnf_stream(example_nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   std::stringstream out;
   out << nnf;
   if (std::string(out.str()) != example_nnf_str)
   {
      std::cout << " [FAILED]" << std::endl;
      std::cout << out.str() << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   std::cout << out.str() << std::endl;
   return true;
}

bool
SUBool::TestNnfNode::TestNnfDecision() const
{
   std::cout << "Testing Nnf with decision nodes" << std::endl;
   std::stringstream nnf_stream(example_dec_nnf_str);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   std::stringstream out;
   out << nnf;
   if (std::string(out.str()) != example_dec_nnf_str)
   {
      std::cout << "Parsing and writing [FAILED]" << std::endl;
      std::cout << out.str() << std::endl;
      return false;
   }
   std::cout << "Parsing and writing [OK]" << std::endl;
   std::cout << out.str() << std::endl;
   std::stringstream ex_nnf_stream(example_nnf_str);
   auto ex_nnf = Nnf::ReadFromStream(ex_nnf_stream);
   std::cout << "Checking values on full assignments";
   Assignment alpha(4);
   do
   {
      if (ex_nnf.Value(alpha) != nnf.Value(alpha))
      {
         std::cout << " [FAILED]" << std::endl;
         std::cout << "... Mismatch on " << alpha << " ( is "
                   << nnf.Value(alpha) << ", should be " << ex_nnf.Value(alpha)
                   << ")" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   std::cout << "Checking values on partial assignments";
   PartialAssignment beta(4);
   do
   {
      if (ex_nnf.Value(beta) != nnf.Value(beta))
      {
         std::cout << " [FAILED]" << std::endl;
         std::cout << "... Mismatch on " << beta << " ( is " << nnf.Value(beta)
                   << ", should be " << ex_nnf.Value(beta) << ")" << std::endl;
         return false;
      }
   } while (beta.Inc() && beta[1] != PartialValue::TRUE);
   // When beta[1] is true, decision nodes offer stronger
   // inference of true values.
   std::cout << " [OK]" << std::endl;

   return true;
}

bool
SUBool::TestNnfNode::TestNnfCopyConstructor() const
{
   std::cout << "Testing Nnf copy constructor" << std::endl;
   std::stringstream nnf_stream(example_decision_str);
   const auto nnf = Nnf::ReadFromStream(nnf_stream);
   auto nnf_copy(nnf);
   std::cout << "Source nnf:" << std::endl;
   std::cout << nnf << std::endl;
   std::cout << "Nnf copy:" << std::endl;
   std::cout << nnf_copy << std::endl;
   std::stringstream nnf_copy_stream;
   nnf_copy_stream << nnf_copy;
   if (std::string(example_decision_str) != nnf_copy_stream.str())
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestNnfNode::TestNnfCopyAssignment() const
{
   std::cout << "Testing Nnf copy assignment" << std::endl;
   std::stringstream nnf_stream(example_decision_str);
   const auto nnf = Nnf::ReadFromStream(nnf_stream);
   Nnf nnf_copy(NnfNode::MakeFalse());
   nnf_copy = nnf;
   std::cout << "Source nnf:" << std::endl;
   std::cout << nnf << std::endl;
   std::cout << "Nnf copy:" << std::endl;
   std::cout << nnf_copy << std::endl;
   std::stringstream nnf_copy_stream;
   nnf_copy_stream << nnf_copy;
   if (std::string(example_decision_str) != nnf_copy_stream.str())
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestNnfNode::RunTests() const
{
   return TestNnfNodeLiteral() && TestNnfNodeGates() && TestNnfWrite()
          && TestNnfRead() && TestNnfDecision() && TestNnfCopyConstructor()
          && TestNnfCopyAssignment();
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   return !SUBool::TestNnfNode().RunTests();
}
