/*
 * Project SUBool
 * Petr Kucera, 2018
 */
/**@file test_nnftraits.h
 * Contains tests for nnf traits checking and manipulation (defined in dnnf).
 */

#include <sstream>

#include "dnnf.h"
#include "dnnf_enc.h"
#include "nnf.h"
#include "nnfnode.h"
#include "normalform.h"
#include "test_nnftraits.h"

const char *nnf_and_decomp = "nnf 3 2 2\n"
                             "L 1\n"
                             "L 2\n"
                             "A 2 0 1";
const char *nnf_and_not_decomp = "nnf 3 2 2\n"
                                 "L 1\n"
                                 "L -1\n"
                                 "A 2 0 1";

const char *example_nnfnode_str = "L 1\n"
                                  "L -2\n"
                                  "L -3\n"
                                  "A 3 0 1 2\n"
                                  "L 3\n"
                                  "O 3 2 3 4\n"
                                  "L -4\n"
                                  "A 2 5 6\n"
                                  "L 4\n"
                                  "A 2 0 8\n"
                                  "A 2 1 4\n"
                                  "L 2\n"
                                  "O 2 2 10 11\n"
                                  "A 2 9 12\n"
                                  "O 4 2 7 13";

const char *example_nnf_str = "nnf 15 17 4\n"
                              "L 1\n"
                              "L -2\n"
                              "L -3\n"
                              "A 3 0 1 2\n"
                              "L 3\n"
                              "O 3 2 3 4\n"
                              "L -4\n"
                              "A 2 5 6\n"
                              "L 4\n"
                              "A 2 0 8\n"
                              "A 2 1 4\n"
                              "L 2\n"
                              "O 2 2 10 11\n"
                              "A 2 9 12\n"
                              "O 4 2 7 13";

const char *example_dec_nnf_str = "nnf 9 10 4\n"
                                  "L 1\n"
                                  "A 0\n"
                                  "L 3\n"
                                  "D 2 1 2\n"
                                  "A 2 0 3\n"
                                  "L -2\n"
                                  "A 2 0 5\n"
                                  "D 3 1 6\n"
                                  "D 4 4 7";

const char *example_decision_str = "nnf 10 12 4\n"
                                   "L 4\n"
                                   "L 3\n"
                                   "O 0 2 0 1\n"
                                   "L -3\n"
                                   "D 2 1 3\n"
                                   "A 2 2 4\n"
                                   "L -4\n"
                                   "O 0 2 6 3\n"
                                   "A 2 4 7\n"
                                   "D 1 5 8";

const char *example_smooth_nnf_str = "nnf 21 28 4\n"
                                     "L 1\n"
                                     "L 2\n"
                                     "L 3\n"
                                     "L 4\n"
                                     "L -2\n"
                                     "L -3\n"
                                     "L -4\n"
                                     "L -1\n"
                                     "O 1 2 0 7\n"
                                     "O 2 2 1 4\n"
                                     "O 3 2 2 5\n"
                                     "A 2 4 2\n"
                                     "A 2 1 10\n"
                                     "O 2 2 12 11\n"
                                     "A 2 0 3\n"
                                     "A 2 13 14\n"
                                     "A 3 0 4 5\n"
                                     "A 3 2 9 8\n"
                                     "O 3 2 17 16\n"
                                     "A 2 6 18\n"
                                     "O 4 2 15 19";

const char *example_smooth_decision_nnf_str = "nnf 19 26 4\n"
                                              "L 1\n"         // 0
                                              "L 2\n"         // 1
                                              "L 3\n"         // 2
                                              "L -1\n"        // 3
                                              "L -2\n"        // 4
                                              "L -3\n"        // 5
                                              "O 1 2 0 3\n"   // 6 | 1 -1
                                              "O 2 2 1 4\n"   // 7 | 2 -2
                                              "O 3 2 2 5\n"   // 8 | 3 -3
                                              "A 2 1 8\n"     // 9
                                              "A 2 2 7\n"     // 10
                                              "O 0 2 9 10\n"  // 11
                                              "A 2 0 11\n"    // 12
                                              "A 2 2 6\n"     // 13
                                              "A 2 0 8\n"     // 14
                                              "O 0 2 14 13\n" // 15
                                              "A 2 2 6\n"     // 16
                                              "D 2 16 15\n"   // 17
                                              "D 4 12 17";    // 18

const char *example_smooth_dnnf_monotone_x = "nnf 11 11 3\n"
                                             "L 1\n"       // 0
                                             "L 2\n"       // 1
                                             "L 3\n"       // 2
                                             "L -1\n"      // 3
                                             "L -2\n"      // 4
                                             "L -3\n"      // 5
                                             "O 2 2 1 4\n" // 6
                                             "O 2 2 2 5\n" // 7
                                             "A 3 0 1 7\n" // 8
                                             "A 3 0 2 6\n" // 9
                                             "O 0 2 8 9";  // 10

bool
SUBool::TestNnfTraits::TestNnfTraitsCheck(const char *nnf_string,
                                          bool decomposable, bool smooth,
                                          bool decision_nnf) const
{
   std::cout << "Testing NNF traits." << std::endl;
   std::stringstream nnf_stream(nnf_string);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   std::cout << "Input NNF:";
   std::cout << nnf << std::endl;
   auto traits = nnf_traits(nnf);
   std::cout << "Decomposable: " << traits.decomposable
             << (traits.decomposable == decomposable ? " [OK]" : " [FAILED]")
             << std::endl;
   std::cout << "Smooth: " << traits.smooth
             << (traits.smooth == smooth ? " [OK]" : " [FAILED]") << std::endl;
   std::cout << "Decision NNF: " << traits.decision_nnf
             << (traits.decision_nnf == decision_nnf ? " [OK]" : " [FAILED]")
             << std::endl;
   return traits.decomposable == decomposable && traits.smooth == smooth
          && traits.decision_nnf == decision_nnf;
}

bool
SUBool::TestNnfTraits::TestNnfTraitsCheck() const
{
   bool success = true;
   success = success && TestNnfTraitsCheck(nnf_and_decomp, true, true, true);
   success =
       success && TestNnfTraitsCheck(nnf_and_not_decomp, false, true, true);
   success = success && TestNnfTraitsCheck(example_nnf_str, true, false, false);
   success =
       success && TestNnfTraitsCheck(example_dec_nnf_str, true, false, true);
   success =
       success && TestNnfTraitsCheck(example_smooth_nnf_str, true, true, false);
   success = success
             && TestNnfTraitsCheck(example_smooth_decision_nnf_str, true, true,
                                   false);
   success =
       success
       && TestNnfTraitsCheck(example_smooth_dnnf_monotone_x, true, true, false);
   return success;
}

bool
SUBool::TestNnfTraits::TestNnfLiteralNodes(
    const char *nnf_string,
    const std::unordered_set<Literal> &literal_nodes) const
{
   std::cout << "Testing literal nodes collection" << std::endl;
   std::stringstream nnf_stream(nnf_string);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   std::cout << "Input NNF:" << std::endl;
   std::cout << nnf << std::endl;
   NnfNodeInputsVisitor visitor(true);
   nnf.AcceptPostorder(visitor);
   auto v_literal_nodes = std::move(visitor).LiteralNodes();
   if (literal_nodes.size() != v_literal_nodes.size())
   {
      std::cout << "[FAILED]" << std::endl;
      std::cout << "... size mismatch (should be " << literal_nodes.size()
                << ", but is " << v_literal_nodes.size() << std::endl;
      return false;
   }
   for (const auto &lit : literal_nodes)
   {
      if (v_literal_nodes.count(lit) == 0)
      {
         std::cout << "[FAILED]" << std::endl;
         std::cout << "... literal " << lit
                   << " not contain in literals determined by visitor"
                   << std::endl;
         return false;
      }
      if (lit != v_literal_nodes[lit]->GetLiteral())
      {
         std::cout << "[FAILED]" << std::endl;
         std::cout << "... unexpected literal" << std::endl;
         return false;
      }
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestNnfTraits::TestNnfLiteralNodes() const
{
   bool success = true;
   success = success
             && TestNnfLiteralNodes(nnf_and_decomp,
                                    {Literal(1, true), Literal(2, true)});
   success = success
             && TestNnfLiteralNodes(nnf_and_not_decomp,
                                    {Literal{1, true}, Literal{1, false}});
   success = success
             && TestNnfLiteralNodes(example_nnf_str,
                                    {Literal{1, true}, Literal{2, false},
                                     Literal{3, false}, Literal{3, true},
                                     Literal{4, false}, Literal{4, true},
                                     Literal{2, true}});
   success = success
             && TestNnfLiteralNodes(example_decision_str,
                                    {Literal{4, true}, Literal{3, true},
                                     Literal{3, false}, Literal{4, false}});
   return true;
}

bool
SUBool::TestNnfTraits::TestNnfSmooth(const char *nnf_string) const
{
   std::cout << "Testing nnf smooth transformation" << std::endl;
   std::stringstream nnf_stream(nnf_string);
   auto nnf = Nnf::ReadFromStream(nnf_stream);
   std::cout << "Input NNF:" << std::endl;
   std::cout << nnf << std::endl;
   auto smooth_nnf = nnf_smooth(nnf);
   std::cout << "Output NNF:" << std::endl;
   std::cout << smooth_nnf << std::endl;
   auto traits = nnf_traits(smooth_nnf);
   if (!traits.smooth)
   {
      std::cout << "It is not smooth [FAILED]" << std::endl;
      return false;
   }
   std::cout << "It is smooth [OK]" << std::endl;
   std::cout << "Checking values on full assignments";
   Assignment alpha(4);
   do
   {
      if (smooth_nnf.Value(alpha) != nnf.Value(alpha))
      {
         std::cout << " [FAILED]" << std::endl;
         std::cout << "... Mismatch on " << alpha << " ( is "
                   << smooth_nnf.Value(alpha) << ", should be "
                   << nnf.Value(alpha) << ")" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << " [OK]" << std::endl;
   std::cout << "Checking values on partial assignments";
   PartialAssignment beta(4);
   do
   {
      auto sv = smooth_nnf.Value(beta);
      auto v = nnf.Value(beta);
      /* Values can be different on partial assignments */
      if (sv != v && sv != PartialValue::UNDEF && v != PartialValue::FALSE)
      {
         std::cout << " [FAILED]" << std::endl;
         std::cout << "... Mismatch on " << beta << " ( is "
                   << smooth_nnf.Value(beta) << ", should be "
                   << nnf.Value(beta) << ")" << std::endl;
         return false;
      }
   } while (beta.Inc());
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestNnfTraits::TestNnfSmooth() const
{
   return TestNnfSmooth(example_nnf_str) && TestNnfSmooth(example_decision_str);
}

bool
SUBool::TestNnfTraits::RunTests() const
{
   return TestNnfTraitsCheck() && TestNnfLiteralNodes() && TestNnfSmooth();
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   return !SUBool::TestNnfTraits().RunTests();
}
