/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file test_normalform.cpp
 * Contains tests for NormalForm.
 */

#include <cmath>
#include <iostream>
#include <sstream>
#include <vector>

#include "dimacs.h"
#include "test_normalform.h"

const char *dimacs_simple = "c\n"
                            "c start with comments\n"
                            "c\n"
                            "c\n"
                            "p cnf 5 3\n"
                            "1 -5 4 0\n"
                            "c comment in the middle\n"
                            "-1 5 3 4 0\n"
                            "-3 -4 0\n"
                            "c at the end and there is an empty line below\n";

const char *dimacs_quinn = "c  quinn.cnf\n"
                           "c\n"
                           "p cnf 16 18\n"
                           "  1    2  0\n"
                           " -2   -4  0\n"
                           "  3    4  0\n"
                           " -4   -5  0\n"
                           "  5   -6  0\n"
                           "  6   -7  0\n"
                           "  6    7  0\n"
                           "  7  -16  0\n"
                           "  8   -9  0\n"
                           " -8  -14  0\n"
                           "  9   10  0\n"
                           "  9  -10  0\n"
                           "-10  -11  0\n"
                           " 10   12  0\n"
                           " 11   12  0\n"
                           " 13   14  0\n"
                           " 14  -15  0\n"
                           " 15   16  0";

const char *dimacs_horn = "c\n"
                          "c A horn formula\n"
                          "c\n"
                          "p cnf 7 8\n"
                          "2 0\n"
                          "-4 0\n"
                          "5 -7 -2 0\n"
                          "-1 3 -4 0\n"
                          "-5 -4 -2 1 0\n"
                          "3 -1 -2 0\n"
                          "5 -2 0\n"
                          "1 -2 -3 -4 0";

bool
SUBool::TestNormalForm::TestCNFDeterminingValue() const
{
   std::cout << "Test determining value of a CNF ";
   bool success = true;
   if (!CNF::DeterminingValue())
   {
      std::cout << "[PASSED]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
      success = false;
   }
   return success;
}

bool
SUBool::TestNormalForm::TestDNFDeterminingValue() const
{
   std::cout << "Test determining value of a DNF ";
   bool success = true;
   if (DNF::DeterminingValue())
   {
      std::cout << "[PASSED]" << std::endl;
   }
   else
   {
      std::cout << "[FAILED]" << std::endl;
      success = false;
   }
   return success;
}

bool
SUBool::TestNormalForm::TestDIMACSSimple() const
{
   std::cout << "Test DIMACS format for CNF.";
   auto cnf = Dimacs::parse_cnf(dimacs_simple);
   std::cout << "Input: " << std::endl;
   std::cout << dimacs_simple << std::endl;
   std::cout << "Output: " << cnf;
   if (cnf.MaxVariable() != 5)
   {
      std::cout << " -- incorrect number of variables [FAILED]" << std::endl;
      return false;
   }
   if (cnf.Size() != 3)
   {
      std::cout << " -- incorrect number of clauses [FAILED]" << std::endl;
      return false;
   }
   std::cout << " [PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestNormalForm::TestDIMACSQuinn() const
{
   std::cout << "Test DIMACS format for CNF.";
   auto cnf = Dimacs::parse_cnf(dimacs_quinn);
   std::cout << "Input: " << std::endl;
   std::cout << dimacs_quinn << std::endl;
   std::cout << "Output: " << cnf;
   if (cnf.MaxVariable() != 16)
   {
      std::cout << " -- incorrect number of variables [FAILED]" << std::endl;
      return false;
   }
   if (cnf.Size() != 18)
   {
      std::cout << " -- incorrect number of clauses [FAILED]" << std::endl;
      return false;
   }
   std::cout << " [PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestNormalForm::TestEvaluationSimple() const
{
   std::cout << "Test evaluation of CNF." << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs_simple);
   Assignment sat_assignment(5, false);
   Assignment unsat_assignment(5, false);
   unsat_assignment[5] = true;

   if (!cnf.Value(sat_assignment))
   {
      std::cout << "Value on a satisfying assignment is false [FAILED]"
                << std::endl;
      return false;
   }
   if (cnf.Value(unsat_assignment))
   {
      std::cout << "Value on a falsifying assignment is true [FAILED]"
                << std::endl;
      return false;
   }
   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestNormalForm::TestClauseAccess() const
{
   std::cout << "Test CNF clause access." << std::endl;
   std::vector<Clause> clauses;
   clauses.push_back(Dimacs::parse_clause("1 -5 4 0"));
   clauses.push_back(Dimacs::parse_clause("-1 5 3 4 0"));
   clauses.push_back(Dimacs::parse_clause("-3 -4 0"));
   CNF cnf(clauses);
   if (cnf[0][2].mVariable != 5)
   {
      std::cout << "3rd literal in 1st clause is not on variable 5 [FAILED]"
                << std::endl;
      return false;
   }
   if (cnf[1][1].mVariable != 3)
   {
      std::cout << "2nd literal in 2nd clause is not on variable 3 [FAILED]"
                << std::endl;
      return false;
   }
   if (cnf[2][0].mVariable != 3)
   {
      std::cout << "1st literal in 3rd clause is not on variable 3 [FAILED]"
                << std::endl;
      return false;
   }
   std::cout << "Input clauses" << std::endl;
   for (const auto &clause : clauses)
   {
      std::cout << clause << std::endl;
   }
   std::cout << "CNF Clauses: " << std::endl;
   for (const auto &clause : cnf)
   {
      std::cout << clause << std::endl;
   }
   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestNormalForm::TestApplyRenamingWithDIMACS(
    const char *dimacs, const Assignment &renaming) const
{
   auto cnf = Dimacs::parse_cnf(dimacs);
   CNF ren_cnf(cnf);
   cnf.ApplyRenaming(renaming);
   std::vector<Clause> ren_clauses;
   for (const auto &clause : cnf)
   {
      std::vector<Literal> literals;
      for (const auto &lit : clause)
      {
         literals.push_back(lit);
         if (renaming[lit.mVariable])
         {
            literals.back().Negate();
         }
      }
      Clause new_clause(literals);
      ren_clauses.push_back(new_clause);
   }
   CNF my_renamed_cnf(ren_clauses);
   return (my_renamed_cnf == ren_cnf);
}

bool
SUBool::TestNormalForm::TestApplyRenaming() const
{
   std::cout << "Test apply renaming" << std::endl;

   Assignment ren_simple(5, false);
   ren_simple[4] = true;
   ren_simple[3] = true;
   std::cout << "Simple dimacs ";
   if (!TestApplyRenamingWithDIMACS(dimacs_simple, ren_simple))
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   else
   {
      std::cout << "[PASSED]" << std::endl;
   }

   Assignment ren_horn(7, true);
   std::cout << "Horn dimacs ";
   if (!TestApplyRenamingWithDIMACS(dimacs_horn, ren_horn))
   {
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   else
   {
      std::cout << "[PASSED]" << std::endl;
   }

   return true;
}

const char *dimacs_apply = "c\n"
                           "p cnf 7 8\n"
                           "1 2 3 -5 0\n"
                           "-3 4 5 -7 0\n"
                           "7 0\n"
                           "5 -6 0\n"
                           "-3 4 6 0\n"
                           "1 3 -4 0\n"
                           "-3 7 0\n"
                           "5 6 -7 0\n";

const char *alpha_apply_1 = "1******";
const char *dimacs_apply_1 = "c\n"
                             "p cnf 7 6\n"
                             "-3 4 5 -7 0\n"
                             "7 0\n"
                             "5 -6 0\n"
                             "-3 4 6 0\n"
                             "-3 7 0\n"
                             "5 6 -7 0\n";

const char *alpha_apply_2 = "1***01*";
const char *dimacs_apply_2 = "c\n"
                             "p cnf 0 1\n"
                             "0\n";

const char *alpha_apply_3 = "1*0*1*1";
const char *dimacs_apply_3 = "c\n"
                             "p cnf 0 0\n";

bool
SUBool::TestNormalForm::TestApply(const PartialAssignment &alpha,
    const std::string &dimacs_in, const std::string &dimacs_out,
    PartialValue val) const
{
   std::cout << "Testing application of a partial assignment." << std::endl;
   auto cnf_in = Dimacs::parse_cnf(dimacs_in);
   auto cnf_out = Dimacs::parse_cnf(dimacs_out);
   std::cout << "Input cnf:" << std::endl;
   std::cout << cnf_in << std::endl;

   std::cout << "Trying assignment " << alpha << std::endl;
   CNF cnf(cnf_in);
   PartialValue v = cnf.Apply(alpha);
   if (v != val)
   {
      std::cout << "Wrong value: " << v << " (expected " << val << ") [FAILED]"
                << std::endl;
      return false;
   }
   if (cnf != cnf_out)
   {
      std::cout << "Wrong cnf: " << std::endl;
      std::cout << cnf << std::endl;
      std::cout << "Expected cnf: " << std::endl;
      std::cout << cnf_out << std::endl;
      std::cout << "[FAILED]" << std::endl;
      return false;
   }
   std::cout << " [OK]" << std::endl;
   return true;
}

bool
SUBool::TestNormalForm::TestApply() const
{
   TestApply(PartialAssignment(7, PartialValue::UNDEF), dimacs_apply,
       dimacs_apply, PartialValue::UNDEF);
   TestApply(PartialAssignment(alpha_apply_1), dimacs_apply, dimacs_apply_1,
       PartialValue::UNDEF);
   TestApply(PartialAssignment(alpha_apply_2), dimacs_apply, dimacs_apply_2,
       PartialValue::FALSE);
   TestApply(PartialAssignment(alpha_apply_3), dimacs_apply, dimacs_apply_3,
       PartialValue::TRUE);
   return false;
}

const char *subsum_cnf_a = "c\n"
                           "p cnf 15 5\n"
                           "1 2 3 4 5 0\n"
                           "2 3 4 5 0\n"
                           "3 4 5 0\n"
                           "4 5 0\n"
                           "5 0\n";

bool
SUBool::TestNormalForm::TestSubsumptions() const
{
   std::cout << "Testing adding without subsumptions" << std::endl;
   auto cnf = Dimacs::parse_cnf(subsum_cnf_a);
   CNF a_cnf;
   a_cnf.PushBack(cnf[0]);
   std::cout << "Start with cnf:\n" << a_cnf << std::endl;
   for (unsigned i = 1; i < cnf.Size(); ++i)
   {
      std::cout << "Adding clause " << cnf[i];
      if (a_cnf.AddWithoutSubsumptions(cnf[i]))
      {
         std::cout << " [OK]" << std::endl;
      }
      else
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
   }
   std::cout << "Final CNF:\n" << a_cnf << std::endl;

   CNF b_cnf;
   b_cnf.PushBack(cnf[cnf.Size() - 1]);
   std::cout << "Start with cnf:\n" << b_cnf << std::endl;
   for (unsigned i = cnf.Size() - 1; i > 0; --i)
   {
      std::cout << "Adding clause " << cnf[i - 1];
      if (b_cnf.AddWithoutSubsumptions(cnf[i - 1]))
      {
         std::cout << " [FAILED]" << std::endl;
         return false;
      }
      else
      {
         std::cout << " [OK]" << std::endl;
      }
   }
   std::cout << "Final CNF:\n" << b_cnf << std::endl;

   return true;
}

const char *dupl_cnf = "c\n"
                       "p cnf 15 8\n"
                       "1 2 3 4 5 0\n"
                       "3 4 5 0\n"
                       "-3 -4 -5 0\n"
                       "1 2 3 4 5 0\n"
                       "3 4 -5 0\n"
                       "-3 -4 -5 0\n"
                       "3 4 5 0\n"
                       "1 2 3 4 5 0\n";

bool
SUBool::TestNormalForm::TestRemoveDuplicities() const
{
   std::cout << "Testing removing duplicities" << std::endl;
   auto cnf = Dimacs::parse_cnf(dupl_cnf);
   std::cout << "Input cnf: " << std::endl;
   std::cout << cnf << std::endl;
   cnf.RemoveDuplicities();
   std::cout << "After removal: " << std::endl;
   std::cout << cnf << std::endl;
   if (cnf.Size() != 4)
   {
      std::cout << "Size mismatch [FAILED]" << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestNormalForm::TestSort() const
{
   std::cout << "Testing removing duplicities" << std::endl;
   auto cnf = Dimacs::parse_cnf(dupl_cnf);
   std::cout << "Input cnf: " << std::endl;
   std::cout << cnf << std::endl;
   cnf.Sort();
   std::cout << "After sort: " << std::endl;
   std::cout << cnf << std::endl;
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestNormalForm::RunTests() const
{
   bool success = TestCNFDeterminingValue() && TestDNFDeterminingValue();
   success = TestDIMACSSimple() && success;
   success = TestDIMACSQuinn() && success;
   success = TestEvaluationSimple() && success;
   success = TestClauseAccess() && success;
   success = TestApplyRenaming() && success;
   success = TestSubsumptions() && success;
   success = TestRemoveDuplicities() && success;
   success = TestApply() && success;
   success = success && TestSort();
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::TestNormalForm test;
   return !test.RunTests();
}
