/*
 * Project SUBool
 * Petr Kucera, 2020
 */
/**@file test_propagate.cpp
 * Contains tests for semantic propagation.
 */

#include "test_propagate.h"
#include "dimacs.h"
#include "propagate.h"

const char *dimacs_simple = "p cnf 8 8\n"
                            "1 2 5 0\n"
                            "-5 3 4 0\n"
                            "1 2 6 0\n"
                            "-6 -3 4 0\n"
                            "1 2 7 0\n"
                            "-7 3 -4 0\n"
                            "1 2 8 0\n"
                            "-8 -3 -4 0\n";

bool
SUBool::TestPropagate::TestPropagatorWithAssumptions(
    const CNF &cnf, AbsPropagator &prop, const PartialAssignment &alpha) const
{
   std::cout << "Check propagator with assumptions " << alpha << std::endl;
   auto beta = prop.Propagate(alpha);
   auto beta_p = propagate2(cnf, alpha, true_pred<unsigned>);
   if (beta != beta_p)
   {
      std::cout << "[FAILED]" << std::endl;
      std::cout << "... expected: " << beta_p << std::endl;
      std::cout << "... received: " << beta << std::endl;
      return false;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestPropagate::TestSimpleSATPropagation() const
{
   std::cout << "Checking simple CNF propagation" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs_simple);
   UnitPropagatorM up(cnf);
   IncSolverWrapper solver(cnf);
   MinimalSubimplicant min_subimplicant(&cnf);
   Propagator prop(solver.GetSolverPtr(), &up, &min_subimplicant);
   PartialAssignment alpha(cnf.MaxVariable());
   do
   {
      if (!TestPropagatorWithAssumptions(cnf, prop, alpha))
      {
         std::cout << "... failed incremental propagator" << std::endl;
         return false;
      }
   } while (alpha.Inc());
   std::cout << "[OK]" << std::endl;
   return true;
}

bool
SUBool::TestPropagate::TestRandomCNFPropagation(const unsigned n_rounds,
    const unsigned n_clauses, const unsigned n_vars,
    const unsigned n_assump_rounds, const double assump_prob_false,
    const double assump_prob_true) const
{
   std::cout << "Checking propagation on random CNFs" << std::endl;
   for (unsigned round = 0; round < n_rounds; ++round)
   {
      std::cout << "round " << round << " started" << std::endl;
      auto cnf = random_cnf(n_clauses, n_vars);
      IncSolverWrapper solver(cnf);
      UnitPropagatorM up(cnf);
      MinimalSubimplicant min_subimplicant(&cnf);
      Propagator prop(solver.GetSolverPtr(), &up, &min_subimplicant);
      for (unsigned assump_round = 0; assump_round < n_assump_rounds;
           ++assump_round)
      {
         PartialAssignment alpha(n_vars, assump_prob_false, assump_prob_true);
         if (!TestPropagatorWithAssumptions(cnf, prop, alpha))
         {
            std::cout << "... failed incremental propagator" << std::endl;
            return false;
         }
      }
      std::cout << "round " << round << " finished" << std::endl;
   }
   std::cout << "[OK]" << std::endl;
   return true;
}

double
SUBool::TestPropagate::TimeOfCNFPropagation(
    AbsPropagator &prop, const std::vector<PartialAssignment> &alpha) const
{
   CPUTime cpu_time;
   unsigned cnt = 0;
   for (const auto &a : alpha)
   {
      auto beta = prop.Propagate(a);
      std::cout << ++cnt << std::endl;
   }
   return cpu_time.TimeSpan();
}

bool
SUBool::TestPropagate::CheckTimes() const
{
   auto cnf = random_sat_cnf(1000, 500);
   std::vector<PartialAssignment> alpha;
   for (unsigned i = 0; i < 100; ++i)
   {
      alpha.emplace_back(500, 0.2, 0.2);
   }
   UnitPropagatorM up(cnf.first);
   IncSolverWrapper solver(cnf.first);
   MinimalSubimplicant min_subimplicant(&cnf.first);
   Propagator prop(solver.GetSolverPtr(), &up, &min_subimplicant);
   auto time = TimeOfCNFPropagation(prop, alpha);
   std::cout << "    time inc: " << time << "s" << std::endl;
   return true;
}

bool
SUBool::TestPropagate::RunTests() const
{
   return TestSimpleSATPropagation()
          && TestRandomCNFPropagation(10, 50, 20, 20, .2, .2) && CheckTimes();
}

int
main(int /*argc*/, char * /*argv*/[])
{
   SUBool::logs.mVerbLevel = 4;
   srand(time(NULL));
   return !SUBool::TestPropagate().RunTests();
}
