/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file test_schoning.cpp
 * Contains tests for Schoning SAT algorithm.
 */

#include <iostream>
#include <sstream>

#include "dimacs.h"
#include "random_utils.h"
#include "schoning.h"
#include "test_schoning.h"

const char *dimacs_simple = "c\n"
                            "c start with comments\n"
                            "c\n"
                            "c\n"
                            "p cnf 5 3\n"
                            "1 -5 4 0\n"
                            "-1 5 3 4 0\n"
                            "-3 4 0";

const char *dimacs_horn = "c\n"
                          "c A horn formula\n"
                          "c\n"
                          "p cnf 7 8\n"
                          "2 0\n"
                          "-4 0\n"
                          "5 -7 -2 0\n"
                          "-1 3 -4 0\n"
                          "-5 -4 -2 1 0\n"
                          "3 -1 -2 0\n"
                          "5 -2 0\n"
                          "1 -2 -3 -4 0";

const char *dimacs_renhorn = "c\n"
                             "p cnf 7 8\n"
                             "2 0\n"
                             "-4 0\n"
                             "-2 5 -7 0\n"
                             "-1 -3 -4 0\n"
                             "1 -2 -4 -5 0\n"
                             "-1 -2 -3 0\n"
                             "-2 5 0\n"
                             "1 -2 3 -4 0";

const char *dimacs_nonrenhorn = "c\n"
                                "c A formula which is not renamable Horn\n"
                                "c\n"
                                "p cnf 8 8\n"
                                "1 -2 0\n"
                                "2 -3 0\n"
                                "3 1 -2 0\n"
                                "-1 4 0\n"
                                "-4 5 0\n"
                                "-5 7 3 0\n"
                                "-7 -1 5 4 0\n"
                                "8 -1 7 0";

const char *dimacs_unsat = "c\n"
                           "c A formula which is not renamable Horn\n"
                           "c\n"
                           "p cnf 8 8\n"
                           "1 -2 0\n"
                           "2 -3 0\n"
                           "3 1 -2 0\n"
                           "-1 4 0\n"
                           "-4 5 0\n"
                           "-5 7 3 0\n"
                           "-7 -1 5 4 0\n"
                           "8 -1 7 0\n"
                           "-5 -1 0\n"
                           "3 1 0";

SUBool::TestSchoning::TestSchoning() {}

bool
SUBool::TestSchoning::TestDimacsCNF(const char *dimacs, bool satisfiable) const
{
   std::cout << "Testing a DIMACS CNF" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs);
   std::cout << "Input: " << std::endl;
   std::cout << dimacs << std::endl;
   Assignment alpha(cnf.MaxVariable());
   if (Schoning::SAT(cnf, &alpha) != satisfiable)
   {
      std::cout << "Schoning SAT claims that the cnf ";
      std::cout << (satisfiable ? "is not" : "is") << " satisfiable while it ";
      std::cout << (satisfiable ? "is" : "is not") << " satisfiable. [FAILED]"
                << std::endl;
      return false;
   }
   if (cnf.Value(alpha) != satisfiable)
   {
      std::cout << "Schoning SAT returned assignment: " << alpha << std::endl;
      std::cout << "Value of CNF on this assignment is " << cnf.Value(alpha)
                << " which is wrong! [FAILED]" << std::endl;
      return false;
   }
   std::cout << "Schoning SAT correctly determined the satisfiability status. "
                "[PASSED]"
             << std::endl;
   return true;
}

bool
SUBool::TestSchoning::TestSimpleCNF() const
{
   /*bool success=TestDimacsCNF(dimacs_simple, true);
   success=TestDimacsCNF(dimacs_horn, true)&&success;
   success=TestDimacsCNF(dimacs_renhorn, true)&&success;
   success=TestDimacsCNF(dimacs_nonrenhorn, true)&&success; */
   bool success = TestDimacsCNF(dimacs_unsat, false); //&&success;
   return success;
}

bool
SUBool::TestSchoning::TestRandomCNF(
    unsigned k, unsigned n_variables, unsigned size) const
{
   CNF cnf = random_k_cnf(size, k, n_variables);
   std::cout << "Testing a random " << k << "-CNF:" << std::endl;
   std::cout << cnf << std::endl;
   Assignment alpha(cnf.MaxVariable());
   bool satisfiable = Schoning::SAT(cnf, &alpha);
   std::cout << "Schoning SAT claims that the cnf ";
   std::cout << (satisfiable ? "is" : "is not") << " satisfiable" << std::endl;
   bool value = cnf.Value(alpha);
   std::cout << "Returned assignment alpha: " << alpha << std::endl;
   std::cout << "Value of CNF on this assignment is " << value;
   if (value != satisfiable)
   {
      std::cout << " which is wrong! [FAILED]" << std::endl;
      return false;
   }
   else
   {
      std::cout << " which is correct." << std::endl;
   }
   if (cnf.Value(alpha) != satisfiable)
   {
      std::cout << "Schoning SAT returned assignment: " << alpha << std::endl;
      std::cout << "Value of CNF on this assignment is " << cnf.Value(alpha)
                << " which is wrong! [FAILED]" << std::endl;
      return false;
   }
   std::cout << "Schoning SAT finished correctly. [PASSED]" << std::endl;

   return false;
}

bool
SUBool::TestSchoning::RunTests() const
{
   bool success = true;
   //  bool success=TestSimpleCNF();
   success = TestRandomCNF(3, 50, 225) && success;
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   unsigned seed = time(nullptr);
   // unsigned seed=1484403005; //time(nullptr);
   std::cout << "Seed: " << seed << std::endl;
   srand(seed);
   SUBool::TestSchoning test;
   return !test.RunTests();
}
