/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file test_smutex.cpp
 * Contains tests for SharedMutex and related classes.
 */
#include <iostream>
#include <thread>

#include "smutex.h"
#include "test_smutex.h"

namespace SUBool
{
   class ThreadSafeCounter
   {
    private:
      mutable SharedMutex mMutex;
      unsigned int mValue{0};

    public:
      ThreadSafeCounter() : mMutex() {}

      // Multiple threads/readers can read the counter's value at the same time.
      unsigned int
      get() const
      {
         SharedLockGuard lock(mMutex);
         return mValue;
      }

      // Only one thread/writer can increment/write the counter's value.
      void
      increment()
      {
         std::unique_lock<SharedMutex> lock(mMutex);
         mValue++;
      }

      // Only one thread/writer can reset/write the counter's value.
      void
      reset()
      {
         std::unique_lock<SharedMutex> lock(mMutex);
         mValue = 0;
      }
   };
} // namespace SUBool

std::mutex out_mutex;

bool
SUBool::TestSharedMutex::TestWithThreads() const
{
   ThreadSafeCounter counter;

   auto increment_and_print = [&counter]() {
      for (int i = 0; i < 30; i++)
      {
         counter.increment();
         std::lock_guard<std::mutex> lock(out_mutex);
         std::cout << std::this_thread::get_id() << ' ' << counter.get()
                   << '\n';
      }
   };

   std::thread thread1(increment_and_print);
   std::thread thread2(increment_and_print);

   thread1.join();
   thread2.join();

   return true;
}

bool
SUBool::TestSharedMutex::RunTests() const
{
   return TestWithThreads();
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::TestSharedMutex test;
   return !test.RunTests();
}
