/*
 * Project SUBool
 * Petr Kucera, 2016
 */
/**@file test_unitprop.h
 * Tests of UnitPropagator.
 */

#include <chrono>
#include <sstream>

#include "dimacs.h"
#include "test_unitprop.h"
#include "unitprop.h"

const char *dimacs_simple = "c\n"
                            "c start with comments\n"
                            "c\n"
                            "c\n"
                            "p cnf 5 3\n"
                            "1 -5 4 0\n"
                            "-1 5 3 4 0\n"
                            "-3 -4 0";

const char *dimacs_quinn = "c  quinn.cnf\n"
                           "c\n"
                           "p cnf 16 18\n"
                           "  1    2  0\n"
                           " -2   -4  0\n"
                           "  3    4  0\n"
                           " -4   -5  0\n"
                           "  5   -6  0\n"
                           "  6   -7  0\n"
                           "  6    7  0\n"
                           "  7  -16  0\n"
                           "  8   -9  0\n"
                           " -8  -14  0\n"
                           "  9   10  0\n"
                           "  9  -10  0\n"
                           "-10  -11  0\n"
                           " 10   12  0\n"
                           " 11   12  0\n"
                           " 13   14  0\n"
                           " 14  -15  0\n"
                           " 15   16  0";

const char *dimacs_horn = "c\n"
                          "c A horn formula\n"
                          "c\n"
                          "p cnf 7 8\n"
                          "2 0\n"
                          "-4 0\n"
                          "5 -7 -2 0\n"
                          "-1 3 -4 0\n"
                          "-5 -4 -2 1 0\n"
                          "3 -1 -2 0\n"
                          "5 -2 0\n"
                          "1 -2 -3 -4 0";

const char *amo_product_10 = "c AMO Product\n"
                             "p cnf 16 24\n"
                             "-2 11 0\n"
                             "-1 11 0\n"
                             "-1 -2 0\n"
                             "-3 12 0\n"
                             "-11 12 0\n"
                             "-3 -11 0\n"
                             "-4 13 0\n"
                             "-12 13 0\n"
                             "-4 -12 0\n"
                             "-5 14 0\n"
                             "-13 14 0\n"
                             "-5 -13 0\n"
                             "-6 15 0\n"
                             "-14 15 0\n"
                             "-6 -14 0\n"
                             "-7 16 0\n"
                             "-15 16 0\n"
                             "-7 -15 0\n"
                             "-9 -10 0\n"
                             "-8 -10 0\n"
                             "-8 -9 0\n"
                             "-10 -16 0\n"
                             "-9 -16 0\n"
                             "-8 -16 0\n";

const char *unitprop_cnf = "c\n"
                           "p cnf 6 5\n"
                           "1 -2 0\n"
                           "1 -3 0\n"
                           "2 3 -5 0\n"
                           "2 3 -4 0\n"
                           "1 2 3 4 5 6 0\n";

bool
SUBool::TestUnitProp::TestDimacsHorn() const
{
   std::cout << "Testing unit propagation on Horn CNF" << std::endl;
   auto cnf = Dimacs::parse_cnf(dimacs_horn);
   std::cout << "Input: " << std::endl;
   std::cout << dimacs_horn << std::endl;

   std::cout << "Trying the empty assignment" << std::endl;
   PartialAssignment alpha(cnf.MaxVariable());
   PartialAssignment beta(cnf.MaxVariable());

   UnitPropagator up(&cnf);
   up.Propagate(alpha, &beta);

   std::cout << "Consequences: " << beta << std::endl;

   return true;
}

bool
SUBool::TestUnitProp::TestAmoProduct10() const
{
   std::cout << "Testing unit propagation on AMO Product with n=10"
             << std::endl;
   auto cnf = Dimacs::parse_cnf(amo_product_10);
   std::cout << "Input: " << std::endl;
   std::cout << amo_product_10 << std::endl;

   UnitPropagator up(&cnf);

   std::cout << "Trying assignments setting one variable to 1" << std::endl;
   for (int i = 1; i <= 10; ++i)
   {
      PartialAssignment alpha(cnf.MaxVariable());
      PartialAssignment beta(cnf.MaxVariable());
      alpha[i] = PartialValue::TRUE;
      std::cout << "--" << std::endl;
      std::cout << "Result: " << up.Propagate(alpha, &beta) << std::endl;
      std::cout << "Assumptions : " << alpha << std::endl;
      std::cout << "Consequences: " << beta << std::endl;
      PartialAssignment gamma(cnf.MaxVariable(), PartialValue::FALSE);
      gamma[i] = PartialValue::TRUE;
      for (int j = 1; j <= 10; ++j)
      {
         if (beta[j] != gamma[j])
         {
            std::cout << "[FAILED]" << std::endl;
            return false;
         }
      }
      std::cout << "[PASSED]" << std::endl;
   }

   std::cout << "Trying assignments setting two variables to 1" << std::endl;
   for (int i = 1; i <= 10; ++i)
   {
      for (int j = 1; j <= 10; ++j)
      {
         if (j == i)
         {
            continue;
         }
         PartialAssignment alpha(cnf.MaxVariable());
         PartialAssignment beta(cnf.MaxVariable());
         alpha[i] = PartialValue::TRUE;
         alpha[j] = PartialValue::TRUE;
         std::cout << "--" << std::endl;
         auto res = up.Propagate(alpha, &beta);
         std::cout << "Result: " << res;
         if (res < 0)
         {
            std::cout << " [FAILED]" << std::endl;
            return false;
         }
         std::cout << " [PASSED]" << std::endl;
      }
   }

   return true;
}

bool
SUBool::TestUnitProp::TestUnitPropCNF() const
{
   std::cout << "Testing unit propagation on CNF" << std::endl;
   auto cnf = Dimacs::parse_cnf(unitprop_cnf);
   std::cout << "Input: " << std::endl;
   std::cout << unitprop_cnf << std::endl;

   std::cout << "Trying the assignment with assumptions -1" << std::endl;
   PartialAssignment alpha(cnf.MaxVariable());
   alpha[1] = PartialValue::FALSE;
   alpha[6] = PartialValue::FALSE;
   PartialAssignment beta(cnf.MaxVariable());

   UnitPropagator up(&cnf);
   std::cout << "Result: " << up.Propagate(alpha, &beta) << std::endl;
   std::cout << "Consequences: " << beta << std::endl;

   PartialAssignment gamma(cnf.MaxVariable(), PartialValue::FALSE);
   gamma[cnf.MaxVariable()] = PartialValue::TRUE;
   std::cout << "Gamma: " << gamma << std::endl;
   for (unsigned k = 1; k <= cnf.MaxVariable(); ++k)
   {
      if (beta[k] != gamma[k])
      {
         std::cout << beta[k] << "!=" << gamma[k] << " k: " << k << std::endl;
         std::cout << "[FAILED]" << std::endl;
         return false;
      }
   }
   std::cout << "[PASSED]" << std::endl;
   return true;
}

bool
SUBool::TestUnitProp::TestStdIn() const
{
   auto cnf = Dimacs::parse_cnf(std::cin);
   PartialAssignment alpha(cnf.MaxVariable());
   PartialAssignment beta(cnf.MaxVariable());

   alpha[1] = PartialValue::FALSE;

   auto start = std::chrono::system_clock::now();
   UnitPropagator up(&cnf);
   auto mid = std::chrono::system_clock::now();
   std::cout << "Result: " << up.Propagate(alpha, &beta) << std::endl;
   auto end = std::chrono::system_clock::now();
   std::chrono::duration<double> diff = end - start;
   std::chrono::duration<double> diff1 = mid - start;
   std::chrono::duration<double> diff2 = end - mid;
   std::cout << "Consequences: " << beta << std::endl;
   std::cout << "Time: " << diff.count() << " (" << diff1.count() << " + "
             << diff2.count() << ")" << std::endl;
   return true;
}

bool
SUBool::TestUnitProp::RunTests() const
{
   bool success = TestDimacsHorn();
   success = success && TestAmoProduct10();
   // success = success && TestUnitPropCNF();
   //   bool success=TestStdIn();
   return success;
}

int
main(int /*argc*/, char * /*argv*/[])
{
   srand(time(NULL));
   SUBool::TestUnitProp test;
   return !test.RunTests();
}
