/*
 * Project SUBool
 * Petr Kucera, 2019
 */
/**@file test_upenc.cpp
 * Contains tests for UPEncoder descendants, size estimates.
 */
#include <ctime>
#include <sstream>

#include "dimacs.h"
#include "drenc.h"
#include "logstream.h"
#include "random_utils.h"
#include "test_upenc.h"
#include "up_impl_level.h"
#include "up_log.h"
#include "up_min_lq.h"
#include "up_onelevel.h"
#include "up_quad.h"
#include "up_quad_impl.h"
#include "up_scc_bin_direct.h"
#include "up_scc_binary.h"
#include "up_scc_unary.h"

const std::string dimacs1 = "p cnf 20 48\n"
                            "5 0\n"
                            "-1 18 0\n"
                            "-1 -4 -6 15 -18 0\n"
                            "-4 -17 18 0\n"
                            "9 0\n"
                            "-1 6 -11 -15 -16 0\n"
                            "18 0\n"
                            "-5 7 -12 -15 -20 0\n"
                            "-2 10 -11 0\n"
                            "-3 -5 17 0\n"
                            "-1 -4 6 -17 -18 0\n"
                            "-4 -7 -16 19 0\n"
                            "-2 -4 7 -9 -10 0\n"
                            "-5 -14 17 0\n"
                            "-5 8 -10 -16 0\n"
                            "-6 12 0\n"
                            "17 0\n"
                            "19 0\n"
                            "14 0\n"
                            "3 0\n"
                            "12 -16 0\n"
                            "10 0\n"
                            "-3 10 -11 0\n"
                            "15 -18 0\n"
                            "-2 20 0\n"
                            "8 -18 0\n"
                            "-1 16 -19 0\n"
                            "-3 -4 12 -15 -19 0\n"
                            "12 -15 -17 0\n"
                            "9 -11 0\n"
                            "-7 13 0\n"
                            "-4 9 0\n"
                            "6 -10 0\n"
                            "-11 13 -16 -18 0\n"
                            "-2 -4 -6 -8 13 0\n"
                            "1 -10 -18 0\n"
                            "-9 -19 20 0\n"
                            "-3 -4 15 -20 0\n"
                            "-6 -11 17 -18 -20 0\n"
                            "4 0\n"
                            "-16 19 0\n"
                            "-1 -7 8 -10 -13 0\n"
                            "-4 -6 -8 14 0\n"
                            "-8 -12 13 0\n"
                            "-6 14 0\n"
                            "-2 -3 16 0\n"
                            "1 0\n"
                            "-3 9 0\n";

bool
SUBool::TestUPEncoder::TestSizeEstimateOfEncoderWithLevel(
    const UPEncoder &enc, unsigned max_level) const
{
   auto sizes = enc.EstimateSize(max_level);
   std::vector<unsigned> inputs;
   std::vector<unsigned> outputs;
   std::vector<Clause> clauses;
   enc.Encode(max_level, &inputs, &outputs, &clauses);
   CNF enc_cnf(clauses);
   std::cout << "... level: " << max_level << std::endl;
   std::cout << "... estimated max_variable: " << sizes.max_variable
             << ", real: " << enc_cnf.MaxVariable() << std::endl;
   std::cout << "... estimated size: " << sizes.size
             << ", real: " << enc_cnf.Size() << std::endl;
   std::cout << "... estimated length: " << sizes.len
             << ", real: " << enc_cnf.Length() << std::endl;
   if (sizes.max_variable != enc_cnf.MaxVariable())
   {
      std::cout << "Checking size estimator of an encoder (max variable "
                   "mismatch) [FAILED]"
                << std::endl;
      return false;
   }
   if (sizes.size != enc_cnf.Size())
   {
      std::cout
          << "Checking size estimator of an encoder (size mismatch) [FAILED]"
          << std::endl;
      return false;
   }
   if (sizes.len != enc_cnf.Length())
   {
      std::cout
          << "Checking size estimator of an encoder (length mismatch) [FAILED]"
          << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestUPEncoder::TestSizeEstimateOfEncoder(
    const UPEncoder &enc, VarStore &var_store) const
{
   bool success = true;
   unsigned level = 0;
   do
   {
      var_store = VarStore(false);
      success = success && TestSizeEstimateOfEncoderWithLevel(enc, level);
      if (!success)
      {
         std::cout << "[TestSizeEstimateOfEncoder] failed at level " << level
                   << std::endl;
         break;
      }
      if (level == 0)
      {
         level = 1;
      }
      else
      {
         level = enc.NextLevelBound(level);
      }
   } while (level > 0 && level < enc.MaxLevelBound());
   return success;
}

bool
SUBool::TestUPEncoder::TestSizeEstimateOneLevel(const CNF &cnf) const
{
   std::cout << "Testing OneLevel encoding" << std::endl;
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   UPEncOneLevel enc(cnf, cl2lit, var_store);
   return TestSizeEstimateOfEncoderWithLevel(enc, 0);
}

bool
SUBool::TestUPEncoder::TestSizeEstimateQuad(
    UPEncQuad::DIRECTION direction, const CNF &cnf) const
{
   std::cout << "Testing Quadratic encoding with direction="
             << static_cast<unsigned>(direction) << std::endl;
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   UPEncQuad enc(direction, cnf, cl2lit, var_store);
   if (!TestSizeEstimateOfEncoder(enc, var_store))
   {
      std::cout << "cnf:\n" << cnf << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestUPEncoder::TestSizeEstimateQuadImpl(const CNF &cnf) const
{
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   auto is = dual_rail(cnf, true);
   is.MakeBodyMinimal();
   is.MakeRightIrredundant();
   UPEncQuadImpl enc(cnf, cl2lit, var_store, is);
   return TestSizeEstimateOfEncoder(enc, var_store);
}

bool
SUBool::TestUPEncoder::TestSizeEstimateImplLevel(const CNF &cnf) const
{
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   auto is = dual_rail(cnf, true);
   is.MakeBodyMinimal();
   is.MakeRightIrredundant();
   UPEncImplLevel enc({true}, cnf, cl2lit, var_store, is);
   return TestSizeEstimateOfEncoder(enc, var_store);
}

bool
SUBool::TestUPEncoder::TestSizeEstimateLogarithmic(
    const CNF &cnf, const UPEncLogarithmic::Config &conf) const
{
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   UPEncLogarithmic enc(conf, cnf, cl2lit, var_store);
   if (!TestSizeEstimateOfEncoder(enc, var_store))
   {
      std::cout << "[TestSizeEstimateLogarithmic] failed with literal_levels="
                << conf.literal_levels << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestUPEncoder::TestSizeEstimateMinLQ(
    const CNF &cnf, bool literal_levels) const
{
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   UPEncLogarithmic::Config log_conf;
   log_conf.literal_levels = literal_levels;
   UPEncMinLQ enc(cnf, cl2lit, var_store, log_conf);
   if (!TestSizeEstimateOfEncoder(enc, var_store))
   {
      std::cout << "[TestSizeEstimateMinLQ] failed with literal_levels="
                << literal_levels << std::endl;
      return false;
   }
   return true;
}

bool
SUBool::TestUPEncoder::TestSizeEstimateSCCUnary(
    const CNF &cnf, LIT_IMPL_SYSTEM is_type) const
{
   std::cout << "Testing SCCUnary encoding" << std::endl;
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   auto is = init_lit_impl_system(cnf, is_type);
   LitISGraph scc_graph(is);
   // std::cout << "cnf:\n" << cnf << std::endl;
   // std::cout << "impl system:\n" << is << std::endl;
   // scc_graph.DumpImplSystem(3, is);
   UPSCCUnary enc(cnf, cl2lit, var_store, is, scc_graph);
   return TestSizeEstimateOfEncoder(enc, var_store);
}

bool
SUBool::TestUPEncoder::TestSizeEstimateSCCBinary(
    const CNF &cnf, bool use_bit_level, LIT_IMPL_SYSTEM is_type) const
{
   std::cout << "Testing SCCBinary encoding with use_bit_level="
             << bool_to_string(use_bit_level) << std::endl;
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   auto is = init_lit_impl_system(cnf, is_type);
   LitISGraph scc_graph(is);
   // std::cout << "cnf:\n" << cnf << std::endl;
   // std::cout << "impl system:\n" << is << std::endl;
   // scc_graph.DumpImplSystem(3, is);
   UPSCCBinary::Config conf{use_bit_level};
   UPSCCBinary enc(conf, cnf, cl2lit, var_store, is, scc_graph);
   return TestSizeEstimateOfEncoder(enc, var_store);
}

bool
SUBool::TestUPEncoder::TestSizeEstimateSCCBinaryDirect(
    const CNF &cnf, bool use_bit_level, LIT_IMPL_SYSTEM is_type) const
{
   std::cout << "Testing SCCBinaryDirect encoding with use_bit_level="
             << bool_to_string(use_bit_level) << std::endl;
   ClausesToLiterals cl2lit(cnf);
   VarStore var_store(false);
   auto is = init_lit_impl_system(cnf, is_type);
   LitISGraph scc_graph(is);
   UPSCCBinaryDirect::Config conf{use_bit_level};
   UPSCCBinaryDirect enc(conf, cnf, cl2lit, var_store, is, scc_graph);
   if (!TestSizeEstimateOfEncoder(enc, var_store))
   {
      std::cout << "cnf:\n" << cnf << std::endl;
      std::cout << "impl system:\n" << is << std::endl;
      scc_graph.DumpImplSystem(3, is);
      return false;
   }
   return true;
}

bool
SUBool::TestUPEncoder::TestSizeEstimateForCNF(const CNF &cnf) const
{
   bool success = true;
   success = success && TestSizeEstimateOneLevel(cnf);
   success = success && TestSizeEstimateQuad(UPEncQuad::DIRECTION::IF, cnf);
   success =
       success && TestSizeEstimateQuad(UPEncQuad::DIRECTION::ONLY_IF, cnf);
   success = success && TestSizeEstimateQuad(UPEncQuad::DIRECTION::BOTH, cnf);
   success = success
             && TestSizeEstimateLogarithmic(
                 cnf, UPEncLogarithmic::Config(
                          /* literal_levels */ false,
                          /* split_size */ 4,
                          /* use_bit_level */ true,
                          /* add_amo_on_fire_vars */ true,
                          /* add_compare_exclusivity */ true,
                          /* add_force_linear_order_on_max_level */ true,
                          /* contradiction_has_max_level */ true));
   // UPEncLogarithmic::Config(false, 4, true, false, true, true))
   success = success
             && TestSizeEstimateLogarithmic(
                 cnf, UPEncLogarithmic::Config(
                          /* literal_levels */ true,
                          /* split_size */ 4,
                          /* use_bit_level */ true,
                          /* add_amo_on_fire_vars */ true,
                          /* add_compare_exclusivity */ true,
                          /* add_force_linear_order_on_max_level */ true,
                          /* contradiction_has_max_level */ true));
   success = success
             && TestSizeEstimateLogarithmic(
                 cnf, UPEncLogarithmic::Config(
                          /* literal_levels */ false,
                          /* split_size */ 4,
                          /* use_bit_level */ false,
                          /* add_amo_on_fire_vars */ true,
                          /* add_compare_exclusivity */ true,
                          /* add_force_linear_order_on_max_level */ true,
                          /* contradiction_has_max_level */ true));
   success = success
             && TestSizeEstimateLogarithmic(
                 cnf, UPEncLogarithmic::Config(
                          /* literal_levels */ true,
                          /* split_size */ 4,
                          /* use_bit_level */ false,
                          /* add_amo_on_fire_vars */ true,
                          /* add_compare_exclusivity */ true,
                          /* add_force_linear_order_on_max_level */ true,
                          /* contradiction_has_max_level */ true));
   success = success && TestSizeEstimateMinLQ(cnf, true);
   success = success && TestSizeEstimateMinLQ(cnf, false);
   success = success && TestSizeEstimateQuadImpl(cnf);
   success = success && TestSizeEstimateImplLevel(cnf);
   success =
       success && TestSizeEstimateSCCUnary(cnf, LIT_IMPL_SYSTEM::BODY_MINIMAL);
   success = success
             && TestSizeEstimateSCCBinary(
                 cnf, false, LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED);
   success = success
             && TestSizeEstimateSCCBinary(
                 cnf, true, LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED);
   success = success
             && TestSizeEstimateSCCBinaryDirect(
                 cnf, false, LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED);
   success = success
             && TestSizeEstimateSCCBinaryDirect(
                 cnf, true, LIT_IMPL_SYSTEM::BODY_MINIMAL_IRRED);
   return success;
}

bool
SUBool::TestUPEncoder::TestSizeEstimate(
    unsigned n_rounds, unsigned n_vars, unsigned n_clauses) const
{
   std::cout
       << "Checking UPEncoder descendants size estimate on random formulas"
       << std::endl;
   bool success = true;
   for (unsigned i = 0; success && i < n_rounds; ++i)
   {
      auto cnf = random_horn_cnf(n_clauses, 5, n_vars, true);
      // auto cnf = random_cnf(n_clauses, n_vars);
      cnf.RemoveDuplicities();
      success = TestSizeEstimateForCNF(cnf);
   }
   std::cout
       << "Checked UPEncoder descendants size estimates on random formulas "
       << (success ? "[OK]" : "[FAILED]") << std::endl;
   return success;
}

bool
SUBool::TestUPEncoder::RunTests() const
{
   return TestSizeEstimateForCNF(Dimacs::parse_cnf(dimacs1))
          && TestSizeEstimate(10, 20, 50);
}

int
main(int /*argc*/, char * /*argv*/[])
{
   SUBool::logs.mVerbLevel = 4;
   srand(time(NULL));
   return !SUBool::TestUPEncoder().RunTests();
}
